#ifndef _PW_CIR_
#define _PW_CIR_

#include "linear.h"
#include "DFfile.h"
#include "DFalloc.h"
#include "memuser.h"
#include "PWilib.h"
#include "logicOpt.h"

/* program function */
#define LOGIC_SIM_1VEC  				1
#define ATPG_BASED_ESTIMATION_GAIN_C 	2
#define MC_LOGIC_SIM_1VEC				3
#define SW_RANDOM_SIM					4
#define NETLIST_READ					5
#define NETLIST_RELEASE					6
#define ATPG_BASED_ESTIMATION_GAIN      7
#define ATPG_BASED_ESTIMATION_FANOUT	8
#define EXPORT_SPICE					9
#define INTEL_NETLIST_READ				10
#define EXHAUSTIVE_SPICE_SIM			11
#define ILIB_READ						12
#define ATPG_BASED_ESTIMATION_WITH_ILIB 13

/* rand input vec generator mode */
#define RAND_INPUT_VEC_ALL_PI   1
#define RAND_INPUT_VEC_REAL_PI	2

/* for debugging */
#define DEBUG_INFO			0
#define TIME_REPORT			1

#define DEBUG_OUT_FILE		"dbgInfo"
#define RESULT_FILE_SUF		".result"
#define SPICE_FILE_SUF		".sp"

/* Circuits */
#define NUM_LIB_CELL		25			/* # of gate types that are supported */
										/* to be changed when new types added */ 
#define MAX_PI				3000 		/* max # of PI, including PSEUDO PI */
#define MAX_PO				3000		/* max # of PO, including PSEUDO PO */
#define MAX_FAN_OUT			2000		/* max # of fanout for one gate */
#define MAX_IN_PER_GATE 	20			/* max # of fanin for one gate */
#define	MAX_ABS_GAIN		28800		/* max absolute value of Gain */
#define INVALID_GAIN		-99999		/* value for invalid gain */

/* For MFFC */
#define MFFC_INITIAL		-1
#define MFFC_COLORED		 1
#define MFFC_DISCARDED		-2

#define MAX_NUM_MFFC		6000

#define INFINITY_TIME		9999E10 	/*assume the sim time not exceed it*/
#define NO_MORE_EXCITATION  -1

#define PW_INST 				struct pw_inst
#define PW_LIB  				struct pw_lib
#define UNCERT_LIST_HEAD    	struct uncert_list_head
#define UNCERT_ITEM				struct uncert_item
#define PW_DOUBLE_LIST			struct pw_double_list
#define PW_INT_DUAL_LIST		struct pw_int_dual_list
#define PW_VECTOR_PAIR			struct pw_vector_pair
#define TRANS_STACK				struct trans_stack


/***************************************************************
	NOTE: THE LIB NO. IS THE ACCESS NO. OF THE POWER LIB,
		  SO CHANGES OF THEM LEAD TO MODIFYING LIBRARY DATA!!!
		  (the inst name in lib is in *PWdeviceName[], index starting from 0)
	NOTE: if you add devices, change the "NUM_LIB_CELL"!!!
*****************************************************************/
#define  NOT_LIB_NO  		1
#define  AND_LIB_NO			2
#define  NAND_LIB_NO		3
#define	 OR_LIB_NO			4
#define	 NOR_LIB_NO			5
#define  XOR_LIB_NO   		6
#define  XNOR_LIB_NO   		7
#define  MUX21_LIB_NO  		8
#define  AOI21_LIB_NO  		9
#define  AOI211_LIB_NO  	10
#define  AOI221_LIB_NO  	11
#define  AOI31_LIB_NO   	12
#define  AOI311_LIB_NO   	13
#define  OAI21_LIB_NO   	14
#define  OAI211_LIB_NO  	15
#define  OAI221_LIB_NO  	16
#define  OAI31_LIB_NO   	17
#define  OAI311_LIB_NO   	18
#define  PI_LIB_NO      	19
#define  PO_LIB_NO      	20

#define  DF_LIB_NO			21
#define  DFR_LIB_NO			22
#define  DFS_LIB_NO			23
#define  DFRS_LIB_NO		24

#define  BUFF_LIB_NO		25

PW_LIB
{									/*char name[MAX_STR_BUF];*/
	long		ld_cap;				/* num of load_cap*/
	double		dur;				/* duration,unit =ns*/
	double		peak;				/* peak current,unit= ma*/
	double		peakT;				/* the time from start to peak */
	double		delay;				/* half-amp to half-amp gate delay*/
};

/* bucket element */
struct pw_bucket_elem
{
	int			instID;				/* instance/gate ID */
	LOGIC_VALUE	assignment;			/* logic value assigned */
};

typedef struct pw_bucket_elem PW_BUCKET_ELEM;

/* bucket for bucket sorting */
struct pw_bucket
{
	bi_lin_table_rec	bucketElemList;	/* the head of the bucket list */
};

typedef struct pw_bucket PW_BUCKET;

/* gate and logic value pair in implication process */
struct pw_imply_stru
{
	PW_INST			*inst;			/* instance */
	LOGIC_VALUE		value;			/* logic value */
};
typedef struct pw_imply_stru PW_IMPLY_STRU;

/* instance/gate */
PW_INST
{	PW_INST 		*in[MAX_IN_PER_GATE];/* fanin links */
	PW_INST 		*out[MAX_FAN_OUT];	 /* fanout links */
	PW_INST			*outz[MAX_FAN_OUT];	/* flip-flop QZ fanout links, */
										/* not used in current version */
	int				zflag;	/* for pseudo PI only, if QZ links consedered */
							/* NOT used in current version */
						    /* showing whether fanin is out or outz */

	char			name[MAX_STR_BUF];	/* instance name */
	int				instID;				/* instance ID # in instance list */
	int				instLibNo;			/* the instance type # in library */
										/* -1 for unidentified inst type */

	PW_INST 		*next; 				/* for ckt levelization list */
	int				maxLev,minLev;		/* max/min level */
	int				fanin,fanout;		/* # of fanin/fanout */
	BOOLEAN 		visited;			/* flag used in levelization */

	gate_current_t  *powerup_current;   /* for current at a specified time*/
	double			instPower;			/* power dissipation of the instance */
										/* obtained from power lib, */
										/* currently not used */

	bi_lin_table_rec InSortedList[2];	/* sorted input pin list */
										/* for backtracing */

	int				gain_st[2];			/* gain for logic value VALUE_0 & 1*/
	int				new_gain_st[2];		/* temporary new gain */
	bi_lin_table_rec pGainUpdt_st[2];	/* ptr for the rec in one gain bucket */
										/* for updating gate gain and bucket */

	stk_rec 		implyStkTop;		/* stack for implied values */
	LOGIC_VALUE		state;				/* logic state */

	int				MFFC_clustering_flag;	/* for MFFC clustering */
	int				MFFC_No;				/* No. of MFFC */
	BOOLEAN			MFFC_root;				/* flag of MFFC root gate */

	pointer			sis_node;				/* direct mem-data transferring from sis */
};

#endif 
