#ifndef _ilib_powerup_current_waveform_
#define _ilib_powerup_current_waveform_

#include "sis.h"


/* the number of fileds for the required information 
 * at the beginning of the ilib file for powerup current 
 */
#define REQUIRED_INFO_FIELD_NUM  	4

/* # of generic type (without fanin#) that could be in the ilib */
#define GENERIC_TYPE_NUM_IN_ILIB  	7

/* # of complex gate types that needn't to exist in the lib file */
#define COMPLEX_GATE_TYPE_NUM		3

/* # of time points in the time range of waveform to be investigated */
#define TIME_POINT_NUM				10

/* index for the gate BUFF in the library */
/* update it when you change the ilib */
#define BUFF_ILIB_INDEX				6

/* the table index is computed by concatenating type_index, nin, and nout 
	type_index: for the generic gate type, starting from 0
	nin:		the # of inputs of the specific gate, starting from 1
	nout:		the # of fanouts of the specific gate, starting from 1
	nin_bits:	the field width for nin
	nout_bits:	the field width for nout
 */
#define lookup_table_index(type_index, nin, nin_bits , nout, nout_bits) \
		((((unsigned) (type_index)) << ((nin_bits)+(nout_bits))) + \
		(((unsigned) (nin-1)) << (nout_bits)) + (unsigned) (nout-1))

/* list type */
typedef enum pair_list_type pair_list_type_t;
enum pair_list_type {
	STABLE0_LIST,		/* 0 */
	RISE_LIST			/* 1 */
};


/* current & time pair */
typedef struct current_time_pair ilib_pair_t ;
struct
current_time_pair
{
	float current; /* unit: uA */
	float time;    /* unit: ps */
};

/* pair list head */
typedef struct ilib_pair_list_head ilib_pair_list_head_t ;
struct
ilib_pair_list_head
{
	float 	turnonTime;				/* for sleep transistor */
	float	leftmostPeakTime[2];	/* used to determine the time range */
	float	rightmostPeakTime[2];
	lsList	pairList[2];			/* two lists of current-time pairs: rise & stable0 */
};

/* one entry of the lookup table corresponds to one particular gate*/
typedef struct ilib_entry ilib_entry_t ;
struct
ilib_entry
{
	lsList	headList;	/* the list of pair-list-head */
};

typedef struct ilib_index_rule ilib_index_rule_t ;
struct
ilib_index_rule
{
	int		fanin_field_bits;
	int		fanout_field_bits;
};

typedef struct ilib_table ilib_table_t ;
struct
ilib_table
{
	char **generic_gate_types; /* gate types in the library */
	int	table_size; 		/* total # of allocated entries */
	int	valid_entry_num;	/* each particular gate (with specified fanin # and fanout #) per entry */

	int max_fanin_num;		/* the max fanin # supported in the lib */
	int max_fanout_num; 	/* the max fanout # supported in the lib */

	lsList turnon_time_list;/* the list recording all the different turn-on time */

	ilib_index_rule_t *rule;	/* the access rule */

	array_t *body;			/* table body */
};

/* the value of power-up current at a specified time point */
typedef  pair_list_type_t current_type_t;
typedef  struct gate_current_value gate_current_t; /* used in atpg_ilib alg */
struct
gate_current_value
{
    /* the current of both types at a specified time */
    float           current[2];
    /* the preferred type of current: the larger one */
    current_type_t  preferred_type;
};


extern int PWilibRead(FILE* ilibFp, FILE* dbgFp);


#endif
