#ifndef _ilib_powerup_current_routines_
#define _ilib_powerup_current_routines_

#include "sis.h"
#include "PWcir.h"

extern int PWilibRead(FILE* ilibFp, FILE* dbgFp);

/* get the pairlist with specified turn-on time
 *
 * NOTE: list_type currently belongs to [0,1]
 */
extern lsList ilibPairList(ilib_table_t *table, int index, float turnon_time, int list_type);

/* get the entry from the ilib table, corresponding to one particular gate */
extern ilib_entry_t * ilibEntry(ilib_table_t *table, int index);

/* get the entry from ilib table, explicitly using the three indexing fields */
extern ilib_entry_t * ilibEntry3(ilib_table_t *table, int type_index, int nin, int nout);

/* get the headList for an entry */
extern lsList ilibEntryHeadList(ilib_entry_t *entry);

/* convert lookup table index to gate information */
extern int index2gate(ilib_table_t *table, unsigned int index, char *type_name, int *nin, int *nout);

/* the index corresponding to the generic gate type */
extern int generic_type_index(char *str, char **generic_type_index_table);

/* allocate & initialize the lookup table */
extern void alloc_init_ilib_table(ilib_table_t **table,  int entry_num, int fanin_num, int fanin_field_bits, int fanout_num, int fanout_field_bits, int flag);

/* get the minimum binary field that can represents the number */
extern int fit_radix2_field(int num, int *bits);

/* dump ilib to file */
extern void dump_ilib(ilib_table_t *table, FILE* dbgFp);

/* get the sum of two current waveform */
extern int sum_current_waveform( lsList new_pairlist, lsList src1_pairlist, lsList src2_pairlist );




/*Start a in-order traversal of two ordered list just as one merged ordered list
 *The order is determined the key in each element of the lists 
 */
extern void lsTraverseInorder2_start(lsList list1, lsList list2, lsGen *gen1, lsGen *gen2);

/* Get the data contained in the current element of the list, 
 * and move to the next element 
 * NOTE: the data field of the element in the list MUST be a pointer!
 */
extern lsStatus
lsTraverseInorder2_next(lsGen gen1, lsGen gen2, lsGeneric* data, lsGeneric (*data_access)(), int (*data_compare)());

/* access the current element specified by the generator */
extern lsStatus
lsCurrent(lsGen generator, lsGeneric *data, lsHandle *itemHandle);

/* get the value from piece-wise linear function */
extern float piece_linear_point(float y1,float x1,float y2,float x2,float x);

/* finish the traversal */
extern void lsTraverseInorder2_finish(lsGen gen1, lsGen gen2);

/* get the value of current from a pairlist at the specified time point */
extern float get_current(lsList pairlist, float time);

/* append a new pair into the pairlist */
extern void pairNewEnd(lsList pairlist, float current, float time);

/* get the time for the left-most & right-most local peak current */
extern void left_right_PeakTime(ilib_table_t *table);

/* get the time range and step for timed ATPG */
/* currently, we investigate all the gates in the library */
/* later, it can be modified to look at only the gates in the circuit */
extern void time_range_step(ilib_table_t *table, float turnon_time,
							PW_INST *instList, 
							float *t_start, float *t_end, float *t_step);

extern int accessCurrentILIB(ilib_table_t *table, PW_INST *inst, float time,
							 float turnon_time);

/* report the result for atpg_ilib */
/* not only combinational block supported in atpg_ilib */
extern float reportMaxCurrent(float time, BOOLEAN write_file_flag);

/* obtain the current at specified time for each gate in the sorting array
 *
 * the corresponding field in the instance will be updated 
 */
extern int currentAtTime(array_t *gate_sort, ilib_table_t *PWilib_table, float time, float turnon_time);

/* Sorting gate's input pins for backtracing 
 * accroding to the power-up current in the library.
 */
extern void PWinputSort_ilib(void);
extern void PWinputSortOneGate_ilib(PW_INST *q);

/* sorted list for inputs and different assignment type*/
/* input index of the gate is stored in the list */
extern void inputListSort_ilib(double InWeight[][2], PW_INST *inst, int inputNum, int assignmentInteger);

#endif
