#ifndef _LEAKAGE_CURRENT_EST_GA_ALGORITHM_
#define _LEAKAGE_CURRENT_EST_GA_ALGORITHM_

#define SIS
#include "sis.h"

#include "ga_alg.h"
#include "logicOpt.h"

/* Genetic Algorithm for leakage current estimation.
 * NOTE: performed at the imaxup data structure.
 */
extern int lkg_ga_estimation(int in_max_gen, int min_max_flag);

/* report estimation results */
extern void GA_report_result_vector(chromosome_t *solution, int min_max_flag, FILE *fp);

/* randomly generate the initial population of vectors for leakage current estimation
 * one bit encoding for one primary input or pseudo primary input */
extern array_t * init_pop_vector(void);

/* randomly set the gene corresponding to the input bit */
extern void random_set_input_bit(chromosome_t *chromo_p, int position);

/* selection scheme */
extern array_t * select_vector1(array_t *pop);

/* tournament selection without replacement */
extern array_t * tournaSel(array_t *pop);

/* assuming only two chromo are visited previously */
extern void remove_visited_chromo(array_t **pop_left, int chromo1, int chromo2);

/* the replacement scheme in leakage current estimation.
 * old_pop will be freed, and completley replaced by the new_pop.*/
extern void replace_vector(array_t **old_pop, array_t *new_pop);

/* the crossover scheme for leakage current estimation.
 * one-point crossover with probability of 1.  */
extern void crossover_vector(array_t *pop);

/* the one-point crossover schem */
extern void one_point_crossover(chromosome_t *p1, chromosome_t* p2);

/* swap the part from index '0' to 'offset_in_word' */
extern void partial_swap(int *word1, int *word2, int offset_in_word);

/* mutation scheme */
extern void mutation_vector(array_t *pop);

/* finish proc for leakage current estimation */
extern BOOLEAN finish_vector(int gen_num, int max_gen, int no_improve_num);

/* get the leakage current for the chromo as its fitness */
extern double fitness_lkg_current(chromosome_t * chr);

/* get the reciprocal of leakage current for the chromo as its fitness */
extern double fitness_lkg_current_reciprocal(chromosome_t * chr);

/* decoding the chromo and convert to an input vector
 * HACK: assuming it is a binary encoding, one input as one bit */
extern void decode_chromo_to_vector(chromosome_t * chr);

/* find the next input which is not VDD/GND, and assign the logic value */
extern void assign_next_input_bit(int * PI_index, LOGIC_VALUE val);

/* get the leakage current for each instance and the total leakage current.
 * Assume one simulation has been performed.  */
extern double compute_lkg_current(void);

/* get pi & ppi number, excluding GND/VDD */
extern int controllable_pi(void);


extern void decode_dump_pop_vector(array_t *pop, FILE* fp);
extern void decode_dump_chromo_vector(chromosome_t *chr, FILE * fp);



#endif
