/***************************************************************************
			FILE:       linear.h
			DESCR:      linear table structure and operation funcs (user)
			AUTH:       Fei Li
			DATE:       April 1999
***************************************************************************/

#ifndef _LINEAR_TABLE_OP
#define	_LINEAR_TABLE_OP

#include "memuser.h"
#include "DFtype.h"

#define	NOT_CARE_KEY	0

/* general uni-directional linear table record item */
struct lin_table_rec_
{	
	double  key;
	pointer member;
	struct 	lin_table_rec_ *next;
};

typedef struct lin_table_rec_ *lin_table_rec;

/* general bi-directional linear table record item */
struct bi_lin_table_rec_
{	
	double	key;
	pointer	member;
	struct	bi_lin_table_rec_ *prev;
	struct	bi_lin_table_rec_ *next;
};

typedef struct bi_lin_table_rec_ *bi_lin_table_rec;

/********************************
* This feature is 
* Temporarily disabled!!
********************************/
/********************************
#if defined(__STDC__)
#define ARGS(args) args
#else
#define ARGS(args) ()
#endif
*********************************/

/*>>> user-visible functions */
/**--------------------------------------------------------------**/
extern	lin_table_rec	lin_table_prev_ins	(lin_table_rec *,lin_table_rec *, double, pointer);
			/*ARGS((lin_table_rec *,lin_table_rec *, double, pointer));*/
			/*    HEAD,				TAIL,			KEY, 	MBR  */

extern	lin_table_rec 	lin_table_prev_del	(lin_table_rec *,lin_table_rec *);
			/*ARGS((lin_table_rec *,lin_table_rec *));*/
			/*		HEAD,			TAIL	*/

extern	lin_table_rec	lin_table_post_ins	(lin_table_rec *,lin_table_rec *, double, pointer);
			/*ARGS((lin_table_rec *,lin_table_rec *, double, pointer));*/
			/*		HEAD,			TAIL,			KEY,	MBR	*/

extern	lin_table_rec	lin_table_look_up	(lin_table_rec, double, lin_table_rec *);
			/*ARGS((lin_table_rec, double, lin_table_rec *));*/
			/*		HEAD,		  KEY,		PREMBR		*/
			/*	and matched MBR is returned */

extern	lin_table_rec	lin_table_arb_del	(lin_table_rec *, lin_table_rec *, lin_table_rec);
			/*ARGS((lin_table_rec *, lin_table_rec *, lin_table_rec));*/
			/*		HEAD,			TAIL,			PREMBR		*/
			/*	and the deleted MBR is returned , NOTE:deleted not freed*/

extern	bi_lin_table_rec bi_lin_table_prev_ins	(bi_lin_table_rec *, bi_lin_table_rec *, double, pointer);
			/*ARGS((bi_lin_table_rec *, bi_lin_table_rec *, double, pointer));*/
			/*		HEAD,					TAIL,			KEY,	MBR */
extern	bi_lin_table_rec bi_lin_table_prev_del	(bi_lin_table_rec *, bi_lin_table_rec *);
			/*ARGS((bi_lin_table_rec *, bi_lin_table_rec *));*/
			/*		HEAD,					TAIL		*/
extern	bi_lin_table_rec bi_lin_table_post_ins	(bi_lin_table_rec *, bi_lin_table_rec *, double, pointer);
			/*ARGS((bi_lin_table_rec *, bi_lin_table_rec *, double, pointer));*/
			/*			HEAD,				TAIL,			 KEY,	  MBR	*/
extern  bi_lin_table_rec bi_lin_table_post_del	(bi_lin_table_rec *, bi_lin_table_rec *);
			/*ARGS((bi_lin_table_rec *, bi_lin_table_rec *));*/
			/*			HEAD,				TAIL		*/
extern	bi_lin_table_rec bi_lin_table_look_up	(bi_lin_table_rec , double);
			/*ARGS((bi_lin_table_rec , double));*/
			/*			HEAD,		 dstKEY   */
			/* and matched member is returned */
extern	bi_lin_table_rec bi_lin_table_arb_del	(bi_lin_table_rec *, bi_lin_table_rec *, bi_lin_table_rec);
			/*ARGS((bi_lin_table_rec *, bi_lin_table_rec *, bi_lin_table_rec));*/
			/*			HEAD,				TAIL,				delMBR	*/
			/* and the deleted MBR is returned, NOTE:deleted but not freed */

/**--------------------------------------------------------------**/

/*==================================================================
	MACRO LIST:
		STK_PUSH(lin_table_rec *TOP_PP, double KEY, pointer MBR)
		STK_POP(lin_table_rec *TOP_PP)
		QUEUE_PUSH(lin_table_rec *HEAD_PP, lin_table_rec *TAIL_PP, double KEY, pointer MBR)
		QUEUE_POP(lin_table_rec *HEAD_PP, lin_table_rec *TAIL_PP)
		
		BI_STK_PUSH(bi_lin_table_rec *TOP_PP, double KEY, pointer MBR)
		BI_STK_POP(bi_lin_table_rec *TOP_PP)
		BI_QUEUE_PUSH(bi_lin_table_rec *HEAD_PP, bi_table_rec *TAIL_PP, double KEY, pointer MBR)
		BI_QUEUE_POP(bi_lin_table_rec *HEAD_PP, bi_lin_table_rec *TAIL_PP)

====================================================================*/

/* renamed type for specific situation */
typedef lin_table_rec stk_rec;
typedef lin_table_rec queue_rec;

typedef bi_lin_table_rec bi_stk_rec;
typedef bi_lin_table_rec bi_queue_rec;

/* stack/queue operation */
#define STK_PUSH(TOP_PP, KEY, MBR) \
		lin_table_prev_ins(TOP_PP, NULL, KEY, MBR)

#define STK_POP(TOP_PP) lin_table_prev_del(TOP_PP, NULL)

#define QUEUE_PUSH(HEAD_PP, TAIL_PP, KEY, MBR) \
		lin_table_post_ins(HEAD_PP, TAIL_PP, KEY, MBR)
#define QUEUE_POP(HEAD_PP, TAIL_PP) \
		lin_table_prev_del(HEAD_PP, TAIL_PP)

#define BI_STK_PUSH(TOP_PP, KEY, MBR) \
		bi_lin_table_prev_ins(TOP_PP, NULL, KEY, MBR)

#define BI_STK_POP(TOP_PP) bi_lin_table_prev_del(TOP_PP,NULL)

#define BI_QUEUE_PUSH(HEAD_PP, TAIL_PP, KEY, MBR) \
		bi_lin_table_post_ins(HEAD_PP, TAIL_PP, KEY, MBR)

#define BI_QUEUE_POP(HEAD_PP, TAIL_PP) \
		bi_lin_table_prev_del(HEAD_PP, TAIL_PP)

#undef ARGS

#endif
