#include <stdio.h>
#include <string.h>
#include "DFtype.h"
#include "DFstring.h"
#include "PWcir.h"
#include "PWroutines.h"
#include "PWglobal_var.h"
#include "PWreadIntel.h"
#include "hashtable.h"

extern void IntelPreProcessInfo(FILE *fin, long *PWinstNum, long *DF_Num);
extern void lastStrInLine(FILE* fin, char* str);
extern PW_INST * isRealPIO(PW_INST* SourceP[], long inNum,
		  					PW_INST* SinkP[],   long outNum, char str[]);
extern void IntelCktConnect(FILE* fin, lin_table_rec *NetHashTable, int hashsize);
extern void IntelConnectInput(char net_name[], lin_table_rec * net_hashtable, int size, PW_INST* currInst);
extern void linkOperation(PW_INST* inst, PW_INST* inInst);
extern PW_INST	* IntelFaninInst(char net_name[], lin_table_rec* net_hashtable, int size, PW_INST* currInst);
extern PW_INST	* pseudoFaninGate(PW_INST* inInst, PW_INST* currInst);
extern void IntelConnectOutput(char out_net_name[], lin_table_rec * net_hashtable, int size, PW_INST* currInst);
extern BOOLEAN      isOutput2PO();

void
IntelNetRead(FILE* fin)
{
	char	str[MAX_STR_BUF],tmpStr[MAX_STR_BUF],ch;
	int		pi_index,po_index,instance_index;
	int		statementFg;
	int		i,j,t;

	int				hashTableSize=100;
	lin_table_rec	*NetHashTable;

	unknownCellList=NULL;
	DFlipList=NULL;
	DF_Num=0;
	PWlineNum=PWinstNum=0;
	PSEUDO_PI_index=PSEUDO_PO_index=0;

	PrimInNum=0;
	PrimOutNum=0;
	vdd_index=gnd_index=-1;
	pi_index=po_index=instance_index=0;

	pseudoCount=0;

	/* initialize the hash table for net name */
	hashTableInit(&NetHashTable, hashTableSize);

	fprintf(stdout,"Reading...\n");

	IntelPreProcessInfo(fin,&PWinstNum,&DF_Num);
	PWlineNum = PWinstNum; /* backward compatible */
	/*fprintf(stdout, "%ld %ld", PWinstNum, DF_Num);*/

	/* allocation: # of instances + # of possible pseudo PI/PO */
    /* assume only output Q of flip-flop is supported */
    /* Note: PI inputs such as VCC/VSS, if any, are created separately, not within 'instList' */
    instList=(PW_INST *)mem_get_block((SIZE_T) (PWinstNum+2*DF_Num)*sizeof(PW_INST));

	/* initialization */
	initPWinst(instList);

	/* reading instances information */
	rewind(fin);
	do{
		if(fscanf(fin,"%s",str)<= 0)
			break;
		strupper(str);

		if(strcmp(str, "$$.INPUT")==0 ||
			(strcmp(str, "$$+")==0 && statementFg==PI_STATEMENT) )
		{
			statementFg=PI_STATEMENT;
			ch=fgetc(fin);
			ungetc(ch,fin);
			while(ch!='\n')
			{
				/*ungetc(ch,fin);*/
				do{
					ch=fgetc(fin);
				  }while(ch=='\r' || ch == '\t' || ch == ' ');

				  ungetc(ch,fin);
				  if(ch != '\n')
				  {
				  	fscanf(fin,"%s",str);
				  	strupper(str);
				  	if(instance_index >= PWinstNum || pi_index >= PrimInNum)
				  	{
				  		fprintf(stderr,"Error in instance/PI number calculation.\n");
				  		exit(1);
					}else{
							strcpy(instList[instance_index].name,str);
				  			instList[instance_index].instLibNo=PI_LIB_NO;
				  			instList[instance_index].maxLev=0; /* for PIs */
				  			instList[instance_index].minLev=0;
				  			SourceP[pi_index++]=&instList[instance_index];
				  			++instance_index;
						 }
				  	ch = fgetc(fin);
				  	ungetc(ch,fin);
				  }
	  		 }

		}else if(strcmp(str, "$$.OUTPUT")==0 ||
				(strcmp(str, "$$+")==0 && statementFg==PO_STATEMENT) )
			  {
			  	statementFg=PO_STATEMENT;

			  	ch=fgetc(fin);
			  	ungetc(ch,fin);
			  	while(ch!='\n')
			  	{
			  		/*ungetc(ch,fin);*/
			  		do{
			  			ch=fgetc(fin);
					  }while( ch=='\r' || ch == '\t' || ch == ' ');

					  ungetc(ch, fin);
					  if(ch != '\n')
					  {
					  	fscanf(fin,"%s",str);
					  	strupper(str);
					  	if(instance_index >= PWinstNum || po_index >= PrimInNum)
					  	{
					  		fprintf(stderr,"Error in instance/PO number calculation.\n");
					  		exit(1);
						}else{
								strcpy(instList[instance_index].name,str);
								instList[instance_index].instLibNo=PO_LIB_NO;
								SinkP[po_index++]=&instList[instance_index];
								++instance_index;
							 }

						ch = fgetc(fin);
						ungetc(ch,fin);
					   }
				}
			  }else
			   if(strcmp(str,"$$.IOPUT") == 0)
			   {
			   		statementFg=PIO_STATEMENT;
			   }else
			    if(strcmp(str,"$$.DIC")==0)
			    {
			    	statementFg=DIC_STATEMENT;
			    	while(strcmp(str, "$$.DICEND")!=0)
			    	{
			    		fscanf(fin,"%s",str);
			    		strupper(str);
					}
				}else
				 if(strcmp(str, "$")==0 ||
				    strcmp(str,"$$")==0)
				    	continue;
				 else if(str[0] == '@' )
				 {
				 	strcpy(tmpStr,&(str[1]));
					strcpy(instList[instance_index].name, tmpStr);

					fscanf(fin,"%s", str);
					if((t=PWgetLibNo(str,PWdeviceName))<0)
					{
						fprintf(stderr,"cell %s not in lib.\n",str);
						fprintf(stderr,"please update lib!\n");
						exit(1);
					}
					instList[instance_index].instLibNo=t;

					if(t==DF_LIB_NO  ||
				       t==DFR_LIB_NO ||
				       t==DFS_LIB_NO ||
				       t==DFRS_LIB_NO)
					{
						/* collect flip-flops */
						PWinstListPREInsert(&(DFlipList),&(instList[instance_index]));
					}
					/* hash table for netName */

					lastStrInLine(fin, str);
					/*assumption: PIs/POs are created before any other inst*/
					if(isRealPIO(SourceP,PrimInNum,SinkP,PrimOutNum,str)==NULL)
					{
						hashTableInsert(NetHashTable, str, 
										(pointer) (&(instList[instance_index])),
										hashTableSize,strHF2);
					}
					instance_index++;

				 }else{
				 		fprintf(stderr,"Incorrect format around %s.\n",str);
				 		exit(1);
					  }
	  }while(toNextLine(fin));

	/* currently, PrimInNum/PrimOutNum is only for real PI/PO */
	PSEUDO_PI_index=PrimInNum;
	PSEUDO_PO_index=PrimOutNum;

	 /* connection */
	 IntelCktConnect(fin,NetHashTable,hashTableSize);

	 /*hashTableRelease1(&NetHashTable, hashTableSize);*/

	PWinstNum_and_pseudo = PWinstNum+pseudoCount;
	PWlabelCirFanOut_FanIn();

	fprintf(stdout,"Netlist has been read.\n");

	 /******
	 if(DEBUG_INFO)
	 {
	 *******/
		for(i=0;i<PWinstNum_and_pseudo;i++)
        {
            fprintf(dbgFp,"\n\n\n");
            fprintf(dbgFp,"Circuit Output for Debugging...\n");

            fprintf(dbgFp,"line:%d instname:%s instLibNo:%d \
                    fanout %d fanin %d\n",
                    i,instList[i].name,instList[i].instLibNo,
                    instList[i].fanout, instList[i].fanin);

            fprintf(dbgFp,"its output to:\n");
            for(j=0;j<MAX_FAN_OUT&&instList[i].out[j]!=NULL;j++)
            {   fprintf(dbgFp,"out%d:%s\n",j+1,instList[i].out[j]->name);
            }
            for(j=0;j<MAX_FAN_OUT&&instList[i].outz[j]!=NULL;j++)
            {   fprintf(dbgFp,"outz%d:%s\n",j+1,instList[i].outz[j]->name);
            }
            fprintf(dbgFp,"its input from:\n");
            for(j=0;j<MAX_IN_PER_GATE&&instList[i].in[j]!=NULL;j++)
            {   fprintf(dbgFp,"in%d:%s\n",j+1,instList[i].in[j]->name);
            }

            fprintf(dbgFp,"\n\n");
        }

        fprintf(dbgFp,"PIs are %ld:(NOTE:PSEUDO_PI counted)\n",PrimInNum);
        for(i=0;i<MAX_PI && SourceP[i]!=NULL;i++)
        {   fprintf(dbgFp,"%d:name %s , libNo %d, \
                            ,fanout %d\n",
                        i,SourceP[i]->name, SourceP[i]->instLibNo,
                        SourceP[i]->fanout);

            fprintf(dbgFp,"\tits output to:\n");
            for(j=0;j<MAX_FAN_OUT&&SourceP[i]->out[j]!=NULL;j++)
            {   fprintf(dbgFp,"out%d:%s\n",j+1,SourceP[i]->out[j]->name);
            }
            fprintf(dbgFp,"\tits input from:\n");
            for(j=0;j<MAX_IN_PER_GATE && SourceP[i]->in[j]!=NULL;j++)
            {   fprintf(dbgFp,"in%d:%s\n",j+1,SourceP[i]->in[j]->name);
            }
            fprintf(dbgFp,"\n\n");
        }


        fprintf(dbgFp,"POs are %ld:(NOTE:PSEUDO_PO  counted)\n",PrimOutNum);
        for(i=0;i<MAX_PO && SinkP[i]!=NULL;i++)
            fprintf(dbgFp,"%d:name %s , libNo %d,maxLev %d,minLev %d\n",
                    i,SinkP[i]->name, SinkP[i]->instLibNo,
                    SinkP[i]->maxLev,SinkP[i]->minLev);


        fflush(dbgFp);
		/******
		}
		******/
}

void
IntelPreProcessInfo(FILE *fin, long *PWinstNum, long *DF_Num)
{
	int	statementFg;
	char ch = ' ';
	int	i=0,j=0;
	char strbuff[MAX_STR_BUF];

	do{	
		if( fscanf(fin,"%s",strbuff) <= 0 )
			break;
		strupper(strbuff);

		if(strcmp(strbuff, "$$.INPUT")==0 || 
			( strcmp(strbuff, "$$+")==0 && statementFg==PI_STATEMENT) )
		{
			statementFg=PI_STATEMENT;

			ch=fgetc(fin);
			ungetc(ch,fin);
			while(ch!='\n')
			{
				/*ungetc(ch,fin);*/
				do{
					ch=fgetc(fin);
			  	}while( ch=='\r' || ch == '\t' || ch == ' ');

				ungetc(ch, fin);
				if(ch != '\n')
				{
					fscanf(fin,"%s",strbuff);
					strupper(strbuff);
					j++;
					if(PrimInNum>=MAX_PI)
					{
						fprintf(stderr,"Not enough space for PI & PSEUDO PI\n");
						fprintf(stderr,"Larger matrix needed.\n");
						exit(1);
					}else{
							PrimInNum++;
						 }

					ch = fgetc(fin);
					ungetc(ch,fin);
				}
			}
		}else if(strcmp(strbuff, "$$.OUTPUT")==0 ||
				 (strcmp(strbuff, "$$+")==0 && statementFg==PO_STATEMENT) )
			  {
			  	statementFg=PO_STATEMENT;

			  	ch=fgetc(fin);
			  	ungetc(ch,fin);
			  	while(ch!='\n')
			  	{
			  		/*ungetc(ch,fin);*/
			  		do{
			  			ch=fgetc(fin);
					  }while( ch=='\r' || ch == '\t' || ch == ' ');

					  ungetc(ch, fin);
					  if(ch != '\n')
					  {
					  	fscanf(fin,"%s",strbuff);
					  	strupper(strbuff);
					  	j++;
					  	if(PrimOutNum>=MAX_PO)
					  	{
					  		fprintf(stderr,"Not enough space for PO & PSEUDO PO.\n");
					  		fprintf(stderr,"Larger matrix needed.\n");
					  		exit(1);
						}else{
								PrimOutNum++;
							 }

					  	ch = fgetc(fin);
					  	ungetc(ch,fin);
					  }
				}
			  }else if(strcmp(strbuff, "$$.IOPUT") == 0)
			  		{
			  			statementFg=PIO_STATEMENT;
			  			ch=fgetc(fin);
			  			while(ch!='\n')
			  			{
			  				if( ch!='\r' && ch!='\t' && ch!=' ' )
			  				{
			  					fprintf(stderr,"Currently IOPUT is not supported.\n");
			  					exit(1);
							}
			  				ch=fgetc(fin);
						}
					}else
			  		if(strcmp(strbuff, "$$.DIC")==0)
			  		{
			  			statementFg=DIC_STATEMENT;
			  			while(strcmp(strbuff, "$$.DICEND")!=0)
			  			{
			  				fscanf(fin,"%s",strbuff);
			  				strupper(strbuff);
						}
					}else

					if(strcmp(strbuff, "$")==0 || 
					   strcmp(strbuff, "$$")==0)
						continue;
					else if(strbuff[0] == '@')
					{
						statementFg=INST_STATEMENT;
					
						j++;
						fscanf(fin,"%s",strbuff);
						strupper(strbuff);
						if(strbuff[0] == '@')
						{
							fprintf(stderr,"Incorrect format around %s.\n", strbuff);
							exit(1);
						}else if(strcmp(strbuff, "DF")==0 ||
								 strcmp(strbuff, "DFR")==0 ||
								 strcmp(strbuff, "DFS")==0 ||
								 strcmp(strbuff, "DFRS")==0)
							  {
								i++;
							  }
					}else {
							fprintf(stderr,"Incorrect format around %s.\n",
									strbuff);
							exit(1);
						  }

	}while(toNextLine(fin));

	*PWinstNum=j; /* # of instances */
	*DF_Num=i;	  /* # of FFs */
}

/* get the last string in the line */
void
lastStrInLine(FILE* fin, char* str)
{
	char ch;

	ch=fgetc(fin);
	ungetc(ch,fin);
	while(ch!='\n')
	{
		/*ungetc(ch,fin);*/
		do{
			ch=fgetc(fin);
		}while( ch=='\r' || ch == '\t' || ch == ' ');

		ungetc(ch, fin);
		if(ch != '\n')
		{
			fscanf(fin,"%s",str);
			strupper(str);
			ch = fgetc(fin);
			ungetc(ch,fin);
		}
	}
}

/* if the name belongs to some PI/PO, then get its address */
PW_INST *
isRealPIO(PW_INST* SourceP[], long inNum,
		  PW_INST* SinkP[],   long outNum, char str[])
{
	int	i;
	for(i=0;i<inNum;i++)
	{
		if(strcmp(SourceP[i]->name, str)==0)
		{
			return SourceP[i];
		}
	}

	for(i=0;i<outNum;i++)
	{
		if(strcmp(SinkP[i]->name, str)==0)
		{
			return SinkP[i];
		}
	}
	return NULL;
}

void
IntelCktConnect(FILE* fin, lin_table_rec *NetHashTable, int hashsize)
{
	char str[MAX_STR_BUF],tmpStr[MAX_STR_BUF],ch;
	int	 gateIndex;

	/* Note: pesuedo PI/POs haven't been created yet */
	gateIndex = PrimInNum + PrimOutNum;

	rewind(fin);
	do{
		if(fscanf(fin,"%s",str)<=0)
			break;
		strupper(str);

		if(strcmp(str,"$$.INPUT")==0 ||
		   strcmp(str,"$$.OUTPUT")==0 ||
		   strcmp(str,"$$.IOPUT")==0 ||
		   strcmp(str,"$$+")==0 ||
		   strcmp(str,"$$")==0 ||
		   strcmp(str,"$")==0 ||
		   strcmp(str,"$$.DIC")==0 ||
		   strcmp(str,"$$.DICEND")==0)
		{
			continue;
		}else if(str[0] == '@')
			  {
			  	fscanf(fin,"%s", str);
			  	fscanf(fin,"%s", str);
			  	strupper(str);

			  	ch=fgetc(fin);
			  	ungetc(ch,fin);
			  	while(ch!='\n')
			  	{
			  		strcpy(tmpStr,str);

			  		/*ungetc(ch,fin);*/
			  		do{
			  			ch=fgetc(fin);
					}while( ch=='\r' || ch == '\t' || ch == ' ');

					ungetc(ch,fin);
					if(ch != '\n')
					{ /* the previous one is an input */
						IntelConnectInput(tmpStr,NetHashTable,hashsize,&(instList[gateIndex]));
						fscanf(fin,"%s",str);
						strupper(str);
						ch = fgetc(fin);
						ungetc(ch,fin);
						if(ch=='\n')
						{ /* it's identified as output immediately */
						  IntelConnectOutput(str,NetHashTable,hashsize,&(instList[gateIndex]));
						}
					}else{
							/* the previous one is the last one,then an output*/
							IntelConnectOutput(tmpStr,NetHashTable,hashsize,&(instList[gateIndex]));
 						 }
				}
				gateIndex++;
			  }
	  }while(toNextLine(fin));
}

void
IntelConnectInput(char net_name[], lin_table_rec * net_hashtable, int size, PW_INST* currInst)
{
	PW_INST	* 	inInst=NULL;

	inInst=IntelFaninInst(net_name, net_hashtable, size, currInst);

	/** link the instance (indicated by lineNumber <l>),
        and its fanin instance <inInst> to form the network
    **/

	linkOperation(currInst,inInst);
}

void
linkOperation(PW_INST* inst, PW_INST* inInst)
{

	switch(inst->instLibNo)
	{
		case PO_LIB_NO:
		case NOT_LIB_NO:
		case BUFF_LIB_NO:
			dualLinkNOTType("I",inInst,inst->instID,FALSE);
			break;
		case AND_LIB_NO:
		case NAND_LIB_NO:
		case OR_LIB_NO:
		case NOR_LIB_NO:
		case XOR_LIB_NO:
		case XNOR_LIB_NO:
			dualLinkAllEqu("I",inInst,inst->instID,FALSE);
			break;
		case DF_LIB_NO:
		case DFR_LIB_NO:
		case DFS_LIB_NO:
		case DFRS_LIB_NO:
			dualLinkDFtype("I",inInst,inst->instID,FALSE);
			break;
		default:
			fprintf(stderr,"unrecognized cell %s\n",inst->name);
			exit(1);
	}
}



PW_INST	*
IntelFaninInst(char net_name[], lin_table_rec* net_hashtable, int size, PW_INST* currInst)
{
	int			i;
	PW_INST		*inInst=NULL;

	PW_INST		*q;

	/* first check reserved keywords, then check pi/po , then hash table */
	/* for sequential ckt, pseudo PI/PO may be added */

	/* NOTE: GND, VDD are not counted into PWinstNum or PWinstNum_pseudo */
	if(strcmp(net_name,"GND")==0 ||
	   strcmp(net_name,"VSS")==0 )
	{
		for(i=0;i<PrimInNum;i++)
		{	/** we need only one GND **/
			if(strcmp(SourceP[i]->name,"GND")==0 ||
			   strcmp(SourceP[i]->name,"VSS")==0)
				return (SourceP[i]);
		}
		q=createPIinst("VSS");
		gnd_index=PrimInNum-1;
		if(gnd_index<0)
		{
			fprintf(stderr,"Error in creating GND input.\n");
			exit(1);
		}
		return q;
	}
	if(strcmp(net_name,"VDD")==0 ||
	   strcmp(net_name,"VCC")==0 )
	{
		for(i=0;i<PrimInNum;i++)
		{
			if(strcmp(SourceP[i]->name,"VDD")==0 ||
			   strcmp(SourceP[i]->name,"VCC")==0 )
				return SourceP[i];
		}
		q=createPIinst("VCC");
		vdd_index=PrimInNum-1;
		if(vdd_index<0)
		{
			fprintf(stderr,"Error in creating VDD input.\n");
			exit(1);
		}
		return q;
	}


	for(i=0;i<PrimInNum + PrimOutNum;i++)
	{
		if(strcmp(instList[i].name, net_name)==0)
		{
			inInst=&(instList[i]);
			break;
		}
	}
	if(inInst == NULL)
	{
		inInst=(PW_INST*) hashTableFind(net_hashtable, net_name, size, strHF2);
		if(inInst == NULL)
		{
			fprintf(stderr,"Error in Looking up Hash Table for %s.\n",net_name);
			fprintf(stderr,"Check whether it's a floating node!!.\n");
			exit(1);
		}
	}

	q=pseudoFaninGate(inInst, currInst);
	return q;
}

/* get the proper fanin gate if pseuod PI/PO needs to be inserted */
PW_INST	*
pseudoFaninGate(PW_INST* inInst, PW_INST* currInst)
{
	PW_INST		*q;
	char		*strNameTmp, strName[20];

	q=inInst;

	if(inInst->instLibNo == DF_LIB_NO  ||
	   inInst->instLibNo == DFR_LIB_NO ||
	   inInst->instLibNo == DFS_LIB_NO ||
	   inInst->instLibNo == DFRS_LIB_NO)
	{
		/* currently we only support O or Q, no QZ is supported */
		if( inInst->out[0]==NULL )
		{
			pseudoCount++;
			strNameTmp=intToStr(pseudoCount);
			strcpy(strName,"PSEUDO");
			strcat(strName,strNameTmp);
			free(strNameTmp);

			q=&(instList[PWinstNum+pseudoCount-1]);
			q->zflag=FALSE; /* QZ not supported */
			strcpy(q->name,strName);
			q->instLibNo=PI_LIB_NO;
			q->instID=PWinstNum+pseudoCount-1;
			q->fanin=1;
			q->maxLev=q->minLev=0;
			if(PrimInNum>=MAX_PI)
			{
				fprintf(stderr,"Not enough space for PI & PSEUDO_PI\n");
				fprintf(stderr,"Larger matrix needed.\n");
				exit(1);
			}
			SourceP[PrimInNum++]=q;

			/* link fanout instances */
			linkOutput(inInst, q, FALSE);
			/* link fanin instance */
			q->in[0]=inInst;
	 	 }else{
	 	 		/*** PSEUDO PI ALREADY EXITS ***/
	 	 		/* QZ not supported */
	 	 		q=inInst->out[0];
			  }
	}else if(currInst->instLibNo == DF_LIB_NO  ||
			 currInst->instLibNo == DFR_LIB_NO ||
			 currInst->instLibNo == DFS_LIB_NO ||
			 currInst->instLibNo == DFRS_LIB_NO)
		  {
		  	/* not supported case */
		  	if(inInst->instLibNo == DF_LIB_NO ||
		  	   inInst->instLibNo == DFR_LIB_NO||
		  	   inInst->instLibNo == DFS_LIB_NO||
		  	   inInst->instLibNo == DFRS_LIB_NO)
			{
				fprintf(stderr,"Currenly we only consider canonical sequential ckt model. No pipeline structure.\n");
				exit(1);
			}

			if( isOutput2PO(inInst,FALSE) == FALSE )
			{
				/* pseudo PO needs to be inserted */
				pseudoCount++;
				strNameTmp=intToStr(pseudoCount);
				strcpy(strName,"PSEUDO");
				strcat(strName,strNameTmp);
				free(strNameTmp);

				/*pseudo PO*/
				q=&(instList[PWinstNum+pseudoCount-1]);
				q->zflag=FALSE;
				strcpy(q->name,strName);
				q->instLibNo=PO_LIB_NO;
				q->instID=PWinstNum+pseudoCount-1;
				q->fanin=1;
				q->fanout=0;
				if(PrimOutNum>=MAX_PO)
				{
					fprintf(stderr,"Not enough space for PO & PSEUDO_PO\n");
					fprintf(stderr,"Larger matrix needed.\n");
					exit(1);
				}
				SinkP[PrimOutNum++]=q;

				linkOutput(inInst,q,FALSE);
				q->in[0]=inInst;
			 }
			 q=inInst;
		  }else q=inInst;

	 return q;
}

/* connect to output */
void
IntelConnectOutput(char out_net_name[], lin_table_rec * net_hashtable, int size, PW_INST* currInst)
{
	int	i;
	PW_INST	*outInst=NULL,*q;

        for(i=0;i<PrimInNum + PrimOutNum;i++)
        {
                if(strcmp(instList[i].name, out_net_name)==0)
                        outInst=&(instList[i]);
        }
	/* if it's not a PO, just do nothing */
	/* the connect will be done when the net becomes an input */
	if(outInst==NULL)
		return;
	
	q=pseudoFaninGate(currInst,outInst);

	linkOperation(outInst,q);
}
