#include <stdio.h>
#include "DFtype.h"
#include "DFstring.h"
#include "DFfile.h"
#include "DFalloc.h"
#include "PWcir.h"
#include <string.h>
#include "sis.h"

#include "PWroutines.h"

extern PW_INST 	*unknownCellList,*instList,*DFlipList;
extern long		PWlineNum,PWinstNum,PrimInNum,PrimOutNum,PSEUDO_PI_index;
extern PW_INST 	* SourceP[],* SinkP[];
extern char    	*PWdeviceName[];
extern FILE	   	*dbgFp;
extern long		PSEUDO_PO_index;
extern long		DF_Num;

extern	int		pseudoCount;
extern  long	PWinstNum_and_pseudo;

extern PW_INST 	*PWgetInputInst(PW_INST *currInst, char *netName, BOOLEAN *zflagp);
extern void		reportCirStatistics(FILE * dbgFp, long realPInum, long realPOnum);
extern int 		getNewInputLinkIndex(int lineNo);
extern BOOLEAN 	isOutput2PO(PW_INST *pInst, BOOLEAN zflag);
extern void 	getPreProcessInfo(FILE *fin, long *PWlineNum, long *DF_Num);
extern void 	PWgetCirNet(FILE *fin);
extern void		dualLinkMUX21(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void		dualLinkEqu21(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void		dualLinkEqu211(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void		dualLinkEqu221(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void		dualLinkEqu31(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void		dualLinkEqu32(char* pinName, PW_INST* inInst, long lineNo, BOOLEAN zflag);
extern void 	clearVisitedFlag(void);


/*----------------------------------------------------------------------------------*/
void	NetRead(FILE *fin)
{
	int		i,j,t;
	char	str[MAX_STR_BUF];


	unknownCellList=NULL;
	DFlipList=NULL;
	DF_Num=0;

	PrimInNum=0;
	PrimOutNum=0;

	pseudoCount=0;

	fprintf(stdout,"Reading...\n");

	getPreProcessInfo(fin,&PWlineNum,&DF_Num);

	/* # of instance 
		requirement: one instance per line in input file
	*/
	PWinstNum=PWlineNum;

	/* allocation: # of instances + # of possible pseudo PI/PO */
	/* assume only output Q of flip-flop is supported */
	instList=(PW_INST *)mem_get_block((SIZE_T) (PWlineNum+2*DF_Num)*sizeof(PW_INST));

	/* initialization */
	initPWinst(instList);
	
	/* reading instances information */
	rewind(fin);
	for(i=0;i<PWlineNum;i++)
	{	fscanf(fin,"%s",str);
		strupper(str);
		strcpy(instList[i].name,str);
		fscanf(fin,"%s",str);
		strupper(str);
		/** manipulate primary IO **/
		if(strcmp(str,"PI")==0)
		{	
			instList[i].instLibNo=PI_LIB_NO;
			if(PrimInNum>=MAX_PI)
			{	fprintf(stderr,"Not enough space for PI & PSEUDO PI\n");
				fprintf(stderr,"Larger matrix needed.\n");
				exit(1);
			}else{
					SourceP[PrimInNum++]=&instList[i];
				 }
			instList[i].maxLev=0;/****for PI ****/
			instList[i].minLev=0;
		}else
		if(strcmp(str,"PO")==0)
		{	
			instList[i].instLibNo=PO_LIB_NO;
			if(PrimOutNum>=MAX_PO)
			{	fprintf(stderr,"Not enough space for PO\n");
				fprintf(stderr,"Larger matrix needed.\n");
				exit(1);
			}else{
					SinkP[PrimOutNum++]=&instList[i];
				 }
		}else 	/** manipulate real gates **/
			{	if((t=PWgetLibNo(str,PWdeviceName))<0)
				{
					fprintf(stderr,"cell %s not in lib at line %d\n",str,i+1);
					fprintf(stderr,"please update lib!\n");
					exit(1);
				}
				instList[i].instLibNo=t;  /* t==-1 if not in lib */

				if(t==DF_LIB_NO  ||
				   t==DFR_LIB_NO ||
				   t==DFS_LIB_NO ||
				   t==DFRS_LIB_NO)
				{	
				    /* collect flip-flops */
				    PWinstListPREInsert(&(DFlipList),&(instList[i]));
				}
			}
		toNextLine(fin);
	}

	/* starting index for pseudo PI/PO, also the # of real primary inputs*/
	PSEUDO_PI_index=PrimInNum;
	PSEUDO_PO_index=PrimOutNum;

	reportCirStatistics(dbgFp, PSEUDO_PI_index, PSEUDO_PO_index);

	if(DEBUG_INFO)
	{ 
		fprintf(dbgFp,"Primary input number(not including PSEUDO PI)=%ld\n",
				PSEUDO_PI_index);

		for(i=0;i<PWlineNum;i++) 
		{	fprintf(dbgFp,"line:%d instname:%s instLibNo:%d\n", 
					i,instList[i].name,instList[i].instLibNo);
		}

		fprintf(dbgFp,"\n\n");

	}

	/* link the gates into a circuit network & update pseudoCount */
	PWgetCirNet(fin);

	PWinstNum_and_pseudo = PWinstNum+pseudoCount;

	PWlabelCirFanOut_FanIn();

	fprintf(stdout,"Netlist has been read.\n");


	/* debugging */
	/***********
	if(DEBUG_INFO)
	{	
	*************/
		for(i=0;i<PWinstNum_and_pseudo;i++)
		{   
			fprintf(dbgFp,"\n\n\n");
			fprintf(dbgFp,"Circuit Output for Debugging...\n");

			fprintf(dbgFp,"line:%d instname:%s instLibNo:%d \
					fanout %d fanin %d\n",
					i,instList[i].name,instList[i].instLibNo,
					instList[i].fanout, instList[i].fanin);

			fprintf(dbgFp,"its output to:\n");
			for(j=0;j<MAX_FAN_OUT&&instList[i].out[j]!=NULL;j++)
			{	fprintf(dbgFp,"out%d:%s\n",j+1,instList[i].out[j]->name);
			}
			for(j=0;j<MAX_FAN_OUT&&instList[i].outz[j]!=NULL;j++)
			{	fprintf(dbgFp,"outz%d:%s\n",j+1,instList[i].outz[j]->name);
			}
			fprintf(dbgFp,"its input from:\n");
			for(j=0;j<MAX_IN_PER_GATE&&instList[i].in[j]!=NULL;j++)
			{	fprintf(dbgFp,"in%d:%s\n",j+1,instList[i].in[j]->name);
			}

			fprintf(dbgFp,"\n\n");
		}


		fprintf(dbgFp,"PIs are %ld:(NOTE:PSEUDO_PI counted)\n",PrimInNum);
		for(i=0;i<MAX_PI && SourceP[i]!=NULL;i++)
		{	fprintf(dbgFp,"%d:name %s , libNo %d, \
							,fanout %d\n",
						i,SourceP[i]->name, SourceP[i]->instLibNo,
						SourceP[i]->fanout);

			fprintf(dbgFp,"\tits output to:\n");
			for(j=0;j<MAX_FAN_OUT&&SourceP[i]->out[j]!=NULL;j++)
			{	fprintf(dbgFp,"out%d:%s\n",j+1,SourceP[i]->out[j]->name);
			}
			fprintf(dbgFp,"\tits input from:\n");
			for(j=0;j<MAX_IN_PER_GATE && SourceP[i]->in[j]!=NULL;j++)
			{	fprintf(dbgFp,"in%d:%s\n",j+1,SourceP[i]->in[j]->name);
			}
			fprintf(dbgFp,"\n\n");
		}


		fprintf(dbgFp,"POs are %ld:(NOTE:PSEUDO_PO  counted)\n",PrimOutNum);
		for(i=0;i<MAX_PO && SinkP[i]!=NULL;i++)
			fprintf(dbgFp,"%d:name %s , libNo %d,maxLev %d,minLev %d\n",
					i,SinkP[i]->name, SinkP[i]->instLibNo,
					SinkP[i]->maxLev,SinkP[i]->minLev);


		fflush(dbgFp);
	/**************
	}
	*****************/
}

/*------------------------------------------------------------------------*/

/** create & initialize PI instance AND INSERT INTO source list **/
/** for pseudo PI, GND, VDD etc. **/
PW_INST * createPIinst(name)
char *name;
{	PW_INST *tmp;
	int i;

	if(PrimInNum>=MAX_PI)
	{	fprintf(stderr,"MAX_PI is exceeded!!\n");
		exit(1);
	}else{
		tmp=(PW_INST *)mem_get_block((SIZE_T) sizeof(PW_INST));

    	/* initialize */
    	for(i=0;i<MAX_IN_PER_GATE;i++)
    		tmp->in[i]=NULL;
		for(i=0;i<MAX_FAN_OUT;i++)
		{
    		tmp->out[i]=NULL;
    		tmp->outz[i]=NULL;
		}
		tmp->next=NULL;
    	tmp->instLibNo=PI_LIB_NO;
    	tmp->instID=-1;
		tmp->maxLev=0;
		tmp->minLev=0;
		tmp->visited=FALSE;
		tmp->fanout=0;
		tmp->fanin=1;	/* special for pseudo PI */
		tmp->instPower=0;
		tmp->InSortedList[0]=NULL;
		tmp->InSortedList[1]=NULL;
		tmp->state=VALUE_UNKNOWN;
		tmp->implyStkTop=NULL;
		tmp->pGainUpdt_st[0]=NULL;
		tmp->pGainUpdt_st[1]=NULL;
		tmp->gain_st[0]=0;
		tmp->gain_st[1]=1;
		tmp->new_gain_st[0]=0;
		tmp->new_gain_st[1]=0;

		tmp->sis_node = NULL;
        tmp->powerup_current =
        (gate_current_t *) mem_get_block((SIZE_T) sizeof(gate_current_t));


		STK_PUSH(&(tmp->implyStkTop),(double)VALUE_UNKNOWN,(pointer)NULL);

    	strcpy(tmp->name,name);
		if(PrimInNum>=MAX_PI)
		{	fprintf(stderr,"Not enought space for PI & PSEUDO_PI\n");
			fprintf(stderr,"Larger matrix needed.\n");
			exit(1);
		}else{
    			SourceP[PrimInNum++]=tmp;
			 }
		return tmp;
		}
}
	
/** create & initialize PO instance AND INSERT into sink list **/
PW_INST *createPOinst(name)
char *name;
{	PW_INST *tmp;
	int i;

	if(PrimOutNum>=MAX_PO)
	{	fprintf(stderr,"MAX_PO is exceeded!!\n");
		exit(1);
	}else{
		tmp=(PW_INST *) mem_get_block((SIZE_T) sizeof(PW_INST));

    	/* initialize */
    	for(i=0;i<MAX_IN_PER_GATE;i++)
			tmp->in[i]=NULL;
		for(i=0;i<MAX_FAN_OUT;i++)
		{
    		tmp->out[i]=NULL;
    		tmp->outz[i]=NULL;
		}
		tmp->next=NULL;
       	tmp->instLibNo=PO_LIB_NO; 
       	tmp->instID=-1;
		tmp->minLev=-1;
		tmp->maxLev=-1;
		tmp->visited=FALSE;
		tmp->fanout=0;
		tmp->fanin=1;
		tmp->instPower=0;
		tmp->InSortedList[0]=NULL;
		tmp->InSortedList[1]=NULL;
		tmp->implyStkTop=NULL;
		tmp->state=VALUE_UNKNOWN;
		tmp->pGainUpdt_st[0]=NULL;
		tmp->pGainUpdt_st[1]=NULL;
		tmp->gain_st[0]=0;
		tmp->gain_st[1]=0;
		tmp->new_gain_st[0]=0;
		tmp->new_gain_st[1]=0;

		tmp->sis_node = NULL;
        tmp->powerup_current =
        (gate_current_t *) mem_get_block((SIZE_T) sizeof(gate_current_t));


		STK_PUSH(&(tmp->implyStkTop),(double)VALUE_UNKNOWN,(pointer) NULL);

       	strcpy(tmp->name,name);
		if(PrimOutNum>=MAX_PO)
		{	fprintf(stderr,"Not enough space for PO\n");
			fprintf(stderr,"Lager matrix needed.\n");
			exit(1);
		}else{
       			SinkP[PrimOutNum++]=tmp;
			 }
		return tmp;
		}
}


/** get the cell No. in lib, -1 means it's not in lib **/
/** First change to upper case for comparison **/
int PWgetLibNo(cell,devNameList)
char	*cell;
char	*devNameList[];
{	int i;
	strupper(cell);	
	for(i=0;i<NUM_LIB_CELL;i++)
	{	
		if(strcmp(cell,devNameList[i])==0)
			return (i+1);
	}
	/* for compatibility */
	if(strcmp(cell,"AND2")==0 ||
		strcmp(cell,"AND3")==0 ||
		strcmp(cell,"AND4")==0 ||
		strcmp(cell,"AND5")==0)
		return AND_LIB_NO;
	if(strcmp(cell,"NAND2")==0 ||
		strcmp(cell,"NAND3")==0 ||
		strcmp(cell,"NAND4")==0 ||
		strcmp(cell,"NAND5")==0)
		return NAND_LIB_NO;
	if(strcmp(cell,"OR2")==0 ||
		strcmp(cell,"OR3")==0 ||
		strcmp(cell,"OR4")==0 ||
		strcmp(cell,"OR5")==0)
		return OR_LIB_NO;
	if(strcmp(cell,"NOR2")==0 ||
		strcmp(cell,"NOR3")==0 ||
		strcmp(cell,"NOR4")==0 ||
		strcmp(cell,"NOR5")==0)
		return NOR_LIB_NO;
	if(strcmp(cell,"INV")==0)
		return NOT_LIB_NO;

	return -1;
}

/*--------------------------------------------------------------*/
/* initialize the list of instances */
void
initPWinst(list)
PW_INST *list;
{	int i,j;
	for(i=0;i<PWlineNum+2*DF_Num;i++)
	{	
		for(j=0;j<MAX_IN_PER_GATE;j++)
			list[i].in[j]=NULL;
		for(j=0;j<MAX_FAN_OUT;j++)
		{	list[i].out[j]=NULL;
			list[i].outz[j]=NULL;
		}
		list[i].next=NULL;
		list[i].maxLev=-1;/** unassigned **/
		list[i].minLev=-1;
		list[i].visited=FALSE;
		list[i].fanout=list[i].fanin=0;
		list[i].InSortedList[0]=NULL;
		list[i].InSortedList[1]=NULL;
		list[i].implyStkTop=NULL;
		list[i].state=VALUE_UNKNOWN;
		list[i].instPower=0;
		list[i].instID=i;	/* ID number */
		list[i].pGainUpdt_st[0]=NULL;
		list[i].pGainUpdt_st[1]=NULL;
		list[i].gain_st[0]=0;
		list[i].gain_st[1]=0;
		list[i].new_gain_st[0]=0;
		list[i].new_gain_st[1]=0;

		list[i].sis_node = NULL;
		list[i].powerup_current = 
		(gate_current_t *) mem_get_block((SIZE_T) sizeof(gate_current_t));
	}
}
/*--------------------------------------------------------------*/
void
PWgetCirNet(fin)
FILE *fin;
{	long 	l;
	int  	j;	
	char 	name[MAX_STR_BUF],IOinfo[MAX_STR_BUF]; 
	char 	*pin,*netName;
	PW_INST *inInst;
	BOOLEAN zflag;

	rewind(fin);
	/** FOR ONE INSTANCE IN NETLIST **/
	for(l=0;l<PWlineNum;l++)
	{	
		if(instList[l].instLibNo<0)
		{	fprintf(stderr,"Unrecognized cell passed to PWgetCirNet()\n");
			exit(1);
		}
		else
		{
			fscanf(fin,"%s",name);
			fscanf(fin,"%s",name);
			strupper(name); /** instance type name **/
			
			if(strcmp(name,"PI")==0)
			{	toNextLine(fin);
				continue;
			}

			/*NOTE:all IO connection info should be in one string*/
			fscanf(fin,"%s",IOinfo);
			strupper(IOinfo);

	    while((j=findSpecChar(IOinfo,'='))>=0) /* while "=" still can be found*/
    	 {  
    	 	pin=getSubStr(IOinfo,0,j-1);	/* pin name */
        	strCutBehindSpecChar(IOinfo,'=');
 
        	if((j=findSpecChar(IOinfo,','))<0)
			{	if((j=findSpecChar(IOinfo,';'))<0)
				{   fprintf(stderr,"Input net file syntax error at line %ld!\n",
							l+1);
					/*return;*/
					exit(1);
				}
			}

			/** fanin net name **/
        	netName=getSubStr(IOinfo,0,j-1);
        	if(strCutBehindSpecChar(IOinfo,',')==FALSE)
        	{   if(strCutBehindSpecChar(IOinfo,';')==FALSE)
            	{   fprintf(stderr,"Input net file syntax error at line %ld!\n",
							l+1);
					/*return;*/
                	exit(1);
            	}
			}

			/** the cases that require no link operation **/
			/** because of the input file format **/
			if(strcmp(pin,"O")==0 || 
			   strcmp(pin,"Q")==0 ||
			   strcmp(pin,"QZ")==0)
			{	continue;
			}

			/** fanin instance **/
			/** the returned <zflag> is always FALSE here, 
				because it's only for pseudo PI 
			**/
			inInst=PWgetInputInst(&(instList[l]),netName,&zflag);

			/** link the instance (indicated by lineNumber <l>),
				and its fanin instance <inInst> to form the network
			**/
			/** MUX21, AOI are not actually used in this version **/
			if(strcmp(name,"PO")==0||strcmp(name,"NOT")==0
				||strcmp(name,"BUFF")==0)
				dualLinkNOTType(pin,inInst,l,zflag);
			else if(strcmp(name,"AND")==0)
				dualLinkAllEqu(pin,inInst,l,zflag);
			else if(strcmp(name,"NAND")==0)
				dualLinkAllEqu(pin,inInst,l,zflag);
			else if(strcmp(name,"OR")==0)
				dualLinkAllEqu(pin,inInst,l,zflag);
			else if(strcmp(name,"NOR")==0)
				dualLinkAllEqu(pin,inInst,l,zflag);
			else if(strcmp(name,"XOR")==0||strcmp(name,"XNOR")==0)
				dualLinkAllEqu(pin,inInst,l,zflag);
			else if(strcmp(name,"MUX21")==0)
				dualLinkMUX21(pin,inInst,l,zflag);
			else if(strcmp(name,"AOI21")==0||strcmp(name,"OAI21")==0)
				dualLinkEqu21(pin,inInst,l,zflag);
			else if(strcmp(name,"AOI211")==0||strcmp(name,"OAI211")==0)
				dualLinkEqu211(pin,inInst,l,zflag);
			else if(strcmp(name,"AOI221")==0||strcmp(name,"OAI221")==0)
            	dualLinkEqu221(pin,inInst,l,zflag);
			else if(strcmp(name,"AOI31")==0||strcmp(name,"OAI31")==0)
            	dualLinkEqu31(pin,inInst,l,zflag);
			else if(strcmp(name,"AOI311")==0||strcmp(name,"OAI311")==0)
            	dualLinkEqu32(pin,inInst,l,zflag);
			else if(strcmp(name,"DF")==0 || strcmp(name,"DFR")==0 ||
					strcmp(name,"DFS")==0 || strcmp(name,"DFRS")==0)
				dualLinkDFtype(pin,inInst,l,zflag);
			else {	fprintf(stderr,"unrecongnized cell \
                        	not filtered at line %ld!\n",l+1);
					/*return;*/
					exit(1);
			 	 }
		 }
		}
		toNextLine(fin);
	}
}

/*-------------------------------------------------------------------------*/

/** link instance <inInst> to its  fanout instance <nextInst> & check fanout **/
/** only forward direction link **/
void
linkOutput(inInst,nextInst,zflag)
PW_INST* inInst,*nextInst;
BOOLEAN      zflag;
{	int o,oz;
	for(o=0;o<MAX_FAN_OUT;o++)
	{	if(inInst->out[o]==NULL)
		break;
	}
	if(o==MAX_FAN_OUT)
	{	fprintf(stderr,"instance %s 'out' fanout exceeds limit!\n",
				inInst->name);
		exit(1);
	}
    for(oz=0;oz<MAX_FAN_OUT;oz++) 
    {   if(inInst->outz[oz]==NULL)  
        break;
    }
    if(oz==MAX_FAN_OUT)
    {   fprintf(stderr,"instance %s 'outz' fanout exceeds limit!\n",
                inInst->name);
        exit(1);
    }
	if(zflag==FALSE)
		inInst->out[o]=nextInst;
	else	inInst->outz[oz]=nextInst;
}

/*-------------------------------------------------------------------------*/
/*   To connect gate input pin to fanin gate <inInst> */

/*** for NOT/BUFF type: one input,one output*/
void
dualLinkNOTType(pinName,inInst,lineNo,zflag)
char*	 pinName;
PW_INST* inInst;
long	 lineNo;/** starting from 0 **/
BOOLEAN		 zflag;
{	
	if(strcmp(pinName,"O")==0)
		return;
	else if(strcmp(pinName,"I")==0)
		instList[lineNo].in[0]=inInst;	
	else{
			fprintf(stderr,"IO info in input file error at line %ld!\n",
					lineNo+1);
            exit(1);
		}
	 linkOutput(inInst,&(instList[lineNo]),zflag);
}
					
/* for instance whose inputs are all equivalent */
void
dualLinkAllEqu(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{
	int index;

	if(strcmp(pinName,"O")==0)
        return;
	else if(strcmp(pinName,"I")==0)
		 {
		 	index=getNewInputLinkIndex(lineNo);
		 	instList[lineNo].in[index]=inInst;
		 }
		else{
				fprintf(stderr,"IO info in input file error!\n");
				exit(1);
			}

	linkOutput(inInst,&(instList[lineNo]),zflag);
}

/*-------------------------------------------*/
/*  below is not used in current version */
void
dualLinkMUX21(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{	
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"I1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"I2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"C")==0)
        instList[lineNo].in[2]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }

	linkOutput(inInst,&(instList[lineNo]),zflag);
}

void
dualLinkEqu21(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{	
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"IA1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"IA2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"IB")==0)
        instList[lineNo].in[2]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }

	linkOutput(inInst,&(instList[lineNo]),zflag);
}

void
dualLinkEqu211(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{   
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"IA1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"IA2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"IB")==0)
        instList[lineNo].in[2]=inInst;
    else if(strcmp(pinName,"IC")==0)
        instList[lineNo].in[3]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }

	linkOutput(inInst,&(instList[lineNo]),zflag);
}

void
dualLinkEqu221(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{   
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"IA1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"IA2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"IB1")==0)
        instList[lineNo].in[2]=inInst;
    else if(strcmp(pinName,"IB2")==0)
        instList[lineNo].in[3]=inInst;
	else if(strcmp(pinName,"IC")==0)
        instList[lineNo].in[4]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }

	linkOutput(inInst,&(instList[lineNo]),zflag);
}

void
dualLinkEqu31(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{   
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"IA1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"IA2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"IA3")==0)
        instList[lineNo].in[2]=inInst;
    else if(strcmp(pinName,"IB")==0)
        instList[lineNo].in[3]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }
	
	linkOutput(inInst,&(instList[lineNo]),zflag);
}

void
dualLinkEqu32(pinName,inInst,lineNo,zflag)
char*    pinName;
PW_INST* inInst;
long     lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{   
	if(strcmp(pinName,"O")==0)
        return;
    else if(strcmp(pinName,"IA1")==0)
        instList[lineNo].in[0]=inInst;
    else if(strcmp(pinName,"IA2")==0)
        instList[lineNo].in[1]=inInst;
    else if(strcmp(pinName,"IA3")==0)
        instList[lineNo].in[2]=inInst;
    else if(strcmp(pinName,"IB1")==0)
        instList[lineNo].in[3]=inInst;
    else if(strcmp(pinName,"IB2")==0)
        instList[lineNo].in[4]=inInst;
    else{
            fprintf(stderr,"IO info in input file error!\n");
            exit(1);
        }

	linkOutput(inInst,&(instList[lineNo]),zflag);
}
/* above is not used in current version */
/*----------------------------------------------------------*/

/* For DF, DFS, DFR, DFRS */
/* in the netlist format, the input pin to DF can be 'I' or 'D'*/
void
dualLinkDFtype(pinName,inInst,lineNo,zflag)
char	*pinName;
PW_INST	*inInst;
long	lineNo;/** starting from 0 **/
BOOLEAN      zflag;
{	
	if(strcmp(pinName,"Q")==0 ||
	   strcmp(pinName,"QZ")==0 ||
	   strcmp(pinName,"O")==0)
	   return;
	else if(strcmp(pinName,"R")==0 ||
			strcmp(pinName,"S")==0)
		 return;
		 else if(strcmp(pinName,"CP")==0)
			  return;
			  else if(strcmp(pinName,"D")==0)
				   instList[lineNo].in[0]=inInst;
				   else if(strcmp(pinName,"I")==0)
				   			instList[lineNo].in[0]=inInst;
				   			else{
								fprintf(stderr,
										"IO info in input file error!\n");
								exit(1);
					   			}
	linkOutput(inInst,&(instList[lineNo]),zflag);
}

/*	fanin gate for current instance <currInst> */
PW_INST *PWgetInputInst(currInst,netName,zflagp)
PW_INST	*currInst;
char 	*netName;
BOOLEAN *zflagp;
{	int 	pos,ln,i;
	char 	*tmp1,*tmp2,*strNameTmp,strName[20];

	PW_INST *q;

	if(strcmp(netName,"GND")==0)
	{	for(i=0;i<PrimInNum;i++)
		{	/** we need only one GND **/
			if(strcmp(SourceP[i]->name,"GND")==0)
			return (SourceP[i]);
		}
		q=createPIinst("GND");	
		*zflagp=FALSE;
		return q;
	}else if(strcmp(netName,"VDD")==0)
		{	/** we need only one VDD **/
			for(i=0;i<PrimInNum;i++)
			{	if(strcmp(SourceP[i]->name,"VDD")==0)
				return SourceP[i];
			}
			q=createPIinst("VDD");
			*zflagp=FALSE;
			return q;
		}else{
				ln=strlen(netName);
				pos=findLastSpecChar(netName,'_');
				if(pos == -1)
				{
					fprintf(stderr,"Syntax Error for %s.\n",netName);
					exit(1);
				}
				tmp1=getSubStr(netName,0,pos-1);
				tmp2=getSubStr(netName,pos+1,ln-1);
				if(strcmp(tmp2,"QZ")==0) /* -> QZ not used in current version */
					*zflagp=TRUE;
				else if(strcmp(tmp2,"Q")==0)
					*zflagp=FALSE;
				else if(strcmp(tmp2,"O")==0)
					*zflagp=FALSE;
				else {	fprintf(stderr,
							"output pin doesn't fall into Q or QZ or O\n");
						exit(1);
					}

				/*** find the fanin instance for current gate ***/
				for(ln=0;ln<PWlineNum;ln++)
				{	if(strcmp(instList[ln].name,tmp1)==0)
					break;
				}
				if(ln==PWlineNum)
				{	fprintf(stderr,"inst %s 's output to an unknown inst\n",
							tmp1);
					exit(1);
				}

				if(instList[ln].instLibNo==-1)
				{	
					fprintf(stderr,"Unregconized cell passed to \
									PWgetInputInst()\n");
					exit(1);

				}else if(instList[ln].instLibNo==DF_LIB_NO  ||
						 instList[ln].instLibNo==DFR_LIB_NO ||
						 instList[ln].instLibNo==DFS_LIB_NO ||
						 instList[ln].instLibNo==DFRS_LIB_NO)
					  {	
					  	/* a pseudo PI to be inserted */
						if( ((*zflagp)==FALSE && instList[ln].out[0]==NULL)||
							((*zflagp)==TRUE && instList[ln].outz[0]==NULL) )
						{
							pseudoCount++;
							strNameTmp=intToStr(pseudoCount);
							strcpy(strName,"PSEUDO");
							strcat(strName,strNameTmp);
							free(strNameTmp);

							/*PSEUDO-PI, and it accepts QZ/Q of DFF*/
							q=&(instList[PWinstNum+pseudoCount-1]);
							q->zflag=(*zflagp);
							strcpy(q->name,strName);
							q->instLibNo=PI_LIB_NO;
							q->instID=PWinstNum+pseudoCount-1;
							q->fanin=1;
							q->maxLev=q->minLev=0;
							if(PrimInNum>=MAX_PI)
							{   fprintf(stderr,"Not enought space for PI & PSEUDO_PI\n");
								fprintf(stderr,"Larger matrix needed.\n");
								exit(1);
							}
							SourceP[PrimInNum++]=q;

							/* link fanout instances*/
							linkOutput(&(instList[ln]),q,(*zflagp));
							/* link fanin instance */
							q->in[0]=&(instList[ln]);
						}else{
								/*** PSEUDO PI ALREADY EXITS ***/
								if((*zflagp)==FALSE)
									q=instList[ln].out[0];
								if((*zflagp)==TRUE)
									q=instList[ln].outz[0];
							 }
					  }
					  else if(currInst->instLibNo==DF_LIB_NO ||
					  		  currInst->instLibNo==DFR_LIB_NO||
					  		  currInst->instLibNo==DFS_LIB_NO||
					  		  currInst->instLibNo==DFRS_LIB_NO)
						   {
						   		 /* not considered cases */
						   		 if((*zflagp)==TRUE||strcmp(tmp2,"Q")==0 ||
						   		 	instList[ln].instLibNo==DF_LIB_NO ||
						   		 	instList[ln].instLibNo==DFR_LIB_NO ||
						   		 	instList[ln].instLibNo==DFS_LIB_NO ||
						   		 	instList[ln].instLibNo==DFRS_LIB_NO)
						   		 {
						   		 	fprintf(stderr,
						   		 		"Currently, we conisder canonical sequential ckt model.\
						   		 		no pipeline structure.\n");
									exit(1);
								 }

						   	     if( isOutput2PO(&(instList[ln]),(*zflagp))==FALSE )
								 {
						   	     	/*pseudo PO need be inserted */
									pseudoCount++;
									strNameTmp=intToStr(pseudoCount);
									strcpy(strName,"PSEUDO");
									strcat(strName,strNameTmp);
									free(strNameTmp);

								 	/*pseudo PO*/
								 	q=&(instList[PWinstNum+pseudoCount-1]);
								 	q->zflag=(*zflagp);
								 	strcpy(q->name,strName);
								 	q->instLibNo=PO_LIB_NO;
								 	q->instID=PWinstNum+pseudoCount-1;
								 	q->fanin=1;
								 	q->fanout=0;
								 	if(PrimOutNum>=MAX_PO)
								 	{
								 		fprintf(stderr,"Not enought space for PO & PSEUDO_PO\n");
								 		fprintf(stderr,"Larger matrix needed.\n");
								 		exit(1);
									}
								 	SinkP[PrimOutNum++]=q;

								 	linkOutput(&(instList[ln]),q,(*zflagp));
								 	q->in[0]=&(instList[ln]);
								 }
								 q=&(instList[ln]);

						   }else q=&(instList[ln]);

 				/* for a real gate, it's always FALSE */
 				/* because we insert one pseudo-PI between QZ output and its fanout */
				*zflagp=FALSE;
				return q;
			}
}

/*--------------------------------------------------------------------------*/
/** fanout & fanin # for each instance in the circuit **/

/** NOTE: flip-flops' fanout # is its fanout # in circuit without any PSEUDO PI inserted **/
void
PWlabelCirFanOut_FanIn()
{
	int i,j,k;

	/* pseudo-PIs are included */
	for(i=0;i<PrimInNum;i++)
	{	for(j=0;j<MAX_FAN_OUT;j++)
		{	if(SourceP[i]->out[j]==NULL)
			break;
		}
		for(k=0;k<MAX_FAN_OUT;k++)
		{	if(SourceP[i]->outz[k]==NULL)
			break;
		}
		(SourceP[i])->fanout=j+k;
	}

    for(i=0;i<PWlineNum;i++)
    {   if(instList[i].instLibNo==DF_LIB_NO ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO)
		{	if(instList[i].out[0]!=NULL)
				j=instList[i].out[0]->fanout;
			else
				j=0;
			if(instList[i].outz[0]!=NULL)
				k=instList[i].outz[0]->fanout;
			else
				k=0;
			instList[i].fanout=j+k;
			/*for DF type, fanin is always 1*/
			instList[i].fanin=1;
		}else{
        		/* fanin number */
        		for(j=0;j<MAX_IN_PER_GATE;j++)
        		{
        			if(instList[i].in[j]==NULL)
        			break;
				}
				instList[i].fanin=j;

        		/* fanout number */
        		for(j=0;j<MAX_FAN_OUT;j++)
        		{   if(instList[i].out[j]==NULL)
            		break;
        		}
        		for(k=0;k<MAX_FAN_OUT;k++)
        		{   if(instList[i].outz[k]==NULL)
            		break;
        		}
        		instList[i].fanout=j+k;
			 }
    }

}

/*-------------------------------------------------------------------------*/

/* clear the visited flag to FALSE */
void
clearVisitedFlag()
{
	int	i;

	for(i=0;i<=PWinstNum;i++)
		instList[i].visited=FALSE;
}

void
reportCirStatistics(dbgFp, realPInum, realPOnum)
FILE * dbgFp;
long	realPInum,realPOnum;
{
	fprintf(dbgFp,"The circuit statistics is:\n");
	fprintf(dbgFp,"Primary input number(not including PSEUDO PI)=%ld\n",
	                realPInum);
	fprintf(dbgFp,"Primary output number(not including PSEDO PO)=%ld\n",
					realPOnum);
}

/* the next available input pin */
int
getNewInputLinkIndex(lineNo)
int	lineNo;
{
	int j;

	for(j=0;j<MAX_IN_PER_GATE;j++)
	{
		if(instList[lineNo].in[j]==NULL)
    		break;
    }
    if(j>=MAX_IN_PER_GATE)
    {
    	fprintf(stderr,"Fan-in # exceeds the maximum number. Use larger matrix.\n");
    	exit(1);
	}
	return j;
}

/* detect if output has been linked to a pseudo PO */
BOOLEAN
isOutput2PO(pInst,zflag)
PW_INST	*pInst;
BOOLEAN	zflag;
{
	int 	i;
	BOOLEAN	flag=FALSE;

	if(zflag==FALSE)
	{
		for(i=0;i<MAX_FAN_OUT;i++)
		{
			if(pInst->out[i] == NULL)
				break;

			/* currently, PO and pseudo-PO have the same lib No. */
			if(pInst->out[i]->instLibNo == PO_LIB_NO)
			{
				flag=TRUE;
				break;
			}
		}
	}else{
			for(i=0;i<MAX_FAN_OUT;i++)
			{
				if(pInst->outz[i] == NULL)
					break;

			 	/* currently, PO and pseudo-PO have the same lib No. */
				if(pInst->outz[i]->instLibNo == PO_LIB_NO)
				{
					flag=TRUE;
					break;
				}
			}
		 }

	return flag;
}

/* get the pre-processing info for reading netlist later:
   # of instances, # of flip-flops
*/
void
getPreProcessInfo(fin,PWlineNum,DF_Num)
FILE 	*fin;
long	*PWlineNum;
long	*DF_Num;
{
	int i=0,j=-1;
	char strbuff[MAX_STR_BUF];

	do{
		fscanf(fin,"%s",strbuff);
		fscanf(fin,"%s",strbuff);
		if(strcmp(strbuff,"DF")==0 ||
			strcmp(strbuff,"DFR")==0 ||
			strcmp(strbuff,"DFS")==0 ||
			strcmp(strbuff,"DFRS")==0)
		{
			i++;
		}
			j++;
	}while(toNextLine(fin));

	*PWlineNum=j;
	*DF_Num=i;
}
