#include "PWcir.h"

#include "PWroutines.h"

#include "logicOpt.h"
extern 	BOOLEAN	backwardImplyNOT(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern  BOOLEAN backwardImplyAND(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	BOOLEAN backwardImplyNAND(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	BOOLEAN backwardImplyOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	BOOLEAN backwardImplyNOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	BOOLEAN backwardImplyXOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	BOOLEAN	backwardImplyXNOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);
extern	void 	inputScan_U(PW_INST *pInst, int *count, int *index, BOOLEAN *allOne,
							BOOLEAN *allZero, BOOLEAN *noOne, BOOLEAN *noZero);
extern	void 	pushProcessElem(stk_rec *stack, PW_INST *inst, LOGIC_VALUE value);
extern  BOOLEAN backwardImplyBUFF(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack);

/* ONE CIRCUIT LEVEL IMPLICATION */
/* gate <pInst> has been assigned with value 'assignment' */
/* all backward implied gates and their implied values are pushed into <stack> */
BOOLEAN
backwardImplication(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
		BOOLEAN	conflictFg;

		switch(pInst->instLibNo)
		{
			case  NOT_LIB_NO:
				conflictFg=backwardImplyNOT(pInst,assignment,stack);
				break;
			case BUFF_LIB_NO:
				conflictFg=backwardImplyBUFF(pInst,assignment,stack);
				break;
			case AND_LIB_NO:
				conflictFg=backwardImplyAND(pInst,assignment,stack);
				break;
			case NAND_LIB_NO:
				conflictFg=backwardImplyNAND(pInst,assignment,stack);
				break;
			case OR_LIB_NO:
				conflictFg=backwardImplyOR(pInst,assignment,stack);
				break;
			case NOR_LIB_NO:
				conflictFg=backwardImplyNOR(pInst,assignment,stack);
				break;
			case XOR_LIB_NO:
				conflictFg=backwardImplyXOR(pInst,assignment,stack);
				break;
			case XNOR_LIB_NO:
				conflictFg=backwardImplyXNOR(pInst,assignment,stack);
				break;

			case PI_LIB_NO:
			case PO_LIB_NO:
				/* no backward implication */
				/* and the pseudo-PI will block the propagation for DFF*/
				conflictFg=FALSE;
				break;
			default:
				fprintf(stderr,"Unexpected Gate Type for %s, ID %d.\n",
							pInst->name,pInst->instID);
				exit(1);
		}
		return conflictFg;
}


BOOLEAN
backwardImplyNOT(PW_INST *pInst, LOGIC_VALUE assignment,stk_rec *stack)
{
	BOOLEAN			conflict;
	LOGIC_VALUE		value;

	value=(*((LOGIC_VALUE *)(pInst->in[0]->implyStkTop->member)));
	if( value==VALUE_UNKNOWN)
	{
		pushProcessElem(stack,pInst->in[0],not_operation(assignment));
		conflict=FALSE;
	}else if( value==assignment)
		  {	/* conflicts: because of redundant circuits */
			conflict=TRUE;
		  }else{	/* no implication */
		  			conflict=FALSE;
			   }

	return conflict;
}

BOOLEAN
backwardImplyBUFF(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
	BOOLEAN         conflict;
	LOGIC_VALUE     value;

	value=(*((LOGIC_VALUE *)(pInst->in[0]->implyStkTop->member)));
	if( value==VALUE_UNKNOWN)
	{
		pushProcessElem(stack,pInst->in[0],assignment);
		conflict=FALSE;
	}else if( value== not_operation(assignment))
		  {
		  	conflict=TRUE;
		  }else{
		  		conflict=FALSE;
			   }

	return conflict;
}

BOOLEAN
backwardImplyAND(PW_INST *pInst, LOGIC_VALUE assignment,stk_rec *stack)
{
	int		i;
	int		count;
	int		index;
	BOOLEAN	allOne,allZero,noOne,noZero;

	/* one case */
	if(assignment==VALUE_1)
	{
		for(i=0;i<pInst->fanin;i++)
		{
			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_0)
				return TRUE;

			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_UNKNOWN)
			{
				pushProcessElem(stack,pInst->in[i],VALUE_1);
			}
		}
	}else{	/* another case: when assigning 0 */
			inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);

			if(allOne==TRUE)	/* conflict */
				return TRUE;
			/* if only one UNKNOWN value ,need implication*/
			if(count==1 && noZero==TRUE)
			{
				pushProcessElem(stack,pInst->in[index],VALUE_0);
			}
		}
	return FALSE;
}

BOOLEAN
backwardImplyNAND(PW_INST *pInst, LOGIC_VALUE assignment,stk_rec *stack)
{
	int		i;
	int		count;
	int		index;
	BOOLEAN allOne,allZero,noOne,noZero;

	/* one case */
	if(assignment==VALUE_0)
	{
		for(i=0;i<pInst->fanin;i++)
		{
			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_0)
				return TRUE;

			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_UNKNOWN)
			{
				pushProcessElem(stack,pInst->in[i],VALUE_1);
			}
		}
	}else{	/* another case: when assigning 1 */
			inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);

			if(allOne==TRUE)	/* conflict */
				return	TRUE;
			/* if only one UNKNOWN value, need implication */
			if(count==1 && noZero==TRUE)
			{
				pushProcessElem(stack,pInst->in[index],VALUE_0);
			}
		 }
	return FALSE;
}

BOOLEAN
backwardImplyOR(PW_INST* pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
	int		i;
	int		count;
	int     index;
	BOOLEAN allOne,allZero,noOne,noZero;

	/* one case */
	if(assignment==VALUE_0)
	{
		for(i=0;i<pInst->fanin;i++)
		{
			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_1)
				return TRUE;

			if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_UNKNOWN)
			{
				pushProcessElem(stack,pInst->in[i],VALUE_0);
			}
		}
	}else{	/* another case: when assigning 1 */
			inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);

			if(allZero==TRUE)	/* conflict */
				return TRUE;
			/* if only one UNKNOWN value */
			if(count==1 && noOne==TRUE)
			{
				pushProcessElem(stack,pInst->in[index],VALUE_1);
			}
		 }
	return FALSE;
}

BOOLEAN
backwardImplyNOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
	int     i;
    int     count;
    int     index;
    BOOLEAN allOne,allZero,noOne,noZero;

    /* one case */
    if(assignment==VALUE_1)
    {
    	for(i=0;i<pInst->fanin;i++)
    	{
    		if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_1)
    			return TRUE;

    		if( (*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)))==VALUE_UNKNOWN)
    		{
    			pushProcessElem(stack,pInst->in[i],VALUE_0);
			}
		}
	}else{	/* another case: when assigning 0 */
			inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);

			if(allZero==TRUE)
				return TRUE;	/* conflict */

			/* if only one UNKNOWN value */
			if(count==1 && noOne==TRUE)
			{
				pushProcessElem(stack,pInst->in[index],VALUE_1);
			}
		 }
	return FALSE;
}

/* only for XOR2 */
BOOLEAN
backwardImplyXOR(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
	int		count;
	int 	index;
	BOOLEAN allOne,allZero,noOne,noZero;

	inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);

	/* one case */
	if(assignment==VALUE_0)
	{
		/* conflict */
		if(count==0 && allOne==FALSE && allZero==FALSE)
			return TRUE;

		if(count==1)
		{
			pushProcessElem(stack,pInst->in[index],(*((LOGIC_VALUE *)(pInst->in[1-index]->implyStkTop->member))) );
		}
	}else{	/* another case: when assigning 1 */

			/* conflict */
			if(count==0 && ( allOne==TRUE || allZero==TRUE) )
				return TRUE;

			if(count==1)
			{
				pushProcessElem(stack,pInst->in[index],
								not_operation( (*((LOGIC_VALUE *)(pInst->in[1-index]->implyStkTop->member))) ) );
			}
		 }
	return FALSE;
}

/* only for XNOR2 */
BOOLEAN
backwardImplyXNOR(PW_INST *pInst, LOGIC_VALUE assignment,stk_rec *stack)
{
	int		count;
	int		index;
	BOOLEAN allOne,allZero,noOne,noZero;

	inputScan_U(pInst,&count,&index,&allOne,&allZero,&noOne,&noZero);
	/* one case */
	if(assignment==VALUE_0)
	{
		/* conflict */
		if(count==0 && ( allOne==TRUE || allZero==TRUE) )
			return TRUE;

		if(count==1)
		{
			pushProcessElem(stack,pInst->in[index],not_operation( (*((LOGIC_VALUE *)(pInst->in[1-index]->implyStkTop->member))) ) );
		}
	}else{	/* another case: when assigning 1 */
			if(count==0 && allOne==FALSE && allZero==FALSE)
				return TRUE;

			if(count==1)
			{
				pushProcessElem(stack,pInst->in[index],(*((LOGIC_VALUE *)(pInst->in[1-index]->implyStkTop->member))) );
			}
		 }
	return FALSE;
}

/* scan the input pins of gate
   get the UNKNOWN Value number and the last one's index
*/
void
inputScan_U(pInst,count,index,allOne,allZero,noOne,noZero)
PW_INST		*pInst;
int			*count;
int			*index;
BOOLEAN		*allOne,*allZero,*noOne,*noZero;
{
	int			i,j,k;
	LOGIC_VALUE	tmpValue;

	(*count)=0;
	j=k=0;

	for(i=0;i<pInst->fanin;i++)
    {
		tmpValue=(*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)));
        if(tmpValue==VALUE_UNKNOWN)
        {
        	(*count)++;
            (*index)=i;
		}else if(tmpValue==VALUE_1)
			  	j++;
			  else	k++;
	}

	if(j==pInst->fanin)
		(*allOne)=TRUE;
	else	(*allOne)=FALSE;

	if(j==0)
		(*noOne)=TRUE;
	else	(*noOne)=FALSE;

	if(k==pInst->fanin)
		(*allZero)=TRUE;
	else	(*allZero)=FALSE;

	if(k==0)
		(*noZero)=TRUE;
	else	(*noZero)=FALSE;

}

/* allocate a processing element, and push into stack */
void
pushProcessElem(stk_rec *stack,PW_INST *inst, LOGIC_VALUE value)
{
	PW_IMPLY_STRU   *processElem;

	processElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
	processElem->inst=inst;
	processElem->value=value;
	STK_PUSH(stack,(double)NOT_CARE_KEY,(pointer)processElem);
}

/* get the logic value from Implication Stack */
LOGIC_VALUE
fetchImplyValueInStk(pInst)
PW_INST	*pInst;
{
	return (*((LOGIC_VALUE *)(pInst->implyStkTop->member)));
}
