#include "PWcir.h"
#include "linear.h"
#include "linearFree.h"
#include "linearExt.h"

#include "PWroutines.h"
#include "PWglobal_var.h"

extern void 		checkBucketIndexRange(int index);
extern void			reportBucketContent(void);
extern void 		reportBucketList(bi_lin_table_rec list);
extern void 		rebuildBucket(void);


/* establish the bucket structure */
void
PWbucketInit()
{
	int	i,k;

	PWbucketList=(PW_BUCKET *)mem_get_block((SIZE_T) ((2*MAX_ABS_GAIN+1)*sizeof(PW_BUCKET)));
	PWbucketTail=(PW_BUCKET *)mem_get_block((SIZE_T) ((2*MAX_ABS_GAIN+1)*sizeof(PW_BUCKET)));

	maxGainIndex=-1;

	k=gain2index(MAX_ABS_GAIN);

	for(i=0;i<=k;i++)
	{
		PWbucketList[i].bucketElemList=NULL;
		PWbucketTail[i].bucketElemList=NULL;
	}

	maxGainIndex=-1;	/* buckets are empty */
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		/* for PI and real gate whose output is NOT
		   determined, it has a valid gain,
		*/
		/* DFF not considered */
		if(instList[i].instLibNo!=PO_LIB_NO &&
			instList[i].instLibNo!=DF_LIB_NO &&
			instList[i].instLibNo!=DFR_LIB_NO &&
			instList[i].instLibNo!=DFS_LIB_NO &&
			instList[i].instLibNo!=DFRS_LIB_NO &&
			fetchImplyValueInStk(&(instList[i]))==VALUE_UNKNOWN)
			assignBucket(&(instList[i]),PWbucketList,PWbucketTail);
	}

	/***
	reportBucketContent();
	***/
}

/* remove all the instances in removeList from BUCKETs;
   move the instances in <reArrangeList> to a new BUCKET-->
   if the new gain equals to the old gain, nothing will be done!
*/
void
PWbucketUpdt(removeList,reArrangeList)
lin_table_rec		removeList;
bi_lin_table_rec	reArrangeList;
{
	PWbucketRemoveGarbage(removeList);
	PWbucketReArrange(reArrangeList);
}

/* currently, gain_st is not updated, and
new gain is temporarily stored in new_gain_st
*/
void
PWbucketRemoveGarbage(removeList)
lin_table_rec   removeList;
{
	bi_lin_table_rec	*head,*tail;
	lin_table_rec		p;
	PW_INST				*pInst;
	int					index,i;
	bi_lin_table_rec	rec;
	PW_BUCKET_ELEM		*elem;

	p=removeList;
	while(p!=NULL)
	{
		pInst=((PW_INST *)(p->member));
		if(pInst->instLibNo != PO_LIB_NO)
		{
		  for(i=0;i<2;i++)
		  {

			if(pInst->gain_st[i]==INVALID_GAIN)
				index=gain2index(0-MAX_ABS_GAIN);
			else
				index=gain2index(pInst->gain_st[i]); 
			checkBucketIndexRange(index);

			head=&(PWbucketList[index].bucketElemList);
			tail=&(PWbucketTail[index].bucketElemList);

			rec=pInst->pGainUpdt_st[i];
			bi_lin_table_arb_del(head,tail,rec);
			elem=(PW_BUCKET_ELEM *)(rec->member);
			mem_free_block((pointer)elem);
			mem_free_block((pointer)rec);
			pInst->pGainUpdt_st[i]=NULL;

			/* update max gain index */
			if((*head)==NULL && maxGainIndex==index)
			{
				maxGainIndex=nextMaxGainIndex(maxGainIndex);
			}
		  }
		}
		p=next_lin_table_rec(p);
	}
}

/* currently, gain_st is not updated, and
new gain is temporarily stored in new_gain_st
*/
/* gain_st is used to locate the bucket element before updating */
void
PWbucketReArrange(reArrangeList)
bi_lin_table_rec	reArrangeList;
{
	bi_lin_table_rec    *head,*tail;
	bi_lin_table_rec	p;
    PW_INST             *pInst;
    int                 index,i;
    bi_lin_table_rec    rec;
    BOOLEAN				invalidFg;

	p=reArrangeList;
	while(p!=NULL)
	{
		for(i=0;i<2;i++)
		{

			pInst=((PW_INST *)(p->member));
			/* if we really need to rearrange it */
			if(pInst->gain_st[i]!=pInst->new_gain_st[i])
			{
				/* detach from the old bucket */
				index=gain2index(pInst->gain_st[i]);
				head=&(PWbucketList[index].bucketElemList);
				tail=&(PWbucketTail[index].bucketElemList);

				rec=pInst->pGainUpdt_st[i];
				bi_lin_table_arb_del(head,tail,rec);

				/* update max gain index */
				if((*head)==NULL && maxGainIndex==gain2index(pInst->gain_st[i]))
				{
					maxGainIndex=nextMaxGainIndex(maxGainIndex);
				}

				/* put into the new bucket */
				if(pInst->new_gain_st[i]==INVALID_GAIN)
				{
					invalidFg=TRUE;
					index=gain2index(0-MAX_ABS_GAIN);/* the last bucket */
				}
				else
				{
					invalidFg=FALSE;
					index=gain2index(pInst->new_gain_st[i]);
				}
				checkBucketIndexRange(index);

				head=&(PWbucketList[index].bucketElemList);
				tail=&(PWbucketTail[index].bucketElemList);


				/* update max gain index */
				if((*head)==NULL)
				{
				 if(maxGainIndex==-1)
					maxGainIndex=index;
				 else
				    maxGainIndex=DF_MAX(maxGainIndex,index);
				}


				if(invalidFg==FALSE)
				pInst->pGainUpdt_st[i]=
				bi_lin_table_prev_ins(head,tail,(double)NOT_CARE_KEY,(pointer)(rec->member));
				else
				pInst->pGainUpdt_st[i]=
				bi_lin_table_post_ins(head,tail,(double)NOT_CARE_KEY,(pointer)(rec->member));
				mem_free_block((pointer)rec);

				/* update gain */
				pInst->gain_st[i]=pInst->new_gain_st[i];
			}
		}
		p=next_bi_lin_table_rec(p);
	}
}



/* convert index in bucket list into gain value */
/* index ranges from 0 to 2*MAX_ABS_GAIN */
/* gain ranges from -MAX_ABS_GAIN to MAX_ABS_GAIN */
int
index2gain(index)
int	index;
{
	return (index-MAX_ABS_GAIN);
}

/* convert gain value into index in bucket list */
int
gain2index(gain)
int	gain;
{
	return (gain+MAX_ABS_GAIN);
}


/* assign a bucket for the instance, and put it in */
void
assignBucket(pInst,PWbucketList,PWbucketTail)
PW_INST		*pInst;
PW_BUCKET	*PWbucketList;
PW_BUCKET	*PWbucketTail;
{
	int				i;
	int				index;
	PW_BUCKET_ELEM	*elem;
	bi_lin_table_rec *head,*tail;
	BOOLEAN			invalidFg;

	if(pInst->gain_st[0]==INVALID_GAIN &&
	   pInst->gain_st[1]==INVALID_GAIN)
	{
		fprintf(stderr,"Try to assign bucket to %s whose output is determined.\n",
				pInst->name);
		exit(1);
	}


	for(i=0;i<2;i++)
	{
		if(pInst->gain_st[i]==INVALID_GAIN)
		{
			invalidFg=TRUE;
			index=gain2index(0-MAX_ABS_GAIN);
		}else{
				invalidFg=FALSE;
				index=gain2index(pInst->gain_st[i]);
			 }
		checkBucketIndexRange(index);

		head=&(PWbucketList[index].bucketElemList);
		tail=&(PWbucketTail[index].bucketElemList);

		elem=(PW_BUCKET_ELEM *)mem_get_block((SIZE_T) sizeof(PW_BUCKET_ELEM));
		elem->instID=pInst->instID;
		elem->assignment=int2logic(i);


		/* update the max gain index */
		if((*head)==NULL)
		{
			if(maxGainIndex==-1)
				maxGainIndex=index;
			else
				maxGainIndex=DF_MAX(maxGainIndex,index);
		}

		if(invalidFg==FALSE)
		pInst->pGainUpdt_st[i]=
		bi_lin_table_prev_ins(head,tail,(double)NOT_CARE_KEY,(pointer)elem);
		else
		pInst->pGainUpdt_st[i]=
		bi_lin_table_post_ins(head,tail,(double)NOT_CARE_KEY,(pointer)elem);
	}
}

int
nextMaxGainIndex(currIndex)
int	currIndex;
{
	int	i;

	for(i=currIndex-1;i>=0;i--)
	{
		if(PWbucketList[i].bucketElemList!=NULL)
			break;
	}
	return i;	/* if the all buckets are empty, -1 is returned */
}

void
checkBucketIndexRange(index)
int	index;
{
	if(index>2*MAX_ABS_GAIN || index<0)
	{
		fprintf(stderr,"Not enough gain range!\n");
    	exit(1);
	}
}

PW_BUCKET_ELEM *
accessMaxBucketElem(bucketlist,maxGainIndex)
PW_BUCKET   *bucketlist;
int         maxGainIndex;
{
    bi_lin_table_rec    rec;
    rec=bucketlist[maxGainIndex].bucketElemList;

    return ((PW_BUCKET_ELEM *)(rec->member));
}

void
reportBucketContent()
{
	int	i;

	if(maxGainIndex==-1)
	{
		fprintf(dbgFp,"All buckets are empty.\n");
	}else{
			for(i=maxGainIndex;i>=0;i--)
			{
				if(PWbucketList[i].bucketElemList!=NULL)
				{
					fprintf(dbgFp,"Bucket content are:\n");
					reportBucketList(PWbucketList[i].bucketElemList);
				}
			}
		 }
}

void
reportBucketList(list)
bi_lin_table_rec	list;
{
	bi_lin_table_rec	p;
	int					instID;
	LOGIC_VALUE			assignment;

	p=list;

	while(p!=NULL)
	{
		instID=((PW_BUCKET_ELEM *) (p->member))->instID;
		assignment=((PW_BUCKET_ELEM *) (p->member))->assignment;
		fprintf(dbgFp,"instance=%s\tassignemnt=%d\n",
				instList[instID].name,logic2int(assignment));

		p=p->next;
	}
}

/* update the buckets accrodint to the new gate gains */
void
updtBucketWholeCircuit(addList,removeList,reArrangeList,PWbucketList,PWbucketTail)
lin_table_rec		addList;
lin_table_rec		removeList;
bi_lin_table_rec	reArrangeList;
PW_BUCKET   		*PWbucketList;
PW_BUCKET   		*PWbucketTail;
{
	lin_table_rec       p;
	PW_INST				*pInst;

	PWbucketUpdt(removeList,reArrangeList);

	p=addList;
	while(p!=NULL)
	{
		pInst=((PW_INST *)(p->member));
		assignBucket(pInst,PWbucketList,PWbucketTail);

		p=next_lin_table_rec(p);
	}
}

/* in case that the bucket is out of order */
void
rebuildBucket()
{
	int	i,k;

	maxGainIndex=-1;
	k=gain2index(MAX_ABS_GAIN);
	/* warning: we don't free memory, and it may leads to memory leakage */
	/*
	for(i=0;i<=k;i++)
	{
		bi_lin_table_rec_Free(&(PWbucketList[i].bucketElemList),
							  &(PWbucketTail[i].bucketElemList));
	}
	*/

	maxGainIndex=-1;
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
			instList[i].instLibNo!=DF_LIB_NO &&
			instList[i].instLibNo!=DFR_LIB_NO &&
			instList[i].instLibNo!=DFS_LIB_NO &&
			instList[i].instLibNo!=DFRS_LIB_NO &&
			fetchImplyValueInStk(&(instList[i]))==VALUE_UNKNOWN)
			assignBucket(&(instList[i]),PWbucketList,PWbucketTail);
	}
}
