#include "PWcir.h"
#include "linearExt.h"
#include "PWroutines.h"

extern void initRoots(int rootsArray[], long totalNum, PW_INST ** PO_list, long PO_num);
extern lin_table_rec MFFCmember(PW_INST* instanceList, long instanceNum,
					 		    lin_table_rec * pMFFClistHead, lin_table_rec * pMFFClistTail,
								int currMFFC_No);
extern void initMFFCflag(PW_INST * instanceList, long instanceNum);
extern void labelRootGate(int rootsArray[], int rootsNum, PW_INST* instanceList, long instanceNum);
extern void backwardColoring(PW_INST *pInst, PW_INST *MFFCroot);
extern void labelMFFC_No(lin_table_rec mffcGateList, int currMFFC_No);
extern void restoreClusterFlag(PW_INST * instanceList, int instanceNum);
extern void reportMFFC(lin_table_rec MFFCrootList, lin_table_rec nonMFFCrootList,
					   long POnum, long instanceNum,
				 	   lin_table_rec MFFClistHead, lin_table_rec MFFClistTail);
extern void separateRootGates(PW_INST *instanceList, int instanceNum, 
							  lin_table_rec *pMFFCrootList, lin_table_rec *pnonMFFCrootList);
extern void getNewRoots(int *roots, int *rootsNum, int instanceNum, lin_table_rec MFFClist_head,
						int startMFFC, int endMFFC);
extern void discardMFFC(lin_table_rec MFFClist_head, int startMFFC, int endMFFC);
extern void newRootGateFromPO(PW_INST ** POlist, int POnum);

extern FILE *dbgFp;
extern char *PWdeviceName[];
extern int	numberofMFFC;
extern long DF_Num;

/* MFFC clustering */
void
PWclusterMFFC(instanceList,instanceNum,POlist,POnum,pMFFClistHead,pMFFClistTail,pMFFCrootList,pnonMFFCrootList)
PW_INST			* instanceList;						/* list of all instances */
long	  		instanceNum;						/* # of instances */
PW_INST			** POlist;							/* list of POs */
long	  		POnum;								/* # of POs */
lin_table_rec	* pMFFClistHead, * pMFFClistTail;	/* list of MFFCs */
lin_table_rec	* pMFFCrootList, * pnonMFFCrootList;
{
	int				*roots;							/* for the inst ID of MFFC root gates */
	int				rootsNum;
	int				startMFFC=0, endMFFC=-1; 
	int				i,j;
	lin_table_rec	mffcGates;

	(*pMFFClistHead)=(*pMFFClistTail)=NULL;
	(*pMFFCrootList)=(*pnonMFFCrootList)=NULL;
	roots=(int *)mem_get_block((SIZE_T) instanceNum*sizeof(int));

	/*get the initial set of MFFC root gates */
	initRoots(roots,instanceNum,POlist,POnum); 
	/* number of root gates */
	rootsNum=POnum;					
	initMFFCflag(instanceList,instanceNum);

	/* label the root gate flag */
	labelRootGate(roots,rootsNum,instanceList,instanceNum);

	while(rootsNum>0)
	{
		for(i=0;i<rootsNum;i++)
		{
			for(j=0;j<rootsNum;j++)
			{
				if(j!=i)
				{
					backwardColoring(&(instanceList[roots[j]]),
									 &(instanceList[roots[i]]));
				}
			}
			/* a new MFFC */
			endMFFC++;
			mffcGates=MFFCmember(instanceList,instanceNum,
								 pMFFClistHead,pMFFClistTail,endMFFC);
			if(mffcGates==NULL)
			{
				fprintf(stderr,"Error:MFFC not found!\n");
				exit(1);
			}
			/* label the MFFC # to MFFC gates */
			labelMFFC_No(mffcGates,endMFFC);
			/* restore clustering flag of all the gates that is not discarded */
			restoreClusterFlag(instanceList,instanceNum);
		}

		/* get new root gates into root array and label the root gate flag */
		getNewRoots(roots,&rootsNum,instanceNum,(*pMFFClistHead),
					startMFFC,endMFFC);
		/* discard the found MFFC gates */
		discardMFFC((*pMFFClistHead),startMFFC,endMFFC);

		startMFFC=endMFFC+1;
	}

	numberofMFFC = endMFFC+1; /* total # of MFFCs */
	if(numberofMFFC > MAX_NUM_MFFC-1)
	{
		fprintf(stderr,"We need larger MAX_NUM_MFFC.\n");
		exit(1);
	}

	/* Althogh PI can be treated as a root gate, */
	/* PO is not treated as a root gate ->*/
	/* Therefore, we have to get new root gates from POs */
	newRootGateFromPO(POlist,POnum);

	/* put root gates and non-root gates into diff. lists */
	separateRootGates(instanceList,instanceNum,pMFFCrootList,pnonMFFCrootList);


	/* check and report MFFC clustering result */
	reportMFFC((*pMFFCrootList),(*pnonMFFCrootList),POnum,instanceNum,(*pMFFClistHead),(*pMFFClistTail));
}

/* get the initial set of MFFC root gates */
void
initRoots(rootsArray,totalNum,PO_list,PO_num)
int		rootsArray[];	/* for the instance ID of the root gate */
long	totalNum;
PW_INST ** PO_list;
long    PO_num;
{
	int i;

	for(i=0;i<totalNum;i++)
	{
		rootsArray[i]=-1;
	}

	for(i=0;i<PO_num;i++)
	{
		rootsArray[i]=PO_list[i]->instID;
	}
}

void
initMFFCflag(instanceList,instanceNum)
PW_INST * instanceList;
long      instanceNum;
{
	int i;

	for(i=0;i<instanceNum;i++)
	{
		instanceList[i].MFFC_clustering_flag=MFFC_INITIAL;
		instanceList[i].MFFC_No=-1;
		instanceList[i].MFFC_root=FALSE;
	}
}

/* label root gates flag according to rootsArray */
void
labelRootGate(rootsArray,rootsNum,instanceList,instanceNum)
int			rootsArray[];
int			rootsNum;
PW_INST 	* instanceList;
long      	instanceNum;
{
	int	i;

	if(rootsNum>instanceNum)
	{
		fprintf(stderr,"Number of MFFC root gates is larger than number of all the gates.\n");
		exit(1);
	}
	for(i=0;i<rootsNum;i++)
	{
		instanceList[rootsArray[i]].MFFC_root=TRUE;
	}
}

/* coloring the reachable gates in the backward signal flow direction of the MFFC root gate */
void
backwardColoring(pInst,MFFCroot)
PW_INST	*pInst;
PW_INST	*MFFCroot;	/* the selected MFFC root gate */
{
	queue_rec		head;
	queue_rec		tail;
	lin_table_rec	tmpRec;
	PW_INST			*tmpInst;
	int				i;

	head=tail=NULL;
	if(pInst->MFFC_clustering_flag==MFFC_DISCARDED)
	{
		fprintf(stderr,"Error in backwardColoring().\n");
		exit(1);
	}

	/* this root gate feeds to another root gate directly or indirectly */
	if(pInst->MFFC_clustering_flag==MFFC_COLORED)
		return;

	QUEUE_PUSH(&head,&tail,(double)NOT_CARE_KEY,(pointer)pInst);

	while(head!=NULL)
	{
		tmpRec=QUEUE_POP(&head,&tail);
		tmpInst=(PW_INST *)(tmpRec->member);
		mem_free_block((pointer)tmpRec);

		tmpInst->MFFC_clustering_flag=MFFC_COLORED;

		if(tmpInst->instLibNo == PI_LIB_NO)
			continue;

		for(i=0;i<tmpInst->fanin;i++)
		{
			if(tmpInst->in[i]->MFFC_clustering_flag==MFFC_DISCARDED)
			{
				/*It's possible that the MFFC whose root gate has smaller
				 circuit level # is clustered before those with larger level # */
				continue;
			}

			/* cut the MFFCroot gate's fanouts, and continue coloring */
			if(tmpInst->in[i]->instID!=MFFCroot->instID &&
			   tmpInst->in[i]->MFFC_clustering_flag==MFFC_INITIAL)
				QUEUE_PUSH(&head,&tail,(double)NOT_CARE_KEY,(pointer)tmpInst->in[i]);
		}
	}
}

/* get the list of gates belonging to one MFFC, and insert into the MFFC list */
lin_table_rec
MFFCmember(instanceList,instanceNum,pMFFClistHead,pMFFClistTail,currMFFC_No)
PW_INST     	* instanceList;
long        	instanceNum;
lin_table_rec * pMFFClistHead, * pMFFClistTail;
int				currMFFC_No;
{
	int	i;

	lin_table_rec gateList=NULL;

	for(i=0;i<instanceNum;i++)
	{
		if(instanceList[i].MFFC_clustering_flag==MFFC_INITIAL &&
			instanceList[i].instLibNo!=DF_LIB_NO &&
			instanceList[i].instLibNo!=DFR_LIB_NO &&
			instanceList[i].instLibNo!=DFS_LIB_NO &&
			instanceList[i].instLibNo!=DFRS_LIB_NO)
			STK_PUSH(&gateList, (double)NOT_CARE_KEY, (pointer) &(instanceList[i]));
	}

	if(gateList==NULL)
	{
		fprintf(stderr,"Error in MFFCmember().\n");
		fprintf(stderr,"At least one MFFC member should be found.\n");
		exit(1);
	}

	QUEUE_PUSH(pMFFClistHead,pMFFClistTail,(double)currMFFC_No,(pointer)gateList);
	return gateList;
}

/* label the MFFC No. */
void
labelMFFC_No(mffcGateList,currMFFC_No)
lin_table_rec 	mffcGateList;
int				currMFFC_No;
{
	lin_table_rec currRec;

	currRec=mffcGateList;
	while(currRec!=NULL)
	{
		((PW_INST *)(currRec->member))->MFFC_No=currMFFC_No;
		currRec=next_lin_table_rec(currRec);
	}
}

void
restoreClusterFlag(instanceList,instanceNum)
PW_INST	* instanceList;
int		instanceNum;
{
	int	i;

	for(i=0;i<instanceNum;i++)
	{
		if(instanceList[i].MFFC_clustering_flag==MFFC_COLORED)
			instanceList[i].MFFC_clustering_flag=MFFC_INITIAL;
	}
}

/* get new roots into root array, and label the root gate flag */
void
getNewRoots(roots,rootsNum,instanceNum,MFFClist_head,startMFFC,endMFFC)
int     		*roots;
int				*rootsNum;
int				instanceNum;
lin_table_rec	MFFClist_head;
int				startMFFC,endMFFC;
{
	lin_table_rec	gateList,currMFFC,currGate;
	int				i,j;
	PW_INST			* pInst;

	for(i=0;i<instanceNum;i++)
	{
		roots[i]=-1;
	}

	currMFFC=MFFClist_head;
	j=0;

	while(currMFFC!=NULL)
	{
		if((int)(currMFFC->key)>=startMFFC && (int)(currMFFC->key)<=endMFFC)
		{
			gateList=((lin_table_rec)(currMFFC->member));
			currGate=gateList;
			while(currGate!=NULL)
			{
				pInst=(PW_INST *)(currGate->member);
				for(i=0;i<pInst->fanin;i++)
				{
					/* new root gates */
					if(pInst->in[i]->instLibNo!=DF_LIB_NO &&
						pInst->in[i]->instLibNo!=DFR_LIB_NO &&
						pInst->in[i]->instLibNo!=DFS_LIB_NO &&
						pInst->in[i]->instLibNo!=DFRS_LIB_NO &&
						pInst->in[i]->MFFC_No<0 && 
						pInst->in[i]->MFFC_root==FALSE)
					{
						roots[j]=pInst->in[i]->instID;
						pInst->in[i]->MFFC_root=TRUE;
						j++;
					}
				}
				currGate=next_lin_table_rec(currGate);
			}
		}
		currMFFC=next_lin_table_rec(currMFFC);
	}
	(*rootsNum) = j;
}

/* label the found MFFC gates as discarded */
void
discardMFFC(MFFClist_head,startMFFC,endMFFC)
lin_table_rec   MFFClist_head;
int             startMFFC,endMFFC;
{
	lin_table_rec   gateList,currMFFC,currGate;
	PW_INST			* pInst;

	currMFFC=MFFClist_head;

	while(currMFFC!=NULL)
	{
		if((int)(currMFFC->key)>=startMFFC && (int)(currMFFC->key)<=endMFFC)
		{
			gateList=((lin_table_rec)(currMFFC->member));
			currGate=gateList;
			while(currGate!=NULL)
			{
				pInst=(PW_INST *)(currGate->member);
				pInst->MFFC_clustering_flag=MFFC_DISCARDED;

				currGate=next_lin_table_rec(currGate);
			}
		}
		currMFFC=next_lin_table_rec(currMFFC);
	}
}

/* put root gates and non-root gates into different lists */
void
separateRootGates(instanceList,instanceNum,pMFFCrootList,pnonMFFCrootList)
PW_INST			*instanceList;
int				instanceNum;
lin_table_rec 	*pMFFCrootList,*pnonMFFCrootList;
{
	int	i;

	for(i=0;i<instanceNum;i++)
	{
		if(instanceList[i].instLibNo==DF_LIB_NO ||
			instanceList[i].instLibNo==DFR_LIB_NO ||
			instanceList[i].instLibNo==DFS_LIB_NO ||
			instanceList[i].instLibNo==DFRS_LIB_NO)
			continue;

		if(instanceList[i].MFFC_root==TRUE)
		{
			if(instanceList[i].instLibNo==PO_LIB_NO)
			{
				/* PO and DFF are not counted in */
				continue;

			}else{
					STK_PUSH(pMFFCrootList,(double)NOT_CARE_KEY,
							 (pointer)(&(instanceList[i])));
				 }
		}else{
				STK_PUSH(pnonMFFCrootList,(double)NOT_CARE_KEY,
						 (pointer)(&(instanceList[i])));
			 }
	}
}

/* check and report clustering result */
void
reportMFFC(MFFCrootList,nonMFFCrootList,POnum,instanceNum,MFFClistHead,MFFClistTail)
lin_table_rec   MFFCrootList,nonMFFCrootList;
long			POnum,instanceNum;
lin_table_rec	MFFClistHead,MFFClistTail;
{
	PW_INST			*pInst;
	lin_table_rec 	currGate;
	long			i=0,j=0,k,l;
	lin_table_rec	tmpRec;
	BOOLEAN			reconvergentFg;
	int				num_of_this_size[MAX_NUM_MFFC],counter;

	fprintf(dbgFp,"\n");
	fprintf(dbgFp,"*********************************************\n");
	fprintf(dbgFp,"The MFFC clustering result:\n");
	fprintf(dbgFp,"MFFC root gates:\n");
	currGate=MFFCrootList;

	while(currGate!=NULL)
	{
		pInst=(PW_INST *)(currGate->member);
		fprintf(dbgFp,"Name=%s\t Type=%s\n",pInst->name,
				PWdeviceName[pInst->instLibNo-1]);
		currGate=next_lin_table_rec(currGate);
		i++;
	}
	fprintf(dbgFp,"---------------------------------------------\n");
	fprintf(dbgFp,"non-MFFC-root gates:\n");
	currGate=nonMFFCrootList;

	while(currGate!=NULL)
	{
		pInst=(PW_INST *)(currGate->member);
		fprintf(dbgFp,"Name=%s\t Type=%s\n",pInst->name,
					PWdeviceName[pInst->instLibNo-1]);
		currGate=next_lin_table_rec(currGate);
		j++;
	}

	fprintf(dbgFp,"---------------------------------------------\n");

	if(instanceNum==i+j+POnum+DF_Num)
	{
		fprintf(dbgFp,"The clustering result passed the simple check:\n");
		fprintf(dbgFp,"instance number matches.\n\n");
	}else{
			fprintf(stderr,"The clustering result failed the simple check:\n");
			fprintf(stderr,"instance number doesn't match!!.\n\n");
			exit(1);
		 }
	fprintf(dbgFp,"*********************************************\n");

	fprintf(dbgFp,"MFFC statistics:\n\n");
	fprintf(dbgFp,"# of MFFC = %ld\n\n",i);

	tmpRec=MFFClistHead;
	k=0;
	for(counter=0;counter<MAX_NUM_MFFC;counter++)
		num_of_this_size[counter]=0;

	while(tmpRec!=NULL)
	{
		k++;	/* NO. of MFFC */
		currGate=(lin_table_rec) (tmpRec->member);
		l=0;	/* # of gates within a MFFC */
		reconvergentFg=FALSE;
		while(currGate!=NULL)
		{
			l++;
			pInst=(PW_INST *)(currGate->member);
			if(pInst->MFFC_root==FALSE && pInst->fanout>1)
				reconvergentFg=TRUE;
			currGate=next_lin_table_rec(currGate);
		}

		fprintf(dbgFp,"MFFC %3ld: size=%ld gates, reconvergent path exists?=%s\n",k,l,((reconvergentFg==TRUE)?("YES"):("NO")));
		if(l>MAX_NUM_MFFC)
		{
			fprintf(stderr,"Larger array needed to report MFFC.\n");
			exit(1);
		}
		num_of_this_size[l]++;
		tmpRec=next_lin_table_rec(tmpRec);
	}
	fprintf(dbgFp,"##################################################\n");
	for(counter=0;counter<MAX_NUM_MFFC;counter++)
	{
		if(num_of_this_size[counter]!=0)
			fprintf(dbgFp,"MFFC size=%d\t# of MFFC=%d\n",counter,num_of_this_size[counter]);
	}
	fprintf(dbgFp,"\n");
}

/* PO is not counted into root gates of an MFFC */
/* so the gate feeding PO should be used as the root gate */
void
newRootGateFromPO(POlist,POnum)
PW_INST         ** POlist;
int             POnum;
{
	int	i;

	for(i=0;i<POnum;i++)
	{
		if(POlist[i]->in[0]==NULL)
        {
           fprintf(stderr,"Error in newRootGateFromPO().\n");
           fprintf(stderr,"PO should have one input.\n");
 	    }

		POlist[i]->in[0]->MFFC_root=TRUE;
	}
}
