/*****************************************************************************
 * FILE:        PWcom_imax_powerup.cc                                          *
 * DESCR:       Maximum Current Estimation Package                           *
 * AUTH:        Fei Li                                                       *
 * DATE:        July 2001                                                    *
 * NOTE:        Nov. 2001 Updated and integrated in SIS                      *
 *****************************************************************************/

#define SIS

#include <stdio.h>
#include "DFfile.h"
#include "sis.h"
#include "PWcir.h"
#include "DFfile.h"
#include "DFstring.h"
#include "PWerrorMsg.h"
#include "PWroutines.h"

#include "PWinitData.h"
#include "PWmemcpy_sis.h"
#include "PWilib.h"

extern int imax_powerup_command(network_t **network, int argc, char **argv);


static int imaxup_read_flag = 0;
static int imaxup_vector_flag = 0;
static int imaxup_rilib_flag = 0;

int
imax_powerup_command(network_t **network, int argc, char **argv)
{
	char str_buff[128];
	char fname[200],real_fname[200];
	static char old_fname[200]; //static var for file name
	char fout_name[300];
	char spice_fname[300];
	char dbg_fname[300];
	int	 prog_func, pos;
	int	 flag;
	FILE *fin=NULL, *sp_fout=NULL, *ilibFp=NULL;
	float turn_on_time=-1;

	if(argc <= 1)
	{
		fprintf(stderr,"No options was given.\n"); 
		fprintf(stderr,"Use \"help\" to get available options.\n");
		return 1;
	}
	strcpy(str_buff,argv[1]);
	if(strcmp(str_buff, "read")==0)
		prog_func = NETLIST_READ;
	else if(strcmp(str_buff, "rilib")==0)
		prog_func = ILIB_READ;
	else if(strcmp(str_buff, "release")==0)
		prog_func = NETLIST_RELEASE;
	else if(strcmp(str_buff,"gain-c")==0)
		prog_func = ATPG_BASED_ESTIMATION_GAIN_C;
	else if(strcmp(str_buff,"sim")==0)
		prog_func = LOGIC_SIM_1VEC;
	else if(strcmp(str_buff,"swsim")==0)
		prog_func = SW_RANDOM_SIM;
    else if(strcmp(str_buff, "MCsim")==0)
		prog_func = MC_LOGIC_SIM_1VEC;
	else if(strcmp(str_buff, "gain")==0)
		prog_func = ATPG_BASED_ESTIMATION_GAIN;
	else if(strcmp(str_buff, "fanout")==0)
		prog_func = ATPG_BASED_ESTIMATION_FANOUT;
	else if(strcmp(str_buff, "atpg_ilib")==0)
		prog_func = ATPG_BASED_ESTIMATION_WITH_ILIB;
	else if(strcmp(str_buff, "exportSp")==0)
		prog_func = EXPORT_SPICE;
	else if(strcmp(str_buff, "readIntel")==0)
		prog_func = INTEL_NETLIST_READ;
	else if(strcmp(str_buff, "exhaustiveSpiceSim")==0)
		prog_func = EXHAUSTIVE_SPICE_SIM;
	else{
			fprintf(stderr,"Unknown options.\n");
			return 1;
		}

	switch(prog_func)
	{
		case NETLIST_READ:
		case INTEL_NETLIST_READ:
			flag = imaxup_read_flag;
			if(flag != 0)
			{
				fprintf(stderr,"You have already read a netlist file\
						%s via <read> option in imaxup tool.\n",
						old_fname);
				fprintf(stderr,"If you want to read a new one, \
						release this one firsts.\n\n");
				return 1;
			}
			if(argc<=2) /* data structure transform from SIS */
			{
				fprintf(stdout,"<read> option is used, but no file is given.\n");
				fprintf(stdout,"Let's try direct data transferring from sis.\n\n");

				strcpy(fname, (*network)->net_name);
				strcpy(fout_name, fname);
				strcat(fout_name, RESULT_FILE_SUF);
				resultFp = fopen(fout_name, "w");
                                if(resultFp == NULL)
                                {
                                	fprintf(stderr,"Cannot open result output file.\n");
                                        return 1;
                                }
				strcpy(dbg_fname,DEBUG_OUT_FILE);
				strcat(dbg_fname,".");
				strcat(dbg_fname,fname);
                                dbgFp=fopen(dbg_fname,"w");
                                if(dbgFp == NULL)
                                {
                                        fprintf(stderr,"Cannot open debugging output file.\n");
                                        return 1;
                                }
           
				if(PWmemcpy_sis(*network) == 1) /* action of data structure transform */
				{
					NetRelease(fin, resultFp, dbgFp);
					fprintf(stderr,"Direct data transferring from sis failed.\n");
					fprintf(stderr,"Solutions: a> read a circuit into SIS.\n");
					fprintf(stderr,"Solutoins: b> map the circuit into basic library with\
						\nonly and, nand, or, nor, xor2, xnor2, buf, inv,pi,po,\
						vdd, gnd, before transferring\n\n");
					return 1;
				}

				/* levelize  a successfully read netlist */
				PWformLevList();
				fprintf(stdout,"Success!\n");
				imaxup_read_flag = 1;
				strcpy(old_fname, fname);
				PWdump_ckt(dbgFp);

			}else{ /* read netlist from file */
				strcpy(fname,argv[2]);
				fin=fopen(fname,"r");
				if(fin == NULL)
				{
					fprintf(stderr,"Cannot open netlist file %s.\n",fname);
					return 1;
				}else{
					/* result file */
					strcpy(fout_name, fname);
					pos=findLastSpecChar(fout_name,'/');
					while(pos!=-1)
					{
						strCutBehindSpecChar(fout_name, '/');
						pos = findLastSpecChar(fout_name, '/');
					}
					strcpy(real_fname,fout_name);
					strcat(fout_name, RESULT_FILE_SUF);
					resultFp = fopen(fout_name, "w");
					if(resultFp == NULL)
					{
						fprintf(stderr,"Cannot open result output file.\n");
						return 1;
					}
					/*File for debugging information output*/
					strcpy(dbg_fname,DEBUG_OUT_FILE);
					strcat(dbg_fname,".");
					strcat(dbg_fname,real_fname);
					dbgFp=fopen(dbg_fname,"w");
					if(dbgFp == NULL)
					{
						fprintf(stderr,"Cannot open debugging output file.\n");
						return 1;
					}

					/* action of reading */
					if(prog_func == NETLIST_READ)
						NetRead(fin);
					else if(prog_func == INTEL_NETLIST_READ)
						IntelNetRead(fin);
					else{
						fprintf(stderr,"unknown program function requested.\n");
				 		return 1;
					    }
					/* levelize */
					PWlabelLevel();
					PWchkLevAssump();
					PWformLevList();

					strcpy(old_fname, fname);
					imaxup_read_flag = 1;
				     }
			     }
			break;
		case ILIB_READ: /* read power-up current library */
			if(imaxup_rilib_flag != 0)
			{
				fprintf(stderr, "ILIB file has already been read.\n");
				return 1;
			}
			FOPEN(ilibFp, "imaxup.ilib", "r");
			if(dbgFp == NULL)
			{
				FOPEN(dbgFp, "ilib_debug.info", "w");
				PWilibRead(ilibFp, dbgFp);
				fclose(dbgFp);
			}else PWilibRead(ilibFp, dbgFp);
			imaxup_rilib_flag = 1;
			break;
		case NETLIST_RELEASE: /* release a netlist from memory */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				fprintf(stderr,"No old netlist file has been read or\
								in-memory copied from SIS part yet.");
				/*cout << "No old netlist file has been read yet." << endl;*/
				return 1;
			}else{
					NetRelease(fin, resultFp, dbgFp);
					strcpy(fname,"");
					imaxup_read_flag = 0;
			     }
			break;
		case ATPG_BASED_ESTIMATION_GAIN_C: /* Imax/gain-c */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				ERROR_MSG(NO_FILE_READ_ERROR);
				/*
				cout << "You have to read a netlist file first before running." << endl;*/
				return 1;
			}else{
					PWimax_gain_c();
					imaxup_vector_flag=1;
			     }
			break;
		case ATPG_BASED_ESTIMATION_GAIN: /* Imax/gain */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				ERROR_MSG(NO_FILE_READ_ERROR);
				/* cout << "You have to read a netlist file first before running." << endl;*/
				return 1;
			}else{
					PWimax_gain();
					imaxup_vector_flag=1;
				 }
			break;
		case ATPG_BASED_ESTIMATION_FANOUT: /* Imax/fanout */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				ERROR_MSG(NO_FILE_READ_ERROR);
				/*cout << "You have to read a netlist file first before running." << endl;*/
				return 1;
			}else{
					PWimax_fanout();
					imaxup_vector_flag=1;
				 }
			break;
		case ATPG_BASED_ESTIMATION_WITH_ILIB: /* Timed ATPG */
			if(imaxup_read_flag == 0 || imaxup_rilib_flag == 0)
			{
				ERROR_MSG(NO_ILIB_OR_NET_READ_ERROR);
				return 1;
			}else{
					if(argc < 3) { fprintf(stderr,"Turn-on time maybe not specified.\n"); return 1; }
			 		sscanf(argv[2],"%f", &turn_on_time);
					if(PWimax_ilib(turn_on_time)==1) return 1;
					imaxup_vector_flag=1;
				 }
			break;
		case LOGIC_SIM_1VEC: /* 1-vector logic sim for power-up current */
			flag = imaxup_read_flag ;
            if(flag == 0)
            {
				ERROR_MSG(NO_FILE_READ_ERROR);
            	/*cout << "You have to read a netlist file first before running." << endl;*/
                return 1; 
            }else{  
                		PWrand_vec_simple_sim(imaxup_rilib_flag, 100);
				 }
			break;
		case SW_RANDOM_SIM: /* random logic sim (2vector) for swiching current */
			flag = imaxup_read_flag ;
            if(flag == 0)
            {
				ERROR_MSG(NO_FILE_READ_ERROR);
                /*cout << "You have to read a netlist file first before running." << endl;*/
                return 1; 
            }else{
                 	PWnormSwitchingSim();
                 }

			break;
		case MC_LOGIC_SIM_1VEC: /* monte carlo 1-vector logic sim */
			flag = imaxup_read_flag ;
            if(flag == 0)
            {
				ERROR_MSG(NO_FILE_READ_ERROR);
                /*cout << "You have to read a netlist file first before running." << endl;*/
                return 1; 
            }else{  
            		fprintf(stderr,"Currently not supported!");
            		/*cout << "Currently not suppored!" << endl;*/
                	return 1;
                 }
		case EXPORT_SPICE: /* export the gate netlist to a spice netlist */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				ERROR_MSG(NO_FILE_READ_ERROR);
                /*cout << "You have to read a netlist file first before exporting." << endl;*/
				return 1;
			}else{				
					strcpy(spice_fname,old_fname);
					pos=findLastSpecChar(spice_fname,'/');
					while(pos!=-1)
					{
						strCutBehindSpecChar(spice_fname, '/');
						pos = findLastSpecChar(spice_fname, '/');
					}
						
					strcat(spice_fname,SPICE_FILE_SUF);
					sp_fout=fopen(spice_fname,"w");
					if(sp_fout == NULL)
					{
						fprintf(stderr,"Cannot export to file %s.\n",
										spice_fname);
						/*cout << "Cannot export to file " << spice_fname << endl;*/
						return 1;
					}
					fprintf(stdout,"exporting spice netlist with power gating...\n");
					/*cout << "exporting spice netlist with power gating...";*/
					PWexportSp(sp_fout, imaxup_vector_flag);
					fprintf(stdout,"Done.\n");
					/*cout << "  done" << endl;*/
					fclose(sp_fout);
				 }
			
			break;
		case EXHAUSTIVE_SPICE_SIM: /* exhaustive spice simulation */
			flag = imaxup_read_flag ;
			if(flag == 0)
			{
				ERROR_MSG(NO_FILE_READ_ERROR);
				/*cout << "You have to read a netlist file first before exporting." << endl;*/
				return 1;
			}else{
					strcpy(spice_fname,old_fname);
                    pos=findLastSpecChar(spice_fname,'/');
                    while(pos!=-1)
                    {
                        strCutBehindSpecChar(spice_fname, '/');
                        pos = findLastSpecChar(spice_fname, '/');
                    }

					strcat(spice_fname,SPICE_FILE_SUF);
					//exhaustive_spice_powerup(spice_fname);
					//exhaustive_spice_switch(spice_fname);
					random_spice_powerup(spice_fname,0);
				}
			break;
		default:
			fprintf(stderr,"Unknown program function. Debug it.\n");
			/*
			cerr << "Unknown program function. Debug it." << endl;
			*/
			return 1;
			break;
	}

	/* the command is successfully executed */
	return 0;
}

int get_imaxup_read_flag() { return imaxup_read_flag; }

int get_imaxup_rilib_flag() { return imaxup_rilib_flag; }
