#include <stdio.h>
#include "PWcir.h"
#include <string.h>
#include "DFtype.h"
#include "DFstring.h"
#include "DFrandom.h"
#include "PWglobal_var.h"
#include "PWroutines.h"
#include "linearExt.h"
#include "linearFree.h"
#include <unistd.h>

extern void PWexportSp(FILE* fout, int vectorFlag);
extern double exhaustive_spice_switch(char *base_file_name, int typeNum);
extern double valueInSpiceListFile(char *lis_fname,char *var, int *errFg, int *errNum);
extern void printVectorInStack(stk_rec stack);
extern void storeVectorFromStack(LOGIC_VALUE *vec, stk_rec stack);
extern void first4vec(LOGIC_VALUE *vec4,int typeNum);
extern void PWexportSp_switch(FILE* fout, int vectorFlag);
extern double random_spice_powerup(char *base_file_name, int typeNum);
extern void storeVectorFromPI(LOGIC_VALUE *vec);
extern void printVectorFromPI(LOGIC_VALUE *vec);
extern void exhaustive_spice_file_generator(char *base_file_name, int typeNum);



/* values of the first 4 PIs are determined by the typeNum */
double
exhaustive_spice_powerup(char *base_file_name, int typeNum)
{

	stk_rec			stack1, stack2;
	LOGIC_VALUE     *pValue;
	int				i,j;
	FILE			*fout;
	char			sp_fname[MAX_STR_BUF], spice_command[MAX_STR_BUF];
	char			lis_fname[MAX_STR_BUF];
	double			Imax_power_up=0, current;
	stk_rec			tmpRec;
	PW_INST	*		tmpPI;
	LOGIC_VALUE		imax_powerup_vec[MAX_PI];
	LOGIC_VALUE		vec4[4];
	int				errNum=0,errFg;

	j=0;
	stack1=stack2=NULL;

	/*PWimplyStackGarbageClean();*/
	PWimplyStackRelease();

	/* including assignment to PI inputs VCC/VSS */
	initImplyStk();


	if(typeNum<0 || typeNum>=16)
	{
		fprintf(stderr,"type number for exhaustive spice simulation is out of normal range.\n");
		exit(1);
	}

	/* first 4 PIs */
	first4vec(vec4,typeNum);

	/* initial stack, excluding VCC/VSS */
	for(i=4;i<PrimInNum;i++)
	{
		if(strcmp(SourceP[i]->name,"VSS")!=0 &&
		   strcmp(SourceP[i]->name,"VCC")!=0)
		{
			/* initial vector */
			pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
			(*pValue)=VALUE_0;
			STK_PUSH(&(SourceP[i]->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
			STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)SourceP[i]);
		}
	}

	/* for debugging */
	fprintf(stdout,"\n\n\nExhaustive Vector List for power-up current.\n");
	printVectorInStack(stack1);

	/* exhaustive simulation for power-up current */
	do{
		j++;
		/* spice netlist file name */
		strcpy(sp_fname,base_file_name);
		strcat(sp_fname,intToStr(typeNum));
		strcat(sp_fname,"_");
		strcat(sp_fname,intToStr(j));
		/* list file name for spice simu output */
		strcpy(lis_fname,sp_fname);
		strcat(lis_fname,".lis");

		fout=fopen(sp_fname,"w");
		if(fout == NULL)
		{
			fprintf(stderr,"Cannot create spice netlist file %s.\n",
							sp_fname);
			return -1.0;
		}
		/* export spice netlist with sleep transistor */
		PWexportSp(fout, 1);
		fflush(fout);
		fclose(fout);

		strcpy(spice_command,"/afs/engr.wisc.edu/apps/bin/hspice ");
		strcat(spice_command,sp_fname);
		strcat(spice_command," > ");
		strcat(spice_command,lis_fname);

		/* spice simulation */
		system(spice_command);

		/* power-up current (positive value) */
		current= valueInSpiceListFile(lis_fname, "imax_powerup=", &errFg, &errNum);
		if(errFg == 1 && errNum <= 20 )
		{	unlink(sp_fname);
			unlink(lis_fname);
			j--;
			continue;
		}
		if(errFg == 1 && errNum > 20)
		{
			exit(1);
		}
		Imax_power_up=DF_MAX(Imax_power_up, current);
		/*fprintf(stdout,"the current for this run: %le\n",current);*/

		/* store the vector that generates max power-up current so far */
		if(Imax_power_up == current)
		{
			storeVectorFromStack(imax_powerup_vec, stack1);
			fprintf(stdout,"imax so far: %le\n",Imax_power_up);
			printVectorInStack(stack1);
			fflush(stdout);
		}


		/* delete old files */
		unlink(sp_fname);
		unlink(lis_fname);

		/* next possible vector */

		tmpRec=STK_POP(&stack1);
		tmpPI=(PW_INST *)(tmpRec->member);
		mem_free_block((pointer)tmpRec);

		if(fetchImplyValueInStk(tmpPI) == VALUE_0)
		{

			stackFree(&(tmpPI->implyStkTop));
			pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
			(*pValue)=VALUE_1;
			STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
			STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)tmpPI);
		}else{
				/* push to stack2 temporarily */
				while(tmpPI!=NULL && fetchImplyValueInStk(tmpPI) == VALUE_1)
				{
					stackFree(&(tmpPI->implyStkTop));
					pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
					(*pValue)=VALUE_0;
					STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
					STK_PUSH(&stack2,(double)NOT_CARE_KEY,(pointer)tmpPI);

					tmpRec=STK_POP(&stack1);
					if(tmpRec!=NULL)
					{
						tmpPI=(PW_INST *)(tmpRec->member);
						mem_free_block((pointer)tmpRec);
					}else tmpPI=NULL;
				}

				/* transfer from stack2 to stack1 */
				if(tmpPI!=NULL)
				{
					stackFree(&(tmpPI->implyStkTop));
					pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
					(*pValue)=VALUE_1;
					STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
					STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)tmpPI);

					stackTransfer(&stack1, &stack2);
				}
			 }
		/*printVectorInStack(stack1);*/

	}while(stack1!=NULL);

	stackFree1(&stack1);
	stackFree1(&stack2);

	fprintf(stdout,"Exhaustive spice simulation for power-up current.\n");
	fprintf(stdout,"\tThe maximum power-up current is %eamp\n", 
					 Imax_power_up);

	fprintf(stdout,"The corresponding input vector is as follows:\n");
	for(i=4;i<PrimInNum;i++)
	{
		fprintf(stdout,"PI_name=%s, value=%d\n",
						  SourceP[i]->name,
						  logic2int(imax_powerup_vec[i]));
	}

	fflush(stdout);
	return Imax_power_up;
}

double
valueInSpiceListFile(char *lis_fname,char *var, int* errFg, int *errNum)
{
	FILE* 	flist;
	char  	str[MAX_STR_BUF];
	double	value;

	FOPEN(flist, lis_fname, "r");

	fscanf(flist,"%s", str);
	while(!feof(flist) && strcmp(str,var)!=0)
	{
		fscanf(flist,"%s", str);
	}

	if(feof(flist)==0)
	{
		/* not end-of-file */
		fscanf(flist,"%le",&value);
		fclose(flist);
		(*errFg)=0;
		(*errNum)=0;
		return DF_ABS(value);
	}else{
			fprintf(stderr,"variable %s not found in list file %s.\n",
							var, lis_fname);
			fprintf(stderr,"Something wrong with the spice netlist.\n");
			(*errNum)++;
			(*errFg)=1;
			return 0;
			/*exit(1);*/
		 }
}

void
printVectorInStack(stk_rec stack)
{
	stk_rec 	tmpRec;
	PW_INST 	*pInst;
	LOGIC_VALUE val;

	tmpRec = stack;
	while(tmpRec != NULL)
	{
		pInst = (PW_INST *) tmpRec->member;
		val = fetchImplyValueInStk(pInst);

		fprintf(stdout,"%d", logic2int(val));
		tmpRec = next_lin_table_rec(tmpRec);
	}
	fprintf(stdout,"\n");
	fflush(stdout);
}

void
storeVectorFromStack(LOGIC_VALUE *vec, stk_rec stack)
{
	int			i=0;
	stk_rec 	tmpRec;
	PW_INST		*pInst;
	LOGIC_VALUE	val;

	tmpRec = stack;
	while(tmpRec != NULL)
	{
	 	if(i>=MAX_PI)
	 	{
	 		fprintf(stderr, "Error in storeVectorFromStack().\n");
	 		fprintf(stderr, "Contact Programmer!.\n");
	 		exit(1);
		}
	 	pInst = (PW_INST *) tmpRec->member;
	 	val = fetchImplyValueInStk(pInst);

	 	vec[PrimInNum-i-1] = val;
	 	tmpRec = next_lin_table_rec(tmpRec);
	}
}

/* including assignment to the first 4 PIs */
void
first4vec(LOGIC_VALUE *vec4,int typeNum)
{
	int	remainder, quotient;
	int	i;

	LOGIC_VALUE *pValue, val;

	quotient = typeNum;
	for(i=0;i<4;i++)
	{
		remainder = quotient % 2;
		quotient = quotient / 2;
		vec4[3-i] = int2logic(remainder);

		pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
		(*pValue)=vec4[3-i];
		STK_PUSH(&(SourceP[3-i]->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
	}

	fprintf(stdout,"\n\n\n");
	fprintf(stdout,"First 4 PIs' logic values:\n");
	for(i=0;i<4;i++)
	{
		val = fetchImplyValueInStk(SourceP[i]);
		fprintf(stdout,"%d", logic2int(val));
	}
	fprintf(stdout,"\n\n\n");

	for(i=0;i<4;i++)
    {
        fprintf(stdout,"PI_name=%s, value=%d\n",
                          SourceP[i]->name,
						  logic2int(vec4[i]));
    }

}


/* actually, it's pseudo exhaustive */
/* currently, combinational ckt only!!! */
double 
exhaustive_spice_switch(char *base_file_name, int typeNum)
{

	int				i,j;
	FILE			*fout;
	char			sp_fname[MAX_STR_BUF], spice_command[MAX_STR_BUF];
	char			lis_fname[MAX_STR_BUF];
	double			Imax_switch=0, current;
	int				errNum=0,errFg;
	int				time_counter = 0;

	j=0;

	/*PWimplyStackGarbageClean();*/
	PWimplyStackRelease();

	/* including assignment to PI inputs VCC/VSS */
	initImplyStk();

	randomGenInit1();

	/* for debugging */
	fprintf(stdout,"\n\n\nPseudo Exhaustive (random) Vector List for switching current.\n");

	/* pseudo exhaustive (random) simulation for switching current */
	for(j=0;j<1000;j++)
	{
		if(time_counter >= 5)
		{
			PWimplyStackGarbageClean();
			time_counter = 0;
		}else time_counter++;

		rand_input_vec_gen(RAND_INPUT_VEC_ALL_PI);
		simple_sim_operation();


		for(i=0;i<1500;i++)
		{
			rand_input_vec_gen(RAND_INPUT_VEC_REAL_PI);
			simple_sim_operation();

			/* spice netlist file name */
			strcpy(sp_fname,base_file_name);
			strcat(sp_fname,"_");
			strcat(sp_fname,intToStr(j+i));
			/* list file name for spice simu output */
			strcpy(lis_fname,sp_fname);
			strcat(lis_fname,".lis");

			fout=fopen(sp_fname,"w");
			if(fout == NULL)
			{
				fprintf(stderr,"Cannot create spice netlist file %s.\n",
								sp_fname);
				return -1.0;
			}
			/* export spice netlist with sleep transistor */
			PWexportSp_switch(fout, 1);
			fflush(fout);
			fclose(fout);

			if(time_counter >=5)
			{
				PWimplyStackGarbageClean();
				time_counter = 0;
			}else time_counter++;

			strcpy(spice_command,"/afs/engr.wisc.edu/apps/bin/hspice ");
			strcat(spice_command,sp_fname);
			strcat(spice_command," > ");
			strcat(spice_command,lis_fname);

			/* spice simulation */
			system(spice_command);

			/* switching current (positive value) */
			current= valueInSpiceListFile(lis_fname, "imax_switch=", &errFg, &errNum);
			if(errFg == 1 && errNum <= 20 )
			{	unlink(sp_fname);
				unlink(lis_fname);
				i--;
				continue;
			}
			if(errFg == 1 && errNum > 20)
			{
				exit(1);
			}
			Imax_switch=DF_MAX(Imax_switch, current);
			/*fprintf(stdout,"the current for this run: %le\n",current);*/

			/* store the vector that generates max power-up current so far */
			if(Imax_switch == current)
			{
				fprintf(stdout,"imax so far: %le\n",Imax_switch);
			    for(i=0;i<PrimInNum;i++)
    			{
					if(strcmp(SourceP[i]->name,"VSS") == 0 ||
					   strcmp(SourceP[i]->name,"VCC") == 0 ||
					   strcmp(SourceP[i]->name,"GND") == 0 ||
					   strcmp(SourceP[i]->name,"VDD") == 0 )
					{
						fprintf(stdout,"PI_name=%s, value1=%d value2=%d\n", SourceP[i]->name,
								logic2int(fetchImplyValueInStk(SourceP[i])),
								logic2int(fetchImplyValueInStk(SourceP[i])));
					}else{
        			fprintf(stdout,"PI_name=%s, value1=%d value2=%d\n", SourceP[i]->name,
								logic2int(fetchImplyValueInStk(SourceP[i])), logic2int((*((LOGIC_VALUE *)(SourceP[i]->implyStkTop->next->member)))) );
						}
				}
				
				fprintf(stdout,"\n");
				fflush(stdout);
			}


			/* delete old files */
			unlink(sp_fname);
			unlink(lis_fname);

		}

	}


	fprintf(stdout,"Pseudo Exhaustive (random) spice simulation for switching current.\n");
	fprintf(stdout,"\tThe maximum switching current is %eamp\n", 
					 Imax_switch);

	fprintf(stdout,"The corresponding input vector is as follows:\n");

	fflush(stdout);

	return Imax_switch;
}

/* random simulation for power-up current */
double
random_spice_powerup(char *base_file_name, int typeNum)
{

    int             i,j;
    FILE            *fout;
    char            sp_fname[MAX_STR_BUF], spice_command[MAX_STR_BUF];
    char            lis_fname[MAX_STR_BUF];
    double          Imax_power_up=0, Imin_power_up=10000000, current;
    int             errNum=0,errFg;
    int             time_counter = 0;
	LOGIC_VALUE     Imax_powerup_vec[MAX_PI], Imin_powerup_vec[MAX_PI];


    j=0;


    /*PWimplyStackGarbageClean();*/
    PWimplyStackRelease();

    /* including assignment to PI inputs VCC/VSS */
    initImplyStk();

    randomGenInit1();

    /* for debugging */
    fprintf(stdout,"\n\n\nPseudo Exhaustive (random) Vector List for powerup current.\n");

    /* pseudo exhaustive (random) simulation for switching current */
    for(j=0;j<15000;j++)
    {
        if(time_counter >= 5)
        {
            PWimplyStackGarbageClean();
            time_counter = 0;
        }else time_counter++;

        rand_input_vec_gen(RAND_INPUT_VEC_ALL_PI);


            /* spice netlist file name */
            strcpy(sp_fname,base_file_name);
            strcat(sp_fname,"_");
            strcat(sp_fname,intToStr(j));
            /* list file name for spice simu output */
            strcpy(lis_fname,sp_fname);
            strcat(lis_fname,".lis");

            fout=fopen(sp_fname,"w");
            if(fout == NULL)
            {
                fprintf(stderr,"Cannot create spice netlist file %s.\n",
                                sp_fname);
                return -1.0;
            }
            /* export spice netlist with sleep transistor */
	        PWexportSp(fout, 1);
		 	fflush(fout);
            fclose(fout);

            strcpy(spice_command,"/afs/engr.wisc.edu/apps/bin/hspice ");
            strcat(spice_command,sp_fname);
            strcat(spice_command," > ");
            strcat(spice_command,lis_fname);

            /* spice simulation */
            system(spice_command);

            current= valueInSpiceListFile(lis_fname, "imax_powerup=",&errFg,&errNum);

			if(errFg == 1 && errNum <= 20 )
            {   unlink(sp_fname);
                unlink(lis_fname);
                i--;
                continue;
            }
            if(errFg == 1 && errNum > 20)
            {
                exit(1);
            }
			Imax_power_up=DF_MAX(Imax_power_up, current);
			Imin_power_up=DF_MIN(Imin_power_up, current);

			if(Imax_power_up == current)
			{
				storeVectorFromPI(Imax_powerup_vec);
			}
			if(Imin_power_up == current)
			{
				storeVectorFromPI(Imin_powerup_vec);
			}

			if(j>=0 && j<=100)
			{
				fprintf(stdout,"current = %le\n",current);
    			for(i=0;i<PrimInNum;i++)
    			{
					fprintf(stdout,"%d",logic2int(fetchImplyValueInStk(SourceP[i])));
    			}
				fprintf(stdout,"\n");
			}else if(j>100 && j<=600 && (j%100)==0)
				  {
						fprintf(stdout,"current per 100 vecs = %le\n",current);
						for(i=0;i<PrimInNum;i++)
						{
                    		fprintf(stdout,"%d",logic2int(fetchImplyValueInStk(SourceP[i])));
                		}
						fprintf(stdout,"\n");
						fprintf(stdout,"The Imax so far: %le\n", Imax_power_up);
						printVectorFromPI(Imax_powerup_vec);
						fprintf(stdout,"The Imin so far: %le\n", Imin_power_up);
						printVectorFromPI(Imin_powerup_vec);

				   }else if(j>600 && j<=2000 && (j%200)==0)
						{
							 fprintf(stdout,"current = per 200  %le\n",current);
                        	for(i=0;i<PrimInNum;i++)
                        	{
                            fprintf(stdout,"%d",logic2int(fetchImplyValueInStk(SourceP[i])));
                        	}
                        	fprintf(stdout,"\n");
							fprintf(stdout,"The Imax so far: %le\n", Imax_power_up);
							printVectorFromPI(Imax_powerup_vec);
                            fprintf(stdout,"The Imin so far: %le\n", Imin_power_up);
                            printVectorFromPI(Imin_powerup_vec);

						}else if((j%1000)==0)
						{
							 fprintf(stdout,"current per 1000 = %le\n",current);
                        	for(i=0;i<PrimInNum;i++)
                        	{
                            fprintf(stdout,"%d",logic2int(fetchImplyValueInStk(SourceP[i])));
                        	}
                        	fprintf(stdout,"\n");
							fprintf(stdout,"The Imax so far: %le\n", Imax_power_up);
							printVectorFromPI(Imax_powerup_vec);
	                        fprintf(stdout,"The Imin so far: %le\n", Imin_power_up);
                            printVectorFromPI(Imin_powerup_vec);

						}
				fflush(stdout);

            if(Imax_power_up == current)
            {
                fprintf(stdout,"imax so far: %le\n",Imax_power_up);
                fflush(stdout);
            }
            /* delete old files */
            unlink(sp_fname);
            unlink(lis_fname);

        }


    fprintf(stdout,"Pseudo Exhaustive (random) spice simulation for power-up current.\n");
    fprintf(stdout,"\tThe maximum power-up current is %eamp\n",
                     Imax_power_up);

    fprintf(stdout,"The corresponding input vector is as follows:\n");
    for(i=0;i<PrimInNum;i++)
    {
        fprintf(stdout,"PI_name=%s, value=%d\n",
                          SourceP[i]->name,
                          logic2int(Imax_powerup_vec[i]));
    }

	fflush(stdout);

    return Imax_power_up;
}

void
storeVectorFromPI(LOGIC_VALUE *vec)
{
	int i;

	for(i=0; i<PrimInNum;i++)
	{
		vec[i]= fetchImplyValueInStk(SourceP[i]);
	}
}

void
printVectorFromPI(LOGIC_VALUE *vec)
{
	int i;
	for(i=0; i<PrimInNum; i++)
	{
		fprintf(stdout,"%d",logic2int(vec[i]));
	}
}

/* powerup or switch */
void
exhaustive_spice_file_generator(char *base_file_name, int typeNum)
{
	stk_rec         stack1, stack2;
    LOGIC_VALUE     *pValue;
    int             i,j;
    FILE            *fout;
    char            sp_fname[MAX_STR_BUF], spice_command[MAX_STR_BUF];
    char            lis_fname[MAX_STR_BUF];
    double          Imax_power_up=0, current;
    stk_rec         tmpRec;
    PW_INST *       tmpPI;
    LOGIC_VALUE     imax_powerup_vec[MAX_PI];
    LOGIC_VALUE     vec4[4];
    int             errNum=0,errFg;

    j=0;
    stack1=stack2=NULL;

    /*PWimplyStackGarbageClean();*/
    PWimplyStackRelease();

    /* including assignment to PI inputs VCC/VSS */
    initImplyStk();

    /* initial stack, excluding VCC/VSS */
    for(i=0;i<PrimInNum;i++)
    {
        if(strcmp(SourceP[i]->name,"VSS")!=0 &&
           strcmp(SourceP[i]->name,"VCC")!=0)
        {
            /* initial vector */
            pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
            (*pValue)=VALUE_0;
            STK_PUSH(&(SourceP[i]->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
            STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)SourceP[i]);
        }
    }


    /* exhaustive simulation for power-up current */
    do{
        j++;
        /* spice netlist file name */
        strcpy(sp_fname,base_file_name);
        strcat(sp_fname,intToStr(typeNum));
        strcat(sp_fname,"_");
        strcat(sp_fname,intToStr(j));
        /* list file name for spice simu output */
        strcpy(lis_fname,sp_fname);
        strcat(lis_fname,".lis");

        fout=fopen(sp_fname,"w");
        if(fout == NULL)
        {
            fprintf(stderr,"Cannot create spice netlist file %s.\n",
                            sp_fname);
            return -1.0;
        }
        /* export spice netlist with sleep transistor */
        PWexportSp(fout, 1);
        fflush(fout);
        fclose(fout);

        strcpy(spice_command,"/afs/engr.wisc.edu/apps/bin/hspice ");
        strcat(spice_command,sp_fname);
        strcat(spice_command," > ");
        strcat(spice_command,lis_fname);

	    /* next possible vector */

        tmpRec=STK_POP(&stack1);
        tmpPI=(PW_INST *)(tmpRec->member);
        mem_free_block((pointer)tmpRec);

        if(fetchImplyValueInStk(tmpPI) == VALUE_0)
        {

            stackFree(&(tmpPI->implyStkTop));
            pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
            (*pValue)=VALUE_1;
            STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
            STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)tmpPI);
        }else{
                /* push to stack2 temporarily */
                while(tmpPI!=NULL && fetchImplyValueInStk(tmpPI) == VALUE_1)
                {
                    stackFree(&(tmpPI->implyStkTop));
                    pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
                    (*pValue)=VALUE_0;
                    STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
                    STK_PUSH(&stack2,(double)NOT_CARE_KEY,(pointer)tmpPI);

                    tmpRec=STK_POP(&stack1);
                    if(tmpRec!=NULL)
                    {
                        tmpPI=(PW_INST *)(tmpRec->member);
                        mem_free_block((pointer)tmpRec);
                    }else tmpPI=NULL;
                }

                /* transfer from stack2 to stack1 */
                if(tmpPI!=NULL)
                {
                    stackFree(&(tmpPI->implyStkTop));
                    pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
                    (*pValue)=VALUE_1;
                    STK_PUSH(&(tmpPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
                    STK_PUSH(&stack1,(double)NOT_CARE_KEY,(pointer)tmpPI);

                    stackTransfer(&stack1, &stack2);
                }
             }
        /*printVectorInStack(stack1);*/

    }while(stack1!=NULL);

    stackFree1(&stack1);
    stackFree1(&stack2);

    fprintf(stdout,"The corresponding input vector is as follows:\n");
    for(i=0;i<PrimInNum;i++)
    {
        fprintf(stdout,"PI_name=%s, value=%d\n",
                          SourceP[i]->name,
                          logic2int(imax_powerup_vec[i]));
    }

	fflush(stdout);
}
