#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "PWcir.h"
#include "PWglobal_var.h"
#include "PWroutines.h"

extern void exportOneGate(FILE*, PW_INST*);
extern void writeGateOutputNode(FILE* fout, char name[]);
extern void writeEquInputNode(FILE* fout, PW_INST* pInst);
extern void writeGateType(FILE* fout, int devNum, int fanin);
extern void spiceHeadOutput(FILE* fout);
extern void paraSubckt(FILE* fout, char* path);
extern void inputSignals(FILE* fout);
extern void inputSignals_switch(FILE* fout);
extern void zeroVolSource(FILE* fout);
extern void zeroVolSource_switch(FILE* fout);
extern void spiceAnalysis(FILE* fout);
extern void spiceAnalysis_switch(FILE* fout);
extern void PWexportSp_switch(FILE* fout, int vectorFlag);


/* with sleep transistor drivers */
static int vccFlag=0, vssFlag=0;
void
PWexportSp(FILE* fout, int vectorFlag)
{
	int 	i;
	PW_INST	*pInst;
	int		sleep_size;
	char 	*path = "/pong/usr3/f/feil/research/ImaxSpiceVerification/";

	spiceHeadOutput(fout);

	/* functional circuit */
	fprintf(fout,"\n* ckt\n");

	/* not including VDD/GND inputs */
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo==PO_LIB_NO)
			continue;

		pInst = &(instList[i]);
		exportOneGate(fout,pInst);
	}

	/* PMOS sleep transistor */
	sleep_size = (PWinstNum * 0.5) * 2;
	fprintf(fout, "\nMsleep sleep_trans_drain sleep_trans_gate VDD_node VDD_node PMOS L=trans_l W=\'%d*trans_w\' \n", sleep_size);

	/* Drviers for sleep transistor */
	/* in order to match the experimental setting for library,
	 * no drivers are used 
	 */
    fprintf(fout,"*The following is commented to match the setting for lib.\n");
	fprintf(fout,"\n* Sleep Signal Drivers\n");
	fprintf(fout,"*XDR1 dr1_out dr1_in NOT_big\n");
    fprintf(fout,"*XDR2 sleep_trans_gate dr1_out NOT_small\n");



	/* include parameters and sub-ckt */
	paraSubckt(fout,path);

	/* signals */
	if(vectorFlag == 0)
	{
		fprintf(fout, "\n");
		fprintf(fout, "* The input vector hasn't been generated.\n");
		fprintf(fout, "* So, the input signals are not generated here.\n");
	}else{
			fprintf(fout, "\n");
			fprintf(fout, "* signals\n");
			inputSignals(fout);
		 }

	/* zero voltage source for connection and monitoring current */
	zeroVolSource(fout);

	/* technology file */
	//fprintf(fout, ".INCLUDE \"%sntrs013.L3\"\n", path);
	fprintf(fout, ".INCLUDE \"%s100nm_nominal_H\"\n", path);


	/* analysis */
	spiceAnalysis(fout);

	/* end */
	fprintf(fout, "\n.END\n");
}

/* one gate per line */
void
exportOneGate(FILE* fout, PW_INST* pInst)
{
	char subckt[MAX_STR_BUF];

	/* pseudo PI */
	if( pInst->instLibNo == PI_LIB_NO && 
		(pInst->in[0]!=NULL || pInst->fanin != 0) )
		return;

	/* VDD/GND inputs */
	if( pInst->instLibNo == PI_LIB_NO &&
		(strcmp(pInst->name,"GND")==0 || strcmp(pInst->name,"VDD")==0 ||
		 strcmp(pInst->name,"VSS")==0 || strcmp(pInst->name,"VCC")==0) )
		return;

	strcpy(subckt,"X");
	strcat(subckt,pInst->name);
	fprintf(fout,"%s", subckt);
	fprintf(fout," ");

	writeGateOutputNode(fout, pInst->name);
	writeEquInputNode(fout, pInst);
	writeGateType(fout, pInst->instLibNo, pInst->fanin);
	fprintf(fout, "\n");
}

/* write the gate output node and one space */
void
writeGateOutputNode(FILE* fout, char name[])
{
	char str[MAX_STR_BUF];

	if(fout == NULL)
	{
		fprintf(stderr,"cannot write to fout, which is NULL\n");
		exit(1);
	}

	strcpy(str,name);
	strcat(str,"_O");
	fprintf(fout,"%s", str);
	fprintf(fout," ");
}

/* write the equivalent inputs of the gate and one space */
void
writeEquInputNode(FILE* fout, PW_INST* pInst)
{
	int  i;
	char str[MAX_STR_BUF];

	if(pInst->fanin  == 0) /* the gate is a real PI */
	{
		strcpy(str, pInst->name);
		strcat(str, "_I");
		fprintf(fout, "%s", str);
	}else{
			for(i=0; i<pInst->fanin; i++)
			{
				if( pInst->in[i]->instLibNo == PI_LIB_NO &&
					strcmp(pInst->in[i]->name,"VSS")==0 )
				{
					writeGateOutputNode(fout, pInst->in[i]->name);
					vssFlag = 1;
				}else
				if( pInst->in[i]->instLibNo == PI_LIB_NO &&
					strcmp(pInst->in[i]->name,"VCC")==0 )
				{
					writeGateOutputNode(fout, pInst->in[i]->name);
					vccFlag = 1;
				}else /* for pseudo PI */
				if(pInst->in[i]->instLibNo == PI_LIB_NO &&
				   pInst->in[i]->fanin != 0)
					writeGateOutputNode(fout, pInst->in[i]->in[0]->name);
				else
					writeGateOutputNode(fout, pInst->in[i]->name);
			}
		 }
	fprintf(fout, " ");
}

/* write gate type and one space */
void
writeGateType(FILE* fout, int devNum, int fanin)
{
	switch(devNum)
	{
		case NOT_LIB_NO:
		case BUFF_LIB_NO:
			fprintf(fout, "%s", PWdeviceName[devNum-1]);
			break;
		case PI_LIB_NO:
		  	fprintf(fout, "BUFF"); /* PI is actually a buffer */
			break;
		case PO_LIB_NO:
			fprintf(stderr, "PO should be ignored in exporting spice netlist.\n");
			exit(1);
			break;
		case DF_LIB_NO:
		case DFR_LIB_NO:
		case DFS_LIB_NO:
		case DFRS_LIB_NO:
			fprintf(fout, "DF"); /* currently only one implementation of FFs */
			break;
		default:
			fprintf(fout, "%s", PWdeviceName[devNum-1]);
			fprintf(fout, "%d", fanin);
	}
	fprintf(fout, " ");
}

void
spiceHeadOutput(FILE* fout)
{
	fprintf(fout, "*Spice Circuit File  generated by Imax export.\n");
}

void
paraSubckt(FILE* fout, char *path)
{
	fprintf(fout, "\n");
	fprintf(fout, ".INCLUDE \"%sparameter.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sglobal.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snand2.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snand3.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snand4.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snand5.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snot.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sand2.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sand3.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sand4.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sand5.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sor2.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sor3.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sor4.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sor5.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snor2.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snor3.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snor4.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snor5.sp\"\n",path);
	/*
	fprintf(fout, ".INCLUDE \"%sxnor2.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%sxor2.sp\"\n",path);
	*/
	fprintf(fout, ".INCLUDE \"%sbuff.sp\"\n",path);

	fprintf(fout, ".INCLUDE \"%snot_small.sp\"\n",path);
	fprintf(fout, ".INCLUDE \"%snot_big.sp\"\n",path);
}

/* input signals and VDD */
void
inputSignals(FILE* fout)
{
	float	powersupply=1.35;
	int i;
	float signal;

	fprintf(fout, "\n");
	fprintf(fout, "Vdd VDD_node 0 DC %gV\n", powersupply);
	/* changed to match the setting for library */
	/*fprintf(fout, "Vgated dr1_in 0 PWL 0ns %gV 0.01ns 0V\n", powersupply);*/

	fprintf(fout, "Vgated sleep_trans_gate 0 PWL 0ns %gV 0.1ns 0V\n",
				  powersupply);

	/* for real PIs */
	for(i=0;i<PSEUDO_PI_index;i++)
	{
		if( fetchImplyValueInStk(SourceP[i]) == VALUE_1)
			signal = powersupply;
		else
			signal = 0;
		/*
		fprintf(fout, "V_%s %s_I 0 PWL 0ns 0V 0.1ns %gV\n", 
						SourceP[i]->name,
						SourceP[i]->name, 
						signal);
		*/
		fprintf(fout, "V_%s %s_I 0 DC %gV\n",
						SourceP[i]->name,
						SourceP[i]->name, 
						signal);
	}
	/* for VDD/GND inputs */
	if(vccFlag != 0)
	{
		fprintf(fout, "V_VCC_INPUT VCC_O 0 DC %gV\n",
						powersupply);
	}
	if(vssFlag != 0)
	{
		fprintf(fout, "V_VSS_INPUT VSS_O 0 DC 0V\n");
	}
}


void zeroVolSource(FILE* fout)
{
	fprintf(fout, "\n");

	/* connect virtual gnd and real gnd */
	fprintf(fout, "Virtual_gnd_source virtual_gnd_node 0 DC 0V\n");

	/* connect virtual vdd and sleep transistor drain */
	fprintf(fout, "Virtual_vdd_source sleep_trans_drain virtual_vdd_node 0 DC 0V\n");

}

void
spiceAnalysis(FILE* fout)
{
	int i;

	fprintf(fout, "\n");
	fprintf(fout, "* Specify initial states\n");

	/* .IC */
	fprintf(fout, ".IC V(sleep_trans_drain)=0V\n");

	fprintf(fout,".IC V(dr1_out)=0V\n");
	fprintf(fout,".IC V(dr1_in)=1.35V\n");
	fprintf(fout,".IC V(sleep_trans_gate)=1.35V\n");


	fprintf(fout, "\n");
	 /* excluding pseudo PIs */
	for(i=0;i<PWinstNum;i++)
	{
		if(instList[i].instLibNo !=PO_LIB_NO)
			fprintf(fout, ".IC V(%s_O)=0V\n", instList[i].name);
	}

	fprintf(fout, ".TEMP 120\n");
	fprintf(fout, ".TRAN 1n 50n UIC\n");
	/*fprintf(fout, ".PROBE I(Virtual_vdd_source)\n");*/
	fprintf(fout, ".MEAS TRAN IMAX_POWERUP  MAX I(Virtual_vdd_source) FROM=0ns TO=50ns\n");
	fprintf(fout, ".PRINT TRAN\n");
	/*fprintf(fout, ".OPTIONS PROBE POST MEASOUT\n");*/
}


void
PWexportSp_switch(FILE* fout, int vectorFlag)
{
	int 	i;
	PW_INST	*pInst;
	int		sleep_size;
	char *path = "/pong/usr3/f/feil/research/ImaxSpiceVerification/";

	spiceHeadOutput(fout);

	/* functional circuit */
	fprintf(fout,"\n* ckt\n");

	/* not including VDD/GND inputs */
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo==PO_LIB_NO)
			continue;

		pInst = &(instList[i]);
		exportOneGate(fout,pInst);
	}

	/* PMOS sleep transistor */
	/*
	sleep_size = (PWinstNum * 0.5) * 2;
	fprintf(fout, "\nMsleep sleep_trans_drain sleep_trans_gate VDD_node VDD_node PMOS L=trans_l W=\'%d*trans_w\' \n", sleep_size);
	*/


	/* include parameters and sub-ckt */
	paraSubckt(fout,path);

	/* signals */
	if(vectorFlag == 0)
	{
		fprintf(fout, "\n");
		fprintf(fout, "* The input vector hasn't been generated.\n");
		fprintf(fout, "* So, the input signals are not generated here.\n");
	}else{
			fprintf(fout, "\n");
			fprintf(fout, "* signals\n");
			inputSignals_switch(fout);
		 }

	/* zero voltage source for connection and monitoring current */
	zeroVolSource_switch(fout);

	/* technology file */
	fprintf(fout, ".INCLUDE \"%sntrs013.L3\"\n", path);


	/* analysis */
	spiceAnalysis_switch(fout);

	/* end */
	fprintf(fout, "\n.END\n");
}

void 
zeroVolSource_switch(FILE* fout)
{
	fprintf(fout, "\n");

	/* connect virtual gnd and real gnd */
	fprintf(fout, "Virtual_gnd_source virtual_gnd_node 0 DC 0V\n");

	/* connect virtual vdd and sleep transistor drain */
	fprintf(fout, "Virtual_vdd_source VDD_node virtual_vdd_node 0 DC 0V\n");

}



void
spiceAnalysis_switch(FILE* fout)
{
	int i;

	fprintf(fout, "\n");
	fprintf(fout, "* Specify initial states\n");

	/* .IC */
	/*fprintf(fout, ".IC V(sleep_trans_drain)=0V\n");*/

	fprintf(fout, "\n");
	 /* excluding pseudo PIs */
	for(i=0;i<PWinstNum;i++)
	{
		if(instList[i].instLibNo !=PO_LIB_NO)
			fprintf(fout, ".IC V(%s_O)=%fV\n", instList[i].name, ((logic2int((*((LOGIC_VALUE*)(instList[i].implyStkTop->next->member))))==0)?0:1.35));
	}

	fprintf(fout, ".TEMP 120\n");
	fprintf(fout, ".TRAN 1n 50n UIC\n");
	/*fprintf(fout, ".PROBE I(Virtual_vdd_source)\n");*/
	fprintf(fout, ".MEAS TRAN IMAX_SWITCH  MAX I(Virtual_vdd_source) FROM=0ns TO=50ns\n");
	fprintf(fout, ".PRINT TRAN\n");
	/*fprintf(fout, ".OPTIONS PROBE POST MEASOUT\n");*/
}


void
inputSignals_switch(FILE* fout)
{
	float	powersupply=1.35;
	int 	i;
	float 	signal1, signal2;

	fprintf(fout, "\n");
	fprintf(fout, "Vdd VDD_node 0 DC %gV\n", powersupply);
	/*fprintf(fout, "Vgated sleep_trans_gate 0 PWL 0ns %gV 0.1ns 0V\n", powersupply);*/
	/* for real PIs */
	for(i=0;i<PSEUDO_PI_index;i++)
	{
		if( fetchImplyValueInStk(SourceP[i]) == VALUE_1)
			signal2 = powersupply;
		else
			signal2 = 0;

		if((*((LOGIC_VALUE *) (SourceP[i]->implyStkTop->next->member))) == VALUE_1)
			signal1 = powersupply;
		else
			signal1 = 0;

		fprintf(fout, "V_%s %s_I 0 PWL 0ns %gV 0.1ns %gV\n", 
						SourceP[i]->name,
						SourceP[i]->name, 
						signal1,
						signal2);
	}
	/* for VDD/GND inputs */
	if(vccFlag != 0)
	{
		fprintf(fout, "V_VCC_INPUT VCC_O 0 DC %gV\n",
						powersupply);
	}
	if(vssFlag != 0)
	{
		fprintf(fout, "V_VSS_INPUT VSS_O 0 DC 0V\n");
	}
}
