#include "PWcir.h"
#include "linearExt.h"

#define	ADD_TO_TRACK_STK     1
#define RM_FROM_TRACK_STK	-1
#define DO_NOTHING_TRACK_STK 0

#include "PWroutines.h"
#include "logicOpt.h"
#include "PWglobal_var.h"

extern	int 		getInDex(PW_INST *sendInst, PW_INST *recieveInst);
extern  BOOLEAN		forwardImplyPO(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack);
extern	BOOLEAN 	forwardImplyNOT(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack);
extern	BOOLEAN 	forwardImplyPrimGate(PW_INST *pInst, stk_rec *stack);
extern	void 		inputScan_U_ctrlVal(PW_INST *pInst, int *count, int *index, BOOLEAN *noCtrlVal);
extern	BOOLEAN 	forwardImplyXOR(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack);
extern	BOOLEAN 	forwardImplyXNOR(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack);
extern  int 		getCirLevFromProcessRec(lin_table_rec   rec);
extern  void 		incrLevInsProcessElem();
extern  void 		incrLevInsLinearTable(lin_table_rec *head, double key, pointer mbr);
extern  void 		incrLevInsAllFanOuts(PW_INST *pInst, lin_table_rec *list);
extern  BOOLEAN 	forwardImplyBUFF(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack);
extern  void 		keepTrackOfGates(stk_rec *trackStkTop, int trackStkFlag, PW_INST *pInst);
extern  int 		setTrackStkFlag(LOGIC_VALUE oldValue, LOGIC_VALUE newValue);
extern 	void		pushProcessElem();
extern 	int			deltaGain(LOGIC_VALUE oldvalue, LOGIC_VALUE newvalue, int fanout);

/* ONE CIRCUIT LEVEL IMPLICATION */
/* the assignment for gate <pInst> has been finished before calling this routine */
/* all forward implied gates and their implied values are pushed into <stack> */
/* now for power-up current, seq ckt is converted into a combinational ckt,
   we don't handle port QZ !!! and propagation stop at pseudo-PIs
*/
BOOLEAN
forwardImplication(PW_INST *pInst, LOGIC_VALUE assignment, stk_rec *stack)
{
	int		i;
	BOOLEAN	conflictFg=FALSE;
	PW_INST	*currInst;

	for(i=0;i<pInst->fanout;i++)
	{
		currInst=pInst->out[i];

		switch(currInst->instLibNo)
		{
			case NOT_LIB_NO:
				conflictFg=forwardImplyNOT(currInst,assignment,stack);
				break;
			case BUFF_LIB_NO:
				conflictFg=forwardImplyBUFF(currInst,assignment,stack);
				break;
			case AND_LIB_NO:
			case NAND_LIB_NO:
			case OR_LIB_NO:
			case NOR_LIB_NO:
				conflictFg=forwardImplyPrimGate(currInst,stack);
				break;
			case XOR_LIB_NO:
				conflictFg=forwardImplyXOR(currInst,assignment,stack);
				break;
			case XNOR_LIB_NO:
				conflictFg=forwardImplyXNOR(currInst,assignment,stack);
				break;
			case PI_LIB_NO:
				/* in combinational circuit, it's impossible come here */
				conflictFg=FALSE;
				break;
			case PO_LIB_NO:
				conflictFg=forwardImplyPO(currInst,assignment,stack);
				break;
			case DF_LIB_NO:
			case DFS_LIB_NO:
			case DFR_LIB_NO:
			case DFRS_LIB_NO:
				conflictFg=FALSE;
				break;
			default:
				fprintf(stderr,"Unexpected Gate Type for %s, ID %d in forwardImplication().\n", currInst->name,currInst->instID);
				exit(1);
		}

		if(conflictFg==TRUE)
			break;
	}
	return conflictFg;
}

/* get the pin index of the gate from where it gets its input signal */
/* NOTE: if multi-input have the same one output,
		 this routine is not correct & may need to be updated
*/
int
getInDex(sendInst,recieveInst)
PW_INST	*sendInst,*recieveInst;
{
	int	i;

	for(i=0;i<recieveInst->fanin;i++)
	{
		if(sendInst==recieveInst->in[i])
			return i;
	}
	fprintf(stderr,"Error in getInDex().\n");
	exit(1);
}

/* one of pInst's INPUTs has been assigned with a logic value*/
/* for Primary Output PO */
BOOLEAN
forwardImplyPO(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack)
{
	BOOLEAN     conflict=FALSE;
	LOGIC_VALUE value;

	value= *((LOGIC_VALUE *)(pInst->implyStkTop->member));
	if( value == VALUE_UNKNOWN )
	{
		pushProcessElem(stack,pInst,inputAssignment);
		conflict=FALSE;
	}else if( value != inputAssignment )
		      conflict=TRUE; /*conflicts*/
		  else  conflict = FALSE;

	return conflict;
}


/* one of pInst's INPUTs has been assigned with a logic value*/
/* for NOT type instance */
BOOLEAN
forwardImplyNOT(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack)
{
	BOOLEAN		conflict=FALSE;
	LOGIC_VALUE	value;

	value= *((LOGIC_VALUE *)(pInst->implyStkTop->member));
	if( value == VALUE_UNKNOWN )
	{
		pushProcessElem(stack,pInst,not_operation(inputAssignment));
		conflict=FALSE;
	}else if( value==inputAssignment )
		  {	/* conflicts */
		  	conflict=TRUE;
		  }else{	/* no implication */
		  			conflict=FALSE;
			   }
	return conflict;
}

/* for Buffer type instance */
BOOLEAN
forwardImplyBUFF(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack)
{
	BOOLEAN		conflict=FALSE;
	LOGIC_VALUE value;

	value= *((LOGIC_VALUE *)(pInst->implyStkTop->member));
	if( value == VALUE_UNKNOWN )
	{
		pushProcessElem(stack,pInst,inputAssignment);
		conflict=FALSE;
	}else if( value == not_operation(inputAssignment) )
		  {
		  	conflict=TRUE;
		  }else{
		  			conflict=FALSE;
			   }

	return conflict;
}

/* forward implication for other primitive gates */
BOOLEAN
forwardImplyPrimGate(pInst,stack)
PW_INST     *pInst;
stk_rec     *stack;
{
	LOGIC_VALUE	ValueBF,ValueAF;
	BOOLEAN		noCtrlVal,conflict=FALSE;
	int			count;
	int			index;

	/* value before assignment */
	ValueBF=( *((LOGIC_VALUE *)(pInst->implyStkTop->member)));
	inputScan_U_ctrlVal(pInst,&count,&index,&noCtrlVal);

	ValueAF=evaluateGate(pInst,IMPLY_STK_TYPE);

	/* one case */
	if(ValueBF==VALUE_UNKNOWN)
	{
		if(ValueAF!=VALUE_UNKNOWN)
		{	
			pushProcessElem(stack,pInst,ValueAF);
		}
		conflict=FALSE;

	}else{ 	/* another case: forward becomes backward */

		if(ValueBF==xor_operation(ctrlValue[pInst->instLibNo-1],inversion[pInst->instLibNo-1]) && count==1 && noCtrlVal==TRUE)
		  {
		  	pushProcessElem(stack,pInst->in[index],ctrlValue[pInst->instLibNo-1]);
		  	conflict=FALSE;

		  }else if(ValueAF!=VALUE_UNKNOWN && ValueAF!=ValueBF)
		  		{
		  			conflict=TRUE;
				}else	
					conflict=FALSE;
		 }
	return conflict;
}

/* scan the input pins of gate
   get the UNKNOWN Value number and the last one's index
*/
/* also collect ctrolling value infomation at inputs */
void
inputScan_U_ctrlVal(pInst,count,index,noCtrlVal)
PW_INST     *pInst;
int         *count;
int         *index;
BOOLEAN     *noCtrlVal;
{
	int     		i;
    LOGIC_VALUE    	tmpValue;

	(*count)=0;
	(*noCtrlVal)=TRUE;

	for(i=0;i<pInst->fanin;i++)
	{
		tmpValue=(*((LOGIC_VALUE *)(pInst->in[i]->implyStkTop->member)));

		if(tmpValue==VALUE_UNKNOWN)
		{
			(*count)++;
			(*index)=i;
		}else if(tmpValue==ctrlValue[pInst->instLibNo-1])
			  (*noCtrlVal)=FALSE;
	}
}

/* for XOR2 gate */
BOOLEAN
forwardImplyXOR(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack)
{
    LOGIC_VALUE    	ValueBF,ValueAF;
    BOOLEAN 	 	noCtrlVal,conflict=FALSE;
    int     		count;
    int     		index;

	/* value before assignment */
	ValueBF=( *((LOGIC_VALUE *)(pInst->implyStkTop->member)));
	inputScan_U_ctrlVal(pInst,&count,&index,&noCtrlVal);

	ValueAF=evaluateGate(pInst,IMPLY_STK_TYPE);
	/* one case */
	if(ValueBF==VALUE_UNKNOWN)
	{
		if(ValueAF!=VALUE_UNKNOWN)
		{
			pushProcessElem(stack,pInst,ValueAF);
		}
		conflict=FALSE;

	}else
	{
		/* another case: forward becomes backward */
		if(count==1)
		{
			pushProcessElem(stack,pInst->in[index],xor_operation(inputAssignment,ValueBF));
			conflict=FALSE;
		}else if(ValueAF!=VALUE_UNKNOWN && ValueAF!=ValueBF)
				conflict=TRUE;
			  else
			  	conflict=FALSE;
	}
	return conflict;
}

/* for XNOR2 gate */
BOOLEAN
forwardImplyXNOR(PW_INST *pInst, LOGIC_VALUE inputAssignment, stk_rec *stack)
{
	LOGIC_VALUE    	ValueBF,ValueAF;
    BOOLEAN 	   	noCtrlVal,conflict=FALSE;
    int     		count;
    int     		index;

	/* value before assignment */
	ValueBF=( *((LOGIC_VALUE *)(pInst->implyStkTop->member)));
	inputScan_U_ctrlVal(pInst,&count,&index,&noCtrlVal);

	ValueAF=evaluateGate(pInst,IMPLY_STK_TYPE);
	/* one case */
	if(ValueBF==VALUE_UNKNOWN)
	{
		if(ValueAF!=VALUE_UNKNOWN)
		{
			pushProcessElem(stack,pInst,ValueAF);
		}
		conflict=FALSE;

	}else
	{	/* another case */
		if(count==1)
		{
			pushProcessElem(stack,pInst->in[index],xnor_operation(inputAssignment,ValueBF));
			conflict=FALSE;
		}else if(ValueAF!=VALUE_UNKNOWN && ValueAF!=ValueBF)
				conflict=TRUE;
			  else
			  	conflict=FALSE;
	}

	return conflict;
}

/* COMPLETE implication AND ASSIGNMENT for current PI */
/* used in PODEM algorithm, 3-value: 0,1,x  */
/* if value x is used, it's a logic value assignment restoration process */
void
forwardImplyPI3V(PW_INST *targetPI, LOGIC_VALUE assignmentPI, stk_rec *trackStkTop, int *gain_MFFC_root)
/*PW_INST		*targetPI;
LOGIC_VALUE	assignmentPI;
stk_rec		*trackStkTop;
int			*gain_MFFC_root;
*/
{
    stk_rec         tmpRec;
    stk_rec   		processStk=NULL;    /* process stack top */
    PW_INST         *tmpInst;
    LOGIC_VALUE     currAssign;
    LOGIC_VALUE     *pValue;
    int				trackStkFlag;

	/* compute gain gradually, but only used for MFFC root gate */
	/* delta gain for PI */
	(*gain_MFFC_root) +=deltaGain(fetchImplyValueInStk(targetPI),
						assignmentPI,targetPI->fanout);

	trackStkFlag=setTrackStkFlag(fetchImplyValueInStk(targetPI),assignmentPI);

	/* assignmentPI could be 0,1, or x */
	pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
	(*pValue)=assignmentPI;
	STK_PUSH(&(targetPI->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);

	/* keep track of the gate which is assigned, in order to restore */
	keepTrackOfGates(trackStkTop,trackStkFlag, targetPI);
	/*STK_PUSH(trackStkTop,(double)NOT_CARE_KEY,(pointer)targetPI);*/

	incrLevInsAllFanOuts(targetPI,&processStk);


    while(processStk!=NULL)
    {
        /* get the current processing element*/
        tmpRec=STK_POP(&processStk);
        tmpInst=(PW_INST *)(tmpRec->member);

        mem_free_block((pointer)tmpRec);


        /* check if the gate has already been assigned */
        /* which is possible in RECONVERGENT circuits because
           we may have MULTIPLE process elements for ONE assignment!!
        */

		currAssign=evaluateGate(tmpInst,IMPLY_STK_TYPE);
		if(currAssign!=fetchImplyValueInStk(tmpInst))
		{
			/* compute gain, but only used for MFFC root gate */
			(*gain_MFFC_root) +=deltaGain(fetchImplyValueInStk(tmpInst),
			                        currAssign,tmpInst->fanout);

			trackStkFlag=setTrackStkFlag(fetchImplyValueInStk(tmpInst),currAssign);

        	/* push logic value for current instance */
        	pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
        	(*pValue)=currAssign;
        	STK_PUSH(&(tmpInst->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);

        	/* keep track of the gate which is assigned, in order to restore */
        	keepTrackOfGates(trackStkTop,trackStkFlag, tmpInst);
        	/*STK_PUSH(trackStkTop,(double)NOT_CARE_KEY,(pointer)tmpInst);*/

        	incrLevInsAllFanOuts(tmpInst,&processStk);
		}
    }
}


/* insert processing element in increasing order of circuit level */
/* stk_rec and lin_table_rec are actually the same type */
/*
void
incrLevInsProcessElem(list,pInst,assignment)
lin_table_rec	*list;
PW_INST			*pInst;
LOGIC_VALUE		assignment;
{
	PW_IMPLY_STRU	*processElem;

	processElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
	processElem->inst=pInst;
	processElem->value=assignment;

	incrLevInsLinearTable(list,(double)NOT_CARE_KEY, (pointer)processElem);

}
*/

/* insert all fanout gates in increasing order of circuit level */
/*  QZ of Dflip-flip is not supported!!!!!!!!!!!!*/
/* stk_rec and lin_table_rec are actually the same data type */
void
incrLevInsAllFanOuts(pInst,list)
PW_INST			*pInst;
lin_table_rec	*list;
{
	int	i;

	for(i=0;i<pInst->fanout;i++)
	{
		/* QZ is not considered, implication within comb block */
		/* PO is processed */
		if(pInst->out[i]->instLibNo==DF_LIB_NO ||
		   pInst->out[i]->instLibNo==DFR_LIB_NO ||
		   pInst->out[i]->instLibNo==DFS_LIB_NO ||
		   pInst->out[i]->instLibNo==DFRS_LIB_NO)
			continue;

		/* only for cominational circuits */
		incrLevInsLinearTable(list,(double)NOT_CARE_KEY,(pointer)(pInst->out[i]));
	}
}

/* insert a linear table with increasing order of key value */
void
incrLevInsLinearTable(head,key,mbr)
lin_table_rec	*head;
double			key;
pointer			mbr;
{
	lin_table_rec	rec,p,q;
	int				dstLev,compLev;
	rec=(lin_table_rec)mem_get_block((SIZE_T)(sizeof(struct lin_table_rec_)));

	rec->key=key;
	rec->member=mbr;
	dstLev=getCirLevFromProcessRec(rec);

	p=(*head);
	while(p!=NULL)
	{
		compLev=getCirLevFromProcessRec(p);

		if(dstLev<compLev)
			break;

		q=p;
		p=next_lin_table_rec(p);
	}

	if(p==(*head))
	{
		rec->next=(*head);
		(*head)=rec;
	}else{
			rec->next=q->next;
			q->next=rec;
		 }
}

int
getCirLevFromProcessRec(rec)
lin_table_rec	rec;
{
	int	lev;
	lev=((PW_INST *)(rec->member))->maxLev;

	return lev;
}

/* for computation of gain of root gate */
/* PODEM process should be carried on until the root gate is assigned with
   specified value and the gain is computed in an updating way */
int
deltaGain(LOGIC_VALUE oldvalue, LOGIC_VALUE newvalue, int fanout)
{
	int deltaValue=0;

	if((oldvalue!=VALUE_0 && oldvalue!=VALUE_1 && oldvalue != VALUE_UNKNOWN)||
		(newvalue!=VALUE_0 && newvalue!=VALUE_1 && newvalue != VALUE_UNKNOWN))
	{
		fprintf(stderr, "Bad logic value type in delta().\n");
		exit(1);
	}

	if(newvalue == VALUE_0)
	{
		if(oldvalue == VALUE_1)
			deltaValue=0-(2*fanout);
		else if(oldvalue == VALUE_UNKNOWN)
				deltaValue=0-fanout;

	}else if( newvalue == VALUE_1)
		  {
		  	if( oldvalue == VALUE_0)
		  		deltaValue= 2*fanout;
			else if(oldvalue == VALUE_UNKNOWN)
				deltaValue= fanout;

		  }
	return deltaValue;
}

/* in 3-value forward implication, you may restore or assign a value */
/* this routine set proper flag for proper action */
int
setTrackStkFlag(LOGIC_VALUE oldValue, LOGIC_VALUE newValue)
/*LOGIC_VALUE	oldValue, newValue;*/
{
	if(oldValue == VALUE_UNKNOWN)
		return ADD_TO_TRACK_STK;
	else if(newValue == VALUE_UNKNOWN)
			return RM_FROM_TRACK_STK;
		 else return DO_NOTHING_TRACK_STK;
}

void
keepTrackOfGates(trackStkTop,trackStkFlag, pInst)
stk_rec     *trackStkTop;
int			trackStkFlag;
PW_INST		*pInst;
{
	stk_rec	tmpRec;

	if(trackStkFlag== ADD_TO_TRACK_STK)
		STK_PUSH(trackStkTop,(double)NOT_CARE_KEY,(pointer)pInst);
	else if(trackStkFlag == RM_FROM_TRACK_STK)
		 {
		  	tmpRec=STK_POP(trackStkTop);
		  	mem_free_block((pointer)tmpRec);
		 }
	     else{}
}
