#include	"PWcir.h"
#include 	"linearExt.h"
#include	"linear.h"

#include 	"PWroutines.h"

extern int shiftMFFCrootGain(lin_table_rec MFFCrootList, lin_table_rec nonMFFCrootList, BOOLEAN initFlag);
extern int shiftMFFCrootGain_cluster(lin_table_rec   MFFCrootList, lin_table_rec nonMFFCrootList);


/*
extern void shiftAction();
extern int minGainRootGates();
extern int maxGainNonRootGates();
extern int getProperMinGain();
extern int getProperMaxGain();
extern void shiftOneGain();
*/

/* Since the root gates are to be processed before non-root gates, */
/* we only shift the gain value of all the root gates before non-root gates */
/*NOTE: only root gates whose logic values are not determined are considered */
int
shiftMFFCrootGain(MFFCrootList,nonMFFCrootList,initFlag)
lin_table_rec	MFFCrootList,nonMFFCrootList;
BOOLEAN			initFlag;
{
	int	rootMinGain,nonRootMaxGain;
	int	increment;

	/* minimum gain among all the larger gain values of the root gates */
	/* explain: because one gate has 2 gain values, we select the larger one */
	rootMinGain=minGainRootGates(MFFCrootList,initFlag);

	nonRootMaxGain=maxGainNonRootGates(nonMFFCrootList,initFlag);

	/* shift is needed */
	if(rootMinGain<=nonRootMaxGain)
	{	

		increment=nonRootMaxGain-rootMinGain+1;

		shiftAction(MFFCrootList,increment,initFlag);

	}else
		increment=0;

	return increment;
}

/* perform the gain shifting */
void
shiftAction(MFFCrootList,increment,initFlag)
lin_table_rec	MFFCrootList;
int				increment;
BOOLEAN			initFlag;
{
	PW_INST         *pInst;
	lin_table_rec	currGate;

	currGate=MFFCrootList;

	while(currGate!=NULL)
	{

		pInst=(PW_INST *)(currGate->member);

		if(fetchImplyValueInStk(pInst) == VALUE_UNKNOWN)
		{
			if(initFlag==TRUE)	/* shift the initial gain values */
			{
				shiftOneGain(&(pInst->gain_st[0]),increment);
				shiftOneGain(&(pInst->gain_st[1]),increment);
	
			}else{	/* shift the to-be-updated gain values */
					shiftOneGain(&(pInst->new_gain_st[0]),increment);
					shiftOneGain(&(pInst->new_gain_st[1]),increment);
	 	 	 	}
		}
		currGate=next_lin_table_rec(currGate);
	}
}

/* INVALID_GAIN is only a flag, and will be left unchanged */
void
shiftOneGain(pGain,increment)
int	*pGain, increment;
{
	if((*pGain)!=INVALID_GAIN)
		(*pGain)+=increment;
}

/* the minimum gain of all the larger gains within 
the two gains associated with the root gates
*/
/* because only the larger one of the two gains is
useful although we put both into the bucket
*/
int
minGainRootGates(MFFCrootList,initFlag)
lin_table_rec	MFFCrootList;
BOOLEAN			initFlag;
{
	PW_INST         *pInst;
    lin_table_rec   currGate;
    int				minGain,tmpGain;
    int				i;

    currGate=MFFCrootList;
    i=0;

    while(currGate!=NULL)
    {

    	pInst=(PW_INST *)(currGate->member);

		if(fetchImplyValueInStk(pInst) == VALUE_UNKNOWN)
		{
    		if(i==0)
    		{
				/* get the min gain based on the larger gain associated with
			   	one gate
				*/
				minGain=getProperMinGain(pInst,initFlag);
	
    		}else{
					tmpGain=getProperMinGain(pInst,initFlag);
					minGain=DF_MIN(minGain,tmpGain);
				}
    		i++;
		}

    	currGate=next_lin_table_rec(currGate);
	}

	return minGain;
}

/* the max gain among all the non-root gates */
int
maxGainNonRootGates(nonMFFCrootList,initFlag)
lin_table_rec   nonMFFCrootList;
BOOLEAN         initFlag;
{
	PW_INST         *pInst;
	lin_table_rec   currGate;
	int				maxGain,tmpGain;
	int				i;

	currGate=nonMFFCrootList;
	i=0;

	while(currGate!=NULL)
	{

		pInst=(PW_INST *)(currGate->member);

		if(fetchImplyValueInStk(pInst) == VALUE_UNKNOWN)
		{
			if(i==0)
			{
				maxGain=getProperMaxGain(pInst,initFlag);
	
			}else{
					tmpGain=getProperMaxGain(pInst,initFlag);
					maxGain=DF_MAX(maxGain,tmpGain);
		     	}
			i++;
		}

		currGate=next_lin_table_rec(currGate);
	}
	return maxGain;
}

/* 1. according to initFlag, get the proper gain,
   2. Because only one gain associated with one gate
      is useful although we put both into bucket,
      the proper min gain is got from the larger gain of
      one gate.
*/
int
getProperMinGain(pInst,initFlag)
PW_INST	*pInst;
BOOLEAN	initFlag;
{
	int	gain;

	if(initFlag==TRUE)
    {   /* For the initial gain values */
    	gain=DF_MAX(pInst->gain_st[0],pInst->gain_st[1]);
    }
    else{
            /* For the to-be-updated gain values */
            gain=DF_MAX(pInst->new_gain_st[0],pInst->new_gain_st[1]);
        }
	return gain;
}



/* the real maximum gain of both gains of one gate */
int
getProperMaxGain(pInst,initFlag)
PW_INST	*pInst;
BOOLEAN initFlag;
{
	int gain;

	if(initFlag==TRUE)
	{   /* For the initial gain values */
		gain=DF_MAX(pInst->gain_st[0],pInst->gain_st[1]);
	}
	else{
			/* For the to-be-updated gain values */
			gain=DF_MAX(pInst->new_gain_st[0],pInst->new_gain_st[1]);
		}

	return gain;
}

/* the shift function for cluster gain definition of MFFC root gates */
/* this function assumes that: 
	the gains for non-root gates stay in
	gain_st[] and those for root gates stay in new_gain_st[]*/
int
shiftMFFCrootGain_cluster(MFFCrootList,nonMFFCrootList)
lin_table_rec   MFFCrootList,nonMFFCrootList;
{
    int rootMinGain,nonRootMaxGain;
    int increment;

    /* minimum gain of all the root gates based on the larger gain 
       associated with one roog gate.
    */
    rootMinGain=minGainRootGates(MFFCrootList,FALSE);

    nonRootMaxGain=maxGainNonRootGates(nonMFFCrootList,TRUE);

    /* shift is needed */
    if(rootMinGain<=nonRootMaxGain)
    {

        increment=nonRootMaxGain-rootMinGain+1;

        shiftAction(MFFCrootList,increment,FALSE);

    }else
        increment=0;

    return increment;
}

