#include <stdio.h>
#include <string.h>
#include "DFstring.h"
#include "PWcir.h"
#include "PWilib.h"
#include "PWilib_routines.h"
#include "PWglobal_var.h"

#include "sis.h"

extern void update_turnontime_list(ilib_table_t *table, float turnon_time);

/* extract fanin # from particular gate type name */
extern void split_gate_name_fields(char *str,int *nin);

/* read the current value & time pair into list */
extern int read_list_pair(FILE* fin, ilib_table_t *table, int index, float turnon_time, pair_list_type_t list_type);

/* insert pair into one lookup table entry */
extern void insert_pair_operation(ilib_table_t *table, int index, float turnon_time, pair_list_type_t list_type, float current, float time);

/* generate the library entries for complex gate types from existing ones
 */
extern int genComplexGateLib(ilib_table_t *table);

/* Generate the lib for the gate which is the cascading form of two simpler gates.
   The second gate is a simple one-input gate.
 */
extern int genLibCascade2(ilib_table_t *table, char *target, char *src1, char *src2);

/* Generate the entry for the gate which is the cascading form of two simpler gates.
 * The second gate is a simple one-input gate.
 */
extern int genEntryCascade2(ilib_table_t *table, char *target, int fanin_num, int fanout_num, char *src1, char *src2);

/* obtain the current library data by combining two cascaed simple gates */
extern int ComplexGateILIB(lsList target_headlist, lsList src1_headlist, lsList src2_headlist);

/* 
 * The generic gate types that is supported in the imax/ilib algorithm.
 * and could be in the lib file.
 *
 * Not all of them have to show up in the lib file, because
 * some can be obtained by others since all the gates are turned on
 * almost at the same time according to spice simulation.
 */
char *generic_type_index_table[GENERIC_TYPE_NUM_IN_ILIB]=
{
	"NOT",
	"NAND",
	"NOR",
	"DF",

	"AND",
	"OR",
	"BUFF"
};

char *complex_gate_type_name[COMPLEX_GATE_TYPE_NUM]=
{
	"AND",
	"OR",
	"BUFF"
};

void
update_counter(int * counter, int * flag, int required_num)
{
	(*counter) ++;
	if((*counter) >= required_num)
		(*flag) = 1;
}

/* the index corresponding to the generic gate type */
int
generic_type_index(char *str, char **generic_type_index_table)
{
	int	i;
	for(i=0; i<GENERIC_TYPE_NUM_IN_ILIB; i++)
	{
		if( strcmp(str, generic_type_index_table[i])==0)
			return i;
	}
	return -1;
}

/* allocate & initialize the lookup table */
void
alloc_init_ilib_table(ilib_table_t **table,  int total_entry_num, int fanin_num, int fanin_field_bits, int fanout_num, int fanout_field_bits, int flag)
{
	int i;
	ilib_entry_t *entry;

	if(flag > 0) 
	{
		(*table) = (ilib_table_t *) mem_get_block ((SIZE_T) sizeof(ilib_table_t));
		(*table)->generic_gate_types = generic_type_index_table;
		(*table)->table_size = total_entry_num;
		(*table)->valid_entry_num = 0;
		(*table)->max_fanin_num = fanin_num;
		(*table)->max_fanout_num = fanout_num;

		(*table)->turnon_time_list = lsCreate();

		(*table)->rule = (ilib_index_rule_t *) mem_get_block ((SIZE_T) sizeof(ilib_index_rule_t));
		(*table)->rule->fanin_field_bits = fanin_field_bits;
		(*table)->rule->fanout_field_bits = fanout_field_bits;

		(*table)->body=array_alloc(ilib_entry_t*, total_entry_num);

		for(i=0; i< total_entry_num; i++)
		{
			entry = (ilib_entry_t *) mem_get_block ((SIZE_T) sizeof(ilib_entry_t));
			entry -> headList = lsCreate();
			array_insert_last(ilib_entry_t *, (*table)->body, entry);
		}
	}
}

/* keep track of all the turn-on time in the lib file */
void
update_turnontime_list(ilib_table_t *table, float turnon_time)
{
	lsList	list;
	lsGen	gen;
	float	*time, *time_alloc;
	int		flag =0;

	list = table->turnon_time_list;

	lsForeachItem(list, gen, time)
	{
		if((*time) == turnon_time)
		{	flag = 1;
			break;
		}
	}
	if(flag != 1)
	{
		time_alloc = (float *) mem_get_block((SIZE_T) (sizeof(float)));
		(*time_alloc) = turnon_time;
		lsNewEnd( (lsList) list, (lsGeneric) time_alloc, LS_NH );
	}
}

/* insert pair into one lookup table entry */
void
insert_pair_operation(ilib_table_t *table, int index, float turnon_time, pair_list_type_t list_type, float current, float time)
{
	ilib_entry_t		  *entry;
	ilib_pair_list_head_t *head;
	lsGen 				  gen;
	lsList				  p;

	entry = array_fetch(ilib_entry_t *, table->body, index);

	if(lsLength(entry->headList) <= 0) table->valid_entry_num++ ;

	update_turnontime_list(table, turnon_time);

	/* find or create the corresponding pair list */
	lsForeachItem(entry->headList, gen, head)
	{
		if(head->turnonTime == turnon_time)
			break;
	}

	if(head == NULL)
	{
		head = (ilib_pair_list_head_t *) mem_get_block((SIZE_T) sizeof(ilib_pair_list_head_t));
		head->turnonTime = turnon_time;
		head->leftmostPeakTime[RISE_LIST] = head->rightmostPeakTime[RISE_LIST] = -1;
		head->leftmostPeakTime[STABLE0_LIST] = head->rightmostPeakTime[STABLE0_LIST] = -1;
		head->pairList[RISE_LIST] = lsCreate();
		head->pairList[STABLE0_LIST] = lsCreate();
		lsNewEnd( (lsList) entry->headList, (lsGeneric) head, LS_NH );
		p = head->pairList[list_type];
	}else{ p = head->pairList[list_type]; }

	pairNewEnd((lsList) p, current, time) ;
}

/* append the new pair */
void
pairNewEnd(lsList pairlist, float current, float time)
{
	ilib_pair_t	  *pair;

        pair = (ilib_pair_t *) mem_get_block((SIZE_T) sizeof(ilib_pair_t));
        pair->current = current;
        pair->time = time;
        lsNewEnd( (lsList) pairlist, (lsGeneric) pair, LS_NH );
}


/* read the current value & time pair into list */
int
read_list_pair(FILE* fin, ilib_table_t *table, int index, float turnon_time, pair_list_type_t list_type)
{

	float			current, time;
	char			ch = ' ';

	ch=fgetc(fin);
	while(ch!='\n')
	{
		do{ ch=fgetc(fin);
		}while(ch=='\r' || ch == '\t' || ch == ' ');

		if(ch == '\n')
			continue;

		if(ch != '(')
		{
			fprintf(stderr,"Wrong ilib file format around %c.\n", ch);
			return 1;
		}else{
				fscanf(fin, "%f", &current);
				fscanf(fin, "%f", &time);
				do{ ch=fgetc(fin);
				}while(ch=='\r' || ch == '\t' || ch == ' ');
				if(ch != ')')
				{
					fprintf(stderr,"Wrong ilib file format around %c.\n", ch);
					return 1;
				}
			 }

		/* insert the current_time pair */
		insert_pair_operation(table, index, turnon_time, list_type, current, time);

	}
	return 0;
}

/* extract fanin # from particular gate type name */
void
split_gate_name_fields(char *str,int *nin)
{
	unsigned int	len;
	char *p;
	len = strlen(str);
	p = &(str[len - 1]);
	while(isdigit((int)(*p)) !=0 ) /* while it's a digit */
	{
		p--;
	}
	p++;
	if(*p == '\0')
		*nin = 1;	/* defaut is 1 : like NOT, DF */
	else
		sscanf(p, "%d", nin);

	while(*p !='\0')
	{
		(*p) = '\0';
		p++;
	}
}

/* get the minimum binary field that can represents the number */
int
fit_radix2_field(int num, int *bits)
{
	int	i;
	int j;
	for(i=0;i<16;i++)
	{
		j = (int) (pow(2,i));
		if(j >= num) 
		{
			*bits = i;
			return j;
		}
	}
	return -1;
}

/* generate the library entries for complex gate types from existing ones
 */
int
genComplexGateLib(ilib_table_t *table)
{
	int				i, type_index;
	ilib_entry_t	*entry1, *entry2;

	for(i=0;i<COMPLEX_GATE_TYPE_NUM;i++)
	{
		/* assume that the lib should have entry with load 1 */
		type_index = generic_type_index(complex_gate_type_name[i], generic_type_index_table);
		entry1 = ilibEntry3(table, type_index, 1, 1);
		entry2 = ilibEntry3(table, type_index, 2, 1);

		if( lsLength( ilibEntryHeadList(entry1) ) <= 0 &&
		    lsLength( ilibEntryHeadList(entry2) ) <= 0 )
		{
			/* generate it from existing entries in the lib */
			if(strcmp(complex_gate_type_name[i], "AND") == 0)
			{
				genLibCascade2(table, "AND", "NAND", "NOT");
			}else if(strcmp(complex_gate_type_name[i], "OR") == 0)
			{
				genLibCascade2(table, "OR", "NOR", "NOT");
			}else if(strcmp(complex_gate_type_name[i], "BUFF") == 0)
			{
				genLibCascade2(table, "BUFF", "NOT", "NOT");
			}else
			{
					fprintf(stderr,"Cannot generate lib entry for %s from other gate types.\n", 
									complex_gate_type_name[i]);
					return 1;
				 }
		}
	}
	return 0;
}

/* 
 * read ilib: the library for power-up current 
 * 
 * NOTE: if not all gate types are given in the lib file,
 *       try to generate for complex gate types 
 *       from the combination of existing ones.
 */

int
PWilibRead(FILE* ilibFp, FILE* dbgFp)
{
	/* type, fanin#, fanout# will be concatenated to index the lookup table */
	int	generic_type_num=-1, fanin_num=-1, fanout_num=-1, turnon_time_num=-1, entry_num=-1;
	int fanin_num_fit=-1, fanout_num_fit=-1;
	int fanin_field_bits, fanout_field_bits;
	int flag, counter;
	int nin, nout, type_index;
	float turnon_time;
	char str[MAX_STR_BUF], generic_type_name[MAX_STR_BUF];

	if(ilibFp == NULL)
		return 1;	/* fail */

	flag = -1;
	counter = 0; /* required fields before reading data */
	do{
		if(fscanf(ilibFp, "%s", str)<=0)
			break;
		strupper(str);

		if(strcmp(str, "#")==0 || str[0] == '#')
			continue;
		if(strcmp(str, "GENERIC_TYPE_NUM=")==0)
		{
			fscanf(ilibFp, "%d", &generic_type_num);
			fprintf(stdout,"There are %d generic gate types in the lib file.\n",
							generic_type_num);
			fprintf(stdout,"But we may generate data for complex gates using these. \n");
			fprintf(stdout, "The total supported # of types is %d.\n",GENERIC_TYPE_NUM_IN_ILIB);
			update_counter(&counter, &flag, REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit * fanout_num_fit;
			alloc_init_ilib_table(&PWilib_table, entry_num, fanin_num, fanin_field_bits, fanout_num, fanout_field_bits, flag);
		}
		else if(strcmp(str, "FANIN_NUM=")==0)
		{
			fscanf(ilibFp, "%d", &fanin_num);
			fanin_num_fit=fit_radix2_field(fanin_num, &fanin_field_bits);
			update_counter(&counter, &flag, REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit * fanout_num_fit;
			alloc_init_ilib_table(&PWilib_table, entry_num, fanin_num, fanin_field_bits, fanout_num, fanout_field_bits, flag);
		}else if(strcmp(str, "FANOUT_NUM=")==0)
		{
			fscanf(ilibFp, "%d", &fanout_num);
			fanout_num_fit= fit_radix2_field(fanout_num, &fanout_field_bits);
			update_counter(&counter, &flag, REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit * fanout_num_fit;
			alloc_init_ilib_table(&PWilib_table, entry_num, fanin_num, fanin_field_bits, fanout_num, fanout_field_bits, flag);
		}
		else if(strcmp(str, "TURNON_TIME_NUM=")==0)
		{
			fscanf(ilibFp, "%d", &turnon_time_num);
			update_counter(&counter, &flag, REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit * fanout_num_fit;
			alloc_init_ilib_table(&PWilib_table, entry_num, fanin_num, fanin_field_bits, fanout_num, fanout_field_bits, flag);
		}else if(strcmp(str, "GATE=")==0)
		{
			while(strcmp(str, "ENDGATE")!=0)
			{
				/* get generic gate type */
			  fscanf(ilibFp, "%s", str);
			  split_gate_name_fields(str,&nin);
			  strcpy(generic_type_name, str);
			  type_index = generic_type_index(str, generic_type_index_table);
			  if(nin<=0 || type_index<0)
			  {
			  	fprintf(stderr, "Wrong with %s%d \n", str, nin);
				return 1;
			  }

			  fscanf(ilibFp, "%s", str);
			  while(strcmp(str, "TURNON_TIME=")==0)
			  {
			  	fscanf(ilibFp, "%f", &turnon_time);
			  	if(turnon_time <= 0)
			  	{
			  		fprintf(stderr,"Negative turnon_time %f for gate %s%d \n", 
			  						turnon_time, generic_type_name, nin);
					return 1;
				}
				while(strcmp(str, "ENDTURNON_TIME")!=0)
			   {
			  	fscanf(ilibFp, "%s", str);
				while(strcmp(str, "LOAD=") == 0)
				{
					fscanf(ilibFp, "%d", &nout);
					if(nout<=0)
					{
						fprintf(stderr, "Worng with %s%d with load %d .\n", generic_type_name, nin, nout);
						return 1;
					}

					while(strcmp(str, "ENDLOAD")!=0)
					{
						fscanf(ilibFp, "%s", str);
						while(strcmp(str, "RISE")==0 || strcmp(str, "STABLE0")==0)
						{
							if(strcmp(str, "RISE")==0)
							{
								toNextLine(ilibFp);
								if( read_list_pair(ilibFp, PWilib_table, 
													lookup_table_index(type_index, nin, fanin_field_bits, nout, fanout_field_bits), 
													turnon_time,
													RISE_LIST) == 1)
									return 1;
													
							}
							if(strcmp(str, "STABLE0")==0)
							{
								toNextLine(ilibFp);
								if( read_list_pair(ilibFp, PWilib_table,
													 lookup_table_index(type_index, nin, fanin_field_bits, nout, fanout_field_bits),
													 turnon_time,
													 STABLE0_LIST) == 1)
									return 1;
							}
							fscanf(ilibFp, "%s", str);
						}
					}
					fscanf(ilibFp, "%s", str);
				}
			   }
			   fscanf(ilibFp, "%s", str);
			  }
			}
		}
	  }while(toNextLine(ilibFp));

	  /* debugging */
	  dump_ilib(PWilib_table, dbgFp);
	  fflush(dbgFp);

	  if(genComplexGateLib(PWilib_table) == 1) /* failed */
	  	return 1;

	  left_right_PeakTime(PWilib_table);

	  /* debugging */
	  dump_ilib(PWilib_table, dbgFp);
	  fflush(dbgFp);

	  return 0;
}


/* dump ilib to file */
void
dump_ilib(ilib_table_t *table, FILE * dbgFp)
{
	int			 			i;
	ilib_entry_t 			*entry;
	lsGen		 			gen1, gen2;
	ilib_pair_list_head_t 	*head;
	ilib_pair_t				*pair;
	float                   *time;
	char                    type_name[MAX_STR_BUF];
	int					nin, nout;
	int					counter;

					fprintf(dbgFp,"\n");
	                              fprintf(dbgFp,"   table size : %d \n", table->table_size);
				      fprintf(dbgFp,"valid entries : %d \n", table->valid_entry_num);
				      fprintf(dbgFp," max fanin #  : %d \n", table->max_fanin_num);
				      fprintf(dbgFp,"max fanout #  : %d \n", table->max_fanout_num);
	fprintf(dbgFp,"Total %d turnon_time are in the lib.\n", lsLength(table->turnon_time_list));
	lsForeachItem(table->turnon_time_list, gen1, time)
	{
		fprintf(dbgFp," turnon time = %f \n", (*time));
	}
	fprintf(dbgFp,"All the turnon time have been printed.\n\n");

	for(i=0; i<array_n(table->body); i++)
	{
		entry = array_fetch(ilib_entry_t *, table->body, i);
		if(lsLength(entry->headList)>0)
		{
			lsForeachItem(entry->headList, gen1, head)
			{
				if(lsLength(head->pairList[RISE_LIST])>0 && lsLength(head->pairList[STABLE0_LIST])>0)
				{
					fprintf(dbgFp,"          index : %d  \n",i);
					index2gate(table, (unsigned)i, type_name, &nin, &nout);
					fprintf(dbgFp,"          name  : %s \n", type_name);
					fprintf(dbgFp,"         fanin# : %d \n", nin);
					fprintf(dbgFp,"        fanout# : %d \n", nout);
					fprintf(dbgFp,"   Turn-on Time : %f \n", head->turnonTime);

					fprintf(dbgFp,"     rise pairs :\n");
				   fprintf(dbgFp,"leftmost PeakTime: %f \n", head->leftmostPeakTime[RISE_LIST]);
				  fprintf(dbgFp,"rightmost PeakTime: %f \n", head->rightmostPeakTime[RISE_LIST]);
					fprintf(dbgFp,"                 ");
					counter = 0;
					lsForeachItem(head->pairList[RISE_LIST], gen2, pair)
					{
						fprintf(dbgFp,"    (%f %f) ", pair->current, pair->time);
						counter ++;
						if(counter%3 == 0) 
						{
							fprintf(dbgFp, "\n");
							fprintf(dbgFp,"                 ");
						}
					}
					fprintf(dbgFp, "\n");

					fprintf(dbgFp,"   stable0 pairs:\n");
				  fprintf(dbgFp,"leftmost PeakTime: %f \n", head->leftmostPeakTime[STABLE0_LIST]);
				 fprintf(dbgFp,"rightmost PeakTime: %f \n", head->rightmostPeakTime[STABLE0_LIST]);
					fprintf(dbgFp,"                 ");

					counter = 0;
					lsForeachItem(head->pairList[STABLE0_LIST], gen2, pair)
					{
						fprintf(dbgFp,"    (%f %f) ", pair->current, pair->time);
						counter ++;
						if(counter%3 == 0) 
						{
							fprintf(dbgFp, "\n");
							fprintf(dbgFp,"                 ");
						}
					}
					fprintf(dbgFp,"\n");
				}
			}
		}
	}
}

/* convert lookup table index to gate information */
int
index2gate(ilib_table_t *table, unsigned int index, char *type_name, int *nin, int *nout)
{
	unsigned int fanin_mask, fanout_mask;
	unsigned int type_index;
	int			 fanin_field_bits;
	int			 fanout_field_bits;

	fanin_mask = fanout_mask = 1;
	fanin_field_bits = table->rule->fanin_field_bits;
	fanout_field_bits = table->rule->fanout_field_bits;

	fanin_mask = (fanin_mask << (fanin_field_bits+fanout_field_bits - 1)) | 
				 (fanin_mask << (fanin_field_bits+fanout_field_bits - 2));
	fanout_mask = (fanout_mask << (table->rule->fanout_field_bits - 1)) | fanout_mask ;

	*nin = ((index & fanin_mask) >> fanout_field_bits) + 1U;
	*nout = (index & fanout_mask) + 1U;
	type_index = index >> (table->rule->fanin_field_bits + table->rule->fanout_field_bits);
	if(type_index < 0 || type_index >= GENERIC_TYPE_NUM_IN_ILIB)
	{
		fprintf(stderr,"Error in index2gate(): conversion from index to type_index error.\n");
		return 1; /* fail */
	}
	
	strcpy(type_name, generic_type_index_table[type_index]);
	return 0; /* OK */
}

/* Generate the lib for the gate which is the cascading form of two simpler gates.
   The second gate is a simple one-input gate.
 */
int
genLibCascade2(ilib_table_t *table, char *target, char *src1, char *src2)
{
	int	i,j;

	if(strcmp(target, "AND") == 0 || strcmp(target, "OR") == 0)
	{
		/* for different fanin number */
		for(i=2; i<=4; i++)
		{
			/* for different fanout number */
			for(j=1; j<=4; j++)
			{
				genEntryCascade2(table, target, i, j, src1, src2);
				table->valid_entry_num ++;
			}
		}
	}else if (strcmp(target, "BUFF") == 0)
	{
		for(j=1; j<=4; j++)
		{
			genEntryCascade2(table, target, 1, j, src1, src2);
			table->valid_entry_num ++;
		}
	}else {
			fprintf(stderr,"Unknown case in genLibCascade2(): %s %s %s\n", target, src1, src2);
			return 1;
		  }
	return 0;
}

/* Generate the entry for the gate which is the cascading form of two simpler gates.
   The second gate is a simple one-input gate.
 */
int
genEntryCascade2(ilib_table_t *table, char *target, int fanin_num, int fanout_num, char *src1, char *src2)
{
	int	   					target_type_index, src1_type_index, src2_type_index;
	lsList 					target_headlist, src1_headlist, src2_headlist;
	ilib_entry_t			*entry_target, *entry_src1, *entry_src2;

	target_type_index = generic_type_index(target, generic_type_index_table);
	src1_type_index = generic_type_index(src1, generic_type_index_table);
	src2_type_index = generic_type_index(src2, generic_type_index_table);

	entry_target = ilibEntry3(table, target_type_index, fanin_num, fanout_num);
	entry_src1 = ilibEntry3(table, src1_type_index, fanin_num, 1);
	entry_src2 = ilibEntry3(table, src2_type_index, 1, fanout_num);

	target_headlist = ilibEntryHeadList(entry_target);
	src1_headlist = ilibEntryHeadList(entry_src1);
	src2_headlist = ilibEntryHeadList(entry_src2);

	if( lsLength(target_headlist) >0 )
	{
		fprintf(stderr," Error: try to generate an entry for gate which already has one.\n");
		return 1;
	}
	if( lsLength(src1_headlist) != lsLength(src2_headlist) )
	{
		fprintf(stderr, "Error: The entries in the lib don't have the same number of turn-on time.\n");
		return 1;
	}

	if( ComplexGateILIB(target_headlist, src1_headlist, src2_headlist) == 1)  return 1;

	return 0;
}

/* obtain the current library data by combining two cascaed simple gates */
/* the second gate should have the inversion property */
int
ComplexGateILIB(lsList target_headlist, lsList src1_headlist, lsList src2_headlist)
{
	ilib_pair_list_head_t *head;
	ilib_pair_list_head_t   *src1_head, *src2_head;
	lsGen gen1, gen2;

	/* for each turn-on time availble, we add up the current waveform */
	for(gen1 = lsStart(src1_headlist), gen2 = lsStart(src2_headlist);
	    ((lsNext(gen1, (lsGeneric *) &src1_head, LS_NH) == LS_OK) && 
		 (lsNext(gen2, (lsGeneric *) &src2_head, LS_NH) == LS_OK))
		|| ((void)lsFinish(gen1),(void)lsFinish(gen2),0); )
	{
		if(src1_head->turnonTime != src2_head->turnonTime)
		{
			fprintf(stderr,"Error in ComplexGateILIB() : the turn-on time must be the same in order to combine.\n");
			return 1;
		}
		head = (ilib_pair_list_head_t *) mem_get_block((SIZE_T) sizeof(ilib_pair_list_head_t));
		head->turnonTime = src1_head->turnonTime;
		head->pairList[RISE_LIST] = lsCreate();
		head->pairList[STABLE0_LIST] = lsCreate();
		lsNewEnd( (lsList) target_headlist, (lsGeneric) head, LS_NH );
		/* we assume that the second gate is an inversion type gate */
		if( sum_current_waveform(head->pairList[RISE_LIST], src1_head->pairList[STABLE0_LIST], src2_head->pairList[RISE_LIST]) == 1)
			return 1;
		if( sum_current_waveform(head->pairList[STABLE0_LIST], src1_head->pairList[RISE_LIST], src2_head->pairList[STABLE0_LIST]) == 1)
			return 1;
	}

	return 0;
}
