#include <stdio.h>
#include "DFtype.h"
#include "Rtime.h"
#include "PWcir.h"
#include "PWglobal_var.h"
#include "PWroutines.h"
#include "PWilib.h"
#include "PWilib_routines.h"

extern int gate_preferred_current_cmp(char **inst1, char **inst2);
extern void dump_gate_sort(array_t* gate_sort);
extern float reportMaxCurrent2(float turnon_time, BOOLEAN write_file_flag);
extern float max_current_in_waveform(float turnon_time);
extern int currentAtTime_DF(ilib_table_t *PWilib_table, float time, float turnon_time);
extern void print_RTL_est_info(float max_powerup_current);
extern void PWmaxCurrent_ilib(array_t *gate_array_sorted);


/* report the result for atpg_ilib */
extern float reportMaxCurrent(float time, BOOLEAN write_file_flag);


int
gate_preferred_current_cmp(char **inst1, char **inst2)
{
	float 					current1, current2;
	gate_current_t 			*p1, *p2;

	p1 = ((PW_INST*)(*inst1))->powerup_current;
	p2 = ((PW_INST*)(*inst2))->powerup_current;
	current1 = p1->current[p1->preferred_type];
	current2 = p2->current[p2->preferred_type];
	if(current1 == current2)
		return 0;
	else if(current1 > current2)
		return 1;
	else return -1;
}

/* Imax with the current waveform library: ilib */
int
PWimax_ilib(float turnon_time)
{
	float	t_start, t_end, t_step, time, *time_p;
	array_t *gate_sort;
	int		i;
	float   current_max, max_powerup_current=0;
	int     vec[MAX_PI];
	lsGen   gen1;
	int     isValid=0;

	fprintf(stdout,"Running Imax/ilib.\n");
	fprintf(resultFp,"\n\nRunning Imax/ilib.\n");
	fprintf(resultFp,"=============================================\n");
	SetTime(TIME_REPORT);

	/* currently, test it for turn-on time 100ps */
	lsForeachItem(PWilib_table->turnon_time_list, gen1, time_p)
	{
		if(turnon_time == (*time_p)) isValid = 1; // a valid turnon time value
	}
	if(isValid == 0)
	{
		fprintf(stderr," Invalid turnon time value.\n");
		return 1; /* fail */
	}
	//turnon_time = 100;

	/* the range of time and step to try timed-atpg */
	time_range_step(PWilib_table, turnon_time, instList, &t_start, &t_end, &t_step);

	gate_sort = array_alloc(PW_INST *, PWinstNum_and_pseudo);
	for(i=0; i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
		   instList[i].instLibNo!=DF_LIB_NO &&
		   instList[i].instLibNo!=DFR_LIB_NO &&
		   instList[i].instLibNo!=DFS_LIB_NO &&
		   instList[i].instLibNo!=DFRS_LIB_NO)
		   {
		   	array_insert_last(PW_INST *, gate_sort, &(instList[i]));
		   }
	}

	/* ATPG-based estimation at each time point */
	for(time = t_start; time <=  t_end; time += t_step)
	{
		PWfieldRelease();
		initImplyStk();

		fprintf(stdout,"Initialization for Estimation at time point %f ... \n", time);

		/* get the current at the time point */
		if(currentAtTime(gate_sort, PWilib_table, time, turnon_time)==1) 
		return 1;
		/* sorting gates according to the current at the time point */
		array_sort(gate_sort, gate_preferred_current_cmp);
		dump_gate_sort(gate_sort);
		fflush(dbgFp);

		/* sorting gate's input pins */
		PWinputSort_ilib();
		fflush(dbgFp);

		fprintf(stdout,"Initialization finished.\n");
		ReportTime(TIME_REPORT);

		/* ATPG based Power-up current Maximization : Imax/ilib */
		fprintf(stdout,"Estimation starting...\n");
		PWmaxCurrent_ilib(gate_sort);

		/* report the max power-up current at the selected time point */
		reportMaxCurrent(time,TRUE);

		current_max = reportMaxCurrent2(turnon_time, TRUE);
		max_powerup_current = DF_MAX(current_max, max_powerup_current);
		if(max_powerup_current == current_max) input_vec_copy(SourceP, vec);


		fflush(dbgFp);
		fflush(resultFp);
		fprintf(stdout,"Estimation is finished for time point %f \n\n.",time);

	}
	fprintf(stdout,"All estimation have been finished. Check output file for result.\n\n\n");
	fprintf(resultFp,"\n=============================================\n");
	fprintf(resultFp,"Final_result_powerup= %f ua.\n",max_powerup_current);
    dump_input_vector(resultFp,vec);

    print_RTL_est_info(max_powerup_current);

	return 0; /* OK */
}

/* currently, we investigate all the gates in the library */
/* later, it can be modified to look at only the gates in the circuit */
void
time_range_step(ilib_table_t *table, float turnon_time, PW_INST *instList, float *t_start, float *t_end, float *t_step)
{
	float tmp1, tmp2;
	int	  flag=0;
	int		i;
	ilib_entry_t 			*entry;
	ilib_pair_list_head_t 	*head;
	lsGen					gen;

	for(i=0; i<array_n(table->body); i++)
	{
		entry = array_fetch(ilib_entry_t *, table->body, i);
		if(lsLength(entry->headList)>0)
		{
			lsForeachItem(entry->headList, gen, head)
			{
			  if(head->turnonTime == turnon_time)
			  {
				flag++;
				if(lsLength(head->pairList[RISE_LIST])>0 && lsLength(head->pairList[STABLE0_LIST])>0)
				{
					tmp1 = DF_MIN(head->leftmostPeakTime[STABLE0_LIST],
									  head->leftmostPeakTime[RISE_LIST]);
					tmp2 = DF_MAX(head->rightmostPeakTime[STABLE0_LIST],
									  head->rightmostPeakTime[RISE_LIST]);
					if(flag == 1)
					{
						*t_start = tmp1;
						*t_end = tmp2;
					}
					else{
						*t_start = DF_MIN((*t_start), tmp1);
						*t_end = DF_MAX((*t_end), tmp2);
						}
				}
			  }
			}
		}
	}
	*t_step = (*t_end - *t_start)/TIME_POINT_NUM;
	if((*t_step) < 0)
	{
		fprintf(stderr,"Error: negative time step in time_range_step().\n");
		exit(1);
	}
}

/* obtain the current at specified time for each gate in the sorting array
 *
 * the corresponding field in the instance will be updated 
 */
int
currentAtTime(array_t *gate_sort, ilib_table_t *PWilib_table, 
			  float time, float turnon_time)
{
	int		i;
	PW_INST	*inst;

	for(i=0; i<array_n(gate_sort); i++)
	{
		inst = array_fetch(PW_INST *, gate_sort, i);
		if(accessCurrentILIB(PWilib_table, inst, time, turnon_time)==1) 
			return 1;
	}
	if(currentAtTime_DF(PWilib_table, time, turnon_time) == 1) 
		return 1;

	return 0;
}

int
currentAtTime_DF(ilib_table_t *PWilib_table, float time, float turnon_time)
{
	int i;

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo==DF_LIB_NO ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO)
		{
			if(accessCurrentILIB(PWilib_table, &(instList[i]),time, turnon_time)==1)
				return 1;
		}
	}
	return 0;
}

/* report the result for atpg_ilib */
float
reportMaxCurrent(float time, BOOLEAN write_file_flag)
{
    int     	i;
    float     	maxCurrent=0;
    float 		maxCurrentWithoutPI=0;
    float     	maxCurrentRealPI=0;
    float		maxCurrentFF=0;
    LOGIC_VALUE lv;
    int     	v;

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
		   instList[i].instLibNo!=DF_LIB_NO &&
		   instList[i].instLibNo!=DFR_LIB_NO &&
		   instList[i].instLibNo!=DFS_LIB_NO &&
		   instList[i].instLibNo!=DFRS_LIB_NO)
		{
			lv=fetchImplyValueInStk(&(instList[i]));
			v=logic2int(lv);	/* either 0 or 1 */

			if(instList[i].instLibNo!=PI_LIB_NO)
				maxCurrentWithoutPI+=
					instList[i].powerup_current->current[v];
		}
	}

	maxCurrentRealPI=0;
	/* for real primary inputs: which are buffers */
	for(i=0;i<PSEUDO_PI_index;i++)
	{
		lv=fetchImplyValueInStk(SourceP[i]);
		v=logic2int(lv);
		maxCurrentRealPI+= SourceP[i]->powerup_current->current[v];
	}

    /* For pseudo_PO, it is counted into a primitive gate's fanout.
       But if you use direct memory copy from sis,
       nothing to worry about. No pseudo_PO after memory copy.
     */

	/* For pseudo_PI: designate the FFs (Flip-flops)
	 */
	for(i=PSEUDO_PI_index;i<PrimInNum;i++)
	{
		if(strcmp(SourceP[i]->name, "VSS") !=0 &&
                   strcmp(SourceP[i]->name, "VCC") !=0 &&
		   strcmp(SourceP[i]->name, "VDD") !=0 &&
		   strcmp(SourceP[i]->name, "GND") !=0)
		{
		lv= fetchImplyValueInStk(SourceP[i]);
		v = logic2int(lv);
		maxCurrentFF += SourceP[i]->in[0]->powerup_current->current[v];
		}
	}

	maxCurrent = maxCurrentWithoutPI + maxCurrentRealPI + maxCurrentFF;

	if(write_file_flag == TRUE)
	{
		fprintf(resultFp,"\n");
		fprintf(resultFp,"The maximum current at time point %fps is %f ua.\n", 
						time,
						maxCurrent);
		fprintf(resultFp,"\n\nThe input vector is:\n");
		for(i=0;i<PrimInNum;i++)
		{
			fprintf(resultFp,"PI_name=%s, value=%d\n",
							SourceP[i]->name,
							logic2int(fetchImplyValueInStk(SourceP[i])));
		}
		fprintf(resultFp,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
	}
	return maxCurrent;
}

void
dump_gate_sort(array_t* gate_sort)
{
	int 	i;
	PW_INST *inst;

	fprintf(dbgFp,"Dump the sorted gate array.\n");
	for(i=array_n(gate_sort)-1; i>=0; i--)
	{
		inst = array_fetch(PW_INST *, gate_sort,i);

		fprintf(dbgFp,"instance: %s, preferred current: %f.\n",
		inst->name, 
		inst->powerup_current->current[inst->powerup_current->preferred_type]);
	}
	fprintf(dbgFp,"\n");
}

float
reportMaxCurrent2(float turnon_time, BOOLEAN write_file_flag)
{
	float maxCurrent;
	int	i;

	maxCurrent = max_current_in_waveform(turnon_time);

    if(write_file_flag == TRUE)
    {
        fprintf(resultFp,"\n");
        fprintf(resultFp,"The maximum current obtained by this run of atpg_ilib is %f ua.\n",
                        maxCurrent);
        fprintf(resultFp,"\n\nThe input vector is:\n");
        for(i=0;i<PrimInNum;i++)
        {
            fprintf(resultFp,"PI_name=%s, value=%d\n",
                            SourceP[i]->name,
                            logic2int(fetchImplyValueInStk(SourceP[i])));
        }
		fprintf(resultFp,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
    }
    return maxCurrent;
}

void
print_RTL_est_info(float max_powerup_current)
{
    int metric=0,i;

    fprintf(resultFp,"Total_instances= %d (real gates, real PI, & flip-flops)\n",
            (PWinstNum_and_pseudo - PrimOutNum - pseudoCount) );

    for(i=0;i<PWinstNum_and_pseudo;i++)
    {
      /* real gates or flip-flops */
      if(instList[i].instLibNo!=PO_LIB_NO && instList[i].instLibNo!=PI_LIB_NO)
      {
        metric += instList[i].fanout;
      }
    }
    /* real PI */
    for(i=0;i<PSEUDO_PI_index;i++)
    {
      metric += SourceP[i]->fanout;
    }

    fprintf(resultFp,"Gate_count= %d.\n", (PWinstNum_and_pseudo - PrimOutNum - pseudoCount));

    fprintf(resultFp,"Avg_powerup_current_per_gate= %f ua.\n",
            max_powerup_current/(PWinstNum_and_pseudo - PrimOutNum - pseudoCount));
    /*
    fprintf(resultFp,"metric= %d.\n",metric);
    fprintf(resultFp,"Avg_powerup_new_metric= %f us.\n", max_powerup_current/metric);
    */
}
