#include "PWcir.h"
#include "linearExt.h"
#include "linearFree.h"

#include "PWroutines.h"
#include "PWglobal_var.h"


extern void restoreImplication(stk_rec *trackStkTop);
extern void impliedValueToState(stk_rec stack);
extern void report_imply_stk_top(void);

/* COMPLETE the LOGIC ASSIGNMENT for current gate and the implication process */
/* return 'gain' value which is based on the implication result */
int
implyGate(pInst, assign_value, trackStkTop,pConflict)
PW_INST     *pInst;     	/* instance to be assigned */
LOGIC_VALUE assign_value;
stk_rec     *trackStkTop;	/* for restoration */
BOOLEAN     *pConflict;		/* flag <conflict> */
{
    stk_rec         tmpRec;
    stk_rec         processStk=NULL;    /* process stack top */
    PW_IMPLY_STRU   *pProcessElem;      /* for allocating process elem */
    PW_IMPLY_STRU   *currProcessElem;   /* current process elem */
    PW_INST         *tmpInst;
    LOGIC_VALUE     currAssign;
    LOGIC_VALUE     *pValue;
    int             gainValue=0;
    BOOLEAN         conflict1,conflict2;

    if(assign_value!=VALUE_0 && assign_value!=VALUE_1)
    {
        fprintf(stderr,"Incorrect call of implyGate():\n");
        fprintf(stderr,"For gate %s, ID %d\n",pInst->name,pInst->instID);
        exit(1);
    }
    /* before assignment, it should be unknown value */
    if( (*(LOGIC_VALUE*)((pInst->implyStkTop)->member))!=VALUE_UNKNOWN )
    {
        fprintf(stderr,"Error Reporte->\n");
        fprintf(stderr,"During implyGate() for gate %s, ID %d\n",
                        pInst->name,pInst->instID);
        fprintf(stderr,"Should not assign a conflicting value!\n\n");
        exit(1);
    }



    /* allocate and push the original one for assignment */
    pProcessElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
    pProcessElem->inst=pInst;
    pProcessElem->value=assign_value;
    STK_PUSH(&processStk,(double)NOT_CARE_KEY,(pointer)pProcessElem);


    while(processStk!=NULL)
    {
        /* get the current processing element*/
        tmpRec=STK_POP(&processStk);
        currProcessElem=(PW_IMPLY_STRU *)(tmpRec->member);
        mem_free_block((pointer)tmpRec);

        /* get the current instance and logic value */
        tmpInst=currProcessElem->inst;
        currAssign=currProcessElem->value;
        mem_free_block((pointer)currProcessElem);

        /* check if the gate has already been assigned */
        /* which is possible in RECONVERGENT circuits because
           we may have MULTIPLE process elements for ONE assignment!!
		*/
		if(fetchImplyValueInStk(tmpInst)==currAssign)
			continue;

        /* assign logic value for current instance */
        pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
        (*pValue)=currAssign;
        STK_PUSH(&(tmpInst->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);

        /* update gain */
        if(tmpInst->instLibNo!=PO_LIB_NO)
        {	if(currAssign==VALUE_0)
        	{
         	   gainValue-=tmpInst->fanout;
        	}else if(currAssign==VALUE_1)
        	      { gainValue+=tmpInst->fanout;
        	      }
		}

        /* keep track of the gate which is assigned, in order to restore */
        STK_PUSH(trackStkTop,(double)NOT_CARE_KEY,(pointer)tmpInst);

        conflict1=backwardImplication(tmpInst,currAssign,&processStk);
        conflict2=forwardImplication(tmpInst,currAssign,&processStk);
        if(conflict1==TRUE || conflict2==TRUE)
        {
            (*pConflict)=TRUE;
            stackFree(&processStk);

            return gainValue;   /* in fact, gain here has no meaning because of the conflict */
        }

    }
        (*pConflict)=FALSE;
        return gainValue;
}

/* restore value, free stack mem space */
void
restoreImplication(trackStkTop)
stk_rec *trackStkTop;
{
    stk_rec tmpRec;
    PW_INST *tmpInst;

    while((*trackStkTop)!=NULL)
    {
        tmpRec=STK_POP(trackStkTop);
        tmpInst=(PW_INST *)(tmpRec->member);
        mem_free_block((pointer)tmpRec);

        tmpRec=STK_POP(&(tmpInst->implyStkTop));
        mem_free_block((pointer)(tmpRec->member));/* for pValue */
        mem_free_block((pointer)tmpRec);
    }
}


/* copy implied value in implyStk to variable state */
void
impliedValueToState(stack)
stk_rec stack;
{
    PW_INST *instance;
    stk_rec tmpRec=stack;

    while(tmpRec!=NULL)
    {
        instance=((PW_INST*)(tmpRec->member));
        instance->state=(*((LOGIC_VALUE *)(instance->implyStkTop->member)));
        tmpRec=NEXT_STK_REC(tmpRec);
    }
}

/* initialize gates' implication stack */
void
initImplyStk()
{
    int         i;
    LOGIC_VALUE *pValue;
    stk_rec		tmpstk=NULL;
    int			tmpval;

    /* for 'instlist' structure, not including VDD/GND inputs */
    for(i=0;i<PWinstNum_and_pseudo;i++)
    {
        pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
        (*pValue)=VALUE_UNKNOWN;
        STK_PUSH(&(instList[i].implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
    }
    /* initialization for VDD/GND inputs */
    if(gnd_index>=0)
    {
		pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
		(*pValue)=VALUE_UNKNOWN;
		STK_PUSH(&(SourceP[gnd_index]->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
    }

    if(vdd_index>=0)
    {
		pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
		(*pValue)=VALUE_UNKNOWN;
		STK_PUSH(&(SourceP[vdd_index]->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
	}

	/* Assignment and implication for VDD/GND inputs */
	if(gnd_index>=0)
	{

		forwardImplyPI3V(SourceP[gnd_index],VALUE_0,&tmpstk,&tmpval);
		stackFree1(&tmpstk);
	}
	if(vdd_index>=0)
	{
		forwardImplyPI3V(SourceP[vdd_index],VALUE_1,&tmpstk,&tmpval);
		stackFree1(&tmpstk);
	}

}


/* print the top element of the implication stack */
void
report_imply_stk_top()
{
    int i;

    fprintf(dbgFp,"\n\n");
    fprintf(dbgFp,"Checking implication value stack top.\n");
    for(i=0;i<PWinstNum_and_pseudo;i++)
    {
        if(instList[i].instLibNo!=DF_LIB_NO &&
        	instList[i].instLibNo!=DFR_LIB_NO &&
        	instList[i].instLibNo!=DFS_LIB_NO &&
        	instList[i].instLibNo!=DFRS_LIB_NO )
        fprintf(dbgFp,"instance name: %s imply_stk_top=%u\n",
                      instList[i].name,
                      *((LOGIC_VALUE *)(instList[i].implyStkTop->member)));
    }
}

/* detect the case that the gate output value assignment is implictely unfeasible */
BOOLEAN
detectImplicitJustifyFailure(pInst,expected_assign)
PW_INST		*pInst;
LOGIC_VALUE	expected_assign;
{
	BOOLEAN conflict;
	stk_rec	trackStkTop=NULL;

	implyGate(pInst,expected_assign,&trackStkTop,&conflict);
	restoreImplication(&trackStkTop);
	if(conflict==TRUE)
		return TRUE;
	else
		return FALSE;
}
