#include "PWcir.h"
#include "linearExt.h"

#include "PWglobal_var.h"
#include "PWroutines.h"

extern 	void 	restoreImplication(stk_rec *trackStkTop);
extern	void 	impliedValueToState();
extern	void	report_gain(void);
extern  void	report_imply_stk_top(void);
extern  int		shiftMFFCrootGain();

/* for Imax/Gain-cluster algorithm */
/* assign the initial gain to all gates */
void
PWgateInitGain()
{
	int		i;
	stk_rec	trackGateList=NULL;

	/* initialize gates' implication stacks ,*/
	/* including pre-implication for VDD & GND inputs */
	initImplyStk();

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
		   instList[i].instLibNo!=DF_LIB_NO &&
		   instList[i].instLibNo!=DFR_LIB_NO &&
		   instList[i].instLibNo!=DFS_LIB_NO &&
		   instList[i].instLibNo!=DFRS_LIB_NO)
		{
			/* for real gates and PI,we have gain metrc.*/ 

			/* initialize gate gain under current situation */
			/* gain bucket has not been established, so the trackGateList is ignored */
			if(fetchImplyValueInStk(&(instList[i])) == VALUE_UNKNOWN)
			{	/* for those not affected by the VDD/GND inputs */
				updtGateGain(&(instList[i]),TRUE,&trackGateList);
				if(trackGateList!=NULL)
				{
					fprintf(stderr,"Error in PWgate InitGain().\n");
					exit(1);
				}
			}
		}
	}

	/* move MFFC root gates before non-MFFC root gates in the gain order */
	gainShiftAmount=shiftMFFCrootGain(MFFCrootList,nonMFFCrootList,TRUE);

	/* for debugging */
	/* report_gain(); */
	/*report_imply_stk_top();*/
}


/*--------------------------------------------------------------------------------*/

/* update gain for one gate */
/* if initFlag=TRUE, it's initialization; otherwise, it's updating */
/* NOTE: if the logic value of one gate can be 1 or 0, it has a gain.
		 But sometimes the logic value of one gate may be implicitely 
		 determined by current state and reconvergent paths, then 
		 its gain for the illegal assignment will be labelled invalid, 
		 and when we update the bucket, its bucket element will be put after the
		 legal assignment rather than be removed!!
*/
void
updtGateGain(pInst,initFlag,trackStk)
PW_INST	*pInst;
BOOLEAN	initFlag;
stk_rec *trackStk;
{
	BOOLEAN	conflict;

	(*trackStk)=NULL;

	/* first gain for gate output logic value 0 */
	
	/* 'conflict' flag means the gate output doesn't depend on PI's value,
		which occurs in redundant circuits */

	/* initFlag is TRUE if it is to initialize gain*/
	if(initFlag==TRUE)
		pInst->gain_st[0]=implyGate(pInst,VALUE_0,trackStk,&conflict);
	else
		pInst->new_gain_st[0]=implyGate(pInst,VALUE_0,trackStk,&conflict);

	/* restore the implication values & free trackStk */
	restoreImplication(trackStk);


	if(conflict==TRUE)
	{
		/* the gate output can only have VALUE_1 */
		pInst->new_gain_st[1]=implyGate(pInst,VALUE_1,trackStk,&conflict);
		pInst->new_gain_st[0]=INVALID_GAIN;

		/* NOTE: if it's gain updating, gain_st[] shouldn't be changed */
		/* because we need it to remove corresponding element in bucket */
		if(initFlag==TRUE)
		{
			pInst->gain_st[0]=pInst->new_gain_st[0];
			pInst->gain_st[1]=pInst->new_gain_st[1];
		}

		/* the gate output should be either VALUE_1 or VALUE_0 */
		if(conflict==TRUE)
		{	fprintf(stderr,"Impossible value for %s,ID %d\n",pInst->name,pInst->instID);
			exit(1);
		}

		restoreImplication(trackStk);

	}else{

			/* for gain of output value 1 */
			if(initFlag==TRUE)
				pInst->gain_st[1]=implyGate(pInst,VALUE_1,trackStk,&conflict);
			else
				pInst->new_gain_st[1]=implyGate(pInst,VALUE_1,trackStk,&conflict);

			if(conflict==TRUE)
			{	/* the gate output can only have VALUE_0 */
				restoreImplication(trackStk);
				pInst->new_gain_st[0]=implyGate(pInst,VALUE_0,trackStk,&conflict);

		        pInst->new_gain_st[1]=INVALID_GAIN;

				/* NOTE: if it's gain updating, gain_st[] shouldn't be changed */
				/* because we need it to remove corresponding element in bucket */
		        if(initFlag==TRUE)
		        {
		        	pInst->gain_st[0]=pInst->new_gain_st[0];
		        	pInst->gain_st[1]=pInst->new_gain_st[1];
				}

				restoreImplication(trackStk);

			}else{

					/* the gate output depends on PIs' value */
					restoreImplication(trackStk);
				 }
		 }
}



void
report_gain()
{
	int	i;

	fprintf(dbgFp,"\n\n");
	fprintf(dbgFp,"The gain of all gates including PIs.\n");
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=DF_LIB_NO &&
			instList[i].instLibNo!=DFR_LIB_NO &&
			instList[i].instLibNo!=DFS_LIB_NO &&
			instList[i].instLibNo!=DFRS_LIB_NO)
		fprintf(dbgFp,"instance name: %s gain0 = %d gain1 = %d\n",
					   instList[i].name,
					   instList[i].gain_st[0],instList[i].gain_st[1]);
	}
}
