#include "PWcir.h"
#include "linearExt.h"
#include "linearFree.h"

#include "PWglobal_var.h"
#include "PWroutines.h"

extern void     InWeightQuickSort(double InWeight[][2], int low, int up);
extern void 	inputListSort(double InWeight[][2], PW_INST *inst, int inputNum, int assignmentInteger);
extern void 	reportInputSortResult(void);
extern void 	updtInputOrderWholeCircuit(lin_table_rec addList, bi_lin_table_rec reArrangeList);
extern void 	PWgateListInputSort(bi_lin_table_rec list);


/* sorting gate's input pins for backtracing */
/* For different gate output assignment, we have different sorting list*/
void
PWinputSort()
{   
	int	k;

    /* FOR EACH INSTANCE EXCEPT FLIP-FLOP, WE HAVE ORDERED INPUTS */
	for(k=0;k<PWinstNum_and_pseudo;k++)
	{	
		if( instList[k].instLibNo!=DF_LIB_NO &&
			instList[k].instLibNo!=DFR_LIB_NO &&
			instList[k].instLibNo!=DFS_LIB_NO &&
			instList[k].instLibNo!=DFRS_LIB_NO &&
			fetchImplyValueInStk(&(instList[k])) == VALUE_UNKNOWN)
			PWinputSortOneGate(&instList[k]);
	}

	/*debugging*/
	/* reportInputSortResult(); */


}

/* non-increasing order */
/* In case that some inputs are from root gates, 
the shifted gain is restored for input pins sorting */
void
PWinputSortOneGate(q)
PW_INST	*q;
{
	int     i,j,k;
	double  InWeight[MAX_IN_PER_GATE][2];

		/* Flip/flop, pseudo-PI may also be filtered here if we consider them*/
		if(q->instLibNo==PI_LIB_NO ||
		   q->instLibNo==PO_LIB_NO)
			return;

		/* for 2 assignment types */
		for(j=0;j<2;j++)
		{
            for(i=0;i<MAX_IN_PER_GATE;i++)
            {   
            	/* the initial WEIGHT for gate's input*/
                InWeight[i][0]=-99999;
            }

            k=0;
            for(i=0;i<q->fanin;i++)
            {   
                /* for unassigned inputs */
                if(fetchImplyValueInStk(q->in[i])==VALUE_UNKNOWN)
                {
                	/* gain value */
                	if(q->in[i]->MFFC_root==TRUE &&
                	   q->in[i]->gain_st[j]!=INVALID_GAIN)
                	{
                		/* INVALID_GAIN is a flag, and cannot be changed */
                		InWeight[k][0]=q->in[i]->gain_st[j]-gainShiftAmount;
					}else{
                			InWeight[k][0]=q->in[i]->gain_st[j];
						 }
					/* input pin position*/
                	InWeight[k][1]=i;
                	k++;
				}
            }

            /* the input ID <i>: starting from 0 to i-1 */
            /* sort in nondecreasing order */
            InWeightQuickSort(InWeight,0,k-1);

            /****** the ordered list is in nonincreasing order ******/
            /******  because PRE insertion is used      	   ******/
            inputListSort(InWeight,q,k,j);

            /***
            if(DEBUG_INFO)
            {
                fprintf(dbgFp,"Inst:%s 's Input Backtrack Order \n",
                        q->name);
                j=0;
                dbg=q->InSortedList;
                while(j==0 || dbg!=q->InSortedList)
                {   fprintf(dbgFp,"\tOrder=%d\tInputId=%d\tweight=%g\n",
                            j+1,dbg->elem,InWeight[--i][0]);
                    dbg=dbg->next;
                    j++;
                }
            }
            ***/
		}
}

/* nondecreasing quick sort */
void    InWeightQuickSort(InWeight,low,up)
double  InWeight[][2];
int     low,up;
{   int     i,j;
    double  w,id;

    if(low<up)
    {   i=low;
        j=up;

        w=InWeight[low][0];
        id=InWeight[low][1];
        while(i!=j)
        {   while(i<j && InWeight[j][0]>w)
                j--;
            if(i<j)
            {   InWeight[i][0]=InWeight[j][0];
                InWeight[i++][1]=InWeight[j][1];
            }
            while(i<j && InWeight[i][0]<w)
                i++;
            if(i<j)
            {   InWeight[j][0]=InWeight[i][0];
                InWeight[j--][1]=InWeight[i][1];
            }
        }
        InWeight[i][0]=w;
        InWeight[i][1]=id;
        InWeightQuickSort(InWeight,low,i-1);
        InWeightQuickSort(InWeight,i+1,up);
    }
}

/* sorted list of inputs for different logic value assignment */
/* indice of the gate inputs are stored in the list */
void
inputListSort(InWeight,inst,inputNum,assignmentInteger)
double	InWeight[][2];
PW_INST	*inst;
int		inputNum;
int		assignmentInteger;
{
	int	i;

	bi_lin_table_rec_Free1(&(inst->InSortedList[assignmentInteger]),NULL);

	/* if inputNum<0, then no x-path input */
	for(i=0;i<inputNum;i++)
	{
		bi_lin_table_prev_ins(&(inst->InSortedList[assignmentInteger]),NULL,
								(double)(InWeight[i][1]),NULL);
	}
}

void
reportInputSortResult()
{
	int	i,j,k;
	bi_lin_table_rec	dbg;

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo==DF_LIB_NO ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO )
		   	continue;

        fprintf(dbgFp,"Inst:%s 's Input Backtrack Order \n",
                instList[i].name);
        for(k=0;k<2;k++)
        {
                fprintf(dbgFp,"for assignemnt of %d:\n",k);
                dbg=instList[i].InSortedList[k];
				j=0;

                while(dbg!=NULL)
                {   fprintf(dbgFp,"\tOrder=%d\tInputId=%d\tinput_inst=%s\n",
                            j+1,(int)(dbg->key),instList[i].in[(int)(dbg->key)]->name);
                    dbg=dbg->next;
                    j++;
                }
		}
	}
}


void
PWgateListInputSort(list)
bi_lin_table_rec   list;
{
    bi_lin_table_rec   	p;
    PW_INST         	*inst;

    p=list;
    while(p!=NULL)
    {
        inst=(PW_INST   *)(p->member);

        PWinputSortOneGate(inst);

        p=next_bi_lin_table_rec(p);
    }
}

void
updtInputOrderWholeCircuit(addList,reArrangeList)
lin_table_rec		addList;
bi_lin_table_rec	reArrangeList;
{
	lin_table_rec	p;
	PW_INST			*inst;

	/* it's a bi-linear table structure */
	PWgateListInputSort(reArrangeList);

	/* it's a linear table structure */
	p=addList;
	while(p!=NULL)
	{
		inst=(PW_INST   *)(p->member);

		PWinputSortOneGate(inst);

		p=next_lin_table_rec(p);
	}
}
