#include "PWcir.h"
#include "linearFree.h"
#include "linearExt.h"

#include "PWglobal_var.h"
#include "PWroutines.h"

extern void     InWeightQuickSort();
extern void 	inputListSort_fanout(double InWeight[][2], PW_INST *inst, 
									int inputNum, int assignmentInteger);
extern void 	reportInputSortResult(void);
extern void		PWinputSortOneGate_fanout(PW_INST *q);
extern void 	reportInputSortResult_fanout(void);
extern void 	PWgateListInputSort_fanout(bi_lin_table_rec   list);
extern void 	updtInputOrderByFanoutForAllGates_fanout(PW_INST *array, int     num);




/* Gate's input sorting for backtracing */
/* For each gate's output assignment, we have its own sorting list*/
void
PWinputSort_fanout()
{   
	int	k;

    /* FOR EACH NON-SEQUENTIAL INSTANCE,WE HAVE ORDERED INPUTS */
	for(k=0;k<PWinstNum;k++)
	{	
		if( instList[k].instLibNo!=DF_LIB_NO &&
		    instList[k].instLibNo!=DFR_LIB_NO &&
		    instList[k].instLibNo!=DFS_LIB_NO &&
		    instList[k].instLibNo!=DFRS_LIB_NO &&
		    fetchImplyValueInStk(&(instList[k])) == VALUE_UNKNOWN )
			PWinputSortOneGate_fanout(&instList[k]);
	}

	/* debugging
	reportInputSortResult();
	*/

    /** FOR PSEUDO PI ,SIMILAR SORT LIST NEEDED **/
    /**
    for(i=PSEUDO_PI_index;i<PrimInNum;i++)
    {   InWeight[0][0]=0;
    **/
        /** only one input **/
        /**
        InWeight[0][1]=1;

        formInSortedList(InWeight,SourceP[i],1);
    }
    **/

}

/* non-increasing order */
/* according to fanout */
void
PWinputSortOneGate_fanout(q)
PW_INST	*q;
{
	int     i,j,k;
	double  InWeight[MAX_IN_PER_GATE][2];

		/* Flip/flop, pseudo-PI should also be filtered here if we consider them*/
		if(q->instLibNo==PI_LIB_NO ||
		   q->instLibNo==PO_LIB_NO)
			return;

		/* for 2 assignment types */
		for(j=0;j<2;j++)
		{
            for(i=0;i<MAX_IN_PER_GATE;i++)
            {   
            	/* the WEIGHT for gate's input*/
                InWeight[i][0]=-99999;
            }

            k=0;
            for(i=0;i<q->fanin;i++)
            {   
                /* for unassigned inputs */
                if(fetchImplyValueInStk(q->in[i])==VALUE_UNKNOWN)
                {
                	InWeight[k][0]=q->in[i]->fanout;
                	InWeight[k][1]=i;
                	k++;
				}
            }

            /* the input ID <i>: from 0 to i-1 */
            /* sort in nondecreasing order */
            InWeightQuickSort(InWeight,0,k-1);

            /* if j=0(assign 0); POST insertion and nondecreasing order */
            /* if j=1(assign 1); PRE insertion and nonincreasing order */
            inputListSort_fanout(InWeight,q,k,j);

            /***
            if(DEBUG_INFO)
            {
                fprintf(dbgFp,"Inst:%s 's Input Backtrack Order \n",
                        q->name);
                j=0;
                dbg=q->InSortedList;
                while(j==0 || dbg!=q->InSortedList)
                {   fprintf(dbgFp,"\tOrder=%d\tInputId=%d\tweight=%g\n",
                            j+1,dbg->elem,InWeight[--i][0]);
                    dbg=dbg->next;
                    j++;
                }
            }
            ***/
		}
}


/* sorted list for inputs and different assignment type*/
/* input index of the gate is stored in the list */
void
inputListSort_fanout(InWeight,inst,inputNum,assignmentInteger)
double	InWeight[][2];
PW_INST	*inst;
int		inputNum;
int		assignmentInteger;
{
	int	i;
	bi_lin_table_rec	tail=NULL;

	bi_lin_table_rec_Free1(&(inst->InSortedList[assignmentInteger]),NULL);

  if(assignmentInteger==1)
  {
	/* if inputNum<0, then no x-path input */
	for(i=0;i<inputNum;i++)
	{
		bi_lin_table_prev_ins(&(inst->InSortedList[assignmentInteger]),NULL,
								(double)(InWeight[i][1]),NULL);
	}
  }else{

  		tail=NULL;
  		for(i=0;i<inputNum;i++)
  		{
  			bi_lin_table_post_ins(&(inst->InSortedList[assignmentInteger]),
  									&tail,
  									(double)(InWeight[i][1]),NULL);
		}
	   }
}

void
reportInputSortResult_fanout()
{
	int	i,j,k;
	bi_lin_table_rec	dbg;

	for(i=0;i<PWinstNum;i++)
	{
		if(instList[i].instLibNo==DF_LIB_NO ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO )
		   continue;

        fprintf(dbgFp,"Inst:%s 's Input Backtrack Order \n",
                      instList[i].name);
        for(k=0;k<2;k++)
        {
                fprintf(dbgFp,"for assignemnt of %d:\n",k);
                dbg=instList[i].InSortedList[k];
				j=0;

                while(dbg!=NULL)
                {   fprintf(dbgFp,"\tOrder=%d\tInputId=%d\tinput_inst=%s\n",
                            j+1,(int)(dbg->key),instList[i].in[(int)(dbg->key)]->name);
                    dbg=dbg->next;
                    j++;
                }
		}
	}
}


void
PWgateListInputSort_fanout(list)
bi_lin_table_rec   list;
{
    bi_lin_table_rec   	p;
    PW_INST         	*inst;

    p=list;
    while(p!=NULL)
    {
        inst=(PW_INST   *)(p->member);

        PWinputSortOneGate_fanout(inst);

        p=next_bi_lin_table_rec(p);
    }
}

void
updtInputOrderByFanoutForAllGates_fanout(array,num)
PW_INST	*array;
int		num;
{
	int	i;
	for(i=0;i<num;i++)
	{
		PWinputSortOneGate_fanout(&(array[i]));
	}
}
