#include "PWcir.h"
#include "linearFree.h"
#include "linearExt.h"

#include "PWglobal_var.h"
#include "PWroutines.h"
#include "PWilib_routines.h"

extern void     InWeightQuickSort(double InWeight[][2], int low, int up);
extern void     reportInputSortResult_ilib(void);

/* Sorting gate's input pins for backtracing 
 * accroding to the power-up current in the library.
 */
void
PWinputSort_ilib()
{
    int k;

    /* FOR EACH NON-SEQUENTIAL INSTANCE,WE HAVE ORDERED INPUTS */
    for(k=0;k<PWinstNum_and_pseudo;k++)
    {
        if( instList[k].instLibNo!=DF_LIB_NO &&
            instList[k].instLibNo!=DFR_LIB_NO &&
            instList[k].instLibNo!=DFS_LIB_NO &&
            instList[k].instLibNo!=DFRS_LIB_NO &&
            fetchImplyValueInStk(&(instList[k])) == VALUE_UNKNOWN )
            PWinputSortOneGate_ilib(&instList[k]);
    }
	reportInputSortResult_ilib();
}

void
PWinputSortOneGate_ilib(PW_INST	*q)
{
    int     i,j,k;
    double  InWeight[MAX_IN_PER_GATE][2];

        /* Flip/flop, pseudo-PI should also be filtered here if we consider them*/
        if(q->instLibNo==PI_LIB_NO ||
           q->instLibNo==PO_LIB_NO)
            return;

        /* for 2 assignment types: */
        /* j==0 : stable0 */
        /* j==1 : rise    */
        for(j=0;j<2;j++)
        {
            for(i=0;i<MAX_IN_PER_GATE;i++)
            {
                /* the WEIGHT for gate's input*/
                InWeight[i][0]=-99999;
            }

            k=0;
            for(i=0;i<q->fanin;i++)
            {
                /* for unassigned inputs: weight is current */
                if(fetchImplyValueInStk(q->in[i])==VALUE_UNKNOWN)
                {
                    InWeight[k][0]=q->in[i]->powerup_current->current[j];
                    InWeight[k][1]=i;
                    k++;
                }
            }

            /* the input ID <i>: from 0 to i-1 */
            /* sort in nondecreasing order */
            InWeightQuickSort(InWeight,0,k-1);

            /* if j=0(assign 0); */
            /* if j=1(assign 1); PRE insertion and nonincreasing order */
            inputListSort_ilib(InWeight,q,k,j);

        }
}

/* sorted list for inputs and different assignment type*/
/* input index of the gate is stored in the list */
void
inputListSort_ilib(InWeight,inst,inputNum,assignmentInteger)
double  InWeight[][2];
PW_INST *inst;
int     inputNum;
int     assignmentInteger;
{
	int 				i;
	bi_lin_table_rec	tail=NULL;

	bi_lin_table_rec_Free1(&(inst->InSortedList[assignmentInteger]), NULL);

	/* for both assignments, use pre-insertion to get nonincreasing order */
    /* if inputNum<0, then no x-path input */
    for(i=0;i<inputNum;i++)
    {
        bi_lin_table_prev_ins(&(inst->InSortedList[assignmentInteger]),NULL,
                                (double)(InWeight[i][1]),NULL);
    }
}

void
reportInputSortResult_ilib()
{
    int i,j,k;
    bi_lin_table_rec    dbg;

    for(i=0;i<PWinstNum_and_pseudo;i++)
    {
        if(instList[i].instLibNo==DF_LIB_NO ||
           instList[i].instLibNo==DFR_LIB_NO ||
           instList[i].instLibNo==DFS_LIB_NO ||
           instList[i].instLibNo==DFRS_LIB_NO )
            continue;

        fprintf(dbgFp,"Inst:%s 's Input Backtrack Order \n",
                instList[i].name);
        for(k=0;k<2;k++)
        {
                fprintf(dbgFp,"for assignemnt of %d:\n",k);
                dbg=instList[i].InSortedList[k];
                j=0;

                while(dbg!=NULL)
                {   fprintf(dbgFp,"\tOrder=%d\tInputId=%d\tinput_inst=%s\tcurrent=%f\n",
                            j+1,(int)(dbg->key),instList[i].in[(int)(dbg->key)]->name,
							instList[i].in[(int)(dbg->key)]->powerup_current->current[k]);
                    dbg=dbg->next;
                    j++;
                }
        }
    }
}

