#include "PWcir.h"

#include "PWglobal_var.h"
#include "PWroutines.h"


extern PW_INST *PWgetLevelItem(PW_INST *list, int levNo);
extern void 	levAllGateOut(PW_INST *instanceP, PW_INST ** levelList, PW_INST *output[]);


/* label the level number for each gate */
void PWlabelLevel()
{
	PW_INST * curLevel,*tmpLevel,*instanceP;
	int		i;

	curLevel=NULL;
	tmpLevel=NULL;

	/** ALL PI(include pseudo) ARE OF THE SAME LEVEL **/
	for(i=0;i<PrimInNum;i++)
		PWinstListPREInsert(&curLevel,SourceP[i]);
	while(curLevel!=NULL)
	{
		while((instanceP=PWinstListPREPop(&curLevel))!=NULL)
		{	
			/* label level number */
			/* gates for the next level is put into tmpLevel */
			levAllGateOut(instanceP,instanceP->out,&tmpLevel);
			levAllGateOut(instanceP,instanceP->outz,&tmpLevel);
		}
		curLevel=tmpLevel;
		tmpLevel=NULL;
	}

	/* debugging info */
	/***
	if(DEBUG_INFO)
	{	
	***/
		for(i=0;i<PWinstNum_and_pseudo;i++)
		{	fprintf(dbgFp,"%d:Instance %s -- maxLevel=%d,minLevel=%d\n",
					i,instList[i].name,instList[i].maxLev,instList[i].minLev);
		}
		fflush(dbgFp);
	/***
	}
	***/
}

/** assign level # to all fanouts of the gate, and put them into <levelList> **/
/** You may select Q or QZ for Dflip-flop **/
void
levAllGateOut(instanceP,output,levelList)
PW_INST *instanceP,** levelList;
PW_INST *output[];
{	int i,j,k;
	PW_INST *tmp;

    for(i=0;i<(instanceP->fanout);i++)
    {
        if((tmp=output[i])==NULL )
            break;
		else if(tmp->instLibNo==DF_LIB_NO  ||
				tmp->instLibNo==DFR_LIB_NO ||
				tmp->instLibNo==DFS_LIB_NO ||
				tmp->instLibNo==DFRS_LIB_NO)
			 	/** flip-flops don't have level numbers **/
			 	continue;
        	 else{      
        	 		/** compute min level number **/

						/* MUX21 is not supported now */
						/* control path doesn't propagate signal */
						/*****
						if(strcmp(tmp->name,"MUX21")==0&&i==2)
								continue;
						******/

	        		if(tmp->minLev<0)
               			tmp->minLev=(instanceP->minLev+1);
            		else{
                 			tmp->minLev=DF_MIN(tmp->minLev,
                                    			instanceP->minLev+1);
                		}
				 }

 		/* <visited>=FALSE initially */
 		/* <visited>=TRUE if the gate's max level has been obtained */
		if(tmp->visited==TRUE) 
			continue;
		else{
				/** compute maxLevel number **/
        		k=-1;
        		for(j=0;j<MAX_IN_PER_GATE;j++)
        		{   if(tmp->in[j]==NULL)    
						break;

					/* MUX21 is not supported now */
					/** eliminate control input **/
					/********************************
					else if(strcmp(tmp->name,"MUX21")==0&&i==2)
							continue;
					***********************************/

					else if(tmp->in[j]->maxLev<0)
				 		{	k=-1;
							break;
				 		}else {
            					k=DF_MAX(tmp->in[j]->maxLev,k);
					   		}
        		}
        		if(k>=0)
        		{  	tmp->maxLev=k+1; 
					PWinstListPREInsert(levelList,tmp);
					tmp->visited=TRUE;
        		}
			}
	}
}


/* check assumption for circuit levelization */
void
PWchkLevAssump()
{	int i,k[100],count=0;

	for(i=0;i<PWinstNum_and_pseudo && count<100;i++)
	{	if(instList[i].instLibNo==DF_LIB_NO  ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO)
			continue;

		/*if any asynchronous feedback, some gates' level number may not be obtained*/
		if((instList[i].maxLev<0 || instList[i].minLev<0)&&
			instList[i].instLibNo>0)
		k[count++]=i;
	}
	if(count==0)
	{	fprintf(stdout,"\nTHE CIRCUIT PASSED THE SIMPLE CHECK:\n");
		fprintf(stdout,"No feedbacks after removing flip-flops.\n\n");
		return ;
	}
	else {

			fprintf(stdout,"\n\nThe circuit doesn't satisfy the assumption:\n");
			fprintf(stdout,"\t\t\tMaybe input file IOinfo has incorrect <space> OR\n");
			fprintf(stdout,"\t\t\tMaybe there are feedback loops among following:\n"); 
			fprintf(stdout,"Following are some instance which may have some problems.\n");
			for(i=0;i<count;i++)
			{	fprintf(stdout,"line:%d instname:%s \n",k[i],instList[k[i]].name);
			}
		}
}

/* construct a group of lists, each of which is for gates with the same level number */
void
PWformLevList()
{	int 	cirMaxLev=0;
	int 	curLev,i,j;
	PW_INST *p,*q;

	LevList=NULL;

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{	
		/** include PSEUDO PI **/
		if(instList[i].instLibNo==PI_LIB_NO ||
		   instList[i].instLibNo==PO_LIB_NO
			|| instList[i].instLibNo<0)
			continue;
		if(instList[i].instLibNo==DF_LIB_NO  ||
		   instList[i].instLibNo==DFR_LIB_NO ||
		   instList[i].instLibNo==DFS_LIB_NO ||
		   instList[i].instLibNo==DFRS_LIB_NO)
		   continue;

		/*	curLev: the circuit level # of the gate that is the max level # */
		/*	cirMaxLev: the largest level head-pointer ever created */
		curLev=instList[i].maxLev;	
		if(curLev>cirMaxLev)
		{	/** CREATE ALL NECESSARY NEW LEVEL HEAD POINTER **/
			for(j=0;j<(curLev-cirMaxLev);j++)
			{	
				p=(PW_INST *) mem_get_block((SIZE_T) sizeof(PW_INST));
				p->next=NULL;
				p->out[0]=NULL;
				PWinstListPOSTInsert(&LevList,p);
			}
			cirMaxLev=curLev;
		}

		p=PWgetLevelItem(LevList,curLev);
		PWinstListPREInsert(&(p->out[0]),&instList[i]);
	}

	/* debugging info */
	/***
	if(DEBUG_INFO)
	{	
	****/
		fprintf(dbgFp,"The combinational circuit is levelized:\n");
		i=1;	
		for(p=LevList;p!=NULL;p=p->next)
		{	q=p->out[0];	
			/* the definition of circuit level is maxLev */
			fprintf(dbgFp,"Circuit Level %d:\n",i);
			while(q!=NULL)
			{	fprintf(dbgFp,"InstName:	%s\n",q->name);
				q=q->next;
			}
			i++;
		}
		fflush(dbgFp);
	/***
	}
	***/
}

/* get the head pointer dedicated for a specific level # from level list */
PW_INST *PWgetLevelItem(list,levNo)
PW_INST *list;
int		levNo;
{	int i;
	for(i=1;i<levNo;i++)  /* if levNo=1,then the pointer needn't move along list**/
		list=list->next;
	return list;
}
