/***********************************************************
	Primitive Logic Operation for:
		3-value Simuation and gate evaluation
************************************************************/

#include "PWcir.h"
#include "PWroutines.h"
#include "PWgateLogic.h"

extern 	LOGIC_VALUE gateOutputValue(PW_INST *pInst, int sourceType);

/* swap for 3-value enconding */
extern LOGIC_VALUE swapHalfByte(LOGIC_VALUE var);

LOGIC_VALUE
not_operation(LOGIC_VALUE var)
{
	return swapHalfByte(~var);
}

LOGIC_VALUE
xor_operation(LOGIC_VALUE varA, LOGIC_VALUE varB)
{
	LOGIC_VALUE	Abar,Bbar;

	Abar=not_operation(varA);
	Bbar=not_operation(varB);
	return ((Abar&varB) | (Bbar&varA));
}

LOGIC_VALUE
xnor_operation(LOGIC_VALUE varA, LOGIC_VALUE varB)
{
	LOGIC_VALUE tmpVar;

	tmpVar=xor_operation(varA,varB);
	return not_operation(tmpVar);
}

LOGIC_VALUE
swapHalfByte(LOGIC_VALUE var)
{
	LOGIC_VALUE	tmp;

	tmp=var & ((LOGIC_VALUE)15);
	tmp=tmp<<4;

	tmp=tmp | ((var>>4)&15);

	return tmp;
}

/* 'sourceType' denotes what date-stru the inputs come from */
/* currently, we only use "IMPLY_STK_TYPE" as the sourceType */
/* evluate the gate output */
LOGIC_VALUE
evaluateGate(pInst,sourceType)
PW_INST	*pInst;
int		sourceType;
{
	int			i;
	LOGIC_VALUE	var;
	LOGIC_VALUE result;

	switch(pInst->instLibNo)
	{
		case NOT_LIB_NO:
			var=gateOutputValue(pInst->in[0],sourceType);
			result=not_operation(var);
    	    break;
		case BUFF_LIB_NO:
			var=gateOutputValue(pInst->in[0],sourceType);
			result=var;
			break;

        case AND_LIB_NO:
        	var=VALUE_1;
        	for(i=0;i<pInst->fanin;i++)
        	{
        		var=var & gateOutputValue(pInst->in[i],sourceType);
			}
			result=var;
			break;

        case NAND_LIB_NO:
        	var=VALUE_1;
        	for(i=0;i<pInst->fanin;i++)
        	{
        		var=var & gateOutputValue(pInst->in[i],sourceType);
			}
			result=not_operation(var);
			break;

        case OR_LIB_NO:
        	var=VALUE_0;
        	for(i=0;i<pInst->fanin;i++)
        	{
        		var=var | gateOutputValue(pInst->in[i],sourceType);
			}
			result=var;
			break;

        case NOR_LIB_NO:
        	var=VALUE_0;
        	for(i=0;i<pInst->fanin;i++)
        	{
        		var=var | gateOutputValue(pInst->in[i],sourceType);
			}
			result=not_operation(var);
            break;

        /* only XOR2 */
        case XOR_LIB_NO:
        	result=xor_operation(gateOutputValue(pInst->in[0],sourceType),
        						 gateOutputValue(pInst->in[1],sourceType));
            break;

        /* only XNOR2 */
        case XNOR_LIB_NO:
        	result=xnor_operation(gateOutputValue(pInst->in[0],sourceType),
        						  gateOutputValue(pInst->in[1],sourceType));
            break;
        case PO_LIB_NO:
        	result=gateOutputValue(pInst->in[0],sourceType);
        	break;
        /*case PI_LIB_NO:*/
        default:
        	fprintf(stderr,"Not valid gate type for evaluation in evaluateGate().\n");
        	exit(1);
        	break;
	}
	return result;
}

/* access the output value of the current gate,
   if it's already determined
*/
LOGIC_VALUE
gateOutputValue(pInst,sourceType)
PW_INST		*pInst;
int			sourceType;
{
	if(sourceType==STATE_TYPE)
	{
		return pInst->state;

	}else if(sourceType==IMPLY_STK_TYPE)
		  {
		  	return (*((LOGIC_VALUE *)(pInst->implyStkTop->member)));

		  }else{
		  		fprintf(stderr,"Incorrect sourceType in gateOutputValue().\n");
		  		exit(1);
			   }
}

/* conversion between integer and logic value */
LOGIC_VALUE
int2logic(i)
int i;
{
    if(i==0)
        return VALUE_0;
    else if(i==1)
        return VALUE_1;
         else{
                fprintf(stderr,"Input out of range in int2logic().\n");
                exit(1);
             }
}

int
logic2int(LOGIC_VALUE l)
{
    if(l==VALUE_0)
        return 0;
    else if(l==VALUE_1)
            return 1;
         else{
                fprintf(stderr,"Input out of range in logic2int().\n");
                exit(1);
             }
}

