#include <stdio.h>
#include "PWcir.h"
#include "PWinitData.h"

extern void reportMaxCharge(BOOLEAN write_file_flag);
extern void PWinitMFFC_Id_Array();
extern void PWrand_vec_simple_sim();
extern void PWmcSim();


main(argc,argv)
int argc;
char *argv[];
{
  	FILE        * fin = stdin;
  	char		fname[200];
  	char		fout_name[207];
  	int			pos;
	int			prog_func;
	float		confd,err;


  	if(argc!=3)
  	{	fprintf(stderr,"Usage: PWmain <func_command> <filename>\n");
		fprintf(stderr,"       available func_command:\n");
		fprintf(stderr,"       gain-c  -- ATPG-based estimation -- Imax/Gain-cluster\n");
		fprintf(stderr,"       sim   -- 1vector random simulation for power-up current\n");
		fprintf(stderr,"       swsim -- random simulation for normal switching current.\n");
		fprintf(stderr,"       MCsim -- Monte Carlo simulation\n");
		exit(1);
	}
	if(strcmp(argv[1],"sim")==0)
		prog_func = LOGIC_SIM_1VEC;
	else if(strcmp(argv[1],"gain-c")==0)
		prog_func = ATPG_BASED_ESTIMATION_GAIN_C;
		 else if(strcmp(argv[1],"MCsim")==0)
		 		prog_func = MC_LOGIC_SIM_1VEC;
		 	else if(strcmp(argv[1],"swsim")==0)
		 			prog_func = SW_RANDOM_SIM;
	     		else {
					fprintf(stderr,"Unknown <func_command>.\n");
					exit(1);
		  			 }

	strcpy(fname,argv[2]);
	strcpy(fout_name,fname);
	pos=findLastSpecChar(fout_name,'/');
	while(pos!=-1)
	{
		strCutBehindSpecChar(fout_name,'/');
		pos=findLastSpecChar(fout_name,'/');
	}
	strcat(fout_name,".result");

	/* debugging info output */
	FOPEN(dbgFp,"dbgInfo","w");
	FOPEN(resultFp, fout_name,"w");

	SetTime(TIME_REPORT);
	/* read netlist */
	NetRead(fname);	

	fprintf(stdout,"Levelizing Ckt.\n");
	/* label gate with level number */
	PWlabelLevel();

	/* check the assumption: 
	   the ckt after removing flip-flops has no feedbacks and is levelizable.
	*/
	PWchkLevAssump();

	/* construct the level list */
	PWformLevList();

	fflush(dbgFp);
	fprintf(stdout,"Levelization finished.\n");
	ReportTime(TIME_REPORT);	/* report time for reading and levelization */

	if(prog_func == ATPG_BASED_ESTIMATION_GAIN_C)
	{
	  /* clustering ckt using MFFC */
	  fprintf(stdout,"MFFC Clustering starting...\n");
	  PWclusterMFFC(instList,PWinstNum_and_pseudo,SinkP,PrimOutNum,
				&MFFClistHead,&MFFClistTail,
				&MFFCrootList,&nonMFFCrootList);
	  fflush(dbgFp);

	  fprintf(stdout,"MFFC Clustering finished.\n");
	  ReportTime(TIME_REPORT);

	  /*  initial gain assignment to gates */
	  fprintf(stdout,"Initialization for Estimation...\n");
	  PWgateInitGain();

	  /* establish the bucket using the initial gain assignment */
	  /* bucket is for dynamic re-ordering of gates */
	  PWbucketInit();
	  fflush(dbgFp);

	  /* ordering the gates' input pins */
	  PWinputSort();
	  fflush(dbgFp);
	  fprintf(stdout,"Initialization finished.\n");
	  ReportTime(TIME_REPORT);

	  /* ATPG-based stored Charge Maximization for power-up current: Imax/gain-cluster */
	  fprintf(stdout,"Estimation starting...\n");
	  PWmaxCharge_gain_c();

	  /* report the max stored charge, which is the metric for power-up current */
	  reportMaxCharge(TRUE);

	  fclose(dbgFp);
	  fprintf(stdout,"Estimation finished.\n");

	}else if(prog_func == LOGIC_SIM_1VEC)
		  {
			/* 1vec random simulation for power-up current */
			PWrand_vec_simple_sim();
			ReportTime(TIME_REPORT);
	      }else if(prog_func == MC_LOGIC_SIM_1VEC)
	            {
	      			fprintf(stdout,"Enter confidence(e.g. 0.95 for 95%):");
	      			scanf("%f",&confd);
	      			fprintf(stdout,"Error bound( e.g. 0.02 for 2%):");
	      			scanf("%f",&err);
	      			/* Monte Carlo Simulation */
	      			PWmcSim(confd,err);
			    }else{	/* random simulation for normal switching current */
			    		PWnormSwitchingSim();
			   	     }

	ReportTime(TIME_REPORT);
}
