#include "PWcir.h"
#include "linear.h"
#include "linearExt.h"
#include "linearFree.h"

#include "logicOpt.h"
#include "PWroutines.h"
#include "PWglobal_var.h"


extern int 			JustifyAssignment(PW_INST *pInst, LOGIC_VALUE assignment, BOOLEAN realJustifyFg);
extern void 		gateListForUpdtGain(stk_rec assigned_gate_list, bi_lin_table_rec *reArrange_gate_list);
extern void 		updtGateListGain(bi_lin_table_rec *reArrangeList, stk_rec *assignedList);
extern void 		PWgateListInputSort();
extern void 		manipulateBucketUpdtList(bi_lin_table_rec *reArrangeList, stk_rec *assignedList, stk_rec *trackGateList);
extern void 		accessPIbacktrackStkTop(stk_rec stack, PW_INST **targetPI, LOGIC_VALUE *assignmentPI, double *triedTimes);
extern void 		nextDecision(PW_INST **targetPI, LOGIC_VALUE *assignmentPI, double *triedTimes);
extern 	bi_lin_table_rec get_bi_lin_table_rec_tail(bi_lin_table_rec head);
extern void			rebuildBucket(void);
extern void 		restoreBySimulation(stk_rec PIbacktrackStk);
extern void 		getBacktraceInput(PW_INST *pInst, bi_lin_table_rec candidate[]);
extern void 		gateStackGarbageClean(PW_INST *pInst);
extern void 		labelMFFCid_to_correct(stk_rec assigned_gate_list, BOOLEAN MFFCid[], int numberofMFFC);
extern void 		copyGain(bi_lin_table_rec reArrangeList);
extern int			reportMaxCharge(BOOLEAN write_file_flag);
extern void			Backtrace(PW_INST *pInst, LOGIC_VALUE assignment, PW_INST **targetPI, LOGIC_VALUE *assignmentPI);
extern void 		clearVisitedFlag(void);
extern void 		PWinitMFFC_Id_Array(BOOLEAN MFFC_Id_Array[]);



/* get the input for backtracing path */
void
getBacktraceInput(pInst,candidate)
PW_INST				*pInst;
bi_lin_table_rec	candidate[];
{
	bi_lin_table_rec	tmpRec;
	int					pinID;

	tmpRec = pInst->InSortedList[0];
	while(tmpRec !=NULL)
	{
		pinID = (int)(tmpRec->key);
		if(fetchImplyValueInStk(pInst->in[pinID]) == VALUE_UNKNOWN)
			break;
		tmpRec = next_bi_lin_table_rec(tmpRec);
	}
	candidate[0] = tmpRec;

	tmpRec = pInst->InSortedList[1];
	while(tmpRec !=NULL)
	{
		pinID = (int)(tmpRec->key);
		if(fetchImplyValueInStk(pInst->in[pinID]) == VALUE_UNKNOWN)
			break;
		tmpRec = next_bi_lin_table_rec(tmpRec);
	}
	candidate[1] = tmpRec;
}

/* Imax/Gain-cluster: ATPG-based Max. Power-up Current Estimation */
void
PWmaxCharge_gain_c(void)
{
	PW_BUCKET_ELEM	*elem;
	int				i=0;

	while(maxGainIndex>=0)	/* Bucket not empty */
	{	/* gate that has the largest gain and needs to be assigned with a logic value */
		elem=accessMaxBucketElem(PWbucketList,
								 maxGainIndex);

		/* justification of assignment */
		JustifyAssignment(&(instList[elem->instID]),elem->assignment,TRUE);
		i++;
		fprintf(stdout,"The gate No. %d is justified.\n",i);

	}
}


/* justify the specified assignment */
/* if <realJustifyFg> = TRUE, it means it's a real justification.  
the returned value will not be used*/
/* otherwise, it's for MFFC root gate gain computation (based on the MFFC) --> */
/*  in this version, the <realJustifyFg> is always used as TRUE */
int
JustifyAssignment(pInst,assignment, realJustifyFg)
PW_INST		*pInst;
LOGIC_VALUE	assignment;
BOOLEAN		realJustifyFg; /* flag: a real justification or 
					a gain calculation for MFFC root gates(restoration needed)*/
{
	LOGIC_VALUE		assignmentPI;
	PW_INST			*targetPI;
	BOOLEAN			conflict;
	BOOLEAN			uniqueValueDetected;
	BOOLEAN			backtrackFg,passFg;
	PW_IMPLY_STRU	*PIbacktrackElem=NULL;
	stk_rec			PIbacktrackStk=NULL;
	stk_rec			tmpRec;
	double			triedTimes;				
	int				time_counter=0;
	int				try_counter=0;
	int				backtrack_try=0;

	stk_rec				assigned_gate_list=NULL;
	bi_lin_table_rec	reArrange_gate_list=NULL;
	stk_rec				addList=NULL;
	stk_rec				removeList=NULL;
	bi_lin_table_rec	reArrangeList=NULL;


	int gain_MFFC_rootgate=0;

	/* failure flag of justification for value may be already implied */
	uniqueValueDetected=FALSE; 
								
	PIbacktrackElem=NULL;
	PIbacktrackStk=NULL;

	if(fetchImplyValueInStk(pInst) == assignment )
	{
		rebuildBucket();
		return 0;
	}

	/* mapping gate assignment to PI assignments & forward implication */
	while(fetchImplyValueInStk(pInst)!=assignment && uniqueValueDetected==FALSE)
	{
		if(fetchImplyValueInStk(pInst)!=VALUE_UNKNOWN)
		{
			/** in case that the bucket is out of order **/
			rebuildBucket();
			break;
		}


		if(time_counter >= 600 )
		{
			PWimplyStackGarbageClean();
			time_counter = 0;
		}
		else time_counter++;

		try_counter++;

		/* get one taget PI */
		Backtrace(pInst,assignment,&targetPI,&assignmentPI);

		/* currently selected PI hasn't trigger backtracking */
		backtrackFg=FALSE;
		passFg=FALSE; /* flag for successfully mapping one PI */

		/* push target into backtracking stack */
		PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
		PIbacktrackElem->inst=targetPI;
		PIbacktrackElem->value=assignmentPI;

		STK_PUSH(&PIbacktrackStk,(double)0,(pointer)PIbacktrackElem);

	    assigned_gate_list=NULL;
	    reArrange_gate_list=NULL;

	    /* mapping to one PI one time */
	    /* backtracking may be expected here */
	    while(PIbacktrackStk!=NULL && passFg==FALSE)
	    {
	  	  if((backtrack_try + try_counter)%60 == 0)
	  	  {
	  		  fprintf(stdout,"\n");
	  		  backtrack_try = 0;
	  		  try_counter = 0;
		  }

	  	  fprintf(stdout,".");
	  	  backtrack_try++;


		  /* access backtrack stack top */
		  /* get the PI for 3-value forward implication, */
		  /* which could be : the original mapped PI or tracktracked PI */
		  accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);


		  /* assignment of PI and IMPLICATION */
		  /* we don't have conflict if using PODEM, but we still need backtracking */
		  forwardImplyPI3V(targetPI,assignmentPI,&assigned_gate_list,&gain_MFFC_rootgate);

		  if(assignmentPI==VALUE_UNKNOWN)	/* it's a assignment restoration process */
		  {
			  /* remove the restored PI from backtrack stack */
			  tmpRec=STK_POP(&PIbacktrackStk);
			  mem_free_block((pointer)(tmpRec->member));
			  mem_free_block((pointer)(tmpRec));

			  if(PIbacktrackStk!=NULL)
			  {	/* still have chance to justify the assignment */
				  accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);
  
				  /* remove the PI in previous time from backtrack stack */
				  tmpRec=STK_POP(&PIbacktrackStk);
				  mem_free_block((pointer)(tmpRec->member));
				  mem_free_block((pointer)(tmpRec));

				  /* decide the next logic value for the PI */
				  nextDecision(&targetPI,&assignmentPI,&triedTimes);
				  PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
				  PIbacktrackElem->inst=targetPI;
				  PIbacktrackElem->value=assignmentPI;
				  STK_PUSH(&PIbacktrackStk,(double)triedTimes,(pointer)PIbacktrackElem);
			  }
			  /* else, the justification failed !!!*/

		  }else	/* the PI is assigned with either VALUE_0 or VALUE_1 */
		  {

			  /*one case:*/
				/* current combination of PIs' assignment failed to justify the gate assignment */
				/* backtracking needed */
			  if( (fetchImplyValueInStk(pInst)!=VALUE_UNKNOWN && fetchImplyValueInStk(pInst)!=assignment)
				  || (fetchImplyValueInStk(pInst)==VALUE_UNKNOWN && detectImplicitJustifyFailure(pInst,assignment)==TRUE))
			  {
				  backtrackFg=TRUE;

				  /* get the wrong decision */
				  accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);
  	
				  /* remove wrong decision from backtrack stack */
				  tmpRec=STK_POP(&PIbacktrackStk);
				  mem_free_block((pointer)(tmpRec->member));
				  mem_free_block((pointer)(tmpRec));

				  /* get next decision according to the previous wrong one */
				  /* push into backtrack stack for next time */
				  nextDecision(&targetPI,&assignmentPI,&triedTimes);
				  PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
				  PIbacktrackElem->inst=targetPI;
				  PIbacktrackElem->value=assignmentPI;
				  STK_PUSH(&PIbacktrackStk,(double)triedTimes,(pointer)PIbacktrackElem);
			  }else
			  {
				  /* the initially selected PI is passed */
				  /* or the backtracking is successful */
				  /* but justification may not be completed */
				  passFg=TRUE;
			  }
		  }
	    }



		if(PIbacktrackStk==NULL)
		{
			/* the justification failed under current situation */
			/* so the gate output has already been implicitely determined */
			/* but we don't detect until here because of incremental gain update and
			   implicitely logic value determination is detected only in gain updating */
		  if(realJustifyFg == TRUE)
		  {
			stackFree1(&assigned_gate_list);
			implyGate(pInst,not_operation(assignment),&assigned_gate_list,&conflict);
			uniqueValueDetected=TRUE;
			/* impossible case, fix the bug */
			if(conflict==TRUE)
			{
				fprintf(stderr,"Impossible value for %s, ID %d\n in JustifAssignment()",
						pInst->name,pInst->instID);
				exit(1);
			}

			addList=removeList=NULL;
			reArrangeList=NULL;
			/* initial gain updating */
			updtGainWholeCircuit(&addList,&removeList,&reArrangeList);

			/* one way to update buckets
			copyGain(reArrangeList);
			rebuildBucket();*/

			/* alternative way to update buckets */
			updtBucketWholeCircuit(addList,removeList,reArrangeList,PWbucketList,PWbucketTail);

			/* ordering inputs using restored(not-shifted) root gate gains */
			updtInputOrderWholeCircuit(addList,reArrangeList);

			stackFree1(&addList);
			stackFree1(&removeList);
			bi_lin_table_rec_Free1(&reArrangeList,NULL);
			bi_lin_table_rec_Free1(&reArrange_gate_list,NULL);
			stackFree1(&assigned_gate_list);


		  }else{
					/* this version (Imax/gain-cluster) doesn't go through this branch */
					fprintf(stderr,"This version donesn't need it!\n");
					gain_MFFC_rootgate = INVALID_GAIN;
					uniqueValueDetected=TRUE;
					exit(1);
			   }


		}else
		{
			/* at this branch, the selected PI should have been passed */
			if(passFg==FALSE)
			{
				fprintf(stderr,"Impossible situation in Justification.\n");
				fprintf(stderr,"Check backtracking process.\n");
				exit(1);
			}

		  if(backtrackFg==FALSE)
		  {
			/* get the gate list for updating */
			gateListForUpdtGain(assigned_gate_list,&reArrange_gate_list);

			/* update the gain of each gate in the list */
			updtGateListGain(&reArrange_gate_list,&assigned_gate_list);
			PWbucketUpdt(assigned_gate_list,reArrange_gate_list);

			/* re-sort the inputs */
			PWgateListInputSort(reArrange_gate_list);

			bi_lin_table_rec_Free1(&reArrange_gate_list,NULL);
			stackFree1(&assigned_gate_list);
		  }else
		  {

			  /* part I of the 1st updating solution : updating after one pass*/
			  /*if(realJustifyFg == TRUE && (
			  	 (pInst->MFFC_root == FALSE && fetchImplyValueInStk(pInst)==assignment)||
			  	(pInst->MFFC_root == TRUE && fetchImplyValueInStk(pInst)==assignment) ) )
			  {*/
			  

			  /*another updating solution: updating after one path is assigned */
			  if(realJustifyFg == TRUE)
			  {
				addList=removeList=NULL;
				reArrangeList=NULL;
				/* gain updating and MFFC root gate gain shifting */
				updtGainWholeCircuit(&addList,&removeList,&reArrangeList);

				/*
				copyGain(reArrangeList);
				rebuildBucket();*/

				updtBucketWholeCircuit(addList,removeList,reArrangeList,PWbucketList,PWbucketTail);

				/* ordering inputs using restored(not-shifted) rootgate gains */
				updtInputOrderWholeCircuit(addList,reArrangeList);

				stackFree1(&addList);
				stackFree1(&removeList);
				bi_lin_table_rec_Free1(&reArrangeList,NULL);
				bi_lin_table_rec_Free1(&reArrange_gate_list,NULL);
				stackFree1(&assigned_gate_list);

			  }else{
			  			fprintf(stderr,"This version won't go here!\n");
			  			exit(1);
				   }

			  /* part II for updating solution No. one */
			  /*
			  }else{*/
			  			/* necessary for next backtracing */
			  			/*PWinputSort();*/

			  			/**
			  			updtInputOrderWholeCircuit(addList,reArrangeList);
			  			stackFree1(&addList);
			  			bi_lin_table_rec_Free1(&reArrangeList,NULL);
			  			**/
				   /*}*/



			/***/
		  }
			/****/
		}
		if((try_counter+backtrack_try)%60 == 0 )
		{
			fprintf(stdout,"\n");
			backtrack_try=0;
			try_counter=0;
		}
		fprintf(stdout,"#");

		/*
		reportBucketContent();
		fflush(dbgFp);
		*/
	}

	if(realJustifyFg == FALSE)
	{
		/* this version doesn't need this */
		fprintf(stderr,"this version doesn't need this!\n");


		/* restore all the assigment for this justification */
		restoreBySimulation(PIbacktrackStk);

		/* necessary for correct backtracing */
		/*PWinputSort();*/

		/**
		updtInputOrderWholeCircuit(addList,reArrangeList);
		stackFree1(&addList);
		bi_lin_table_rec_Free1(&reArrangeList,NULL);
		**/

		exit(1);
	}

	/* justification succeed */
	stackFree(&PIbacktrackStk);

	return gain_MFFC_rootgate;
}

/* after backtracing, we get target PI and its expected assignment */
void
Backtrace(pInst,assignment,targetPI,assignmentPI)
PW_INST		*pInst;
LOGIC_VALUE	assignment;
PW_INST		**targetPI;
LOGIC_VALUE	*assignmentPI;
{
	LOGIC_VALUE		inv,inVal1,inVal2;
	int				pinID;
	int				index[2];

	bi_lin_table_rec candidate_rec[2];


	(*targetPI)=pInst;
	(*assignmentPI)=assignment;

	while((*targetPI)->instLibNo!=PI_LIB_NO)
	{
		getBacktraceInput((*targetPI),candidate_rec);


	  /* for gates with controlling value and inversion parity*/
	  if( 	(*targetPI)->instLibNo == NOT_LIB_NO ||
	  		(*targetPI)->instLibNo == AND_LIB_NO ||
	  		(*targetPI)->instLibNo == NAND_LIB_NO ||
	  		(*targetPI)->instLibNo == OR_LIB_NO ||
	  		(*targetPI)->instLibNo == NOR_LIB_NO ||
	  	 	(*targetPI)->instLibNo==BUFF_LIB_NO
	  	  )
	  {
		inv=inversion[(*targetPI)->instLibNo-1];
		/* input value */
		(*assignmentPI)=xor_operation(inv,(*assignmentPI));

		/*if((*targetPI)->InSortedList[logic2int(*assignmentPI)]==NULL)*/
		if(candidate_rec[logic2int(*assignmentPI)]==NULL)
		{
			fprintf(stderr,"No x-path for backtracing of %s with \
							assignment %d\n",
							pInst->name, logic2int(assignment));
			exit(1);
		}

		/*pinID=(int) (((*targetPI)->InSortedList[logic2int(*assignmentPI)])->key);*/
		pinID=(int) (candidate_rec[logic2int(*assignmentPI)]->key);
		(*targetPI)=(*targetPI)->in[pinID];
	  }

	  else 		/* gates with no controlling value */
	  		if((*targetPI)->instLibNo==XOR_LIB_NO ||
	  		  (*targetPI)->instLibNo==XNOR_LIB_NO)
		   {

				/* both inputs are determined already */
				/*if( (*targetPI)->InSortedList[0]==NULL ||
					(*targetPI)->InSortedList[1]==NULL)
					*/
				if( candidate_rec[0] == NULL ||
					candidate_rec[1] == NULL)
				{
					fprintf(stderr,"No x-path for backtracing of %s with \
		                            assignment %d\n",
		                            pInst->name, logic2int(assignment));
                    exit(1);
				}

				/* we have x-path for XOR or XNOR gate */
				/*index[0]=(int) (*targetPI)->InSortedList[0]->key;
				index[1]=(int) (*targetPI)->InSortedList[1]->key;
				*/
				index[0]=(int) candidate_rec[0]->key;
				index[1]=(int) candidate_rec[1]->key;

				inVal1=fetchImplyValueInStk((*targetPI)->in[0]);
				inVal2=fetchImplyValueInStk((*targetPI)->in[1]);

				if(inVal1==VALUE_UNKNOWN && inVal2==VALUE_UNKNOWN)
				{
					if((*targetPI)->in[index[0]]->gain_st[0] >
						(*targetPI)->in[index[1]]->gain_st[1])
					{
						(*targetPI)=(*targetPI)->in[index[0]];
						(*assignmentPI)=VALUE_0;
					}else{
							(*targetPI)=(*targetPI)->in[index[1]];
							(*assignmentPI)=VALUE_1;
					 	 }
				}else{
						/* only one uncertain input, 
						   so another can be determined now
						*/
						if( (*targetPI)->instLibNo==XOR_LIB_NO )
							(*assignmentPI)=
							xor_operation( ((inVal1==VALUE_UNKNOWN)?inVal2:inVal1),
									   	   (*assignmentPI) );
						else
							(*assignmentPI)=
							xnor_operation( ((inVal1==VALUE_UNKNOWN)?inVal2:inVal1),
											(*assignmentPI) );

						/*pinID=(*targetPI)->InSortedList[logic2int(*assignmentPI)]->key;*/
						pinID=(int)candidate_rec[logic2int(*assignmentPI)]->key;
						(*targetPI)=(*targetPI)->in[pinID];
					 }

		   }else{
		   			fprintf(stderr,"Unknown instance in Backtrace() for %s!!\n",
		   						pInst->name);
					exit(1);
				}


	}
}

void
gateListForUpdtGain(assigned_gate_list,reArrange_gate_list)
stk_rec 			assigned_gate_list;
bi_lin_table_rec 	*reArrange_gate_list;
{
	stk_rec	p;
	PW_INST	*inst;
	int		i;

	clearVisitedFlag();

	p=assigned_gate_list;
	while(p!=NULL)
	{
		inst=(PW_INST *)(p->member);
		for(i=0;i<inst->fanin;i++)
		{
			if(inst->in[i]->instLibNo == DF_LIB_NO ||
			   inst->in[i]->instLibNo == DFR_LIB_NO ||
			   inst->in[i]->instLibNo == DFS_LIB_NO ||
			   inst->in[i]->instLibNo == DFRS_LIB_NO)
				continue;

			if(fetchImplyValueInStk(inst->in[i])==VALUE_UNKNOWN &&
				inst->in[i]->visited==FALSE && 
				inst->in[i]->instLibNo!=PO_LIB_NO)
			{
				bi_lin_table_prev_ins(reArrange_gate_list,NULL,(double)NOT_CARE_KEY,(pointer)(inst->in[i]));
				/* set flag so that it won't be added once more */
				inst->in[i]->visited=TRUE;
			}
		}

		/* qz/ outz is not considered currently*/
		for(i=0;i<inst->fanout;i++)
		{
			if(inst->out[i]->instLibNo == DF_LIB_NO ||
                           inst->out[i]->instLibNo == DFR_LIB_NO ||
                           inst->out[i]->instLibNo == DFS_LIB_NO ||
                           inst->out[i]->instLibNo == DFRS_LIB_NO)
                                continue;

			if(fetchImplyValueInStk(inst->out[i])==VALUE_UNKNOWN &&
				inst->out[i]->visited==FALSE &&
				inst->out[i]->instLibNo!=PO_LIB_NO)
			{
				bi_lin_table_prev_ins(reArrange_gate_list,NULL,(double)NOT_CARE_KEY,(pointer)(inst->out[i]));
				inst->out[i]->visited=TRUE;
			}
		}

		p=NEXT_STK_REC(p);
	}
}

void
updtGateListGain(reArrangeList,assignedList)
bi_lin_table_rec	*reArrangeList;
stk_rec				*assignedList;
{
	bi_lin_table_rec	p,q;
	PW_INST				*inst;
	stk_rec				trackGateList=NULL;

	p=*reArrangeList;
	while(p!=NULL)
	{
		inst=(PW_INST	*)(p->member);

		/* it's updating, not initializing */
		trackGateList=NULL;
		updtGateGain(inst,FALSE,&trackGateList);


		q=p->prev;

		p=next_bi_lin_table_rec(p);

	}
	/*gainShiftAmount=shiftMFFCrootGain(MFFCrootList,nonMFFCrootList,FALSE);*/
	/*gainShiftAmount=0;*/
}

/* move all gates from reArrangeList which are also in trackGateList,
   to assignedList.
   get more gates to be updated.
*/
void
manipulateBucketUpdtList(reArrangeList,assignedList,trackGateList)
bi_lin_table_rec    *reArrangeList;
stk_rec             *assignedList;
stk_rec             *trackGateList;
{
	bi_lin_table_rec	p;
	bi_lin_table_rec	reArrangeTail;
	stk_rec				q;
	PW_INST				*inst;
	int					i;

	reArrangeTail=get_bi_lin_table_rec_tail((*reArrangeList));

	q=(*trackGateList);
	while(q!=NULL)
	{
		/* the new uniquely determined gates should be removed from reArrangeList */
		p=bi_lin_table_lookup_by_member((*reArrangeList),q->member);
		if(p!=NULL)
			bi_lin_table_arb_del(reArrangeList,NULL,p);
		mem_free_block((pointer)p);

		/* the new uniquely determined gates should be put into assignedList */
		STK_PUSH(assignedList,(double)NOT_CARE_KEY, (pointer)(q->member));

		/* new gates that should be updated are put into reArrangeList */
	   	inst=(PW_INST *)(q->member);
        for(i=0;i<inst->fanin;i++)
        {
            if(fetchImplyValueInStk(inst->in[i])==VALUE_UNKNOWN &&
                inst->in[i]->instLibNo!=PO_LIB_NO)
            {
            	p=bi_lin_table_lookup_by_member((*reArrangeList),inst->in[i]);
            	if(p==NULL)
                	bi_lin_table_post_ins(reArrangeList,&reArrangeTail,(double)NOT_CARE_KEY,(pointer)(inst->in[i]));
            }
        }

        /* qz/ outz is not considered currently*/
        for(i=0;i<inst->fanout;i++)
        {
            if(fetchImplyValueInStk(inst->out[i])==VALUE_UNKNOWN &&
                inst->out[i]->instLibNo!=PO_LIB_NO)
            {
                p=bi_lin_table_lookup_by_member((*reArrangeList),inst->out[i]);
                if(p==NULL)
                	bi_lin_table_post_ins(reArrangeList,&reArrangeTail,(double)NOT_CARE_KEY,(pointer)(inst->out[i]));
            }
        }


		q=NEXT_STK_REC(q);
	}

	stackFree1(trackGateList);
}



/* report the results */
int
reportMaxCharge(write_file_flag)
BOOLEAN write_file_flag;
{
	int	i;
	int	maxCharge=0;
	int maxChargeWithoutPI=0;
	int	maxChargeRealPI=0;
	int charge_pseudo_PO=0;
	int     charge_second_latch=0;
	LOGIC_VALUE	lv;
	int	v;

	maxChargeWithoutPI=0;
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
			instList[i].instLibNo!=DF_LIB_NO &&
			instList[i].instLibNo!=DFR_LIB_NO &&
			instList[i].instLibNo!=DFS_LIB_NO &&
			instList[i].instLibNo!=DFRS_LIB_NO )
		{
			lv=fetchImplyValueInStk(&(instList[i]));
			v=logic2int(lv);

			if(instList[i].instLibNo!=PI_LIB_NO)
				maxChargeWithoutPI+=v*(instList[i].fanout);

		}
	}

	maxChargeRealPI=0;
	/* for real primary inputs */
	for(i=0;i<PSEUDO_PI_index;i++)
	{
		lv=fetchImplyValueInStk(SourceP[i]);
		v=logic2int(lv);
		maxChargeRealPI+=v*(SourceP[i]->fanout);
	}

	/* for pseudo_PO */
	/* since a pseudo_PO is counted into a primitive gate's
	   fanout, we have to remove this part of charge */
	for(i=PSEUDO_PO_index;i<PrimOutNum;i++)
	{
		lv=fetchImplyValueInStk(SinkP[i]);
		v=logic2int(lv);
		charge_pseudo_PO+=v;
	}

	/* for pseudo_PI: second latch of DFF */
	for(i=PSEUDO_PI_index;i<PrimInNum;i++)
	{
		lv=fetchImplyValueInStk(SourceP[i]);
		v=logic2int(lv);
		if(v==0)
			charge_second_latch+=1;
		else
			charge_second_latch+=SourceP[i]->fanout;
	}

	maxCharge = maxChargeWithoutPI - charge_pseudo_PO +
			maxChargeRealPI + DF_Num +charge_second_latch;

    if(write_file_flag == TRUE)
    {
		fprintf(resultFp,"\n");
		fprintf(resultFp,"The maximum charge is%d.\n" ,maxCharge);
		fprintf(resultFp,"\n\nThe input vector is:\n");
		for(i=0;i<PrimInNum;i++)
		{
			fprintf(resultFp,"PI_name=%s, value=%d\n",
							SourceP[i]->name,
							logic2int(fetchImplyValueInStk(SourceP[i])));
		}
		fprintf(resultFp,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");

		/*fprintf(resultFp,"The max charge without PI's output is %d.\n" ,maxChargeWithoutPI);*/
    }

	return maxCharge;
}

void
accessPIbacktrackStkTop(stack,targetPI,assignmentPI,triedTimes)
stk_rec		stack;
PW_INST		**targetPI;
LOGIC_VALUE	*assignmentPI;
double		*triedTimes;
{
	(*triedTimes)=stack->key;
	(*targetPI)=((PW_IMPLY_STRU *)(stack->member))->inst;
	(*assignmentPI)=((PW_IMPLY_STRU *)(stack->member))->value;
}

/* get next decision of logic value for the PI from preious one */
void
nextDecision(targetPI,assignmentPI,triedTimes)
PW_INST		**targetPI;
LOGIC_VALUE	*assignmentPI;
double		*triedTimes;
{
	if((*assignmentPI)==VALUE_UNKNOWN)
	{
		fprintf(stderr,"It's impossible to have assignmentPI=X in nextDecision().\n");
		fprintf(stderr,"Check the error!.\n");
		exit(1);
	}else{
			if((*triedTimes)==0)
			{
				/* bit reversion */
				(*assignmentPI)=not_operation((*assignmentPI));
				(*triedTimes)++;
			}else if((*triedTimes)==1)
				  {
					/* restore */
					(*assignmentPI)=VALUE_UNKNOWN;
					(*triedTimes)++;
				  }else{
				  		fprintf(stderr,"triedTimes unpected Number!.\n");
				  		exit(1);
				 	   }
		 }
}

bi_lin_table_rec
get_bi_lin_table_rec_tail(head)
bi_lin_table_rec	head;
{
	bi_lin_table_rec	p;

	p=head;
	if(p==NULL)
		return p;
	else{
			while(p->next!=NULL)
			{
				p=p->next;
			}

			return p;
		}
}


void
PWimplyStackGarbageClean(void)
{
	int	k;

	for(k=0;k<PWinstNum_and_pseudo;k++)
	{
		if(instList[k].instLibNo!=DF_LIB_NO &&
			instList[k].instLibNo!=DFR_LIB_NO &&
			instList[k].instLibNo!=DFS_LIB_NO &&
			instList[k].instLibNo!=DFRS_LIB_NO)
			{ gateStackGarbageClean(&(instList[k])); }
	}
}

void
gateStackGarbageClean(pInst)
PW_INST	*pInst;
{
	LOGIC_VALUE	value, *pValue;

    if(pInst->implyStkTop != NULL)
    {
      value = 0;
	  value = fetchImplyValueInStk(pInst);
  
	  stackFree(&(pInst->implyStkTop));
  
	  pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
	  (*pValue)=value;
	  STK_PUSH(&(pInst->implyStkTop),(double)NOT_CARE_KEY,(pointer)pValue);
    }
}


/* restoring the assignment by simulate x */
void
restoreBySimulation(PIbacktrackStk)
stk_rec         PIbacktrackStk;
{
	stk_rec tmpRec;
	PW_INST	*targetPI;
	int		tmp;
	stk_rec assigned_gate_list=NULL;


    tmpRec=PIbacktrackStk;
    while(tmpRec != NULL)
    {
    	targetPI=((PW_IMPLY_STRU *)(tmpRec->member))->inst;
    	assigned_gate_list=NULL;
    	forwardImplyPI3V(targetPI,VALUE_UNKNOWN,&assigned_gate_list,&tmp);
		stackFree1(&assigned_gate_list);

    	tmpRec=next_lin_table_rec(tmpRec);
	}
}

void
labelMFFCid_to_correct(assigned_gate_list,MFFCid,numberofMFFC)
stk_rec assigned_gate_list;
BOOLEAN		MFFCid[];
int			numberofMFFC;
{
	stk_rec tmpRec;
	int		i;
	PW_INST	*pInst;

	if(numberofMFFC > MAX_NUM_MFFC)
	{
		fprintf(stderr,"number of MFFC exceeds MAX_NUM_MFFC\n");
		exit(1);
	}

	for(i=0;i<MAX_NUM_MFFC;i++)
		MFFCid[i]=FALSE;

	tmpRec = assigned_gate_list;
	while(tmpRec != NULL)
	{
		pInst = (PW_INST *) (tmpRec->member);
		if(pInst->MFFC_No > MAX_NUM_MFFC-1)
		{
			fprintf(stderr,"Larger MAX_NUM_MFFC is needed.\n");
			exit(1);
		}
		MFFCid[pInst->MFFC_No]=TRUE;
		tmpRec=next_lin_table_rec(tmpRec);
	}
}

void
PWinitMFFC_Id_Array(MFFC_Id_Array)
BOOLEAN		MFFC_Id_Array[];
{
	int	i;

	for(i=0;i<MAX_NUM_MFFC;i++)
		MFFC_Id_Array[i]=TRUE;
}

void
copyGain(reArrangeList)
bi_lin_table_rec    reArrangeList;
{
	bi_lin_table_rec	tmpRec;
	PW_INST				*pInst;

	tmpRec = reArrangeList;
	while(tmpRec !=NULL)
	{
		pInst= (PW_INST *)(tmpRec->member);
		pInst->gain_st[0] = pInst->new_gain_st[0];
		pInst->gain_st[1] = pInst->new_gain_st[1];

		tmpRec = next_bi_lin_table_rec(tmpRec);
	}
}

