#include <stdio.h>
#include "PWcir.h"
#include "linearFree.h"
#include "linearExt.h"

#include "PWroutines.h"
#include "logicOpt.h"
#include "PWglobal_var.h"

extern void PWmaxCurrent_ilib(array_t *gate_array_sorted);
extern void JustifyAssignment_ilib(PW_INST *pInst, LOGIC_VALUE assignment);
extern void updtInputOrderByCurrentForAllGates_ilib(PW_INST *array, int num);


/* input: the sorted array of gates in non-increasing order */
void
PWmaxCurrent_ilib(array_t *gate_array_sorted)
{
	int		i;
	PW_INST	*inst;

	/* fetch in the non-increasing order */
	for(i=array_n(gate_array_sorted)-1; i>=0; i--)
	{
		inst = array_fetch(PW_INST *, gate_array_sorted, i);

		/* justification of assignment */
		if(inst->powerup_current->preferred_type == RISE_LIST)
			JustifyAssignment_ilib(inst, VALUE_1);
		else JustifyAssignment_ilib(inst, VALUE_0);

	}
}

void
JustifyAssignment_ilib(pInst,assignment)
PW_INST     *pInst;
LOGIC_VALUE assignment;
{
    LOGIC_VALUE     assignmentPI;
    PW_INST         *targetPI;
    BOOLEAN         conflict;
    BOOLEAN         uniqueValueDetected;
    BOOLEAN         backtrackFg,passFg;
    PW_IMPLY_STRU   *PIbacktrackElem=NULL;
    stk_rec         PIbacktrackStk=NULL;
    stk_rec         tmpRec;
    double          triedTimes;
    int				tmpval;

    stk_rec             assigned_gate_list;
    bi_lin_table_rec    reArrange_gate_list;

    uniqueValueDetected=FALSE;
    PIbacktrackElem=NULL;
    PIbacktrackStk=NULL;

    /* assignemnt has already done because of pervious justification */
    if(fetchImplyValueInStk(pInst)!=VALUE_UNKNOWN)
        return;

    while(fetchImplyValueInStk(pInst)!=assignment && uniqueValueDetected==FALSE)
    {

        /* get one taget PI */
        Backtrace_fanout(pInst,assignment,&targetPI,&assignmentPI);

        /* currently selected PI hasn't trigger backtracking */
        backtrackFg=FALSE;
        passFg=FALSE;

        /* push target into backtracking stack */
        PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
        PIbacktrackElem->inst=targetPI;
        PIbacktrackElem->value=assignmentPI;

        STK_PUSH(&PIbacktrackStk,(double)0,(pointer)PIbacktrackElem);

      assigned_gate_list=NULL;
      reArrange_gate_list=NULL;

      /* backtracking may be expected here */
      while(PIbacktrackStk!=NULL && passFg==FALSE)
      {

        /* access backtrack stack top */
        /* get the PI for 3-value forward implication */
        /* it could be for : first time or tracktracked target */
        accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);


        /* assignment of PI and IMPLICATION */
        /* we don't have conflict if using PODEM, 
           But we still need backtracking 
        */
        forwardImplyPI3V(targetPI,assignmentPI,&assigned_gate_list,&tmpval);

        if(assignmentPI==VALUE_UNKNOWN)
        {
            /* remove PI in this time from backtrack stack */
            tmpRec=STK_POP(&PIbacktrackStk);
            mem_free_block((pointer)(tmpRec->member));
            mem_free_block((pointer)(tmpRec));

            if(PIbacktrackStk!=NULL)
            {   /* still have chance to justify */
                accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);

                /* remove PI in previous time from backtrack stack */
                tmpRec=STK_POP(&PIbacktrackStk);
                mem_free_block((pointer)(tmpRec->member));
                mem_free_block((pointer)(tmpRec));

                nextDecision(&targetPI,&assignmentPI,&triedTimes);
                PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
                PIbacktrackElem->inst=targetPI;
                PIbacktrackElem->value=assignmentPI;
                STK_PUSH(&PIbacktrackStk,(double)triedTimes,(pointer)PIbacktrackElem);
            }
            /* else, the justification failed */

        }else

        {

            /* present PIs' assignment combination failed to justif assignment *
/
            /* backtracking needed */
            if( (fetchImplyValueInStk(pInst)!=VALUE_UNKNOWN && fetchImplyValueInStk(pInst)!=assignment)
                || (fetchImplyValueInStk(pInst)==VALUE_UNKNOWN && detectImplicitJustifyFailure(pInst,assignment)==TRUE))
            {
                backtrackFg=TRUE;

                /* get one more wrong decision */
                accessPIbacktrackStkTop(PIbacktrackStk,&targetPI,&assignmentPI,&triedTimes);

                /* remove wrong decision from backtrack stack */
                tmpRec=STK_POP(&PIbacktrackStk);
                mem_free_block((pointer)(tmpRec->member));
                mem_free_block((pointer)(tmpRec));

                /* get next decision according to the previous wrong one */
                /* push into backtrack stack for net trying */
                nextDecision(&targetPI,&assignmentPI,&triedTimes);
                PIbacktrackElem=(PW_IMPLY_STRU *)mem_get_block((SIZE_T) sizeof(PW_IMPLY_STRU));
                PIbacktrackElem->inst=targetPI;
                PIbacktrackElem->value=assignmentPI;
                STK_PUSH(&PIbacktrackStk,(double)triedTimes,(pointer)PIbacktrackElem);
            }else
            {
                /* the initially selected PI is passed */
                /* or the backtracking is successful */
                /* but justification not completed */
                passFg=TRUE;
            }
        }
      }

        if(PIbacktrackStk==NULL)
        {
            /* the justification cannot be performed successfully
               under current situation
            */
            /* so the gate output has already been implicitely determined */
            /* but we don't detect until here because of incremental gain updating */
            implyGate(pInst,not_operation(assignment),&assigned_gate_list,&conflict);
            uniqueValueDetected=TRUE;
            if(conflict==TRUE)
            {
                fprintf(stderr,"Impossible value for %s, ID %d\n in JustifAssignment()",pInst->name,pInst->instID);
                exit(1);
            }

            updtInputOrderByCurrentForAllGates_ilib(instList,PWinstNum);


        }else
        {
            /* the selected PI should be passed here */
            if(passFg==FALSE)
            {
                fprintf(stderr,"Impossible situation in Justification.\n");
                fprintf(stderr,"Check backtracking process.\n");
                exit(1);
            }

            if(backtrackFg==FALSE)
            {

                updtInputOrderByCurrentForAllGates_ilib(instList,PWinstNum);

            }else
            {

                updtInputOrderByCurrentForAllGates_ilib(instList,PWinstNum);

            }
        }

        /*reportBucketContent();*/
        fflush(dbgFp);
    }

    /* justification succeed */
    stackFree(&PIbacktrackStk);
}

void
updtInputOrderByCurrentForAllGates_ilib(array, num)
PW_INST	*array;
int		num;
{
	int	i;
	for(i=0; i<num; i++)
	{
		PWinputSortOneGate_ilib(&(array[i]));
	}
}
