#include "PWcir.h"
#include <stdio.h>
#include "PWmc.h"
#include "DFtype.h"
#include "PWmcData.h"
#include "math.h"
#include "linear.h"
#include "linearExt.h"

#include "PWroutines.h"

extern double	norm_dist[][2];
extern BOOLEAN  isMCconvg(stk_rec MClistHead, double mean, 
						  int num_of_results, int num_of_sim, 
						  double  norm_x, double err);
extern double 	lookupNormDist(double  confd, double  normTable[][2]);
extern double 	getStdDev(stk_rec listHead, double mean, int num);

extern void PWmcSim(double  confd, double err);




void
PWmcSim(confd,err)
double	confd,err;
{
	int				N;
	int 			rand_max=0, currCharge;
	int 			time_counter = 0;
	stk_rec   		mc_list;
	int				num_mc_list;
	double			mean = 0,stdDev;
	BOOLEAN			converge_flag;

	double norm_x;

	norm_x = lookupNormDist(confd,norm_dist);
	mc_list=NULL;
	N=0;
	num_mc_list=0;
	while(N<30)
	{
		N++;
		rand_input_vec_gen(RAND_INPUT_VEC_REAL_PI);

		simple_sim_operation();

		currCharge = reportMaxCharge(FALSE);
		rand_max = DF_MAX(rand_max,currCharge);
		mean = ( mean* ((double)num_mc_list) + currCharge )/(num_mc_list+1);
		STK_PUSH(&mc_list, (double)currCharge, (pointer)NULL);

		num_mc_list++;

	}

	converge_flag=FALSE;
	while(converge_flag == FALSE)
	{

		if( time_counter >= 5 )
        {
		    PWimplyStackGarbageClean();
	        time_counter = 0;
	    }
	    else time_counter++;

		N++;
		fprintf(stderr,"Iteration:%d\n",N);

		rand_input_vec_gen(RAND_INPUT_VEC_REAL_PI);
		simple_sim_operation();

		currCharge = reportMaxCharge(FALSE);
		rand_max = DF_MAX(rand_max,currCharge);
		mean = ( mean* ((double)num_mc_list) + currCharge )/(num_mc_list+1);
		STK_PUSH(&mc_list, (double)currCharge, (pointer)NULL);

		num_mc_list++;
		converge_flag = isMCconvg(mc_list,mean,num_mc_list,N,norm_x,err);
	}
	stdDev=getStdDev(mc_list, mean, num_mc_list);
	fprintf(stdout,"The mean:%f\n",mean);
	fprintf(stdout,"The standard deviation:%f\n",stdDev);
	fprintf(stdout,"The simulated max is %d\n",rand_max);

}

BOOLEAN
isMCconvg(MClistHead, mean, num_of_results, num_of_sim, norm_x, err)
stk_rec	MClistHead;
double	mean;
int		num_of_results,num_of_sim;
double	norm_x,err;
{
	double	stdDev;
	double	param;
	double  epsilon;

	epsilon=err/(1+err);
	stdDev=getStdDev(MClistHead, mean, num_of_results);
	param=norm_x*stdDev;
	param=param/(mean*epsilon);
	param=param*param;
	if(((double)num_of_sim) >= param)
		return TRUE;
	else
		return FALSE;
}




double
lookupNormDist(confd,normTable)
double	confd;
double	normTable[][2];
{
	double  leftArea,alpha;
    int     i,indexMin=0;
    double  dst;
	double  dstMin=2;

	alpha = (1.0-confd)/2.0; /* right area */
	leftArea = 1 - alpha;	 /* left area */

	for(i=0;i<NORM_DSTR_GRID_NUM;i++)
	{
		dst = DF_ABS(normTable[i][0]-leftArea);
		if(dst < dstMin)
		{
			dstMin = dst;
			indexMin = i;
		}
	}
	return normTable[indexMin][1];
}

double
getStdDev(listHead,mean,num)
stk_rec	listHead;
double	mean;
int		num;
{
	stk_rec	tmpRec;
	double	sum,t;

	sum=0;

	for(tmpRec=listHead; tmpRec!=NULL; tmpRec=next_lin_table_rec(tmpRec))
	{
		t=(double)(tmpRec->key) - mean;
		t=t*t;
		sum+=t;
	}

	sum = sum/((double) num);
	sum = sqrt(sum);
	return sum;
}
