#include <stdio.h>
#include <string.h>

#ifndef SIS
#define SIS
#endif

#include "sis.h"
#include "DFtype.h"
#include "PWmemcpy_sis.h"
#include "PWcir.h"
#include "PWglobal_var.h"
#include "PWroutines.h"

extern int mape_get_node_level(node_t *) ;
extern void report_memcpy_statis(network_t *network, FILE *resultFp);


static int copy_index;

/* memory copy: transfer sis's data structure to imaxup's data structure */
int
PWmemcpy_sis(network_t *network)
{
    int     pi_index,po_index,instance_index;

    /*
    int             hashTableSize=300;
    lin_table_rec   *NodeHashTable;
    */


	if(network == NULL)
	{
		fprintf(stderr,"Before transferring data from SIS,\
						please load a file into SIS first.\n");
		return 1;
	}

    copy_index = 0;
    unknownCellList=NULL;
    DFlipList=NULL;
    DF_Num=0;
    PWlineNum=PWinstNum=0;
    PSEUDO_PI_index=PSEUDO_PO_index=0;

	
    PrimInNum=0;
    PrimOutNum=0;
    vdd_index=gnd_index=-1;
    pi_index=po_index=instance_index=0;

    pseudoCount=0;

    /* initialize the hash table for net name */
    /*hashTableInit(&NodeHashTable, hashTableSize);*/

	DF_Num = lsLength(network->latch);

	/* # of PPI : no PPO if transferred from sis*/
	pseudoCount = DF_Num;	

	/* network->nodes: PI + PPI + PO(PPO) + gates ; 
	   no lathes are counted in */
	PWinstNum_and_pseudo = lsLength(network->nodes) + DF_Num;

	PWlineNum = PWinstNum = PWinstNum_and_pseudo - pseudoCount;
	PrimInNum = lsLength(network->pi); /* PI + PPI */
	PrimOutNum = lsLength(network->po);

	PSEUDO_PI_index = PrimInNum - pseudoCount;
	PSEUDO_PO_index = PrimOutNum;

	/* alloc & init */
	instList=(PW_INST*)mem_get_block((PWinstNum+2*DF_Num)*sizeof(PW_INST));
	initPWinst(instList);

	if(InstMemCopy(network)==1) return 1;

	if(FormPIPOlist(network)==1) return 1;

	if(CktConnectCopy(network)==1) return 1;

    report_memcpy_statis(network,resultFp);

	return 0; /* return success: sis's convention */
}

int
InstMemCopy(network_t *network)
{
	if(GateMemCopy(network)==1) return 1;

	if(LatchMemCopy(network)==1) return 1;
	if(PO_MemCopy(network)==1) return 1;
	if(PI_MemCopy(network)==1) return 1;
	if(PPI_MemCopy(network)==1) return 1;

	return 0;
}

/* copy all the gates in the combinational block*/
int
GateMemCopy(network_t *network) 
{
	lsGen 			gen;
	node_t 			*node;
	node_function_t type;

	foreach_node(network, gen, node)
	{
		type=node_function(node);
		switch(type)
		{
			case NODE_BUF:
			case NODE_INV:
			case NODE_AND:
			case NODE_OR:
			case NODE_COMPLEX: /* XOR, XNOR need to be identified */
				if(memCopyOneGate(node) == 1) return 1;
				break;
			case NODE_UNDEFINED:
				return 1; /* unmapped gates not supported */
				break;
			case NODE_PI:
			case NODE_PO:
			case NODE_0:
			case NODE_1:
				break;
			default:
				return 1; /* wrong function */
				break;
		}
	}
	return 0;
}

int
memCopyOneGate(node_t *node)
{
	node_function_t type;
	int		imaxup_lib_no;

	if(copy_index >= PWinstNum+2*DF_Num)
	{
		fprintf(stderr,"copy_index exceeds range!\n\n");
		return 1;
	}
	strcpy(instList[copy_index].name, node->name);
	/*instList[copy_index].instID = copy_index;*/
	instList[copy_index].fanin =  node->nin;
	instList[copy_index].fanout = lsLength(node->fanout);

	/* NOTE: min level is not necessary to get*/
	instList[copy_index].maxLev = mape_get_node_level(node);
	instList[copy_index].minLev = 0;


	type = node_function(node);
	switch(type)
	{
		case NODE_BUF:
			instList[copy_index].instLibNo = BUFF_LIB_NO;
			break;
        case NODE_INV:
        	instList[copy_index].instLibNo = NOT_LIB_NO;
        	break;
        case NODE_AND:
        	imaxup_lib_no = sis_node_func_AND_type(node);
        	if(imaxup_lib_no <= 0)
        	{
        		fprintf(stderr,"Cannot further identify \
        						a AND_type sis node.\n\n");
        		return 1;
        	}else 
        		instList[copy_index].instLibNo = imaxup_lib_no;
        	break;
        case NODE_OR:
        	imaxup_lib_no = sis_node_func_OR_type(node);
        	if(imaxup_lib_no <= 0)
        	{
        		fprintf(stderr,"Cannot further identify \
        						a OR_type sis node.\n\n");
				return 1;
			}else 
				instList[copy_index].instLibNo = imaxup_lib_no;
        	break;
        case NODE_COMPLEX:
        	/* identify XOR2 , XNOR2 */
        	if(node->nin != 2)
        	{
        		fprintf(stderr,"Complex gates other than xor2, xnor2\
        						cannot be handled by imaxup.\n");
				fprintf(stderr,"Try remapping the circuit.\n\n");
				return 1;
			}else{
				imaxup_lib_no = sis_node_func_complex_type(node);
				if(imaxup_lib_no <= 0)
				{
					fprintf(stderr,"Complex gates other than xor2, xnor2\
									cannot be handled by imaxup.\n");
					fprintf(stderr,"Try remapping the circuit.\n\n");
					return 1;
				}else 
					instList[copy_index].instLibNo = imaxup_lib_no;
				 }
        	break;
		default:
			fprintf(stderr,"Wrong Function Type: memCopyOneGate().\n");
			return 1;
			break;
	}

	xlink_imaxup_sis_node(&(instList[copy_index]),node);

	copy_index++;
	return 0;
}

/* cross link an imaxup instance and a sis node */
void
xlink_imaxup_sis_node(PW_INST *inst, node_t *node)
{
	node->imaxup = (char *) (inst);
	inst->sis_node = (pointer) (node);
}

void
xlink_imaxup_sis_latch(PW_INST *inst, latch_t *latch)
{
	latch->imaxup = (char *) (inst);
	inst->sis_node = (pointer) (latch);
}

/* input: must be a NODE_AND node in sis. */
/*        a NODE_AND node in sis denotes a node
          whose ON set only has one implicant and
          whose input number is larger than one.
 */
/* output: AND gate or NOR gate in imaxup */
/* we assume all the inputs are equivalent 
 */
int
sis_node_func_AND_type(node_t *node)
{
	register pset p;
	p = GETSET(node->F,0); /* the only implicant */
	switch(GETINPUT(p,0))  /* the first input in the implicant */
	{
		case ZERO:
			return NOR_LIB_NO;
			break;
		case ONE:
			return AND_LIB_NO;
			break;
		default:
			break;
	}
	return -1;
}

/* input: must be a NODE_OR node in sis */
/*        a NODE_OR node in sis denotes a node
          whose OFF set only has one implicate and
          whose input number is larger than one.
 */
/* output: OR gate or NAND gate in imaxup */
/* we assume all the inputs are equivalent */
int
sis_node_func_OR_type(node_t *node)
{
	int	i;
	register pset p;

	p = GETSET(node->F,0);
	for(i=0;i< node->nin; i++)
	{
		switch(GETINPUT(p,i))
		{
			case ZERO:
				return NAND_LIB_NO;
				break;
			case ONE:
				return OR_LIB_NO;
				break;
			default:
				break;
		}
	}
	return -1;
}

/* input: must be a NODE_COMPLEX node with two inputs in sis */
/* output: XOR2 gate or XNOR2 gate */
int
sis_node_func_complex_type(node_t *node)
{
	register pset p0,p1;

	p0 = GETSET(node->F,0);
	p1 = GETSET(node->F,1);

	if( ((GETINPUT(p0,0)==ZERO && GETINPUT(p0,1)==ONE) &&
		(GETINPUT(p1,0)==ONE && GETINPUT(p1,1)==ZERO))  ||
		((GETINPUT(p0,0)==ONE && GETINPUT(p0,1)==ZERO) &&
		(GETINPUT(p1,0)==ZERO && GETINPUT(p1,1)==ONE)) )
		return XOR_LIB_NO;
	else
	if( ((GETINPUT(p0,0)==ZERO && GETINPUT(p0,1)==ZERO) &&
		(GETINPUT(p1,0)==ONE && GETINPUT(p1,1)==ONE))  ||
		((GETINPUT(p0,0)==ONE && GETINPUT(p0,1)==ONE) &&
		(GETINPUT(p1,0)==ZERO && GETINPUT(p1,1)==ZERO)) )
		return XNOR_LIB_NO;
	else
		return -1;
}

/* we assume all the latches are controlled by global clock */
int
LatchMemCopy(network_t *network)
{
	lsGen 	gen;
	latch_t *l;
	char	latch_name[300];

	foreach_latch(network, gen,l)
	{
		if(copy_index >= PWinstNum+2*DF_Num)
		{
			fprintf(stderr,"copy_index exceeds range!\n\n");
			return 1;
		}

		strcpy(latch_name,"DF__");
		strcat(latch_name,l->latch_output->name);
		strcpy(instList[copy_index].name, latch_name);

		instList[copy_index].fanin = 1;
		instList[copy_index].fanout = lsLength(l->latch_output->fanout);

		instList[copy_index].instLibNo = DF_LIB_NO;

		xlink_imaxup_sis_latch(&(instList[copy_index]),l);
		copy_index++;
	}

	return 0;
}

int
PO_MemCopy(network_t *network)
{
	node_t *node;
	lsGen gen;

	foreach_primary_output(network, gen, node)
	{
		if(copy_index >= PWinstNum+2*DF_Num)
        {
            fprintf(stderr,"copy_index exceeds range!\n\n");
            return 1;
        }

		strcpy(instList[copy_index].name, node->name);
		instList[copy_index].instLibNo = PO_LIB_NO;
		instList[copy_index].fanin = node->nin;
		instList[copy_index].fanout = lsLength(node->fanout);
		instList[copy_index].maxLev = mape_get_node_level(node);
		instList[copy_index].minLev = 0;

		xlink_imaxup_sis_node(&(instList[copy_index]),node);
		copy_index++;
	}
	return 0;
}

/* copy real primary inputs */
int
PI_MemCopy(network_t *network)
{
	node_t *node;
	lsGen gen;

	foreach_primary_input(network, gen, node)
	{
		if(isPPI_node(node, network) == FALSE)
		{
        		if(copy_index >= PWinstNum+2*DF_Num)
        		{
            			fprintf(stderr,"copy_index exceeds range!\n\n");
            			return 1;
        		}

			strcpy(instList[copy_index].name, node->name);
			instList[copy_index].instLibNo = PI_LIB_NO;
			instList[copy_index].fanin = node->nin;
			instList[copy_index].fanout = lsLength(node->fanout);
			instList[copy_index].maxLev = 0;
			instList[copy_index].minLev = 0;

			xlink_imaxup_sis_node(&(instList[copy_index]),node);
			copy_index++;
		}
	}

	return 0;
}

int
PPI_MemCopy(network_t *network)
{
	latch_t *l;
	node_t	*node;
	lsGen gen;

	foreach_latch(network, gen,l)
	{
        if(copy_index >= PWinstNum+2*DF_Num)
        {
            fprintf(stderr,"copy_index exceeds range!\n\n");
            return 1;
        }

		node = l->latch_output;
		strcpy(instList[copy_index].name, node->name);
		instList[copy_index].instLibNo = PI_LIB_NO;
		instList[copy_index].maxLev = 0;
		instList[copy_index].minLev = 0;
		instList[copy_index].fanin = 1; /* from output of DF */
        instList[copy_index].fanout = lsLength(node->fanout);


		xlink_imaxup_sis_node(&(instList[copy_index]),node);
		copy_index++;
	}
	return 0;
}

/* input must be a PI or PPI node */
BOOLEAN
isPPI_node(node_t *node, network_t *network)
{
	node_function_t type;
	lsGen 	gen;
	latch_t *l;

	type = node_function(node);
	if(type != NODE_PI)
	{
		fprintf(stderr,"input should be a PI or PPI node.\n\n");
		exit(1);
	}

	foreach_latch(network, gen, l)
	{
		if(node == l->latch_output)
		{
			return TRUE;
		}
	}
	return FALSE;
}

int
FormPIPOlist(network_t *network)
{
	node_t *node;
	latch_t *l;
	lsGen 	gen;
	PW_INST	*inst;
	int	i,j;

	/* PO list */
	i=0;
	foreach_primary_output(network, gen, node)
	{
		if(i>=MAX_PO)
		{
			fprintf(stderr,"Not enough space for PO & PSEUDO_PO\n");
			fprintf(stderr,"Larger matrix needed.\n");
			return 1;
		}

		inst = (PW_INST *)(node->imaxup);
		SinkP[i]=inst;
		i++;
	}

	/* PI list */
	j=0;
	foreach_primary_input(network, gen, node) 
	{
		if(j>= MAX_PI)
		{
			fprintf(stderr,"Not enough space for PI & PSEUDO_PI\n");
			fprintf(stderr,"Larger matrix needed.\n");
			return 1;
		}

		if(isPPI_node(node, network) == FALSE)
		{
			inst = (PW_INST *)(node->imaxup);
			SourceP[j]=inst;
			j++;
			if(node_function(node) == NODE_0)
				gnd_index = inst->instID;
			if(node_function(node) == NODE_1)
				vdd_index = inst->instID;
		}
	}
	foreach_latch(network, gen, l) /* PPI */
	{
		if(j>= MAX_PI)
                {
                        fprintf(stderr,"Not enough space for PI & PSEUDO_PI\n");
                        fprintf(stderr,"Larger matrix needed.\n");
                        return 1;
                }
		inst = (PW_INST *) (l->latch_output->imaxup);
		SourceP[j]=inst;
		j++;
	}

	return 0;
}

void
report_memcpy_statis(network_t *network, FILE *resultFp)
{

  fprintf(resultFp,"Network \"%s\" transferred in the memory.\n", network->net_name);
  fprintf(resultFp,"Primary Inputs & Pseudo PI (controllable) : %ld\n", PrimInNum);
  fprintf(resultFp,"        Pseudo Primary Inputs : %d\n", pseudoCount);
  fprintf(resultFp,"                   Flip-flops : %ld\n", DF_Num);
  fprintf(resultFp,"              Primary outputs : %ld\n", PrimOutNum);

  if(vdd_index >= 0) fprintf(resultFp,"            Uncontrollable PI : one VDD\n");
  if(gnd_index >= 0) fprintf(resultFp,"            Uncontrollable PI : one GND\n");
  fprintf(resultFp,"      Gates (inc. flip-flops) : %ld\n", PWinstNum_and_pseudo - PrimOutNum - PrimInNum);
  fprintf(resultFp,"\n");
}
