#include "PWcir.h"
#include "Rtime.h"

#include "PWroutines.h"
#include "PWglobal_var.h"
#include <stdio.h>
#include "DFrandom.h"

#define SEQUENCE_LEN	50


void
PWnormSwitchingSim()
{
	int time_counter = 0;
	int i,j;
	int rand_max_sw=0, curr_sw;

	fprintf(stdout,"Running Logic Simulation for Normal Switching Current (2 vectors)\n");
	SetTime(TIME_REPORT);

	randomGenInit1();

	/* # of sequences */
	for(i=0;i<500;i++)
	{
		if(time_counter >= 5)
		{
			PWimplyStackGarbageClean();
			time_counter = 0;
		}else time_counter++;

		rand_input_vec_gen(RAND_INPUT_VEC_ALL_PI);
		simple_sim_operation();
		
		for(j=0;j<SEQUENCE_LEN;j++)
		{
			rand_input_vec_gen(RAND_INPUT_VEC_REAL_PI);
			state_vec_propagate();
			simple_sim_operation();

			curr_sw = reportMaxSW();
			rand_max_sw = DF_MAX(rand_max_sw, curr_sw);
			if(time_counter >=5)
			{
				PWimplyStackGarbageClean();
				time_counter = 0;
			}else time_counter++;
		}
		
	}
	fprintf(stdout,"Logic simulation finished.\n");
	fprintf(stdout,"the max switching activity for random vectors is %d.\n",rand_max_sw);
	fprintf(resultFp,"the max switching activity for random vectors is %d.\n", rand_max_sw);
	
	ReportTime(TIME_REPORT);
}

void
state_vec_propagate()
{
	PW_INST		*pInst;
	LOGIC_VALUE *pValue;

	for(pInst=DFlipList;pInst!=NULL;pInst=pInst->next)
	{
		pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
		(*pValue)=fetchImplyValueInStk(pInst->in[0]);
		/* for pseudo-PIs */
		STK_PUSH(&(pInst->out[0]->implyStkTop),(double)NOT_CARE_KEY,
				 (pointer)pValue);
	}
}

int
reportMaxSW()
{
	int i;
	int maxSW=0;
	int charge_pseudo_PO = 0;
	BOOLEAN	swFg;

	/* primitive gates and real primary inputs, only pseudo_POs are
		mistakenly counted into primitive gates' fanout, 
		which is amened in the just following code*/
	/* also, pseudo PIs counted for the second latch of DFF and
		exter 2 for the first latch*/
	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if(instList[i].instLibNo!=PO_LIB_NO &&
           instList[i].instLibNo!=DF_LIB_NO &&
		   instList[i].instLibNo!=DFR_LIB_NO &&
		   instList[i].instLibNo!=DFS_LIB_NO &&
		   instList[i].instLibNo!=DFRS_LIB_NO )
	 	{
	 		swFg = detectSW(&(instList[i]));
	 		if(swFg == TRUE)
	 		{
	 			maxSW = maxSW + instList[i].fanout;

				/* for the first latch of DFF */
				if(instList[i].instLibNo == PI_LIB_NO &&
					instList[i].in[0] != NULL)
					maxSW = maxSW + 1;	/* assume 1st latch is stable at the starting of the negative edge*/
											/* 1 for part of the second latch*/
			}
	 	}
	}

	for(i=PSEUDO_PO_index;i<PrimOutNum;i++)
    {
    	swFg = detectSW(SinkP[i]);
    	if(swFg == TRUE)
            charge_pseudo_PO+=1;
	}

	maxSW = maxSW - charge_pseudo_PO;
	return maxSW;
}



BOOLEAN
detectSW(pInst)
PW_INST	*pInst;
{
	LOGIC_VALUE lv1,lv2;

	lv1=fetchImplyValueInStk(pInst);
	lv2=(*((LOGIC_VALUE *)(pInst->implyStkTop->next->member)));
	if(lv1== VALUE_UNKNOWN || lv2 == VALUE_UNKNOWN)
    {
    	fprintf(stderr,"Error detected for sim results in reportMaxSW().\n");
    	exit(1);
    }
    if(lv1 != lv2)
    	return TRUE;	/* a switching detected */
	else return FALSE;
}
