#include "PWcir.h"
#include "linearExt.h"
#include "linearFree.h"

#include "PWglobal_var.h"
#include "PWroutines.h"


extern void 	getRearrangeMFFCrootList(lin_table_rec rootList, bi_lin_table_rec *newList);
extern int      JustifyAssignment();
extern int		shiftMFFCrootGain_cluster();
extern void		PWbucketReArrange();
extern int		shiftMFFCrootGain();
extern void 	correctMFFCrootGateGain(lin_table_rec MFFCrootList, BOOLEAN MFFC_Id[]);


/* update the gain for all the gates */
/* gates will be put into different gate lists after get the new gain */
/*For those gates in reArrangeList, their new gain, which is
  stored in new_gain temporarily, may equal to the old gain.
  This situation will be detected in updtBucketWholeCircuit().
*/
/* gain for MFFC root gates will be shifted */
void
updtGainWholeCircuit(addList,removeList,reArrangeList)
stk_rec				*addList;	/* gates that should be added to buckets */
stk_rec				*removeList; /* gates that should be removed from buckets */
bi_lin_table_rec 	*reArrangeList; /* gates that should be re-arranged in the buckets */
{
	int		i;
	stk_rec	trackStk;

	for(i=0;i<PWinstNum_and_pseudo;i++)
	{
		if( instList[i].instLibNo!=PO_LIB_NO &&
			instList[i].instLibNo!=DF_LIB_NO &&
			instList[i].instLibNo!=DFR_LIB_NO &&
			instList[i].instLibNo!=DFS_LIB_NO &&
			instList[i].instLibNo!=DFRS_LIB_NO )
		{
			/* detect error */
			if( (instList[i].pGainUpdt_st[0]==NULL &&
			   instList[i].pGainUpdt_st[1]!=NULL) ||
			    (instList[i].pGainUpdt_st[0]!=NULL &&
			     instList[i].pGainUpdt_st[1]==NULL) )
			{
				fprintf(stderr,"Impossible situation in updtGainWholeCircuit().\n");
				exit(1);
			}


			if(fetchImplyValueInStk(&(instList[i]))==VALUE_UNKNOWN)
			{
				if( instList[i].pGainUpdt_st[0]==NULL )
				{
					trackStk=NULL;

					/* assign gain to gain_st */
					updtGateGain(&(instList[i]),TRUE,&trackStk);
					STK_PUSH(addList,(double)NOT_CARE_KEY,(pointer)(&(instList[i])));
				}else{
						trackStk=NULL;
						/* re-arrange it */
						updtGateGain(&(instList[i]),FALSE,&trackStk);
						bi_lin_table_prev_ins(reArrangeList,NULL,(double)NOT_CARE_KEY,(pointer)(&(instList[i])));
					 }

			}else
 				/* gate has a determined value */
			{
				if(instList[i].pGainUpdt_st[0]==NULL)
				{
					/* nothing should be done */
				}else{
						/* be removed from buckets */
						STK_PUSH(removeList,(double)NOT_CARE_KEY,(pointer)(&(instList[i])));
					 }
			}
		}
	}

	/* move root gates of MFFC ahead of non-MFFC root gates by shifting the gains */
	gainShiftAmount=shiftMFFCrootGain(MFFCrootList,nonMFFCrootList,FALSE);
}

/* based on the init-definition gain, 
we re-compute the gain for MFFC root gates 
by pre-justify possible assignment*/
/* Three operations done here:
	1) re-compute the gain,
	2) shift the gain of MFFC root gates,
	3) adjust the bucket for MFFC root gates.
*/
void
correctMFFCrootGateGain(MFFCrootList,MFFC_Id)
lin_table_rec MFFCrootList;
BOOLEAN			MFFC_Id[];
{
	lin_table_rec 		tmpRec=NULL;
	PW_INST		  		*pInst;
	bi_lin_table_rec    reArrangeList=NULL;
	int					i;


	/* 1)re-compute the gain */
	tmpRec = MFFCrootList;
	i=0;
	while(tmpRec != NULL)
	{
		pInst=((PW_INST*)(tmpRec->member));

		if(fetchImplyValueInStk(pInst)== VALUE_UNKNOWN &&
			MFFC_Id[pInst->MFFC_No] == TRUE)
		{
			i++;
			pInst->new_gain_st[1]=JustifyAssignment(pInst,VALUE_1,FALSE);
			pInst->new_gain_st[0]=JustifyAssignment(pInst,VALUE_0,FALSE);
		}

		tmpRec = next_lin_table_rec(tmpRec);
	}

	if(i==0) return;

	/* 2)shift the gain */
	/* MFFC root gate gain is always an updating, therefore  the initFlag
	is always FALSE */
	gainShiftAmount=shiftMFFCrootGain_cluster(MFFCrootList,nonMFFCrootList);

	/* 3)adjust the bucket */
	reArrangeList=NULL;
	getRearrangeMFFCrootList(MFFCrootList,&reArrangeList);

	PWbucketReArrange(reArrangeList);
	bi_lin_table_rec_Free1(&reArrangeList,NULL);
}

void
getRearrangeMFFCrootList(rootList,newList)
lin_table_rec	 rootList;
bi_lin_table_rec *newList;
{
	lin_table_rec	 	rec1;
	PW_INST			 	*pInst;

	rec1=rootList;
	while(rec1!=NULL)
	{
		pInst=(PW_INST *)(rec1->member);
		if(fetchImplyValueInStk(pInst)==VALUE_UNKNOWN)
		{
			bi_lin_table_prev_ins(newList,NULL,(double)NOT_CARE_KEY,
									(pointer)(pInst));
		}

		rec1=next_lin_table_rec(rec1);
	}
}
