#define SIS
#include "sis.h"
#include <stdio.h>
#include "DFfile.h"
#include "leakage_ilib.h"
#include "leakage_ga.h"
#include "PWroutines.h"

#define LEAKAGE_ILIB_READ 1
#define LEAKAGE_GA_ESTIMATION 2
#define LEAKAGE_RANDOM_SIM 3

extern int leakage_command(network_t **network, int argc, char **argv);
extern void LKG_usage (char **usage, char  *cmd);

extern FILE* dbgFp;

static int lkg_rilib_flag = 0;
static int lkg_est_min_max_flag = 0;

static char *lkg_usage[] =
{
    "usage: %s [rilib] [ga -g [maximum generation] [-u -l]] [rand -m [matched generations for comparison with GA]]",
    "   rilib: read in the leakage current library",
    "   ga: Genetic algorithm for leakage current estimation or",
    "   rand: Random simulation algorithm for leakage current estimation",
    "   -g: maximum number of generations",
    "   -u: maximum leakage current estimation",
    "   -l: minimum leakage current estimation",
    "   Default otpions: -g100, -m100",
    NULL
};


int
leakage_command(network_t **network, int argc, char **argv)
{
    char str_buff[128];
    char fname[200],real_fname[200];
    static char old_fname[200]; //static var for file name
    char fout_name[300];
    char spice_fname[300];
    char dbg_fname[300];
    int  prog_func, pos;
    int  flag;
    FILE *fin=NULL, *sp_fout=NULL, *ilibFp=NULL;

    char ** tmp_argv, c;
    int tmp_argc, max_gen=-1, matched_gen=-1;

	if(argc <= 1)
    {
        fprintf(stderr,"No options was given.\n\n");
        LKG_usage(lkg_usage, argv[0]);
        fprintf(stderr,"Or use \"help\" to get available options.\n");
        return 1;
    }
	strcpy(str_buff,argv[1]);
	if(strcmp(str_buff, "rilib")==0)
		prog_func = LEAKAGE_ILIB_READ;
    else if(strcmp(str_buff, "ga")==0)
        prog_func = LEAKAGE_GA_ESTIMATION;
    else if(strcmp(str_buff, "rand")==0)
        prog_func = LEAKAGE_RANDOM_SIM;
    else{
            fprintf(stderr,"Unknown options.\n\n");
            LKG_usage(lkg_usage, argv[0]);
            return 1;
        }

    /* hack: sub-options */
    tmp_argc = argc - 1;
    tmp_argv = &(argv[1]);

    switch(prog_func)
    {
        case LEAKAGE_ILIB_READ: /* read power-up current library */
            if(lkg_rilib_flag != 0)
            {
                fprintf(stderr, "LEAKAGE ILIB file has already been read.\n");
                return 1;
            }
            FOPEN(ilibFp, "leakage.ilib", "r");
            if(dbgFp == NULL)
            {
                FOPEN(dbgFp, "ilib_debug.info", "w");
                if(read_leakage_ilib(ilibFp, dbgFp)==0) lkg_rilib_flag = 1;
                fclose(dbgFp);
            }else{ 
		           if(read_leakage_ilib(ilibFp, dbgFp)==0) lkg_rilib_flag = 1;
		         }
            break;
        case LEAKAGE_GA_ESTIMATION:
            if(get_imaxup_read_flag() == 0 || lkg_rilib_flag == 0)
            {  fprintf(stderr," Either netlist or leakage current library is not read in.\n");
               return 1;
            }else{
                   util_getopt_reset();
                   while((c=util_getopt(tmp_argc,tmp_argv,"g:ul")) != EOF)
                   {
                     switch(c)
                     {
                       case 'g': max_gen = atoi (util_optarg); break;
                       case 'u': 
                       			if(lkg_est_min_max_flag != 0)
                       			{
                       				fprintf(stderr,"The switches 'u' and 'l' cannot be used together.\n");
                       				return 1;
								}
								lkg_est_min_max_flag = 1; // max current (upper bound)
								break;
                       case 'l':
                       			if(lkg_est_min_max_flag != 0)
                       			{
                       				fprintf(stderr,"The switches 'u' and 'l' cannot be used together.\n");
                       				return 1;
								}
								lkg_est_min_max_flag = -1; //min current (lower bound)
								break;
                       case '?':
                       default: LKG_usage(lkg_usage, argv[0]); return 1;
                     }
                   }
                   if(lkg_est_min_max_flag == 0)
                   {
                   	fprintf(stderr,"Either the switch 'u' or 'l' must be specified\n");
                   	return 1;
				   }
                   if(lkg_ga_estimation(max_gen, lkg_est_min_max_flag) == 1) return 1;
                   // clear flag
                   lkg_est_min_max_flag = 0;
                 }
            break;
        case LEAKAGE_RANDOM_SIM:
            if(get_imaxup_read_flag() == 0 || lkg_rilib_flag == 0)
            {  fprintf(stderr," Either netlist or leakage current library is not read in.\n");
               return 1;
            }else{
                   util_getopt_reset();
                   while((c=util_getopt(tmp_argc,tmp_argv,"m:")) != EOF)
                   {
                     switch(c)
                     {
                       case 'm': matched_gen = atoi (util_optarg); break;
                       case '?':
                       default: LKG_usage(lkg_usage, argv[0]); return 1;
                     }
                   }
                   if(lkg_rand_estimation(matched_gen) == 1) return 1;
                 }
             break;
        default:
            fprintf(stderr,"Unknown program function. Debug it.\n");
            /*
            cerr << "Unknown program function. Debug it." << endl;
            */
            return 1;
    }
	
	/* the command is successfully executed */
    return 0;
}

void LKG_usage (usage,cmd)
char **usage;
char  *cmd;
{
    /*  Print a set of usage strings, replacing the first %s with cmd.  A
    newline is automatically added to each line in usage; you can put
    more for special formatting.  The last element in the usage array
    must be a NULL pointer. */

    char **p;

    for (p=usage; *p != NULL; p++) {
    printf(*p,cmd);
    fputs ("\n",stdout);
    }
}
