/***************************************************************
		FILE:		hashtable.c
		DESCR:		hash table operation
		AUTH:		Fei Li
		DATE:		July 2001
****************************************************************/

#include "memuser.h"
#include "linear.h"
#include "linearExt.h"
#include "linearFree.h"
#include "hashtable.h"
#include "DFfile.h"

void
hashTableInit(lin_table_rec **table, int size)
{
	int	i;
	(*table) = (lin_table_rec *)mem_get_block((SIZE_T)(size * sizeof(lin_table_rec)));
	for(i=0;i<size;i++)
	{
		(*table)[i]=NULL;
	}
}

void
hashTableInsert(lin_table_rec *table, char *key, pointer hashdata, int size, unsigned long (*hashfunc)(char *hashkey))
{
	int				hashval;
	lin_table_rec	p, q;
	char			*elemStr;
	int				done;
	HASH_TABLE_ELEM *hashelem;

	done=0;
	hashval = (int)( hashfunc(key) % size );

	p=table[hashval];
	q=NULL;

	while(p!=NULL)
	{
		elemStr=((HASH_TABLE_ELEM *)(p->member))->hashkey;
		if(strcmp(elemStr,key)==0)
		{	done=1;
			break;		/* we already have this key string, nothing to be done */
		}

		q=p;
		p=next_lin_table_rec(p);
	}

	if(done==0)
	{
		hashelem = (HASH_TABLE_ELEM*)mem_get_block((SIZE_T)(sizeof(HASH_TABLE_ELEM)));
		strcpy(hashelem->hashkey, key);
		hashelem->dataP = hashdata;
		/* insert */
		lin_table_prev_ins( &(table[hashval]), NULL,(double) NOT_CARE_KEY, (pointer) hashelem);
	}
}

pointer
hashTableFind(lin_table_rec *table, char *key,  int size, unsigned long (*hashfunc)(char *hashkey))
{
	int				hashval;
	lin_table_rec	p;
	char			*elemStr;

	hashval = (int) ( hashfunc(key) % size );

	p=table[hashval];

	while(p!=NULL)
	{
		elemStr=((HASH_TABLE_ELEM *)(p->member))->hashkey;
		if(strcmp(elemStr,key)==0)
		{
			return ((HASH_TABLE_ELEM *)(p->member))->dataP;
		}else{
				p=next_lin_table_rec(p);
			 }
	}
	return NULL;
}

/* release hashTable element, but not the data in the element */
void
hashTableRelease1(lin_table_rec **table, int size)
{
	int				i;
	lin_table_rec	p;

	for(i=0;i<size;i++)
	{
		p=(*table)[i];
		/*
		while(p!=NULL)
		{
			p=next_lin_table_rec(p);
		}
		*/
		lin_table_rec_Free(&((*table)[i]), NULL);
 	}
 	mem_free_block((pointer) (*table));
 	(*table)=NULL;
}

/* release hashTable element, including the data in the element */
void
hashTableRelease(lin_table_rec **table, int size)
{
	int				i;
	lin_table_rec	p;

	for(i=0;i<size;i++)
	{
		p=(*table)[i];
		while(p!=NULL)
		{
			mem_free_block((pointer) ((HASH_TABLE_ELEM*)(p->member))->dataP);
			p=next_lin_table_rec(p);
		}
		lin_table_rec_Free(&((*table)[i]), NULL);
 	}
 	mem_free_block((pointer) (*table));
 	(*table)=NULL;
}

unsigned long strHF1(char *key)
{
	int len= strlen(key);
    int hashf=0;

	/*if length is 0 or 1, return key[0]; otherwise,
    compute the sum of the first and last chracters*/
    if(len<=1)
        hashf=key[0];
    else
        hashf=key[0]+key[len-1];

    return (unsigned long) (hashf%101);
}

unsigned long strHF2(char *key)
{
	int hashf=0;

    /*sum of characters of key*/
    while(*key)
            hashf += *key++;

    return (unsigned long) (hashf%101);
}

 

