#include "PWcir.h"
#include "PWglobal_var.h"
#include "PWroutines.h"
#include "leakage_ga.h"

extern int lkg_rand_estimation(int matched_gen);
extern void input_vec_copy(PW_INST *SourceP[], int vec[]);

/* lkg_rand_estimation()
 *
 * random simulation for leakage current estimation.
 * Used for comparison with Genetic Algorithm.
 */
int
lkg_rand_estimation(int matched_gen)
{
  double ckt_lkg,rand_max=0, rand_min=-1;
  int i,num_runs, num_control_pi;
  int vec[MAX_PI], vec_min[MAX_PI];

  /* use the same number input vectors of GA as a comparison */
  /* matched_gen should be equal to the max_gen in GA algorithm for comparison */
  num_control_pi = controllable_pi();
  num_runs = matched_gen * DF_MAX(num_control_pi, 24);

  randomGenInit1();

  for(i=0;i<num_runs;i++)
  {
    PWimplyStackGarbageClean();

    /* for leakage current, we decide PI & PPI */
    rand_input_vec_gen(RAND_INPUT_VEC_ALL_PI);

    /* 1-vec simulation */
    simple_sim_operation();

    /* get leakage current: NOTE- PIs are treated as buffers for current value*/
    ckt_lkg = compute_lkg_current();

    rand_max = DF_MAX(rand_max,ckt_lkg);
    if(rand_min == -1) 
      rand_min = ckt_lkg;
	else rand_min = DF_MIN(rand_min,ckt_lkg);

    if(rand_max == ckt_lkg)
    {
      input_vec_copy(SourceP, vec);
    }
    if(rand_min == ckt_lkg)
    {
      input_vec_copy(SourceP, vec_min);
    }
    fprintf(stdout, "Run=%d\tlkg_current=%f\n",i,rand_max);
  }

  fprintf(resultFp,"the max leakage current for random simulation is %f.\n", rand_max);
  dump_input_vector(resultFp,vec);
  fprintf(resultFp,"the min leakage current for random simulation is %f.\n", rand_min);
  dump_input_vector(resultFp,vec_min);

  return 0; /* success */
}
