/************************************************************************
            FILE:       DFfile.c
            DESCR:      linear table operation
            AUTH:       Fei Li
            DATE:       April 1999
*************************************************************************/
 
#include "memuser.h"
#include "linear.h"


/* prev-insertion operation in linear table */
/* new record is created */
lin_table_rec
/***
#if defined(__STDC__)
lin_table_prev_ins(lin_table_rec *head, lin_table_rec *tail, double key, pointer insMbr)
#else
***/
lin_table_prev_ins(head, tail, key, insMbr)
	lin_table_rec *head;
	lin_table_rec *tail;
	double		  key;
	pointer		  insMbr;
/***
#endif
***/
{	lin_table_rec rec;
	rec=(lin_table_rec)mem_get_block((SIZE_T)(sizeof(struct lin_table_rec_)));

	rec->key=key;
	rec->member=insMbr;

	if(*head==NULL)
	{	*head=rec;
		rec->next=NULL;
		if(tail!=NULL)
			*tail=rec;
	}else{
			rec->next=(*head);
			*head=rec;
		 }

	return rec;
}

/*-------------------------------------------------------------*/

/* prev-deletion operation for linear table */
/* record is deleted from lin-table, but memory NOT freed */
lin_table_rec
/***
#if defined(__STDC__)
lin_table_prev_del(lin_table_rec *head, lin_table_rec *tail)
#else
***/
lin_table_prev_del(head, tail)
	lin_table_rec *head;
	lin_table_rec *tail;
/***
#endif
***/
{	lin_table_rec tmpRec;

	if(*head==NULL)
		return NULL;
	else{
			tmpRec=(*head);
			*head=(*head)->next;
			if(*head==NULL && tail!=NULL)
				*tail=NULL;

			return tmpRec;
		}
}

/*--------------------------------------------------------------*/

/* post-insertion operation for linear table */
/* new record is created */
lin_table_rec
/***
#if defined(__STDC__)
lin_table_post_ins(lin_table_rec *head, lin_table_rec *tail, double key, pointer insMbr)
#else
***/
lin_table_post_ins(head, tail, key, insMbr)
	lin_table_rec *head;
	lin_table_rec *tail;
	double		  key;
	pointer		  insMbr;
/***
#endif
***/
{	lin_table_rec rec;
	
	rec=(lin_table_rec)mem_get_block((SIZE_T)(sizeof(struct lin_table_rec_)));
	rec->key=key;
	rec->member=insMbr;

	if(*tail==NULL)
	{	*tail=rec;
		rec->next=NULL;
		if(head!=NULL)
			*head=rec;
	}else{
			(*tail)->next=rec;
			rec->next=NULL;
			*tail=rec;
		 }
	
	return rec;
}

/*--------------------------------------------------------------*/

/* look up a record in a linear table, FROM head TO tail!
	<key> is used as search ID

  !!>>if the <member> used as key, refer to linearExt.c
*/
lin_table_rec
/***
#if defined(__STDC__)
lin_table_look_up(lin_table_rec head, double dstKey, lin_table_rec *prevMbr)
#else
***/
lin_table_look_up(head, dstKey, prevMbr)
	lin_table_rec head;
	double		  dstKey;
	lin_table_rec *prevMbr;
/***
#endif
***/
{	lin_table_rec rec;
	
	/* if not match, or if the matched member is the head element,
	   prevMbr remains NULL
	*/
	*prevMbr=NULL;

	for(rec=head;rec!=NULL;rec=rec->next)
	{	if(rec->key==dstKey)
			break;
		else
			*prevMbr=rec;
	}

	return rec;
}

/*--------------------------------------------------------------*/

/* delete an arbitrary member of a linear table */
/* deleted record is NOT freed */
/* >>>>arg:preDel==NULL, is interpreted as the first rec to be deleted */
lin_table_rec
/***
#if defined(__STDC__)
lin_table_arb_del(lin_table_rec *head, lin_table_rec *tail, lin_table_rec prevDel)
#else
***/
lin_table_arb_del(head, tail, prevDel)
	lin_table_rec *head;
	lin_table_rec *tail;
	lin_table_rec prevDel;
/***
#endif
***/
{	lin_table_rec tmpRec;

	/* the first rec to be deleted */
	if(prevDel==NULL)
	{	tmpRec=*head;
		*head=(*head)->next;
		if(*head==NULL && tail!=NULL)
			*tail=NULL;
	}else{
			tmpRec=prevDel->next;
			prevDel->next=prevDel->next->next;
			if(tail!=NULL && *tail==tmpRec)
				*tail=prevDel;
		 }
	return tmpRec;
}

/* prev-insertion operation in bi-linear-table */
/* new record is created */
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_prev_ins(bi_lin_table_rec *head, bi_lin_table_rec *tail, double key, pointer insMbr)
#else
***/
bi_lin_table_prev_ins(head, tail, key, insMbr)
	bi_lin_table_rec *head;
	bi_lin_table_rec *tail;
	double key;
	pointer insMbr;
/***
#endif
***/
{
	bi_lin_table_rec rec;
    rec=(bi_lin_table_rec)mem_get_block((SIZE_T)(sizeof(struct bi_lin_table_rec_)));
 
    rec->key=key;
    rec->member=insMbr;
 
    if(*head==NULL)
    {   *head=rec;
        rec->next=NULL;
		rec->prev=NULL;

        if(tail!=NULL)
            *tail=rec;
    }else{
            rec->next=(*head);
            *head=rec;
			rec->next->prev=rec;
			rec->prev=NULL;
         }
 
    return rec;
}

/* prev-deletion operation for bi-linear-table */
/* record is deleted from bi-linear-table, but NOT freeed */
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_prev_del(bi_lin_table_rec *head, lin_table_rec *tail)
#else
***/
bi_lin_table_prev_del(head, tail)
    bi_lin_table_rec *head;
    bi_lin_table_rec *tail;
/***
#endif
***/
{   bi_lin_table_rec tmpRec;
 
    if(*head==NULL)
        return NULL;
    else{
            tmpRec=(*head);
            *head=(*head)->next;
			if((*head)!=NULL)
				(*head)->prev=NULL;
			/* the last member deleted */
            if(*head==NULL && tail!=NULL)
                *tail=NULL;
 
            return tmpRec;
        }
}

/* post-insertion operation for bi-linear-table */
/* new record is created */
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_post_ins(bi_lin_table_rec *head, bi_lin_table_rec *tail, double key, pointer insMbr)
#else
***/
bi_lin_table_post_ins(head, tail, key, insMbr)
    bi_lin_table_rec *head;
    bi_lin_table_rec *tail;
    double        	key;
    pointer       	insMbr;
/***
#endif
***/
{   bi_lin_table_rec rec;
   
    rec=(bi_lin_table_rec)mem_get_block((SIZE_T)(sizeof(struct bi_lin_table_rec_)));
    rec->key=key;
    rec->member=insMbr;
 
    if(*tail==NULL)
    {   *tail=rec;
		rec->next=NULL;
		rec->prev=NULL;
        if(head!=NULL)
            *head=rec;
    }else{
			rec->prev=(*tail);
            (*tail)->next=rec;
            rec->next=NULL;
            *tail=rec;
         }
 
    return rec;
}

/* post-deleteion operation for bi-linear-table */
/* record is deleted from bi-linear-table but NOT freed */
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_post_del(bi_lin_table_rec *head, bi_lin_table_rec *tail)
#else
***/
bi_lin_table_post_del(head, tail)
	bi_lin_table_rec 	*head;
	bi_lin_table_rec	*tail;
/***
#endif
***/
{	bi_lin_table_rec tmpRec;

	if(*tail==NULL)
		return NULL;
	else{
			tmpRec=(*tail);
			(*tail)=(*tail)->prev;
			if((*tail)!=NULL)
				(*tail)->next=NULL;
			if((*tail)==NULL && head!=NULL)
				*head=NULL;

			return tmpRec;
		}
}


/* look up a record in a bi-linear-table, FROM head TO tail!
    <key> is used as search ID
 
  !!>>if the <member> used as key, refer to linearExt.c
*/
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_look_up(bi_lin_table_rec head, double dstKey)
#else
***/
bi_lin_table_look_up(head, dstKey)
    bi_lin_table_rec head;
    double        	dstKey;
/***
#endif
***/
{   bi_lin_table_rec rec;
 
    for(rec=head;rec!=NULL;rec=rec->next)
    {   if(rec->key==dstKey)
            break;
    }
 
    return rec;
}


/* delete an arbitrary member of a bi-linear-table */
/* deleted record is NOT freeed */
/* >>>>arg:preDel==NULL, is interpreted as the first rec to be deleted */
/* YOU SHOULD MAKE SURE THE MEMBER IS MATCHED */
bi_lin_table_rec
/***
#if defined(__STDC__)
bi_lin_table_arb_del(bi_lin_table_rec *head, bi_lin_table_rec *tail, bi_lin_table_rec delMbr)
#else
***/
bi_lin_table_arb_del(head, tail, delMbr)
    bi_lin_table_rec *head;
    bi_lin_table_rec *tail;
    bi_lin_table_rec delMbr;
/***
#endif
***/
{   bi_lin_table_rec tmpRec;
 
    if(delMbr==NULL)
       tmpRec=NULL;
    else{
            tmpRec=delMbr;

			/* the last member */
			if(delMbr->next==NULL && tail!=NULL)
				(*tail)=delMbr->prev;
			/* the first member */
			if(delMbr->prev==NULL)
			(*head)=delMbr->next;
			
			if(delMbr->prev!=NULL)
				delMbr->prev->next=delMbr->next;
			if(delMbr->next!=NULL)
				delMbr->next->prev=delMbr->prev;

       }

    return tmpRec;
}

