/*
 * com_mape.c 
 *
 * used to insert the mape command into sis. 
 * Changbo Long(clong@cae.wisc.edu)
 */
#include "mape_int.h"

#define MAX_GATE_LINE	2048

static int com_mape();
static void mape_destroy_user_option(void) ;

FILE *mapeout ;
mape_option_t *user_option = NULL;

void
init_mape()
{
	com_add_command("mape", com_mape, 1) ;
		/*
	if((mapeout = fopen(MAPEOUTPUT, "w+")) == NULL){
		fprintf(siserr, "Warning: Can not open %s, using sisout as output 
			for mape\n", MAPEOUTPUT) ;  
		mapeout = sisout ;
	}
		*/
}

void
end_mape()
{
	mape_destroy_user_option() ;
}

void
init_cmssc(void)
{
}

void
end_cmssc(void)
{
}

void
init_imax_powerup()
{
}

void
end_imax_powerup()
{
}

/*
 * create_default_option
 *
 * create the default user's option
 */
void 
mape_create_default_option(void)
{
	user_option = ALLOC(mape_option_t, 1) ;

	user_option->method 		= USING_PODEM ;
	user_option->rdm_option.times 	= 100 ;
	user_option->rdm_option.perc	= 100  ; 
	user_option->verb		= FALSE ;
	user_option->output		= NULL ;
	user_option->given_gates 	= NULL ;	
	user_option->delay_model 	= ZERO_DELAY_MODEL ;
	user_option->ssc_obj 		= SSCOBJ_GROUP_SUM ;
	user_option->mug		= 0 ;	/* max unchangeable group */
	user_option->version		= 3 ;	/* version of the user's option */ 
	user_option->max_cp		= 3.0 ;	/*default required critical path */ 
	user_option->podem_option.fanout_priority	= 8;
	user_option->podem_option.level_priority	= 2;
	user_option->podem_option.backtrack_limit	= 100;
	user_option->podem_option.backtrace_limit	= 40; /*40 percent maximum pi could be tried when backtrace*/

	user_option->curga_option.generation		= 50; /*default generation number*/
	user_option->curga_option.pi_bit		= 0 ; /*the bit number to specify the switching time in each pi*/
	user_option->curga_option.bit_unit		= 100 ;	/* How many percent of a gate delay for the distance
								   between two adjacent switchs. */
}

/*
 * mape_usage
 *
 * how to use the command of mape. 
 */
void 
mape_usage(void)
{
	(void) fprintf(stderr, "usage: mape [-m method] [-v] [-o filename] [-f #] [-l #] [-t #] [-p #]\n") ;
	(void) fprintf(stderr, "\t-m\tspecify method, could be iteration, podem or random.\n") ;
	(void) fprintf(stderr, "\t-v\tverbose mode.\n") ;
	(void) fprintf(stderr, "\t-o\toutput file\n") ;
	(void) fprintf(stderr, "\t-s\tgiven gate set file\n");
	(void) fprintf(stderr, "\t-f\t0~10, for podem method, fanout priority.\n");
	(void) fprintf(stderr, "\t-l\t0~10, for podem method, level priority.\n");
	(void) fprintf(stderr, "\t-t\t1~infinity, for random method, simulation times.\n");
	(void) fprintf(stderr, "\t-p\t0~100, for random method, input transition probability.\n") ;
	(void) fprintf(stderr, "\t-k\t1~infinity, for random method, backtrack limitation of podem.\n") ;
	(void) fprintf(stderr, "\t-e\t1~100, for random method, accessible primary input percent in backtrace.\n") ;
	(void) fprintf(stderr, "\t-d\tzero, unit or real\n") ;
	(void) fprintf(stderr, "\t-S\tobjective function type for SSC, group_sum, group_max or msp_max.\n") ;
}

/*
 * mape_is_given_node
 *
 * is the node a user specified node?
 */
int
mape_is_given_node(node)
node_t *node ;
{
	if(node_type(node) == PRIMARY_INPUT)
		return FALSE ;
	if(!(user_option->given_gates))
		return TRUE ;
	if(lsLength(user_option->given_gates) == 0)
		return TRUE ;
	return mape_node_in_list(node, user_option->given_gates) ;
}

/*
 * mape_clean_given_gate
 *
 * clean the given gate list. 
 */
void 
mape_clean_given_gate(void)
{
	node_t *tmp ;

	if(!(user_option->given_gates))
		return ;
	while(lsLength(user_option->given_gates))
		LS_ASSERT(lsDelEnd(user_option->given_gates, (lsGeneric *)&tmp)) ;
}

/*
 * mape_push_given_gatep
 *
 * push user defined gates into given_gates ;
 */
void 
mape_push_given_gatep(node)
node_t *node ;
{
	if(!node)
		return ;

	if(!(user_option->given_gates))
		user_option->given_gates = lsCreate() ;
	if(!mape_node_in_list(node, user_option->given_gates))
		lsNewEnd(user_option->given_gates, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_push_given_gate
 *
 * push user defined gates into the 
 * user_option's given_gates ;
 */
void 
mape_push_given_gate(network, name)
network_t *network ;
char *name ;
{
	node_t *node ;

	node = network_find_node(network, name) ; 
	if(node){
		if(!(user_option->given_gates))
			user_option->given_gates = lsCreate() ;
		if(!mape_node_in_list(node, user_option->given_gates))
			lsNewEnd(user_option->given_gates, (lsGeneric)node, LS_NH) ;
	}else 
		fprintf(siserr, "Warning: can't find specified node: %s, discard it!\n", name) ;
}


/*
 * mape_parse_gate_file
 *
 * parse the the gate file where some gates 
 * are defined by the user. 
 */
void 
mape_parse_gate_file(network, fp)
network_t *network ;
FILE *fp ;
{
	char line[MAX_GATE_LINE] ;
	char *word ;

	while(!feof(fp)){
		fgets(line, MAX_GATE_LINE, fp) ;
		word = strtok(line, " ,\n") ;
		while(word != NULL){
			if(strlen(word) > 0) 
				mape_push_given_gate(network, word) ;
			word = strtok(NULL, " ,\n") ;
		}
	}
}

/*
 * mape_get_given_gates
 *
 *
 * get the user specified gates. 
 */
void 
mape_get_given_gates(network, s)
network_t *network ;
char *s ;
{
	FILE *fp ;

	if((fp = fopen(s, "r")) == NULL){
		fprintf(siserr, "Warning: can't open specified file %s, discarding -s option!\n", s) ;
		return ;
	}
	mape_parse_gate_file(network, fp) ;
}

/*
 * mape_get_user_option
 * 
 * get the user's option's. 
 */
int 
mape_get_user_option(network, argc, argv)
network_t *network ;
int argc ;
char **argv ;
{
	int c ;	
	FILE *fp ;

	util_getopt_reset() ;
	while((c = util_getopt(argc, argv, "vm:g:b:u:p:t:f:l:o:k:e:s:d:S:G:V:")) != EOF){
		switch(c){
			case 'v':
				user_option->verb = TRUE ;
				break ;
			case 'o':
				if((fp = fopen(util_optarg, "w+")) == NULL){
					mape_usage() ;
					return FALSE ;
				}
				user_option->output = fp ;
				break ;
			case 'm':
				if(!strcmp(util_optarg, USING_PODEM_S)){
					user_option->method = USING_PODEM ;
					break ;
				}else if(!strcmp(util_optarg, USING_RDM_S)){
					user_option->method = USING_RDM ;
					break ;
				}else if(!strcmp(util_optarg, USING_ITR_S)){
					user_option->method = USING_ITR ;
					break ;
				}else if(!strcmp(util_optarg, USING_UPB_S)){
					user_option->method = USING_UPB ;
					break ;
				}else if(!strcmp(util_optarg, USING_GROUP_S)){
					user_option->method = USING_GROUP ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURRENT_S)){
					user_option->method = USING_CURRENT ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURITR_S)){
					user_option->method = USING_CURITR ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURGRP_S)){
					user_option->method = USING_CURGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_SPICE_S)){
					user_option->method = USING_SPICE ;
					break ;
				}else if(!strcmp(util_optarg, USING_SPIGRP_S)){
					user_option->method = USING_SPIGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_OPTGRP_S)){
					user_option->method = USING_OPTGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_SAGRP_S)){
					user_option->method = USING_SAGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_SAPLT_S)){
					user_option->method = USING_SAPLT ;
					break ;
				}else if(!strcmp(util_optarg, USING_SASSC_S)){
					user_option->method = USING_SASSC ;
					break ;
				}else if(!strcmp(util_optarg, USING_MSPPLT_S)){
					user_option->method = USING_MSPPLT ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURSA_S)){
					user_option->method = USING_CURSA ;
					break ;
				}else if(!strcmp(util_optarg, USING_TATPG_S)){
					user_option->method = USING_TATPG ;
					break ;
				}else if(!strcmp(util_optarg, USING_PIPELINE_S)){
					user_option->method = USING_PIPELINE ;
					break ;
				}else if(!strcmp(util_optarg, USING_CLOCKSKEW_S)){
					user_option->method = USING_CLOCKSKEW;
					break ;
				}else if(!strcmp(util_optarg, USING_CURGA_S)){
					user_option->method = USING_CURGA;
					break ;
				}else if(!strcmp(util_optarg, USING_CURRDM_S)){
					user_option->method = USING_CURRDM;
					break ;
				}else if(!strcmp(util_optarg, USING_TATPGGRP_S)){
					user_option->method = USING_TATPGGRP;
					break ;
				}else if(!strcmp(util_optarg, USING_CURGAGRP_S)){
					user_option->method = USING_CURGAGRP;
					break ;
				}else if(!strcmp(util_optarg, USING_CURLOC_S)){
					user_option->method = USING_CURLOC ;
					break ;
				}else if(!strcmp(util_optarg, USING_PBSSCE_S)){
					user_option->method = USING_PBSSCE ;
					break ;
				}else if(!strcmp(util_optarg, USING_PBSSCEGRP_S)){
					user_option->method = USING_PBSSCEGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_RDMAVG_S)){
					user_option->method = USING_RDMAVG ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURPLOC_S)){
					user_option->method = USING_CURPLOC ;
					break ;
				}else if(!strcmp(util_optarg, USING_DUMP_S)){
					user_option->method = USING_DUMP ;
					break ;
				}else if(!strcmp(util_optarg, USING_CURRDMGRP_S)){
					user_option->method = USING_CURRDMGRP ;
					break ;
				}else if(!strcmp(util_optarg, USING_SOSA_S)){
					user_option->method = USING_SOSA ;
					break ;
				}else if(!strcmp(util_optarg, USING_DEPSA_S)){
					user_option->method = USING_DEPSA ;
					break ;
				}else if(!strcmp(util_optarg, USING_DISKSIM_S)){
					user_option->method = USING_DISKSIM;
					break ;
				}else if(!strcmp(util_optarg, USING_STN_S)){
					user_option->method = USING_STN;
					break ;
				}else{ 
					mape_usage() ;
					return FALSE ;
				}
			case 'g':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->curga_option.generation = atoi(util_optarg) ;
				break ;
			case 'b':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->curga_option.pi_bit = atoi(util_optarg) ;
				break ;
			case 'u':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->curga_option.bit_unit = atoi(util_optarg) ;
				break ;
			case 'p':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->rdm_option.perc = atoi(util_optarg) ;
				break ;
			case 't':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->rdm_option.times = atoi(util_optarg) ;
				break ;
			case 'f':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->podem_option.fanout_priority = atoi(util_optarg) ;
				break ;
			case 'l':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->podem_option.level_priority = atoi(util_optarg) ;
				break ;
			case 'k':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->podem_option.backtrack_limit = atoi(util_optarg) ;
				break ;
			case 'e':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->podem_option.backtrace_limit = atoi(util_optarg) ;
				break ;
			case 's': 
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}
				mape_get_given_gates(network, util_optarg) ;
				break ;
			case 'd': 
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}
				if(!strcmp(util_optarg, ZERO_DELAY_MODEL_S)){
					user_option->delay_model = ZERO_DELAY_MODEL;
					break ;
				}else if(!strcmp(util_optarg, UNIT_DELAY_MODEL_S)){
					user_option->delay_model = UNIT_DELAY_MODEL;
					break ;
				}else if(!strcmp(util_optarg, REAL_DELAY_MODEL_S)){
					user_option->delay_model = REAL_DELAY_MODEL;
					break ;
				}else {
					mape_usage() ;
					return FALSE ;
				}
			case 'S': 
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}
				if(!strcmp(util_optarg, SSCOBJ_GROUP_SUM_S)){
					user_option->ssc_obj = SSCOBJ_GROUP_SUM ;
					break ;
				}else if(!strcmp(util_optarg, SSCOBJ_GROUP_MAX_S)){
					user_option->ssc_obj = SSCOBJ_GROUP_MAX ;
					break ;
				}else if(!strcmp(util_optarg, SSCOBJ_MSP_MAX_S)){
					user_option->ssc_obj = SSCOBJ_MSP_MAX ;
					break ;
				}else {
					mape_usage() ;
					return FALSE ;
				}
			case 'G':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->mug = atoi(util_optarg) ;
				break ;
			case 'V':
				if(strlen(util_optarg) == 0){
					mape_usage() ;
					return FALSE ;
				}	
				user_option->version = atoi(util_optarg) ;
				break ;
			default:
				mape_usage();
				return FALSE;
					
		}
	}
	return TRUE ;
}

/*
 * mape_destroy_user_options
 *
 * destroy the user_options structure. 
 */
void 
mape_destroy_user_option(void)
{
	if(user_option->output != NULL)
		fclose(user_option->output) ;
	free(user_option) ;
	user_option = NULL ;
}

/*
 * test the control value function. 
 *
 */
void
test_control_value(void)
{
	int b[4][4] = { {ZERO, TWO, ONE, ONE}, 
			{TWO, ZERO, TWO, TWO}, 
			{ZERO, TWO, ONE, ONE}, 
			{ZERO, TWO, ONE, ONE} } ; 
	pcube cb[4] ;
	pcover cover ;
	int i, j, cv, pos;

	for(i=0;i<4;i++) 
		cb[i] = set_new(4) ;
	cover = new_cover(MAXFAMILYCUBES) ;

	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			PUTINPUT(cb[i], j, b[i][j]) ;
	dump_s_set("cb[0]=", 4, cb[0]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[1]=", 4, cb[1]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[2]=", 4, cb[2]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[3]=", 4, cb[3]);
	fprintf(mapeout, "\n") ;
	for(i=0;i<4;i++)
		mape_sf_append_set(cover, cb[i]) ;

	cube.num_vars = 4 ;
	switch(mape_node_ctrl_value(cover, &pos, &cv)){
		case NO_CTRL_VALUE:
			fprintf(mapeout, "no control value for the above cubes\n") ;
			break ;
		case NORMAL_CTRL_VALUE:
			fprintf(mapeout, "normal ctrl value is %d, position is %d\n", cv-1, pos) ;
			break ;
		case REVERSE_CTRL_VALUE:
			fprintf(mapeout, "reverse ctrl value is %d, position is %d\n", cv-1, pos) ;
			break ;
		default:
			break ;	
	}
	fflush(mapeout) ;
}

/*
 * test_cube_operations
 *
 * This function is used to test the cube's operation. 
 */
void 
test_cube_operations(network)
network_t *network ;
{
	int a[4] = {ONE, TWO, TWO, TWO} ;
	int b[4][4] = { {ZERO, ZERO, ZERO, ZERO}, 
			{ZERO, ZERO, ZERO, ONE}, 
			{ZERO, ONE, ONE, ONE}, 
			{ZERO, ZERO, ONE, ONE} } ; 
	int i, j ;
	pcube ca ;
	pcube cb[4] ;
	pcover f ;
	pcover r ;


	cube.num_vars = 4 ;
	ca = set_new(4) ;
	for(i=0;i<4;i++) 
		cb[i] = set_new(4) ;
	f = new_cover(MAXFAMILYCUBES) ;

	for(i=0;i<4;i++) 
		PUTINPUT(ca, i, a[i]) ;
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			PUTINPUT(cb[i], j, b[i][j]) ;
	dump_s_set("ca=", 4, ca);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[0]=", 4, cb[0]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[1]=", 4, cb[1]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[2]=", 4, cb[2]);
	fprintf(mapeout, "\n") ;
	dump_s_set("cb[3]=", 4, cb[3]);
	fprintf(mapeout, "\n") ;
	for(i=0;i<4;i++)
		mape_sf_append_set(f, cb[i]) ;

	r = cb_sharp(ca, f) ;

	fprintf(mapeout, "r=") ;
	dump_sf(4, r) ;
	fprintf(mapeout, "\n") ;
	fflush(mapeout) ;

	fprintf(mapeout, "f=") ;
	dump_sf(4, f) ;
	fprintf(mapeout, "\n") ;
	fflush(mapeout) ;
}

/*
 * test_backtrace. 
 *
 * test the backtrace function. 
 */
void 
test_backtrace(node, value)
node_t *node ;
mape_node_value_t value ;
{
	node_t *pi ;
	mape_node_value_t pi_v ;
	char text[128] ;
	
	if(node == NULL){
		fprintf(mapeout, "Please don't pass me NULL node\n") ;
		fflush(mapeout) ;
		return ;
	}
	mape_podem_backtrace(node, value, &pi, &pi_v) ;
	fprintf(mapeout, "backtrace: %s(%s)->...", 
		node->name, mape_node_value_string(value, text)) ;
	fprintf(mapeout, "->%s(%s)\n", 
		((pi == NULL) ? "NULL":pi->name), mape_node_value_string(pi_v, text)) ;
	fflush(mapeout) ;
}


/*
 * function to test the backtrack in the circuit. 
 *
 */
void 
test_backtrack(node, value)
node_t *node ;
mape_node_value_t value ;
{
	char text[128] ;
	fprintf(mapeout, "Test backtrack function, initial condition: %s=%s\n", 
		node->name, mape_node_value_string(value, text))  ;
	fflush(mapeout) ;
	mape_podem_backtrack(node, value) ;
}

/*
 * function to test the implication in the circuit. 
 *
 */
void 
test_implication(node, value)
node_t *node ;
mape_node_value_t value ;
{
	char text[128] ;
	fprintf(mapeout, "Test implication function, initial condition: %s=%s\n", 
		node->name, mape_node_value_string(value, text))  ;
	mape_podem_implication(node, value) ;
}

/*
 * pick_ith_input
 *
 * pick up the ith input for backtrace. 
 */
node_t *
pick_ith_input(network, n) 
network_t *network ;
int n ;
{
	lsGen gen ;
	node_t *p ;
	int i = 0 ;

	foreach_primary_input(network, gen, p){
		i++ ;
		if(i == n)
			return p ;
	}
	return p ;
}

/*
 * pick_ith_output
 *
 * pick up the ith input for backtrace. 
 */
node_t *
pick_ith_output(network, n) 
network_t *network ;
int n ;
{
	lsGen gen ;
	node_t *p ;
	int i = 0 ;

	foreach_primary_output(network, gen, p){
		i++ ;
		if(i == n)
			return p ;
	}
	return p ;
}

/*
 * com_mape
 * 
 * The entrance of the mape package appended to sis. 
 */
int 
com_mape(network, argc, argv)
network_t **network; 
int argc ;
char **argv ;
{
	long last_time ;
	long x ;
	int i ;
	int tmp ;
	double time ;

	mape_create_default_option() ;

	if ((*network) == NIL(network_t)) 
		return 0;
	if(!mape_get_user_option(*network, argc, argv))
		return 1;
	if (network_num_internal(*network) == 0) 
		return 0;

	global_network = *network ;


	mapeout = user_option->output?user_option->output:sisout ;
	user_option->podem_option.backtrace_limit_a=
		(int)((float)(network_num_pi(*network)*(user_option->podem_option.backtrace_limit))/100.0) ;
	tmp = user_option->podem_option.backtrace_limit_a ;
	user_option->podem_option.backtrace_limit_a= tmp > 5 ? tmp : 5 ;

	last_time = util_cpu_time() ;
	if(user_option->method == USING_PODEM){
		//dump_network(*network) ;
		mape(*network) ;
		last_time = util_cpu_time() - last_time ;
		mape_report(*network, user_option->verb, last_time) ;
	}else if(user_option->method == USING_RDM){
		//dump_network(*network) ;
		mape_rdm(*network, user_option->rdm_option.times, user_option->rdm_option.perc) ; 
		last_time = util_cpu_time() - last_time ;
		mape_rdm_report(*network, user_option->verb, last_time) ;
	}else if(user_option->method == USING_ITR){
		dump_network(*network) ;
		mape_itr(*network) ;
		last_time = util_cpu_time() - last_time ;
		mape_report(*network, user_option->verb, last_time) ;
	}else if(user_option->method == USING_UPB){
		//dump_network(*network) ;
		mape_upb(*network) ;
	}else if(user_option->method == USING_GROUP){
		//dump_network(*network) ;
		mape_group(*network) ;
	}else if(user_option->method == USING_CURRENT){
		//dump_network(*network) ;
		mape_current(*network) ;
	}else if(user_option->method == USING_CURITR){
		//dump_network(*network) ;
		mape_curitr(*network, &time, NULL) ;
	}else if(user_option->method == USING_CURGRP){
		//dump_network(*network) ;
		mape_curgrp(*network) ;
	}else if(user_option->method == USING_SPICE){ //dump the circuit information into spice format
		mape_spice(*network) ;
	}else if(user_option->method == USING_SPIGRP){ //dump the circuit information into spice format
		mape_spigrp(*network) ;
	}else if(user_option->method == USING_OPTGRP){ //dump the circuit information into spice format
		//mape_optgrp(*network) ;
		//mape_sagrp(*network) ;
	}else if(user_option->method == USING_SAGRP){ //dump the circuit information into spice format
		//mape_optgrp(*network) ;
		mape_sagrp(*network) ;
	}else if(user_option->method == USING_SAPLT){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_saplt(*network) ;
	}else if(user_option->method == USING_SASSC){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_sassc(*network) ;
	}else if(user_option->method == USING_MSPPLT){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_mspplt(*network) ;
	}else if(user_option->method == USING_CURSA){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_cursa(*network, &time, NULL) ;
	}else if(user_option->method == USING_TATPG){ //dump the circuit information into spice format
		dump_network(*network) ;
		mape_tatpg(*network) ;
	}else if(user_option->method == USING_PIPELINE){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_pipeline(*network) ;
	}else if(user_option->method == USING_CLOCKSKEW){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_clockskew(network) ;
	}else if(user_option->method == USING_CURGA){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_curga(*network) ;
	}else if(user_option->method == USING_CURRDM){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_currdm(*network) ;
	}else if(user_option->method == USING_CURGAGRP){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_curga_grp(*network) ;
	}else if(user_option->method == USING_TATPGGRP){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_tatpg_grp(*network) ;
	}else if(user_option->method == USING_CURLOC){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_curloc(*network) ;
	}else if(user_option->method == USING_PBSSCE){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_pbssce(*network) ;
	}else if(user_option->method == USING_PBSSCEGRP){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_pbssce_grp(*network) ;
	}else if(user_option->method == USING_RDMAVG){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_rdmavg(*network) ;
	}else if(user_option->method == USING_CURPLOC){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_curploc(*network) ;
	}else if(user_option->method == USING_CURRDMGRP){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_currdm_grp(*network) ;
	}else if(user_option->method == USING_SOSA){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_sosa(*network) ;
	}else if(user_option->method == USING_DEPSA){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_depsa(*network) ;
	}else if(user_option->method == USING_DISKSIM){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_disksim(*network) ;
	}else if(user_option->method == USING_STN){ //dump the circuit information into spice format
		//dump_network(*network) ;
		mape_stn(*network) ;
	}else if(user_option->method == USING_DUMP){ //dump the circuit information into spice format
		mape_stn_test() ;
		//dump_network(*network) ;
	}
	return 1;
}
