#ifndef _GENETIC_ALGORITHM_TEMPLATE_
#define _GENETIC_ALGORITHM_TEMPLATE_

#include "DFtype.h"
#include "mape_int.h"

typedef struct chromosome{
int     *data;      /* the data in the chromosome */
double  fitness;    /* the fitness value */
BOOLEAN visited;    /* flag */
} chromosome_t;


/* initialize the Genetic Algorithm */
extern void init_ga_alg(int gene_count_l,int gene_bits_l,int chromo_wsize_l,int pop_size_l,int max_generation_l);

/* access to the static vars */
extern int get_gene_count(void);
extern int get_gene_bits(void);
extern int get_chromo_wsize(void);
extern int get_pop_size(void);
extern int get_max_generation(void);
extern int get_cur_generation(void);


/* a genetic algorithm template */
extern chromosome_t * ga_alg( array_t * (*init_pop)(void),
        double fitness_func(chromosome_t *chr),
        array_t * (*select)(array_t *pop),
        void (*crossover)(array_t *pop),
        void (*mutation)(array_t *pop),
        void (*replace)(array_t **old_pop, array_t *new_pop),
        BOOLEAN (*finish)(int gen_num, int max_gen, int no_improve_num) );

/* compute the fitness values and return the best chromosome */
extern chromosome_t* eval_fitness( array_t *pop, double (*fitness)(chromosome_t *) );

/* update the best chromo so far
   and return the number of consecutive generations with no improvement 
 */
extern void update_best_chromo(chromosome_t **best, chromosome_t *current_best, int *no_improve_num);

extern void clear_visited_flag(array_t *pop);

/* HACK: set one gene bit (binary) in a chromosome */
extern void set_chromo_bit(chromosome_t *chromo_p, int position, int value);

extern int get_word_index(int gene_position);
extern int get_word_offset(int gene_position);

extern chromosome_t * new_chromosome(void);

/* duplicate the chromo and insert into the array as the last element.
 * global var 'chromo_wsize' is used */
extern void chromo_dup_and_ins_last(array_t* chr_array, chromosome_t * chr);

/* duplicate a chromosome. global var 'chromo_wsize' is used */
extern chromosome_t * dup_chromo(chromosome_t * chr);


#endif
