/*
 * mape.c
 *
 * The file includes the mape function of the 
 * mape package, which is the master function 
 * in the package. 
 */
#include "mape_int.h"

static lsList trans_frontier = NULL ;
static int backtrack_limit = 0 ;
network_t *global_network = NULL ;
int backtrace_num = 0 ;

static int mape_podem_pick_frontier(node_t **, node_t **, mape_node_value_t *) ;
static float mape_podem_node_priority(node_t *);
static void mape_podem_pop_frontier(node_t *) ;

/*
 * mape_podem_pop_frontier
 *
 * pop the node in the frontier while backtrack.  
 */
void 
mape_podem_pop_frontier(node)
node_t *node ;
{
	int del = FALSE ;
	lsGen gen ;
	node_t *p, *tmp  ;

	/*
	fprintf(mapeout, "enter into pop frontier\n") ;
	fflush(mapeout) ;
	*/
	if(node == NULL)
		mape_error("Unexpected NULL node\n", __FILE__, __LINE__) ;
	if(trans_frontier == NULL)
		mape_error("Unexpected NULL list\n", __FILE__, __LINE__) ;

	if(lsLength(trans_frontier) != 0)
		lsForeachItem(trans_frontier, gen, p)
			if(node == p){
				LS_ASSERT(lsDelBefore(gen, (lsGeneric *)&tmp));
				/*
				fprintf(mapeout, "pop frontier: node = %s, tmp = %s\n", node->name, tmp->name) ;
				fflush(mapeout) ;
				*/
				return ;
			}
	mape_error("Can not find the node to be removed\n", __FILE__, __LINE__);
}

/*
 * mape_podem_push_frontier
 *
 * push a node into the transition's frontier list. 
 */
void 
mape_podem_push_frontier(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p ;

	if(node == NULL)
		return ;
	if(trans_frontier == NULL) 
		trans_frontier = lsCreate() ;
	
	/* if node contained, return */
	/*
	if(lsLength(trans_frontier) != 0)
		lsForeachItem(trans_frontier, gen, p)
			if(node == p)
				return ;
	*/

	lsNewEnd(trans_frontier, (lsGeneric)node, NULL) ;
}

/*
 * mape_podem_pack_frontier
 *
 * remove the non-frontier from the trans_frontier list. 
 */
void 
mape_podem_pack_frontier(void)
{
	lsGen gen ;
	node_t *p ;

	if(!trans_frontier)
		return ;

	if(lsLength(trans_frontier) != 0)
		lsForeachItem(trans_frontier, gen, p)
			if(mape_value_determined(mape_top(p)))
				LS_ASSERT(lsDelBefore(gen, (lsGeneric *)&p));
}

/*
 * mape_simulate_node
 *
 * simulate the node according to the value in the podem's stack. 
 * Just using the cude's sharp product operation to finish it. 
 * It is 9 value based. 
 */
mape_node_value_t 
mape_simulate_node(node)
node_t *node ;
{
	int value ;
	pcube input ;
	int i ;
	int oF, oS ; //first and second output. 
	mape_node_value_t o ;
	node_t *p ;

	o = mape_top(node) ;

	/* primary input's value must be set */
	if(node_type(node) == PRIMARY_INPUT)
		return o ; 

	/* value already determined, don't need simulate again */
	if(mape_value_determined(o))
		return o ;

	cube.num_vars = node->nin ;
	define_cube_size (node_num_fanin(node)) ;

	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_top_first(p) ;
		PUTINPUT(input, i, value) ;
	}
	oF = mape_set_simulate_node(input, node) ; 
	foreach_fanin(node, i, p){
		value = mape_top_second(p) ;
		PUTINPUT(input, i, value) ;
	}
	oS = mape_set_simulate_node(input, node) ; 
	o = mape_value_convert(oF, oS) ;

	free_cube(input) ;
	return o ;
}


/*
 * mape_set_simulate_node
 *
 * simulate the node under the input set mode. 
 */
int 
mape_set_simulate_node(input, node)
pcube input;  
node_t *node ;
{
	pcover rtv ;
	pcube first ;

	switch(node_function(node)){
		case NODE_PI:
		case NODE_PO:
		case NODE_UNDEFINED:
		case NODE_BUF: 
			return GETINPUT(input, 0) ;
		case NODE_INV:
			return mape_set_value_not(GETINPUT(input, 0)) ;
		case NODE_0:
			return ZERO ;
		case NODE_1:
			return ONE ;
			break ;
		/*need sharp product*/
		case NODE_AND:
		case NODE_OR:
		case NODE_COMPLEX:
		default:
			rtv = cb_sharp(input, node->F);
			break ;
	}

	if(rtv->count == 0){ 	// all input all covered by function. 
		free_cover(rtv) ;
		return ONE ; 
	}else if(rtv->count >= 2){ //cube are partially covered by function. 
		free_cover(rtv) ;
		return TWO ;
	}else {
		first = (pcube)rtv->data ;	
		if(setp_equal(first, input)){
			free_cover(rtv) ;
			return ZERO ;
		}else {
			free_cover(rtv) ;
			return TWO ;
		}
	}
}

/*
 * mape_podem_pick_frontier
 *
 * 
 * pick a frontier from the frontier list. 
 */
int 
mape_podem_pick_frontier(frontier, fanin, fanin_value)
node_t **frontier ;		/*transistion frontier*/
node_t **fanin ;		/*selected fanin to make the transistion propogation*/
mape_node_value_t *fanin_value ; /*fanin's value*/
{
	lsGen gen ;
	node_t *p ;
	node_t *backup = NULL ;
	float priority, new_prio;
	int i = 0 ;
	int cv ;
	int backup_cv ;
	char text[128] ;

	if(trans_frontier == NULL)
		return FALSE ;

	if(lsLength(trans_frontier) == 0)
		return FALSE ;

	*frontier = NULL ;

	/* select the frontier which has the largest priority */
	lsForeachItem(trans_frontier, gen, p){
		if(mape_value_determined(mape_top(p)))
			continue ;
		if(*frontier == NULL){
			*frontier = p ;
			priority = mape_podem_node_priority(p);
		}else if((new_prio = mape_podem_node_priority(p)) > priority){
			*frontier = p ;
			priority = new_prio ;
		}
		i++ ;
	}

	if(*frontier == NULL)
		return FALSE ;

	/*
	 * if this gate has a controlling value, 
	 * fanin is determined by the controlling value. 
	 */
	cube.num_vars = (*frontier)->nin;
	define_cube_size((*frontier)->nin) ;
	foreach_fanin((*frontier), i, p){
		if(!mape_value_determined(mape_top(p))){
			if(mape_node_ctrl_input((*frontier)->F, i, &cv) != NO_CTRL_VALUE){ 
				*fanin_value = mape_value_convert(mape_set_value_not(cv), cv) ;
				*fanin = p ;
				return TRUE;
			}
			if(!backup){
				backup = p ;
				backup_cv = cv ;
			}
		}
	}

	/* no ctrl value in any input position, pick the last fanin tried above*/
	*fanin = backup;
	*fanin_value = MAPE_RISE ;
	/*
	fprintf(mapeout, "pick up frontier %s \n", (*frontier)->name) ;
	fflush(mapeout) ;
	*/
	return TRUE ;
}

/*
 * mape_podem_backtrack_do
 *
 * for the utility of mape_imp. 
 */
int
mape_podem_backtrack_do(node, arg)
node_t *node ;
char *arg ;
{
	node_t *pi = ((mape_nv_pair_t *)arg)->node ;
	mape_node_value_t pi_value = ((mape_nv_pair_t *)arg)->value ;

	if(mape_pop(node, pi, pi_value) == TRUE){
		return TRUE ;
	}else{ 
		mape_podem_pop_frontier(node) ;
		return FALSE ;
	}
}

/*
 * mape_podem_backtrack 
 *
 * recursively backtrack the last decision made before. 
 */
void 
mape_podem_backtrack(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_nv_pair_t pi_pair ;
	/*
	fprintf(mapeout, "backtrack is called\n") ;
	fflush(mapeout) ;
	*/

	/* can not implicate any more gates */
	if(pi_value == MAPE_U_U)
		return ;

	pi_pair.node = pi ;
	pi_pair.value = pi_value ;

	backtrack_limit ++ ;
	mape_imp_reset(mape_podem_backtrack_do, (char *)&pi_pair) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_podem_imp_do
 *
 * for the mape_imp utility. 
 */
int 
mape_podem_imp_do(node, arg)
node_t *node ;
char *arg ;
{
	mape_node_value_t value ;
	if(node == NULL)
		return FALSE ;

	value = mape_simulate_node(node) ;

	/*processing stop*/
	if(value == MAPE_U_U){
		mape_podem_push_frontier(node);
		return FALSE ;
	}

	if(node->type != PRIMARY_INPUT)
		mape_push(node, value, ((mape_nv_pair_t *)arg)->node, ((mape_nv_pair_t *)arg)->value) ;

	return TRUE ;
}

/*
 * mape_podem_implication 
 *
 * recursively implication of a determined gate value in the circuit 
 * from a pi. 
 */
void 
mape_podem_implication(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_nv_pair_t pi_pair ;

	/*can not implicate any more gates */
	if(pi_value == MAPE_U_U)
		return ;

	pi_pair.node = pi ; 
	pi_pair.value = pi_value ;

	mape_push(pi, pi_value, pi, pi_value) ;
	mape_imp_reset(mape_podem_imp_do ,(char *)&pi_pair) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_node_not_family
 *
 * used in backtrace, determine the node is inverter like or not. 
 */
int 
mape_node_not_family(node)
node_t *node ;
{
	node_function_t func ;
	pset ps, last_ps ;
	int pos, cv, output ;
	pcube *cubes ;

	if(!node->F)
		return FALSE ;

	cube.num_vars = node->nin ;
	define_cube_size(node->nin) ;
	func = node_function(node) ;
	if(func == NODE_INV){
		return TRUE ; 
	}else if(func == NODE_1){
		return FALSE ;
	}else if(func == NODE_0){
		return TRUE ;
	}else{
		/*find the control value from the node's function*/
		switch(mape_node_ctrl_value(node->F, &pos, &cv)){
			case REVERSE_CTRL_VALUE :
				return (cv == ONE) ;
			case NORMAL_CTRL_VALUE:
				return (cv == ZERO);
			case NO_CTRL_VALUE :
			default:
				return FALSE ;
		}
	}	
	return FALSE ;
}

/*
 * mape_node_loading_cap
 *
 * get node's loading capacitance
 */
float
mape_node_loading_cap(node)
node_t *node ;
{
	return lsLength(node->fanout)*1.0 ;
}

/*
 * mape_podem_node_priority
 *
 *
 * determine the priority of the node in the circuit. 
 * according to the paper, it is :
 *
 *  priority = LEVEL_WEIGHT*level + FANOUT_WEIGHT*fanout_number ;
 */
float 
mape_podem_node_priority(node)
node_t *node ;
{
	float rtv ;

	rtv = user_option->podem_option.level_priority*mape_get_node_level(node) + 
	      user_option->podem_option.fanout_priority*mape_node_loading_cap(node) ;

	/*rtv = LEVEL_WEIGHT*mape_get_node_level(node) + FANOUT_WEIGHT*lsLength(node->fanout) ;*/
	return rtv ;
}


/*
 * mape_all_pi_assigned
 *
 * return ture is all the PI has been assign value
 */
int
mape_all_pi_assigned(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_primary_input(network, gen, node) {
		if( !mape_value_determined(mape_top(node)) )
			return 0 ;
	}
	return 1;
}

/*
 * mape_podem_backtrace
 *
 *
 * backtrace from a node to a primary input. 
 */
int 
mape_podem_backtrace(node, value, pi, pi_v)
node_t *node; 
mape_node_value_t value;
node_t **pi ;
mape_node_value_t *pi_v ;
{
	int i ;
	node_t *p ;
	node_t *next = NULL;
	float priority = -1 ;
	float new_prio = -1 ;
	char text[128] ;

	if(node == NULL)
		return FALSE;

	/*
	fprintf(mapeout, "\tbacktrace: %s\t(%s)\n", node->name, mape_print_node_value(value)) ;
	fflush(mapeout) ;
	*/

	if(backtrace_num > user_option->podem_option.backtrace_limit_a)
		return FALSE ;

	if(node_type(node) == PRIMARY_INPUT){
		*pi = node ;
		*pi_v = value ;
		backtrace_num ++ ;
		return TRUE ;
	}
	
	foreach_fanin(node, i, p){
		if(mape_value_determined(mape_top(p)))
			continue ;
		new_prio = mape_podem_node_priority(p) ;
		if(!next || new_prio > priority){
			next = p ;	
			priority = new_prio ;
		}
	}

	if(mape_node_not_family(node))
		value = mape_value_not(value) ;
	return mape_podem_backtrace(next, value, pi, pi_v) ;
}

/*
 * mape_podem
 *
 * using podem algorithm to justify the node's value. 
 * must under the current stack status. 
 */
int 
mape_podem(node, value)
node_t *node ;
mape_node_value_t value ;
{
	node_t *pi ;
	mape_node_value_t pi_v ;

	mape_podem_backtrace(node, value, &pi, &pi_v);
	return mape_subpodem(node, value, pi, pi_v) ;
}

/*
 * mape_subpodem
 *
 * subpodem is the recursive part of podem. 
 */
int 
mape_subpodem(node, value, pi, pi_v)
node_t *node ; 
mape_node_value_t value ;
node_t *pi ;
mape_node_value_t pi_v ;
{
	node_t *next_pi ;
	mape_node_value_t next_pi_v ;
	node_t *frontier ;
	node_t *fanin ;
	char text[2][128] ;
	mape_node_value_t fanin_value ;

	/*
	fprintf(mapeout, "mape_subpodem %s, %s, %s, %s\n", node->name, 
		mape_node_value_string(value, text[0]), pi->name, mape_node_value_string(pi_v, text[1])) ; 
	fflush(mapeout) ;
	*/

	/*try the first value of the pi*/
	mape_podem_implication(pi, pi_v);	

	if(mape_top(node) == value)
		return TRUE ;

	if(backtrack_limit < user_option->podem_option.backtrack_limit && !mape_conflict(value, mape_top(node))){
		//mape_podem_pack_frontier() ;
		if(mape_podem_pick_frontier(&frontier, &fanin, &fanin_value)){ 
			if(mape_podem_backtrace(fanin, fanin_value, &next_pi, &next_pi_v))
				if(mape_subpodem(node, value, next_pi, next_pi_v))
					return TRUE ;
		}
	}

	/*
	fprintf(mapeout, "2 mape_subpodem %s, %s, %s, %s\n", node->name, 
		mape_node_value_string(value, text[0]), pi->name, mape_node_value_string(mape_value_not(pi_v), text[1])) ; 
	fflush(mapeout) ;
	*/

	/*try first value failed, so try the second pi's value*/
	mape_podem_backtrack(pi, pi_v) ;

	mape_podem_implication(pi, mape_value_not(pi_v));	

	if(mape_top(node) == value)
		return TRUE ;

	if(backtrack_limit < user_option->podem_option.backtrack_limit && !mape_conflict(value, mape_top(node))){
		//mape_podem_pack_frontier() ;
		if(mape_podem_pick_frontier(&frontier, &fanin, &fanin_value)){ 	
			if(mape_podem_backtrace(fanin, fanin_value, &next_pi, &next_pi_v))
				if(mape_subpodem(node, value, next_pi, next_pi_v))
					return TRUE ;
		}
	}

	/*
	fprintf(mapeout, "3 mape_subpodem %s, %s, %s, %s\n", node->name, 
		mape_node_value_string(value, text[0]), pi->name, mape_node_value_string(mape_value_first_not(pi_v), text[1])) ; 
	fflush(mapeout) ;
	*/

	/*The first two values fail, try the third value */
	mape_podem_backtrack(pi, mape_value_not(pi_v)) ;

	mape_podem_implication(pi, mape_value_first_not(pi_v)) ;

	if(mape_top(node) == value)
		return TRUE ;

	if(backtrack_limit < user_option->podem_option.backtrack_limit && !mape_conflict(value, mape_top(node))){
		//mape_podem_pack_frontier() ;

		if(mape_podem_pick_frontier(&frontier, &fanin, &fanin_value)){ 	
			if(mape_podem_backtrace(fanin, fanin_value, &next_pi, &next_pi_v))
				if(mape_subpodem(node, value, next_pi, next_pi_v))
					return TRUE ;
		}
	}


	/*The first 3 values fail, try the fourth value */
	mape_podem_backtrack(pi, mape_value_first_not(pi_v)) ;
	return FALSE ;

	mape_podem_implication(pi, mape_value_second_not(pi_v)) ;

	if(mape_top(node) == value)
		return TRUE ;

	/*
	fprintf(mapeout, "4 mape_subpodem %s, %s, %s, %s\n", node->name, 
		mape_node_value_string(value, text[0]), pi->name, mape_node_value_string(mape_value_second_not(pi_v), text[1])) ; 
	fflush(mapeout) ;
	*/
	if(backtrack_limit < user_option->podem_option.backtrack_limit && !mape_conflict(value, mape_top(node))){
		//mape_podem_pack_frontier() ;

		if(mape_podem_pick_frontier(&frontier, &fanin, &fanin_value)){ 	
			if(mape_podem_backtrace(fanin, fanin_value, &next_pi, &next_pi_v))
				if(mape_subpodem(node, value, next_pi, next_pi_v))
					return TRUE ;
		}
	}
	/* It's impossible to excite the fault, return false */
	/*
	mape_podem_backtrack(pi, mape_value_second_not(pi_v));
	return FALSE ;
	*/
}

/*
 * mape_justify
 *
 * justify the node's value. when success, all the 
 * value should be kept in the node. when fail, all
 * the values should be undo. 
 */
int 
mape_justify(node, value)
node_t *node; 
mape_node_value_t value ;
{
	backtrack_limit = 0 ;
	return mape_podem(node, value);
}

/*
 * mape_node_sort_value
 *
 * return the node's sort value for take the priority. 
 */
int 
mape_node_sort_value(node)
node_t *node ;
{
	/*
	if(!node)
		return 0 ;

	return node_num_fanout(node) ;
	*/

	/*
	 * 
	 *
	 *
	 */
	int node_c = mape_get_node_C1(node) + mape_get_node_C0(node) ;
	
	if(node_c > (70 * mape_ctrl_max())/100)
		return 0 ;
	else 
		return (node_num_fanout(node)*3) + 100 - (node_c*100)/mape_ctrl_max() ;
}

/*
 * mape_node_set_visit. 
 *
 * set the node's visit. 
 */
void 
mape_node_set_visit(node)
node_t *node ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)(node->mape))->visit = TRUE ; 
}

/*
 * mape_node_is_visit
 *
 */
int
mape_node_is_visit(node)
node_t *node ;
{
	if(node->mape == NULL)
		return FALSE ;

	return ((mape_t *)(node->mape))->visit  ;
}

/*
 * take_max_priority_node
 * 
 * take the node out which has the maximum priority
 */
node_t *
take_max_priority_node(list, c_list, p_function)
lsList list ;
lsList c_list ;
int (*p_function)() ;
{
	lsGen gen ;
	node_t *p ;
	int value = 0 ;
	int new_value = 0 ;
	node_t *rtv = NULL ;
	node_t *tmp ;

	if(lsLength(c_list) > 0){ /*there are still node in the c_list*/
		for(;;){
			if(lsLength(c_list) == 0)
				break ;
			lsDelEnd(c_list, (lsGeneric *)(&p)) ;
			if(!mape_node_is_visit(p) && !mape_value_determined(mape_top(p))){
				return p ;
			}
		}
		return take_max_priority_node(list, c_list, p_function) ;
	}else { /*there are no candidate nodes in candidate_list*/
		if(lsLength(list) == 0){ /*no any node left */
			return NULL ;
		}
		lsForeachItem(list, gen, p){
			if(mape_node_is_visit(p) || mape_value_determined(mape_top(p)))
				continue ;

			new_value = (*p_function)(p) ;
			if(new_value > value){
				value = new_value ;
				while(lsLength(c_list))
					lsDelEnd(c_list, (lsGeneric *)&tmp) ;
				rtv = p ;
			}else if(new_value == value)
				lsNewEnd(c_list, (lsGeneric)p, LS_NH) ;
		}
		return rtv ;
	}
}

/*
 * the master loop of the mape package. 
 *
 */
int
mape(network)
network_t *network ;
{
	lsList all_nodes;
	lsList candidate_nodes;
	lsGen gen ;
	node_t *node;
	int i ;
	mape_node_value_t value ;

	/* get all node's controllability */
	mape_ctrl_get_all(network) ;
	/*
	 * The following algorithm is descriped in:  
	 * 
	 * <Maximum Power Estimation for CMOS Circuits
	 * Using Deterministic and Static Approaches>
	 *                  -- IEEE Trans. VLSI March 1998.
	 */

	all_nodes = lsCreate() ;
	candidate_nodes = lsCreate() ;
	mape_timer_reset() ;

	foreach_node(network, gen, node)
		lsNewEnd(all_nodes, (lsGeneric)node, LS_NH) ;

	while(node = take_max_priority_node(all_nodes, candidate_nodes, mape_node_sort_value)){
		backtrace_num = 0 ;
		/*
		fprintf(mapeout, "Try to assign %s, #Fanout=%d, C0=%d, C1=%d\n", 
			node->name, node_num_fanout(node), mape_get_node_C0(node), mape_get_node_C1(node)) ; 
		fflush(mapeout) ;
		*/
		mape_node_set_visit(node) ;
		//dump_network(network) ;

		/* assign rising to the node */
		if(! mape_conflict(MAPE_RISE, value)){
			if(mape_justify(node, MAPE_RISE)){
				if(mape_all_pi_assigned(network))
					break ;
				else 
					continue ;
			}
		}

		/* assign falling to the node */
		if(! mape_conflict(MAPE_FALL, mape_top(node))){
			if(mape_justify(node, MAPE_FALL)){
				if(mape_all_pi_assigned(network))
					break ;
				else 
					continue ;
			}
		}
	}
	mape_timer_report("Entity") ; 
	lsDestroy(all_nodes, NULL) ;
	lsDestroy(candidate_nodes, NULL) ;
	return 1;
}
