/*
 * mape_clockskew
 *
 *
 * for ECE734, get the clock skew scheduling for each latch 
 * in the circuit. This technique is used to speedup the 
 * circuit. 
 */
#include "mape_int.h"

/*copy from retime.h */
#define RETIME_DEFAULT_REG_AREA 1.0     /* Area of a latch */
#define RETIME_DEFAULT_REG_DELAY 0.0    /* Delay through a latch */
#define MAX_LATCH_TO_INITIALIZE 16 /* No of latches acceptable to extract STG */

#define RETIME_NOT_SET -1
#define RETIME_USER_NOT_SET -100000.0
#define RETIME_TEST_NOT_SET -50000.0    /* Half of the USER_NOT_SET value */
#define RETIME_DEF_TOL  0.1     /* Tolerence of the binary search */

typedef struct mape_clockskew_latch_struct {
	lsList  ed_list; /* edge's list */
	re_node *source ;
	int 	matrix_id ;
}mape_clockskew_latch_t ;

int retime_debug = 0 ;
double *lm ; 	/* long path matrix */
double *sm ; 	/* short path matrix */
double *csl ;	/* clockskew list */
double *cslb ;  /* backup clock skew list */
lsList cs_latch_list = NULL; /* the latch's list */

extern  char *  mape_clockskew_sa_create(void) ;
extern  double  mape_clockskew_sa_cost(char *) ;
extern  void    mape_clockskew_sa_destory(char *) ;
extern  void    mape_clockskew_sa_move(char *, char *) ;
extern  int     mape_clockskew_sa_equilibrum(int, int, int) ;
extern  int     mape_clockskew_sa_stop(double, int, double) ;
extern  void    mape_clockskew_sa_accept(int) ;

/*
 * mape_clockskew_set_node_delay
 *
 * set the delay value for the node. 
 */
void 
mape_clockskew_set_node_delay(node, l_delay, s_delay)
node_t *node ;
double l_delay, s_delay ; 
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;
		
	((mape_t *)(node->mape))->cs->l_delay = l_delay ;
	((mape_t *)(node->mape))->cs->s_delay = s_delay ;

	/*
	fprintf(mapeout, "set node delay: %s, %lf, %lf\n", node->name, l_delay, s_delay) ;
	*/
}

/*
 * mape_clockskew_get_node_delay
 *
 * return the delay value for the node. 
 */
void 
mape_clockskew_get_node_delay(node, l_delay, s_delay)
node_t *node ;
double *l_delay, *s_delay ;
{
	if(node->mape == NULL){
		*l_delay = -1.0 ;
		*s_delay = -1.0 ;
		return ;
	}

	if( ((mape_t *)(node->mape))->cs == NULL){
		*l_delay = -1.0 ;
		*s_delay = -1.0 ;
		return ;
	}

	*l_delay = ((mape_t *)(node->mape))->cs->l_delay ;
	*s_delay = ((mape_t *)(node->mape))->cs->s_delay ;
}

/*
 * mape_clockskew_is_original_node
 *
 * is the original node 
 */
int 
mape_clockskew_is_original_node(re_no)
re_node *re_no ;
{
	if(re_no->node->mape == NULL)
		return FALSE ;
	if( ((mape_t *)(re_no->node->mape))->cs == NULL )
		return FALSE;

	return ( ((mape_t *)(re_no->node->mape))->cs->l_delay == 0.0 &&
		 ((mape_t *)(re_no->node->mape))->cs->s_delay == 0.0) ;
}

/*
 * mape_clockskew_is_delay_set
 *
 * is the delay already set for node. 
 */
int 
mape_clockskew_is_delay_set(re_no)
re_node *re_no ;
{
	if(re_no->node->mape == NULL)
		return FALSE ;
	if( ((mape_t *)(re_no->node->mape))->cs == NULL )
		return FALSE;

	return ( ((mape_t *)(re_no->node->mape))->cs->l_delay >= 0.0 &&
		 ((mape_t *)(re_no->node->mape))->cs->s_delay >= 0.0) ;
}

/*
 * mape_clockskew_create_clockskew_t
 *
 * create a clockskew_t structure. 
 */
mape_clockskew_t *
mape_clockskew_create_clockskew_t(void)
{
	mape_clockskew_t * rtv ;

	rtv 		= ALLOC(mape_clockskew_t, 1) ;
	rtv->l_delay	= -1.0 ;
	rtv->s_delay	= -1.0 ;

	return rtv ;
}

/*
 * mape_clockskew_create_latch_t
 *
 * create a mape_clockskew_latch_t structure. 
 */
mape_clockskew_latch_t *
mape_clockskew_create_latch_t(ed, source, id)
re_edge *ed ;
re_node *source ;
int id ;
{
	mape_clockskew_latch_t *rtv ;

	rtv 		= ALLOC(mape_clockskew_latch_t, 1) ;
	rtv->source	= source ;
	rtv->matrix_id	= id ;
	rtv->ed_list	= lsCreate() ;	

	lsNewEnd(rtv->ed_list, (lsGeneric)ed, LS_NH) ;

	return rtv ;
}

/*
 * mape_clockskew_push_le
 *
 * push one re_edge into the latch structure. 
 */
void 
mape_clockskew_push_le(ed)
re_edge *ed ;
{
	lsGen gen, gen1 ;
	lsList el ;
	mape_clockskew_latch_t *lp ;
	static int id = 0 ; 
	re_edge *edp ;

	if(cs_latch_list == NULL)
		cs_latch_list = lsCreate() ;	

	lsForeachItem(cs_latch_list, gen, lp)
		if(ed->source == lp->source){
			lsNewEnd(lp->ed_list, (lsGeneric)ed, LS_NH) ;
			lsFinish(gen) ;
			return ;
		}

	lp = mape_clockskew_create_latch_t(ed, ed->source, ++id) ;
	lsNewEnd(cs_latch_list, (lsGeneric)lp, LS_NH) ;
}

/*
 * mape_clockskew_dump
 *
 * dump all the information. 
 */
void 
mape_clockskew_dump(void)
{
	lsGen gen, gen1 ;
	mape_clockskew_latch_t *latch ;
	re_edge *ed ;
	int i, j , n ;

	fprintf(mapeout, "Dump the latch's list: \n\n") ;
	lsForeachItem(cs_latch_list, gen, latch){
		fprintf(mapeout, "Id: %d\n", latch->matrix_id) ;
		fprintf(mapeout, "\tSource: %s\n", latch->source->node->name) ;
		fprintf(mapeout, "\tSinks: ") ;
		lsForeachItem(latch->ed_list, gen1, ed)
			fprintf(mapeout, "%s ", ed->sink->node->name) ;
		fprintf(mapeout, "\n") ;
	}

	n = lsLength(cs_latch_list) ;
	fprintf(mapeout, "Dump the long path matrix: \n") ; 
	for(i=0;i<n;i++){
		fprintf(mapeout, "\t") ;
		for(j=0;j<n;j++)
			fprintf(mapeout, "%-4.4lf ", lm[i*n+j]) ;
		fprintf(mapeout, "\n") ;
	}
	fprintf(mapeout, "Dump the short path matrix:\n") ; 
	for(i=0;i<n;i++){
		fprintf(mapeout, "\t") ;
		for(j=0;j<n;j++)
			fprintf(mapeout, "%-4.4lf ", sm[i*n+j]) ;
		fprintf(mapeout, "\n") ;
	}
}

/*
 * mape_clockskew_level_init   
 *
 * initialize all the level's value for each node. 
 */
void 
mape_clockskew_delay_init(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node){
		if(node->mape == NULL)
			node->mape = (char *)mape_create_mape_t() ;
		if(((mape_t *)(node->mape))->cs == NULL)
			((mape_t *)(node->mape))->cs = mape_clockskew_create_clockskew_t() ;

		((mape_t *)(node->mape))->cs->l_delay = -1.0 ;
		((mape_t *)(node->mape))->cs->s_delay = -1.0 ;
	}
}


/*
 * mape_clockskew_set_lmatrix_entry
 *
 * set the long path matrix entry's value
 */
void 
mape_clockskew_set_lmatrix_entry(i, j, value)
int i, j ; 
double value ;
{
	int n = lsLength(cs_latch_list) ;

	lm[(i-1)*n + (j-1)] = value ;
}

/*
 * mape_clockskew_set_mmatrix_entry
 *
 * set the long path matrix entry's value
 */
void 
mape_clockskew_set_mmatrix_entry(i, j, value)
int i, j ; 
double value ;
{
	int n = lsLength(cs_latch_list) ;

	sm[(i-1)*n + (j-1)] = value ;
}

/*
 * mape_clockskew_update_node_delay
 *
 * update node's delay value. 
 */
void 
mape_clockskew_update_node_delay(re_no)
re_node *re_no ;
{
	re_edge *ed;
	double l_m, s_m, l_delay, s_delay ; 
	int i = 0 ;
	lsGen gen ;

	fprintf(mapeout, "Recur_set_delay: %s\n", re_no->node->name) ;

	l_m = -1.0;
	s_m = 9999.0 ;
	re_foreach_fanin(re_no, i, ed){
		mape_clockskew_get_node_delay(ed->source->node, &l_delay, &s_delay) ;
		fprintf(mapeout, "\t%s, %lf, %lf\n", ed->source->node->name, l_delay, s_delay) ;
		if(l_delay >= 0.0)
			l_m = MAX(l_m, l_delay) ;
		if(s_delay >= 0.0)
			s_m = MIN(s_m, s_delay) ;
	}

	if(l_m == -1.0 && s_m == 9999.0){
		mape_clockskew_set_node_delay(re_no->node, -1.0, -1.0) ;
		return ;
	}

	if(s_m >= 9998.0)
		s_m = 0.0 ;

	mape_clockskew_set_node_delay(re_no->node, l_m + re_no->final_delay, 
		s_m + re_no->final_delay) ;
}

/*
 * mape_clockskew_recur_set_delay
 *
 * recursively set the delay for each node in the re_graph. 
 */
void 
mape_clockskew_recur_set_delay(re_no)
re_node *re_no ;
{
	int i = 0 ;
	re_edge *ed ;

	mape_clockskew_update_node_delay(re_no) ;
	re_foreach_fanout(re_no, i, ed)
		if(ed->weight == 0 && !mape_clockskew_is_original_node(ed->sink))
			mape_clockskew_recur_set_delay(ed->sink) ;
}


/*
 * mape_clockskew_lp_build_matrix 
 * build matrix for a particular latch
 */
void 
mape_clockskew_latch_build_matrix(network, lp)
network_t *network ;
mape_clockskew_latch_t *lp ;
{
	lsGen gen ;
	re_edge *ed ;
	int i, j ;
	mape_clockskew_latch_t *p ;
	
	/* inintialize all the path's value*/
	mape_clockskew_delay_init(network) ;

	mape_clockskew_set_node_delay(lp->source->node, 0.0, 0.0) ;
	lsForeachItem(lp->ed_list, gen, ed)
		mape_clockskew_recur_set_delay(ed->sink) ;
	mape_clockskew_update_node_delay(lp->source) ;
	
	
	/* set the matrix entries values */
	i = lp->matrix_id ;
	lsForeachItem(cs_latch_list, gen, p){
		j = p->matrix_id ;
		mape_clockskew_set_lmatrix_entry(i, j, 
			((mape_t *)(p->source->node->mape))->cs->l_delay) ;
		mape_clockskew_set_mmatrix_entry(i, j, 
			((mape_t *)(p->source->node->mape))->cs->s_delay) ;
	}
}

/*
 * mape_clockskew_build_matrix
 *
 * build the long and short path matrix ;
 */
void 
mape_clockskew_build_matrix(network, graph)
network_t *network ;
re_graph *graph ;
{
	int latch_num ;
	lsList ln_list ; /*latch node list*/
	lsGen gen ;
	re_edge *re_ed ;
	mape_clockskew_latch_t *lp ;
	int i ;

	
	/* find all the node whose fanout is a latch in the graph*/  
	re_foreach_edge(graph, i, re_ed)
		if(re_ed->weight >= 1)
			/* push latch edge to create the latch's list */ 
			mape_clockskew_push_le(re_ed) ;

	/* allocate the long and short latch matrix */
	latch_num = lsLength(cs_latch_list) ;
	if(!latch_num)
		mape_error("No latch exist!", __FILE__, __LINE__) ;

	lm = ALLOC(double, latch_num * latch_num) ;
	sm = ALLOC(double, latch_num * latch_num) ;

	csl = ALLOC(double, latch_num) ; 
	cslb = ALLOC(double, latch_num) ; 
	memset(csl, 0, sizeof(double)*latch_num) ;
	memset(cslb, 0, sizeof(double)*latch_num) ;

	/* from each latch, label all the node's until another latch */
	lsForeachItem(cs_latch_list, gen, lp){
		fprintf(mapeout, "\n\n\nStart Setting: %s\n", lp->source->node->name) ;
		mape_clockskew_latch_build_matrix(network, lp) ;
	}

	mape_clockskew_dump() ;
}

/*
 * mape_clockskew_sa_create
 *
 * just return the network's address.
 */
char *
mape_clockskew_sa_create(void)
{
        return (char *)global_network ;
}

/*
 * mape_cursa_sa_destroy
 *
 * do nothing.
 */
void
mape_clockskew_sa_destroy(nouse)
char *nouse ;
{
        return ;
}

/*
 * mape_cursa_stop
 *
 * stop the process of the sa.
 */
int
mape_clockskew_sa_stop(temp, total, reject_ratio)
double temp, reject_ratio ;
int total ;
{
        if(temp <= 0.0001)
                return TRUE ;
        else
                return FALSE;
}

/*
 * mape_clockskew_sa_accept
 *
 * callback function for informing the accept of the move.
 */
void
mape_clockskew_sa_accept(ac)
int ac ;
{
	int i, n ;

	n = lsLength(cs_latch_list) ;

	if(!ac){
		for(i=0;i<n;i++)
			csl[i] = cslb[i] ;
	}
}

/*
 * mape_clockskew_node_boundary
 *
 * get the node's boundary for clockskew. 
 */
void 
mape_clockskew_node_boundary(lp, low, high) 
mape_clockskew_latch_t *lp ;
double *low, *high ;
{
	int n, i, j;
	mape_clockskew_latch_t *p ;
	lsGen gen ;

	*high = 9999.0 ;
	*low = -9999.0 ;

	n = lsLength(cs_latch_list) ;
	i = lp->matrix_id ;

	lsForeachItem(cs_latch_list, gen, p){
		j = p->matrix_id ;
		if(j == i)
			continue ;
		if(sm[(i-1)+(j-1)*n] >= 0.0 && sm[(j-1)+(i-1)*n] >= 0.0){
			*low  = MAX(*low, csl[j-1] - sm[(i-1) + (j-1)*n]) ;
			*high = MIN(*high, csl[j-1] + sm[(j-1) + (i-1)*n]) ;
		}
	}

	if(*high == 9999.0)
		*high = 0.0 ;
	if(*low == -9999.0)	
		*low = 0 ;
}

/*
 * mape_clockskew_sa_move
 *
 * make some change to the pi.
 */
void
mape_clockskew_sa_move(from, to)
char *from, *to ;
{
	double low_bound, high_bound ;
	double cs ;
	int i, n ;
	lsGen gen ;
	long r ;
	mape_clockskew_latch_t *lp = NULL ;

	/* get the target latch to set clockskew */
	n = lsLength(cs_latch_list) ;
	r = random() ;
	i = (int)(r%n) ;

	if(i == 0)
		return ;

	lsForeachItem(cs_latch_list, gen, lp){
		if(lp->matrix_id == i+1) {
			lsFinish(gen) ;
			break ;
		}
	}
	
	if(lp == NULL){
		fprintf(mapeout, "i=%d\n", i) ;
		lsForeachItem(cs_latch_list, gen, lp)
			fprintf(mapeout, "%d \n", lp->matrix_id) ;

		mape_error("Selecting latch\n", __FILE__, __LINE__) ;
	}
	/* get the clock skew's boundary for the latch */
	mape_clockskew_node_boundary(lp, &low_bound, &high_bound) ;

	/* set the move to be in the boundary */
	r = random() ;
	r = r%100000 ;  

	cs = low_bound + (high_bound - low_bound)*r/100000 ;
	cslb[lp->matrix_id-1] = csl[lp->matrix_id-1] ;
	csl[lp->matrix_id-1] = cs ;
}

/*
 * mape_clockskew_sa_cost
 *
 * the cost function.
 */
double
mape_clockskew_sa_cost(nouse)
char *nouse ;
{
	double rtv = 0.0 ;
	int i, j ;
	int n ;

	n = lsLength(cs_latch_list) ;

	for(i=0;i<n;i++)
		for(j=0;j<n;j++){
			if(i == j)
				continue ;
			rtv = MAX(rtv, lm[i*n+j]+ csl[j] - csl[i]) ;
		}
	return rtv ;
}


/*
 * mape_clockskew_sa_equilibrum
 *
 * equilibrum stop
 */
int
mape_clockskew_sa_equilibrum(up, total, reject)
int up, total , reject;
{
        if(up >= 100 || total >= 200)
                return TRUE ;
        else
                return FALSE ;
}

/*
 * mape_clockskew_optimze
 *
 * Using SA to optimize the clockskew. 
 */
void 
mape_clockskew_optimize(network)
network_t *network ;
{
	char *sol ;

	mape_sa_set(100, 0.9, mape_clockskew_sa_create, mape_clockskew_sa_destroy,	
		mape_clockskew_sa_cost, mape_clockskew_sa_move, mape_clockskew_sa_equilibrum,
		mape_clockskew_sa_stop, mape_clockskew_sa_accept) ;
	sol = (char *)network ;
	mape_sa_start(&sol) ;
}

/*
 * mape_clockskew
 *
 * main function for clockskew. 
 */
void 
mape_clockskew(network)
network_t **network ;
{
	
	re_graph *graph ;
	int use_mapped = TRUE ;
	int i, n ;

	srandom(util_cpu_time()) ;
	graph = retime_network_to_graph(*network, 1, use_mapped);
	if(graph == NIL(re_graph)){
		(void)fprintf(siserr,"%s", error_string());
		return ;
	}
	retime_dump_graph(mapeout, graph) ;

	/* execute the retiming */
	(void)com_execute(network, "retime") ;	


	graph = retime_network_to_graph(*network, 1, use_mapped);
	if(graph == NIL(re_graph)){
		(void)fprintf(siserr,"%s", error_string());
		return ;
	}
	global_network	= *network ;

	/*build the register matrix*/
	mape_clockskew_build_matrix(*network, graph) ;

	/* optimize the clockskew */
	mape_clockskew_optimize(*network) ;

	/* dump the clock skew for each latch */
	n = lsLength(cs_latch_list) ;
	for(i=0;i<n;i++)
		fprintf(mapeout, "cs[%d]=%.4lf,", i, csl[i]) ; 
	fprintf(mapeout, "\n") ;

	//retime_dump_graph(mapeout, graph) ;
}
