/*
 * mape_ctrl
 *
 * The analysis of the controllability for each node in 
 * the circuit. 
 * 
 * Changbo Long   Nov. 15, 2001 (Have a good day). 
 */
#include "mape_int.h"

#define	PI_C0		10
#define PI_C1		10
#define NODE_C0 	1
#define NODE_C1 	1
#define CTRL_MIN	1
#define CTRL_MAX	9999

int ctrl_max = 0 ;


/*
 * mape_ctrl_max
 *
 * get the maximum control value. 
 */
int 
mape_ctrl_max(void)
{
	return ctrl_max ;
}

/*
 * mape_ctrl_get_C0
 *
 * get the C0 value of the node. 
 */
int 
mape_get_node_C0(node)
node_t *node ;
{
	if(!(node->mape))
		return 0 ;
	return ((mape_t *)(node->mape))->C0 ;
}

/*
 * mape_ctrl_get_C1
 *
 * get the C0 value of the node. 
 */
int 
mape_get_node_C1(node)
node_t *node ;
{
	if(!(node->mape))
		return 0 ;
	return ((mape_t *)(node->mape))->C1 ;
}

/*
 * set the controllability for the node. 
 *
 */
void 
mape_set_ctrl_value(node, C0, C1)
node_t *node ;
int C0 ;
int C1 ;
{
	if(!node)
		mape_error("Unexpected NULL node\n", __FILE__, __LINE__) ;
	if(!(node->mape))
		node->mape = (char *)mape_create_mape_t() ;

	/* set the C0 and C1's value for the node */
	((mape_t*)(node->mape))->C0 = C0 ;
	((mape_t*)(node->mape))->C1 = C1 ;
}

/*
 * mape_ctrl_cal_cube
 *
 * calculate the control value for a cube. 
 */
int 
mape_ctrl_cal_cube(a, s, nin)
array_t *a ;
pcube s ;
int nin ;
{
	int i ;
	int rtv = 0 ;
	
	for(i=0;i<nin;i++){
		switch (GETINPUT(s, i)){
			case ONE: 
				rtv += array_fetch(int, a, i*2 + 1) ;
				break ;
			case ZERO:
				rtv += array_fetch(int, a, i*2) ;
				break ;
			case TWO:
			default: 
				break ;
		}
	}
	return (rtv + NODE_C0) ;
}

/*
 * mape_ctrl_cal_cover
 *
 * calculate the control value under a cover. 
 * input's value are provided in the array. 
 */
int 
mape_ctrl_cal_cover(a, c, nin)
array_t *a ;
pcover c ;
int nin ;
{
	int C0 , C1 ;
	int rtv = CTRL_MAX ;
	pcube ps , ps_last;
	int cv ;
	
	foreach_set(c, ps_last, ps){
		cv = mape_ctrl_cal_cube(a, ps, nin) ;
		rtv = MIN(rtv, cv) ;
	}
	return rtv ;
}

/*
 * mape_ctrl_cal_C0
 *
 * calculate the control 0 of a specified node. 
 */
int 
mape_ctrl_cal_C0(node)
node_t *node ;
{
	array_t *a ;
	node_t *p ;
	int i, j , rtv;
	pcube all_x ;
	pcover comp ;
	
	/* compositing the input control value array */
	a = array_alloc(int, 0) ;
	i = 0 ;
	foreach_fanin(node, i, p){
		array_insert_last(int, a, mape_get_node_C0(p)) ;
		array_insert_last(int, a, mape_get_node_C1(p)) ;
	}

	/* get the sigle cover for output 0 */
	cube.num_vars = node->nin ;

	all_x = new_cube() ;
	for(i=0;i<node->nin;i++)
		PUTINPUT(all_x, i, TWO);

	/* get the complement of the node's 1 cube, that's 0's cube*/
	comp = cb_sharp(all_x, node->F) ;

	rtv = mape_ctrl_cal_cover(a, comp, node->nin) ;

	array_free(a) ;
	free_cube(all_x) ;
	free_cover(comp) ;

	return rtv ;
}

/*
 * mape_ctrl_cal_C0
 *
 * calculate the control 0 of a specified node. 
 */
int 
mape_ctrl_cal_C1(node)
node_t *node ;
{
	array_t *a ;
	node_t *p ;
	int i, j , rtv;
	
	a = array_alloc(int, 0) ;
	foreach_fanin(node, i, p){
		array_insert_last(int, a, mape_get_node_C0(p)) ;
		array_insert_last(int, a, mape_get_node_C1(p)) ;
	}
	rtv = mape_ctrl_cal_cover(a, node->F, node->nin) ;
	array_free(a) ;

	return rtv ;
}

/*
 * mape_ctrl_imp_do
 *
 * for the mape_imp utility. each node will do this function. 
 * if return TRUE, the fanout of this node will be push into
 * the imp_list in the mape_imp utility. Otherwise, will not. 
 */
int 
mape_ctrl_imp_do(node, arg)
node_t *node ;
char *arg ; /*no use here*/
{
	int C0 ;
	int C1 ;

	switch(node_function(node)){
		case NODE_PI:
		case NODE_UNDEFINED:
			return ;
		case NODE_0:
			mape_set_ctrl_value(node, CTRL_MIN, CTRL_MAX) ;
			return ;
		case NODE_1:
			mape_set_ctrl_value(node, CTRL_MAX, CTRL_MIN) ;
			return ;
		case NODE_BUF:
			mape_set_ctrl_value(
				node, 
				mape_get_node_C0(*(node->fanin)) + NODE_C0, 
				mape_get_node_C1(*(node->fanin)) + NODE_C1
			);
			return ;
		case NODE_PO:
			mape_set_ctrl_value(
				node, 
				mape_get_node_C0(*(node->fanin)), 
				mape_get_node_C1(*(node->fanin))
			);
			return ;
		/*need detail calculation */
		case NODE_AND:
		case NODE_OR:
		case NODE_INV:
		case NODE_COMPLEX:
		default:
			break ;
	}

	C0 = mape_ctrl_cal_C0(node) ;
	C1 = mape_ctrl_cal_C1(node) ;

	ctrl_max = MAX(ctrl_max, C0+C1) ;

	mape_set_ctrl_value(node, C0, C1) ;
	return TRUE ;
}

/*
 * mape_cal_ctrl_value 
 *
 * calculate all the ctrl value all the node in the circuit. 
 */
void 
mape_ctrl_get_all(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	mape_imp_reset(mape_ctrl_imp_do, NULL);

	foreach_primary_input(network, gen, node){
		mape_set_ctrl_value(node, PI_C0, PI_C1) ;
		mape_imp_push(node) ;
	}

	/* start implication */
	mape_imp_start() ;
}

/*
 * mape_ctrl_push_pi_fanin
 *
 * push the pi_fanin to the pi_fanin list. 
 */
void 
mape_ctrl_push_pi_fanin(node, pi)
node_t *node, *pi ;
{
	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;
	
	if(!((mape_t *)(node->mape))->pi_fanin)
		((mape_t*)(node->mape))->pi_fanin = lsCreate() ;

	if(!mape_node_in_list(pi, ((mape_t*)(node->mape))->pi_fanin))
		lsNewEnd(((mape_t*)(node->mape))->pi_fanin, (lsGeneric)pi, LS_NH) ;
}

/*
 * mape_ctrl_get_pi_fanin
 *
 * get the pi fanin for each node. 
 */
void 
mape_ctrl_get_pi_fanin(node)
node_t *node ;
{
	int i ;
	node_t *fanin, *pi ;
	lsGen gen ;

	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)node->mape)->pi_fanin )
		return ;

	if(node->type == PRIMARY_INPUT){
		mape_ctrl_push_pi_fanin(node, node) ;
		return ;
	}

	foreach_fanin(node, i, fanin){
		mape_ctrl_get_pi_fanin(fanin) ;

		lsForeachItem( ((mape_t*)(fanin->mape))->pi_fanin, gen, pi )
			mape_ctrl_push_pi_fanin(node, pi) ;	
	}
}

/*
 * mape_ctrl_get_all_pi_fanin
 *
 * get all the nodes fanin pi. 
 */
void 
mape_ctrl_get_all_pi_fanin(network)
network_t *network ;
{
	lsGen gen ;
	node_t *po ;

	foreach_primary_output(network, gen, po)
		mape_ctrl_get_pi_fanin(po) ;
}

/*
 * mape_ctrl_dump_node_pi_fanin
 *
 * dump node's pi_fanin. 
 */
void 
mape_ctrl_dump_node_pi_fanin(node)
node_t *node ;
{
	lsGen gen ;
	node_t *pi ;

	fprintf(mapeout, "node %s's pi_fanin\n\t", node->name) ;

	lsForeachItem( ((mape_t*)(node->mape))->pi_fanin, gen, pi)	
		fprintf(mapeout, "%s ", pi->name) ;
	
	fprintf(mapeout, "\n") ;
}


/*
 * mape_ctrl_dump_pi_fanin
 *
 * dump all the node's pi_fanin list. 
 */
void 
mape_ctrl_dump_pi_fanin(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node)
		mape_ctrl_dump_node_pi_fanin(node) ;
}
