/*
 * mape_cur.c 
 *
 * current estimation. 
 *
 * Changbo Long (Dec, 16, 2001)
 */
#include "mape_int.h"

/*
#define CUR_DEBUG
*/
/*
#define NO_GLITCH
*/

typedef struct map_struct map_t;
struct map_struct {
	lib_gate_t *gate;           /* best gate which matches */
	char *binding;       /* used during matching */
	node_t *match_node;         /* used to re-build the graph */
	int ninputs;                /* number of inputs for this gate */
	node_t **save_binding;      /* bindings for the gate inputs */
	delay_time_t *arrival_info; /* copy of arrival times of inputs. used in fanout opt (array of size ninputs) */
	double map_area;            /* current area cost function */
	delay_time_t map_arrival;   /* current arrival time cost function */
	double load;                /* record load on this pin (from last map) */
	int slack_is_critical;      /* slack at node is 'critical' */
	int inverter_paid_for;
	delay_time_t inverter_arrival;
	int index;
	delay_time_t required;
	st_table *gate_link;        /* used in bin_delay_build_network: bindings gate input -> gate root */
	st_generator *gen;          /* to be used in conjunction with gate_link */
	node_t *node_y;             /* used in bin_power.c to cache information */
	array_t *fanout_tree;
	node_t *fanout_source;
	char *pin_info;   /* only allocated when necessary */
};

#define TIME_BOUNDARY 20
#define VOL_BOUNDARY 0.00001

lsList latch_list = NULL ;
lsList cone_pi = NULL ;
int dump_flag = FALSE ;
int total_switch = 0 ;
int cone_mode = FALSE ;
int cone_pi_num = 0 ;

mape_delay_value_t * mape_create_delay_value_t(mape_node_value_t, double, node_t *, mape_node_value_t) ;
double mape_cur_point_value(lsList, double) ;
int mape_cur_next_trans(lsList, double *, mape_node_value_t *, double *, double *, int *) ;
void mape_cur_near_node_voltage(node_t *, double, double *, double *) ;
int mape_cur_is_node_in_cone(node_t *) ;
void mape_cur_set_cone_mode(int) ;

/*
 * mape_cur_is_cone_mode. 
 *
 * is the cone mode. 
 */
int 
mape_cur_is_cone_mode(void)
{
	return cone_mode ;
}

/*
 * mape_cur_set_cone_mode
 *
 * set the cone mode. 
 */
void 
mape_cur_set_cone_mode(v)
int v ;
{
	cone_mode = v ;
}

 
/*
 * mape_cur_set_dump_flag
 *
 * set the dump flag
 */
void 
mape_cur_set_dump_flag(f)
int f ;
{
	dump_flag = f ;
}

/*
 * mape_cur_create_node_value_t ;
 *
 * create the node's value structure. 
 */
mape_cur_node_value_t *
mape_cur_create_node_value_t(void)
{
	mape_cur_node_value_t *rtv ;

	rtv 			= ALLOC(mape_cur_node_value_t, 1) ; 
	rtv->vol_points 	= lsCreate() ;
	rtv->pi			= NULL ;
	rtv->pi_value		= MAPE_FALL ;
	rtv->delay		= 0.0 ;
	rtv->trans_values	= NULL ;

	return rtv ;
}

/*
 * mape_cur_node_delay
 *
 * get node's approximate delay. 
 */
double
mape_cur_top_node_delay(node)
node_t *node ;
{
	mape_cur_node_value_t *nv ;

	nv = mape_cur_top_node_value(node) ;
	return nv->delay ;
}

/*
 * mape_cur_create_cur_point_t 
 *
 * create a current point. 
 */
mape_cur_point_t *
mape_cur_create_cur_point_t(time, value)
double time, value ;
{
	mape_cur_point_t *rtv ;

	rtv 		= ALLOC(mape_cur_point_t, 1) ;
	rtv->time 	= time ;
	rtv->value	= value ;

	return rtv ;
}

/*
 * mape_cur_create_vol_point_t 
 *
 * create a current point. 
 */
mape_vol_point_t *
mape_cur_create_vol_point_t(value, time, multi)
double time, value ;
int multi ;
{
	mape_vol_point_t *rtv ;

	rtv 		= ALLOC(mape_vol_point_t, 1) ;
	rtv->time 	= time ;
	rtv->value	= value ;
	rtv->multi	= multi ;

	return rtv ;
}

/*
 * mape_cur_push_node_value
 *
 * save the node's value. 
 */
void 
mape_cur_push_node_value(node, cur_node_value)
node_t *node;
mape_cur_node_value_t *cur_node_value ;
{
	lsList cur_values ;
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t*)(node->mape))->cur_values == NULL )
		((mape_t*)(node->mape))->cur_values = lsCreate() ;
	
	cur_values = ((mape_t*)(node->mape))->cur_values ;
	lsNewEnd(cur_values, (lsGeneric)cur_node_value, NULL) ;
}

/*
 * mape_cur_new_node_value
 *
 * create a new node_value in the node's mape structure. 
 */
void 
mape_cur_new_node_value(node)
node_t *node ;
{
	mape_cur_node_value_t *node_value ;

	node_value = mape_cur_create_node_value_t() ;
	mape_cur_push_node_value(node, node_value) ;
}

/*
 * mape_cur_push_trans_value
 *
 * push a transition value into the top of the node_value's list. 
 */
void 
mape_cur_push_trans_value(node, value, delay, pi, pi_value)
node_t *node ;
mape_node_value_t value ;
double delay ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_cur_node_value_t *top ;
	mape_delay_value_t *trans_value ;

	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;

	if(! ((mape_t*)(node->mape))->cur_values )
		mape_cur_new_node_value(node) ;

	/*get the top node value */
	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 

	/*create and push trans_value*/
	/*
	trans_value = mape_create_delay_value_t(value, delay, pi, pi_value) ;
	lsNewEnd(top->trans_values, (lsGeneric)trans_value, LS_NH) ;
	*/
}

/*
 * mape_cur_push_vol_point
 *
 * push a voltage point in the node's points list. 
 */
void 
mape_cur_push_vol_point(node, value, time, multi)
node_t *node ;
double value, time ;
int multi ;
{
	mape_cur_node_value_t *top ;
	mape_vol_point_t *point ;

	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;

	if(! ((mape_t*)(node->mape))->cur_values)
		mape_cur_new_node_value(node) ;

	if(lsLength(((mape_t*)(node->mape))->cur_values ) == 0)
		mape_cur_new_node_value(node) ;

	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 
	point = mape_cur_create_vol_point_t(value, time, multi) ;
	lsNewEnd(top->vol_points, (lsGeneric)point, LS_NH) ;
}

/*
 * mape_cur_pop_vol_points
 *
 * pop the node's top voltage points list. 
 */
lsList
mape_cur_pop_vol_points(node)
node_t *node ;
{
	mape_cur_node_value_t *top ;
	lsList rtv  ;

	if(!node->mape)
		return NULL ;
	if(! ((mape_t *)(node->mape))->cur_values )
		return NULL ;
	
	lsDelEnd( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top) ;
	rtv = top->vol_points ;
	free(top) ;
	if(top->trans_values)
		free(top->trans_values) ;
	return rtv ;
}

/*
 * mape_cur_pop_trans_values. 
 *
 * pop the node's top trans values. 
 */
lsList
mape_cur_pop_trans_values(node)
node_t *node ;
{
	mape_cur_node_value_t *top ;
	lsList rtv  ;

	if(!node->mape)
		return NULL ;
	if(! ((mape_t *)(node->mape))->cur_values )
		return NULL ;
	
	lsDelEnd( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top) ;
	//rtv = top->trans_values ;
	free(top) ;
	return rtv ;
}

/*
 * mape_cur_top_node_value
 *
 * get node's top node value. 
 */
mape_cur_node_value_t *
mape_cur_top_node_value(node)
node_t *node ;
{
	mape_cur_node_value_t *top ;

	if(!node->mape)
		return NULL ;
	if(! ((mape_t*)(node->mape))->cur_values )
		return NULL ;

	/*get the top node value */
	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 
	return top ;
}

/*
 * mape_cur_top_vol_points. 
 *
 * get the node's top voltage point list. 
 */
lsList 
mape_cur_top_vol_points(node)
node_t *node ;
{
	mape_cur_node_value_t *top ;

	if(!node->mape)
		return NULL ;
	if(! ((mape_t*)(node->mape))->cur_values )
		return NULL ;
	if(lsLength(((mape_t *)(node->mape))->cur_values) == 0)
		return NULL ;

	/*get the top node value */
	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 
	return top->vol_points ;
}

/*
 * mape_cur_top_trans_values. 
 *
 * get the node's top transition node value's list. 
 */
lsList 
mape_cur_top_trans_values(node)
node_t *node ;
{
	mape_cur_node_value_t *top ;

	if(!node->mape)
		return NULL ;
	if(! ((mape_t*)(node->mape))->cur_values )
		return NULL ;

	/*get the top node value */
	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 
	/*
	return top->trans_values ;
	*/
	return NULL ;
}

/*
 * mape_cur_create_delay_value_t
 *
 * create and fill in a delay_value_t
 */
mape_delay_value_t *
mape_create_delay_value_t(value, delay, pi, pi_value)
mape_node_value_t value ;
double delay ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_delay_value_t *rtv ;

	rtv 		= ALLOC(mape_delay_value_t, 1) ;
	rtv->value	= value ;
	rtv->delay	= delay ;
	rtv->pi		= pi ;	
	rtv->pi_value	= pi_value ;
	rtv->times 	= 1 ;
	rtv->propogate 	= TRUE ;

	return rtv ;
}

/*
 * mape_cur_curlib_gate
 *
 * get the gate's curlib. 
 */
mape_curlib_gate_t *
mape_cur_curlib_gate(node_t *node)
{
	lib_gate_t *lib_gate ;
	mape_curlib_gate_t *rtv ;

	if(!node)
		return NULL ;

	if(!(node->mape)){
		lib_gate = lib_gate_of(node) ;
		if(!lib_gate)
			return mape_curlib_find_gate(DEFAULT_LIB_GATE) ;
		else 
			return mape_curlib_find_gate(lib_gate_name(lib_gate))  ;
	}

	if( ((mape_t*)(node->mape))->curlib_gate )
		return ((mape_t*)(node->mape))->curlib_gate ;
	else{
		lib_gate = lib_gate_of(node) ;
		if(!lib_gate)
			return mape_curlib_find_gate(DEFAULT_LIB_GATE) ;
		else {
			rtv = mape_curlib_find_gate(lib_gate_name(lib_gate)) ;
			((mape_t*)(node->mape))->curlib_gate = rtv ;
			return rtv ;
		}
	}
}

/*
 * mape_node_is_latch
 *
 * return TRUE if the node is a latch
 */
int 
mape_node_is_latch(node)
node_t *node ;
{
	char text[100] ;
	mape_curlib_gate_t *curlib_gate ;
	
	curlib_gate = mape_cur_curlib_gate(node) ;
	if(curlib_gate == NULL)
		return FALSE ;

	if(!strcmp(curlib_gate->name, "dff"))
		return TRUE;
	else 
		return FALSE ;
}

/*
 * mape_cur_make_latch_list. 
 *
 * make the latch list. 
 */
void 
mape_cur_make_latch_list(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ; 
	lsHandle h ;

	foreach_node(network, gen, node)
		if(mape_node_is_latch(node)){
			if(latch_list == NULL)
				latch_list = lsCreate() ;
			lsNewEnd(latch_list, (lsGeneric)node, &h) ;
		}
}

/*
 * mape_cur_is_seq_circuit
 *
 * is the circuit a sequential circuit.  
 * this function is effective only after the function of
 * mape_cur_make_latch_list. 
 */
int 
mape_cur_is_seq_circuit(void)
{
	return (latch_list != NULL) ;
}

/*
 * mape_cur_node_loading
 *
 * get the loading capacitance of the node. 
 */
double
mape_cur_node_loading(node)
node_t *node ;
{
	double cload = 0.0 ;
	lsGen gen ;
	node_t *fanout ;
	mape_curlib_gate_t *curlib_gate ;

	foreach_fanout(node, gen, fanout){
		curlib_gate = mape_cur_curlib_gate(fanout) ;
		if(curlib_gate)
			cload += curlib_gate->effin ;
	}
	return cload ;
}

/*
 * mape_cur_dump_node
 *
 * dump the node's transistion list 
 */
void 
mape_cur_dump_node(node)
node_t *node  ;
{
	lsList vol_points ;
	lsGen gen ;
	mape_vol_point_t  *vp ;
	mape_cur_node_value_t *nv ;
	int i=0 ;

	vol_points = mape_cur_top_vol_points(node) ;
	nv = mape_cur_top_node_value(node) ;

	//fprintf(mapeout, "%s(%.2lf)", node->name, nv->delay) ;
	fprintf(mapeout, "%s", node->name) ;
	if(vol_points == NULL){
		fprintf(mapeout, "\n") ;
		return ;
	}

	lsForeachItem(vol_points, gen, vp){
		if(i%10 == 0)
			fprintf(mapeout, "\n\t") ;
		else if(i!=0)
			fprintf(mapeout, ",") ;

		fprintf(mapeout, "%.2f(t=%.2f)", vp->value, vp->time) ;
		i++ ;
	}

	fprintf(mapeout, "\n") ;
}

/*
 * mape_cur_dump
 *
 * dump all the information. 
 */
void 
mape_cur_dump(void)
{
	lsGen gen ;
	node_t *node ;

	fprintf(mapeout, "Current package dump node's values information\n") ;
	foreach_node(global_network, gen, node)
		mape_cur_dump_node(node) ;
}


/*
 * mape_cur_simulate_node_after
 *
 * simulate the node just after the time.
 */
int
mape_cur_simulate_node_after(node, time)
node_t *node ;
double time ;
{
	int value ;
	pcube input ;
	int i ;
	mape_node_value_t o ;
	node_t *p ;
	int oS ;

	cube.num_vars = node->nin ;
	define_cube_size (node->nin) ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_cur_node_value_after(p, time) ;
		PUTINPUT(input, i, value) ;
	}

	oS = mape_set_simulate_node(input, node) ;

	free_cube(input) ;
	return oS ;
}

/*
 * mape_cur_simulate_node_before
 *
 * simulate the node just before the time.
 */
int
mape_cur_simulate_node_before(node, time)
node_t *node ;
double time ;
{
	int value ;
	pcube input ;
	int i ;
	node_t *p ;
	int oF ;

	cube.num_vars = node->nin ;
	define_cube_size (node->nin) ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_cur_node_value_before(p, time) ;
		PUTINPUT(input, i, value) ;
	}

	oF = mape_set_simulate_node(input, node) ;
	free_cube(input) ;
	return oF ;
}

/*
 * mape_cur_node_value_after
 *
 *
 * return the node's value just after the time. 
 */
int
mape_cur_node_value_after(node, time)
node_t *node  ; 
double time ;
{
	lsList trans_values ;
	lsGen gen ;
	mape_delay_value_t *p, *dv = NULL ;
	mape_node_value_t value ;
	double dis = 9999, ndis ;

	trans_values = mape_cur_top_trans_values(node) ;

	/* find the nearest transistion value */
	lsForeachItem(trans_values, gen, p){
		ndis = fabs(p->delay - time) ;	
		if(ndis < dis){
			dv = p ;
			dis = ndis ;
		}
	}

	if(dv == NULL)
		mape_error("unexpected NULL value!", __FILE__, __LINE__) ;

	if(dv->delay >= time + TIME_BOUNDARY)
		return mape_value_first(dv->value) ;
	else if(dv->delay <= time - TIME_BOUNDARY)
		return mape_value_second(dv->value) ;
	else 
		return mape_value_second(dv->value) ;
}

/*
 * mape_cur_trans_value_contains
 *
 * is the node contains one value? 
 */
mape_delay_value_t *
mape_cur_trans_value_contains(node, value, delay)
node_t *node ;
mape_node_value_t value ;
double delay ;
{
	lsGen gen ;
	mape_delay_value_t *dv ;
	lsList tl ;

	tl = mape_cur_top_trans_values(node);
	if(!tl)
		return FALSE ;
	lsForeachItem(tl, gen, dv){
		if(dv->value == value && fabs(dv->delay - delay) < TIME_BOUNDARY){
			lsFinish(gen) ;	
			return dv ;
		}
	}
	return NULL ;
}


/*
 * mape_cur_node_value_before
 *
 *
 * return the node's value just before the time. 
 */
int
mape_cur_node_value_before(node, time)
node_t *node  ; 
double time ;
{
	lsList trans_values ;
	lsGen gen ;
	mape_delay_value_t *p, *dv = NULL ;
	mape_node_value_t value ;
	double dis = 9999, ndis ;

	trans_values = mape_cur_top_trans_values(node) ;

	/* find the nearest transistion value */
	lsForeachItem(trans_values, gen, p){
		ndis = fabs(p->delay - time) ;	
		if(ndis < dis){
			dv = p ;
			dis = ndis ;
		}
	}

	if(dv == NULL)
		mape_error("unexpected NULL value!", __FILE__, __LINE__) ;

	if(dv->delay >= time + TIME_BOUNDARY)
		return mape_value_first(dv->value) ;
	else if(dv->delay <= time - TIME_BOUNDARY)
		return mape_value_second(dv->value) ;
	else 
		return mape_value_first(dv->value) ;
}

/*
 * mape_cur_find_near_trans_pair
 *
 * find a pair of transition values which are quite near.  
 */
int 
mape_cur_find_near_trans_pair(trans_values, h1, h2)
lsList trans_values ;
lsHandle *h1, *h2 ;
{
}


/*
 * post process the transistion values. 
 *
 *
 */
void 
mape_cur_post_process(trans_values)
lsList trans_values ;
{
	mape_delay_value_t *dv1, *dv2 ;
	lsHandle h1, h2 ;

	if(lsLength(trans_values) == 2){
		lsFirstItem(trans_values, (lsGeneric *)&dv1, &h1) ; 
		lsLastItem(trans_values, (lsGeneric *)&dv2, &h2) ; 

		if(fabs(dv1->delay - dv2->delay) <= HALF_SLOP_TIME){
			/*
			dv1->propogate = FALSE ;
			dv1->propogate = FALSE ;
			*/
			lsRemoveItem(h1, (lsGeneric *)&dv1) ;
			lsRemoveItem(h2, (lsGeneric *)&dv1) ;
		}
	}

	/*
	while( mape_cur_find_near_trans_pair(trans_values, &h1, &h2) ){
		lsRemove(h2, (lsGeneric *)&dv1) ;
		lsRemove(h1, (lsGeneric *)&dv1) ;
	}
	*/
}


/*
 * mape_cur_simulate_node
 *
 * simulate the node. 
 */
void 
mape_cur_simulate_node(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	int i ;
	lsList otl ; /*output transition list*/
	lsList itl ; /*input transition list*/
	node_t *fanin ;
	lsGen gen ;
	mape_delay_value_t *dv, *find ;
	double cload ;
	mape_curlib_gate_t *curlib_gate ;
	mape_node_value_t value_before, value_after, ov ;
	double odlay ;

	/*
	fprintf(mapeout, "\n\nSimulate node: %s\n", node->name) ;
	*/

	/*primary input's value should have been set*/
	if(node_type(node) == PRIMARY_INPUT)
		return ;

	/* for each input calculate the transition it will cause */ 
	cload = mape_cur_node_loading(node) ;
	curlib_gate = mape_cur_curlib_gate(node) ;

	foreach_fanin(node, i, fanin){
		itl = mape_cur_top_trans_values(fanin);
		lsForeachItem(itl, gen, dv){
			/* not transition value, don't calculate the transition */
			if(dv->value != MAPE_RISE && dv->value != MAPE_FALL && dv->propogate == FALSE)
				continue ;
			/* calculate the transition value */
			value_before = mape_cur_simulate_node_before(node, dv->delay) ;
			value_after = mape_cur_simulate_node_after(node, dv->delay) ;	
			/*
			fprintf(mapeout, "calculate the values: time=%.2lf before=%d, after=%d\n", dv->delay, 
				value_before, value_after) ; 
			*/
			
			if(value_before != value_after){
				ov = mape_value_convert(value_before, value_after) ;
				if(ov == MAPE_RISE)
					odlay = mape_curlib_value(curlib_gate, MAPE_CURLIB_RDLAY, cload) ;
				else 
					odlay = mape_curlib_value(curlib_gate, MAPE_CURLIB_FDLAY, cload) ;

				find = mape_cur_trans_value_contains(node, ov, odlay + dv->delay) ;
				if(!find)
					mape_cur_push_trans_value(node, ov, odlay + dv->delay, pi, pi_value) ;
				else 
					find->times ++ ;
			}
		}
	}

	/* if the output of the transistion still empty, we need to calculate the static value */
	otl = mape_cur_top_trans_values(node);
	if(otl == NULL){
		mape_cur_new_node_value(node) ;	
		otl = mape_cur_top_trans_values(node);
	}else { /*post process */
		mape_cur_post_process(otl) ;
	}

	if(lsLength(otl) == 0){
		value_before = mape_cur_simulate_node_before(node, 0.0) ;
		mape_cur_push_trans_value(node, mape_value_convert(value_before, value_before), 0.0, pi, pi_value) ;
	}

	/* dump the transistion value of this node*/
}

/*
 * mape_cur_node_voltage
 *
 * get node's voltage at time. 
 */
double
mape_cur_node_voltage(node, time)
node_t *node ;
double time ;
{
	lsList l ;

	l = mape_cur_top_vol_points(node) ;
	if(l == NULL)
		return UNKNOWN_VOL ;

	return mape_cur_point_value(l, time) ;
}

/*
 * mape_cur_near_node_voltage
 *
 * return the voltage near one point. 
 */
void
mape_cur_near_node_voltage(node, time, vol1, vol2)
node_t *node ;
double time, *vol1, *vol2 ;
{
	lsList cur_points ;
	lsGen gen ;
	mape_cur_point_t *p1 = NULL ;
	mape_cur_point_t *p2 = NULL ;
	mape_cur_point_t *point ;
	double t ;

	cur_points = mape_cur_top_vol_points(node) ;
	if(cur_points == NULL){
		*vol1 = UNKNOWN_VOL ;
		*vol2 = UNKNOWN_VOL ;
		return ;
	}

	lsForeachItem(cur_points, gen, point){
		if(point->time > time){
			p2 = point ;
			lsFinish(gen) ;
			break ;
		}
		p1 = point ;
	}

	if(p1 == NULL && p2 == NULL){
		*vol1 = UNKNOWN_VOL ;
		*vol2 = UNKNOWN_VOL ;
		return  ;
	}if(p1 == NULL && p2 != NULL){
		*vol1 = p2->value ;
		*vol2 = p2->value ;
		return  ;
	}if(p2 == NULL && p1 != NULL){
		*vol1 = p1->value ;
		*vol2 = p1->value ;
		return  ;
	}

	if(p1->time == time){
		*vol1 = p1->value ;
		*vol2 = p1->value ;
		return ;
	}
	if(p2->time == time){
		*vol1 = p2->value ;
		*vol2 = p2->value ;
		return  ;
	}
	if(p1->value == UNKNOWN_VOL || p2->value == UNKNOWN_VOL){
		if(time > (p1->time+p2->time)/2){
			*vol1 = p2->value ;
			*vol2 = p2->value ;
			return  ;
		}else{ 
			*vol1 = p1->value ;
			*vol2 = p1->value ;
			return  ;
		}
	}

	//fprintf(mapeout, "p1->value=%.2lf(%.2lf), p2->value=%.2lf(%.2lf)\n",
		//p1->value, p1->time, p2->value, p2->time) ;
	t = time - SLOP_TIME/10 ;
	*vol1 = (p1->value) + (p2->value - p1->value) * (t - p1->time) / (p2->time - p1->time) ;
	t = time + SLOP_TIME/10 ;
	*vol2 = (p1->value) + (p2->value - p1->value) * (t - p1->time) / (p2->time - p1->time) ;
}



/*
 * mape_cur_node_logic_value
 *
 * get node's logic value at one time. 
 */
int 
mape_cur_node_logic_value(node, time)
node_t *node ;
double time ;
{
	int value ;
	pcube input ;
	int i ;
	node_t *p ;
	int oF ;
	double vol ;

	cube.num_vars = node->nin ;
	define_cube_size (node->nin) ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		vol = mape_cur_node_voltage(p, time) ;
		//fprintf(mapeout, "time=%lf, voltage=%lf\n", time, vol) ;
		if(vol == UNKNOWN_VOL)
			value = TWO ;
		else 
			value = ((vol <= MAX_VOLTAGE/2)?ZERO:ONE) ;
		PUTINPUT(input, i, value) ;
	}

	oF = mape_set_simulate_node(input, node) ;
	free_cube(input) ;
	return oF ;
}

/*
 * mape_cur_simulation_start_value
 *
 * get the node's starting point value. 
 */
double
mape_cur_simulation_start_value(node)
{
	int lv ;

	lv = mape_cur_node_logic_value(node, 0.0) ;

	if(lv == ONE)
		return MAX_VOLTAGE ;
	else if(lv == TWO)
		return UNKNOWN_VOL ;
	else 
		return 0.0;
}

/*
 * mape_cur_simulation_hv_time
 *
 * get the node's nearest half maximum value time after the start time.
 */
double
mape_cur_simulation_hv_time_s(node, start)
node_t *node ;
double start ;
{
	lsList vol_points ;
	lsGen gen ;
	mape_vol_point_t *prev = NULL ;
	mape_vol_point_t *curr ;


	if(NODE_SIM_CURR(node) == NULL)  
		return 10000.0 ;

	curr = (mape_vol_point_t *)lsFetchHandle(NODE_SIM_HC(node)) ;
	prev = (mape_vol_point_t *)lsFetchHandle(NODE_SIM_HP(node)) ;

	while(TRUE){
		if(prev->time >= start){
			if( (prev->value != UNKNOWN_VOL && curr->value == UNKNOWN_VOL) ||
			    (prev->value == UNKNOWN_VOL && curr->value != UNKNOWN_VOL) ){
			    break ;
			}
			if( (prev->value - MAX_VOLTAGE/2)*(curr->value - MAX_VOLTAGE/2) < 0 ){
				break ;
			}
		}
		NODE_SIM_HP(node) = NODE_SIM_HC(node) ;
		prev = (mape_vol_point_t *)lsFetchHandle(NODE_SIM_HP(node)) ;

		if(lsNext(NODE_SIM_CURR(node), (lsGeneric *)&curr, &NODE_SIM_HC(node)) != LS_OK){
			lsFinish(NODE_SIM_CURR(node)) ;
			NODE_SIM_CURR(node) = NULL ;
			return 10000.0 ;
		}
	}
	return (prev->time + curr->time)/2 ;
}

/*
 * mape_cur_simulation_hv_time
 *
 * get the node's nearest half maximum value time after the start time.
 */
double
mape_cur_simulation_hv_time(node, start)
node_t *node ;
double start ;
{
	lsList vol_points ;
	lsGen gen ;
	mape_vol_point_t *prev = NULL ;
	mape_vol_point_t *curr ;

	vol_points = mape_cur_top_vol_points(node) ;

	if(vol_points == NULL)
		return 10000.0 ;

	lsForeachItem(vol_points, gen, curr){
		if(prev == NULL && curr->time >= start){
			prev = curr ;
			continue ;
		}
		if(prev != NULL){
			if( (prev->value != UNKNOWN_VOL && curr->value == UNKNOWN_VOL) ||
			    (prev->value == UNKNOWN_VOL && curr->value != UNKNOWN_VOL) ){
			    lsFinish(gen) ;
			    //fprintf(mapeout, "prev->time=%.2lf, curr->time=%.2lf\n", 
			    //	prev->time, curr->time) ;
			    return (prev->time + curr->time)/2 ;
			}

			if( (prev->value - MAX_VOLTAGE/2)*(curr->value - MAX_VOLTAGE/2) < 0 ){
				lsFinish(gen) ;
				return (prev->time + curr->time)/2 ;
			}else {
				prev = curr ;
			}
		}
	}

	return 10000.0 ;
}

/*
 * mape_cur_simulation_trans_time
 *
 * from start time any pi's value get half MAX_VOLTAGE. 
 */
int 
mape_cur_simulation_trans_time(node, start, multi)
node_t *node ;
double *start ;
int *multi ;
{
	int i ;
	node_t *fanin, *n_f ;
	double time = 9999.0 ;
	double ntime ;

	foreach_fanin(node, i, fanin){
		ntime = mape_cur_simulation_hv_time_s(fanin, *start) ; // simple version of hv time
		//ntime = mape_cur_simulation_hv_time(fanin, *start) ;
		if(ntime < time){
			time = ntime ;
			*multi = 1 ;
			n_f = fanin ;
		}else if(ntime == time){
			(*multi) ++ ;
		}
	}

	if(time == 9999.0)
		return FALSE;
	else {
		*start = time ;
		return TRUE ;
	}
}

/*
 * mape_cur_simulation_near_logic_value
 *
 * logic value near a time.
 */
void 
mape_cur_simulation_near_logic_value(node, start, v1, v2)
node_t *node ;
double start ;
int *v1, *v2 ;
{
	int value ;
	pcube input, input1, input2 ;
	int i ;
	node_t *p ;
	int oF ;
	double vol, vol1, vol2 ;
	int dif = FALSE ;

	cube.num_vars = node->nin ;
	define_cube_size (node->nin) ;
	input1 = new_cube() ;
	input2 = new_cube() ;

	foreach_fanin(node, i, p){
		mape_cur_near_node_voltage(p, start, &vol1, &vol2) ;
		if(fabs(vol1 - vol2) >= 0.001 && dif == FALSE)
			dif = TRUE ;
		vol = mape_cur_node_voltage(p, start) ;
		if(vol1 == UNKNOWN_VOL)
			value = TWO ;
		else 
			value = ((vol1 <= MAX_VOLTAGE/2)?ZERO:ONE) ;
		PUTINPUT(input1, i, value) ;

		if(vol2 == UNKNOWN_VOL)
			value = TWO ;
		else 
			value = ((vol2 <= MAX_VOLTAGE/2)?ZERO:ONE) ;
		PUTINPUT(input2, i, value) ;
	}

	*v1 = mape_set_simulate_node(input1, node) ;
	if(dif)
		*v2 = mape_set_simulate_node(input2, node) ;
	else 
		*v2 = *v1 ;
		
	free_cube(input1) ;
	free_cube(input2) ;
}


/*
 * mape_cur_simulation_critical_time
 *
 * return the time point that the voltage will change. 
 */
int 
mape_cur_simulation_critical_time_vx(node, ctime, multi, fv)
node_t *node ;
double *ctime ;
double *fv ;
int *multi ;
{
	double start = *ctime ;
	int v1, v2, lv ;
	mape_cur_node_value_t *top;

	*fv = UNKNOWN_VOL ;
	top = mape_cur_top_node_value(node) ;
	while(mape_cur_simulation_trans_time(node, &start, multi)){
		v1 = mape_cur_node_logic_value(node, start - SLOP_TIME/10 ) ;
		v2 = mape_cur_node_logic_value(node, start + SLOP_TIME/10 ) ;
		top->delay = start ;
		if(v1 != v2){
			*ctime = start ;
			if(v1 == ONE){
				*fv = MAX_VOLTAGE ;
			}else if(v1 == ZERO){
				*fv = 0.0 ;
			}else{
				*fv = UNKNOWN_VOL ;
			}
			return TRUE ;
		}
	}

	lv = mape_cur_node_logic_value(node, 99999.0) ;
	if(lv == ONE){
		*fv = MAX_VOLTAGE ;		
	}else if(lv == ZERO){
		*fv = 0.0 ;
	}else {
		*fv = UNKNOWN_VOL ;
	}
	*ctime = -1.0 ;
	return FALSE ;
}

/*
 * mape_cur_simulation_critical_time
 *
 * return the time point that the voltage will change. 
 */
int 
mape_cur_simulation_critical_time(node, ctime, multi)
node_t *node ;
double *ctime ;
int *multi ;
{
	double start = *ctime ;
	int v1, v2 ;
	mape_cur_node_value_t *top;

	top = mape_cur_top_node_value(node) ;
	while(mape_cur_simulation_trans_time(node, &start, multi)){
		/*
		v1 = mape_cur_node_logic_value(node, start - SLOP_TIME/10 ) ;
		v2 = mape_cur_node_logic_value(node, start + SLOP_TIME/10 ) ;
		*/
		mape_cur_simulation_near_logic_value(node, start, &v1, &v2) ; 
		top->delay = start ;
		if(v1 != v2){
			*ctime = start ;
			return TRUE ;
		}
	}
	*ctime = -1.0 ;
	return FALSE ;
}

/*
 * mape_cur_set_change_pi
 *
 * set change pi and its value. 
 */
void 
mape_cur_set_change_pi(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_cur_node_value_t *top ;
	mape_vol_point_t *point ;

	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;

	if(! ((mape_t*)(node->mape))->cur_values)
		mape_cur_new_node_value(node) ;

	if( lsLength(((mape_t *)(node->mape))->cur_values) == 0 ){
		top = mape_cur_create_node_value_t() ;
		lsNewEnd( ((mape_t *)(node->mape))->cur_values, (lsGeneric)top, NULL ) ;
	}
	
	(void)lsLastItem( ((mape_t *)(node->mape))->cur_values, (lsGeneric *)&top, LS_NH ) ; 
	top->pi = pi ;
	top->pi_value = pi_value ;
}

/*
 * mape_cur_node_delay
 *
 * get node's delay. 
 */
double 
mape_cur_node_delay(node, dir)
node_t *node ;
int dir ;
{
	double cload ;
	mape_curlib_gate_t *curlib_gate ;

	cload = mape_cur_node_loading(node) ;
	curlib_gate = mape_cur_curlib_gate(node) ;

	if(dir == RISE_DIRECTION)
		return mape_curlib_value(curlib_gate, MAPE_CURLIB_RDLAY, cload) ;
	else 
		return mape_curlib_value(curlib_gate, MAPE_CURLIB_FDLAY, cload) ;
}
 

/*
 * mape_cur_glitch_value_vx.
 *
 * get the glitch's peak voltage value. 
 */
double 
mape_cur_glitch_value_vx(start, end, gtime, pfv, fv)
double start, end, gtime ;
double pfv, fv ;
{
	if(fv == UNKNOWN_VOL){
		return UNKNOWN_VOL ;
	}
	if(pfv == UNKNOWN_VOL){
		return fv ;
	}

	if(fabs(pfv - fv) <= 0.0001){
		if(fabs(fv - 0.0) <= 0.0001){
			return ( gtime - start) * MAX_VOLTAGE / (end - start) ;
		}else {
			return MAX_VOLTAGE - ( gtime - start) * MAX_VOLTAGE / (end - start) ;
		}
	}else {
		return ( pfv + (gtime-start)*(fv-pfv)/(end-start) ) ;
	}
}

/*
 * mape_cur_glitch_value.
 *
 * get the glitch's peak voltage value. 
 */
double 
mape_cur_glitch_value(start, end, gtime, direction)
double start, end, gtime ;
int direction ; 
{
	if(direction == RISE_DIRECTION){
		return ( gtime - start) * MAX_VOLTAGE / (end - start) ;
	}else if(direction == FALL_DIRECTION){
		return MAX_VOLTAGE - ( gtime - start) * MAX_VOLTAGE / (end - start) ;
	}
}

/*
 * mape_cur_latch_cur_logic
 *
 * return the current logic value for a latch. 
 */
int 
mape_cur_latch_cur_logic(node)
node_t *node ;
{
	lsList vol_points ;
	mape_vol_point_t *vp ;
	lsHandle h ;

	if(!mape_node_is_latch(node))	
		mape_error("unexpected combinational component!", __FILE__, __LINE__) ;

	vol_points = mape_cur_top_vol_points(node) ;
	if(vol_points == NULL)
		return ZERO ;
	if(lsLength(vol_points) == 0)
		return ZERO ;

	lsLastItem(vol_points, (lsGeneric *)&vp, &h) ;
	return (vp->value >= MAX_VOLTAGE?ONE:ZERO) ;
}

/*
 * mape_cur_simulate_latch_v
 *
 * simulate a latch. 
 */
void 
mape_cur_simulate_latch_v(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t *pi_value ;
{
	int cur_value ;
	int n_value ;
	int dir ;
	double etime ;
	double stime ;
	lsList vol_points ;

	cur_value = mape_cur_latch_cur_logic(node) ;
	n_value = mape_cur_node_logic_value(node, (double)(CLOCK_CYCLE / 2))  ;
	/* assign a transistion to the latch after the half_cycle since the original 
	   state could be any */

	if( 1 /*cur_value != n_value*/){
		/* assign a transistion to the latch after the half_cycle */
		stime = (double)(CLOCK_CYCLE/2 + HALF_SLOP_TIME) ;
		dir = (cur_value==ZERO?RISE_DIRECTION:FALL_DIRECTION) ; 
		etime = (double)(stime + 2*mape_cur_node_delay(node, dir)) ;
		mape_cur_push_vol_point(node, (cur_value == ZERO?0.0:MAX_VOLTAGE), stime, 1) ;
		mape_cur_push_vol_point(node, (cur_value == ZERO?MAX_VOLTAGE:0.0), etime, 1) ;
	}else{
		vol_points = mape_cur_top_vol_points(node) ;

		if(vol_points == NULL){
			mape_cur_push_vol_point(node, cur_value, 0.0, 1) ;
			return ;
		}
		if(lsLength(vol_points) == 0){
			mape_cur_push_vol_point(node, cur_value, 0.0, 1) ;
			return ;
		}
	}
}

/*
 * mape_cur_simulate_node_vx
 *
 * simulate the node's voltage with waveform,
 * unknown voltage cases are considered. 
 */
int
mape_cur_simulate_node_vx(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	double ctime = -0.0 ; 	// critical time 
	double pctime = -1.0 ;	// previous critical time
	double etime ;		// desired end time of the transition. 
	double pvalue, gvalue, fv, pfv;
	double target_time ;
	int multi ;
	int pmulti ;
	int dir; 
	int rtv = FALSE ;

	//fprintf(mapeout, "\n\tSimulate node %s\n", node->name) ;

	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;


	/* get the initial value and direction */
	pvalue = mape_cur_simulation_start_value(node) ; 
	mape_cur_set_change_pi(node, pi, pi_value) ;

	pctime = -1.0 ;
	etime  = 0.0 ;

	if(pvalue == MAX_VOLTAGE)
		dir = FALL_DIRECTION ;
	else if(pvalue == UNKNOWN_VOL)
		dir = UNKNOWN_DIRECTION ;
	else 
		dir = RISE_DIRECTION ;
	
	if(pvalue != UNKNOWN_VOL)
		rtv = TRUE ;

	//fprintf(mapeout, "\t\tGet inintial value: %lf, rtv=%d\n", pvalue, rtv) ; 
	target_time = mape_tatpg_get_target_time() ;

	while(TRUE){
		if(!mape_cur_simulation_critical_time_vx(node, &ctime, &multi, &fv)){ 
			if(etime == 0.0){
				if(pvalue != UNKNOWN_VOL)
					mape_cur_push_vol_point(node, pvalue, etime, 1) ;
			}else {
				mape_cur_push_vol_point(node, fv, etime, 1) ;
			}

			//fprintf(mapeout, "\t\t 1 fv=%lf\n", fv) ;
			if(fv != UNKNOWN_VOL)
				rtv = TRUE ;
			break ;
		}
		if(ctime > target_time)
			break ;

		//fprintf(mapeout, "\t\t 2 fv=%lf\n", fv) ;
		if(fv != UNKNOWN_VOL)
			rtv = TRUE ;
		if(pctime == -1.0){ // the first critical time
			pctime = ctime ;
			mape_cur_push_vol_point(node, fv, pctime, multi) ; 
			etime = pctime + 2*mape_cur_node_delay(node, dir) ; // desired end time
			continue ;
		}
		if(ctime - etime >= 0.0){
			mape_cur_push_vol_point(node, fv, etime, 1) ;
			dir = (dir == RISE_DIRECTION?FALL_DIRECTION:RISE_DIRECTION) ; // set new direction. 
			mape_cur_push_vol_point(node, fv, ctime, multi) ;
			pctime = ctime ;
			etime = pctime + 2*mape_cur_node_delay(node, dir) ; // desired end time
		}else{  // find a glitch 
			//gvalue = mape_cur_glitch_value(pctime, etime, ctime, dir) ;
			gvalue = mape_cur_glitch_value_vx(pctime, etime, ctime, pfv, fv) ;
			dir = (dir == RISE_DIRECTION?FALL_DIRECTION:RISE_DIRECTION) ; 
			mape_cur_push_vol_point(node, gvalue, ctime, multi) ;
			if(fabs(pfv - fv) < 0.001)
				etime = ctime + (ctime - pctime) ;
			else 
				etime = ctime + (SLOP_TIME - (ctime - pctime)) ;
			pctime = ctime ;
			/*
			if(dir == RISE_DIRECTION){
				etime = ctime + 2*mape_cur_node_delay(node, dir)*(MAX_VOLTAGE - gvalue)/MAX_VOLTAGE ;
			}else 
				etime = ctime + 2*mape_cur_node_delay(node, dir)*gvalue/MAX_VOLTAGE;
			*/
		}
		pfv=fv ;
	}
	//mape_cur_dump_node(node) ;

	return rtv ;
}

/*
 * mape_cur_reset_node_sim
 *
 * reset node's simulation structure. 
 */
void 
mape_cur_reset_node_sim(node) 
node_t *node ;
{
	lsList vol_points ;
	mape_cur_point_t *p ;

	vol_points = mape_cur_top_vol_points(node) ;

	if(vol_points == NULL){
		NODE_SIM_CURR(node) = NULL ;
		return ;
	}
	if(lsLength(vol_points) < 2){
		NODE_SIM_CURR(node) = NULL ;
		return ;
	}
	if(NODE_SIM_CURR(node))
		lsFinish(NODE_SIM_CURR(node)) ;

	NODE_SIM_CURR(node) = lsStart(vol_points) ;

	lsNext(NODE_SIM_CURR(node), (lsGeneric *)&p, &NODE_SIM_HP(node)) ;
	lsNext(NODE_SIM_CURR(node), (lsGeneric *)&p, &NODE_SIM_HC(node)) ;
}

/*
 * mape_cur_simulate_node_v
 *
 * simulate the node's voltage with waveform. 
 */
void
mape_cur_simulate_node_v(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	double ctime = -0.0 ; 	// critical time 
	double pctime = -1.0 ;	// previous critical time
	double etime ;		// desired end time of the transition. 
	double pvalue, gvalue;
	int multi ;
	int pmulti ;
	int dir; 
	node_t *fanin ;
	int i ;

	//fprintf(mapeout, "\n\tSimulate node %s\n", node->name) ;

	if(node_type(node) == PRIMARY_INPUT){
		//mape_cur_dump_node(node) ;
		return  ;
	}

	/* get the initial value and direction */
	pvalue = mape_cur_simulation_start_value(node) ; 
	//fprintf(mapeout, "\t\tGet start value:%lf\n", pvalue) ;
	mape_cur_set_change_pi(node, pi, pi_value) ;

	foreach_fanin(node, i, fanin)
		mape_cur_reset_node_sim(fanin) ;

	if(mape_node_is_latch(node)){
		mape_cur_simulate_latch_v(node, pi, pi_value) ;
		return ;
	}

	pctime = -1.0 ;
	etime  = 0.0 ;

	if(pvalue == MAX_VOLTAGE)
		dir = FALL_DIRECTION ;
	else 
		dir = RISE_DIRECTION ;

	while(TRUE){
		if(!mape_cur_simulation_critical_time(node, &ctime, &multi)){ //get transition start point. 
			if(etime == 0.0)
				mape_cur_push_vol_point(node, pvalue, etime, 1) ;
			else
				mape_cur_push_vol_point(node, (dir==RISE_DIRECTION?MAX_VOLTAGE:0.0), etime, 1) ;
			break ;
		}
		if(pctime == -1.0){ // the first critical time
			pctime = ctime ;
			mape_cur_push_vol_point(node, (dir==RISE_DIRECTION?0.0:MAX_VOLTAGE), pctime, multi) ; 
			etime = pctime + 2*mape_cur_node_delay(node, dir) ; // desired end time
			continue ;
		}
		if(ctime > etime){
			mape_cur_push_vol_point(node, (dir==RISE_DIRECTION?MAX_VOLTAGE:0.0), etime, 1) ;
			dir = (dir == RISE_DIRECTION?FALL_DIRECTION:RISE_DIRECTION) ; // set new direction. 
			mape_cur_push_vol_point(node, (dir==RISE_DIRECTION?0.0:MAX_VOLTAGE), ctime, multi) ;
			pctime = ctime ;
			etime = pctime + 2*mape_cur_node_delay(node, dir) ; // desired end time
		}else{  // find a glitch 
			gvalue = mape_cur_glitch_value(pctime, etime, ctime, dir) ;
			dir = (dir == RISE_DIRECTION?FALL_DIRECTION:RISE_DIRECTION) ; // set new direction. 
			mape_cur_push_vol_point(node, gvalue, ctime, multi) ;
			pctime = ctime ;
			if(dir == RISE_DIRECTION){
				etime = ctime + 2*mape_cur_node_delay(node, dir)*(MAX_VOLTAGE - gvalue)/MAX_VOLTAGE ;
			}else 
				etime = ctime + 2*mape_cur_node_delay(node, dir)*gvalue/MAX_VOLTAGE;
		}
	}
	//mape_cur_dump_node(node) ;
}

/*
 * mape_cur_imp_do
 *
 *
 *
 */
int
mape_cur_imp_do(node, arg)
node_t *node ;
char *arg ; /*no use*/
{
	/*simulate the node and store the value into top delay_value list*/
	//mape_cur_simulate_node(node, node, MAPE_0_0) ;  /* the pi and pi's value is not used here.*/

	if(node_type(node) == PRIMARY_INPUT){
		mape_cur_simulate_node_v(node, node, MAPE_0_0) ; /* the pi and pi's value is not used here.*/
		return TRUE ;
	}

	if(cone_mode && !mape_cur_is_node_in_cone(node))
		return FALSE ;

	//fprintf(mapeout, "Simulate node: %s\n", node->name) ;
	mape_cur_simulate_node_v(node, node, MAPE_0_0) ; /* the pi and pi's value is not used here.*/
	return TRUE ;
}

/*
 * mape_cur_test
 *
 * test the current library. 
 */
void 
mape_cur_initialize_random(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;
	int i = 0 ;
	long r ;
	mape_node_value_t n_value ;
	static int first_time = TRUE ;
	unsigned seed ;
	int v ;
	long tmp ;

	if(first_time){
		tmp = util_cpu_time() ;
		seed = (unsigned int)(tmp%256) ;
		srandom(seed) ;
		first_time = FALSE ;
	}

	/*calculate the loading capacitance for each node*/
	mape_imp_reset(mape_cur_imp_do, (char *)NULL) ;
	foreach_primary_input(network, gen, pi){
		r = random() ;
		v = (int)r%100 ;
		if(v >= 50){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}
		/*
		mape_cur_push_trans_value(pi, n_value, 0.0, pi, n_value) ;
		*/
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
}

/*
 * mape_cur_test
 *
 * test the current library. 
 */
void 
mape_cur_initialize(network, is_fall)
network_t *network ;
int is_fall ;
{
	lsGen gen ;
	node_t *pi, *latch ;
	int i = 0 ;

	/*calculate the loading capacitance for each node*/
	mape_imp_reset(mape_cur_imp_do, (char *)NULL) ;
	if(is_fall){
		foreach_primary_input(network, gen, pi){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
			mape_imp_push(pi) ;
		}
		/*
		if(mape_cur_is_seq_circuit()){
			lsForeachItem(latch_list, gen, latch){
				mape_cur_push_vol_point(latch, (double)MAX_VOLTAGE, (double)0.0, 1) ;
				mape_cur_push_vol_point(latch, (double)0.0, (double)SLOP_TIME, 1) ;
				mape_cur_set_change_pi(latch, latch, MAPE_FALL) ;
				mape_imp_push(latch) ;
			}
		}
		*/
	}else{ 
		foreach_primary_input(network, gen, pi){
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
			mape_imp_push(pi) ;
		}
		/*
		if(mape_cur_is_seq_circuit()){
			lsForeachItem(latch_list, gen, latch){
				mape_cur_push_vol_point(latch, (double)0.0, (double)0.0, 1) ;
				mape_cur_push_vol_point(latch, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
				mape_cur_set_change_pi(latch, latch, MAPE_FALL) ;
				mape_imp_push(latch) ;
			}
		}
		*/
	}
	mape_imp_start() ;
}

/*
 * mape_cur_triangle_value
 *
 * get the time point's current value in the current triangle. 
 */
double 
mape_cur_triangle_value(time, start, height, pos, dis)
double time, start, height, pos, dis ;
{
	if(time >= start + dis || time <= start)
		return 0.0 ;
	if(time >= start + pos)
		return (height * (start + dis - time) / (dis - pos) ) ;
	else
		return (height * (time - start) / pos ) ;
}

/*
 * mape_cur_point_value. 
 *
 * get a time point's current value in the current describing way of points list. 
 */
double
mape_cur_point_value(cur_points, time)
lsList cur_points  ; 
double time ;
{
	lsGen gen ;
	mape_cur_point_t *p1 = NULL ;
	mape_cur_point_t *p2 = NULL ;
	mape_cur_point_t *point ;

	lsForeachItem(cur_points, gen, point){
		if(point->time > time){
			p2 = point ;
			lsFinish(gen) ;
			break ;
		}
		p1 = point ;
	}

	if(p1 == NULL && p2 == NULL)
		return UNKNOWN_VOL ;
	if(p1 == NULL && p2 != NULL)
		return p2->value ;
	if(p2 == NULL && p1 != NULL)
		return p1->value ;

	if(p1->time == time)
		return p1->value ;
	if(p2->time == time)
		return p2->value ;
	if(p1->value == UNKNOWN_VOL || p2->value == UNKNOWN_VOL){
		if(time > (p1->time+p2->time)/2)
			return p2->value ;
		else 
			return p1->value ;
	}

	//fprintf(mapeout, "p1->value=%.2lf(%.2lf), p2->value=%.2lf(%.2lf)\n",
		//p1->value, p1->time, p2->value, p2->time) ;
	return (p1->value) + (p2->value - p1->value) * (time - p1->time) / (p2->time - p1->time) ;
}

/*
 * mape_cur_next_add_time. 
 *
 *
 * reassign the next add time. 
 */
void 
mape_cur_next_add_time(add_time, start, pos, dis)
double *add_time ;
double start, pos, dis ;
{
	if(*add_time == start){
		*add_time = start + pos ;
		return ;
	}
	if(*add_time == start+pos){
		*add_time = start + dis ;
		return ;
	}
	if(*add_time == start+dis){
		*add_time = -1.0 ;
		return ;
	}
	*add_time = -1.0 ;
}
 

/*
 * mape_cur_add_triangle_current 
 *
 *
 * add a triangle current to the points. 
 */
void 
mape_cur_add_triangle_current(start, height, pos, dis, cur_points)
double start, height, pos, dis ; 
lsList cur_points ;
{
	lsGen gen ;
	mape_cur_point_t *p = NULL, *tmp ;
	lsHandle h ;
	mape_cur_point_t *new_point ; 
	double add_time = start ;
	double add_value ;
	int value_added ;
	int i ;
	lsList back_list ;
	int list_over = FALSE ;
	int data_used = TRUE ;
	double cur_point_value ;

#ifdef CUR_DEBUG
	fprintf(mapeout, "Add triangle current %.2lf, %.2lf, %.2lf, %.2lf\n", start, height, pos, dis) ;
#endif
	if(dump_flag)
		fprintf(mapeout, "Add triangle current %.2lf, %.2lf, %.2lf, %.2lf\n", start, height, pos, dis) ;

	/* transfter point from cur_points to back_list */
	back_list = lsCreate() ;
	while(lsLength(cur_points)){
		lsDelBegin(cur_points, (lsGeneric *)&p) ;
		LS_ASSERT(lsNewEnd(back_list, (lsGeneric)p, LS_NH)) ;
	}

	/* reconstruct cur_points list */
	if(lsLength(back_list) == 0)
		list_over = TRUE ;
	gen = lsStart(back_list) ;
	for(;;){
		/* next point in the list. */
		if(!list_over && data_used){
			if(lsNext(gen, (lsGeneric *)&p, &h) != LS_OK)
				list_over = TRUE ;
			data_used = FALSE ;
		}
		if(list_over && add_time == -1.0)
			break ;

		if(list_over && add_time != -1.0){
			cur_point_value = mape_cur_point_value(back_list, add_time) ;
			add_value = mape_cur_triangle_value(add_time, start, height, pos, dis) +
				    ((cur_point_value == UNKNOWN_VOL)?0.0:cur_point_value) ;
			new_point = mape_cur_create_cur_point_t(add_time, add_value) ;
			lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ;
			mape_cur_next_add_time(&add_time, start, pos, dis) ;
			continue ;
		}
		if(add_time == -1.0 && !list_over){
			add_value = mape_cur_triangle_value(p->time, start, height, pos, dis) + p->value ;
			new_point = mape_cur_create_cur_point_t(p->time, add_value) ;
			lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ;
			data_used = TRUE ;
			continue ;
		}
		if(p->time < add_time){
			add_value = mape_cur_triangle_value(p->time, start, height, pos, dis) + p->value ;
			new_point = mape_cur_create_cur_point_t(p->time, add_value) ;
			lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ;
			data_used = TRUE ;
			continue ;
		}
		if(p->time == add_time){
			add_value = mape_cur_triangle_value(p->time, start, height, pos, dis) + p->value ;
			new_point = mape_cur_create_cur_point_t(p->time, add_value) ;
			lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ;
			data_used = TRUE ;
			mape_cur_next_add_time(&add_time, start, pos, dis) ;
			continue ;
		}
		if(p->time > add_time){
			cur_point_value = mape_cur_point_value(back_list, add_time) ;
			add_value = mape_cur_triangle_value(add_time, start, height, pos, dis) +
				    ((cur_point_value == UNKNOWN_VOL)?0.0:cur_point_value) ;
			new_point = mape_cur_create_cur_point_t(add_time, add_value) ;
			lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ;
			mape_cur_next_add_time(&add_time, start, pos, dis) ;
			continue ;
		}
	}
	lsFinish(gen) ;
	lsDestroy(back_list, free);

	/*dump the current's point in the curlist*/
#ifdef CUR_DEBUG
	i = 0 ;
	fprintf(mapeout, "dump current's point\n\t") ;
	lsForeachItem(cur_points, gen, p){
		i ++ ;
		if(i % 5 == 0)
			fprintf(mapeout, "\n\t") ;
		if(i == 1)
			fprintf(mapeout, "(%.2lf,%.2lf)", p->time, p->value) ;
		else 
			fprintf(mapeout, ",(%.2lf,%.2lf)", p->time, p->value) ;
	}
	fprintf(mapeout, "\n") ;
#endif
	if(dump_flag){
		i = 0 ;
		fprintf(mapeout, "dump current's point\n\t") ;
		lsForeachItem(cur_points, gen, p){
			i ++ ;
			if(i % 5 == 0)
				fprintf(mapeout, "\n\t") ;
			if(i == 1)
				fprintf(mapeout, "(%.2lf,%.2lf)", p->time, p->value) ;
			else 
				fprintf(mapeout, ",(%.2lf,%.2lf)", p->time, p->value) ;
		}
		fprintf(mapeout, "\n") ;
	}
}

/*
 * mape_cur_next_trans
 *
 *
 * get the next transition value from the start time. 
 */
int
mape_cur_next_trans(vol_points, start, type, delay, times, multi)
lsList vol_points ;
mape_node_value_t *type ;
double *start, *delay, *times ;
int *multi ;
{
	lsGen gen ;
	mape_vol_point_t *prev = NULL ;
	mape_vol_point_t *curr ;

	if(vol_points == NULL)
		return FALSE ;

	lsForeachItem(vol_points, gen, curr){
		if(prev == NULL && curr->time >= *start){
			prev = curr ;
			continue ;
		}
		if(prev != NULL){
#ifdef NO_GLITCH
			if((fabs(prev->value - MAX_VOLTAGE) <= VOL_BOUNDARY && fabs(curr->value - 0.0) <= VOL_BOUNDARY) || 
			    (fabs(prev->value - 0.0) <= VOL_BOUNDARY && fabs(curr->value - MAX_VOLTAGE) <= VOL_BOUNDARY )){
#else
			if(prev->value != curr->value){
#endif
				*start = curr->time - 0.01 ;
				*delay = prev->time ;
				*type  = (curr->value > prev->value)?MAPE_RISE:MAPE_FALL ;
				*times = fabs((curr->value - prev->value)/MAX_VOLTAGE) ; 
				*multi = prev->multi ; 
				lsFinish(gen) ;
				return TRUE ;
			}else {
				prev = curr ;
			}
		}
	}

	return FALSE ;
}

/*
 * mape_cur_add_node_current
 *
 * add the node's current up . 
 */
void 
mape_cur_add_node_current_v(node, cur_points) 
node_t *node ;
lsList cur_points ;
{
	lsList vol_points ;
	lsGen gen ;
	double cload ;
	mape_curlib_gate_t *curlib_gate ;
	double pos, height, dis ;
	char *one_fanout = NULL ;
	node_t *fanout ;
	double start = -0.0 ;
	mape_node_value_t type ;
	double delay, times ;
	int multi ;

	vol_points = mape_cur_top_vol_points(node) ;
	if(vol_points == NULL) 
		return;

	/* only one fanout, get the fanout curlib type*/
	if(node_num_fanout(node) == 1){
		foreach_fanout(node, gen, fanout){
			curlib_gate = mape_cur_curlib_gate(fanout) ;
			if(curlib_gate != NULL)
				one_fanout = curlib_gate->name ;
		}
	}

	curlib_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;

#ifdef CUR_DEBUG
	fprintf(mapeout, "Add node %s's current\n", node->name) ;
#endif 
	if(dump_flag)
		fprintf(mapeout, "Add node %s's current\n", node->name) ;

	while(mape_cur_next_trans(vol_points, &start, &type, &delay, &times, &multi)){
		if(type == MAPE_FALL){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFDIS, one_fanout) ;
			}else{ 
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFDIS, cload) ;
			}
		}else if(type == MAPE_RISE){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRDIS, one_fanout) ;
			}else{
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRDIS, cload) ;
			}
		}else 
			continue ;

		/* revise the positive curve*/
		/*
		if(height >= 0){
			if(delay >= 45)
				delay -= 40 ;
			pos = 10 ; 
			dis /= 2 ;
		}
		if(times < 0.99 && times >= 0.70) 
			times = 1.0 ;
		*/
		/*
		if(height >= 0)
			continue ;
		*/
		/*
		if(multi > 1)
			mape_cur_add_triangle_current(delay, height*times*multi*0.8, pos, dis, cur_points) ; 
		else 
			mape_cur_add_triangle_current(delay, height*times, pos, dis, cur_points) ; 
		*/
		if(height >= 0)
			continue ;

		mape_cur_add_triangle_current(delay, height, pos, dis, cur_points) ; 
	}
}

/*
 * mape_cur_add_node_current
 *
 * add the node's current up . 
 */
void 
mape_cur_add_node_current(node, cur_points) 
node_t *node ;
lsList cur_points ;
{
	lsList trans_values ;
	mape_delay_value_t *dv ;
	lsGen gen ;
	double cload ;
	mape_curlib_gate_t *curlib_gate ;
	double pos, height, dis ;
	char *one_fanout = NULL ;
	node_t *fanout ;

	trans_values = mape_cur_top_trans_values(node) ;
	if(trans_values == NULL) 
		return;

	/* only one fanout, get the fanout curlib type*/
	if(node_num_fanout(node) == 1){
		foreach_fanout(node, gen, fanout){
			curlib_gate = mape_cur_curlib_gate(fanout) ;
			if(curlib_gate != NULL)
				one_fanout = curlib_gate->name ;
		}
	}

	curlib_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;

#ifdef CUR_DEBUG
	fprintf(mapeout, "Add node %s's current\n", node->name) ;
#endif 
	if(dump_flag)
		fprintf(mapeout, "Add node %s's current\n", node->name) ;

	lsForeachItem(trans_values, gen, dv){
		if(dv->value == MAPE_FALL){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFDIS, one_fanout) ;
			}else{ 
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFDIS, cload) ;
			}
		}else if(dv->value == MAPE_RISE){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRDIS, one_fanout) ;
			}else{
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRDIS, cload) ;
			}
		}else 
			continue ;

		/* ignore the overshort*/
		if(height >= 0){
			/*continue ;*/
			pos = 5 ; 
		}

		//mape_cur_add_triangle_current(0.0, height, pos, dis, cur_points) ; 
		//mape_cur_add_triangle_current(dv->delay, height, pos, dis, cur_points) ; 
		mape_cur_add_triangle_current(MAX(dv->delay - HALF_SLOP_TIME, 0), height*(dv->times), pos, dis, cur_points) ; 
	}
}

/*
 * mape_cur_accumulate_current
 *
 * accumulate the current into the cur_list.  
 */
void
mape_cur_accumulate_current(network, cur_point)
network_t *network ;
lsList cur_point ;
{
	node_t *node ;
	lsGen gen ;

	foreach_node(network, gen, node) {
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT
		   || ! mape_is_given_node(node))
			continue ;
		mape_cur_add_node_current_v(node, cur_point) ;
	}
}

/*
 * mape_cur_cal_current
 *
 * calculate the total current created by the transistion 
 * in the circuit. 
 */
double
mape_cur_max_current(network, time, switch_list)
network_t *network ;
double *time ;
lsList switch_list ;
{
	/* record the transition point in the current's waveform */
	lsList cur_points ;	
	node_t *node ;
	lsGen gen ;
	double rtv = 0.0 ;
	mape_cur_point_t *point ;
	lsList vol_points ;
	double point_value, point_value_b, point_value_a;
	mape_curlib_gate_t *curlib_gate ;
	int f = FALSE ;

#ifdef CUR_DEBUG
	mape_cur_dump() ;	
#endif
	if(dump_flag)
		mape_cur_dump() ;

	cur_points = lsCreate() ;
	foreach_node(network, gen, node) {
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT
		   || ! mape_is_given_node(node))
			continue ;
		//mape_cur_add_node_current(node, cur_points) ;	
		mape_cur_add_node_current_v(node, cur_points) ;
	}

	lsForeachItem(cur_points, gen, point){
#ifdef CUR_DEBUG
		fprintf(mapeout, "current point : time=%.2lf, value=%.2lf\n", point->time, point->value) ;
#endif
	if(dump_flag)
		fprintf(mapeout, "current point : time=%.2lf, value=%.2lf\n", point->time, point->value) ;
		
		/*
		if(point->time > CLOCK_CYCLE/2) 
			continue ;
		*/
		if(fabs(point->value) > fabs(rtv)){
			rtv = point->value ;
			*time = point->time ;
		}
	}
	//mape_cur_dump() ;

	if(switch_list){

		foreach_node(network, gen, node) {
			if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT
			   || ! mape_is_given_node(node))
				continue ;

			vol_points = mape_cur_top_vol_points(node) ;
			if(vol_points == NULL) 
				continue;

			while(lsLength(cur_points)){
				lsDelBegin(cur_points, (lsGeneric *)&point) ;
				free(point) ;
			}
			mape_cur_add_node_current_v(node, cur_points) ;
			point_value = mape_cur_point_value(cur_points, *time) ;
			if(point_value != UNKNOWN_VOL && point_value < 0)
					lsNewEnd(switch_list, (lsGeneric)node, LS_NH) ;
		}
	}

	lsDestroy(cur_points, free) ;
	return rtv ;
}

/*
 * Using test filter to change the 
 * circuit for special circuit test. 
 *
 */
void 
mape_cur_test_filter(network)
network_t *network ;
{
	lsGen gen ;
	lib_gate_t *not ;
	node_t *p ;

	/*Test filter for invertor_chain */
	not = lib_get_gate(lib_get_library(), "not") ;
	foreach_node(network, gen, p){
		if(node_type(p) == PRIMARY_INPUT || node_type(p) == PRIMARY_OUTPUT)
			continue ;
		p->map = (char *)ALLOC(map_t, 1) ;
		((map_t *)(p->map))->gate = not ;
	}
}

/*
 * mape_cur_clear
 *
 * clear the simulation result last time. 
 */
void 
mape_cur_clear(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;

	foreach_node(network, gen, node)
		mape_curitr_node_clear(node) ;
}

/*
 * maximum current estimation. 
 *
 */
void
mape_current(network)
network_t *network ;
{
	double max_cur ;
	double time ;
	int i=0 ;

	/*build the current lib*/
	mape_curlib_build_lib() ;
	//mape_curlib_set_unit_mode(TRUE) ;

	/*test filter to change the network*/
	/*
	mape_cur_test_filter(network) ;
	*/

	/*make latch list*/
	/*
	mape_cur_make_latch_list(network) ;
	*/

	mape_timer_reset() ;
	mape_timer_start() ;

	/*test the current library*/
	/*
	for(i=0;i<200;i++){
		mape_cur_initialize(network, FALSE) ;
		mape_cur_clear(network) ;
	}
	*/
	mape_timer_stop() ;
	mape_cur_initialize(network, FALSE) ;
	mape_timer_report("1 times of inintialize\n") ;

	//mape_cur_initialize_random(network) ;

	/*
	mape_cur_dump();
	*/
	max_cur = mape_cur_max_current(network, &time, NULL) ;
	mape_cur_clear(network) ;
	return ;

	fprintf(mapeout, "Maximum rise current estimation: %.2lf\n", max_cur) ;

	mape_cur_initialize(network, TRUE) ;
	/*
	mape_cur_dump() ;
	*/
	mape_timer_reset() ;
	mape_timer_start() ;
	for(i=0;i<=200;i++)
		max_cur = mape_cur_max_current(network, &time, NULL) ;
	mape_timer_stop() ;
	mape_timer_report("200 times of maximum current estimation\n") ;

	fprintf(mapeout, "Maximum fall current estimation: %.2lf\n", max_cur) ;
	mape_cur_clear(network) ;
}

/*
 * mape_cur_is_node_in_cone
 *
 * is the node in the cone list. 
 */
int
mape_cur_is_node_in_cone(node)
node_t *node ;
{
	return ((mape_t *)(node->mape))->C1 ;
}


/*
 * mape_cur_node_cone
 *
 * make the cone's list 
 */
void 
mape_cur_node_cone(node)
node_t *node ;
{
	node_t *fanin ;
	int i ;

	if( ((mape_t *)(node->mape)) -> C1 == 1 )
		return ;

	((mape_t *)(node->mape))->C1 =  1 ;

	foreach_fanin(node, i, fanin)
		mape_cur_node_cone(fanin) ;
}

/*
 * mape_cur_cone_pi_num
 *
 * return the pi number for a cone
 */
int
mape_cur_cone_pi_num(void)
{
	if(cone_pi)
		return lsLength(cone_pi) ;
	else 
		return 0 ;
}

/*
 * mape_cur_cone_pi
 *
 * return the cone_pi ;
 */
lsList
mape_cur_cone_pi(void)
{
	return cone_pi ;
}

/*
 * mape_cur_group_cone
 *
 *
 * make the gate cone list for a group of circuit, which 
 * includes all the necessary gates for determine the value
 * of the gates in the group. 
 */
void 
mape_cur_group_cone(group)
lsList group;
{
	node_t *node ;
	lsGen gen ;
	int count ;

	foreach_node(global_network, gen, node){
		if(!((mape_t *)(node->mape)))
			(mape_t *)(node->mape) = mape_create_mape_t() ;

		((mape_t *)(node->mape))->C1 = 0 ;
	}

	lsForeachItem(group, gen, node){
		mape_cur_node_cone(node) ;
	}

	if(cone_pi){
		while(lsLength(cone_pi))
			lsDelEnd(cone_pi, (lsGeneric *)&node) ;
	}
	if(!cone_pi)
		cone_pi = lsCreate() ;

	foreach_primary_input(global_network, gen, node)
		if( (mape_cur_is_node_in_cone(node)) )
			lsNewEnd(cone_pi, (lsGeneric)node, LS_NH) ;
}
