/*
 * mape_curga
 *
 * maximum current estimation under GA . 
 * Partially copied from Fei Li's source code. . 
 * 
 * Changbo Long. 
 */
#include "mape_int.h"
#include "ga_alg.h"

/*
#define PBSSCE
*/

#define GLOBAL_LOCAL

#define PI_VEC_LENGTH	10	/* used in the curloc */
#define PI_POS_BIT	0	/* the number of the bit for specifying the switching position */
#define PPI_TRANS_NUM   10	/* transition number in the pseudo primary input */

/*
#define PI_POS_UNIT	HALF_SLOP_TIME
#define PI_POS_UNIT	SLOP_TIME
*/

array_t *	mape_curga_init_pop_vector(void) ;
array_t *	mape_curga_select(array_t *) ;
void 		mape_curga_crossover(array_t *) ;
void 		mape_curga_mutation(array_t *) ;
void 		mape_curga_replace(array_t **, array_t *) ;
double		mape_curga_fitness(chromosome_t *) ;
double		mape_curploc_ga_fitness(chromosome_t *) ;
double		mape_curloc_ga_fitness(chromosome_t *) ;
BOOLEAN		mape_curga_finish(int, int, int) ;
void 		mape_curga_parse_code_unit(chromosome_t *, int, int *, int *) ;
double		mape_curploc_code_switch_pos(node_t *, int, int) ;
int 		mape_curga_is_special_group(node_t *) ;

lsList 	pseudo_pi_list = NULL ;


/* assuming binary encoding */
void
decode_dump_chromo_vector(chromosome_t *chr, FILE * fp)
{
	int i,j;
	int word_index, word_offset;

	fprintf(fp, "\t") ;
	for(i=0;i<get_gene_count();i++) {
		word_index = get_word_index(i);
		word_offset = get_word_offset(i);
		if( (chr->data[word_index] & (1<<word_offset)) == 0) {
			fprintf(fp, "0") ;
		}else{
			fprintf(fp,"1") ;
		}
	}
	fprintf(fp, "\n") ;
}

/* decode & dump a popualation */
void
decode_dump_pop_vector(array_t *pop, FILE* fp)
{
	int i;
	chromosome_t *chr;

	fprintf(fp, "Dumping the population.\n");
	for(i=0;i<array_n(pop);i++) {
		chr = array_fetch(chromosome_t *, pop, i);
		fprintf(fp,"Dumping chromosome No. %d\n", i);
		decode_dump_chromo_vector(chr,fp);
	}
	fflush(fp);
}

/*
 * remove_visited_chromo
 *
 * remove the chromosome that has been visited. 
 */
void 
mape_curga_remove_visited_chromo(pop_left, chromo1, chromo2)
array_t **pop_left ;
int chromo1, chromo2;
{
	array_t *tmp_array;
	int i,k,n;
	chromosome_t *chr;

	n = array_n(*pop_left);
	tmp_array = array_alloc(chromosome_t *, n-2);

	for(k=0,i=0;k<array_n(*pop_left);k++) {
		if(k != chromo1 && k != chromo2){
			chr = array_fetch(chromosome_t*, *pop_left, k);
			array_insert(chromosome_t*, tmp_array, i, chr);
			i++;
		}
	}
	array_free(*pop_left);
	(*pop_left) = tmp_array;
}

/*
 * mape_curga_tourna
 *
 * (1)tournament selection without replacement.
 * (2)a new array with new chromo instances is returned.
 */
array_t *
mape_curga_tourna(pop)
array_t *pop ;
{
	int n,i,j;
	array_t * pop_left;
	array_t * better;
	chromosome_t *p1, *p2;
	long r1, r2 ;

	better = array_alloc(chromosome_t *, 0);
	pop_left = array_dup(pop); /* pointer addresses are copied */

	while(array_n(pop_left) > 0) {
		n = array_n(pop_left);
		do{
			r1 = random() ;
			r2 = random() ;
			i = (int)r1%n ;
			j = (int)r2%n ;
		}while(i == j) ;
		p1 = array_fetch(chromosome_t*, pop_left, i);
		p2 = array_fetch(chromosome_t*, pop_left, j);

		if(p1->fitness > p2->fitness)
			chromo_dup_and_ins_last(better, p1);
		else 
			chromo_dup_and_ins_last(better, p2);
		p1->visited = p2->visited = TRUE;	
		mape_curga_remove_visited_chromo(&pop_left,i,j);
	}

	return better;
}


/*
 * mape_curga_select
 *
 * selct, copied from Fei Li's code. 
 */
array_t *
mape_curga_select(pop)
array_t *pop ;
{
	array_t * better1, *better2;

	//fprintf(stdout, "Get into selection:\n");
	better1 = mape_curga_tourna(pop);
	better2 = mape_curga_tourna(pop);

	array_append(better1, better2);
	array_free(better2) ;

	//fprintf(stdout, "Population after selection:\n");
	//decode_dump_pop_vector(better1, stdout);
	return better1 ;
}

/* partial_swap()
 * swap the part from index '0' to 'offset_in_word'
 */
void
partial_swap(word1, word2, offset_in_word)
int *word1; 
int *word2; 
int offset_in_word;
{
	int i;
	int tmp1,tmp2;

	for(i=offset_in_word;i>=0;i--) {
		tmp1 = (*word1) & (1<<i);
		tmp2 = (*word2) & (1<<i);
		(*word1) = ((*word1) & (~(1<<i))) | tmp2;
		(*word2) = ((*word2) & (~(1<<i))) | tmp1;
	}
}

/*
 * mape_curga_one_point_crossover
 *
 * one point crossover for two chromosome. 
 */
void
mape_curga_one_point_crossover(p1, p2)
chromosome_t *p1; 
chromosome_t *p2;
{
	int gene_position;
	int word_index, offset_in_word;
	int tmp,i;
	long r ;

	/* randomly get a position */
	r = random() ;
	gene_position = r % get_gene_count();
	word_index = get_word_index(gene_position);
	offset_in_word = get_word_offset(gene_position);

	partial_swap(&(p1->data[word_index]), &(p2->data[word_index]), offset_in_word);
	for(i=word_index-1;i>=0;i--) {
		tmp = p1->data[i];
		p1->data[i] = p2->data[i];
		p2->data[i] = tmp;
	}
}

/*
 * mape_curga_crossover
 *
 * selct, copied from Fei Li's code. 
 */
void
mape_curga_crossover(pop)
array_t *pop ;
{
	array_t * pop_left;
	int n;
	int i,j;
	chromosome_t *p1, *p2;
	long r1, r2 ;


	clear_visited_flag(pop);
	pop_left = array_dup(pop); /* pointer addresses are copied */

	while(array_n(pop_left) > 0){
		n = array_n(pop_left);

		/* two different random numbers from 0 to n-1 */
		do{
			r1 = random() ;
			r2 = random() ;
			i = (int)r1%n ;
			j = (int)r2%n ;
		}while(i == j) ;
		p1 = array_fetch(chromosome_t*, pop_left, i);
		p2 = array_fetch(chromosome_t*, pop_left, j);

		mape_curga_one_point_crossover(p1, p2);
		p1->visited = p2->visited = TRUE;

		mape_curga_remove_visited_chromo(&pop_left,i,j);
	}

	/* debugging */
	//fprintf(stdout,"Population after crossover:\n");
	//decode_dump_pop_vector(pop, stdout);
}

/*
 * mape_curga_mutation
 *
 * selct, copied from Fei Li's code. 
 */
void
mape_curga_mutation(pop)
array_t *pop ;
{
	double prob;
	double randout;
	chromosome_t *chr;
	int i,j;
	int word_index, offset_in_word;
	long r ;
	int cur_gen, max_gen ;

	/* mutation probability */
	prob = 10000 / get_gene_count();
	cur_gen = get_cur_generation() ;
	max_gen = get_max_generation() ;

	//prob = prob * (1 + (get_gene_count()*(max_gen - cur_gen))/(2 * max_gen)) ;

	/* for each chromosome */
	for(i=0;i<array_n(pop);i++) {
		chr = array_fetch(chromosome_t*, pop, i);

		for(j=0;j<get_gene_count();j++) {
			r = random() ;
			randout = (double)(r%10000) ;
			if( randout < prob ){
				/* flip the bit */
				word_index = get_word_index(j);
				offset_in_word = get_word_offset(j);
				chr->data[word_index] = chr->data[word_index] ^ (1<<offset_in_word);
			}
		}

	}

	/* debugging */
	//fprintf(stdout,"Population after mutation:\n");
	//decode_dump_pop_vector(pop, stdout);
}

/*
 * mape_curga_replace
 *
 * replace, copied from Fei Li's code. 
 */
void
mape_curga_replace(old_pop, new_pop)
array_t **old_pop ;
array_t *new_pop ;
{
	int i;
	chromosome_t *chr;

	for(i=0;i<array_n(*old_pop);i++) {
		chr = array_fetch(chromosome_t*, *old_pop, i);
		FREE(chr->data);
		FREE(chr);
	}
	array_free(*old_pop);
	(*old_pop) = new_pop;
}

/*
 * mape_curga_finish
 *
 * finish, copied from Fei Li's code. 
 */
BOOLEAN
mape_curga_finish(gen_num, max_gen, no_improve_num)
int gen_num, max_gen, no_improve_num ;
{
	if(gen_num >= max_gen)
		return TRUE ;
	else 
		return FALSE ;
	/*
	if(gen_num > max_gen) 
		return TRUE;
	else 
		return FALSE;
	*/
}

/*
 * mape_curloc_sim_imp_do
 *
 * simulation implication do
 */
int
mape_curloc_sim_imp_do(node, arg)
node_t *node ;
char *arg ; /* no use */
{
	int is_given, is_ppi;


	//mape_cur_dump() ;

	is_given = mape_is_given_node(node) ;
	is_ppi = mape_node_in_list(node, pseudo_pi_list) ;

	if(is_ppi){
		return TRUE ;
	}else if(is_given){
		//fprintf(mapeout, "Simulate node %s\n", node->name) ;
		mape_cur_simulate_node_v(node, node, MAPE_0_0) ;
		return TRUE ;
	}else 
		return FALSE ;
}

/*
 * mape_curploc_code_switch_pos
 *
 * translate the code in GA into the real switching pos
 */
double
mape_curploc_code_switch_pos(node, pos, total)
node_t *node ;
int pos, total ;
{
	lsList list ;
	int count ;
	long r ;
	lsGen gen ;
	int i, unit, offset; 
	mape_pbssce_st_t *st ;

	if(node->mape == NULL)
		return -1.0 ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		return -1.0 ;

	list = ((mape_t *)(node->mape))->plt->st_list ;

	if(list == NULL)
		return -1.0 ;
	
	if(lsLength(list) == 0)
		return -1.0 ;

	if(lsLength(list) == 1){
		lsFirstItem(list, (lsGeneric *)&st, LS_NH) ;
		return st->time ;
	}


	unit = lsLength(list)/total ;
	r = random() ;
	if(unit <= 1)
		count = r%(lsLength(list)) ;
	else {
		offset = r%unit ;
		count = unit*pos + offset  ;
		count = count%lsLength(list) + 1 ;
	}

	i = 0 ;
	lsForeachItem(list, gen, st){
		i ++ ;	
		if(i == count){
			lsFinish(gen) ;
			if(st -> rp + st -> fp >= 0.0000001)
				return st->time ;
			else 
				return -1.0 ;
		}
	}
	fprintf(mapeout, "Can't find the count %d, length=%d\n", count, lsLength(list)) ;
	return -1.0 ;
}

/*
 * mape_curploc_assign_ppi
 *
 * Assign transistions to the PPI.  
 */
void 
mape_curploc_assign_ppi(ppi, pos, switch_type)
node_t *ppi ;
double *pos ;
int switch_type ;
{
	int i, ns, nns, j ;
	double np = 0.0, pnp = -10.0 ;
	double ratio ;
	int count = 0 ;
	int total ; /*total possible switch*/


	total = lsLength(((mape_t*)(ppi->mape))->plt->st_list) ;
	ns = switch_type ;
	//fprintf(mapeout, "%s\n", ppi->name ) ;
	for(;;){
		np = 9999.0 ;
		for(i=0;i<PPI_TRANS_NUM;i++){
			if( ((float)(count))*1.0/total >= 0.49 )
				break ;
			if(pos[i] == np && pos[i] > 0.0){
				pos[i] = -1.0 ;
			}else if(pos[i] < np && pos[i] > 0.0){
				np = pos[i] ;
				j = i ;
			}
		}
		if(np == 9999.0){
			if(ns == 1){
				mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(0.0 + pnp - HALF_SLOP_TIME), 1) ;
			}else{
				mape_cur_push_vol_point(ppi, (double)0.0, (double)(0.0 + pnp - HALF_SLOP_TIME), 1) ;
			}
			//fprintf(mapeout, "Assign last time :%.2lf\n", pnp) ;
			return ;
		}
		count ++ ;

		if(pnp <= -1.0){ //first time
			if(ns == 1){
				mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}else{
				mape_cur_push_vol_point(ppi, (double)0.0, (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}
			//fprintf(mapeout, "Assign first time :%.2lf\n", np) ;
			pnp = np + SLOP_TIME ;
		}else if(np > pnp){
			if(ns == 1){
				mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(0.0 + pnp-HALF_SLOP_TIME), 1) ;
				mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}else{
				mape_cur_push_vol_point(ppi, (double)0.0, (double)(0.0 + pnp-HALF_SLOP_TIME), 1) ;
				mape_cur_push_vol_point(ppi, (double)0.0, (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}
			//fprintf(mapeout, "Assign time :%.2lf and %.2lf\n", pnp, np) ;
			pnp = np + SLOP_TIME ;
		}else if(np <= pnp){
			ratio = 1.0 - (pnp - np)/SLOP_TIME ;
			if(ratio < 0.5){
				pos[j] = -1.0 ;
				continue ;
			}

			if(ns == 1){
				mape_cur_push_vol_point(ppi, (double)(MAX_VOLTAGE*ratio), (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}else{
				mape_cur_push_vol_point(ppi, (double)(MAX_VOLTAGE*(1-ratio)), (double)(0.0 + np-HALF_SLOP_TIME), 1) ;
			}
			//fprintf(mapeout, "Assign time :%.2lf \n", np) ;
			pnp = np + SLOP_TIME * ratio ;
		}
		ns = !ns ;
		pos[j] = -1.0 ;
	}
}

/*
 * mape_curploc_ga_fitness
 *
 * get the fitness for a chromosome.
 */
double
mape_curploc_ga_fitness(chr)
chromosome_t *chr ;
{
	int i,j, in, io;
	int word_index, word_offset, count, pos, switch_type, pi_pos_unit, pswitch;
	lsGen gen ;
	node_t *ppi = NULL ;
	lsHandle h ;
	double time ;
	double rtv ;
	double real_pos[PPI_TRANS_NUM];
	int pos_total = 1 ;

	for(i=0;i<user_option->curga_option.pi_bit;i++)
		pos_total = 2 * pos_total ;

	count = get_gene_count() ;
	gen = lsStart(pseudo_pi_list) ;

	mape_imp_reset(mape_curloc_sim_imp_do, (char *)NULL) ;
	//pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	j=0 ;
	for(i=0;i<count;i+=(1+user_option->curga_option.pi_bit)) {
		mape_curga_parse_code_unit(chr, i, &switch_type, &pos) ;
		if(j == 0){
			if(lsNext(gen, (lsGeneric *)&ppi, &h) != LS_OK)
				mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;
			j == 1 ;
		}

		if(j > PPI_TRANS_NUM){
			mape_curploc_assign_ppi(ppi, real_pos, switch_type) ;
			mape_imp_push(ppi) ;
			if(lsNext(gen, (lsGeneric *)&ppi, &h) != LS_OK)
				mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;
			j = 1 ;
		}
		real_pos[j-1] = mape_curploc_code_switch_pos(ppi, pos, pos_total) ;
		if(real_pos[j-1] < 0.0){
			fprintf(mapeout, "unused pos %.2lf\n", real_pos[j-1]) ;
		}
		j++ ;
	}
	/* the last one */
	mape_curploc_assign_ppi(ppi, real_pos, switch_type) ;
	mape_imp_push(ppi) ;

	lsFinish(gen) ;
	mape_imp_start() ;

	rtv=mape_cur_max_current(global_network, &time, NULL) ;
	mape_cur_clear(global_network) ;

	return -rtv ;
}

/*
 * mape_curloc_ga_fitness
 *
 * get the fitness for a chromosome.
 */
double
mape_curloc_ga_fitness(chr)
chromosome_t *chr ;
{
	int i,j, in, io;
	int word_index, word_offset, count, pos, switch_type, pi_pos_unit;
	lsGen gen ;
	node_t *ppi = NULL ;
	lsHandle h ;
	double time ;
	double rtv ;

	count = get_gene_count() ;
	gen = lsStart(pseudo_pi_list) ;

	mape_imp_reset(mape_curloc_sim_imp_do, (char *)NULL) ;
	pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	for(i=0;i<count;i+=(1+user_option->curga_option.pi_bit)) {
		mape_curga_parse_code_unit(chr, i, &switch_type, &pos) ;
		if(lsNext(gen, (lsGeneric *)&ppi, &h) != LS_OK)
			mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;

		if(switch_type){
			mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(ppi, (double)0.0, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(ppi, ppi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(ppi, (double)0.0, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(ppi, ppi, MAPE_RISE) ;
		}
		mape_imp_push(ppi) ;

		/*
		in = i/(PI_VEC_LENGTH) ; 
		io = i%(PI_VEC_LENGTH) ;
		word_index = get_word_index(i);
		word_offset = get_word_offset(i);

		if(io == 0){
			if(lsNext(gen, (lsGeneric *)&ppi, &h) != LS_OK)
				mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;
		}

		if( (chr->data[word_index] & (1<<word_offset)) == 0) {
			mape_cur_push_vol_point(ppi, (double)0.0, (double)(SLOP_TIME*io), 1) ;
			mape_cur_set_change_pi(ppi, ppi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(ppi, (double)MAX_VOLTAGE, (double)(SLOP_TIME*io), 1) ;
			mape_cur_set_change_pi(ppi, ppi, MAPE_RISE) ;
		}
		*/
	}
	lsFinish(gen) ;
	mape_imp_start() ;


	rtv=mape_cur_max_current(global_network, &time, NULL) ;

	mape_cur_clear(global_network) ;
	return -rtv ;
}

/*
 * mape_curga_parse_code_unit
 *
 * parse a code unit to get the switch_type 
 * and switching position. 
 */
void 
mape_curga_parse_code_unit(chr, start, switch_type, pos)
chromosome_t *chr ;
int start ;
int *switch_type, *pos ;
{
	int i ;
	int word_index, word_offset, value;

	*switch_type = 0 ;
	*pos = 0 ;
	for(i=start;i<=(start+user_option->curga_option.pi_bit);i++){
		word_index = get_word_index(i);
		word_offset = get_word_offset(i);
		if( (chr->data[word_index] & (1<<word_offset)) == 0) 
			value = 0 ;
		else 
			value = 1 ;
		if(value == 1 && i == start){
			*switch_type = 1 ;
			continue ;
		}
		if(i > start)
			*pos = (*pos)*2 + value ;
	}
}

/*
 * mape_curga_fitness
 *
 * get the fitness for a chromosome.
 */
double
mape_curga_fitness_detail(chr, time, l)
chromosome_t *chr ;
double *time ;
lsList l ;
{
	int i,j;
	int word_index, word_offset, count, switch_type, pos;
	lsGen gen ;
	node_t *pi ;
	lsHandle h ;
	double rtv ;
	int pi_pos_unit ;
	lsList cpi ;


	count = get_gene_count() ;
	cpi = mape_cur_cone_pi() ;
	gen = lsStart(cpi) ;

	mape_imp_reset(mape_cur_imp_do, (char *)NULL) ;
	pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	for(i=0;i<count;i+=(1+user_option->curga_option.pi_bit)) {
		/*
		word_index = get_word_index(i);
		word_offset = get_word_offset(i);
		*/
		mape_curga_parse_code_unit(chr, i, &switch_type, &pos) ;
		//fprintf(mapeout, "Get switch_type=%d and pos=%d\n", switch_type, pos) ;
		if(lsNext(gen, (lsGeneric *)&pi, &h) != LS_OK)
			mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;

		if(switch_type){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}

		/*
		if( (chr->data[word_index] & (1<<word_offset)) == 0) {
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;			
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}
		*/
		if(mape_cur_is_node_in_cone(pi))
			mape_imp_push(pi) ;
	}
	lsFinish(gen) ;
	mape_imp_start() ;

	rtv=mape_cur_max_current(global_network, time, l) ;
	mape_cur_clear(global_network) ;

	return -rtv ;
}

/*
 * mape_curga_stn_update
 *
 *
 * update the stn values. 
 */
void 
mape_curga_stn_update(total)
double total ;
{
	mape_stn_curga_update(total) ;
}

/*
 * mape_curga_fitness
 *
 * get the fitness for a chromosome.
 */
double
mape_curga_fitness(chr)
chromosome_t *chr ;
{
	int i,j;
	int word_index, word_offset, count, switch_type, pos;
	lsGen gen ;
	node_t *pi ;
	lsHandle h ;
	double time ;
	double rtv ;
	int pi_pos_unit ;
	static double max_value = -10.0 ;
	lsList cpi ;

	count = get_gene_count() ;
	if(mape_cur_is_cone_mode()){
		cpi = mape_cur_cone_pi() ;
		gen = lsStart(cpi) ;
	}else 
		gen = lsStart(global_network->pi) ;

	mape_imp_reset(mape_cur_imp_do, (char *)NULL) ;
	pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	for(i=0;i<count;i+=(1+user_option->curga_option.pi_bit)) {
		/*
		word_index = get_word_index(i);
		word_offset = get_word_offset(i);
		*/
		mape_curga_parse_code_unit(chr, i, &switch_type, &pos) ;
		//fprintf(mapeout, "Get switch_type=%d and pos=%d\n", switch_type, pos) ;
		if(lsNext(gen, (lsGeneric *)&pi, &h) != LS_OK)
			mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;

		if(switch_type){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}

		/*
		if( (chr->data[word_index] & (1<<word_offset)) == 0) {
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;			
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}
		*/
		mape_imp_push(pi) ;
	}
	lsFinish(gen) ;
	mape_imp_start() ;

	rtv=mape_cur_max_current(global_network, &time, NULL) ;

	//mape_curga_stn_update(-rtv) ;
	mape_cur_clear(global_network) ;

	//fprintf(mapeout, "Return fitness: %.2lf\n", -rtv) ;
	return -rtv ;
}

/*
 * mape_curga_init_pop_vector
 *
 * initialize the population vector. 
 */
array_t *
mape_curga_init_pop_vector(void)
{
	int i,j;
	array_t * pop;
	chromosome_t *chromo_p;
	long r ;
	int value ;

	/* seed */
	pop = array_alloc(chromosome_t *, get_pop_size());

	for(i=0;i<get_pop_size();i++){
		chromo_p = new_chromosome();
		for(j=0;j<get_gene_count();j++) {
			/* randomly generating input value */
			if(i == 0)
				value = 1 ;
			else if(i == 1)
				value = 0 ;
			else 
				r = random() ;
				value = (int)(r%2) ;
			set_chromo_bit(chromo_p, j, value) ;
		}
		array_insert(chromosome_t *, pop, i, chromo_p);
	}

	/* debugging */
	//fprintf(stdout,"Initial population:\n");
	//decode_dump_pop_vector(pop, stdout);
	return pop;
}

/*
 * mape_curloc_push_pseudo_pi
 *
 * push a pseudo pi into pseudo_pi_list ;
 */
void 
mape_curloc_push_pseudo_pi(node)
node_t *node ;
{
	if(pseudo_pi_list == NULL)
		pseudo_pi_list = lsCreate() ;
	
	//fprintf(mapeout, "Push pseudo pi: %s\n", node->name) ;
	lsNewEnd(pseudo_pi_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_curloc_pi_imp
 *
 * implication to make the pseudo_pi_list. 
 */
int 
mape_curloc_pi_imp_do(node, arg)
node_t *node ;
char *arg ; /* no use */
{
	int fout_given = FALSE ;
	lsGen gen ;
	node_t *p ;

	if(mape_is_given_node(node)){
		if(node_type(node) == PRIMARY_INPUT)
			mape_curloc_push_pseudo_pi(node) ;
		return TRUE ;
	}

	foreach_fanout(node, gen, p)
		if(!fout_given && mape_is_given_node(p))
			fout_given = TRUE ;

	if(fout_given)
		mape_curloc_push_pseudo_pi(node) ;

	return TRUE ;
}


/*
 * mape_curloc_make_pseudo_pi
 *
 * make the pseudo pi's list. 
 */
void 
mape_curloc_make_pseudo_pi(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_curloc_pi_imp_do, NULL) ;
	foreach_primary_input(network, gen, pi)
		mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_curloc_ga
 *
 * ga algorithm for maximum current extimation 
 * with localization mode. 
 */
void 
mape_curloc_ga(network)
network_t *network ;
{
	int population_size;	
	int max_gen;
	int chromo_word_size;
	int bits_per_pi;
	int num_control_pi;
	chromosome_t * solution;

	mape_timer_reset() ;
	mape_timer_start() ;
	mape_curlib_set_unit_mode(TRUE) ;

	mape_clear_node_list(pseudo_pi_list) ;
	mape_curloc_make_pseudo_pi(network) ;

	srandom(util_cpu_time()) ;
	num_control_pi = lsLength(pseudo_pi_list) ;
	//num_control_pi *= PI_VEC_LENGTH ;
	population_size = MAX(num_control_pi, 24);

	num_control_pi *= (user_option->curga_option.pi_bit + 1) ; 

	if( (population_size % 2) != 0 ) 
		population_size++;
	bits_per_pi = 1;
	chromo_word_size = ceil( (double)((num_control_pi*1.0) / ((sizeof(int)*8) / bits_per_pi)) );
	max_gen = user_option->curga_option.generation ;

	/* initialization */
	init_ga_alg(num_control_pi, bits_per_pi, chromo_word_size , population_size, max_gen);
	solution = ga_alg(mape_curga_init_pop_vector, mape_curloc_ga_fitness,  
			mape_curga_select, mape_curga_crossover, mape_curga_mutation,  		
			mape_curga_replace, mape_curga_finish) ;
	mape_timer_stop() ;
	mape_timer_report("GA time:") ;  
}

/*
 * mape_curploc_ga
 *
 * ga algorithm for maximum current extimation 
 * with localization mode after possibility estimation. 
 */
void 
mape_curploc_ga(network)
network_t *network ;
{
	int population_size;	
	int max_gen;
	int chromo_word_size;
	int bits_per_pi;
	int num_control_pi;
	chromosome_t * solution;

	mape_timer_reset() ;
	mape_timer_start() ;

	mape_clear_node_list(pseudo_pi_list) ;
	mape_curloc_make_pseudo_pi(network) ;

	srandom(util_cpu_time()) ;
	num_control_pi = lsLength(pseudo_pi_list) ;
	//num_control_pi *= PI_VEC_LENGTH ;
	population_size = MAX(num_control_pi, 24);

	num_control_pi *= (user_option->curga_option.pi_bit + 1) ; 
	num_control_pi *= PPI_TRANS_NUM ;

	if( (population_size % 2) != 0 ) 
		population_size++;
	bits_per_pi = 1;
	chromo_word_size = ceil( (double)((num_control_pi*1.0) / ((sizeof(int)*8) / bits_per_pi)) );
	max_gen = user_option->curga_option.generation ;

	/* initialization */
	init_ga_alg(num_control_pi, bits_per_pi, chromo_word_size, population_size, max_gen);
	solution = ga_alg(mape_curga_init_pop_vector, mape_curploc_ga_fitness,  
			mape_curga_select, mape_curga_crossover, mape_curga_mutation,  		
			mape_curga_replace, mape_curga_finish) ;
	mape_timer_stop() ;
	mape_timer_report("GA time:") ;  
}

/*
 * mape_curga_global_cluster
 *
 * evaluate the effectiveness of the global vector 
 * to the clusters. 
 */
#ifdef GLOBAL_LOCAL
void 
mape_curga_global_cluster(chr, groups)
chromosome_t * chr;
lsList groups ;
{
	int i,j;
	int count, switch_type, pos;
	lsGen gen, gen1, gen2 ;
	node_t *pi, *node ;
	lsHandle h ;
	double time ;
	double cur ;
	lsList group ;
	int pi_pos_unit ;

	count = get_gene_count() ;
	gen = lsStart(global_network->pi) ;

	mape_imp_reset(mape_cur_imp_do, (char *)NULL) ;
	pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	for(i=0;i<count;i+=(1+user_option->curga_option.pi_bit)) {
		mape_curga_parse_code_unit(chr, i, &switch_type, &pos) ;
		//fprintf(mapeout, "Get switch_type=%d and pos=%d\n", switch_type, pos) ;
		if(lsNext(gen, (lsGeneric *)&pi, &h) != LS_OK)
			mape_error("mape_curga_fitness\n", __FILE__, __LINE__) ;

		if(switch_type){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)(0.0 + pos*pi_pos_unit), 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)(pos*pi_pos_unit + SLOP_TIME), 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		}
		mape_imp_push(pi) ;
	}
	lsFinish(gen) ;
	mape_imp_start() ;
	fprintf(mapeout, "Entering into global_local\n") ;

	lsForeachItem(groups, gen1, group){
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		cur = mape_cur_max_current(global_network, &time, NULL) ;
		fprintf(mapeout, "Cluster current is: %lf\n", -cur) ;
	}
}
#endif GLOBAL_LOCAL

/*
 * mape_curga
 *
 * ga algorithm for maximum current extimation. 
 */
double 
mape_curga(network)
network_t *network ;
{
	int population_size;	
	int max_gen;
	int chromo_word_size;
	int bits_per_pi;
	int num_control_pi;
	chromosome_t * solution;
	double Iavg ;
	lsList trans_list;
	lsGen gen ;
	node_t *p ;
	double time ;
	int pop ;
	double rtv ;

#ifdef  GLOBAL_LOCAL
	lsList groups, group ;
#endif  GLOBAL_LOCAL

	mape_timer_reset() ;
	mape_timer_start() ;

#ifdef GLOBAL_LOCAL
	groups = mape_group_get_groups(network) ;
#endif GLOBAL_LOCAL

	mape_curlib_set_various_mode(TRUE) ;
	//srandom(util_cpu_time()) ;
	srandom(100) ;
	//num_control_pi = network_num_pi(network) ;
	if(mape_cur_is_cone_mode())
		num_control_pi = mape_cur_cone_pi_num() ;
	else 
		num_control_pi = network_num_pi(network) ;

	population_size = (int) (sqrt(network_num_internal(network) + network_num_pi(network)) * 3);	
	population_size/=2 ;
	population_size = (population_size < 24)?24:population_size ;

	/* add the bit for specifying the position of the switch */
	num_control_pi *= (user_option->curga_option.pi_bit + 1) ; 
	//population_size *= 2 ;
	if( (population_size % 2) != 0 ) 
		population_size++;
	bits_per_pi = 1;
	chromo_word_size = ceil( (double)((num_control_pi*1.0) / ((sizeof(int)*8) / bits_per_pi)) );
	max_gen = user_option->curga_option.generation ;

	/* initialization */
	init_ga_alg(num_control_pi, bits_per_pi, chromo_word_size , population_size, max_gen);

	solution = ga_alg(mape_curga_init_pop_vector, mape_curga_fitness,  
			mape_curga_select, mape_curga_crossover, mape_curga_mutation,  		
			mape_curga_replace, mape_curga_finish) ;

	/*
	fprintf(mapeout, "Iavg=%.2lf\n", mape_curga_fitness(solution)/network_num_internal(network)) ; 
	mape_spice_set_ga_solution(solution) ;
	mape_spice(network) ;
	*/

	rtv = mape_curga_fitness(solution) ;



	//trans_list = lsCreate() ;
	//mape_cur_set_dump_flag(TRUE) ;
	//mape_curga_fitness_detail(solution, &time, trans_list) ;
	//mape_cur_set_dump_flag(FALSE) ;

	/*
	fprintf(mapeout, "Transistion nodes at time %.2lf: ", time) ;
	lsForeachItem(trans_list, gen, p)
		fprintf(mapeout, "%s ", p->name) ;
	fprintf(mapeout, "\n") ;
	lsDestroy(trans_list, NULL) ; 
	*/

#ifdef PBSSCE
	mape_clean_given_gate() ;
	lsForeachItem(trans_list, gen, p)
		mape_push_given_gatep(p) ;
	mape_pbssce_itr_estimate_current(network, trans_list) ;
#endif

	mape_timer_stop() ;
	mape_timer_report("GA time:") ;  

#ifdef GLOBAL_LOCAL
	mape_curga_global_cluster(solution, groups) ; 
#endif GLOBAL_LOCAL
	return rtv ;
	
}

/*
 * curga_grp
 *
 * group ga.  
 */
void 
mape_curga_grp(network)
network_t *network ;
{
	lsList groups, group ;
	lsGen gen, gen1, gen2 ;
	node_t *node ;
	int i ;

	groups = mape_group_get_groups(network) ;
	mape_curlib_set_various_mode(TRUE) ;	
	mape_cur_set_cone_mode(TRUE) ;
	fprintf(mapeout, "after get groups\n") ;

#ifdef PBSSCE
	mape_pbssce_build_coswitch_db(network) ;
	mape_pbssce_st_possibility(network) ;
	mape_cur_clear(network) ;
#endif

	//mape_curlib_set_unit_mode(TRUE) ;	

	/*
	foreach_node(network, gen, node){
		if(mape_curga_is_special_group(node)){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
	}
	fprintf(mapeout, "\n") ;
	mape_curga(network) ;
	*/

	i = 0 ;
	lsForeachItem(groups, gen1, group){
		i++ ;
		/*
		if(i%GROUP_SAMPLE_RATE != 0)
			continue ;
		*/
		mape_cur_group_cone(group) ;
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		mape_curga(network) ;
	}
}

/*
 * mape_curloc
 *
 * to get the maximum current using localization method.  
 */
void 
mape_curloc(network)
network_t *network ;
{
	lsList groups, group ;
	lsGen gen, gen1, gen2 ;
	node_t *node ;

	groups = mape_group_get_groups(network) ;
	mape_curlib_set_various_mode(TRUE) ;	

	lsForeachItem(groups, gen1, group){
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		mape_curloc_ga(network) ;
	}
}

/*
 * mape_curploc
 *
 * to get the maximum current of the localized circuit after 
 * possibility estimation
 */
void 
mape_curploc(network)
network_t *network ;
{
	lsList groups, group ;
	lsGen gen, gen1, gen2 ;
	node_t *node ;

	groups = mape_group_get_groups(network) ;
	mape_curlib_set_various_mode(TRUE) ;	

	/* get the switching possibility for each gate */
	mape_pbssce_st_possibility(network) ;

	/*
	foreach_node(network, gen, node){
		if(mape_curga_is_special_group(node)){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
			//mape_pbssce_node_dump_st(node) ;
		}
	}
	fprintf(mapeout, "\n") ;
	mape_curploc_ga(network) ;
	*/

	lsForeachItem(groups, gen1, group){
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		mape_curploc_ga(network) ;
	}
}

/*
 * mape_curga_is_special_group
 *
 * make a special group for examination in
 * detail. 
 */
int 
mape_curga_is_special_group(node)
node_t *node ;
{
	char *s ;
	char *tmp[16] = {"[1420]", "[1421]", "[780]", "[738]", "[794]", "[766]", "[3254]", 
			 "[556]", "[564]", "[572]", "[580]", "[752]", "[1476]", "[1480]", 
			 "[3385]", "13(814)"};
	int  i ;

	s = node->name ;
	for(i=0;i<16;i++)
		if(!strcmp(s, tmp[i]))
			return TRUE ;

	return FALSE ;
}
