/*
 * mape_cursa.c
 *
 * simulated annealing for maximum current estimation. 
 * This method is special for group's maximum current estimation.  
 * 
 * For one group, just use SA to obtain the maximum current, for 
 * other group, just get the maximum current among all the result 
 * calculated. 
 *
 * Changbo Long,  Mar. 24, 2002. 
 */
#include "mape_int.h"

extern 	char * 	mape_cursa_sa_create(void) ;
extern 	double 	mape_cursa_sa_cost(char *) ;
extern 	void 	mape_cursa_sa_destory(char *) ;
extern 	void 	mape_cursa_sa_move(char *, char *) ;
extern 	int 	mape_cursa_sa_equilibrum(int, int, int) ;
extern 	int 	mape_cursa_sa_stop(double, int, double) ;
extern 	void 	mape_cursa_sa_accept(int) ;

node_t *last_pi 	= NULL ;

/*
 * mape_cursa_sa_create
 *
 * just return the network's address.
 */
char *
mape_cursa_sa_create(void)
{
	return (char *)global_network ;
}

/*
 * mape_cursa_sa_destroy
 *
 * do nothing.
 */
void
mape_cursa_sa_destroy(nouse)
char *nouse ;
{
	return ;
}

/*
 * mape_cursa_sa_cost
 *
 * the cost function. 
 */
double 
mape_cursa_sa_cost(nouse)
char *nouse ;
{
	double time ;
	//return 0.0 ;

	return mape_cur_max_current(global_network, &time, NULL) ;
}

/*
 * mape_cursa_random_pi
 *
 * randomly pickup a pi. 
 */
node_t *
mape_cursa_random_pi(network)
network_t *network ;
{
	int num_pi ;
	long v ;

	num_pi = network_num_pi(network) ;
	v = random() ;
	return network_get_pi(network, (int)(v%num_pi)) ;
}

/*
 * mape_cursa_sa_move
 *
 * make some change to the pi. 
 */
void 
mape_cursa_sa_move(from, to)
char *from, *to ;
{
	last_pi = mape_cursa_random_pi(global_network) ;
	mape_curitr_change_pi(last_pi) ;
}

/*
 * mape_cursa_sa_accept 
 *
 * callback function for informing the accept of the move. 
 */
void 
mape_cursa_sa_accept(ac)
int ac ;
{
	if(ac){
		mape_curitr_update_untran_list() ;
	}else {
		mape_curitr_backtrack_change(last_pi) ;
		mape_curitr_clear_change_list() ;
	}
}

/*
 * mape_cursa_sa_equilibrum
 *
 * equilibrum stop 
 */
int 
mape_cursa_sa_equilibrum(up, total, reject)
int up, total , reject;
{
	if(up >= 100 || total >= 200)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_cursa_stop
 *
 * stop the process of the sa. 
 */
int 
mape_cursa_sa_stop(temp, total, reject_ratio)
double temp, reject_ratio ;
int total ;
{
	if(temp <= 1)
		return TRUE ;
	else 
		return FALSE;
}

/*
 * randomly assign value to the input. 
 */
void 
mape_cursa_initialize(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi, *latch ;
	int i = 0 ;
	long v ;
	int is_fall ;

	/*calculate the loading capacitance for each node*/
	mape_imp_reset(mape_curitr_imp_do, (char *)NULL) ;
	foreach_primary_input(network, gen, pi){
		v = random() ;
		if(v%2 == 0)
			is_fall = TRUE ;
		else 
			is_fall = FALSE ;
		if(is_fall){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
			mape_imp_push(pi) ;
		}else{
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
			mape_imp_push(pi) ;
		}
	}
	mape_imp_start() ;
}

/*
 * set all the group's list.
 *
 * mape_cursa_set_group
 */


/*
 * using sa to obtain the maximum current for all the groups 
 * at the same time. 
 */
double
mape_cursa(network, time, list)
network_t *network ;
double *time ;
lsList list ;
{
	double t_w ;
	char *sol ;

	mape_curlib_set_various_mode(TRUE) ;
	mape_cursa_initialize(network) ; 
	mape_sa_set(100, 0.9, mape_cursa_sa_create, mape_cursa_sa_destroy, 
		mape_cursa_sa_cost, mape_cursa_sa_move, mape_cursa_sa_equilibrum, 
		mape_cursa_sa_stop, mape_cursa_sa_accept) ;
	sol = (char *)network ;
	mape_sa_start(&sol) ;
	*time = 0 ;
	return mape_cursa_sa_cost(sol) ;
}
