/*
 * mape_depsa.c
 *
 * sa for minimize the switching time overlapping. 
 *
 * Changbo Long.  Nov. 24, 2002. (clong@cae.wisc.edu)
 * switching time overlapping is used in the paper in DAC 1998. 
 */

#include "mape_int.h"

#define MAXDEPTH 256
#define DEPSA(n) ((int *)(((mape_t *)(n->mape))->depsa))

extern char * 	mape_depsa_sa_create(void) ;
extern double 	mape_depsa_sa_cost(char *) ;
extern void 	mape_depsa_sa_destroy(char *) ;
extern void 	mape_depsa_sa_move(char *, char *) ;
extern int 	mape_depsa_sa_equilibrum(int, int, int) ;
extern int 	mape_depsa_sa_stop(double, int, double) ;
extern void 	mape_depsa_sa_accept(int) ;

extern node_t  *mape_sosa_random_take(lsList) ;

lsList depsa_clusters = NULL;
int max_depth = 0 ;
lsList g1, g2 ; 	//cluster is being exchanging. 
node_t *n1, *n2 ;	//node is being exchanging. 
int *cur_cost = NULL;	//cost for all clusters.  

/*
 * mape_depsa_sa_create
 *
 * jsut return the groups address. 
 */
char *
mape_depsa_sa_create(void)
{
	return (char *) depsa_clusters ;
}

/*
 * mape_depsa_sa_destroy
 *
 * do nothing. 
 */
void 
mape_depsa_sa_destroy(nouse)
char *nouse ;
{
	return ;
}

/*
 * mape_depsa_sw_overlap
 *
 * switch overlap
 */
int 
mape_depsa_sw_overlap(l)
lsList l ;
{
	lsGen gen ;	
	node_t *node ;
	int i, rtv, t, c, j ;
	int *x[128] ;

	memset(x, 0, sizeof(int *)*128) ;
	t = 0 ;
	lsForeachItem(l, gen, node){
		//fprintf(mapeout, "%s,", node->name) ;
		x[t] = DEPSA(node) ;
		t++ ;
	}
		
	rtv = 0 ;
	for(i=0;i<=max_depth;i++){
		c = 0 ;
		for(j=0;j<t;j++)
			if(x[j][i])	
				c++ ; 
		rtv += ((c-1)>0?(c-1):0)  ;
	}
	//fprintf(mapeout, "overlaping is: %d\n", rtv) ; 

	return rtv ;
}

/*
 * mape_depsa_sw_overlap
 *
 * switch overlap
 */
int 
mape_depsa_sw_overlap_s(l, n)
lsList l ;
node_t *n ;
{
	lsList tmp ;
	int rtv ;

	tmp = lsCopy(l, NULL) ;
	lsNewEnd(tmp, (lsGeneric)n, LS_NH) ;
	rtv = mape_depsa_sw_overlap(tmp) ;
	lsDestroy(tmp, NULL) ;
	return rtv ;
}

/*
 * mape_depsa_sa_cost. 
 *
 * the cost function. 
 */
double 
mape_depsa_sa_cost(nouse)
char *nouse ;
{
	int i ;
	int total_o = 0 ;
	lsGen gen ;
	lsList l ;
	int o ;
	lsList tmp ;

	if(cur_cost == NULL){
		cur_cost = ALLOC(int, lsLength(depsa_clusters)) ;
		for(i=0;i<lsLength(depsa_clusters);i++) 
			cur_cost[i] = -1 ;
	}

	i = 0 ;
	lsForeachItem(depsa_clusters, gen, l){
		if(l != g1 && l != g2){
			if(cur_cost[i] == -1){
				o = mape_depsa_sw_overlap(l) ;
				cur_cost[i] = o ;
				total_o += o ;
			}else{
				total_o += cur_cost[i] ;
			}
		}else if(l == g1){
			o = mape_depsa_sw_overlap_s(l, n2) ;
			total_o += o ;
		}else if(l == g2){
			o = mape_depsa_sw_overlap_s(l, n1) ;
			total_o += o ;
		}
		i ++ ;
	}

	return (double)total_o ;
}

/*
 * mape_depsa_sa_move
 *
 * exchange two gates. 
 */
void 
mape_depsa_sa_move(from, to)
char *from, *to ;
{
	long v ;
	int x1, x2 ;
	lsGen gen ;
	int i ;
	lsList g ;

	v = random() ;
	x1 = (int)(v%lsLength(depsa_clusters)) ;
	do{
		v = random() ;
		x2 = (int)(v%lsLength(depsa_clusters)) ;
	}while(x1 == x2) ;

	g1 = g2 = NULL ;
	n1 = n2 = NULL ;

	i = 0 ;
	lsForeachItem(depsa_clusters, gen, g){
		if(g1 && g2){
			lsFinish(gen) ;
			break ;
		}
		if(i == x1) g1 = g ;
		if(i == x2) g2 = g ;
		i ++ ;
	}
	if(g1 == NULL || g2 == NULL)
		mape_error("g1 and g2 shouldn't be zero !", __FILE__, __LINE__) ;

	n1 = mape_sosa_random_take(g1) ;
	n2 = mape_sosa_random_take(g2) ;
}

/*
 * mape_depsa_sa_accept. 
 *
 * callback function for informing the accept of the move. 
 */
void 
mape_depsa_sa_accept(ac)
int ac ;
{
	int i, i1, i2 ;
	lsGen gen ;
	lsList l ;

	if(ac){
		i = 0 ;
		lsForeachItem(depsa_clusters, gen, l){
			if(l == g1)
				i1 = i ;
			if(l == g2)
				i2 = i ;
			i++ ;
		}

		lsNewEnd(g1, (lsGeneric)n2, LS_NH) ;
		lsNewEnd(g2, (lsGeneric)n1, LS_NH) ;
		cur_cost[i1] = mape_depsa_sw_overlap(g1) ;
		cur_cost[i2] = mape_depsa_sw_overlap(g2) ;
	}else{
		lsNewEnd(g1, (lsGeneric)n1, LS_NH) ;
		lsNewEnd(g2, (lsGeneric)n2, LS_NH) ;
	}

	g1 = g2 = NULL ;
	n1 = n2 = NULL ;
}

/*
 * mape_depsa_sa_equilibrum
 *
 * equilibrum stop
 */
int 
mape_depsa_sa_equilibrum(up, total, reject)
int up, total, reject ;
{
	int times ;
	times = (int)sqrt(lsLength(depsa_clusters)) ;

	if(up >= 100*times || total >= 200*times)
		return TRUE ;
	else
		return FALSE ;
}

/*
 * mape_depsa_sa_stop
 *
 * stop function. 
 */
int 
mape_depsa_sa_stop(temp, total, reject_ratio)
double temp, reject_ratio ;
int total ;
{
	if(temp <= 0.1)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_depsa_set_node_sw
 *
 * set node's sw
 */
void 
mape_depsa_set_node_sw(node, sw, n)
node_t *node ;
int *sw ;
int n ;
{
	int i ;

	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if(DEPSA(node) == NULL)
		DEPSA(node) = ALLOC(int, MAXDEPTH) ;

	memset(DEPSA(node), 0, sizeof(int)*MAXDEPTH) ;
	
	for(i=0;i<n;i++)
		DEPSA(node)[i] = sw[i] ;
}

/*
 * mape_depsa_imp_do
 *
 * do the implication to calculate the switch time for all the 
 * nodes. 
 */
int
mape_depsa_imp_do(node, arg)
node_t *node ;
char *arg ;
{
	int i, j ;
	node_t *fanin ;
	int sw[MAXDEPTH] ;


	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;

	memset(sw, 0, sizeof(int) * MAXDEPTH) ;

	foreach_fanin(node, i, fanin){
		for(j=0;j<MAXDEPTH;j++){
			if(DEPSA(fanin)[j]){
				sw[j+1] = 1 ;
				if(max_depth < j+1)
					max_depth = j ;
			}
		}
	}
	mape_depsa_set_node_sw(node, sw, MAXDEPTH) ;

	return TRUE ;
}


/*
 * mape_depsa_init_sw_time
 *
 * get all the switching time for all the gates. 
 */
void 
mape_depsa_init_sw_time(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;
	int sw[1] = {1} ;

	mape_imp_reset(mape_depsa_imp_do, NULL) ;
	foreach_primary_input(network, gen, pi){
		mape_depsa_set_node_sw(pi, sw, 1) ;
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
}


/*
 * mape_depsa_init
 *
 * initialize the depsa 
 */
void 
mape_depsa_init(network)
network_t *network ;
{
	lsList groups, g ;
	lsGen gen ;

	if(depsa_clusters == NULL)
		depsa_clusters = lsCreate() ;
	groups = mape_group_get_groups(network) ;
	// get all node's switching time. 
	mape_depsa_init_sw_time(network) ;

	lsForeachItem(groups, gen, g)
		lsNewEnd(depsa_clusters, (lsGeneric)g, LS_NH);
}

/*
 * mape_depsa_dump
 *
 * dump the depsa
 */
void 
mape_depsa_dump(network)
network_t *network ;
{
	node_t *node ;
	int i ;
	lsGen gen ;

	foreach_node(network, gen, node){
		fprintf(mapeout, "node %s's depsa \n\t", node->name) ;  
		for(i=0;i<MAXDEPTH;i++){
			if(DEPSA(node)[i])
				fprintf(mapeout, "%d ", i) ;
		}
		fprintf(mapeout, "\n") ;
	}
}

/*
 * mape_depsa_get_clusters
 *
 */
lsList 
mape_depsa_get_clusters()
{
	return depsa_clusters ;
}


/*
 * mape_depsa
 *
 * switching time overlapping minimization. 
 */
void
mape_depsa(network)
network_t *network ;
{
	double t_w ;
	char *sol ;

	mape_depsa_init(network) ;
	//mape_depsa_dump(network) ;

	srandom(util_cpu_time()) ;
	sol = (char *)depsa_clusters ;

	//fprintf(mapeout, "Cost before SA: %.2f\n", mape_depsa_sa_cost(sol)) ;
	mape_sa_set(75, 0.8, mape_depsa_sa_create, mape_depsa_sa_destroy, 
		mape_depsa_sa_cost, mape_depsa_sa_move, mape_depsa_sa_equilibrum, 
		mape_depsa_sa_stop, mape_depsa_sa_accept) ;
	sol = (char *)depsa_clusters ; 
	mape_sa_start(&sol) ;
	//fprintf(mapeout, "Cost after SA: %.2f\n", mape_depsa_sa_cost(sol)) ;
}
