/*
 * mape_disksim
 *
 * simulation circuit and store the waveform in the
 * disk. The simulation result could be extracted back
 * from the simulation and used to verify the sosa 
 * and depsa package. 
 *
 * Changbo Long  clong@cae.wisc.edu.  	Nov. 25, 2002
 */
#include "mape_int.h"
#include <sys/wait.h>
#include <dirent.h>

#define	SIMDIR		"/home/clong/work/sis/sis_uw/mape/src/sim"
#define SIMTIMES  	5000

#define MAXLINELENGTH	100
#define WAVEFORM_BEGIN	"##"
#define NODEWAVEEND	";"

char *file_path[32] ;	 	// the path for the files which stores waveform.  
FILE *cur_fp = NULL ;	 	// current fp, which is reading.   
int file_ind = 0 ;		// current file index. 
int new_read = TRUE ;	

/*
 * mape_disksim_compress 
 * 
 * compress the file. 
 */
void 
mape_disksim_compress(s)
char *s ;
{
	int forkpid, status ;
	char tmp[256] ;

	if((forkpid = vfork()) == 0){
		sprintf(tmp, "/home/clong/bin/gzip %s* \n", s) ; 
		system(tmp) ; 
		exit(0) ;
	}

	waitpid(forkpid, &status,  WNOWAIT) ;
}

/*
 * mape_disksim_decompress
 *
 *
 * decompress the file. 
 */
void
mape_disksim_decompress(s)
char *s ;
{
	int forkpid, status ; 
	char tmp[256] ;

	if((forkpid = vfork()) == 0){
		sprintf(tmp, "/home/clong/bin/gzip -d %s* \n", s) ;
		system(tmp) ;
		exit(0) ;
	}

	waitpid(forkpid, &status, WNOWAIT) ;
}

/*
 * mape_disksim_read
 *
 * read the node's waveform from the file. 
 */
void 
mape_disksim_build_fn(network)
network_t *network ;
{
	DIR *dirp ;
	struct dirent *dp ;
	char name[128], *p ;
	char path[128] ;
	int i ;

	memset(file_path, 0, sizeof(char *)*32) ;

	i = 0 ;
	dirp = opendir(SIMDIR) ;
	while(dirp){
		if ((dp = readdir(dirp)) != NULL) {
			if(strstr(dp->d_name, network->net_name) == NULL)
				continue ;
			sprintf(path, "%s/%s", SIMDIR, dp->d_name) ;
			file_path[i ++ ] = strdup(path) ; 
		} else {
			closedir(dirp) ;
			return ;
		}
	}
}

/*
 * mape_disksim_token_next(fp, token)
 *
 * get the nexttoken. 
 *
 * Returning false means reaching end of the file. 
 */
int
mape_disksim_token_next(fp, token)
FILE *fp ;
char *token ;
{
	static int last = -1 ;
	int r = -1 ;
	char *p = token ;

	if(feof(fp) && last == -1)  // end of file and no last left.  
		return FALSE ;

	if(last != -1)
		r = last ;
	else 
		r = fgetc(fp) ;

	while(r == ' ' || r == '\n' || r == -1){
		if(feof(fp))
			return FALSE ;
		r = fgetc(fp) ;
	}

	*(p++) = r ;
	r = -1 ;
	for(;;){
		if(feof(fp))
			break ;
		r = fgetc(fp) ;
		if(r != ' ' && r != '\n' && r != ';')
			*(p++) = r ;
		else 
			break ;
	}
	*p = '\0' ;
	last = r ;
	return TRUE ;
}

/*
 * mape_disksim_check_nodes
 *
 * check if the nodes's order is the same. 
 */
int 
mape_disksim_check_nodes(network, fp)
network_t *network ;
FILE *fp ;
{
	char token[128] ;
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node){
		mape_disksim_token_next(fp, token) ;
		//fprintf(mapeout, "toke = %s, node->name = %s\n", token, node->name) ;
		if(strcmp(token, node->name))
			return FALSE ;
	}
	return TRUE ;
}

/*
 * mape_disksim_open_next_file.
 *
 * open next file.  
 */
FILE *
mape_disksim_open_next_file(void)
{
	FILE *fp ;
	char tmp[128], *p ;

	if(file_path[file_ind] == NULL)
		return NULL ;

	if(strstr(file_path[file_ind], ".gz") == NULL){
		fp = fopen(file_path[file_ind], "r+") ;
	}else{
		//mape_disksim_decompress(file_path[file_ind]);
		strcpy(tmp, file_path[file_ind]) ;
		p = strstr(tmp, ".gz") ;
		*p = '\0' ;
		fp = fopen(tmp, "r+") ;
	}
	if(!mape_disksim_check_nodes(global_network, fp)){
		fclose(fp);
		fprintf(mapeout, "check file %s \n", file_path[file_ind]) ;
		//mape_disksim_compress(file_path[file_ind]) ;
		mape_error("Disksim check nodes fail!", __FILE__, __LINE__) ;
	}else 
		//fprintf(mapeout, "%s check nodes passed \n", file_path[file_ind]) ; 
	file_ind ++ ;
	return fp ;
}

/*
 * mape_disksim_read_node_waveform
 *
 * read node's waveform
 */
int 
mape_disksim_read_node_waveform(node)
node_t *node ;
{
	char token[128] ;
	double time ;
	double value ;

	for(;;){
		if(!mape_disksim_token_next(cur_fp, token))
			return FALSE ;
		if(!strcmp(token, NODEWAVEEND))
			return TRUE ;
		time = atof(token) ;
		if(!mape_disksim_token_next(cur_fp, token))
			return FALSE ;
		value = atof(token) ;
		mape_cur_push_vol_point(node, value, time, 1); 
	}
	return TRUE ;
}

/*
 * mape_disksim_read_next_simulation
 *
 * read the next simulation result.  
 */
int 
mape_disksim_read_next_simulation(network) 
network_t *network ;
{
	char token[128] ;
	lsGen gen ;
	node_t *node ;

	if(!mape_disksim_token_next(cur_fp, token)){
		fclose(cur_fp) ;
		cur_fp = NULL ;
		return FALSE ;
	}
	if(strcmp(token, WAVEFORM_BEGIN)){
		fprintf(mapeout, "Got token: [%s]\n", token) ;
		mape_error("waveform begin symbol is expected\n", __FILE__, __LINE__) ;
	}

	foreach_node(network, gen, node){
		if(!mape_disksim_read_node_waveform(node))
			mape_error("waveform for node is expected to end \n", __FILE__, __LINE__) ;
	}
	return TRUE ;
}

/*
 * mape_disksim_read
 *
 * read_file ; 
 */
int 
mape_disksim_read(network)
network_t *network ;
{
	static int b_fn = FALSE ;
	char s[128] ;

	if(new_read){
		mape_disksim_build_fn(network) ;
		sprintf(s, "%s/%s", SIMDIR, network->net_name) ;
		mape_disksim_decompress(s) ;
		new_read = FALSE ;
	}
	for(;;){
		if(cur_fp == NULL){
			cur_fp = mape_disksim_open_next_file() ;
			if(!cur_fp){
				sprintf(s, "%s/%s", SIMDIR, network->net_name) ;
				mape_disksim_compress(s) ;
				return FALSE ;
			}
		}
		if(mape_disksim_read_next_simulation(network));	
			return TRUE ;
	}
}

/*
 * mape_disksim_read_reset
 *
 * read reset. 
 */
void 
mape_disksim_read_reset(network)
network_t *network ;
{
	int i ;

	for(i=0;i<32;i++){
		if(file_path[i]){
			free(file_path[i]) ;
			file_path[i] = NULL ;
		}
	}

	file_ind = 0 ;
	if(cur_fp)
		fclose(cur_fp) ;
	cur_fp = NULL ;
	new_read = TRUE ;
}

/* 
 * mape_disksim_write
 *
 * write the simulation result into the disk. 
 */
void 
mape_disksim_write_nodes(network, fp)
network_t *network ;
FILE *fp ;
{
	lsGen gen ;
	node_t *node ;
	int count ;
	char l[256] ;

	l[0] = '\0' ;
	foreach_node(network, gen, node){
		if(strlen(l) + strlen(node->name) > MAXLINELENGTH){
			fprintf(fp, "%s\n", l) ;
			l[0] = '\0' ;
		}
		sprintf(l, "%s %s", l, node->name) ;   
	}
	if(strlen(l))
		fprintf(fp, "%s\n", l) ;
}

/* 
 * mape_disksim_write
 *
 * write the simulation result into the disk. 
 */
void 
mape_disksim_write_waveform(network, fp)
network_t *network ;
FILE *fp ;
{
	lsGen gen, gen1 ;
	char s[2048], point[1024] ;
	node_t *node ;
	lsList vol_points ;
	lsList nvl ;
	mape_vol_point_t *p ;

	fprintf(fp, "%s\n", WAVEFORM_BEGIN) ;
	*s = '\0' ;
	nvl = lsCreate() ;
	foreach_node(network, gen, node){
		vol_points = mape_cur_top_vol_points(node) ;
		while(lsLength(nvl)){
			lsDelEnd(nvl, (lsGeneric *)&p) ;	
			free(p) ;
		}
		mape_pbssce_add_node_current_v(node, vol_points, nvl) ;
		lsForeachItem(nvl, gen1, p){
			sprintf(point, "%.2f %.2f ", p->time, p->value) ;
			if(strlen(point) + strlen(s) > MAXLINELENGTH){
				fprintf(fp, "%s\n", s) ;
				*s = '\0' ;
				strcpy(s, point) ;
			}else
				strcat(s, point) ;
		}
		strcat(s, NODEWAVEEND) ;
		strcat(s, " ") ;
	}
	lsDestroy(nvl, free) ;
	if(strlen(s) > 0)
		fprintf(fp, "%s\n", s) ;
}

/*
 * open a new file for writting the result.
 *
 */
FILE *
mape_disksim_opennew(network, fn)
network_t *network ;
char *fn ;
{
	int i ;
	char path[128], path1[128] ;
	FILE *fp ;

	i = 1 ;	
	for(;;){
		sprintf(path, "%s/%s.%d", SIMDIR, network->net_name, i) ; 
		sprintf(path1, "%s/%s.%d.gz", SIMDIR, network->net_name, i) ; 
		strcpy(fn, path) ;
		if(!mape_check_regfile(path) && !mape_check_regfile(path1)){
			fp = fopen(path, "w+") ;
			//fprintf(mapeout, "open file : %s\n", path) ;
			return fp ;
		}
		i++ ;
	}
}


/* 
 * mape_disksim. 
 *
 * simulation on disk. 
 */
void 
mape_disksim(network)
network_t *network ;
{
	FILE *fp ;
	int i ;
	char fn[128] ;

	if(!mape_check_dir(SIMDIR)){
		fprintf(mapeout, "Error when open directory %s!\n", SIMDIR) ;
		return ;
	}

	fp = mape_disksim_opennew(network, fn) ;
	mape_disksim_write_nodes(network, fp) ;

	mape_curlib_set_various_mode(TRUE) ;
	for(i=0;i<SIMTIMES;i++){
		if(fp == NULL){
			fprintf(mapeout, "Error when open new file !\n") ;
			return ;
		}
		if(i%1000 == 0 && i!=0)
			fprintf(mapeout, "%dth simulation stored\n", i) ;
			
		mape_cur_clear(network) ;
		mape_cur_initialize_random(network) ;
		mape_disksim_write_waveform(network, fp) ;
	}
	fclose(fp) ;

	mape_disksim_compress(fn) ;
}
