/*
 * mape_dump.c
 *
 * used to dump the information of the circuit, just for debuggering. 
 * Changbo Long(clong@cae.wisc.edu)
 */
#include "mape_int.h"

static char * node_type_string(node_type_t, char *) ;
static char * node_function_string(node_type_t, char *) ;

/*
 * node_function_string 
 *
 * convert the node function into a
 * string to display. 
 */
char *
node_function_string(type, string)
node_function_t type ;
char *string ;
{
	switch(type){
		case NODE_PI :
			strcpy(string, "NODE_PI") ;
			break ;
		case NODE_PO :
			strcpy(string, "NODE_PO") ;
			break ;
		case NODE_0 :
			strcpy(string, "NODE_0") ;
			break ;
		case NODE_1 :
			strcpy(string, "NODE_1") ;
			break ;
		case NODE_BUF :
			strcpy(string, "NODE_BUF") ;
			break ;
		case NODE_INV :
			strcpy(string, "NODE_INV") ;
			break ;
		case NODE_AND :
			strcpy(string, "NODE_AND") ;
			break ;
		case NODE_OR :
			strcpy(string, "NODE_OR") ;
			break ;
		case NODE_COMPLEX :
			strcpy(string, "NODE_COMPLEX") ;
			break ;
		case NODE_UNDEFINED :
			strcpy(string, "NODE_UNDEFINED") ;
			break ;
		default:
			strcpy(string, "ERROR_FUNCTION") ;
			break ;
	}
	return string ;
}

/*
 * node_type_string 
 *
 * convert the node type into a
 * string to display. 
 */
char *
node_type_string(type, string)
node_type_t type ;
char *string ;
{
	switch(type){
		case PRIMARY_INPUT :
			strcpy(string, "PRIMARY_INPUT") ;
			break ;
		case PRIMARY_OUTPUT :
			strcpy(string, "PRIMARY_OUTPUT") ;
			break ;
		case INTERNAL :
			strcpy(string, "INTERNAL") ;
			break ;
		case UNASSIGNED :
			strcpy(string, "UNASSIGNED") ;
			break ;
		default:
			strcpy(string, "ERROR_TYPE") ;
	}
	return string ;
}

/*
 * mape_print_node_type ;
 *
 *
 * print node's type. 
 */
char *
mape_print_node_type(type)
node_type_t type ;
{
	static char string[128] ;
	return node_type_string(type, string) ;
}

/*
 * mape_print_node_value
 *
 * print the node's value in the returned string. 
 */
char *
mape_print_node_value_c(value)
mape_node_value_t value ;
{
	static char string[10] ;
	return mape_node_value_string_c(value, string) ;
}

/*
 * mape_node_value_string
 *
 *
 * convert the node's value into a string
 */
char *
mape_node_value_string_c(mape_node_value_t value, char *string)
{
	switch(value){
		case MAPE_FALL:
			strcpy(string, "1/0") ;
			return string ;
		case MAPE_RISE:
			strcpy(string, "0/1") ;
			return string ;
		case MAPE_U_0:
			strcpy(string, "X/0") ;
			return string ;
		case MAPE_0_U:
			strcpy(string, "0/X") ;
			return string ;
		case MAPE_0_0:
			strcpy(string, "0/0") ;
			return string ;
		case MAPE_U_U:
			strcpy(string, "X/X") ;
			return string ;
		case MAPE_1_1:
			strcpy(string, "1/1") ;
			return string ;
		case MAPE_1_U:
			strcpy(string, "1/X") ;
			return string ;
		case MAPE_U_1:
			strcpy(string, "X/1") ;
			return string ;
		default:
			mape_error("An error node's value detected", __FILE__, __LINE__) ;
	}
}

/*
 * mape_print_node_value
 *
 * print the node's value in the returned string. 
 */
char *
mape_print_node_value(value)
mape_node_value_t value ;
{
	static char string[10] ;
	return mape_node_value_string(value, string) ;
}

/*
 * mape_node_value_string
 *
 *
 * convert the node's value into a string
 */
char *
mape_node_value_string(mape_node_value_t value, char *string)
{
	switch(value){
		case MAPE_FALL:
			strcpy(string, "1/0") ;
			return string ;
		case MAPE_RISE:
			strcpy(string, "0/1") ;
			return string ;
		case MAPE_U_0:
			strcpy(string, "X/0") ;
			return string ;
		case MAPE_0_U:
			strcpy(string, "0/X") ;
			return string ;
		case MAPE_0_0:
			strcpy(string, "0/0") ;
			return string ;
		case MAPE_U_U:
			strcpy(string, "X/X") ;
			return string ;
		case MAPE_1_1:
			strcpy(string, "1/1") ;
			return string ;
		case MAPE_1_U:
			strcpy(string, "1/X") ;
			return string ;
		case MAPE_U_1:
			strcpy(string, "X/1") ;
			return string ;
		default:
			mape_error("An error node's value detected", __FILE__, __LINE__) ;
	}
}

/*
 * dump_s_set
 *
 * dump a special set with a string. 
 */
void 
dump_s_set(s, nin, set)
char *s ;
int nin ;
pset set ;
{
	int i ;
	if(set == NULL){
		mape_error("dump's set is NULL", __FILE__, __LINE__) ; 
		return ;
	}
	fprintf(mapeout, "%s", s) ;
	dump_set(nin, set) ;
	fflush(mapeout) ;
}

/*
 * dump_set ;
 *
 * dump a special set 
 * of the node's function. 
 */
void 
dump_set(nin, set)
int nin ;
pset set ;
{
	int i  ;
	if(set == NULL){
		mape_error("dump's set is NULL", __FILE__, __LINE__) ; 
		return ;
	}

	for(i=0;i<nin;i++){
		switch(GETINPUT(set, i)){
		case ZERO:
			fprintf(mapeout, "0") ;
			break ;
		case ONE:
			fprintf(mapeout, "1") ;
			break ;
		case TWO:
			fprintf(mapeout, "X") ;
			break ;
		default:
			fprintf(mapeout, "Z") ;
			break ;
		}
	}
}

/*
 * dump a set family. 
 *
 */
void
dump_sf(n, cover)
int n ;
pcover cover ;
{
	pset ps, ps_last ;

	fprintf(mapeout, "{") ;
	foreach_set(cover, ps_last, ps){
		dump_set(n, ps) ;
		fprintf(mapeout, ",") ;
	}
	fprintf(mapeout, "}") ;
}

/*
 * dump_node 
 *
 * This function is used for debugging
 * It will dump all the information of the node 
 * in the circuit
 */
void 
dump_node(node)
node_t *node ;
{
	char string[128] ;
	int i, length ;
	node_t *p ;
	lsGen gen ;
	lsList stack ;
	mape_imp_value_t *imp_value ;
	pset last_ps, ps ;
	lib_gate_t *gate ;

	fprintf(mapeout, "%s\n", node->name) ;
	fprintf(mapeout, "\t    short_name : %-s\n", node->short_name) ;
	fprintf(mapeout, "\t          type : %-s\n", node_type_string(node->type, string)) ;
	gate = lib_gate_of(node) ;
	fprintf(mapeout, "\t      lib type : %-s\n", (gate == NULL?"<none>":lib_gate_name(gate)) ) ;
	fprintf(mapeout, "\t        sis_id : %-d\n", node->sis_id);
	fprintf(mapeout, "\t      function : %-s\n", node_function_string(node_function(node), string)) ;
	fprintf(mapeout, "\t         level : %-d\n", mape_get_node_level(node)) ;
	fprintf(mapeout, "\t          mape : %-p\n", node->mape) ;
	fprintf(mapeout, "\t            C0 : %-d\n", mape_get_node_C0(node)) ;
	fprintf(mapeout, "\t            C1 : %-d\n", mape_get_node_C1(node)) ;
	fprintf(mapeout, "\t     fanins(%d) : ", node->nin) ;
	/*dump the fanins, ten one line */
	i=0 ;
	foreach_fanin(node, i, p){
		if(i%10==0 && i!=0)
			fprintf(mapeout, "\n\t                 ") ; 
		fprintf(mapeout, "%s", p->name) ; 
		if(i < node->nin - 1)
			fprintf(mapeout, ",") ;
	}
	fprintf(mapeout, "\n") ;
	/*dump the fanouts, ten one line */
	i=0 ;
	length = lsLength(node->fanout) ;
	fprintf(mapeout, "\t   fanouts(%2d) : ", length) ;
	foreach_fanout(node, gen, p){
		if(i%10==0 && i!=0)
			fprintf(mapeout, "\n\t                 ") ; 
		fprintf(mapeout, "%s", p->name) ; 
		if(i < length - 1)
			fprintf(mapeout, ",") ;
	}
	fprintf(mapeout, "\n") ;

	/*dump the mape value, ten one line */  
	if(node->mape == NULL){  //not allocated yet. 
		fprintf(mapeout, "\t     stack(%2d) : ", 0) ;
	} else {
		i=0 ;
		stack = ((mape_t *)(node->mape))->values ;
		length = lsLength(stack) ;
		fprintf(mapeout, "\t     stack(%2d) : ", length) ;
		lsForeachItem(stack, gen, imp_value){
			if(i%10 == 0 && i!=0)
				fprintf(mapeout, "\n\t                 ") ; 
			fprintf(mapeout, "%s", mape_node_value_string(imp_value->value, string)) ; 
			fprintf(mapeout, "(%s=%s)", imp_value->pi->name, 
				mape_node_value_string(imp_value->pi_value, string)) ; 
			if(i < length - 1)
				fprintf(mapeout, ",") ;
			i++ ;
		}
	}
	/* now dumping the cube of the node*/
	fprintf(mapeout, "\n") ;
	/*on-set cube*/
	if(node->F){
		fprintf(mapeout, "\t      cubs(ON) : ") ;
		i=0 ;
		foreach_set(node->F, last_ps, ps){
			if(i%10 == 0 && i!=0)
				fprintf(mapeout, "\n\t                 ") ; 
			dump_set(node->nin, ps);
			if(ps != last_ps)
				fprintf(mapeout, ",") ;
			i++ ;
		}
	}

	/*off-set cube*/
	if(node->R){
		fprintf(mapeout, "\t     cubs(OFF) : ") ;
		i=0 ;
		foreach_set(node->R, last_ps, ps){
			if(i%10 == 0 && i!=0)
				fprintf(mapeout, "\n\t                 ") ; 
			dump_set(node->nin, ps);
			if(ps != last_ps)
				fprintf(mapeout, ",") ;
			i++ ;
		}
	}
	fprintf(mapeout, "\n") ;
	fflush(mapeout) ;
}

/*
 * dump_network
 *
 * This function is used for debugging. 
 * It will dump all the information in the network. 
 */
void 
dump_network(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;

	fprintf(mapeout, "\n******** Dumping the Network %-5s***************\n", network->net_name) ; 
	fprintf(mapeout, "\t         #Node : %-d\n", lsLength(network->nodes)) ;
	fprintf(mapeout, "\t           #PI : %-d\n", lsLength(network->pi)) ;
	fprintf(mapeout, "\t           #PO : %-d\n", lsLength(network->po)) ;
	fprintf(mapeout, "----------- Dumping the Nodes -------------------\n") ;

	foreach_node(network, gen, node)
		dump_node(node) ;

	/*
	fprintf(mapeout, "-------- Dumping the Primary Input --------------\n") ;
	foreach_primary_input(network, gen, node)
		dump_node(node) ;
	fprintf(mapeout, "-------- Dumping the Primary Output -------------\n") ;
	foreach_primary_output(network, gen, node)
		dump_node(node) ;
	*/
	fprintf(mapeout, "********** End of Dumping Network ***************\n") ; 
}

/*
 * mape_report
 *
 * The final report for the 
 * result. 
 */
void 
mape_report(network, verb, time)
network_t *network ;
int verb ;
long time ;
{
	lsGen gen ;
	node_t *p ;
	node_type_t type ;
	long tTF = 0, aTF = 0 ;
	float per ;
	int count = 0 ;

	if(verb){
		fprintf(mapeout, "Maximum power estimation for circuit %s by ATPG-PODEM\n\n", network->net_name) ;
		fprintf(mapeout, "%20s : %-10s%-10s%-10s\n", "Name","Type", "Value", "#Fanout") ;
	}
	/* the primary input */
	foreach_primary_input(network, gen, p){
		/*
		if(!mape_is_given_node(p))
			continue ;
		*/
		if(verb)
			fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name,"PI", mape_print_node_value(mape_top(p)), LSLENGTH(p->fanout)) ;
		/*
		tTF += LSLENGTH(p->fanout);
		aTF += (TRANSITION_VALUE(mape_top(p))?1:0) * LSLENGTH(p->fanout) ;
		*/
		count ++ ;
		if(verb && count %5 == 0)
			fprintf(mapeout, "\n") ;
	}

	/* the primary output */
	foreach_primary_output(network, gen, p){
		if(!mape_is_given_node(p))
			continue ;
		if(verb)
			fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name,"PO", mape_print_node_value(mape_top(p)), LSLENGTH(p->fanout)) ;
		/*
		tTF += LSLENGTH(p->fanout);
		aTF += (TRANSITION_VALUE(mape_top(p))?1:0) * LSLENGTH(p->fanout) ;
		*/
		count ++ ;
		if(verb && count %5 == 0)
			fprintf(mapeout, "\n") ;
	}

	/* internal node */
	foreach_node(network, gen, p){
		if(!mape_is_given_node(p))
			continue ;
		type = node_type(p) ;
		if(type != PRIMARY_INPUT && type != PRIMARY_OUTPUT){
			if(verb)
				fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name, mape_print_node_type(type), 
					mape_print_node_value(mape_top(p)), LSLENGTH(p->fanout)) ;
			tTF += LSLENGTH(p->fanout);
			aTF += (TRANSITION_VALUE(mape_top(p))?1:0) * LSLENGTH(p->fanout) ;
			count ++ ;
			if(verb && count %5 == 0)
				fprintf(mapeout, "\n") ;
		}
	}

	/* static calculation */
	per = (100.0 * aTF)/(float)(tTF) ;
	fprintf(mapeout, "\nTotal #Node = %d\n", network_num_internal(network)) ;
	fprintf(mapeout, "Total Tansition * #Fanout = %d\n", tTF) ; 
	fprintf(mapeout, "Assigned Tansition * #Fanout = %ld (%2.1f%%)\n", aTF, per) ; 
	if(verb)
		fprintf(mapeout, "Time %s\n", util_print_time(time)) ;
	fflush(mapeout) ;
}

/*
 * mape_rdm_report
 *
 * The final report for the mape-random
 * result. 
 */
void 
mape_rdm_report(network, verb, time)
network_t *network ;
int verb ;
long time ;
{
	lsGen gen ;
	node_t *p ;
	node_type_t type ;
	long tTF = 0, aTF = 0 ;
	float per ;
	int count = 0 ;

	/* the primary input */
	if(verb){
		fprintf(mapeout, "Maximum power estimation for circuit %s by random input\n\n", network->net_name) ;
		fprintf(mapeout, "%20s : %-10s%-10s%-10s\n", "Name","Type", "Value", "#Fanout") ;
	}
	foreach_primary_input(network, gen, p){
		if(verb)
			fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name,"PI", 
				mape_print_node_value(mape_rdm_value_max(p)), LSLENGTH(p->fanout)) ;
		/*
		tTF += LSLENGTH(p->fanout);
		aTF += (TRANSITION_VALUE(mape_rdm_value_max(p))?1:0) * LSLENGTH(p->fanout) ;
		*/
		count ++ ;
		if(verb && count %5 == 0)
			fprintf(mapeout, "\n") ;
	}

	/* the primary output */
	foreach_primary_output(network, gen, p){
		if(verb)
			fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name,"PO", 
				mape_print_node_value(mape_rdm_value_max(p)), LSLENGTH(p->fanout)) ;
		/*
		tTF += LSLENGTH(p->fanout);
		aTF += (TRANSITION_VALUE(mape_rdm_value_max(p))?1:0) * LSLENGTH(p->fanout) ;
		*/
		count ++ ;
		if(verb && count %5 == 0)
			fprintf(mapeout, "\n") ;
	}

	/* internal node */
	foreach_node(network, gen, p){
		type = node_type(p) ;
		if(type != PRIMARY_INPUT && type != PRIMARY_OUTPUT){
			if(verb)
				fprintf(mapeout, "%20s : %-10s%-10s%-10d\n", p->name, mape_print_node_type(type), 
						mape_print_node_value(mape_rdm_value_max(p)), LSLENGTH(p->fanout)) ;
			tTF += LSLENGTH(p->fanout);
			aTF += (TRANSITION_VALUE(mape_rdm_value_max(p))?1:0) * LSLENGTH(p->fanout) ;
			count ++ ;
			if(verb && count %5 == 0)
				fprintf(mapeout, "\n") ;
		}
	}

	/* static calculation */
	per = (100.0 * aTF)/(float)(tTF) ;
	fprintf(mapeout, "\nTotal #Node = %d\n", lsLength(network->nodes)) ;
	fprintf(mapeout, "Total Tansition * #Fanout = %d\n", tTF) ; 
	fprintf(mapeout, "Assigned Tansition * #Fanout = %ld (%2.1f%%)\n", aTF, per) ; 
	if(verb)
		fprintf(mapeout, "Time %s\n", util_print_time(time)) ;
	fflush(mapeout) ;
}
