/*
 * implication.c 
 *
 * this file is used to provide a utility to finsh a 
 * calculation in the whole circuit using levelization
 * order of the circuit. 
 *
 * Initializly, you should reset the imp utility, 
 * which specify the operation to every node and 
 * clear the imp_list. Then you should provide 
 * some initilization node in the imp_list by function
 * of mape_imp_push. then you use mape_imp_start 
 * to start the process. which recursively select 
 * the lowest node from the imp_list and do the funciton
 * you specifid. If the return value of the function 
 * is FALSE, it will not push any more node into the 
 * imp_list. Otherwise(TRUE), it wll push all the fanout
 * of this node into the imp_list. This process will 
 * end when the imp_list is empty. 
 *
 * This utility could be used to implication, simulation, 
 * backtrack and controllability calculation. 
 *
 * Changbo Long  Nov.15, 2001(have a nice day)
 */

#include "mape_int.h"

static lsList imp_list = NULL ;
static int (*imp_do)() = NULL ;
static char *imp_do_arg = NULL ;

static void mape_imp(void) ;

/*
 * mape_imp_reset
 *
 *
 * reset the implicaiton utilities. 
 */
void 
mape_imp_reset(f, arg)
int (*f)() ;
char *arg ;
{
	node_t *tmp ;

	imp_do = f ;
	imp_do_arg = arg ;


	if(!imp_list)
		return ;
	while(lsLength(imp_list))
		(void)lsDelEnd(imp_list, (lsGeneric *)(&tmp)) ;
}

/*
 * mape_imp_push
 *
 * push a node into the imp_list
 */
void
mape_imp_push(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p ;

	if(node == NULL)
		return ;
	if(imp_list == NULL)
		imp_list = lsCreate() ;

	/* if node contained, return */
	if(lsLength(imp_list) != 0)
		lsForeachItem(imp_list, gen, p){
			if(node == p){
				lsFinish(gen) ;
				return ;
			}
		}

	/* push node into the list */
	lsNewEnd(imp_list, (lsGeneric)node, NULL) ;
}

/*
 * mape_imp_take
 *
 * take the node out which has the minimal level from the imp_list. 
 */
void 
mape_imp_take(node)
node_t **node ;
{
	int level = -1 ;
	int new_level ;
	lsGen gen ;
	lsHandle h, h_t ;
	node_t *p ;

	if(lsLength(imp_list) == 0){
		*node = NULL ;
		return ;
	}

	gen = lsStart(imp_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		new_level = mape_get_node_level(p) ;
		if(level == -1 || new_level < level){
			level = new_level ;
			h_t = h;
		}
	}
	lsFinish(gen) ;
	LS_ASSERT(lsRemoveItem(h_t, (lsGeneric *)node)) ;
}

/*
 * mape_imp_start
 *
 * start the implication processing. 
 */
void
mape_imp_start(void)
{
	mape_imp() ;
}

/*
 * mape_podem_imp_list
 *
 * implcation all the node in the node's array. 
 */
void 
mape_imp(void)
{
	node_t *node ;
	node_t *p ;
	lsGen gen ;
	mape_node_value_t value ; 
	int rtv ;
	int i = 0 ;
	
	if(imp_list == NULL)
		return ;

	if(lsLength(imp_list) == 0) /* no node in the array */
		return ;

	/* take the node out which has the minimal level */
	mape_imp_take(&node) ;
	if(node == NULL)
		return ;

	rtv = (*imp_do)(node, imp_do_arg) ;
	if(!rtv)
		return ;

	foreach_fanout(node, gen, p){
		mape_imp_push(p) ;
	}

	while(lsLength(imp_list) != 0){
		mape_imp() ;
	}
}
