/*
 * mape_msp.c 
 *
 * This file is used to implement the MSP algorithm published in
 *
 *   A matrix synthesis approach to thermal placement
 *     TCAD, Volume: 17 Issue:11 , Nov. 1998 Page(s): 1166 -1174.  
 *
 * Mar. 11, 2002. 
 */
#include "mape_int.h"

typedef struct mape_msp_param_struct {
	int n ;	/* size of the matrix, which contains nXn elements.*/
	int t ;	/* size of the sub-matrix, which contains tXt elements.*/
	char **matrix ; /*The matrix*/
	double (*value)(char *) ; /*The value function used to calculate element's value*/
}mape_msp_param_t ;

mape_msp_param_t *msp_param = NULL;
lsList 	data_list = NULL;

/*
 * mape_msp_clear
 *
 * clear all the data and space used last time. 
 */
void 
mape_msp_clear(void)
{
	if(msp_param){
		if(msp_param->matrix)
			free(msp_param->matrix) ;
		free(msp_param) ;
	}
	msp_param = NULL ;

	if(data_list)	
		lsDestroy(data_list, NULL) ;
	data_list = NULL ;
}

/*
 * mape_msp_set
 *
 * set all the parameters for the MSP algorithm. 
 */
void 
mape_msp_reset(t, value)
int t ;
double (*value)(char *) ;
{
	/* clear all the datas used in last time */
	mape_msp_clear() ;

	msp_param 		= ALLOC(mape_msp_param_t, 1) ;
	msp_param->t		= t ;	
	msp_param->value	= value ;	
	msp_param->matrix 	= NULL ;
	msp_param->n		= 0 ;
}

/*
 * mape_msp_push_data
 *
 * push data into the data's list. 
 */
void 
mape_msp_push_data(data)
char *data ;
{
	if(!data_list)
		data_list = lsCreate() ;
	lsNewEnd(data_list, (lsGeneric)data, LS_NH) ;
}

/*
 * mape_msp_get_matrix_size
 *
 * return the matrix's size. 
 */
int 
mape_msp_get_matrix_size(void)
{
	int r ;
	if(msp_param)
		if(msp_param->n)
			return msp_param->n ;

	r = (int)sqrt((double)lsLength(data_list)) ;
	return (r*r == lsLength(data_list))?r:r+1 ; 
}

/*
 * mape_msp_compare_data
 *
 * compare function for sort.  
 */
int mape_msp_compare_data(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	return (int)( (*(msp_param->value))((char *)obj1), 
		(*(msp_param->value))((char *)obj2) ) ;
}

/*
 * mape_msp_a1
 *
 * the a1 algorithm. 
 */
void 
mape_msp_a1(void)
{
	int i, j ;

	/* get the size of the matrix */
	msp_param->n = mape_msp_get_matrix_size() ;
	msp_param->matrix = ALLOC(char *, (msp_param->n)*(msp_param->n))  ;

	/*Sort the data according to the value */
	lsSort(data_list, mape_msp_compare_data) ; 

	for(i-0;i<msp_param->n;i++) 
		for(j=0;j<msp_param->n;j++){
			
		}
}

/*
 * mape_msp_a2
 *
 * the a2 algorithm. 
 */
void 
mape_msp_a2(void)
{
}

/*
 * mape_msp_a3
 *
 * the recursive algorithm. 
 */
mape_msp_a3(void)
{
}

