/*
 * mape_mspplt.c 
 *
 * This file implements the MSP algorithm published in
 *
 *   A matrix synthesis approach to thermal placement
 *     TCAD, Volume: 17 Issue:11 , Nov. 1998 Page(s): 1166 -1174.  
 *
 * and use it to generate a placement. 
 * 
 * Mar. 11, 2002. 
 */
#include "mape_int.h"

#define PGSP_HEAD "/home/clong/work/sis/sis_uw/mape/src/spice/pg_head.sp"
#define PGSP_TAIL "/home/clong/work/sis/sis_uw/mape/src/spice/pg_tail.sp"


node_t 	**mspplt_array = NULL ;		/*node's array for placement*/
int	mspplt_size ;	
lsList 	mspplt_groups = NULL ;
int 	max_st_id = 0;

void mape_mspplt_dump(void) ;
void mape_mspplt_place_node(node_t *, int, int) ;

/*
 * mape_msp_create_st 
 *
 */
mape_mspplt_st_t *
mape_msp_create_st(step, rp, fp, hp, lp)
int step ;
float rp, fp, hp, lp ;
{
	mape_mspplt_st_t *rtv ;

	rtv		= ALLOC(mape_mspplt_st_t, 1) ;
	rtv->step	= step ;
	rtv->rp		= rp ;
	rtv->fp		= fp ;
	rtv->hp		= hp ;
	rtv->lp		= lp ;

	return rtv ;
}

/*
 * set node's st_id
 * for the conviniet of coprob. 
 */
void 
mape_mspplt_set_st_id(node, id)
node_t *node ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;
	
	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;

	((mape_t *)(node->mape))->plt->st_id = id ;

	if(max_st_id < id + 1)
		max_st_id = id + 1 ;

}

/*
 * get node's st_id
 *
 * return node's st_id. 
 */
int 
mape_mspplt_st_id(node, id)
node_t *node ;
{
	return  ((mape_t *)(node->mape))->plt->st_id  ;
}

/*
 * mape_msp_create_null_node
 *
 * create a null node, which recode it's group.  
 */
node_t *
mape_mspplt_create_null_node(gn)
int gn ;
{
	node_t *node ;

	node = node_alloc() ;

	node->mape = (char *)mape_create_mape_t() ;
	((mape_t *)(node->mape))->plt = mape_create_plt_t() ;
	((mape_t *)(node->mape))->plt->st_list = lsCreate() ;
	((mape_t *)(node->mape))->plt->group = gn ;

	node->short_name = strdup("NA") ;
	node->name = strdup("NA") ;

	return node ;
}

/*
 * mape_mspplt_null_node
 *
 * is the node null node. 
 */
int 
mape_mspplt_null_node(node)
node_t *node ;
{
	if(!node)
		return TRUE ;

	if(!strcmp(node->name, "NA")) 
		return TRUE ;

	return FALSE ;
}

/*
 * mape_mspplt_push_st
 *
 * push the switching time and the value's possibility
 * to the node. 
 */
void 
mape_mspplt_push_st(node, step, rp, fp, hp, lp) 
node_t *node ;
int step ;
float rp, fp, hp, lp ; /*rise, fall, high, low's possibility*/
{
	mape_mspplt_st_t *st ;

	st = mape_msp_create_st(step, rp, fp, hp, lp) ;

	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;

	if(!((mape_t *)(node->mape))->plt->st_list)
		((mape_t *)(node->mape))->plt->st_list = lsCreate() ;

	lsNewEnd(((mape_t *)(node->mape))->plt->st_list, (lsGeneric)st, LS_NH) ;
}

/*
 * mape_mspplt_node_st_possibility
 *
 * node's value's possibility at a particalar switching time
 */
float
mape_mspplt_node_st_possibility(node, step, value)
node_t *node ;
int step ;
mape_node_value_t value ;
{
	lsList st_list ;
	lsGen gen ;
	mape_mspplt_st_t *st ;
	mape_mspplt_st_t *equ ;

	if(!node)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	if(!node->mape){
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;
	}

	if(!((mape_t *)(node->mape))->plt){
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;
	}

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(!st_list)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	equ = NULL ;
	lsForeachItem(st_list, gen, st){
		if(st->step <= step)
			equ = st ;
	}
	if(!equ)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	if(equ->step == step){
		switch(value){
			case MAPE_FALL:
				return equ->fp ;
			case MAPE_RISE:
				return equ->rp ;
			case MAPE_1_1:
				return equ->hp ;
			case MAPE_0_0:
				return equ->lp ;
			default: 
				mape_error("unexpected node value", __FILE__, __LINE__) ;
		}
	}
	if(equ->step < step){
		switch(value){
			case MAPE_FALL:
				return 0.0 ;
			case MAPE_RISE:
				return 0.0 ;
			case MAPE_1_1:
				return equ->rp + equ->hp ;
			case MAPE_0_0:
				return equ->fp + equ->lp ;
			default: 
				mape_error("unexpected node value", __FILE__, __LINE__) ;
		}
	}
}

/*
 * mape_mspplt_index_possibility
 *
 * return the possibility for such index. 
 */
float 
mape_mspplt_index_possibility(node, step, index)
node_t *node ;
int step ;
int *index ;
{
	int i ;
	float rtv=1.0;
	mape_node_value_t v ;
	node_t *p ;

	foreach_fanin(node, i, p){
		switch(index[i]){
			case 0: 
				v = MAPE_FALL ;
				break ;
			case 1:
				v = MAPE_RISE ;
				break ;
			case 2:
				v = MAPE_1_1 ;
				break ;
			case 3: 
				v = MAPE_0_0 ;
				break ;
			default:
				mape_error("unexpected index!", __FILE__, __LINE__) ;
		}
		rtv *= mape_mspplt_node_st_possibility(p, step, v) ;
	}
	return rtv ;
}

/*
 * mape_mspplt_node_index_input
 *
 * translate a index to the input's index.  
 */
void 
mape_mspplt_node_index_input(ind, nin, input)
int ind, nin, *input ;
{
	int i, j, k = 1;

	for(i=0;i<nin-1;i++)
		k*=4 ; 
	//fprintf(mapeout, "\t%d=", ind) ;
	for(i=0;i<nin;i++){
		input[i] = ind/k ;
		ind -= input[i]*k ;
		k/=4 ;
		//fprintf(mapeout, "%d", input[i]) ;
	}
	//fprintf(mapeout, "\n") ;
}

/*
 * mape_mspplt_index_simulate_node. 
 *
 * using index to simulate node. 
 */
mape_node_value_t
mape_mspplt_index_simulate_node(node, index)
node_t *node  ;
int *index ;
{
	int i, j = 0 ;
	float rtv=1.0;
	mape_node_value_t v, o ;
	node_t *p ;
	pcube input ;
	int oF, oS ;

	cube.num_vars = node->nin ;
	define_cube_size (node_num_fanin(node)) ;

	input = new_cube() ;
	/* simulate the first half */
	foreach_fanin(node, i, p){
		switch(index[i]){
			case 0: 
				PUTINPUT(input, i, ONE) ; 
				break ;
			case 1:
				PUTINPUT(input, i, ZERO) ; 
				break ;
			case 2:
				PUTINPUT(input, i, ONE) ; 
				break ;
			case 3: 
				PUTINPUT(input, i, ZERO) ; 
				break ;
			default:
				mape_error("unexpected index!", __FILE__, __LINE__) ;
		}
	}
	oF = mape_set_simulate_node(input, node) ;

	/* simulate the second half */
	foreach_fanin(node, i, p){
		switch(index[i]){
			case 0: 
				PUTINPUT(input, i, ZERO) ; 
				break ;
			case 1:
				PUTINPUT(input, i, ONE) ; 
				break ;
			case 2:
				PUTINPUT(input, i, ONE) ; 
				break ;
			case 3: 
				PUTINPUT(input, i, ZERO) ; 
				break ;
			default:
				mape_error("unexpected index!", __FILE__, __LINE__) ;
		}
	}
	oS = mape_set_simulate_node(input, node) ;
	o = mape_value_convert(oF, oS);

	free_cube(input) ;
	return o ;
}

/*
 * mape_mspplt_time_st_possibility. 
 *
 * get value's possibility at a particular switching time. 
 */
void
mape_mspplt_time_st_possibility(node, step, rp, fp, hp, lp)
node_t *node ;
int step ;
float *rp, *fp, *hp, *lp ;
{
	int i, j, t, k ;
	int index[20] ;
	mape_node_value_t o ; 
	float p ;

	/* initially, set all possibility to be zero*/
	*rp = *fp = *hp = *lp = 0.0 ;

	t = 1 ;
	for(i=1;i<=node->nin;i++)
		t *= 4 ;
	for(i=0;i<t;i++){
		mape_mspplt_node_index_input(i, node->nin, index) ;
		p = mape_mspplt_index_possibility(node, step, index) ;
		//fprintf(mapeout, "\t\tpossibility = %f\n", p) ;
		if(p == 0.0)
			continue ;
		o = mape_mspplt_index_simulate_node(node, index) ;
		//fprintf(mapeout, "\t\toutput = %s\n", mape_print_node_value_c(o)) ;
		switch(o){
			case MAPE_FALL: 
				*rp += p ;
				break ;
			case MAPE_RISE:
				*fp += p ;
				break ;
			case MAPE_1_1:
				*hp += p ;
				break ;
			case MAPE_0_0:
				*lp += p ;
				break ;
			case MAPE_U_U:
				mape_error("simulatin error!", __FILE__, __LINE__) ;
		}
	}
}

/*
 * mape_mspplt_list_next_st
 *
 * find the next st for a list of switching time. 
 */
int 
mape_mspplt_list_next_st(list, step)
lsList list ;
int *step ;
{
	int rtv = FALSE ;
	lsGen gen ;
	mape_mspplt_st_t *st ;

	//fprintf(mapeout, "\tlist_next_st:") ;
	lsForeachItem(list, gen, st){
		//fprintf(mapeout, "\t\tst->step=%d\n", st->step) ;
		if(st->step > *step){
			*step = st->step ;
			lsFinish(gen) ;
			rtv = TRUE ;
			break ;
		}
	}
	//fprintf(mapeout, "\n") ;
	return rtv ;
}

/* 
 * mape_mspplt_next_st
 *
 * return the next switching time that we should consider. 
 */
int 
mape_mspplt_next_st(node, step)
node_t *node ;
int *step ;
{
	int i ;	
	node_t *fanin ;
	lsList st_list ;
	int next_step, list_nt ;

	next_step = 9999 ;
	foreach_fanin(node, i, fanin){
		list_nt = *step ;
		if(!(fanin->mape))
			continue ;
		if(!((mape_t *)(fanin->mape))->plt)
			continue ;
		st_list = ((mape_t *)(fanin->mape))->plt->st_list ;
		if(!st_list)
			continue ;
		if(mape_mspplt_list_next_st(st_list, &list_nt))
			next_step = MIN(next_step, list_nt) ;
	}

	if(next_step == 9999)
		return FALSE ;

	*step = next_step ;
	return TRUE ;
}

/*
 * mape_mspplt_st_do
 *
 * for mape_imp utility, get all possible switching time
 * and the value's possibility at that point. 
 */
int 
mape_mspplt_st_do(node, arg)
node_t *node ;
char *arg ; /*no use */
{
	int time_step = -1 ;
	float rp, fp, hp, lp ;
	lib_gate_t *gate ;

	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;

	gate = lib_gate_of(node) ;
	//fprintf(mapeout, "Node: %s(%s)\n", node->name, (gate == NULL?"<none>":lib_gate_name(gate))) ; 
	while(mape_mspplt_next_st(node, &time_step)){
		mape_mspplt_time_st_possibility(node, time_step, &rp, &fp, &hp, &lp) ;
		//fprintf(mapeout, "\tpush st: time=%d, rp=%f, fp=%f, hp=%f, lp=%f\n", time_step+1, rp,fp, hp, lp) ;
		mape_mspplt_push_st(node, time_step+1, rp, fp, hp, lp) ;
	}
	return TRUE ;
}

/*
 * mape_mspplt_node_assign_st_value
 *
 * assign st_value to a node. 
 */
void 
mape_mspplt_node_assign_st_value(node)
node_t *node ;
{
	lsGen gen ;
	lsList st_list   ;
	mape_mspplt_st_t *st, *mst ;
	float rfp = -1.0 ;
	int st_value ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	lsForeachItem(st_list, gen, st){
		if(st->rp + st->fp > rfp){
			mst = st ;
			rfp = st->rp + st->fp ;
		}
	}
	((mape_t *)(node->mape))->plt->st_value = (mst->step)*100 + (int)(rfp*99.0) ;
	//fprintf(mapeout, "node %s's st_value is : %d\n", node->name, ((mape_t *)(node->mape))->plt->st_value) ; 
}

/*
 * mape_mspplt_plt_node_assign_st_value
 *
 * assign st_value to a node. 
 */
void 
mape_mspplt_plt_node_assign_st_value(node, t1, t2, t3)
node_t *node ;
int t1, t2, t3 ;
{
	lsGen gen ;
	lsList st_list   ;
	mape_mspplt_st_t *st, *mst ;
	float rfp = -1.0 ;
	int st_value ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	st_value = 0 ;
	lsForeachItem(st_list, gen, st){
		if(st->step == t1 || st->step == t2 || st->step == t3)
			st_value += (int)((st->rp + st->fp)*100) ;
	}
	((mape_t *)(node->mape))->plt->st_value = st_value ;
	//fprintf(mapeout, "node %s's st_value is : %d\n", node->name, ((mape_t *)(node->mape))->plt->st_value) ; 
}


/*
 * mape_mspplt_assign_st_value
 *
 * assign st_value to every node. 
 */
void 
mape_mspplt_assign_st_value(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		mape_mspplt_node_assign_st_value(node) ;
	}
}

/*
 * mape_mspplt_plt_assign_st_value
 *
 * assign st_value to every node. 
 */
void 
mape_mspplt_plt_assign_st_value(list, t1, t2, t3)
lsList list ;
int t1, t2, t3 ;
{
	lsGen gen ;
	node_t *node ;

	lsForeachItem(list, gen, node)
		mape_mspplt_plt_node_assign_st_value(node, t1, t2, t3) ;
}

/*
 * compact node's st_list. 
 *
 *  
 */
void 
mape_mspplt_compact_st_list(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;
	lsGen gen1 ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
		
	foreach_node(network, gen, node){
		if(!node->mape)
			continue ;
		if(!((mape_t*)(node->mape))->plt)
			continue ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		if(!st_list)
			continue ;
		lsForeachItem(st_list, gen1, st){
			if(st->rp + st->fp <= 0.00001)
				lsDelBefore(gen1, (lsGeneric *)&st) ;
		}
	}
}

/*
 * mape_mspplt_st_possibility
 *
 * get all gate's switching time and possibility
 * under unit delay model. 
 */
void 
mape_mspplt_st_possibility(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_mspplt_st_do, NULL) ;
	foreach_primary_input(network, gen, pi){
		mape_mspplt_push_st(pi, 0, 0.5, 0.5, 0.0, 0.0) ;
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
	mape_mspplt_assign_st_value(network) ;
	mape_mspplt_compact_st_list(network) ;
}

/*
 * mape_mspplt_level_cmp
 *
 * compare two nodes according to the level. 
 */
int
mape_mspplt_level_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	return mape_get_node_level((node_t *)obj1) - mape_get_node_level((node_t *)obj2) ;
}

/*
 * mape_mspplt_st_cmp 
 *
 * compare two nodes according to the st 
 */
int
mape_mspplt_st_value_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	node_t *n1, *n2 ;

	n1 = (node_t *)obj1 ;
	n2 = (node_t *)obj2 ;

	return -(((mape_t *)(n1->mape))->plt->st_value - ((mape_t *)(n2->mape))->plt->st_value) ;
}


/*
 * mape_mspplt_push_group
 *
 * push one group into mspplt_groups. 
 */
void 
mape_mspplt_push_group(group)
lsList group ;
{
	if(!mspplt_groups)
		mspplt_groups = lsCreate() ;
	lsNewEnd(mspplt_groups, (lsGeneric)group, LS_NH) ;
}

/*
 * mape_mspplt_rc_group
 *
 * specify the group index for any r and c. 
 */
int 
mape_mspplt_rc_group(row, col)
int row, col ;
{
	int k, r, c , t;
	
	r = row-1 ;
	c = col-1 ;
	t = FIX_GROUP_SIZE ;

	return (c%t)+(r%t)*t ;
	/*
	for(k=0;k<=t*t-1;k++){
		if(r%t != k/t)
			continue ;
		if(c%t != k%t)
			continue ;
		return k ;
	}
	return -1 ;
	*/
}

/*
 * mape_mspplt_array_node. 
 *
 * return the node in the array node by specifying the 
 * row and col. 
 */
node_t *
mape_mspplt_array_node(row, col)
int row, col ;
{
	return mape_saplt_index_node(row, col) ;
	//return mspplt_array[mspplt_size*(row - 1) + col - 1] ;
}

/*
 * mape_mspplt_can_place_latch
 *
 * can I place a latch in the row and col? 
 */
int 
mape_mspplt_can_place_latch(row, col)
int row, col ;
{
	if(row == mspplt_size || col == mspplt_size)
		return FALSE ;
	return (mape_mspplt_array_node(row, col+1) == NULL &&
		mape_mspplt_array_node(row+1, col) == NULL &&
		mape_mspplt_array_node(row+1, col+1) == NULL) ;
}


/*
 * mape_mspplt_take_group_node. 
 *
 * take one node out from the group. 
 */
node_t *
mape_mspplt_take_group_node(gn, can_latch)
int gn, can_latch ;
{
	int count = 0 ;
	lsList group = NULL ;
	lsGen gen ;
	long v;
	int ind ;
	node_t *rtv ;
	int try ;

	lsForeachItem(mspplt_groups, gen, group){
		if(count == gn){
			lsFinish(gen) ;
			break ;
		}
		count ++ ;
	}
	if(group == NULL)
		return NULL ;
	if(lsLength(group) == 0)
		return NULL ;
	if(lsLength(group) == 1){
		lsFirstItem(group, (lsGeneric *)&rtv, LS_NH) ;
		if(!mape_node_is_latch(rtv) || can_latch){
			lsDelEnd(group, (lsGeneric*)&rtv) ;
			return rtv ;
		}else
			return NULL ;
	}

	try = 0 ;
	for(;;){
		try ++ ;
		if(try == 2*lsLength(group))
			return NULL ;
		v = random() ;
		ind = v%lsLength(group) ;
		count = 0 ;
		lsForeachItem(group, gen, rtv){
			if(count == ind){
				if(!mape_node_is_latch(rtv) || can_latch){
					lsDelBefore(gen, (lsGeneric *)&rtv) ;
					lsFinish(gen) ;
					return rtv ;
				}else{
					lsFinish(gen) ;
					break ;
				}
			}
			count ++ ;
		}
	}
}

/*
 * mape_mspplt_place_node
 *
 * place node 
 */
void 
mape_mspplt_place_node(node, row, col)
node_t *node ;
int row ;
int col ;
{
	/*
	mspplt_array[mspplt_size*(row-1)+col-1] = node ;	
	if(mape_node_is_latch(node)){
		mspplt_array[mspplt_size*(row-1)+col] = node ;	
		mspplt_array[mspplt_size*row+col-1] = node ;	
		mspplt_array[mspplt_size*row+col] = node ;	
	}

	if(node == NULL)
		return ;

	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;

	((mape_t *)(node->mape))->plt->row = row ;
	((mape_t *)(node->mape))->plt->col = col ;
	*/
	mape_saplt_place_node(node, row, col) ;
}

/*
 * mape_mspplt_find_slot. 
 *
 * find a suitable slot for the node. 
 */
int
mape_mspplt_find_slot(node, row, col)
node_t *node ; 
int *row, *col ;
{
	int level, r, c, is_latch ;

	level = mape_get_node_level(node) ;
	is_latch = mape_node_is_latch(node) ;

	/* find suitable node*/
	for(r=1;r<=mspplt_size;r++)
		for(c=1;c<=mspplt_size;c++){
			if(mape_mspplt_array_node(r, c))
				continue ;
			if(is_latch){
				if(mape_mspplt_array_node(r, c+1) == NULL &&
				   mape_mspplt_array_node(r+1, c) == NULL &&
				   mape_mspplt_array_node(r+1, c+1) == NULL){
				   	*row = r ;	
					*col = c ;
					return TRUE;
				}
			}else{
				*row = r ;	
				*col = c ;
				return TRUE ;
			}
		}

	return FALSE ;
}

/*
 * mape_mspplt_assign_residual_node
 *
 * assige the residual node. 
 */
void 
mape_mspplt_assign_residual_node(void)
{
	lsList group = NULL ;
	lsGen gen ;
	int count = 0 ;
	node_t *node ;
	int row, col ;

	lsForeachItem(mspplt_groups, gen, group){
		if(lsLength(group) == 0)
			continue ;
		while(lsLength(group)){
			lsDelEnd(group, (lsGeneric *)&node) ;
			if(mape_mspplt_find_slot(node, &row, &col))
				mape_mspplt_place_node(node, row, col) ;
		}
	}
	/* show the final residual node */
	/*
	lsForeachItem(mspplt_groups, gen, group){
		count ++ ;
		fprintf(mapeout, "Group(%d), lsLength=%d\n", count, lsLength(group)) ; 
	}
	*/
}

/*
 * mape_mspplt_random_group_gate
 *
 * random take a gate out from one group
 */
void
mape_mspplt_random_group_gate(gn, row, col, v1, v2)
int gn, *row, *col ;
long v1, v2 ;
{
	long v ;	
	int r_size, c_size, count ;

	mape_saplt_plt_size(&r_size, &c_size) ;
	count = r_size/FIX_GROUP_SIZE ;

	*row = (gn-1)/FIX_GROUP_SIZE + (v1%count)*FIX_GROUP_SIZE + 1 ;
	*col = (gn-1)%FIX_GROUP_SIZE + (v2%count)*FIX_GROUP_SIZE + 1 ;
	//fprintf(mapeout, "random group gate: gn=%d, %d, %d\n", gn, *row, *col) ;


	/*
	v  = random() ;
	*row = (v%r_size)+1 ;
	v  = random() ;
	*col = (v%r_size)+1 ;

	for(;;){
		if(mape_mspplt_rc_group(*row, *col) == gn)
			return ;
		(*col) ++ ;
		if(*col >= r_size){
			(*row) ++ ;
			*col = 1 ;
			if(*row >= r_size)
				*row = 1 ;
		}
	}
	*/
}

/*
 * mape_mspplt_make_exindex
 *
 * make the exchangeable index . 
 */
void 
mape_mspplt_make_exindex(gn, ex_count, ex_row, ex_col, fix_count, fix_row, fix_col)
int gn, *ex_count, *fix_count;  
int **ex_row, **ex_col, **fix_row, **fix_col ;
{
	int r_size, c_size, ign, i, j ;

	*ex_count = 0 ;
	*fix_count = 0 ;

	mape_saplt_plt_size(&r_size, &c_size) ;

	*ex_row = ALLOC(int, r_size*c_size) ;
	*ex_col = ALLOC(int, r_size*c_size) ;
	*fix_row = ALLOC(int, r_size*c_size) ;
	*fix_col = ALLOC(int, r_size*c_size) ;

	for(i=1;i<=r_size;i++)
		for(j=1;j<=c_size;j++){
			ign = mape_mspplt_rc_group(i, j) + 1;
			if(ign > gn){
				(*ex_row)[*ex_count] = i ;
				(*ex_col)[*ex_count] = j ;
				(*ex_count) ++ ;
			}else{
				(*fix_row)[*fix_count] = i ;
				(*fix_col)[*fix_count] = j ;
				(*fix_count) ++ ;
			}
		}

	/* dump for debugging */
	/*
	fprintf(mapeout, "Information for mape_mspplt_make_exindex: r_size=%d, c_size=%d\n", r_size, c_size) ;
	fprintf(mapeout, "ex_count=%d\n", *ex_count) ;
	for(i=0;i<*ex_count;i++)
		fprintf(mapeout, "\tex_row[%d]=%d, ex_col[%d]=%d\n", i, (*ex_row)[i], i, (*ex_col)[i]) ; 
	fprintf(mapeout, "fix_count=%d\n", *fix_count) ;
	for(i=0;i<*fix_count;i++)
		fprintf(mapeout, "\tfix_row[%d]=%d, fix_col[%d]=%d\n", i, (*fix_row)[i], i, (*fix_col)[i]) ; 
	*/
}

/*
 * mape_mspplt_random_same_matrix
 *
 *
 * find two indexs which has the same 
 * matrix. 
 */
void 
mape_mspplt_random_same_matrix(row1, col1, row2, col2)
int *row1, *col1, *row2, *col2 ;
{
	long v ;
	int ind1, ind2 ;
	int max_size, offset ;

	do{
		v = random() ;
		ind1 = v%FIX_GROUP_GATE ;
		v = random() ;
		ind2 = v%FIX_GROUP_GATE ;
	}while(ind1 == ind2) ;
	*row1 = ind1/FIX_GROUP_SIZE + 1 ;	
	*col1 = ind1%FIX_GROUP_SIZE + 1 ;
	*row2 = ind2/FIX_GROUP_SIZE + 1 ;	
	*col2 = ind2%FIX_GROUP_SIZE + 1 ;

	max_size = plt_size/FIX_GROUP_SIZE ;

	v = random() ;
	offset = v%max_size ;
	*row1 += offset * FIX_GROUP_SIZE ;
	*row2 += offset * FIX_GROUP_SIZE ;

	if(*row1 > plt_size || *row2 > plt_size){
		*row1 -= max_size*FIX_GROUP_SIZE ; 
		*row2 -= max_size*FIX_GROUP_SIZE ; 
	}

	v = random() ;
	offset = v%max_size ;
	*col1 += offset * FIX_GROUP_SIZE ;
	*col2 += offset * FIX_GROUP_SIZE ;

	if(*col1 > plt_size || *col2 > plt_size){
		*col1 -= max_size*FIX_GROUP_SIZE ; 
		*col2 -= max_size*FIX_GROUP_SIZE ; 
	}
}

/*
 * mape_mspplt_random_sam_group
 *
 * randomly get the index for one gate in
 * the same  group ;
 */
void 
mape_mspplt_random_same_group(row1, col1, row2, col2, gn) 
int *row1, *col1, *row2, *col2, gn ;
{
	int ind, max_size, r_size, c_size , offset;
	long v ;

	static int exindex_count = 0, fixindex_count = 0;
	static int *exindex_row = NULL, *exindex_col = NULL ; /* all the exchangable index */
	static int *fixindex_row = NULL, *fixindex_col = NULL ; /* fix group index */

	if(!(exindex_count)&&!(fixindex_count))
		mape_mspplt_make_exindex(gn, &exindex_count, &exindex_row, 
			&exindex_col, &fixindex_count, &fixindex_row, &fixindex_col) ; 
	
	v = random() ;
	if((v%(FIX_GROUP_GATE)+1) > gn){
		ind = v%exindex_count ;
		*row1 = exindex_row[ind] ;
		*col1 = exindex_col[ind] ;
		v = random() ;
		ind = v%exindex_count ;
		*row2 = exindex_row[ind] ;
		*col2 = exindex_col[ind] ;
	}else{
		ind = v%fixindex_count ;
		*row1 = fixindex_row[ind] ;
		*col1 = fixindex_col[ind] ;

		mape_saplt_plt_size(&r_size, &c_size) ;
		max_size = r_size/FIX_GROUP_SIZE ;

		offset = v%max_size ;
		*row2 = *row1 + offset*FIX_GROUP_SIZE ;
		if(*row2 > r_size)
			*row2 -= max_size*FIX_GROUP_SIZE ;

		v = random() ;
		offset = v%max_size ;
		*col2 = *col1 + offset*FIX_GROUP_SIZE ;
		if(*col2 > r_size)
			*col2 -= max_size*FIX_GROUP_SIZE ;
	}
}

/*
 * mape_mspplt_exchange. 
 *
 *
 * exchange two node. 
 */
void 
mape_mspplt_exchange(row1, col1, row2, col2)
int row1, col1, row2, col2 ;
{
	node_t *node1, *node2 ;

	node1 = mape_mspplt_array_node(row1, col1) ;
	node2 = mape_mspplt_array_node(row2, col2) ;
	mape_mspplt_place_node(node2, row1, col1) ;
	mape_mspplt_place_node(node1, row2, col2) ;
}

/*
 * mape_mspplt_redistribute_na_node. 
 *
 */
void 
mape_mspplt_redistribute_na_node(void)
{
	node_t *node ;
	int row, col ;
	int row1, col1 ;

	for(row=1;row<=mspplt_size;row++)
		for(col=1;col<=mspplt_size;col++){
			if(mape_mspplt_array_node(row, col))
				continue ;
			mape_mspplt_random_same_group(&row, &col, &row1, &col1, user_option->mug) ;
			mape_mspplt_exchange(row, col, row1, col1) ;
		}
}

/*
 * mape_mspplt_assign_node. 
 *
 * assign nodes into the mspplt_array. 
 */
void 
mape_mspplt_assign_node(void)
{
	int row, col ;
	int gn ;
	node_t *node ;
	int can_latch ;

	for(row=1;row<=mspplt_size;row++){
		for(col=1;col<=mspplt_size;col++){
			if(mape_mspplt_array_node(row, col))
				continue ;
			gn = mape_mspplt_rc_group(row, col) ;
			can_latch = mape_mspplt_can_place_latch(row, col) ;
			node = mape_mspplt_take_group_node(gn, can_latch) ;
			mape_mspplt_place_node(node, row, col) ;
			/*
			fprintf(mapeout, "After place node=%s, row=%d, col=%d, gn=%d\n", node?node->short_name:"NA", row, col, gn) ;
			mape_mspplt_dump() ;
			*/
		}
	}

	/*
	 * After the first assignment, there are 
	 * still nodes which has not been assigned
	 */
	mape_mspplt_assign_residual_node() ;

	for(row=1;row<=mspplt_size;row++){
		for(col=1;col<=mspplt_size;col++){
			if(mape_mspplt_array_node(row, col))
				continue ;
			gn = mape_mspplt_rc_group(row, col);
			node = mape_mspplt_create_null_node(gn) ;
			mape_mspplt_place_node(node, row, col) ;
		}
	}

	/*
	 * redistribute the NA node. 
	 *
	 */
	 /*
	 fprintf(mapeout, "Array before redistribute\n") ;
	 mape_mspplt_dump() ;
	 */
	 //mape_mspplt_redistribute_na_node() ;
	 /*
	 fprintf(mapeout, "Array after redistribute\n") ;
	 mape_mspplt_dump() ;
	 */
}

/*
 * mape_mspplt_group_node
 *
 * group the node together according to the timing information. 
 */
void 
mape_mspplt_group_node(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen, gen1 ;
	int g_size ;
	lsList node_list ;
	int count = 0 ;
	lsList group = NULL ;
	int res, gn = 1; 


	node_list = lsCreate() ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		lsNewEnd(node_list, (lsGeneric)node, LS_NH) ;
	}
	/* sort the nodes according to the level */
	//lsSort(node_list, mape_mspplt_level_cmp) ;
	/* sort the nodes according to the switching possibility*/
	lsSort(node_list, mape_mspplt_st_value_cmp) ;

	/* build the groups */
	fprintf(mapeout, "node_length=%d, mspplt_size=%d, FIX_GROUP_SIZE=%d\n", lsLength(node_list), mspplt_size, FIX_GROUP_SIZE) ;
	g_size = lsLength(node_list)/(FIX_GROUP_SIZE*FIX_GROUP_SIZE) ;
	res = lsLength(node_list)%(FIX_GROUP_SIZE*FIX_GROUP_SIZE) ;
	lsForeachItem(node_list, gen, node){
		count ++ ;
		if(group == NULL)
			group = lsCreate() ;
		lsNewEnd(group, (lsGeneric)node, LS_NH) ;
		fprintf(mapeout, "node=%s(%d), count = %d, g_size = %d\n", node->name, 
			((mape_t*)(node->mape))->plt->st_value, count, g_size) ;
		if( (gn > res && count == g_size) || (gn <= res && count == g_size+1) ){
			mape_mspplt_push_group(group) ;
			group = NULL ;			
			count = 0 ;
			gn ++ ;
		}
	}
	if(group)
		if(lsLength(group))
			mape_mspplt_push_group(group) ;

	count = 0 ;
	fprintf(mapeout, "The groups(%p): \n", mspplt_groups);
	lsForeachItem(mspplt_groups, gen, group){
		count ++ ;
		fprintf(mapeout, "Group %d(%d): \n", count, lsLength(group)) ;
		lsForeachItem(group, gen1, node){
			fprintf(mapeout, "%s(%d) ", node->name, mape_get_node_level(node)) ;
		}
		fprintf(mapeout, "\n") ;
	}
	lsDestroy(node_list, NULL);
}

/*
 * mape_mspplt_build_size
 *
 * build size of the plt. 
 */
void 
mape_mspplt_build_array(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;
	int space = 0 ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(!mape_node_is_latch(node))
			space += CGATE_AREA  ;
		else 
			space += LATCH_AREA ;
	}
	mspplt_size 	= (int)(sqrt((double)space)) + 1 ;
	mape_saplt_alloc_array(mspplt_size) ;

	//mspplt_array	= ALLOC(node_t *, mspplt_size*mspplt_size) ;
	//memset(mspplt_array, 0, mspplt_size*mspplt_size*sizeof(node_t *)) ;
}

/*
 * mape_mspplt_dump
 *
 * dump the node array. 
 */
void 
mape_mspplt_dump(void)
{
	int i, j ;
	node_t *node ;

	for(i=1;i<=mspplt_size;i++){
		for(j=1;j<=mspplt_size;j++){
			node = mape_mspplt_array_node(i, j) ;
			if(!node)
				fprintf(mapeout, "%-5s", "NA") ;
			else 
				fprintf(mapeout, "%-5s", node->short_name) ;
		}
		fprintf(mapeout, "\n") ;
	}
}

/*
 * mape_mspplt_list_total_st
 *
 * calculate the possibility at each time step in a list. 
 */
void 
mape_mspplt_list_total_st(list, x, max_step)
lsList list ;
double *x ;
int *max_step ;
{
	lsGen gen ;
	mape_mspplt_st_t *st ;
	double rfp ;

	if(list == NULL)
		return ;
	lsForeachItem(list, gen, st){
		if(st->step > *max_step)
			*max_step = st->step ;
		x[st->step] += (st->rp + st->fp) ;
	}
}

/*
 * mape_mspplt_st_first_three
 *
 * first three st value. 
 */
void 
mape_mspplt_st_first_three(x, max_step, t1, t2, t3)
double *x ;
int max_step ;
int *t1, *t2, *t3 ;
{
	double x1, x2, x3 ;
	int i ;

	*t1 = 0 ; x1 = 0 ;
	*t2 = 0 ; x2 = 0 ;
	*t3 = 0 ; x3 = 0 ;

	for(i=0;i<=max_step;i++){
		if(x[i] > x1){
			*t3 = *t2; x3 = x2 ; 
			*t2 = *t1; x2 = x1 ;
			*t1 = i ;  x1 = x[i] ;
			continue ;
		}
		if(x[i] > x2){
			*t3 = *t2; x3 = x2 ; 
			*t2 = i ; x2 = x[i] ;
			continue ;
		}
		if(x[i] > x3){
			*t3 = i ; x3 = x[i] ;
		}
	}
}

/*
 * mape_mspplt_place_node_random
 *
 * place the node randomly. 
 */
void 
mape_mspplt_place_node_random(node) 
node_t *node ;
{
	long v ;
	int row, col ;

	for(;;){
		v = random() ;
		row = v%mspplt_size ;
		v = random() ;
		col = v%mspplt_size ;
		if(!mape_mspplt_array_node(row+1, col+1)){
			mape_mspplt_place_node(node, row+1, col+1) ;
			return ;
		}
	}
}

/*
 * mape_mspplt_place_square
 *
 */
void 
mape_mspplt_place_square(list, l, r, t, d)
lsList list ;
int l, r, t, d ;
{
	node_t *node ;

	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, t, l) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, t, r) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, d, l) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, d, r) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, t, (l+r)/2) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, (t+d)/2, l) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, (t+d)/2, r) ;
	lsDelBegin(list, (lsGeneric *)&node) ;
	mape_mspplt_place_node(node, d, (l+r)/2) ;
}


/*
 * mape_msp_plt_assign_node
 *
 * place node into grid. 
 */
void
mape_mspplt_plt_assign_node(list)
lsList list ;
{
	node_t *node ;
	int length, k  ;
	int row, col  ;
	long v ;

	length = lsLength(list) ;

	while(lsLength(list) >= 3*length/4+5){
		lsDelBegin(list, (lsGeneric *)&node) ;
		fprintf(mapeout, "Place node %s\n", node->short_name) ;
		for(;;){
			v = random() ;
			row = v%(mspplt_size/2) ;
			v = random() ;
			col = v%(mspplt_size/2) ;
			row = (row+1)*2 ;
			col = (col+1)*2 ;
			row = row>mspplt_size?mspplt_size:row ;
			col = col>mspplt_size?mspplt_size:col ;
			if(!mape_mspplt_array_node(row, col)){
				mape_mspplt_place_node(node, row, col) ;
				break ;
			}
		}
	}
	fprintf(mapeout, "After place 1/4\n") ;
	while(lsLength(list) >= length/2 + 20){
		lsDelBegin(list, (lsGeneric *)&node) ;
		fprintf(mapeout, "Place node %s\n", node->short_name) ;
		for(;;){
			v = random() ;
			row = v%(mspplt_size/2) ;
			v = random() ;
			col = v%(mspplt_size/2) ;
			row = (row+1)*2 + 1;
			col = (col+1)*2 + 1;
			row = row>mspplt_size?(row-2):row ;
			col = col>mspplt_size?(col-2):col ;
			if(!mape_mspplt_array_node(row, col)){
				mape_mspplt_place_node(node, row, col) ;
				break ;
			}
		}
	}
	/*place the remaining node randomly */
	while(lsLength(list)){
		lsDelBegin(list, (lsGeneric *)&node) ;
		mape_mspplt_place_node_random(node) ;
	}
}

/*
 * mape_mspplt_plt. 
 *
 * place the nodes in a special way. 
 */
void 
mape_mspplt_plt(network) 
network_t *network ;
{
	double x[256] ;	
	lsGen gen ;
	node_t *node ;
	lsList st_list ;
	int max_step = 0 ;
	int i ;
	int t1, t2, t3 ;
	lsList node_list ;

	/* calculate the possibility at each time step */
	memset(x, 0, sizeof(double)*256) ;
	node_list = lsCreate() ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		mape_mspplt_list_total_st(st_list, x, &max_step) ;
		lsNewEnd(node_list, (lsGeneric)node, LS_NH) ;
	}
	/* find the 3 time steps which has the largest switching possibility */
	mape_mspplt_st_first_three(x, max_step, &t1, &t2, &t3) ;
	fprintf(mapeout, "t1=%d, x1=%.2lf, t2=%d, x2=%.2lf, t3=%d, x3=%.2lf\n", 
		t1, x[t1], t2, x[t2], t3, x[t3]) ;
	mape_mspplt_plt_assign_st_value(node_list, t1, t2, t3) ;
	lsSort(node_list, mape_mspplt_st_value_cmp) ;
	/*
	lsForeachItem(node_list, gen, node){
		fprintf(mapeout, "node %s(%d)\n", node->short_name, ((mape_t *)(node->mape))->plt->st_value) ;
	}
	*/
	mape_mspplt_plt_assign_node(node_list) ;
	mape_mspplt_dump() ;
	lsDestroy(node_list, NULL) ;
}

/*
 * mape_mspplt_list_remove
 *
 * remove the node in list2 from list1.  
 */
void 
mape_mspplt_list_remove(l1, l2)
lsList l1, l2 ;
{
	lsGen gen ;
	node_t *node ;
	node_t *tmp ;

	lsForeachItem(l1, gen, node){
		if(mape_node_in_list(node, l2))
			lsDelBefore(gen, (lsGeneric*)&tmp) ;
	}
}

/*
 * mape_mspplt_another_plt
 *
 * another placement's method. 
 */
void 
mape_mspplt_another_plt(network)
network_t *network ;
{
	double time ;
	lsList switch_list ;
	lsList sort_list , node_list;
	lsGen gen ;
	double total  ;
	node_t *node ;

	switch_list = lsCreate() ;
	node_list = lsCreate();
	sort_list = lsCreate() ;

	mape_curlib_set_unit_mode(TRUE) ;
	node_list = lsCreate() ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		lsNewEnd(node_list, (lsGeneric)node, LS_NH) ;
	}

	/* construct the sorted list */
	while(lsLength(node_list) >= FIX_GROUP_GATE){
		mape_sassc_group_max_ssc(node_list, UNIT_DELAY_MODEL, &time, switch_list);
		lsForeachItem(switch_list, gen, node)
			lsNewEnd(sort_list, (lsGeneric)node, LS_NH) ;
		mape_mspplt_list_remove(node_list, switch_list) ;
	}
	lsForeachItem(node_list, gen, node)
		lsNewEnd(sort_list, (lsGeneric)node, LS_NH) ;
	mape_mspplt_plt_assign_node(sort_list) ;
	mape_mspplt_dump() ;
	lsDestroy(node_list, NULL) ;
	lsDestroy(switch_list, NULL) ;
	lsDestroy(sort_list, NULL) ;
}

/*
 * mape_mspplt_st_max_step
 *
 *
 * maximum step ;
 */
int 
mape_mspplt_st_max_step(node_list, g_size)
lsList node_list ;
int g_size ;
{
	int m_step ;
	int max_time ;
	lsGen gen, gen1 ;
	int x[256], i, j ;
	node_t *node ;
	lsList st_list ;
	mape_mspplt_st_t *st ;

	max_time = 0 ;

	memset(x, 0, sizeof(int)*256) ;
	for(i=0;i<256;i++){
		/* no step any more */
		if(max_time && i > max_time)
			break ;
		/*calculate st_value for every gate */
		lsForeachItem(node_list, gen, node){
			((mape_t*)(node->mape))->plt->st_value = 0 ;
			st_list = ((mape_t *)(node->mape))->plt->st_list ;
			lsForeachItem(st_list, gen1, st){
				if(i == 0 && st->step > max_time)
					max_time = st->step ;
				if(st->step == i) {
					((mape_t *)(node->mape))->plt->st_value = (int)(1000*(st->rp + st->fp)) ;
					lsFinish(gen1) ;
					break ;
				}
			}
		}
		lsSort(node_list, mape_mspplt_st_value_cmp) ;
		j = 0 ;

		/*get ith total first g_size's value*/
		lsForeachItem(node_list, gen1, node){
			x[i] += ((mape_t *)(node->mape))->plt->st_value ;
			j++  ;
			if(j>=g_size){
				lsFinish(gen1) ;
				break ;
			}
			
		}
	}
	m_step = 0  ;
	for(i=0;i<256;i++){
		//fprintf(mapeout, "\t\tx[%d]=%d\n",i, x[i]) ;  
		if(x[i] > x[m_step])
			m_step = i ;
	}
	return m_step ;
}

/*
 * mape_mspplt_st_coprob
 *
 * get the coprobability for two node. 
 */
int 
mape_mspplt_st_coprob(n1, n2)
node_t *n1, *n2 ;
{
	lsGen gen1, gen2 ;
	lsList l1, l2 ;
	int rtv = 0 ;
	int next1, next2 ;
	mape_mspplt_st_t *st1, *st2 ;
	int i, j ;
	int id1, id2 ;
	static int *prob_array = NULL ;

	if(!n1->mape || !n2->mape)
		return 0 ;

	if(((mape_t*)(n1->mape))->plt == NULL || ((mape_t*)(n2->mape))->plt == NULL)
		return 0 ;

	if(prob_array == NULL){
		prob_array = ALLOC(int, max_st_id * max_st_id) ;
		for(i=0;i<max_st_id;i++)
			for(j=0;j<max_st_id;j++)
				prob_array[i*max_st_id+j] = -1 ;
	}
	id1 = mape_mspplt_st_id(n1) ;
	id2 = mape_mspplt_st_id(n2) ;
	if(prob_array[id1*max_st_id + id2] != -1)
		return prob_array[id1*max_st_id+id2] ;
	
	l1 = ((mape_t *)(n1->mape))->plt->st_list ;
	l2 = ((mape_t *)(n2->mape))->plt->st_list ;

	if(l1 == NULL || l2 == NULL)
		return 0 ;

	gen1 = lsStart(l1) ;
	gen2 = lsStart(l2) ;

	next1 = TRUE ;
	next2 = TRUE ;

	while(TRUE){
		if(next1)
			if(lsNext(gen1, (lsGeneric *)&st1, LS_NH) != LS_OK)
				break ;
		if(next2)
			if(lsNext(gen2, (lsGeneric *)&st2, LS_NH) != LS_OK)
				break ;
		if(st1->step == st2->step){
			rtv += (int)(1000*(st1->rp + st1->fp))*(1000*(st2->rp+st2->fp)) ;
			next1 = TRUE ;
			next2 = TRUE ;
		}else if(st1->step > st2->step){
			next1 = FALSE ;
			next2 = TRUE ;
		} else {
			next1 = TRUE ;
			next2 = FALSE ;
		}
	}

	lsFinish(gen1) ;
	lsFinish(gen2) ;

	prob_array[id1*max_st_id+id2] = rtv ;
	prob_array[id2*max_st_id+id1] = rtv ;
	return rtv ;
}

/*
 * mape_mspplt_st_group_coprob
 *
 * copobability for one node to a group. 
 */
int
mape_mspplt_st_group_coprob(node, group)
node_t *node ;
lsList group ;
{
	lsGen gen ;
	node_t *p ;
	int rtv = 0 ;

	lsForeachItem(group, gen, p){
		rtv += mape_mspplt_st_coprob(node, p) ;
	}
	
	return rtv ;
}

/*
 * mape_mspplt_st_scg
 *
 * strong connected group ; 
 * keep gn strong connected nodes in group
 * and transfer the rest to node_list. 
 */
void 
mape_mspplt_st_scg(group, node_list, gn)
lsList group, node_list ;
int gn ;
{
	node_t *node, *p ;
	int count = 0 ;
	lsGen gen ;

	while(lsLength(group) > gn){
		if(count == 0){
			lsForeachItem(group, gen, node)
				((mape_t*)(node->mape))->plt->st_value = mape_mspplt_st_group_coprob(node, group) ;
		}
		lsSort(group, mape_mspplt_st_value_cmp) ;
		lsDelEnd(group, (lsGeneric *)&node) ;
		lsNewBegin(node_list, (lsGeneric)node, LS_NH) ;

		lsForeachItem(group, gen, p)
			((mape_t *)(node->mape))->plt->st_value -= mape_mspplt_st_coprob(p, node) ;
		count ++  ;
	}
}

/*
 * mape_mspplt_take_group_nodes_c
 *
 *
 * complex version for taking group nodes out from node_list 
 */
void 
mape_mspplt_st_take_group_nodes_c(node_list, group, g_size)
lsList node_list, group ;
int g_size ;
{
	lsGen gen, gen1 ;
	node_t *node ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	int count ;

	if(lsLength(node_list) <= g_size){
		while(lsLength(node_list)){
			lsDelEnd(node_list, (lsGeneric *)&node) ;
			lsNewEnd(group, (lsGeneric)node, LS_NH) ;
		}
		return ;
	}

	/* sorting by the switching possibility */
	lsForeachItem(node_list, gen, node){
		((mape_t*)(node->mape))->plt->st_value = 0 ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st){
			((mape_t *)(node->mape))->plt->st_value += (int)(1000*(st->rp + st->fp)) ;
		}
	}
	lsSort(node_list, mape_mspplt_st_value_cmp) ;

	/* take the first 2*/
	count = 0 ;
	while(lsLength(node_list)){
		/*
		count++ ;
		if(count >= 12*g_size)
			break ;
		*/
		lsDelBegin(node_list, (lsGeneric *)&node) ;
		lsNewEnd(group, (lsGeneric)node, LS_NH) ;
	}
	mape_mspplt_st_scg(group, node_list, g_size) ;

	/*
	fprintf(mapeout, "take group : \n") ;
	lsForeachItem(group, gen, node){ 
		fprintf(mapeout, "Node: %s\n", node->name) ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st){
			if(st->rp + st->fp > 0.0)
				fprintf(mapeout, "\t%d: %e\n", st->step, (st->rp + st->fp)) ;   
		}
	}
	*/
}


/*
 * mape_mspplt_take_group_nodes. 
 *
 * take a group of nodes out
 */
void 
mape_mspplt_st_take_group_nodes(node_list, group, g_size)
lsList node_list, group ;
int g_size ;
{
	double x[256] ;
	lsGen gen, gen1 ;
	node_t *node ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	int m_step ;
	int i, j;

	if(lsLength(node_list) <= g_size){
		while(lsLength(node_list)){
			lsDelEnd(node_list, (lsGeneric *)&node) ;
			lsNewEnd(group, (lsGeneric)node, LS_NH) ;
		}
		return ;
	}

	/* get the maximum step */
	memset(x, 0, sizeof(double)*256) ;
	m_step = mape_mspplt_st_max_step(node_list, g_size) ;


	//fprintf(mapeout, "mstep=%d\n", m_step) ;
	/* assign st_value for sorting */
	lsForeachItem(node_list, gen, node){
		((mape_t*)(node->mape))->plt->st_value = 0 ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st){
			if(st->step == m_step) {
				((mape_t *)(node->mape))->plt->st_value += (int)(1000*(st->rp + st->fp)) ;
			}
		}
	}
	lsSort(node_list, mape_mspplt_st_value_cmp) ;
	i = 0 ;
	while(i < g_size){
		lsDelBegin(node_list, (lsGeneric *)&node) ;
		lsNewEnd(group, (lsGeneric)node, LS_NH) ;
		//fprintf(mapeout, "\tpush node %s, st_value=%d\n", node->name, ((mape_t*)(node->mape))->plt->st_value) ;
		i ++ ;
	}
}

/*
 * mape_mspplt_st_group_node
 *
 * group node according to the st value. 
 */
void 
mape_mspplt_st_group_node(network)
network_t *network ;
{
	int g_size ;
	int res, gn = 0 ;
	lsList node_list ;
	lsGen gen ;
	node_t *node ;
	lsList group = NULL ;
	int id = 0 ;

	node_list = lsCreate() ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		lsNewEnd(node_list, (lsGeneric)node, LS_NH) ;
		mape_mspplt_set_st_id(node, id) ;
		id ++ ;
	}

	g_size = lsLength(node_list)/(FIX_GROUP_SIZE*FIX_GROUP_SIZE) ;
	res = lsLength(node_list)%(FIX_GROUP_SIZE*FIX_GROUP_SIZE) ;

	while(lsLength(node_list)){
		group = lsCreate() ;
		if(gn <= res)
			mape_mspplt_st_take_group_nodes_c(node_list, group, g_size + 1) ;
		else 
			mape_mspplt_st_take_group_nodes_c(node_list, group, g_size) ;

		lsForeachItem(group, gen, node)
			((mape_t *)(node->mape))->plt->group = gn ;

		mape_mspplt_push_group(group) ;
		gn ++ ;
	}
	lsDestroy(node_list, NULL) ;
}

/*
 * mape_mspplt_dump_st
 *
 * dump the st's information. 
 */
void 
mape_mspplt_dump_st(network) 
network_t *network ;
{
	lsGen gen, gen1 ;
	node_t *node ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	double x[256] ;
	int c[256] ;
	int i ;
	
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		fprintf(mapeout, "Node: %s\n", node->name) ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st){
			if(st->rp + st->fp > 0.0)
				fprintf(mapeout, "\t%d: %e\n", st->step, (st->rp + st->fp)) ;   
		}
	}

	memset(x, 0, sizeof(double)*256) ;
	memset(c, 0, sizeof(int)*256) ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st){
			x[st->step] += (st->rp + st->fp) ;
			if(st->rp + st->fp >= 0.001)
				c[st->step] ++ ;
		}
	}
	for(i=0;i<256;i++)
		fprintf(mapeout, "step=%d, x[%d]=%.2lf, c[%d]=%d\n", i, i, x[i], i, c[i]) ; 
}

/*
 * mape_mspplt_node_exchangeable.  
 *
 * if these two node are exchangeable. 
 */
int 
mape_mspplt_node_exchangeable(row1, col1, row2, col2, mug)
int row1, col1, row2, col2, mug ;
{
	node_t *node1 ;
	node_t *node2 ;
	int group1, group2 ;

	/*in the same submatrix, it is exchangeable */
	if(((row1-1)/FIX_GROUP_SIZE == (row2-1)/FIX_GROUP_SIZE) &&
	   ((col1-1)/FIX_GROUP_SIZE == (col2-1)/FIX_GROUP_SIZE))
		return TRUE ;
	
	node1 = mape_saplt_index_node(row1, col1) ;
	node2 = mape_saplt_index_node(row2, col2) ;

	group1 = ((mape_t *)(node1->mape))->plt->group ;
	group2 = ((mape_t *)(node2->mape))->plt->group ;

	if(group1 == group2 || (group1 >= mug && group2 >= mug))
		return TRUE ;

	return FALSE ;
}

/*
 * mape_mspplt_move. 
 *
 */
void 
mape_mspplt_move(from, to)
char *from, *to ; /* no use argument */
{
	int row1, col1, row2, col2 ;
	long v ;

	do {
		v = random() ;
		row1 = v%plt_size + 1 ;
		v = random() ;
		col1 = v%plt_size + 1 ;
		v = random() ;
		row2 = v%plt_size + 1 ;
		v = random() ;
		col2 = v%plt_size + 1 ;
	}while(!mape_mspplt_node_exchangeable(row1, col1, row2, col2, user_option->mug)) ;
	mape_saplt_exchange_index(row1, col1, row2, col2) ;
	//fprintf(mapeout, "mape_mspplt_move: (%d, %d)->(%d, %d)\n", row1, col1, row2, col2) ;
	mape_saplt_set_exindex(row1, col1, row2, col2) ;

	/*
	mape_mspplt_random_same_group(&row1, &col1, &row2, &col2, user_option->mug) ;
	//fprintf(mapeout, "mape_mspplt_move: (%d, %d)->(%d, %d)\n", row1, col1, row2, col2) ;
	mape_saplt_exchange_index(row1, col1, row2, col2) ;
	mape_saplt_set_exindex(row1, col1, row2, col2) ;
	*/
}

/*
 * mape_mspplt_matrix_move
 *
 * change node within the same matrix. 
 */
void 
mape_mspplt_matrix_move(from, to)
char *from, *to ; /* no use argument */
{
	int row1, col1, row2, col2 ;

	mape_mspplt_random_same_matrix(&row1, &col1, &row2, &col2) ;
	mape_saplt_exchange_index(row1, col1, row2, col2) ;
	//fprintf(mapeout, "matrix_move : %d, %d, %d, %d\n", row1, col1, row2, col2) ;
	mape_saplt_set_exindex(row1, col1, row2, col2) ;
}

/*
 * mape_saplt_sa_equilibrum
 *
 * equilibrum
 */
int 
mape_mspplt_sa_equilibrum(up, total)
int up, total ;
{
	if(up >= mspplt_size*500 || total >= mspplt_size*1000)
		return TRUE ;
	else 
		return FALSE;
}

/*
 * mape_mspplt_design_pg
 * 
 * design the power/ground network 
 */
void 
mape_mspplt_design_pg(void)
{
	int r_size, c_size ;
	int t_row, t_col ;
	int i, j, index, r_index ;

	mape_print_file(PGSP_HEAD) ;

	mape_saplt_plt_size(&r_size, &c_size) ;
	t_row = r_size/FIX_GROUP_SIZE ;
	t_col = c_size/FIX_GROUP_SIZE ;

	t_row += (t_row * FIX_GROUP_SIZE == r_size ? 0 : 1) ;
	t_col += (t_row * FIX_GROUP_SIZE == r_size ? 0 : 1) ;

	fprintf(mapeout, "num_node %d\n", t_row * t_col+1) ;
	fprintf(mapeout, "matrix_size %d\n\n", t_row * t_col+2) ;


	r_index = 1 ;
	for(j=0;j<t_row;j++)
		for(i=1;i<=t_col;i++){
			index = j*t_row + i ; 
			if(i != t_col)
				fprintf(mapeout, "R%d %d %d 1.68 0.8 13.6 37\n", r_index++, index, index+1) ;
			if(j != t_row-1)
				fprintf(mapeout, "R%d %d %d 1.744 0.8 13.6 37\n", r_index++, index, index+t_col) ; 
			fprintf(mapeout, "I%d %d 0 %d\n", index, index, 0) ;  
		}
	fprintf(mapeout, "R%d %d %d 1.696 0.8 13.6 37\n", t_row*t_col+1, t_row*t_col+1, t_row*t_col) ; 
	fprintf(mapeout, "V1 %d 0 3", t_row*t_col + 1) ;  

	mape_print_file(PGSP_TAIL) ;
}



/*
 * mape_mspplt
 *
 * implement the MSP algorithm and use it to construct 
 * the placement. 
 */
void 
mape_mspplt(network)
network_t *network ;
{
	char *sol ;
	double assc, mssc ;

	/*build the array and specify the size*/
	mape_timer_reset() ;
	mape_timer_start() ;
	mape_mspplt_build_array(network) ;

	/*calculate the all possible switching time for each gate 
	and the switching possibility */
	mape_mspplt_st_possibility(network) ;
	//mape_mspplt_dump_st(network) ;

	/* group nodes together according to the switching time*/
	mape_mspplt_st_group_node(network) ;
	mape_mspplt_assign_node() ;
	mape_timer_stop() ;
	mape_timer_report("runing time") ;
	return ;

	/*
	mape_mspplt_design_pg() ;
	return  ;
	*/

	/* calculate the cost */

	/* dump the array */
	//mape_saplt_dump() ;

	//mape_saplt_init_plt(network) ;

	//fprintf(mapeout, "before sa_set\n") ;
	mape_sa_set(80, 0.9, mape_saplt_sa_create, mape_saplt_sa_destroy, 
		mape_saplt_cost, mape_mspplt_move, mape_saplt_sa_equilibrum, 
		mape_saplt_sa_stop, mape_saplt_sa_accept_s) ;
	mape_sa_start(&sol) ;

	//mape_mspplt_dump_st(network) ;

	/*split the nodes into different group according to the timing information */
	//mape_mspplt_group_node(network) ;

	/* assign the node in different group into array*/
	//mape_mspplt_assign_node() ;

	/* try to place the nodes by special way */
	//A2 mape_mspplt_plt(network) ;

	/* another placement scheme */ 
	//A3 mape_mspplt_another_plt(network) ;

	/* calculate the cost */
	fprintf(mapeout, "Internal node #%d, PI #%d, PO #%d\n", network_num_internal(network), network_num_pi(network),   
		network_num_po(network)) ;
	mape_saplt_cal_ssc(network, &assc, &mssc, FIX_GROUP) ;
	fprintf(mapeout, "After MSP+SAPLT TWL=%lf, ASSC(MSG)=%lf, MSSC(MSG)=%lf\n", mape_saplt_wire_length((char *)network), 
			assc, mssc) ;
	/* dump the array */
	//mape_saplt_dump() ;
}
