/*
 * mape_optgrp.c 
 *
 * create optimized group which will generate totally 
 * minimized current. 
 *
 * Changbo Long(Feb. 27, 2002)
 */
#include "mape_int.h"

lsList optgrps ;

/*
 * iteratively improve the solution.  
 */
void 
mape_optgrp_iter(network)
network_t *network ;
{
}

/*
 * return node's transistion start time. 
 *
 */
void 
mape_optgrp_node_trans_time(node, time)
node_t *node ;
double *time ;
{
	int i ;
	node_t *fanin ;
	double ntime ;
	*time = 0.0 ;

	foreach_fanin(node, i, fanin){
		ntime = mape_cur_simulation_hv_time(fanin, 0.0) ;
		if(ntime > *time)
			*time = ntime ;
	}
}

/*
 * mape_optgrp_collect_node_data
 *
 * collect the node's data. 
 */
void 
mape_optgrp_collect_node_data(node)
node_t *node;
{
	double ctime = 0.0 ;

	if(node_type(node) == PRIMARY_INPUT)
		return  ;

	mape_cur_push_vol_point(node, MAX_VOLTAGE, 0.0, 1) ;

	/*get node's transition start time, ignore the logic*/
	if(mape_node_is_latch(node))
		ctime = (double)(CLOCK_CYCLE/2 + HALF_SLOP_TIME) ;		
	else 
		mape_optgrp_node_trans_time(node, &ctime) ;

	mape_cur_push_vol_point(node, MAX_VOLTAGE, ctime, 1) ;
	mape_cur_push_vol_point(node, 0.0, ctime + 2*mape_cur_node_delay(node, FALL_DIRECTION), 1);
}

/*
 * mape_optgrp_init_do
 *
 * do the implication. 
 */
int 
mape_optgrp_init_do(node, arg)
node_t *node ;
char *arg ; /* no use here */
{
	mape_optgrp_collect_node_data(node) ;
	return TRUE ;
}

/*
 * mape_optgrp_init_data
 *
 * get the time delay and switch current for 
 * each node. 
 */
void 
mape_optgrp_init_data(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ; 
	int i = 0 ;

	mape_imp_reset(mape_optgrp_init_do, (char *)NULL) ;
	foreach_primary_input(network, gen, pi){
		mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
		mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
}

/*
 * mape_optgrp_dis_nodes_equ
 *
 * distribute the nodes into optgrp 
 */

/*
 * mape_optgrp_take_mcn
 *
 * take the maximum point contributor out from list all_nodes to list dis_nodes. 
 */
void 
mape_optgrp_take_mcn(all_nodes, dis_nodes) 
lsList all_nodes, dis_nodes ;
{
}

/*
 * initially devide the circuit into 
 * many groups.  
 */
void 
mape_optgrp_init_solution(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;
	lsList all_nodes, dis_nodes ;

	/*initialize all node's data*/
	mape_optgrp_init_data(network) ;

	/*
	mape_cur_dump() ;
	*/
	dis_nodes = lsCreate() ;

	/*copy all nodes in the network into all_nodes list*/
	all_nodes = lsCreate() ;
	foreach_node(network, gen, node)
		lsNewEnd(all_nodes, (lsGeneric)node, LS_NH) ;

	while(lsLength(all_nodes) != 0){
		mape_optgrp_take_mcn(all_nodes, dis_nodes) ;
		/*
		mape_optgrp_dis_nodes_equ(dis_nodes, optgrps) ;
		*/
	}

	lsDestroy(all_nodes, NULL) ;
	lsDestroy(dis_nodes, NULL) ;
}

/*
 * mape_optgrp
 *
 * optimized group for reducing the current 
 * drawn by each module in the circuit. 
 */
void mape_optgrp(network)
network_t *network ;
{
	/* get the initial solution */
	mape_optgrp_init_solution(network) ;

	/* iterative improvement */
	mape_optgrp_iter(network) ;
}
