/*
 * mape_pbssce
 *
 * Possibility based SSC estimation: using possibility to estimate the 
 * maximum current for a group of of the gates under real circuit model. 
 *
 * Changbo Long,  July 15, 2002. 
 */
#include "mape_int.h"

#define INVALID_DB_ID -1 
#define COSWITCH_THRESHOLD  0.001
#define COSWITCH_STRONG  0.01
#define ACCEPT_PECT 8
#define NUM_CANDIDATE_TIME 5

#define PB_INDEX(n)	(((mape_t *)(n->mape))->C1)
#define PB_GEN(n)  	(((mape_t *)(n->mape))->pbssce.gen)
#define PB_CL(n)	(((mape_t *)(n->mape))->pbssce.cl)
#define PB_ZOL(n)	(((mape_t *)(n->mape))->pbssce.zol)
#define PB_TMPR(n)	(((mape_t *)(n->mape))->pbssce.tmpr)
#define PB_STATIC(n)	(((mape_t *)(n->mape))->pbssce.sw_static)
#define PB_CANTL(n)	(((mape_t *)(n->mape))->pbssce.cantl)
#define PB_CURZO(n)	((mape_pbssce_zo_t *)(((mape_t *)(n->mape))->pbssce.curzo))

#define THIRD_ORDER
#define MAXGROUPSIZE 40

typedef struct mape_pbssce_tv_struct{
	double time ;
	double value ;
	double l_value ; /* lower boundary current value */
	int    index ;
}mape_pbssce_tv_t ;

/* zero and one order logic conflication information struct */
typedef struct mape_pbssce_zo_struct{
	double time ;	 	/* target candidate time */
	double current ;	/* maximum possible current */
	lsList co_list ;	/* coswitch node list */
} mape_pbssce_zo_t ;

/* static_t struct, for candidate time calculation */
typedef struct mape_pbssce_static_struct{
	double time ;
	double s_time ;
	double e_time ;
	double m_value ;
	int repeat ;
} mape_pbssce_static_t  ;

#ifdef THIRD_ORDER
typedef struct mape_pbssce_value3_co_struct{
	int nodes[MAXGROUPSIZE] ;
	double max_value ;
}mape_pbssce_value3_co_t ;

typedef struct mape_pbssce_value3_struct{
	double max_value ;	/* max current value */ 	
	int gsize ;		/* group size */
	int *co_matrix	;	/* co-switch matrix */	
	lsList ol ;		/* orginal list */
	lsList gl ;		/* grow up list */
	lsList sl ;		/* stale list */
}mape_pbssce_value3_t ;

typedef struct mape_pbssce_value3_cant_struct{
	double time ;	// the candidate time. 
	float *cv ;	// the current values at the candidate time ;
}mape_pbssce_value3_cant_t ;

mape_pbssce_value3_t value3 = {0.0, NULL, NULL, NULL, NULL} ;
#endif

int 	mape_pbssce_st_do(node_t *, char *) ;
void 	mape_pbssce_dump_codb(void) ;
void 	mape_pbssce_build_coswitch_db(network_t *) ;
double 	mape_pbssce_itr_node_current(node_t *, double) ;
double 	mape_pbssce_one_order_valid_current(double, node_t *, lsList l) ;
double	mape_pbssce_time_list_coswitch_current(lsList) ;
mape_pbssce_static_t *mape_pbssce_create_static_t(double, double, double, double, int) ;

int *codb, codb_maxid, codb_times;
lsList esl = NULL ;
int use_cant = TRUE ;

/*
 * mape_pbssce_set_use_cant
 *
 * set the use_cant flag. 
 */
void 
mape_pbssce_set_use_cant(s)
int s ;
{
	use_cant = s ;
}

/*
 * mape_pbssce_clean_esl
 *
 * clean the esl list. 
 */
void 
mape_pbssce_clean_esl(void)
{
	int x ;

	if(esl == NULL)
		return ;

	while(lsLength(esl)) 
		lsDelEnd(esl, (lsGeneric *)&x) ;
}

/*
 * mape_pbssce_push_esl
 *
 * push a value into the esl. 
 */
void 
mape_pbssce_push_esl(x)
int x ;
{
	if(esl == NULL)
		esl = lsCreate() ;

	lsNewEnd(esl, (lsGeneric)x, LS_NH) ;
}

/*
 * mape_pbssce_sort_esl
 *
 * sorting the esl list. 
 */
int
mape_pbssce_esl_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	int i1, i2 ;

	i1 = (int)obj1 ;
	i2 = (int)obj2 ;

	return i1 >= i2 ;
}

/*
 * mape_pbssce_sort_esl
 *
 * sorting the esl list. 
 */
void 
mape_pbssce_sort_esl(void)
{
	lsSort(esl, mape_pbssce_esl_cmp) ;
}


/*
 * mape_pbssce_is_esl
 *
 * is the index in esl? 
 */
int 
mape_pbssce_is_esl(x)
int x ;
{
	lsGen gen ;
	int p ;

	lsForeachItem(esl, gen, p){
		if(p == x){
			lsFinish(gen) ;
			return TRUE ;
		}
	}
	return FALSE;
}

/*
 * mape_pbssce_create_zo_t
 *
 * create a zo struct. 
 */
mape_pbssce_zo_t *
mape_pbssce_create_zo_t(time) 
double time ;
{
	mape_pbssce_zo_t *rtv ;

	rtv = ALLOC(mape_pbssce_zo_t, 1) ;	
	rtv->time = time ;
	rtv->current = 0.0 ;
	rtv->co_list = NULL ;

	return rtv ;
}

/*
 * mape_pbssce_destroy_zo_t
 *
 * destroy a zo structure. 
 */
void 
mape_pbssce_destroy_zo_t(zo)
mape_pbssce_zo_t *zo ;
{
	if(zo->co_list)
		lsDestroy(zo->co_list, NULL) ;
	free(zo) ;
}


mape_pbssce_st_t *
mape_pbssce_create_st(time, rp, fp, hp, lp)
double time ; 
float rp, fp, hp, lp ;
{
	mape_pbssce_st_t *rtv ;	
	
	rtv 		= ALLOC(mape_pbssce_st_t, 1) ;
	rtv->time	= time ;
	rtv->rp		= rp ;
	rtv->fp		= fp ;
	rtv->hp 	= hp ;
	rtv->lp		= lp ;

	return rtv ;
}

/* 
 * mape_pbssce_push_st
 *
 * push one switching possibility into the node. 
 */
void 
mape_pbssce_push_st(node, time, rp, fp, hp, lp)
node_t *node ;
double time  ;
float rp, fp, hp, lp ; /*rise, fall, high and low's possibility*/
{
	mape_pbssce_st_t *st ;

	st = mape_pbssce_create_st(time, rp, fp, hp, lp) ;

	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;

	if(!((mape_t *)(node->mape))->plt->st_list)
		((mape_t *)(node->mape))->plt->st_list = lsCreate() ;

	lsNewEnd(((mape_t *)(node->mape))->plt->st_list, (lsGeneric)st, LS_NH) ;
}

/*
 * mape_pbssce_list_next_st
 *
 * find the next st for a list of switching time.  
 */
int 
mape_pbssce_list_next_st(list, time)
lsList list ;
double *time ;
{
	int rtv = FALSE ;
	lsGen gen ;
	mape_pbssce_st_t *st ;

	lsForeachItem(list, gen, st){
		if(st->time > *time){
			*time = st->time ;
			lsFinish(gen) ;
			rtv = TRUE ;
			break ;
		}
	}
	return rtv ;
}

/*
 * mape_pbssce_next_st
 *
 * return the next calculation time point from the values in the inputs. 
 */
int
mape_pbssce_next_st(node, time)
node_t *node ;
double *time ;
{
	int i ;
	node_t *fanin ;
	lsList st_list ;
	double next_time, list_nt ;

	next_time = 10000.1 ;
	foreach_fanin(node, i, fanin){
		list_nt = *time ;
		if(!fanin->mape)
			continue ;
		if(!((mape_t *)(fanin->mape))->plt)
			continue ;
		st_list = ((mape_t *)(fanin->mape))->plt->st_list ;
		if(!st_list)
			continue ;
		if(mape_pbssce_list_next_st(st_list, &list_nt))
			next_time = MIN(next_time, list_nt) ;
	}

	if(next_time >= 10000.0)
		return FALSE ;
	*time = next_time ;
	return TRUE ;
}

/*
 * mape_pbssce_node_index_input 
 *
 * translate a index to the input's index. 
 */
void 
mape_pbssce_node_index_input(ind, nin, input)
int ind, nin, *input ;
{
	return mape_mspplt_node_index_input(ind, nin, input) ;
}

/*
 * mape_pbssce_index_simulate_node
 *
 * simulate node according to the index. 
 */
mape_node_value_t 
mape_pbssce_index_simulate_node(node, index)
node_t *node ;
int *index ;
{
	return mape_mspplt_index_simulate_node(node, index) ;
}

/*
 * mape_pbssce_node_st_possibility
 *
 * node's value's possibility at a particular switching time. 
 */
float 
mape_pbssce_node_st_possibility(node, time, value)
node_t *node;
float time ;
mape_node_value_t value ;
{
	lsList st_list ;	
	lsGen gen ;
	mape_pbssce_st_t *st ;
	mape_pbssce_st_t *equ ;

	if(!node)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	if(!node->mape)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	if(!((mape_t *)(node->mape))->plt)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(!st_list)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	equ = NULL ;
	lsForeachItem(st_list, gen, st)
		if(st->time <= time)
			equ = st ;

	if(!equ)
		return (value == MAPE_0_0 || value == MAPE_1_1)?0.5:0.0 ;

	if(fabs(equ->time - time) <= 0.00001){
		switch(value){
			case MAPE_FALL:
				return equ->fp ;
			case MAPE_RISE:
				return equ->rp ;
			case MAPE_1_1:
				return equ->hp ;
			case MAPE_0_0:
				return equ->lp ;
			default: 
				mape_error("unexpected node value", __FILE__, __LINE__) ;
		}
	}
	if(equ->time < time){
		switch(value){
			case MAPE_FALL:
				return 0.0 ;
			case MAPE_RISE:
				return 0.0 ;
			case MAPE_1_1:
				return equ->rp + equ->hp ;
			case MAPE_0_0:
				return equ->fp + equ->lp ;
			default: 
				mape_error("unexpected node value", __FILE__, __LINE__) ;
		}
	}
}

/*
 * mape_pbssce_index_possibility
 *
 * return the possibility for an index. 
 */
float 
mape_pbssce_index_possibility(node, time, index)
node_t *node ;
double time ;
int *index ;
{
	int i ;
	float rtv = 1.0 ;
	mape_node_value_t v ;
	node_t *p ;

	foreach_fanin(node, i, p){
		switch(index[i]){
			case 0: 
				v = MAPE_FALL ;
				break ;
			case 1: 
				v = MAPE_RISE ;
				break ;
			case 2:
				v = MAPE_1_1 ;
				break ;
			case 3:
				v = MAPE_0_0 ;
				break ;
			default:
				mape_error("unexpected index !", __FILE__, __LINE__) ;
		}
		rtv *= mape_pbssce_node_st_possibility(p, time, v) ;
	}
	return rtv ;
}

/*
 * mape_pbssce_time_st_possibility
 *
 * calculate the new possibility for a special time. 
 */
void 
mape_pbssce_time_st_possibility(node, time, rp, fp, hp, lp)
node_t *node ;
double time ;
float *rp, *fp, *hp, *lp ;
{
	int i, j, t, k ;	
	int index[20] ;

	mape_node_value_t o ;
	float p ;

	/* initially, set all possibility to be zero */
	*rp = *fp = *hp = *lp = 0.0 ;

	t = 1 ;
	for(i=1;i<=node->nin;i++)
		t*=4 ;
	for(i=0;i<t;i++){
		mape_pbssce_node_index_input(i, node->nin, index) ;
		p = mape_pbssce_index_possibility(node, time, index) ;
		if(p == 0.0)
			continue ;
		o = mape_pbssce_index_simulate_node(node, index) ;
		switch(o){
			case MAPE_FALL:
				*rp += p ;
				break ;
			case MAPE_RISE:
				*fp += p ;
				break ;
			case MAPE_1_1:
				*hp += p ;
				break ;
			case MAPE_0_0:
				*lp += p ;
				break ;
			case MAPE_U_U:
				mape_error("simulation error!", __FILE__, __LINE__) ;
		}
	}
}


/*
 * mape_pbssce_st_do
 *
 * do switching possibility calculation for every gate. 
 */
int 
mape_pbssce_st_do(node, arg)
node_t *node ;
char *arg ;	/* no use here*/
{
	mape_curlib_gate_t *c_gate ;
	double time = -1.0, atime, xtime ;
	float rp, fp, hp, lp, delay;
	double cload ;
	int count = 0 ;


	//fprintf(mapeout, "Handle node %s, ", node->name) ; 

	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;
	c_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;

	/* under the assumption of various delay model */
	while(mape_pbssce_next_st(node, &time)){
		count ++ ;
		mape_pbssce_time_st_possibility(node, time, &rp, &fp, &hp, &lp) ; 
		atime = time + mape_curlib_value(c_gate, MAPE_CURLIB_RDLAY, cload) ;
		mape_pbssce_push_st(node, atime, rp, fp, hp, lp) ; 
		//fprintf(mapeout, "Push node %s, (%lf, %f, %f, %f, %f)\n",node->name, atime, rp, fp, hp, lp) ;
	}

	//fprintf(mapeout, "count = %d\n", count) ;
	return TRUE ;
}


/*
 * mape_pbssce_st_possibility
 * 
 * Get all the gate's switching time and possibility under the 
 * real delay model. Please refer to mape_mspplt_st_possibility
 * in mape_mspplt.c for comparison with the unit delay model. 
 */
void 
mape_pbssce_st_possibility(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;
	double time ;
	int tmp ;

	mape_imp_reset(mape_pbssce_st_do, NULL) ;
	foreach_primary_input(network, gen, pi){
		if(CURLIB_APPROX_DELAY_UNIT > 0){
			tmp = (int)((HALF_SLOP_TIME + CURLIB_APPROX_DELAY_UNIT/2)/CURLIB_APPROX_DELAY_UNIT) ;
			time = (double)(tmp*CURLIB_APPROX_DELAY_UNIT) ;
			mape_pbssce_push_st(pi, time, 0.5, 0.5, 0.0, 0.0) ;
		}else 
			mape_pbssce_push_st(pi, HALF_SLOP_TIME, 0.5, 0.5, 0.0, 0.0) ;
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
}

/*
 * mape_pbssce_node_dump_st
 *
 * dump the node's st list. 
 */
void 
mape_pbssce_node_dump_st(node)
node_t *node ;
{
	lsGen gen ;
	lsList st_list ;
	mape_pbssce_st_t *st ;
	
	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	lsForeachItem(st_list, gen, st)
		fprintf(mapeout, "\t(%5.2lf, %5.2f, %5.2f, %5.2f, %5.2f)\n", st->time, st->rp, st->fp, 
			st->hp, st->lp) ;
}

/*
 * mape_pbssce_dump_st
 *
 * dump the st's information. 
 */
void 
mape_pbssce_dump_st(network)
network_t *network ;
{
	lsGen gen, gen1;
	node_t *node ;
	lsList st_list ;
	mape_pbssce_st_t *st ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		fprintf(mapeout, "Node: %s\n", node->name) ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st)
			fprintf(mapeout, "\t(%5.2lf, %5.2f, %5.2f, %5.2f, %5.2f)\n", st->time, st->rp, st->fp, 
				st->hp, st->lp) ;
	}
}

/*
 * mape_pbssce_st2current
 *
 * calculate the current from a switching possibility 
 * structure and put it into the current list. 
 */
void 
mape_pbssce_st2current(stp, curlib_gate, one_fanout, cload, cur_points) 
mape_pbssce_st_t *stp ;
mape_curlib_gate_t *curlib_gate ;
char *one_fanout ;
double cload ;
lsList cur_points ;
{
	double delay, height, pos, dis ;

	if(stp->rp > 0.01){
		delay = stp->time - HALF_SLOP_TIME ;
		if(one_fanout){
			height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRMAX, one_fanout) ;
			pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRPOS, one_fanout) ;
			dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRDIS, one_fanout) ;
		}else{
			height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRMAX, cload) ;
			pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRPOS, cload) ;
			dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRDIS, cload) ;
		}
		if(height >= 0){
			if(delay >= 45)
				delay -= 40 ;
			pos = 10 ;
			dis /= 2 ;
		}
		//if(height <= 0){
			/*
			if(stp->rp >= 0.1)
				mape_cur_add_triangle_current(delay, height*(stp->rp)*2, pos, dis, cur_points) ;
			else
				mape_cur_add_triangle_current(delay, height*(stp->rp), pos, dis, cur_points) ;
			*/
			/*
			mape_cur_add_triangle_current(delay, height*(stp->rp)*(stp->rp)*10, pos, dis, cur_points) ;
			*/
			/*
			mape_cur_add_triangle_current(delay, height*(stp->rp), pos, dis, cur_points) ;
			*/
		//}
		if(height < 0)
			mape_cur_add_triangle_current(delay, height, pos, dis, cur_points) ;
	}

	if(stp->fp > 0.01){
		delay = stp->time - HALF_SLOP_TIME ;
		if(one_fanout){
			height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFMAX, one_fanout) ;
			pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFPOS, one_fanout) ;
			dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFDIS, one_fanout) ;
		}else{
			height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFMAX, cload) ;
			pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFPOS, cload) ;
			dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFDIS, cload) ;
		}
		if(height >= 0){
			if(delay >= 45)
				delay -= 40 ;
			pos = 10 ;
			dis /= 2 ;
		}
		//if(height <= 0){
			/*
			if(stp->rp >= 0.1)
				mape_cur_add_triangle_current(delay, height*(stp->rp)*2, pos, dis, cur_points) ;
			else
				mape_cur_add_triangle_current(delay, height*(stp->rp), pos, dis, cur_points) ;
			*/
			/*
			mape_cur_add_triangle_current(delay, height*(stp->rp)*(stp->rp)*10, pos, dis, cur_points) ;
			*/
			mape_cur_add_triangle_current(delay, height*(stp->fp), pos, dis, cur_points) ;
		//}
		if(height < 0)
			mape_cur_add_triangle_current(delay, height, pos, dis, cur_points) ;
	}
}

/*
 * mape_pbssce_boundary_st2current. 
 *
 * add st's correponding current with 
 * boundary.  
 */
double
mape_pbssce_current_add_triangle(stp, curlib_gate, one_fanout, cload, pre_time, node_current) 
mape_pbssce_st_t *stp ;
mape_curlib_gate_t *curlib_gate ;
char *one_fanout ;
double cload ;
double pre_time ;
lsList node_current ;
{
	double delay, pos, dis ;
	double height1= 0.0, height2=0.0, height, t;
	mape_cur_point_t *new_point ;

	delay = stp->time - HALF_SLOP_TIME ;
	/* get the height, pos, dis */
	if(stp->rp > 0.1){
		if(one_fanout){
			height1 = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRMAX, one_fanout) ;
		}else{
			height1 = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRMAX, cload) ;
		}
	}

	if(stp->fp > 0.1){
		if(one_fanout){
			height2 = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFMAX, one_fanout) ;
		}else{
			height2 = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFMAX, cload) ;
		}
	}
	if(height1 == 0.0 && height2 == 0.0){
		return ;
	}else if(height1 < height2){
		height = height1 ;
		if(one_fanout){
			pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRPOS, one_fanout) ;
			dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRDIS, one_fanout) ;
		}else{
			pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRPOS, cload) ;
			dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRDIS, cload) ;
		}
	}else {
		height = height2 ;
		if(one_fanout){
			pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFPOS, one_fanout) ;
			dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFDIS, one_fanout) ;
		}else{
			pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFPOS, cload) ;
			dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFDIS, cload) ;
		}
	}

	if(pre_time < 0){
		new_point = mape_cur_create_cur_point_t(delay, 0.0);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
		new_point = mape_cur_create_cur_point_t(delay+pos, height);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
	}else if(pre_time < delay){
		new_point = mape_cur_create_cur_point_t(pre_time, 0.0);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
		new_point = mape_cur_create_cur_point_t(delay, 0.0);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
		new_point = mape_cur_create_cur_point_t(delay+pos, height);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
	}else{
		height1 = height - height * (delay + pos - (pre_time - dis)) / dis ; 
		t = (delay + pos - (pre_time - dis)) * height1 / height ; 
		new_point = mape_cur_create_cur_point_t(pre_time - dis + t, height1) ;
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
		new_point = mape_cur_create_cur_point_t(delay+pos, height);
		lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;
	}
	return delay + dis ;
}

/*
 * mape_pbssce_node_current
 *
 * get node's current wave with maximum value boundary. 
 */
void 
mape_pbssce_node_current(node, node_current)
node_t *node ;
lsList node_current ;
{
	lsList st_list ;
	char *one_fanout = NULL ;
	double cload ;
	mape_curlib_gate_t *curlib_gate ;
	mape_pbssce_st_t *stp ;
	lsGen gen ;
	double pre_time = -1.0 ;
	double end_time = -1.0 ;
	mape_cur_point_t *new_point, *p ;
	int i ;
	node_t *fanout ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(node_num_fanout(node) == 1){
		foreach_fanout(node, gen, fanout){
			curlib_gate = mape_cur_curlib_gate(fanout) ;
			if(curlib_gate != NULL) 
				one_fanout = curlib_gate->name ;
		}
	}

	curlib_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;
	lsForeachItem(st_list, gen, stp){
		end_time = mape_pbssce_current_add_triangle(stp, curlib_gate, one_fanout, cload, pre_time, node_current) ;
		pre_time = end_time ;
	}
	/* handle the last end time */
	new_point = mape_cur_create_cur_point_t(end_time, 0.0);
	lsNewEnd(node_current, (lsGeneric)new_point, LS_NH) ;

	/* dump node's */
	i = 0 ;
	/*
	fprintf(mapeout, "dump node %s current's point\n\t", node->name) ;
	lsForeachItem(node_current, gen, p){
		i ++ ;
		if(i % 5 == 0)
			fprintf(mapeout, "\n\t") ;
		if(i == 1)
			fprintf(mapeout, "(%.2lf,%.2lf)", p->time, p->value) ;
		else
			fprintf(mapeout, ",(%.2lf,%.2lf)", p->time, p->value) ;
	}
	fprintf(mapeout, "\n") ;
	*/
}

/*
 * mape_pbssce_merge_current
 *
 * merge the node's current into the total current curve.   
 */
void 
mape_pbssce_merge_current(node_current, cur_points)
lsList node_current, cur_points ;
{
	lsList back_list ;
	lsGen gen1, gen2, gen ;
	mape_cur_point_t *p = NULL, *tmp, *p1 = NULL, *p2 = NULL, *tp1, *tp2, *new_point ;
	lsHandle h ;
	int over1 = FALSE, over2 = FALSE;
	double next_time, next_value, cur_point_value ;
	int i ;

	back_list = lsCreate() ;
	while(lsLength(cur_points)){
		lsDelBegin(cur_points, (lsGeneric *)&p) ;
		LS_ASSERT(lsNewEnd(back_list, (lsGeneric)p, LS_NH)) ;
	}

	/*reconstruct cur_points list */
	gen1 = lsStart(node_current) ; 
	gen2 = lsStart(back_list) ;

	for(;;){
		if(over1 && over2)	
			break ;

		/* get the next two points to be considered*/
		if(!over1 && p1 == NULL){
			if(lsNext(gen1, (lsGeneric *)&p1, &h) != LS_OK){
				p1 = NULL ;
				over1 = TRUE ;
			}
		}
		if(!over2 && p2 == NULL){
			if(lsNext(gen2, (lsGeneric *)&p2, &h) != LS_OK){
				p2 = NULL ;
				over2 = TRUE ;
			}
		}

		tp1 = p1 ; tp2 = p2 ;
		/* handle the two points */
		if(tp1 != NULL && tp2 != NULL){
			if(tp1->time == tp2->time){
				next_time = p1->time ;
				next_value = p1->value + p2->value ;
				p1 = NULL; 
				p2 = NULL;
			}else if(tp1->time < tp2->time){
				next_time = p1->time ;
				cur_point_value = mape_cur_point_value(back_list, next_time) ;
				next_value = p1->value + ((cur_point_value == UNKNOWN_VOL)?0.0:cur_point_value) ;
				p1 = NULL ;
			}else if(tp2->time < tp1->time){
				next_time = p2->time ;
				cur_point_value = mape_cur_point_value(node_current, next_time) ;
				next_value = p2->value + ((cur_point_value == UNKNOWN_VOL)?0.0:cur_point_value) ;
				p2 = NULL ;
			}
		}else if(tp1 == NULL && tp2 != NULL){
			next_time = p2->time ;
			next_value = p2->value ;
			p2 = NULL ;
		}else if(tp1 != NULL && tp2 == NULL){
			next_time = p1->time ;
			next_value = p1->value ;
			p1 = NULL ;
		}else if(tp1 == NULL && tp2 == NULL){
			break ;
		}

		new_point = mape_cur_create_cur_point_t(next_time, next_value) ;		
		lsNewEnd(cur_points, (lsGeneric)new_point, LS_NH) ; 
	}

	lsFinish(gen1) ;
	lsFinish(gen2) ;

	lsDestroy(back_list, free) ;

	/* dump node's */
	/*
	i = 0 ;
	fprintf(mapeout, "dump cur_points \n\t") ;
	lsForeachItem(cur_points, gen, p){
		i ++ ;
		if(i % 5 == 0)
			fprintf(mapeout, "\n\t") ;
		if(i == 1)
			fprintf(mapeout, "(%.2lf,%.2lf)", p->time, p->value) ;
		else
			fprintf(mapeout, ",(%.2lf,%.2lf)", p->time, p->value) ;
	}
	fprintf(mapeout, "\n") ;
	*/
}

/*
 * mape_pbssce_node_estimate_current
 *
 * estimate node's current. 
 */
void 
mape_pbssce_node_estimate_current(node, cur_points)
node_t *node;  
lsList cur_points ;
{
	lsList st_list ;
	char *one_fanout = NULL ;
	mape_curlib_gate_t *curlib_gate ;
	double cload ;
	mape_pbssce_st_t *stp ;
	lsGen gen ;
	node_t *fanout ;
	lsList node_current ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(st_list == NULL)
		return ;

	if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
		return ;

	node_current = lsCreate() ;
	mape_pbssce_node_current(node, node_current) ;
	mape_pbssce_merge_current(node_current, cur_points) ;
	lsDestroy(node_current, free) ;

	/*
	if(node_num_fanout(node) == 1){
		foreach_fanout(node, gen, fanout){
			curlib_gate = mape_cur_curlib_gate(fanout) ;
			if(curlib_gate != NULL) 
				one_fanout = curlib_gate->name ;
		}
	}

	curlib_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;

	lsForeachItem(st_list, gen, stp)
		mape_pbssce_st2current(stp, curlib_gate, one_fanout, cload, node_current) ;
	*/
}

/*
 * mape_pbssce_estimate_current 
 *
 * estimate the current by add the random result together. 
 */
void 
mape_pbssce_estimate_current(network)
network_t *network ;
{
	lsList cur_points ;
	node_t *node ;
	lsGen gen ;
	int i ;
	double max = 0.0 ;
	double time = 0.0 ;
	mape_cur_point_t *p ;

	cur_points = lsCreate() ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT
			|| !mape_is_given_node(node))
				continue ;
		mape_pbssce_node_estimate_current(node, cur_points) ;
	}

	lsForeachItem(cur_points, gen, p){
		if(fabs(p->value) > fabs(max)){
			max = p->value ;
			time = p->time ;
		}
	}

	fprintf(mapeout, "Maximum pbssce estimated current value is: %.2lf at time %.2lf\n", -max, time) ;


	/* dump the information */	
	/*
	i = 0 ;
	lsForeachItem(cur_points, gen, p){
		i++ ;
		if(i%10 == 1)
			fprintf(mapeout, "x(%d:%d) = [%.2lf, ", i, i+9, p->time) ;
		else if(i%10 == 0)
			fprintf(mapeout, "%.2lf]\n", p->time) ;
		else 
			fprintf(mapeout, "%.2lf,", p->time) ;
	}
	fprintf(mapeout, "\n") ;

	i = 0 ;
	lsForeachItem(cur_points, gen, p){
		i++ ;
		if(i%10 == 1)
			fprintf(mapeout, "y(%d:%d) = [%.2lf, ", i, i+9, (p->value)) ;
		else if(i%10 == 0)
			fprintf(mapeout, "%.2lf]\n", (p->value)) ;
		else 
			fprintf(mapeout, "%.2lf,", (p->value)) ;
	}
	fprintf(mapeout, "\n") ;
	*/

	lsDestroy(cur_points, free) ;
}

/*
 * mape_pbssce_group_coswitch
 *
 * get the coswitch value for a group of node. 
 */
double 
mape_pbssce_group_coswitch(group)
lsList group ;
{
	node_t *node1, *node2 ;
	int rtv = 0  ;
	lsGen gen1, gen2 ;

	lsForeachItem(group, gen1, node1)
		lsForeachItem(group, gen2, node2){
			if(node1 != node2){
				rtv += mape_pbssce_codb_value(mape_pbssce_get_dbid(node1), mape_pbssce_get_dbid(node2)) ;
			}
		}
	return (double)(rtv * 1.0)/codb_times ;
}

/*
 * mape_pbssce_node_group_coswitch 
 *
 * the sum of the coswitch between 
 * one node with a group of node. 
 */
float
mape_pbssce_node_group_coswitch(n, g)
node_t * n ;
lsList g ;
{
	lsGen gen ;
	node_t *node ;
	int rtv ;
	int id, id1 ;

	rtv = 0 ;
	id = mape_pbssce_get_dbid(n) ;
	lsForeachItem(g, gen, node){
		if(node != n){
			id1 = mape_pbssce_get_dbid(node);
			rtv += mape_pbssce_codb_value(MAX(id, id1), MIN(id, id1)) ;
		}
	}

	return (float)((float)(rtv*1.0))/codb_times ;
}

/*
 * mape_pbssce_nn_coswitch
 *
 * get the coswitch between two nodes. 
 */
float
mape_pbssce_nn_coswitch(n1, n2)
node_t *n1, *n2 ;
{
	int id1, id2 ;
	int rtv ;

	if(n1 == n2)
		return 1.0;

	id1 = mape_pbssce_get_dbid(n1) ;
	id2 = mape_pbssce_get_dbid(n2) ;

	rtv = mape_pbssce_codb_value(MAX(id1, id2), MIN(id1, id2)) ;

	return (float)((float)(rtv*1.0))/codb_times ;
}

/*
 * mape_pbssce_group_satisfied
 *
 * if the group satisfy the strong connect constraint
 */
int 
mape_pbssce_group_satisfied(g, tl)
lsList g ;
int tl ;
{
	lsGen gen1 ;
	lsGen gen2 ;
	node_t *n1 ;
	node_t *n2 ;
	int length ;
	long r ;
	float v ;

	length = lsLength(g) ;
	if(length <= 1)
		return TRUE ;

	lsForeachItem(g, gen1, n1){
		if(mape_pbssce_node_group_coswitch(n1, g) <= length*COSWITCH_STRONG){
			lsFinish(gen1) ;
			return FALSE ;
		}
		lsForeachItem(g, gen2, n2){
			if(n1 != n2 && mape_pbssce_nn_coswitch(n1, n2) <= COSWITCH_THRESHOLD){
				lsFinish(gen1) ;
				lsFinish(gen2) ;
				return FALSE ;
			}
		}
	}


	return TRUE ;

	/*
	if(length <= tl/3){
		return TRUE ;
	}else{
		r = random() ;
		r = r%1000 ;
		v = (float)(r*1.0)/1000 ;

		if(v >= (float)((length*1.0)/tl) - 0.2)
			return TRUE ;
		else 
			return FALSE ;
	}
	*/
}

/*
 * mape_pbssce_biggest_group_satisfied
 *
 * find the biggest_group_satisfied
 */
void 
mape_pbssce_biggest_group_satisfied(que, sg, tl)
lsList que, sg ;
int tl ;
{
	lsGen gen1 ;
	lsGen gen2 ;
	node_t *p, *q;
	lsList tmp ;
	lsList newl ;

	lsDelBegin(que, (lsGeneric *)&tmp) ;

	/* if the current poped list satisfied */
	if(mape_pbssce_group_satisfied(tmp, tl)){
		while(lsLength(tmp)){
			lsDelBegin(tmp, (lsGeneric *)&p) ;
			lsNewEnd(sg, (lsGeneric)p, LS_NH) ;
		}
		lsDestroy(tmp, NULL) ;
		return ;
	}

	/* put all the child in the end of the que */
	lsForeachItem(tmp, gen1, p){
		newl = lsCreate() ;
		lsForeachItem(tmp, gen2, q)
			if(p != q)
				lsNewEnd(newl, (lsGeneric)q, LS_NH) ;
		lsNewEnd(que, (lsGeneric)newl, LS_NH) ;
	}

	lsDestroy(tmp, NULL) ;
	return mape_pbssce_biggest_group_satisfied(que, sg, tl) ;
}

/*
 * mape_pbssce_strong_coswitch_group
 *
 * get the strongest coswitch group. 
 */
void 
mape_pbssce_strong_coswitch_group(g, sg) 
lsList g, sg ;
{
	lsGen gen ;
	lsGen gen1 ;
	lsHandle h ;
	node_t *p, *q ;
	int count, index ;
	lsList tsg, que ;
	int length ;

	tsg = lsCreate() ;
	que = lsCreate() ;

	while(lsLength(sg))
		lsDelEnd(sg, (lsGeneric *)&p) ;

	/* first get ride of those gates with too many small values*/
	length = lsLength(g) ;
	lsForeachItem(g, gen, p){
		count = 0 ;
		lsForeachItem(g, gen1, q)
			if(mape_pbssce_nn_coswitch(p, q) <= COSWITCH_THRESHOLD) 
				count ++ ;
		if(count <= length*0.8)
			lsNewEnd(tsg, (lsGeneric)p, NULL) ;
	}

	lsNewEnd(que, (lsGeneric)tsg, NULL) ;
	mape_pbssce_biggest_group_satisfied(que, sg, lsLength(tsg)) ;
	lsDestroy(que, free) ;
}

/*
 * mape_pbssce_dump_group_codb
 *
 * dump a group's codb
 */
void 
mape_pbssce_dump_group_codb(group)
lsList group ;
{
	lsGen gen1, gen2 ;
	node_t *p1, *p2 ;
	int id1, id2 ;
	float value ;

	lsForeachItem(group, gen1, p1)
		fprintf(mapeout, "\t%s", p1->name) ;
		
	lsForeachItem(group, gen1, p1){
		fprintf(mapeout, "\n%s", p1->name) ;
		id1 = mape_pbssce_get_dbid(p1) ;
		lsForeachItem(group, gen2, p2){
			id2 = mape_pbssce_get_dbid(p2) ;
			value = ((float)(1.0 * mape_pbssce_codb_value(MAX(id1, id2), MIN(id1, id2))))/codb_times ;
			fprintf(mapeout, "\t%.4f",  value);
		}
	}
	fprintf(mapeout, "\n") ;
}

/*
 * mape_pbssce_backup_vol_points. 
 *
 * get the ith backup vol points. 
 */
lsList 
mape_pbssce_backup_vol_points(node, i)
node_t *node ;
int i ;
{
	int count = 0 ;
	lsGen gen ;
	lsList rtv = NULL ;
	lsList l ;
	lsList lp ;
	
	l = ((mape_t *)(node->mape))->pbssce.cl ;

	lsForeachItem(l, gen, lp){
		if(count == i){
			lsFinish(gen) ;
			return lp ;
		}
		count ++ ;
	}
	return NULL ;
}

/*
 * mape_pbssce_add_node_current_v
 *
 * add node's 
 */
void 
mape_pbssce_add_node_current_v(node, vol_points, cur_points)
node_t *node ;
lsList vol_points, cur_points ;
{
	lsGen gen ;
	double cload ;
	mape_curlib_gate_t *curlib_gate ;
	double pos, height, dis ;
	char *one_fanout = NULL ;
	node_t *fanout ;
	double start = -0.0 ;
	mape_node_value_t type ;
	double delay, times ;
	int multi ;

	/* only one fanout, get the fanout curlib type*/
	if(node_num_fanout(node) == 1){
		foreach_fanout(node, gen, fanout){
			curlib_gate = mape_cur_curlib_gate(fanout) ;
			if(curlib_gate != NULL)
				one_fanout = curlib_gate->name ;
		}
	}

	curlib_gate = mape_cur_curlib_gate(node) ;
	cload = mape_cur_node_loading(node) ;

#ifdef CUR_DEBUG
	fprintf(mapeout, "Add node %s's current\n", node->name) ;
#endif 
	while(mape_cur_next_trans(vol_points, &start, &type, &delay, &times, &multi)){
		if(type == MAPE_FALL){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IFDIS, one_fanout) ;
			}else{ 
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IFDIS, cload) ;
			}
		}else if(type == MAPE_RISE){
			if(one_fanout){
				height = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRMAX, one_fanout) ;
				pos = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRPOS, one_fanout) ;
				dis = mape_curlib_loadg_value(curlib_gate, MAPE_CURLIB_IRDIS, one_fanout) ;
			}else{
				height = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRMAX, cload) ;
				pos = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRPOS, cload) ;
				dis = mape_curlib_value(curlib_gate, MAPE_CURLIB_IRDIS, cload) ;
			}
		}else 
			continue ;

		/* revise the positive curve*/
		/*
		if(height >= 0){
			if(delay >= 45)
				delay -= 40 ;
			pos = 10 ; 
			dis /= 2 ;
		}
		if(times < 0.99 && times >= 0.70) 
			times = 1.0 ;
		*/
		if(height >= 0)
			continue ;
		/*
		if(multi > 1)
			mape_cur_add_triangle_current(delay, height*times*multi*0.8, pos, dis, cur_points) ; 
		else 
			mape_cur_add_triangle_current(delay, height*times, pos, dis, cur_points) ; 
		*/
		mape_cur_add_triangle_current(delay, height, pos, dis, cur_points) ; 
	}
}

/*
 * mape_pbssce_lists_max_current
 *
 * add a list of current waveform together and get the maximum current. 
 */
void 
mape_pbssce_lists_max_current(vpl, current, time)
lsList vpl ;
double *current, *time ;
{
	lsList cur_points, cl ;
	lsGen gen ;
	mape_cur_point_t *cp ;

	cur_points = lsCreate();

	lsForeachItem(vpl, gen, cl)
		mape_pbssce_merge_current(cl, cur_points) ;

	*current = 0.0 ;
	*time = 0.0 ;
	lsForeachItem(cur_points, gen, cp){
		if(cp->value < *current){
			*current = cp->value ;
			*time = cp->time ;
		}
	}
	lsDestroy(cur_points, free) ;
}

/*
 * mape_pbssce_group_max_current
 *
 * get group maximum current from the backup simulation. 
 */
int 
mape_pbssce_group_max_current(l, current, time)
lsList l;
double *current, *time ;
{
	lsList vol_points, vpl ;
	lsGen gen ;
	node_t *node ;
	mape_cur_point_t *point ;
	lsHandle h ;

	vpl = lsCreate() ;
	lsForeachItem(l, gen, node){
		if(lsNext(PB_GEN(node), (lsGeneric *)&vol_points, &h) != LS_OK){
			lsFinish(gen) ;
			lsDestroy(vpl, NULL) ;
			return FALSE ;
		}
		lsNewEnd(vpl, (lsGeneric)vol_points, LS_NH) ;
	}

	mape_pbssce_lists_max_current(vpl, current, time);
	lsDestroy(vpl, NULL) ;
	return TRUE ;
}

/*
 * mape_pbssce_tv_current_cmp
 *
 * compare tv value
 */
int 
mape_pbssce_tv_current_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	mape_pbssce_tv_t *tv1, *tv2 ;

	tv1 = (mape_pbssce_tv_t *)obj1 ;
	tv2 = (mape_pbssce_tv_t *)obj2 ;

	return (tv1->l_value >= tv2->l_value) ;
}

/*
 * create the tv_t
 *
 */
mape_pbssce_tv_t *
mape_pbssce_create_tv_t(time, value, l_value, index)
double time, value, l_value ;
int index ;
{
	mape_pbssce_tv_t *rtv;

	rtv 	= ALLOC(mape_pbssce_tv_t, 1) ;
	rtv->time = time ;
	rtv->value = value ;
	rtv->l_value = l_value ;
	rtv->index = index ;

	return rtv ;
}

/*
 * mape_pbssce_time_exist
 *
 * check if the time already exist. 
 */
int 
mape_pbssce_time_exist(time, l)
double time ;
lsList l ;
{
	lsGen gen ;
	mape_pbssce_tv_t *tvp ;

	lsForeachItem(l, gen, tvp){
		if(fabs(time - tvp->time) <= 0.01){
			lsFinish(gen) ;
			return TRUE ;
		}
	}
	return FALSE ;
}

#ifdef THIRD_ORDER

/*
 * mape_pbssce_value3_create_cant_t 
 *
 * create a cant_t structure.  
 */
mape_pbssce_value3_cant_t *
mape_pbssce_value3_create_cant_t(void)
{
	mape_pbssce_value3_cant_t *rtv ;

	rtv 		= ALLOC(mape_pbssce_value3_cant_t, 1) ;
	rtv->time 	= 0.0 ;
	rtv->cv		= NULL ;

	return rtv ;
}

/*
 * mape_pbssce_node_destroy_cantl
 *
 * destroy node's cantl ; 
 */
void 
mape_pbssce_node_destroy_cantl(node)
node_t *node ;
{
	mape_pbssce_value3_cant_t *c ;

	if(PB_CANTL(node) == NULL)
		return ;

	while(lsLength(PB_CANTL(node))){
		lsDelEnd(PB_CANTL(node), (lsGeneric *)&c) ;
		if(c->cv)
			free(c->cv) ;
		free(c) ;	
	}
}

/*
 * mape_pbssce_static_cmp
 *
 * compare two static structure. 
 */
int
mape_pbssce_static_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	mape_pbssce_static_t *s1, *s2 ;
	
	s1 = (mape_pbssce_static_t *)obj1 ;
	s2 = (mape_pbssce_static_t *)obj2 ;

	return ((s1->m_value)*(1+ 1.0*(s1->repeat)/codb_times) > 
		(s2->m_value)*(1+ 1.0*(s2->repeat)/codb_times));
}


/*
 * mape_pbssce_match_static_t
 *
 * match the static list 
 */
int 
mape_pbssce_match_static_t(s, l, r)
mape_pbssce_static_t *s, **r ;
lsList l ;
{
	lsGen gen ;
	mape_pbssce_static_t *p ;

	lsForeachItem(l, gen, p){
		if(fabs(p->time - s->time) <= 0.5){
			*r = p ;
			lsFinish(gen) ;
			return TRUE ;
		}
	}

	return FALSE ;
}

/*
 * mape_pbssce_itr_estimate_current_time3
 *
 * estimate the time for maximum current 
 */
void
mape_pbssce_itr_estimate_current_time3(network, tl, l)
network_t *network ;
lsList tl, l ;
{
	lsGen gen, gen1, gen2 ;
	node_t *node ;
	mape_pbssce_static_t *s, *ms ;
	mape_cur_point_t *cp ;
	lsList cl ;
	int i, total_i, cn ;
	lsList sl ;
	mape_pbssce_tv_t *tvp ;
	lsHandle h ;

	/*
	lsForeachItem(l, gen, node){
		fprintf(mapeout, "Node %s static information\n", node->name) ;
		i = 0 ;
		lsForeachItem(PB_CL(node), gen1, cl){
			i++ ;
			fprintf(mapeout, "%dth ", i) ;
			lsForeachItem(cl, gen2, cp){
				fprintf(mapeout, "[%.2f, %.2f]", cp->time, cp->value) ; 
			}
			fprintf(mapeout, "\n") ;
		}
		lsForeachItem(PB_STATIC(node), gen1, s){
			fprintf(mapeout, "Static time=%.2f, s_time=%.2f, e_time=%.2f, m_value=%.2f, repeat=%d\n", 
				s->time, s->s_time, s->e_time, s->m_value, s->repeat) ;
		}
	}
	fprintf(mapeout, "Begin estimate current time\n") ;
	*/

	lsFirstItem(l, (lsGeneric *)&node, &h) ;
	total_i = lsLength(PB_CL(node)) ;
	sl = lsCreate() ;
	lsForeachItem(l, gen, node){
		lsForeachItem(PB_STATIC(node), gen1, s){
			if(mape_pbssce_match_static_t(s, sl, &ms)){
				if(!mape_pbssce_static_cmp(s, ms)){
					ms->time = s->time ;
					ms->s_time = s->s_time ;
					ms->e_time = s->e_time ;
				}
				ms->repeat += s->repeat ;
			}else{
				ms = mape_pbssce_create_static_t(s->time, s->s_time, s->e_time, s->m_value, s->repeat) ; 
				lsNewEnd(sl, (lsGeneric)ms, LS_NH) ;
			}
		}
	}
	lsSort(sl, mape_pbssce_static_cmp) ;
	cn = lsLength(sl) ;
	cn = (cn > NUM_CANDIDATE_TIME)?NUM_CANDIDATE_TIME:cn ;
	i = 0 ;
	lsForeachItem(sl, gen, s){
		i++ ;
		tvp=mape_pbssce_create_tv_t(s->time, 0.0, 0.0, 0) ;
		lsNewEnd(tl, (lsGeneric)tvp, LS_NH) ; 
		//fprintf(mapeout, "candidate time=%.2lf\n", s->time) ;
		if(i >= cn){
			lsFinish(gen) ;
			break ;
		}
	}
	lsDestroy(sl, free) ;
}
#endif

/*
 * mape_pbssce_itr_estimate_current_time
 *
 * estimate the time for maximum current. 
 */
void
mape_pbssce_itr_estimate_current_time(network, tl, l)
network_t *network ;
lsList tl, l;
{
	int i ;
	lsList al ;
	double time, current ;
	mape_pbssce_tv_t *tvp ;
	lsGen gen ;
	int length, count ;
	node_t *node ;
	int empty ;

	al = lsCreate() ;
	i = 0 ;

	lsForeachItem(l, gen, node)
		PB_GEN(node) = lsStart(PB_CL(node)) ;

	empty = FALSE ;
	for(;;){
		if(!mape_pbssce_group_max_current(l, &current, &time))
			break ;
		//fprintf(mapeout, "i=%d, max_current=%.2lf, time=%.2lf\n", i, current, time) ;
		tvp = mape_pbssce_create_tv_t(time, 0.0, current, i) ;
		lsNewEnd(al, (lsGeneric)tvp, LS_NH) ;
		i++ ;
	}

	lsForeachItem(l, gen, node)
		lsFinish(PB_GEN(node)) ;

	lsSort(al, mape_pbssce_tv_current_cmp) ;

	/*clean the effective simulation list */
	mape_pbssce_clean_esl() ;
	i=0 ;
	length = lsLength(al) ;
	//count = 2*(int)sqrt(length) ;
	count = length - 10 ;

	lsForeachItem(al, gen, tvp){
		if(i >= count){
			lsFinish(gen) ;
			break ;
		}
		mape_pbssce_push_esl(tvp->index) ;
		//fprintf(mapeout, "Push esl : index = %d, value=%.2lf\n", tvp->index, tvp->l_value) ;
		i++ ;
	}
	mape_pbssce_sort_esl() ;

	for(i=0;i<NUM_CANDIDATE_TIME;i++){
		if(lsLength(al) == 0)
			break ;
		lsDelBegin(al, (lsGeneric *)&tvp) ;
		if(!mape_pbssce_time_exist(tvp->time, tl)){
			lsNewEnd(tl, (lsGeneric)tvp, LS_NH) ;
			//fprintf(mapeout, "candidate time=%.2lf, value=%.2lf\n", tvp->time, tvp->l_value) ;  
		}
	}
	lsDestroy(al, free) ;
}

/*
 * mape_pbssce_itr_vol_node_current
 *
 * Knowing the voltage curve, time and node, get the current value
 */
double
mape_pbssce_itr_vol_node_current(node, vol_points, time)
node_t *node ;
lsList vol_points;
double time ;
{
	lsList cur_points ;
	double rtv = 0.0 ;

	cur_points = lsCreate() ;
	mape_pbssce_add_node_current_v(node, vol_points, cur_points) ;
	rtv = mape_cur_point_value(cur_points, time) ;

	if(rtv == UNKNOWN_VOL)
		rtv = 0.0 ;

	lsDestroy(cur_points, free) ;
	return rtv ;
}

/*
 * mape_pbssce_itr_node_current
 *
 * get the node's maximum current at the special time point. 
 */
double
mape_pbssce_itr_node_current(node, time)
node_t *node ;
double time ;
{
	int i ;
	lsList vol_points ;
	double rtv = 0.0 ;
	double current ;

	i = 0 ;
	for(;;){
		vol_points = mape_pbssce_backup_vol_points(node, i) ;
		if(vol_points == NULL)
			break ;
		
		if(mape_pbssce_is_esl(i)){
			current = mape_pbssce_itr_vol_node_current(node, vol_points, time) ; 
			if(fabs(current) > fabs(rtv))
				rtv = current ;
			//fprintf(mapeout, "node %s, current %.2lf\n", node->name, current) ;
		}
		i++ ;
	}
	return rtv ;
}

/*
 * mape_pbssce_itr_nn_coswitch
 *
 * judge if two nodes can coswitch together
 * according to the simulation result. 
 */
int
mape_pbssce_itr_nn_coswitch(node1, node2, time)
node_t *node1, *node2 ;
double time ;
{
	int total_times, times = 0 ; 
	lsList cur_points, vol_point1, vol_point2 ;
	int index ;
	double height1, height2 ;
	mape_cur_point_t *point ;
	double value ;
	lsGen gen, gen1, gen2, gen3 ;
	lsHandle h ;
	int i ;

	cur_points = lsCreate() ;
	i = -1 ;

	gen1 = lsStart( ((mape_t *)(node1->mape))->pbssce.cl ) ;
	gen2 = lsStart( ((mape_t *)(node2->mape))->pbssce.cl ) ;
	gen3 = lsStart( esl ) ; 

	total_times = lsLength(esl) ;
	for(;;){
		if(lsNext(gen3, (lsGeneric *)&index, &h) != LS_OK)
			break ;

		while(i < index){
			if(lsNext(gen1, (lsGeneric *)&vol_point1, &h) != LS_OK)
				break ;
			if(lsNext(gen2, (lsGeneric *)&vol_point2, &h) != LS_OK)
				break ;
			i++ ;
		}

		/* get the current value at the time point */
		while(lsLength(cur_points)){
			lsDelEnd(cur_points, (lsGeneric *)&point) ;
			free(point) ;
		}
		mape_pbssce_add_node_current_v(node1, vol_point1, cur_points) ;
		value = mape_cur_point_value(cur_points, time) ;
		if(value == UNKNOWN_VOL)
			value = 0.0 ;
		if(value >= -5.0)
			continue ;

		/* get the current value at the time point */
		while(lsLength(cur_points)){
			lsDelEnd(cur_points, (lsGeneric *)&point) ;
			free(point) ;
		}
		mape_pbssce_add_node_current_v(node2, vol_point2, cur_points) ;
		value = mape_cur_point_value(cur_points, time) ;
		if(value == UNKNOWN_VOL)
			value = 0.0 ;
		if(value >= -5.0)
			continue ;

		lsFinish(gen1) ;	
		lsFinish(gen2) ;	
		lsFinish(gen3) ;
		lsDestroy(cur_points, free) ;
		return TRUE ;
	}

	lsFinish(gen1) ;
	lsFinish(gen2) ;
	lsFinish(gen3) ;

	lsDestroy(cur_points, free) ;
	return FALSE ;
}

/*
 * mape_pbssce_set_node_value
 *
 * set a temperate value for the node. 
 */
void 
mape_pbssce_set_node_value(node, value)
node_t *node ;
double value ;
{
	((mape_t *)(node->mape))->pbssce.tmpr = value ;
}

/*
 * mape_pbssce_get_node_value
 *
 * get a temperate value for the node. 
 */
double 
mape_pbssce_get_node_value(node)
node_t *node ;
{
	
	return ((mape_t *)(node->mape))->pbssce.tmpr  ;
}

/*
 * mape_pbssce_node_value_cmp
 *
 * compare node's value
 */
int 
mape_pbssce_node_value_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	return PB_TMPR( ((node_t *)(obj1)) ) >  PB_TMPR( ((node_t *)(obj2)) )  ;
}


/*
 * mape_pbssce_update_co_list
 *
 * up date co list. 
 */
void 
mape_pbssce_update_co_list(zo, node, l)
mape_pbssce_zo_t *zo ;
node_t *node ;
lsList l ;
{
	lsGen gen ;
	node_t *p ;
	
	if(zo->co_list == NULL)
		zo->co_list = lsCreate() ;

	lsForeachItem(l, gen, p)
		if(p != node && !mape_node_in_list(p, zo->co_list))
			lsNewEnd(zo->co_list, (lsGeneric)p, LS_NH) ;
}

/*
 * mape_pbssce_itr_update_zol
 *
 * iteratively update node's zol 
 */
int
mape_pbssce_itr_update_zol(network, tl, l, ci, index)
network_t *network ;
lsList l ;
int *ci ;
int index ;
{
	lsGen gen, gen1, gen2 ;
	node_t *node, *tmp ;
	lsList cur_points ;
	mape_pbssce_zo_t *zo ;
	double current ;
	double total_current = 0.0 ;
	int i ;
	lsList sl[NUM_CANDIDATE_TIME] ;
	lsHandle h ;
	int bi ;

	for(i=0;i<NUM_CANDIDATE_TIME;i++)
		sl[i] = NULL ;

	//fprintf(mapeout, "\nUpdate zol ci=%d, index=%d:\n", *ci, index) ;
	lsForeachItem(l, gen, node){
		bi = *ci ;

		while(bi < index){
			if(lsNext(PB_GEN(node), (lsGeneric *)&cur_points, &h) != LS_OK){
				lsFinish(gen) ;
				return FALSE ;
			}
			bi ++ ;
		}

		i = 0 ;
		//fprintf(mapeout, "\tnode=%s\n", node->name) ; 
		lsForeachItem(PB_ZOL(node), gen1, zo){
			current = mape_cur_point_value(cur_points, zo->time) ;
			if(current == UNKNOWN_VOL)
				current = 0.0 ;
			total_current += current ;
			/*
			fprintf(mapeout, "\t\ttime=%.2lf, current=%.2lf, zo->current=%.2lf\n", 
				zo->time, current,  zo->current) ;
			*/
			if(current < -0.5){
				if(sl[i] == NULL)
					sl[i] = lsCreate() ;
				lsNewEnd(sl[i], (lsGeneric)node, LS_NH) ; 
			}
			if(current  < zo->current )
				zo->current = current ;
			i++ ;
		}
	}

	*ci = index ;

	//fprintf(mapeout, "total current = %.2lf\n", total_current) ;
	/*
	if(sl[0] != NULL){
		fprintf(mapeout, "sl[0]=[") ;
		lsForeachItem(sl[0], gen, node)
			fprintf(mapeout, "%s\t", node->name) ;
		fprintf(mapeout, "]\n") ;
	}
	*/


	lsForeachItem(l, gen, node){
		i=0 ;
		lsForeachItem(PB_ZOL(node), gen1, zo){
			if(mape_node_in_list(node, sl[i])){
				mape_pbssce_update_co_list(zo, node, sl[i]) ;
			}
			/*
			fprintf(mapeout, "node %s's co_list: ", node->name) ;
			if(zo->co_list){
				lsForeachItem(zo->co_list, gen2, tmp)
					fprintf(mapeout, "%s\t", tmp->name) ;
			}
			fprintf(mapeout, "\n") ;
			*/
			i++ ;
		}
	}
	
	for(i=0;i<NUM_CANDIDATE_TIME;i++){
		if(sl[i] != NULL)
			lsDestroy(sl[i], NULL) ;
	}
	return TRUE ;
}

/*
 * mape_pbssce_one_order_valid_current
 *
 * return the maximum current value of a node, considering
 * first order logic conflication
 */
double
mape_pbssce_one_order_valid_current(t, node, l)
double t ;
node_t *node ;
lsList l ;
{
	lsGen gen, gen1 ;
	mape_pbssce_zo_t *zo, *zo1;
	node_t *p ;
	lsList co_list ;

	lsForeachItem(PB_ZOL(node), gen, zo){
		if(zo->time == t){
			lsFinish(gen) ;
			break ;
		}
	}

	if(zo->current >= -1.0)
		return zo->current ;

	lsForeachItem(l, gen, p){
		if(p == node)
			continue ;

		if(!mape_node_in_list(p, zo->co_list)){
			lsForeachItem(PB_ZOL(p), gen1, zo1){
				if(zo1->time != t)
					continue ;
				if(zo1->current < zo->current){
					lsFinish(gen1);
					lsFinish(gen) ;
					return 0.0 ;
				}
			}
		}
	}

	return zo->current ;
}

/*
 * mape_pbssce_node_zo
 *
 * return the zo. 
 */
mape_pbssce_zo_t *
mape_pbssce_node_zo(node, t)
node_t *node ;
double t ;
{
	lsGen gen ;
	mape_pbssce_zo_t *rtv ;

	lsForeachItem(PB_ZOL(node), gen, rtv){
		if(fabs(rtv->time - t) <= 0.001){
			lsFinish(gen) ;
			return rtv ;
		}
	}
	return rtv ;
}

/*
 * mape_pbssce_time_list_coswitch_current
 *
 * get the coswitch current for a list of gate. 
 */
double 
mape_pbssce_time_list_coswitch_current(l)
lsList l ;
{
	lsGen gen, gen1 ;
	lsList bl, tl ;
	node_t *node, *p ;
	double current, mc ;
	mape_pbssce_zo_t *zo, *zo1 ;
	lsHandle h ;
	
	/*
	fprintf(mapeout, "get coswitch current for: \n") ;
	lsForeachItem(l, gen, node)
		fprintf(mapeout, "\t%s", node->name) ;
	fprintf(mapeout, "\n") ;
	*/

	mc = 0.0 ;
	bl = lsCreate() ;
	tl = lsCopy(l, NULL) ;

	while(lsLength(tl)){
		lsDelEnd(tl, (lsGeneric *)&node) ;
	//lsForeachItem(l, gen, node){
		while(lsLength(bl))
			lsDelEnd(bl, (lsGeneric *)&p) ;

		zo = PB_CURZO(node) ;
		if(zo->co_list){
			lsForeachItem(zo->co_list, gen1, p)
				if(mape_node_in_list(p, tl) && p!= node)
				//if(mape_node_in_list(p, l) && p!= node)
					lsNewEnd(bl, (lsGeneric)p, LS_NH) ;
		}

		if(lsLength(bl) == 0){
			current = zo->current ;
		}else if(lsLength(bl) == 1){
			lsFirstItem(bl, (lsGeneric *)&p, LS_NH) ;
			zo1 = PB_CURZO(p) ;
			current = zo->current + zo1->current ;
		}else 
			current = zo->current + mape_pbssce_time_list_coswitch_current(bl) ;

		if(mc > current)
			mc = current ;
	}
	lsDestroy(bl, NULL) ;
	lsDestroy(tl, NULL) ;
	return mc ;
}

#ifdef THIRD_ORDER
/*
 * mape_pbssce_value3_clean 
 *
 * clean all the previous result. 
 */
mape_pbssce_value3_clean(l)
lsList l ;
{
	value3.max_value = 0.0 ;
	value3.gsize = lsLength(l) ;
	if (value3.co_matrix != NULL){
		free(value3.co_matrix) ;
		value3.co_matrix = NULL ;
	}

	if(value3.ol != NULL)
		lsDestroy(value3.ol, free) ;
	value3.ol = lsCreate() ;
	if(value3.gl != NULL)
		lsDestroy(value3.gl, free) ;
	value3.gl = lsCreate() ;
	if(value3.sl != NULL)
		lsDestroy(value3.sl, free) ;
	value3.sl = lsCreate() ;
}

/*
 * mape_pbssce_value3_create_co_t
 *
 *
 * create value3_co_t structure. 
 */
mape_pbssce_value3_co_t *
mape_pbssce_value3_create_co_t(void)
{
	mape_pbssce_value3_co_t *rtv ;

	rtv		= ALLOC(mape_pbssce_value3_co_t, 1) ;
	rtv->max_value 	= 0.0 ;
	memset(rtv->nodes, 0, sizeof(int)*MAXGROUPSIZE) ;

	return rtv ;
}

/*
 * mape_pbssce_value3_matrix_index
 *
 *
 * return the matrix index. 
 */
int
mape_pbssce_value3_matrix_index(i, j, k)
int i, j, k ;
{
	int m, n, u ;

	m = MAX(i, MAX(j, k)) ;
	n = MIN(i, MIN(j, k)) ;
	u = (i+j+k)-m-n ;

	return m-2 + (u-1)*(value3.gsize) + n*(value3.gsize)*value3.gsize ;
}

/*
 * mape_pbssce_value3_update_matrix
 *
 * update the coswitch matrix for third order. 
 */
void 
mape_pbssce_value3_update_matrix(void)
{
	lsGen gen ;
	mape_pbssce_value3_co_t *co ;
	int i, j, k ;

	value3.co_matrix = ALLOC(int, value3.gsize*value3.gsize*value3.gsize) ;
	memset(value3.co_matrix, 0, sizeof(int)*(value3.gsize*value3.gsize*value3.gsize)) ;

	lsForeachItem(value3.ol, gen, co){
		for(i=0;i<=value3.gsize-3;i++){
			if((co->nodes)[i] == 0)
				continue ;
			for(j=i+1;j<=value3.gsize-2;j++){
				if((co->nodes)[j] == 0)
					continue ;
				for(k=j+1;k<=value3.gsize-1;k++){
					if((co->nodes)[k] == 0)
						continue ;
					(value3.co_matrix)[mape_pbssce_value3_matrix_index(i, j, k)] = 1 ;
				}
			}
		}
	}

	/* dump the matrix */
	/*
	for(i=0;i<value3.gsize*value3.gsize*value3.gsize;i++)
		fprintf(mapeout, "co_matrix[%d] = %d\n", i, (value3.co_matrix)[i]) ; 
	*/
}

/*
 * mape_pbssce_value3_list_contain
 *
 * if co1 is being contained in co2 ;
 */
int 
mape_pbssce_value3_list_contain(co1, co2)
mape_pbssce_value3_co_t *co1, *co2 ;
{
	int i ;

	for(i=0;i<value3.gsize;i++)
		if((co1->nodes)[i] == 1 && (co2->nodes)[i] != 1)
			return FALSE ;

	return TRUE ;
} 

/*
 * mape_pbssce_value3_update_ol
 *
 * update value3's original list. 
 */
int 
mape_pbssce_value3_update_ol(co)
mape_pbssce_value3_co_t *co ;
{
	mape_pbssce_value3_co_t *nco, *p, *tmp ;
	int contain = FALSE ;
	lsGen gen ;
	int i ;
	int rtv = FALSE ;

	lsForeachItem((value3.ol), gen, p){
		if(mape_pbssce_value3_list_contain(co, p)){
			lsFinish(gen) ;
			contain = TRUE ;
			if(fabs(p->max_value) < fabs(co->max_value)){
				p->max_value = co->max_value ;
				rtv = TRUE ;
			}
			break ;
		}else if(mape_pbssce_value3_list_contain(p, co)){
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
			free(tmp) ;
			rtv = TRUE ;
		}
	}

	if(!contain){
		rtv = TRUE ;
		nco = mape_pbssce_value3_create_co_t() ;
		for(i=0;i<value3.gsize;i++)
			(nco->nodes)[i] = (co->nodes)[i] ;
		nco->max_value = co->max_value ;
		lsNewEnd(value3.ol, (lsGeneric)nco, LS_NH) ;
	}

	return rtv ;
}

/*
 * mape_pbssce_value3_co_current
 *
 * return the co current. 
 */
double
mape_pbssce_value3_co_current(co, l) 
mape_pbssce_value3_co_t *co ;
lsList l ;
{
	int i ;
	lsGen gen ;
	node_t *node ;
	double mc[MAXGROUPSIZE], total_current;

	i = 0 ;
	lsForeachItem(l, gen, node)
		mc[i++] = PB_TMPR(node) ;

	/* calculate the max current value */
	total_current = 0.0 ;
	for(i=0;i<value3.gsize;i++){
		if((co->nodes)[i])
			total_current += mc[i];
	}
	return total_current ;
}

/*
 * mape_pbssce_static_cur_value
 *
 * get the static current value. 
 */
double 
mape_pbssce_static_cur_value(s, t)
mape_pbssce_static_t *s ;
double t ;
{
	double rtv = 0.0 ;

	if(s->s_time < t && t <= s->time ){
		rtv = s->m_value * (t - s->s_time) / (s->time - s->s_time) ;
	}else if( s->time < t && t < s->e_time){
		rtv = s->m_value * (s->e_time - t) / (s->e_time - s->time) ;
	}

	return rtv ;
}

/*
 * mape_pbssce_value3_node_add_cant. 
 *
 * add cant into node's cantl. 
 */
void 
mape_pbssce_value3_node_add_cant(node, cant)
node_t *node ;
mape_pbssce_value3_cant_t *cant ;
{
	if(PB_CANTL(node) == NULL)
		PB_CANTL(node) = lsCreate() ;
	lsNewEnd(PB_CANTL(node), (lsGeneric)cant, LS_NH) ;
}

/*
 * mape_pbssce_value3_checkout_cant
 *
 * find a cant by time. 
 */
int 
mape_pbssce_value3_checkout_cant(node, t, cant)
node_t *node ;
double t ;
mape_pbssce_value3_cant_t **cant ;
{
	lsGen gen ;
	mape_pbssce_value3_cant_t *p ;

	if(PB_CANTL(node) == NULL){
		*cant = NULL ;
		return FALSE;
	}

	lsForeachItem(PB_CANTL(node), gen, p){
		if(fabs(p->time - t) <= 0.01){
			lsFinish(gen) ;
			*cant = p  ;
			return TRUE ;
		}
	}

	*cant = NULL ;
	return FALSE ;
}

/*
 * mape_pbssce_value3_pre
 *
 * prepare pharse for value3. 
 */
void
mape_pbssce_value3_pre(network, l, t)
network_t *network ;
lsList l ;
double t ;
{
	lsGen gen, gen1 ;
	node_t *node ;
	mape_pbssce_value3_co_t *co;
	double current ;
	int finished ;
	lsList cur_list ;
	int i, j ;
	mape_cur_point_t *p ;
	lsHandle h ;
	mape_pbssce_static_t *s ;
	mape_pbssce_value3_cant_t *cants[MAXGROUPSIZE], *ncants[MAXGROUPSIZE];
	mape_pbssce_value3_cant_t *cant ;

	/* set gen value and set node's max current value to be zero */
	lsForeachItem(l, gen, node){
		PB_GEN(node) = lsStart(PB_CL(node)) ;
		PB_TMPR(node) = 0.0 ;
	}

	/* get node's maximum value according to the switching statics*/
	lsForeachItem(l, gen, node){
		lsForeachItem(PB_STATIC(node), gen1, s){
			current = mape_pbssce_static_cur_value(s, t) ;
			if(fabs(PB_TMPR(node)) < fabs(current))
				PB_TMPR(node) = current ;
		}
	}
	
	/* check if we can find the candidate time stored */
	memset(cants, 0, sizeof(mape_pbssce_value3_cant_t *)*MAXGROUPSIZE) ;
	memset(ncants, 0, sizeof(mape_pbssce_value3_cant_t *)*MAXGROUPSIZE) ;
	lsForeachItem(l, gen, node){
		if(mape_pbssce_value3_checkout_cant(node, t, &cant)){
			//fprintf(mapeout, "find cant: %.2lf\n", cant->time) ;
			cants[PB_INDEX(node)] = cant ;
		}
	}
	
	finished = FALSE ;
	co = mape_pbssce_value3_create_co_t();
	for(i=0;i<codb_times;i++){
		memset(co->nodes, 0, sizeof(int)*MAXGROUPSIZE) ;
		co->max_value = 0.0 ;
		lsForeachItem(l, gen, node){
			if( ((cant = cants[PB_INDEX(node)]) != NULL) && use_cant){
				if(cant->cv == NULL)
					continue ;
				current = (double)((cant->cv)[i]) ;
				if(fabs(current) >= 0.5) {
					(co->nodes)[PB_INDEX(node)] = 1 ;
					co->max_value += current ;
				}
				continue ;
			}
			if(fabs(PB_TMPR(node)) <= 0.5 ){
				if(i == 0){
					cant = mape_pbssce_value3_create_cant_t() ; 	
					ncants[PB_INDEX(node)] = cant ;
					cant->time = t ;  
					mape_pbssce_value3_node_add_cant(node, cant) ;
				}
				continue ;	
			}
			if(lsNext(PB_GEN(node), (lsGeneric *)&cur_list, &h) != LS_OK){
				finished = TRUE ;
				break ;
			}
			current = mape_cur_point_value(cur_list, t) ;

			/*
			lsForeachItem(cur_list, gen1, p){
				fprintf(mapeout, "[%.2lf, %.2lf]", p->time, p->value) ;
			}
			fprintf(mapeout, "\n\tcurrent = %.2lf\n", current) ;
			*/

			if(i == 0){
				cant = mape_pbssce_value3_create_cant_t();
				ncants[PB_INDEX(node)] = cant ;
				cant->time = t ;
				cant->cv = ALLOC(float, codb_times) ;
				mape_pbssce_value3_node_add_cant(node, cant) ;
			}
			if(current == UNKNOWN_VOL)
				current = 0.0 ;
			if(fabs(current) >= 0.5) {
				(co->nodes)[PB_INDEX(node)] = 1 ;	/* add node into co switch list*/
				co->max_value += current ;
			}
			((ncants[PB_INDEX(node)])->cv)[i] = (float)current ; 
		}
		/*
		fprintf(mapeout, "\t %dth value co switch list: (current = %.2f) ", i, co->max_value) ;
		for(j=0;j<value3.gsize;j++)
			fprintf(mapeout, "%d", (co->nodes)[j]) ;	
		fprintf(mapeout, "\n") ;
		*/

		mape_pbssce_value3_update_ol(co) ;
	}
	free(co);
	lsForeachItem(l, gen, node)
		lsFinish(PB_GEN(node)) ;
	/*
	i = 0 ;
	lsForeachItem(value3.ol, gen, co){
		i ++ ;
		fprintf(mapeout, "\t\t Final %dth ol: (current = %.2f)", i, co->max_value) ; 
		for(j=0;j<value3.gsize;j++)
			fprintf(mapeout, "%d", (co->nodes)[j]) ;	
		fprintf(mapeout, "\n") ;
	}
	*/

	mape_pbssce_value3_update_matrix() ;
}

/*
 * mape_pbssce_co_num_cmp
 *
 * compare two co by it's switch node's number. 
 */
int 
mape_pbssce_co_num_cmp(obj1, obj2)
lsGeneric *obj1, *obj2 ;
{
	mape_pbssce_value3_co_t *co1, *co2 ;
	int n1, n2, i ;

	co1 = (mape_pbssce_value3_co_t *)obj1 ;
	co2 = (mape_pbssce_value3_co_t *)obj2 ;

	n1 = n2 = 0 ;

	for(i=0;i<value3.gsize;i++){
		if((co1->nodes)[i])
			n1 ++ ;
		if((co2->nodes)[i])
			n2 ++ ;
	}
	return (n1 >= n2) ; 
}

/*
 * mape_pbssce_value3_matrix_value
 *
 * return the matrix value after the i, j, k specification. 
 */
int 
mape_pbssce_value3_matrix_value(i, j, k)
int i, j, k ;
{
	if(i == j || j == k || i == k)
		return 1 ;

	return (value3.co_matrix)[mape_pbssce_value3_matrix_index(i, j, k)] ;
}


/*
 * mape_pbssce_value3_compatible. 
 *
 * if a position is compatible with a co
 */
int 
mape_pbssce_value3_compatible(pos, co) 
int pos ;
mape_pbssce_value3_co_t *co ;
{
	int i, j ;	


	for(i=0;i<value3.gsize-2;i++){
		if((co->nodes)[i] == 0)
			continue ;
		for(j=i+1;j<value3.gsize-1;j++){
			if((co->nodes)[j] == 0)
				continue ;
			if(mape_pbssce_value3_matrix_value(pos, i, j) == 0)
				return FALSE ;
		}
	}

	/*
	fprintf(mapeout, "\tFind compatible position with co pos=%d: ", pos) ;  
	for(i=0;i<=value3.gsize-1;i++)
		fprintf(mapeout, "%d", (co->nodes)[i]) ;
	fprintf(mapeout, "\n") ;
	*/

	return TRUE ;
}

/*
 * mape_pbssce_value3_co_grow
 *
 * grow of one co to gl list. 
 */
int 
mape_pbssce_value3_co_grow(co, l)
mape_pbssce_value3_co_t *co ;
lsList l ;
{
	lsGen gen ;
	int rtv = FALSE, i, j ;
	mape_pbssce_value3_co_t *nco, *p ;
	double mc[MAXGROUPSIZE] ;
	node_t *node ;

	/* co is contained by one co in gl */
	lsForeachItem(value3.gl, gen, p){
		if(mape_pbssce_value3_list_contain(co, p)){
			lsFinish(gen) ;
			free(co) ;
			return TRUE ;
		}
	}
	
	i = 0 ;
	lsForeachItem(l, gen, node){
		mc[i] = PB_TMPR(node) ; 
		i++ ;
	}

	/* not contained */
	for(i=0;i<value3.gsize;i++){
		if((co->nodes)[i] != 0)
			continue ;
		if(mape_pbssce_value3_compatible(i, co)){
			nco = mape_pbssce_value3_create_co_t() ;
			for(j=0;j<value3.gsize;j++)
				(nco->nodes)[j] = (co->nodes)[j] ;
			nco->max_value = co->max_value ;
			(nco->nodes)[i] = 1 ;
			nco->max_value += mc[i] ;
			lsNewEnd(value3.gl, (lsGeneric)nco, LS_NH) ;
			rtv = TRUE ;
		}
	}
	if(rtv)
		free(co) ;
	return rtv ;
}

/*
 * mape_pbssce_value3_group_grow
 *
 * group grow process. 
 */
void 
mape_pbssce_value3_group_grow(l)
lsList l ;
{
	lsGen gen ;
	mape_pbssce_value3_co_t *co ;
	int i ;
	double total_current, mc[MAXGROUPSIZE] ;
	node_t *node ;

	/* grow process */
	while(lsLength(value3.ol)){
		lsSort(value3.ol, mape_pbssce_co_num_cmp) ;
		while(lsLength(value3.ol)){
			lsDelEnd(value3.ol, (lsGeneric *)&co) ; 
			if(!mape_pbssce_value3_co_grow(co, l))
				lsNewEnd(value3.sl, (lsGeneric)co, LS_NH) ;
		}
		while(lsLength(value3.gl)){
			lsDelEnd(value3.gl, (lsGeneric *)&co) ;
			lsNewEnd(value3.ol, (lsGeneric)co, LS_NH) ;
		}
	}

	/* put the maximum current value into mc matrix*/
	/*
	i = 0 ;
	lsForeachItem(l, gen, node){
		mc[i] = PB_TMPR(node) ;
		i++ ;
	}
	*/

	/* calculate the max current value */
	value3.max_value = 0.0 ;
	while(lsLength(value3.sl)){
		lsDelEnd(value3.sl, (lsGeneric *)&co) ;
		/*
		for(i=0;i<value3.gsize;i++){
			if((co->nodes)[i]) 
				total_current += mc[i];
		}
		*/
		if(fabs(value3.max_value) < fabs(co->max_value))
			value3.max_value = co->max_value ;
		free(co) ;
	}
}

/*
 * mape_pbssce_value3
 *
 * under a specific time, calcualte the 3rd order 
 * maximum coswitching current for node list l ;
 */
double
mape_pbssce_value3(network, l, time)
network_t *network ;
lsList l ;
double time ;
{
	lsGen gen ;
	node_t *node ;
	int i = 0 ;

	/* clean the previous result */
	mape_pbssce_value3_clean(l) ;

	/* set node's index */
	lsForeachItem(l, gen, node){
		PB_INDEX(node) = i ;
		i ++ ;
	}

	/* preparation pharse */
	mape_pbssce_value3_pre(network, l, time) ;

	/* grow the groups */
	mape_pbssce_value3_group_grow(l) ;

	/* return the max current value */
	return value3.max_value ;
}

/*
 * mape_pbssce_itr_estimate_current_value3
 *
 * estimate the current value in the 3rd order. 
 */
void
mape_pbssce_itr_estimate_current_value3(network, tl, l)
network_t *network ;	// network ;
lsList tl ;	// time list. 
lsList l ;	// target node list ;
{
	lsGen gen ;	
	mape_pbssce_tv_t *tvp ;

	lsForeachItem(tl, gen, tvp){
		//fprintf(mapeout, "Value3 estimation maximum current at time: %.2lf\n", tvp->time) ; 
		tvp->value = mape_pbssce_value3(network, l, tvp->time) ;
	}
}
#endif

/* 
 * mape_pbssce_itr_estimate_current_value
 *
 * estimate the maximum current value for time 
 */
void 
mape_pbssce_itr_estimate_current_value(network, tl, l) 
network_t *network ;  // network ;
lsList tl ; 	// time list; 
lsList l ;	// target node list ; 
{
	lsGen gen, gen1, gen2;
	node_t *node, *p ;
	mape_pbssce_tv_t *tvp ;
	mape_pbssce_zo_t *zo ;
	int i  ;
	double current ;
	lsHandle h ;
	int index ;

	lsForeachItem(l, gen, node){
		PB_GEN(node) = lsStart(PB_CL(node)) ;
		PB_ZOL(node) = lsCreate() ;
		lsForeachItem(tl, gen1, tvp){
			zo = mape_pbssce_create_zo_t(tvp->time);
			lsNewEnd(PB_ZOL(node), (lsGeneric)zo, LS_NH) ;
		}
	}

	gen = lsStart( esl ) ; 
	i = -1 ;
	for(;;){
		//fprintf(mapeout, "i= %d\t", i) ;
		if(lsNext(gen, (lsGeneric *)&index, &h) != LS_OK)
			break ;
		if(!mape_pbssce_itr_update_zol(network, tl, l, &i, index))
			break ;
	}

	lsFinish(gen) ;

	lsForeachItem(tl, gen, tvp){
		lsForeachItem(l, gen1, node){
			PB_CURZO(node) = mape_pbssce_node_zo(node, tvp->time) ;  
			/*
			fprintf(mapeout, "node %s's co_list time=%.2lf\n", node->name, tvp->time) ;
			if(PB_CURZO(node)->co_list)
				lsForeachItem(PB_CURZO(node)->co_list, gen2, p)
					fprintf(mapeout, "\t%s", p->name) ;
			fprintf(mapeout, "\n") ;
			*/
		}
		tvp->value = mape_pbssce_time_list_coswitch_current(l) ;
	}

	lsForeachItem(l, gen, node){
		lsFinish(PB_GEN(node)); 
		lsForeachItem(PB_ZOL(node), gen, zo)
			mape_pbssce_destroy_zo_t(zo) ;
	}
}

/*
 * mape_pbssce_itr_estimate_current_value
 *
 * estimate the maximum current value for time 
 */
void
mape_pbssce_itr_estimate_current_value_o(network, tl, l) 
network_t *network ;  // network ;
lsList tl ; 	// time list; 
lsList l ;	// target node list ; 
{
	lsGen gen, gen1, gen2 ;
	mape_pbssce_tv_t *tvp ;
	node_t *node, *node1, *node2 ;
	double current ;
	double c1, c2 ;

	lsForeachItem(tl, gen, tvp){
		tvp->value = 0.0 ;
		lsForeachItem(l, gen1, node){
			current = mape_pbssce_itr_node_current(node, tvp->time) ;
			mape_pbssce_set_node_value(node, current) ;
			tvp->value += current ;
		}
		/*if the estimated value is 10% higher than the low boundary 
		  reduce the value further */
		if(100 * (tvp->value - tvp->l_value) / tvp->l_value > ACCEPT_PECT ){
			lsForeachItem(l, gen1, node1){
				c1 = mape_pbssce_get_node_value(node1) ;
				if(fabs(c1) <= 5.0)
					continue ;
				lsForeachItem(l, gen2, node2){
					if(node1 == node2)
						continue ;
					c1 = mape_pbssce_get_node_value(node1) ;
					if(fabs(c1) <= 5.0){
						lsFinish(gen2) ;
						break ;
					}
					c2 = mape_pbssce_get_node_value(node2) ;
					if(fabs(c2) <= 5.0)
						continue ;
					if(!mape_pbssce_itr_nn_coswitch(node1, node2, tvp->time)){
						if(fabs(c1) < fabs(c2))
							mape_pbssce_set_node_value(node1, 0.0) ;
						else 
							mape_pbssce_set_node_value(node2, 0.0) ;
					}
				}
			}
			tvp->value = 0.0 ;
			lsForeachItem(l, gen1, node)
				tvp->value += mape_pbssce_get_node_value(node) ;

			if(fabs(tvp->value) < fabs(tvp->l_value))
				tvp->value = tvp->l_value ;
		}
	}
}

/*
 * mape_pbssce_itr_estimate_current
 *
 * iterative method to estimate the current 
 */
double
mape_pbssce_itr_estimate_current(network, l)
network_t *network ;
lsList l ;
{
	double time, cur, l_cur;
	lsList tl ;
	lsGen gen ;
	mape_pbssce_tv_t *tvp ;

	tl = lsCreate() ;
	//mape_timer_reset() ;
	//mape_timer_start() ;
#ifndef THIRD_ORDER
	mape_pbssce_itr_estimate_current_time(network, tl, l) ;
#else
	mape_pbssce_itr_estimate_current_time3(network, tl, l) ;	
#endif
	//mape_timer_stop() ;
	//mape_timer_report("EStimate_current_time") ; 

	//mape_timer_reset() ;
	//mape_timer_start() ;
#ifndef THIRD_ORDER
	mape_pbssce_itr_estimate_current_value(network, tl, l) ;
#else
	mape_pbssce_itr_estimate_current_value3(network, tl, l) ;
#endif
	//mape_timer_stop() ;
	//mape_timer_report("EStimate_current_value") ; 

	cur = 0.0 ;
	time = 0.0 ;
	lsForeachItem(tl, gen, tvp){
		if(fabs(tvp->value) > fabs(cur)){
			cur = tvp->value ;
			l_cur = tvp->l_value ;
			time = tvp->time ;
		}
	}
	//fprintf(mapeout, "PBSSCE Estimate maximum current %.2lf at time %.2lf\n", cur, time) ;


	/*
	if( 100*(cur - l_cur)/l_cur <= ACCEPT_PECT )
		fprintf(mapeout, "PBSSCE Estimate maximum current %.2lf at time %.2lf\n", cur, time) ;
	else {
		mape_pbssce_itr_estimation_reduce(network, tl, l) ;
		cur = 0.0 ;
		time = 0.0 ;
		lsForeachItem(tl, gen, tvp){
			if(fabs(tvp->value) > fabs(cur)){
				cur = tvp->value ;
				l_cur = tvp->l_value ;
				time = tvp->time ;
			}
		}
		fprintf(mapeout, "PBSSCE Estimate maximum current %.2lf at time %.2lf\n", cur, time) ;
	}
	*/

	lsDestroy(tl, free) ;
	return cur ;
}

/*
 * mape_pbssce_grp
 *
 * possibility based SSC estimation for a group. 
 */
void 
mape_pbssce_grp(network)
network_t *network ;
{
	lsList groups, group, sgroup ;
	lsGen gen2, gen1 ;
	node_t *node ;
	double cs ;
	int i ;

	groups = mape_group_get_groups(network) ;

	/* strongly co-switch group */
	sgroup = lsCreate() ;
	srandom(util_cpu_time()) ;

	/*set the lib to be various delay mode, where 
	  rising and falling have the same delay a gate. */
	mape_curlib_set_various_mode(TRUE) ;
	mape_pbssce_build_coswitch_db(network) ;
	//mape_curlib_set_unit_mode(TRUE) ;

	/* get the switching possibility for every gate at every time point*/
	//mape_pbssce_st_possibility(network) ;

	i = 0 ;
	lsForeachItem(groups, gen1, group){
		i ++ ;
		if(i%GROUP_SAMPLE_RATE != 0) 
			continue ;
		mape_clean_given_gate() ;

		/*cs = mape_pbssce_group_coswitch(group) ;
		fprintf(mapeout, "coswitch possibility is: %.2lf\n", cs) ; 
		*/

		fprintf(mapeout, "Group: ") ;
		lsForeachItem(group, gen2, node)
			fprintf(mapeout, "%s ", node->name) ;
		fprintf(mapeout, "\n") ;

		//mape_pbssce_dump_group_codb(group) ;
		//mape_pbssce_strong_coswitch_group(group, sgroup) ; 

		while(lsLength(sgroup))
			lsDelEnd(sgroup, (lsGeneric *)&node) ;

		lsForeachItem(group, gen2, node) 
			if(node_type(node) != PRIMARY_INPUT && node_type(node) != PRIMARY_OUTPUT)
				lsNewEnd(sgroup, (lsGeneric)node, LS_NH) ;	

		fprintf(mapeout, "Strong coswitch group: ") ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;

		/*add the possibility together to get the current*/
		//mape_pbssce_estimate_current(network) ;
		mape_pbssce_itr_estimate_current(network, sgroup) ;
	}
	/* dump the st list information */ 
	/*
	mape_pbssce_dump_st(network) ;
	*/
	lsDestroy(sgroup, NULL) ;
}

/*
 * mape_pbssce_set_dbid
 *
 * set node's dbid ;
 */
void 
mape_pbssce_set_dbid(node, id)
node_t *node ;
int id ;
{
	if(!node->mape)
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t*)(node->mape))->C0 = id ;
}

/*
 * mape_pbssce_get_dbid
 * get node's dbid ;
 */
int 
mape_pbssce_get_dbid(node)
node_t *node ;
{
	return ((mape_t*)(node->mape))->C0 ;
}

/*
 * return the value in the db entry. 
 *
 */
int
mape_pbssce_codb_value(id1, id2)
int id1, id2 ;
{
	int index ;
	int a, b, c ;

	if(id1 == INVALID_DB_ID	|| id2 == INVALID_DB_ID || id1 < id2)
		return 0  ;
	if(id1 == id2)
		return codb_times;

	a = id2 ;
	b = id1 ;
	c = codb_maxid ;

	index = (2*c-a)*(a-1)/2 + (b - a) ;
	return codb[index - 1] ;
}

/*
 * mape_pbssce_update_coswitch_db_entry
 *
 * update a entry in the coswitch db. 
 */
void 
mape_pbssce_update_coswitch_db_entry(id1, id2, co)
int id1, id2, co ;
{
	int index ;
	int a, b, c ;

	if(id1 == INVALID_DB_ID	|| id2 == INVALID_DB_ID || id1 == id2 || id1 < id2)
		return  ;

	a = id2 ;
	b = id1 ;
	c = codb_maxid ;

	index = (2*c-a)*(a-1)/2 + (b - a) ;
	codb[index - 1] += co ;
}

/*
 * mape_pbssce_is_coswitch
 *
 * return if coswitch happens for two gates. 
 */
int 
mape_pbssce_is_coswitch(node1, node2)
node_t *node1, *node2 ;
{
	lsList vl1, vl2 ;
	lsGen gen1, gen2 ;
	int rtv = 0 ;
	mape_vol_point_t *vp1 = NULL, *vpp1 = NULL, *vp2 = NULL, *vpp2 = NULL ;
	lsHandle h1, h2 ;
	double ctime1 = 0.0, ctime2 = 0.0, octime1 = 0.0, octime2 = 0.0 ;
	mape_curlib_gate_t *curlib_gate1, *curlib_gate2;
	int f1 = FALSE, f2 = FALSE ;

	vl1 = mape_cur_top_vol_points(node1) ;
	vl2 = mape_cur_top_vol_points(node2) ;
	curlib_gate1 = mape_cur_curlib_gate(node1) ;
	curlib_gate2 = mape_cur_curlib_gate(node2) ;

	if(mape_curlib_value(curlib_gate1, MAPE_CURLIB_IFMAX, 5.0) < 0.0)
		f1 = TRUE ;	
	if(mape_curlib_value(curlib_gate2, MAPE_CURLIB_IFMAX, 5.0) < 0.0)
		f2 = TRUE ;	


	gen1 = lsStart(vl1) ;
	gen2 = lsStart(vl2) ;

	//fprintf(mapeout, "node1=%s, node2=%s\n", node1->name, node2->name) ;
	for(;;){
		if(ctime1 < 0 || ctime2 < 0)
			break ;
		if(ctime1 >= HALF_SLOP_TIME && ctime2 >= HALF_SLOP_TIME && ctime1 <= HALF_SLOP_TIME*20 && ctime2 <= HALF_SLOP_TIME*20
			&& fabs(ctime1 - ctime2) <= 2*SLOP_TIME ){
			rtv = 1;
			break ;
		}

		//fprintf(mapeout, "ctime1=%.2lf, ctime2=%.2lf\n", ctime1, ctime2) ;
		vpp1 = vp1 ; vpp2 = vp2;
		octime1 = ctime1; octime2 = ctime2 ; 
		if(octime1 <= octime2){
			/* get the next transition time point */
			for(;;){
				if(vp1 != NULL && vpp1 != NULL && vp1 != vpp1){
					if(vpp1->value - vp1->value >= MAX_VOLTAGE/2 && f1){
						ctime1 = (vp1->time + vpp1->time)/2 ;
						break ;
					}else if(vp1->value - vpp1->value >= MAX_VOLTAGE/2 && !f1){
						ctime1 = (vp1->time + vpp1->time)/2 ;
						break ;
					}
				}		
				vpp1 = vp1 ;
				if(lsNext(gen1, (lsGeneric *)&vp1, &h1) != LS_OK){
					ctime1 = -1.0 ;
					break ;
				}
			}
		}else{
			/* get the next transistion time point */
			for(;;){
				if(vp2 != NULL && vpp2 != NULL && vp2 != vpp2){
					if(vpp2->value - vp2->value >= MAX_VOLTAGE/2 && f2){
						ctime2 = (vp2->time + vpp2->time)/2 ;
						break ;
					}else if(vp2->value - vpp2->value >= MAX_VOLTAGE/2 && !f2){
						ctime2 = (vp2->time + vpp2->time)/2 ;
						break ;
					}
				}		
				vpp2 = vp2 ;
				if(lsNext(gen2, (lsGeneric *)&vp2, &h2) != LS_OK){
					ctime2 = -1.0 ;
					break ;
				}
			}
		}
	}

	lsFinish(gen1) ;
	lsFinish(gen2) ;
	return rtv ;
}

/*
 * mape_pbssce_update_coswitch_db
 *
 * update the switch db after one random simulation 
 */
void 
mape_pbssce_update_coswitch_db(network) 
network_t *network ;
{
	lsGen gen1, gen2 ;
	node_t *node1, *node2 ;
	int id1, id2 ;

	foreach_node(network, gen1, node1)
		foreach_node(network, gen2, node2){
			id1 = mape_pbssce_get_dbid(node1) ;	
			id2 = mape_pbssce_get_dbid(node2) ;	
			if(id1 == INVALID_DB_ID	|| id2 == INVALID_DB_ID || id1 == id2 || id1 < id2)
				continue  ;
			mape_pbssce_update_coswitch_db_entry(id1, id2, 
				mape_pbssce_is_coswitch(node1, node2) 
			) ;
		}
}

/*
 * mape_pbssce_assign_db_id
 *
 * assign db id to every gate in the circuit. 
 */
void 
mape_pbssce_assign_dbid(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	codb_maxid = 0 ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			mape_pbssce_set_dbid(node, INVALID_DB_ID) ; 
		else 
			mape_pbssce_set_dbid(node, ++codb_maxid) ;
	}
}

/*
 * mape_pbssce_create_coswitch_db
 *
 * create the coswitch db. 
 */
void 
mape_pbssce_create_coswitch_db(network)
network_t *network ;
{
	codb = ALLOC(int, (codb_maxid*(codb_maxid -1))/2) ;
	memset(codb, 0, (codb_maxid*(codb_maxid -1))/2*sizeof(int)) ;
	fprintf(mapeout, "Total number of index: %d\n", codb_maxid*(codb_maxid -1)/2) ;
}

/*
 * mape_pbssce_dump_codb
 *
 * dump the coswitch db for debugging.  
 */
void 
mape_pbssce_dump_codb(void)
{
	int i ;

	for(i=0;i<codb_maxid*(codb_maxid -1)/2;i++){
		if(i%15 == 0)
			fprintf(mapeout, "\n") ;
		fprintf(mapeout, "%.3f ", (float)(codb[i]*1.0/codb_times)) ;  
	}
	fprintf(mapeout, "\n") ;
}

/*
 * mape_pbssce_creat_static_t 
 *
 * create the static time value. 
 */
mape_pbssce_static_t *
mape_pbssce_create_static_t(time, s_time, e_time, m_value, repeat)
double time, m_value, s_time, e_time ;
int repeat ;
{
	mape_pbssce_static_t *rtv ;	

	rtv		= ALLOC(mape_pbssce_static_t, 1) ;
	rtv->time	= time ;
	rtv->s_time 	= s_time ;
	rtv->e_time 	= e_time ;
	rtv->m_value	= m_value ;
	rtv->repeat	= repeat ;

	return rtv ;
}

/*
 * mape_pbssce_tv_statics_update
 *
 * update tv statics. 
 */
void 
mape_pbssce_tv_statics_update(time, s_time, e_time, value, l)
double time, s_time, e_time, value ;
lsList l ;
{
	lsGen gen ;
	mape_pbssce_static_t *s ;

	lsForeachItem(l, gen, s){
		if(fabs(s->time - time) <= 0.5){
			s->repeat ++ ;
			if(fabs(value) > fabs(s->m_value))
				s->m_value = value ;
			lsFinish(gen) ;
			return ;
		}
	}
	
	s = mape_pbssce_create_static_t(time, s_time, e_time, value, 1) ;
	lsNewEnd(l, (lsGeneric)s, LS_NH) ;
}

/*
 * mape_pbssce_node_statics_update
 *
 * update the statics information for node switching. 
 */
void 
mape_pbssce_node_statics_update(node, cl)
node_t *node ; // node. 
lsList cl ; // current list. 
{
	lsGen gen, gen1 ;
	mape_cur_point_t *cp = NULL ;
	mape_cur_point_t *pre1 = NULL ;
	mape_cur_point_t *pre2 = NULL ;
	

	if(PB_STATIC(node) == NULL)
		PB_STATIC(node) = lsCreate() ;

	lsForeachItem(cl, gen, cp){
		if(pre1 && pre2 && cp){
			if(fabs(pre1->value) >= 5.0 && fabs(pre2->value) <= 0.1 && fabs(cp->value) <= 0.1){
				mape_pbssce_tv_statics_update(pre1->time, pre2->time, 
					cp->time, pre1->value, PB_STATIC(node)) ; 
			}
		}
		pre2 = pre1 ;
		pre1 = cp ;
	}
}

/*
 * mape_pbssce_node_backup_simulation
 *
 * try to backup node's simulation
 */
void 
mape_pbssce_node_backup_simulation(node)
node_t *node ;
{
	lsList vol_points, nvl ;
	mape_vol_point_t *vp, *nvp ;
	lsGen gen ;


	vol_points = mape_cur_top_vol_points(node) ;
	nvl = lsCreate() ;
	mape_pbssce_add_node_current_v(node, vol_points, nvl) ;

	/*
	lsForeachItem(vol_points, gen, vp){
		nvp = (mape_vol_point_t *)mape_cur_create_vol_point_t(vp->value, vp->time, vp->multi) ;
		lsNewEnd(nvl, (lsGeneric)nvp, LS_NH) ; 
	}
	*/
	if( ((mape_t *)(node->mape))->pbssce.cl == NULL )
		((mape_t *)(node->mape))->pbssce.cl = lsCreate() ;

	lsNewEnd( ((mape_t *)(node->mape))->pbssce.cl, (lsGeneric)nvl, LS_NH ) ; 
	mape_pbssce_node_statics_update(node, nvl) ;
}

/*
 * mape_pbssce_backup_simulation
 * backup the simulation result for further information. 
 */
void 
mape_pbssce_backup_simulation(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node)
		if(node_type(node) != PRIMARY_INPUT && node_type(node) != PRIMARY_OUTPUT)
			mape_pbssce_node_backup_simulation(node) ;
}

/*
 * mape_pbssce_determ_codb_times
 *
 * determinate the coswitch times. 
 */
void 
mape_pbssce_determ_codb_times(network)
network_t *network ;
{
	int npi, nnode ;

	npi = network_num_pi(network) ;
	nnode = network_num_internal(network) ;
	//codb_times = user_option->rdm_option.times ;
	//codb_times = (int)(80 * sqrt((network_num_internal(network) + network_num_pi(network)))) ;
	codb_times = (int)(60 * sqrt((network_num_internal(network) + network_num_pi(network)))) ;
	if(codb_times > 1500)
		codb_times = 1500 ;
	fprintf(mapeout, "Simulation total times: %d\n", codb_times) ;
}


/*
 * mape_pbssce_coswitch_db
 *
 * build the coswitch database for all the gates 
 * inside the circuit. 
 */
void 
mape_pbssce_build_coswitch_db(network)
network_t *network ;
{
	int i ;
	int sn ;
	int length ;

	/* Assign db id to every gate */
	mape_pbssce_assign_dbid(network) ;
	/* create the coswitch db */
	//mape_pbssce_create_coswitch_db(network) ;

	mape_pbssce_determ_codb_times(network) ;

	mape_timer_reset() ;
	mape_timer_start() ;
	for(i=0;i<codb_times;i++){
		mape_cur_clear(network) ;

		/* simulation and get the total switch number */
		mape_cur_initialize_random(network) ;
		mape_pbssce_backup_simulation(network) ;
		if(i%50 == 0)
			fprintf(mapeout, "Simulation %d\n", i) ;
	}
	mape_timer_stop() ;
	mape_timer_report("ASB Preparation:") ;
}

/*
 * mape_pbssce
 *
 * possibility based SSC estimation. 
 */
void 
mape_pbssce(network)
network_t *network ;
{
	srandom(util_cpu_time()) ;
	/*set the lib to be various delay mode, where 
	  rising and falling have the same delay a gate. */
	mape_curlib_set_various_mode(TRUE) ;
	mape_pbssce_build_coswitch_db(network) ;
	return ;

	/* get the switching possibility for every gate at every time point*/
	mape_pbssce_st_possibility(network) ;

	/*add the possibility together to get the current*/
	mape_pbssce_estimate_current(network) ;

	/* dump the st list information */ 
	/*
	mape_pbssce_dump_st(network) ;
	*/
}

/* 
 * mape_currdmavg
 *
 * random simulation to get the average current. 
 */
void 
mape_rdmavg(network)
network_t *network ;
{
	int i ;
	double time ;
	lsList cur_point ;
	lsGen gen ;
	mape_cur_point_t *p ;

	/*set the lib to be various delay mode, where 
	  rising and falling have the same delay a gate. */
	mape_curlib_set_various_mode(TRUE) ;

	cur_point = lsCreate() ;

	srandom(util_cpu_time()) ;

	for(i=0;i<user_option->rdm_option.times;i++){
		mape_cur_clear(network) ;
		mape_cur_initialize_random(network) ;
		mape_cur_accumulate_current(network, cur_point) ;
	}

	/* dump the information */	
	i = 0 ;
	lsForeachItem(cur_point, gen, p){
		i++ ;
		if(i%10 == 1)
			fprintf(mapeout, "x(%d:%d) = [%.2lf, ", i, i+9, p->time) ;
		else if(i%10 == 0)
			fprintf(mapeout, "%.2lf]\n", p->time) ;
		else 
			fprintf(mapeout, "%.2lf,", p->time) ;
	}
	fprintf(mapeout, "\n") ;

	i = 0 ;
	lsForeachItem(cur_point, gen, p){
		i++ ;
		if(i%10 == 1)
			fprintf(mapeout, "y(%d:%d) = [%.2lf, ", i, i+9, (p->value)/user_option->rdm_option.times) ;
		else if(i%10 == 0)
			fprintf(mapeout, "%.2lf]\n", (p->value)/user_option->rdm_option.times) ;
		else 
			fprintf(mapeout, "%.2lf,", (p->value)/user_option->rdm_option.times) ;
	}
	fprintf(mapeout, "\n") ;

	lsDestroy(cur_point, free) ;
}
