/*
 * mape_pipeline
 *
 * pipelining a combinational circuit by inserting registers.   
 * this file is for ECE734. 
 */
#include "mape_int.h"

lsList cutsets = NULL ;

/*
 * mape_pipeline_push_cutset
 *
 * push a list of node as a cut set. 
 */
void 
mape_pipeline_push_cutset(list)
lsList list ;
{
	if(!cutsets)
		cutsets = lsCreate() ;
	lsNewEnd(cutsets, (lsGeneric)lsCopy(list, NULL), LS_NH) ;
}

/*
 * mape_pipeline_level_cmp
 *
 * compare function for the nodes. 
 */
int
mape_pipeline_level_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	return mape_get_node_level((node_t *)obj1) - 
		mape_get_node_level((node_t *)obj2) ;
}

/*
 * insert register into the node. 
 *
 * 
 */
void 
mape_pipeline_insert_register(node)
node_t *node ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;
		
	((mape_t *)(node->mape))->C0 ++ ;
}

/*
 * is the node has already been inserted register? 
 */
int 
mape_pipeline_is_insert_register(node)
node_t *node ;
{
	if(node->mape == NULL)
		return 0 ;

	return (((mape_t *)(node->mape))->C0 > 0) ;
}

/*
 * mape_pipeline_satisfy_node. 
 *
 *
 * satisfy the node level value to be below the maximum critical path's value
 * by inserting register. We must cut set this target node. 
 */
void 
mape_pipeline_satisfy_node(node, mcp, iNodes)
node_t *node ;
float mcp ;
lsList iNodes ;
{
	node_t *np, *fanin ;
	lsList p_list ;
	lsList n_list ;
	int i, j, step ;
	int all_pi = TRUE ;
	lsGen gen ;

	p_list 	= lsCreate() ;
	n_list 	= lsCreate() ;
	step = (int)mcp ;

	lsNewEnd(n_list, (lsGeneric)node, LS_NH) ; 
	for(i=0;i<step;i++){
		fprintf(mapeout, "\ti=%d\n", i) ;
		while(lsLength(n_list)){
			lsDelEnd(n_list, (lsGeneric *)&np) ;
			lsNewEnd(p_list, (lsGeneric)np, LS_NH) ;
		}
		while(lsLength(p_list)){
			lsDelEnd(p_list, (lsGeneric *)&np) ;
			fprintf(mapeout, "\t\tprocessing node %s\n", np->name) ; 
			if(node_type(np) == PRIMARY_INPUT){
				lsNewEnd(n_list, (lsGeneric)fanin, LS_NH) ;	
			}else {
				foreach_fanin(np, j, fanin)
					if(!mape_pipeline_is_insert_register(fanin))
						lsNewEnd(n_list, (lsGeneric)fanin, LS_NH) ;
			}
		}
	}

	/* are all the nodes PI? */
	lsForeachItem(n_list, gen, np){
		if(node_type(np) != PRIMARY_INPUT){
			all_pi = FALSE ;
			lsFinish(gen) ;
			break ;
		}
	}

	/*transfer the nodes in the n_list into iNodes 
	  if all the nodes are PI, don't do any thing. 
	*/
	if(!all_pi){
		while(lsLength(n_list)){
			lsDelEnd(n_list, (lsGeneric *)&np) ;
			if(mape_node_in_list(np, iNodes))
				continue ;
			mape_pipeline_insert_register(np) ;
			lsNewEnd(iNodes, (lsGeneric)np, LS_NH) ;
		}
	}
	lsDestroy(p_list, NULL) ;
	lsDestroy(n_list, NULL) ;
}

/*
 * mape_pipeline_dump
 *
 * dump the information for register insertion. 
 */
void 
mape_pipeline_dump(network)
network_t *network ;
{
	lsGen gen, gen1 ;
	node_t *node ;
	lsList cut ;

	if(cutsets == NULL)
		return ;

	lsForeachItem(cutsets, gen, cut){
		fprintf(mapeout, "cut: ") ;
		lsForeachItem(cut, gen1, node)
			fprintf(mapeout, "%s ", node->name) ;
		fprintf(mapeout, "\n") ;
	}
}


/*
 * pipeline the circuit until the required critical time 
 * or the minimal time can be achieved. 
 */
void 
mape_pipeline(network)
network_t *network ;
{
	lsList targets; /* current target set of gates to reduce critical path value */
	lsList iNodes;  /* the nodes where register is inserted */
	lsGen gen ;
	node_t *node ;

	targets = lsCreate() ;
	iNodes	= lsCreate() ;

	foreach_primary_output(network, gen, node)
		lsNewEnd(targets, (lsGeneric)node, LS_NH) ;

	for(;;){
		lsSort(targets, mape_pipeline_level_cmp) ;
		while(lsLength(targets)){
			lsDelEnd(targets, (lsGeneric *)&node) ;

			/* reduce the level of the node to be below specified critical path value */
			fprintf(mapeout, "Satisfying node %s\n", node->name) ;
			mape_pipeline_satisfy_node(node, user_option->max_cp, iNodes) ;
		}

		/* finished */
		if(!lsLength(iNodes))
			break ;

		mape_pipeline_push_cutset(iNodes) ;
		/* transfer all the nodes in the iNode into target */
		while(lsLength(iNodes)){
			lsDelEnd(iNodes, (lsGeneric *)&node) ;
			lsNewEnd(targets, (lsGeneric)node, LS_NH) ;
		}
	}

	mape_pipeline_dump(network) ;

	lsDestroy(targets, NULL) ;
	lsDestroy(iNodes, NULL) ;
}
