/*
 * mape_random.c
 * 
 * Maximum power estimation by random input
 *
 * Nov. 5, 2001  Changbo Long(clong@ece.wisc.edu)
 */
#include "mape_int.h"

/*
 * simulate a specific node in the circuit. 
 *
 * mape_rdm_simulate_node
 */
mape_node_value_t
mape_rdm_simulate_node(node)
node_t *node ;
{
	mape_node_value_t o ;
	pcube input ;
	int i ;
	node_t *p ;
	int value, oF, oS ;

	o = mape_rdm_value(node) ;

	if(node_type(node) == PRIMARY_INPUT)
		return o ;

	cube.num_vars = node->nin ;
	define_cube_size(node->nin) ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_rdm_value_first(p) ;
		PUTINPUT(input, i, value) ;
	}
	oF = mape_set_simulate_node(input, node) ;
	foreach_fanin(node, i, p){
		value = mape_rdm_value_second(p) ;
		PUTINPUT(input, i, value) ;
	}
	oS = mape_set_simulate_node(input, node) ;
	o = mape_value_convert(oF, oS) ;

	free_cube(input) ;
	return o ;
}

/*
 * mape_rdm_imp_do
 *
 *
 * for using of the mape_imp utility. 
 */
int 
mape_rdm_imp_do(node, arg)
node_t *node ;
char *arg ;
{
	mape_node_value_t value ;

	value = mape_rdm_simulate_node(node) ;

	if(!mape_value_determined(value))
		mape_error("Determined value expected in random simulation\n", __FILE__, __LINE__) ;
	
	/*input's value already been put into the stack*/
	if(node->type != PRIMARY_INPUT)
		mape_push_rdm(node, value);
	return TRUE ;
}

/*
 * simulate the circuit with all the input node's 
 * value. 
 *
 */
void 
mape_rdm_simulate(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	mape_imp_reset(mape_rdm_imp_do, (char *)NULL) ;

	foreach_primary_input(network, gen, node)
		mape_imp_push(node);
	mape_imp_start() ;
}

/*
 * mape_rdm_transition_fanout
 *
 * calculate the sum of all the transition * #fanout. 
 */
long
mape_rdm_transition_fanout(network)
network_t *network ;
{
	long rtv = 0;
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node){
		if(!mape_value_determined(mape_rdm_value(node))){
			fprintf(mapeout, "node: %s , value: %s\n", node->name, mape_print_node_value(
				mape_rdm_value(node))) ;
			mape_error("Find undetermined value after simulation", __FILE__, __LINE__) ;
		}
		rtv += (TRANSITION_VALUE(mape_rdm_value(node))?1:0) * LSLENGTH(node->fanout) ; 
	}
	return rtv ;
}

/*
 * mape_rdm_copy_to_max
 *
 * copy the random instance into the max
 */
int
mape_rdm_copy_to_max(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node)
		mape_push_rdm_max(node, mape_rdm_value(node)) ;
}

/*
 * mape_rdm_clear_all_nodes
 *
 * clear all the value in the nodes to prepare for the next simulation. 
 */
void 
mape_rdm_clear_all_nodes(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node)
		mape_push_rdm(node, MAPE_U_U) ;
}


/*
 * mape_generate_input_1
 *
 * generate the circuit's input with only FALL
 */
void 
mape_generate_input_1(network)
network_t *network ;
{
	lsGen gen ;
	node_t *p ;

	foreach_primary_input(network, gen, p)
		mape_push_rdm(p, MAPE_RISE) ;
}

/*
 * mape_rdm_generate_input
 *
 * generate the circuit's input according to the parameter. 
 */
void 
mape_rdm_generate_input(network, trans_perc)
network_t *network ;
int trans_perc ;
{
	int first, second ;
	long r ;
	lsGen gen ;
	node_t *p ;
	mape_node_value_t value ;

	srandom(util_cpu_time()) ;

	foreach_primary_input(network, gen, p){
		r = random() ;
		if(r%2 == 0)
			first = ZERO ;
		else 
			first = ONE ;

		if(r%100 >= trans_perc)
			second = first ;
		else 
			second = mape_set_value_not(first) ;

		value = mape_value_convert(first, second) ;
		mape_push_rdm(p, value) ;
	}
}

/*
 * mape_rdm
 *
 * get the maximum power estimation by random input. 
 */
void 
mape_rdm(network, times, trans_perc)
network_t *network ;	/*The network */
int times ;		/*How many random instance should I simulate? */
int trans_perc ;	/*percentage of the transistion in the input */
{
	int i ;
	lsList input_values ;
	long tf = 0 ; /* transition * fanout */
	long tfM = 0 ;

	for(i=0;i<times;i++){
		mape_rdm_clear_all_nodes(network) ;
		if(i == 0)
			mape_generate_input_1(network);
		else 
			mape_rdm_generate_input(network, trans_perc) ;
		mape_rdm_simulate(network);
		tf = mape_rdm_transition_fanout(network) ;
		if(tf > tfM || i == 0){
			mape_rdm_copy_to_max(network) ;
			tfM = tf ;
		}
	}
}
