/*
 * mape_sagrp.c
 *
 * Optimize group using simulated annealing method. 
 * 
 * Changbo Long (Mar. 5, 2002) 
 */
#include "mape_int.h"

#define MAXGROUP_NUM  200

void 	mape_sagrp_move(char *, char *) ;
double 	mape_sagrp_cost(char *) ;
void 	mape_sagrp_destroy(char *) ;
char*	mape_sagrp_create(void) ;
int 	mape_sagrp_equilibrum(int, int, int) ;
int 	mape_sagrp_stop(double, int, double) ;
void 	mape_sagrp_accept(int) ;

double 	last_cost[MAXGROUP_NUM] ;
double 	last_cost_backup[MAXGROUP_NUM] ;
int 	last_switch[2] = {-1, -1}  ;
int 	group_length = 0 ;

/*
 * mape_sagrp_create
 *
 * Allocate memory for the solution handle. 
 */
char *
mape_sagrp_create(void)
{
	char *rt ;

	rt = (char *)lsCreate() ;
	return rt;
}

/*
 * mape_sagrp_destroy
 *
 * destroy the memory space for the solution handle. 
 */
void 
mape_sagrp_destroy(sol)
char *sol ;
{
	lsList list ;

	list = (lsList)sol ;
	lsDestroy(list, free) ;
}

/*
 * mape_sagrp_cost. 
 *
 * get the cost function for the one solution handle. 
 */
double 
mape_sagrp_cost(sol)
char *sol ;
{
	lsList groups ;
	lsList group ;
	lsGen gen, gen1 ;
	node_t *node ;
	double total ;
	double gc ;
	int count = 0 ;
	int i=0 ;
	double time ;

	total = 0.0 ;
	groups = (lsList)sol ;

	/* backup the cost for each group */
	if(last_switch[0] != -1){
		for(i=0;i<lsLength(groups);i++){
			last_cost_backup[i] = last_cost[i] ;
			//fprintf(mapeout, "last_cost[%d] = %.2lf\n",i, last_cost[i]) ;
		}
	}

	lsForeachItem(groups, gen, group){
		if(last_switch[0] == -1){
			mape_clean_given_gate() ;
			lsForeachItem(group, gen1, node) 
				mape_push_given_gatep(node) ;
			mape_curitr_clear(global_network) ;
			gc = mape_curitr(global_network, &time, NULL) ;
			last_cost[count] = gc ;
			total += gc ;
		}else if(count != last_switch[0] && count != last_switch[1]){
			total += last_cost[count] ; 
		}else if(count == last_switch[0] || count == last_switch[1]){
			mape_clean_given_gate() ;
			lsForeachItem(group, gen1, node) 
				mape_push_given_gatep(node) ;
			mape_curitr_clear(global_network) ;
			gc = mape_curitr(global_network, &time, NULL) ;
			last_cost[count] = gc ;
			total += gc ;
		}
		count ++ ;
	}
	fprintf(mapeout, "Cost function return %.2lf \n", -total) ;
	return -total ;
}

/*
 * mape_sagrp_move
 *
 * make modification to the solution. 
 */
void 
mape_sagrp_move(from, to) 
char *from, *to ;
{
	lsList lf, lt ; /*list from and list to*/
	lsGen gen, gen1;
	lsList group, n_group ;
	lsList tmp ;
	long r1, r2 ;
	int g1, g2, n1, n2, length, length1, length2, count, count1, count2 ;
	node_t *node1, *node2, *node ;
	lsList group1, group2 ; 

	lf = (lsList)from ;
	lt = (lsList)to ;
	node1 = NULL ;
	node2 = NULL ;

	/* get two random and different group */
	length = lsLength(lf) ; 
	do{
		r1 = random() ;
		r2 = random() ;
		g1 = (int)(r1%length) ;
		g2 = (int)(r2%length) ;
	}while(g1 == g2) ;

	/* clear the to's data */ 
	while(lsLength(lt)){
		lsDelEnd(lt, (lsGeneric *)&group) ;
		free(group) ;
	}

	/* create the to solution while make node exchange in g1 and g2 */
	count = 0 ;
	lsForeachItem(lf, gen, group){
		n_group = lsCreate() ;
		if(count == g1){ 
			r1 = random() ;
			group1 = n_group ;
			length = lsLength(group) ;
			n1 = (int)(r1%length) ;
			count1 = 0 ;
			lsForeachItem(group, gen1, node){
				if(count1 != n1)
					lsNewEnd(n_group, (lsGeneric)node, LS_NH) ;
				else 
					node1 = node ;
				count1 ++ ;
			}
		}else if(count == g2){
			r2 = random() ;
			group2 = n_group ;
			length = lsLength(group) ;
			n2 = (int)(r2%length) ;
			count2 = 0 ;
			lsForeachItem(group, gen1, node){
				if(count2 != n2)
					lsNewEnd(n_group, (lsGeneric)node, LS_NH) ;
				else 
					node2 = node ;
				count2 ++ ;
			}
		}else{
			lsForeachItem(group, gen1, node)
				lsNewEnd(n_group, (lsGeneric)node, LS_NH) ;
		}
		lsNewEnd(lt, (lsGeneric)n_group, LS_NH) ;
		count ++ ;
	}

	/* exchange the two node */
	if(node1 == NULL || node2 == NULL || group1 == NULL || group2 == NULL){
		mape_error("stupid programming! \n", __FILE__, __LINE__) ;
	}else {
		lsNewEnd(group1, (lsGeneric)node2, LS_NH) ;
		lsNewEnd(group2, (lsGeneric)node1, LS_NH) ;
		last_switch[0] = g1 ;
		last_switch[1] = g2 ;
	}
}

/*
 * mape_sagrp_accept
 *
 * accept the move or not. 
 */
void 
mape_sagrp_accept(ac)
int ac ;
{
	int i=0 ;

	if(!ac){
		for(i=0;i<group_length;i++){
			last_cost[i] = last_cost_backup[i] ;
			//fprintf(mapeout, "last_cost_backup[%d] = %.2lf\n", i, last_cost[i]) ;
		}
	}
}


/*
 * mape_sagrp_equilibrum
 *
 * Judge if reach equilibrum
 */
int 
mape_sagrp_equilibrum(up, total, reject_time)
int up, total, reject_time ;
{
	if(up >= 50 || total >= 100)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_sagrp_stop
 *
 * stop the process of the sa. 
 */
int 
mape_sagrp_stop(temp, total, reject_ratio)
double temp, reject_ratio ;
int total ;
{
	if(temp <= 0.001)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_sagrp 
 *
 * Using simulated annealing method to obtain
 * optimized group.  
 */
void 
mape_sagrp(network)
network_t *network ;
{
	lsList groups ;
	char *sol ;
	int i = 0 ;
	double *x ;

	/* get the initial group method */
	groups = mape_group_get_groups(network) ;
	group_length = lsLength(groups) ;

	for(i=0;i<group_length;i++)
		last_cost[i]=last_cost_backup[i]=0.0 ;
		
	/* set all the parameters for the simulated annealing template*/
	mape_sa_set(100.0, 0.9, mape_sagrp_create, mape_sagrp_destroy, 
		mape_sagrp_cost, mape_sagrp_move, mape_sagrp_equilibrum, mape_sagrp_stop, mape_sagrp_accept) ;

	sol = (char *)(groups) ;
	fprintf(mapeout, "Initial cost is: %.2lf \n", mape_sagrp_cost(sol)) ;
	mape_sa_start(&sol) ;
	fprintf(mapeout, "Cost after optimization: %.2lf \n", mape_sagrp_cost(sol)) ;
}

