/*
 * mape_saplt.c
 *
 * Using simulated annealing algorithm to get the 
 * placement of the circuit. 
 * 
 * Changbo Long. 	Mar. 14, 2002. 
 */
#include "mape_int.h"

#define RANDOM_SPACE_ALLOC	1
#define SEQUENCE_SPACE_ALLOC	2

#define CGATE_AREA		1
#define LATCH_AREA		4


extern 	char * 	mape_saplt_sa_create(void) ;
extern 	double 	mape_saplt_cost(char *) ;
extern 	void 	mape_saplt_sa_destory(char *) ;
extern 	void 	mape_saplt_move(char *, char *) ;
extern 	void 	mape_saplt_move_s(char *, char *) ;
extern 	int 	mape_saplt_sa_equilibrum(int, int, int) ;
extern 	int 	mape_saplt_sa_stop(double, int, double) ;
extern 	void 	mape_saplt_sa_accept(int) ;
extern 	void 	mape_saplt_sa_accept_s(int) ;
extern 	double	mape_saplt_wire_change(node_t *, node_t *) ;
extern 	node_t* mape_saplt_index_node(int row, int col) ; 	
double	mape_saplt_node_prob(node_t *) ;
double	mape_saplt_prob_change(node_t *, node_t *) ;
double	mape_saplt_prob_change_group(int, int) ;
double  mape_saplt_group_prob(lsList) ;

/* a rectangle space */
typedef struct mape_saplt_space_struct{
	int row_s ;
	int row_e ;
	int col_s ;
	int col_e ;
} mape_saplt_space_t ;

void mape_saplt_dump(void) ;

lsList space_list = NULL ;
node_t **plt_array = NULL ;
int plt_size ;
node_t *ex_node1, *ex_node2 ;
double wire_length = -1.0 ;
double total_prob = -1.0 ;
int ex_row1, ex_col1, ex_row2, ex_col2 ;
double *group_prob = NULL ;
int group_count = 0 ;

/*
 * create the mape_saplt_space_t 
 */
mape_saplt_space_t *
mape_saplt_create_space_t(row_s, row_e, col_s, col_e)
int row_s, row_e, col_s, col_e ;
{
	mape_saplt_space_t *rtv ;

	rtv		= ALLOC(mape_saplt_space_t, 1) ;
	rtv->row_s	= row_s ;
	rtv->row_e	= row_e ;
	rtv->col_s	= col_s ;
	rtv->col_e	= col_e ;

	return rtv ;
}

/*
 * set the node array. 
 */
void 
mape_saplt_set_node_array(array, size)
node_t **array ;
int size ;
{
	plt_array 	= array ;
	plt_size	= size ;
}

/*
 * create the mape_node_plt_t struct. 
 */
mape_node_plt_t *
mape_create_plt_t(void)
{
	mape_node_plt_t *rtv ;

	rtv 		= ALLOC(mape_node_plt_t, 1) ; 
	rtv->row	= 0 ; 
	rtv->col	= 0 ; 
	rtv->area	= 0 ;
	rtv->wire	= 0 ;
	rtv->st_list	= NULL ;
	rtv->st_value	= 0 ;
	rtv->prob	= 0.0 ;
	rtv->group	= 0 ;

	return rtv ;
}

/*
 * mape_saplt_index_node
 *
 * return the node in certain row and col . 
 */
node_t *
mape_saplt_index_node(row, col)
int row , col ;
{
	return plt_array[(row-1)*plt_size + col - 1] ;
}

/*
 * mape_saplt_switched_node
 *
 * return the two switched nodes. 
 */
void 
mape_saplt_switched_node(n1, n2)
node_t **n1, **n2 ;
{
	*n1 	= ex_node1 ;
	*n2 	= ex_node2 ;
}

/*
 * mape_saplt_plt_size
 *
 * return the size of the placement. 
 */
void
mape_saplt_plt_size(r, c)
int *r ;
int *c ;
{
	*r = plt_size ;
	*c = plt_size ;
}

/*
 * mape_saplt_alloc_array
 *
 * allocate the array. 
 */
void 
mape_saplt_alloc_array(size)
int size ;
{
	
	plt_size = size ;
	plt_array = ALLOC(node_t *, size*size) ; 
	memset(plt_array, 0, size*size*sizeof(node_t *)) ;
}

/*
 * mape_saplt_set_exnode
 *
 * set the exchange node. 
 */
void 
mape_saplt_set_exnode(node1, node2) 
node_t *node1, *node2 ;
{
	ex_node1 = node1 ;
	ex_node2 = node2 ;
}

/*
 * mape_saplt_set_exindex
 *
 * set the exchange index. 
 */
void 
mape_saplt_set_exindex(row1, col1, row2, col2) 
{
	ex_row1 = row1 ;
	ex_col1 = col1 ;
	ex_row2 = row2 ;
	ex_col2 = col2 ;
}


/*
 * mape_saplt_node_area
 *
 * get the node's area according to type. 
 */
int 
mape_saplt_node_area(node)
node_t *node ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;
	
	if( ((mape_t *)(node->mape))->plt->area > 0)
		return ((mape_t *)(node->mape))->plt->area ;

	if( node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT ) 
		return 0 ;

	if(mape_node_is_latch(node)){
		((mape_t *)(node->mape))->plt->area = LATCH_AREA ;
		return LATCH_AREA ;
	}

	((mape_t *)(node->mape))->plt->area = CGATE_AREA ; 
	return CGATE_AREA ;
}

/*
 * mape_saplt_place_node
 *
 * place node in by assign it's row and col's value. 
 */
void 
mape_saplt_place_node(node, row, col)
node_t *node ;
int row, col ;
{
	if(node == NULL){
		plt_array[plt_size*(row-1)+col-1] = node ;
		return  ;
	}
		
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	if( ((mape_t *)(node->mape))->plt == NULL )
		((mape_t *)(node->mape))->plt = mape_create_plt_t() ;
	
	((mape_t *)(node->mape))->plt->row = row ;
	((mape_t *)(node->mape))->plt->col = col ;

	//fprintf(mapeout, "place node %s to %d, %d, plt_array=%p\n", node->name, row, col, plt_array) ;
	plt_array[plt_size*(row-1)+col-1] = node ;
	//fprintf(mapeout, "place node %s to %d, %d, plt_array=%p\n", node->name, row, col, plt_array) ;
	if(mape_node_is_latch(node)){
		plt_array[plt_size*(row-1)+col] = node ;
		plt_array[plt_size*(row)+col-1] = node ;
		plt_array[plt_size*(row)+col] = node ;
	}
}

/*
 * mape_saplt_rect_alloc_space
 *
 * alloc a space from a rectangle. 
 */
void 
mape_saplt_rect_alloc_space(rect, area, type, row, col, split_list)
mape_saplt_space_t *rect ;
int area, type ;  
int *row, *col ;
lsList split_list ;
{
	int size = 0 ;
	mape_saplt_space_t *new_space ;
	int r, c ;
	int rs, re, cs, ce ;
	long v  ;

	if(area == 1)
		size = 1 ;
	else if(area == 4) 
		size = 2 ;

	rs = rect->row_s ;
	re = rect->row_e ;
	cs = rect->col_s ;
	ce = rect->col_e ;
	if(type == RANDOM_SPACE_ALLOC){
		do{
			v = random () ;
			if(re-rs >= 2)
				r = (int)(v%(re-rs)) ;
			else 
				r = 0 ;
			v = random () ;
			if(ce-cs >= 2)
				c = (int)(v%(ce-cs)) ;
			else 
				c = 0 ;
		}while(rs+r+size-1 > re && cs+c+size-1 > ce) ;
	}else{ 
		r = 0 ;
		c = 0 ;
	}

	if(r > 0){
		new_space = mape_saplt_create_space_t(rs, rs+r-1, cs, ce) ;
		lsNewEnd(split_list, (lsGeneric)new_space, LS_NH) ;
		//fprintf(mapeout, "insert (%2d, %2d, %2d, %2d)\n", rs, rs+r-1, cs, ce) ;
	}
	if(c > 0){
		new_space = mape_saplt_create_space_t(rs+r, rs+r+size-1, cs, cs+c-1) ;
		lsNewEnd(split_list, (lsGeneric)new_space, LS_NH) ;
		//fprintf(mapeout, "insert (%2d, %2d, %2d, %2d)\n", rs+r, rs+r+size-1, cs, cs+c-1) ;
	}
	if(cs+c+size <= ce){
		new_space = mape_saplt_create_space_t(rs+r, rs+r+size-1, cs+c+size, ce) ;
		lsNewEnd(split_list, (lsGeneric)new_space, LS_NH) ;
		//fprintf(mapeout, "insert (%2d, %2d, %2d, %2d)\n", rs+r, rs+r+size-1, cs+c+size, ce) ;
	}
	if(rs+r+size <= re){
		new_space = mape_saplt_create_space_t(rs+r+size, re, cs, ce) ;
		lsNewEnd(split_list, (lsGeneric)new_space, LS_NH) ;
		//fprintf(mapeout, "insert (%2d, %2d, %2d, %2d)\n", rs+r+size, re, cs, ce) ;
	}

	/* relative position to absolute position*/
	*row = rs + r ;
	*col = cs + c ;
}


/*
 * mape_saplt_space_joint
 *
 * joint the split list into the space_list. 
 */
void 
mape_saplt_space_joint(split, space)
lsList split, space ;
{
	lsGen gen ;
	lsHandle h ;
	mape_saplt_space_t *p ;

	gen = lsEnd(split) ;
	while(lsPrev(gen, (lsGeneric *)&p, &h) == LS_OK)
		lsNewBegin(space, (lsGeneric)p, LS_NH) ;
	lsFinish(gen) ;
}

/*
 * mape_saplt_alloc_possible
 *
 * is the allocation from the rectangle is possible. 
 */
int 
mape_saplt_alloc_possible(rect, area)
mape_saplt_space_t *rect ;
int area; 
{
	int size ;

	if(area == 1)
		size = 1 ;
	else if(area == 4)
		size = 2 ;

	return (((rect->row_e - rect->row_s) + 1 >= size) && 
		((rect->col_e - rect->col_s) + 1 >= size)) ; 
}

/*
 * mape_saplt_alloc_space
 *
 * alloc space from the placement. 
 */
int 
mape_saplt_alloc_space(area, type, row, col)
int area, type ;
int *row, *col ; 
{
	lsGen gen, gen1;
	mape_saplt_space_t *rect, *tmp ;
	lsList split_list ;
	int count, length, start ;
	long v ;
	lsHandle h ;
	mape_saplt_space_t *p ;
	int i ;

	if(area != CGATE_AREA && area != LATCH_AREA)
		mape_error("Now I can only handle area 1 and 4 !\n", __FILE__, __LINE__) ;

	v = random() ;
	start = (int)(v%lsLength(space_list)) ;
	start ++ ;
	if(type == SEQUENCE_SPACE_ALLOC)
		start = 1 ;
		
	count = 0 ;
	split_list = lsCreate() ;

	while(TRUE){
		lsForeachItem(space_list, gen, rect){
			count ++ ;
			if(!mape_saplt_alloc_possible(rect, area) || start > count)
				continue ;
			mape_saplt_rect_alloc_space(rect, area, type, row, col, split_list) ;
			/* remove the original rect from the space_list */
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
			free(tmp) ;
			/* add all the splited rect into the space_list */
			//mape_saplt_space_joint(split_list, space_list) ;

			gen1 = lsStart(split_list) ;
			tmp = NULL ;
			while(lsNext(gen1, (lsGeneric *)&p, &h) == LS_OK){
				/*
				fprintf(mapeout, "(%2d, %2d, %2d, %2d) \n", 
					p->row_s, p->row_e, p->col_s, p->col_e) ;
				*/
				if(tmp == NULL){
					lsInAfter(gen, (lsGeneric)p, LS_NH) ;
					lsFinish(gen) ; 
					tmp = p ;
				}else{
					lsForeachItem(space_list, gen, rect){
						if(rect == tmp){
							lsInAfter(gen, (lsGeneric)p, LS_NH) ;
							tmp = p ;
							lsFinish(gen) ;
							break ;
						}
					}
				}
			}
			lsFinish(gen1) ;

			lsDestroy(split_list, NULL) ;
			return TRUE ;
		}
		count = 0 ;
		lsForeachItem(space_list, gen, rect){
			count ++ ;
			if(!mape_saplt_alloc_possible(rect, area) || start <= count)
				continue ;
			mape_saplt_rect_alloc_space(rect, area, type, row, col, split_list) ;

			/* remove the original rect from the space_list */
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
			free(tmp) ;
			/* add all the splited rect into the space_list */
			//mape_saplt_space_joint(split_list, space_list) ;

			gen1 = lsStart(split_list) ;
			tmp = NULL ;
			while(lsNext(gen1, (lsGeneric *)&p, &h) == LS_OK){
				fprintf(mapeout, "(%2d, %2d, %2d, %2d)\n ", 
					p->row_s, p->row_e, p->col_s, p->col_e) ;
				if(tmp == NULL){
					lsInAfter(gen, (lsGeneric)p, LS_NH) ;
					lsFinish(gen) ; 
					tmp = p ;
				}else{
					lsForeachItem(space_list, gen, rect){
						if(rect == tmp){
							lsInAfter(gen, (lsGeneric)p, LS_NH) ;
							tmp = p ;
							lsFinish(gen) ;
							break ;
						}
					}
				}
			}
			lsFinish(gen1) ;
			lsDestroy(split_list, NULL) ;
			return TRUE ;
		}
	}
	lsDestroy(split_list, NULL) ;
	return FALSE ;
}

/*
 * initial the space for the circuit. 
 */
void 
mape_saplt_init_space(network)
network_t *network ;
{
	int space = 0 ;
	lsGen gen, gen1 ;
	node_t *node ;
	int size = 0 ;
	mape_saplt_space_t *sp ;

	/*destroy the space list */
	if(space_list)
		lsDestroy(space_list, free) ;

	/*initialize the space according to the node number */
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(!mape_node_is_latch(node))
			space += CGATE_AREA ;
		else 
			space += LATCH_AREA ;
	}
	size = (int)(sqrt((double)space)) + 1 ;
	/*
	fprintf(mapeout, "Get space needed is %d and size is %d\n", space, size) ;
	*/
	sp = mape_saplt_create_space_t(1, size, 1, size) ; 
	space_list = lsCreate() ;
	lsNewEnd(space_list, (lsGeneric)sp, LS_NH) ;
	plt_array = ALLOC(node_t *, size*size) ; 
	memset(plt_array, 0, size*size*sizeof(node_t *)) ;
	plt_size  = size ;
}

/*
 * initial placement for node in the circuit. 
 */
void 
mape_saplt_init_plt(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;
	int row, col ;
	int type ;

	mape_saplt_init_space(network) ;

	type = RANDOM_SPACE_ALLOC ;
	foreach_node(network, gen, node){
		if(!mape_node_is_latch(node))
			continue ;
		mape_saplt_alloc_space(mape_saplt_node_area(node), type, &row, &col) ; 	
		mape_saplt_place_node(node, row, col) ;
		/*
		fprintf(mapeout, "After place latch %s\n", node->short_name) ;  
		mape_saplt_dump() ;
		*/
	}
	type = RANDOM_SPACE_ALLOC ;
	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(mape_node_is_latch(node))
			continue ;
		mape_saplt_alloc_space(mape_saplt_node_area(node), type, &row, &col) ;
		mape_saplt_place_node(node, row, col) ;
		/*
		fprintf(mapeout, "After place cgate %s\n", node->short_name) ;  
		mape_saplt_dump() ;
		*/
	}
}

/*
 * mape_saplt_net_wire_length
 *
 * caculate the wire length for a net. 
 */
int 
mape_saplt_net_wire_length(node)
node_t *node ;
{
	int w, l,r,t,d;
	mape_node_plt_t *p ;
	int size ;
	lsGen gen ;
	node_t *fanout ;

	if(mape_mspplt_null_node(node))
		return 0 ;

	if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
		return 0 ;

	p = ((mape_t *)(node->mape))->plt ;
	if(node_num_fanout(node) == 0 || !p)
		return 0;
	
	if(mape_saplt_node_area(node) == 1)
		size = 1 ;
	else 
		size = 2 ;

	l=p->col ;
	r=p->col + size - 1 ;
	t=p->row ;
	d=p->row + size - 1 ;

	foreach_fanout(node, gen, fanout){
		if(!fanout->mape)
			continue ;
		if(node_type(fanout) == PRIMARY_OUTPUT)
			continue ;

		p = ((mape_t *)(fanout->mape))->plt ;
		if(!p)
			continue ;
		if(mape_saplt_node_area(node) == 1)
			size = 1 ;
		else 
			size = 2 ;
		if(p->col < l)
			l = p->col ;
		if(p->row < t)
			t = p->row ;
		if(p->col + size - 1 > r)
			r = p->col + size - 1 ;
		if(p->row + size - 1 > d)
			d = p->row + size - 1 ;
	}
	return (r-l)+(d-t) ;
}

/*
 * mape_saplt_wire_length
 *
 * get the total wire length of the network 
 * under the placement. 
 */
double 
mape_saplt_wire_length(nouse)
char *nouse ; /* nouse solution space */
{
	network_t *network ;
	int t_w ; /*total_wire_length*/
	int n_w ;
	lsGen gen ;
	node_t *node ;

	network = global_network ;
	t_w = 0 ;

	foreach_node(network, gen, node){
		n_w = mape_saplt_net_wire_length(node) ;
		t_w += n_w ;
		if(node->mape)
			if(((mape_t *)(node->mape))->plt)
				((mape_t *)(node->mape))->plt->wire = n_w ;
	}

	return (double)t_w ;
}

/*
 * mape_saplt_exchange_latch
 *
 * exchange two latches. 
 */
void 
mape_saplt_exchange_latch(node1, node2)
node_t *node1, *node2 ;
{
	int r, c ;

	r = ((mape_t *)(node1->mape))->plt->row ; 
	((mape_t *)(node1->mape))->plt->row = ((mape_t *)(node2->mape))->plt->row ;  
	((mape_t *)(node2->mape))->plt->row = r ;

	c = ((mape_t *)(node1->mape))->plt->col ; 
	((mape_t *)(node1->mape))->plt->col = ((mape_t *)(node2->mape))->plt->col ;  
	((mape_t *)(node2->mape))->plt->col = c ;

	/* exchange in the node array */
	r = ((mape_t *)(node1->mape))->plt->row ;
	c = ((mape_t *)(node1->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = node1 ;
	plt_array[(r-1)*plt_size + c] = node1 ;
	plt_array[r*plt_size + c-1] = node1 ;
	plt_array[(r)*plt_size + c] = node1 ;

	r = ((mape_t *)(node2->mape))->plt->row ;
	c = ((mape_t *)(node2->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = node2 ;
	plt_array[(r-1)*plt_size + c] = node2 ;
	plt_array[r*plt_size + c-1] = node2 ;
	plt_array[r*plt_size + c] = node2 ;
}

/*
 * mape_saplt_exchange_gate_latch
 *
 * exchange four gates with one latch
 */
void 
mape_saplt_exchange_gate_latch(node1, node2)
node_t *node1, *node2 ;
{
	int r, c ;
	node_t *node, *latch, *node_r, *node_d, *node_c ;

	if(mape_node_is_latch(node1)){
		latch = node1 ;
		node = node2 ;
	}else if(mape_node_is_latch(node2)){
		latch = node2 ;
		node = node1 ;
	}else
		mape_error("wrong pointer!\n", __FILE__, __LINE__) ;

	r = ((mape_t *)(node->mape))->plt->row ;
	c = ((mape_t *)(node->mape))->plt->col ;
	node_r = plt_array[(r-1)*plt_size + c] ;
	node_d = plt_array[r*plt_size + c - 1] ;
	node_c = plt_array[r*plt_size + c] ;

	r = ((mape_t *)(latch->mape))->plt->row ; 
	((mape_t *)(latch->mape))->plt->row = ((mape_t *)(node->mape))->plt->row ;  
	((mape_t *)(node->mape))->plt->row = r ;
	if(node_r)
		((mape_t *)(node_r->mape))->plt->row = r ;
	if(node_d)
		((mape_t *)(node_d->mape))->plt->row = r+1 ;
	if(node_c)
		((mape_t *)(node_c->mape))->plt->row = r+1 ;

	c = ((mape_t *)(latch->mape))->plt->col ; 
	((mape_t *)(latch->mape))->plt->col = ((mape_t *)(node->mape))->plt->col ;  
	((mape_t *)(node->mape))->plt->col = c ;
	if(node_r)
		((mape_t *)(node_r->mape))->plt->col = c+1 ;
	if(node_d)
		((mape_t *)(node_d->mape))->plt->col = c ;
	if(node_c)
		((mape_t *)(node_c->mape))->plt->col = c+1 ;

	/* exchange in the node array */
	r = ((mape_t *)(latch->mape))->plt->row ;
	c = ((mape_t *)(latch->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = latch ;
	plt_array[(r-1)*plt_size + c] = latch ;
	plt_array[r*plt_size + c-1] = latch ;
	plt_array[r*plt_size + c] = latch ;

	r = ((mape_t *)(node->mape))->plt->row ;
	c = ((mape_t *)(node->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = node ;
	plt_array[(r-1)*plt_size + c] = node_r ;
	plt_array[r*plt_size + c-1] = node_d ;
	plt_array[r*plt_size + c] = node_c ;
}

/*
 * mape_saplt_exchange_gate
 *
 * exchange two combinational gates. 
 */
void 
mape_saplt_exchange_gate(node1, node2)
node_t *node1, *node2 ;
{
	int r, c ;

	r = ((mape_t *)(node1->mape))->plt->row ; 
	((mape_t *)(node1->mape))->plt->row = ((mape_t *)(node2->mape))->plt->row ;  
	((mape_t *)(node2->mape))->plt->row = r ;

	c = ((mape_t *)(node1->mape))->plt->col ; 
	((mape_t *)(node1->mape))->plt->col = ((mape_t *)(node2->mape))->plt->col ;  
	((mape_t *)(node2->mape))->plt->col = c ;

	/* exchange in the node array */
	r = ((mape_t *)(node1->mape))->plt->row ;
	c = ((mape_t *)(node1->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = node1 ;
	r = ((mape_t *)(node2->mape))->plt->row ;
	c = ((mape_t *)(node2->mape))->plt->col ;
	plt_array[(r-1)*plt_size + c-1] = node2 ;
}

/*
 * mape_saplt_exchange_index
 *
 * exchange the point in the index. 
 */
void 
mape_saplt_exchange_index(row1, col1, row2, col2) 
int row1, col1, row2, col2 ;
{
	node_t *node1, *node2 ;

	node1 = mape_saplt_index_node(row1, col1) ;
	node2 = mape_saplt_index_node(row2, col2) ;

	mape_saplt_place_node(node1, row2, col2) ;
	mape_saplt_place_node(node2, row1, col1) ;
}

/*
 * mape_saplt_is_exchangeable
 *
 * are the two nodes exchangeable. 
 */
int 
mape_saplt_is_exchangeable(node1, node2)
node_t *node1, *node2 ;
{
	int r, c ;

	if(node1 == NULL || node2 == NULL)
		return FALSE ;
	if(node1 == node2)
		return FALSE ;

	if(!mape_node_is_latch(node1) && !mape_node_is_latch(node2))
		return TRUE ;
	if(mape_node_is_latch(node1) && mape_node_is_latch(node2))
		return TRUE ;

	if(!mape_node_is_latch(node1) && mape_node_is_latch(node2)){
		r = ((mape_t *)(node1->mape))->plt->row ;
		c = ((mape_t *)(node1->mape))->plt->col ;
	}else if(mape_node_is_latch(node1) && !mape_node_is_latch(node2)){
		r = ((mape_t *)(node2->mape))->plt->row ;
		c = ((mape_t *)(node2->mape))->plt->col ;
	}

	if(r == plt_size || c == plt_size)
		return FALSE ;

	if(mape_node_is_latch(plt_array[(r-1)*plt_size + c]))
		return FALSE ;
	if(mape_node_is_latch(plt_array[r*plt_size + c-1]))
		return FALSE ;
	if(mape_node_is_latch(plt_array[r*plt_size + c]))
		return FALSE ;

	return TRUE ;
}

/*
 * mape_saplt_move_s
 *
 * move the node's placement in the way of mspplt move
 */
void 
mape_saplt_move_s(from, to)
char *from, *to ; 	/* no use argument */
{
	int row1, col1, row2, col2 ;

	mape_mspplt_random_same_group(&row1, &col1, &row2, &col2, 0) ;
	mape_saplt_exchange_index(row1, col1, row2, col2) ;
	mape_saplt_set_exindex(row1, col1, row2, col2) ;

}


/*
 * mape_saplt_move
 *
 * move node's placement. 
 */
void 
mape_saplt_move(from, to)
char *from, *to ; /*no use argument */
{
	long v ;
	int ind1, ind2 ;
	node_t *node1, *node2 ;
	
	/* change to exchange index */
	do{
		v = random() ;
		ind1 = (int)(v%(plt_size*plt_size)) ;
		v = random() ;
		ind2 = (int)(v%(plt_size*plt_size)) ;
		node1 = plt_array[ind1] ;
		node2 = plt_array[ind2] ;
	}while(!mape_saplt_is_exchangeable(node1, node2)) ;

	ex_node1 = node1 ;
	ex_node2 = node2 ;
	//fprintf(mapeout, "move node %s and %s\n", node1->short_name, node2->short_name) ;

	/* two nodes are all not latch */
	if(!mape_node_is_latch(node1) && !mape_node_is_latch(node2)){
		mape_saplt_exchange_gate(node1, node2) ;
		//mape_saplt_dump() ;
		return ;
	}

	/* two nodes are all latches */
	if(mape_node_is_latch(node1) && mape_node_is_latch(node2)){
		mape_saplt_exchange_latch(node1, node2) ;
		//mape_saplt_dump() ;
		return ;
	}

	/* one is combinational get and another is latch */
	mape_saplt_exchange_gate_latch(node1, node2) ;
	//mape_saplt_dump() ;
}

/*
 * mape_saplt_dump
 *
 * dump the plt of the circuit. 
 */
void 
mape_saplt_dump(void)
{
	int i, j, count = 0 ;
	lsGen gen ;
	mape_saplt_space_t *sp ;

	if(space_list){
		lsForeachItem(space_list, gen, sp){
			count ++ ;
			if(count %5 == 0)
				fprintf(mapeout, "\n") ;
			fprintf(mapeout, "(%2d, %2d, %2d, %2d) ", 
				sp->row_s, sp->row_e, sp->col_s, sp->col_e) ;
		}
		fprintf(mapeout, "\n") ;
	}

	for(i=0;i<plt_size;i++){
		for(j=0;j<plt_size;j++){
			if(!plt_array[i*plt_size+j])
				fprintf(mapeout, "%-5s", "NA") ;
			else 
				fprintf(mapeout, "%5s(%2d)", (plt_array[i*plt_size+j])->short_name, 
					((mape_t*)((plt_array[i*plt_size+j])->mape))->plt->group) ;
		}
		fprintf(mapeout, "\n") ;
	}
}

/*
 * mape_saplt_sa_create
 *
 * just return the network's address. 
 */
char *
mape_saplt_sa_create(void)
{
	return (char *)global_network ;
}

/*
 * mape_saplt_sa_destroy
 *
 * do nothing. 
 */
void 
mape_saplt_sa_destroy(nouse)
char *nouse ;
{
	return ;
}

/*
 * mape_saplt_sa_equilibrum
 *
 * equilibrum stop 
 */
int 
mape_saplt_sa_equilibrum(up, total, accept)
int up, total, accept;
{
	if(total >= plt_size*plt_size*plt_size*50) 
		return TRUE ;

	if(accept >= plt_size*plt_size*plt_size && total >= plt_size*plt_size*plt_size*10)
		return TRUE ;	
	else 
		return FALSE ;
}

/*
 * mape_saplt_stop
 *
 * stop the process of the sa. 
 */
int 
mape_saplt_sa_stop(temp, total, accept_ratio)
double temp, accept_ratio ;
int total ;
{
	static int times = 0 ;

	if(temp <= 0.001)
		return TRUE ;
	else 
		return FALSE;

	if(accept_ratio <= 0.01 && temp <= 0.01){
		if(times >= 2)
			return TRUE ;
		else {
			times ++ ;
			return FALSE ;
		}
	}else{
		times = 0 ;
		return FALSE;
	}
}

/*
 * mape_saplt_sa_accept
 *
 * SA package callback function of accept the move. 
 */
void 
mape_saplt_sa_accept(ac)
int ac ;
{
	double diff ;
	node_t *node1 ;
	node_t *node2 ;

	if(ac)
		return ;

	//mape_saplt_exchange_index(ex_row1, ex_col1, ex_row2, ex_col2) ;

	/* the move are not accepted */
	if(!mape_node_is_latch(ex_node1) && !mape_node_is_latch(ex_node2)){
		mape_saplt_exchange_gate(ex_node1, ex_node2) ;
	}else if(mape_node_is_latch(ex_node1) && mape_node_is_latch(ex_node2)){
		mape_saplt_exchange_latch(ex_node1, ex_node2) ;
	}else 
		mape_saplt_exchange_gate_latch(ex_node1, ex_node2) ;
	/*
	node1 = mape_saplt_index_node(ex_row1, ex_col1) ;
	node2 = mape_saplt_index_node(ex_row2, ex_col2) ;
	*/

	/* change the wire length */
	diff = mape_saplt_wire_change(ex_node1, ex_node2) ;
	wire_length += diff ;
}

/*
 * mape_saplt_sa_accept
 *
 * SA package callback function of accept the move. 
 */
void 
mape_saplt_sa_accept_p(ac)
int ac ;
{
	double diff ;
	int g1, g2 ;

	if(ac)
		return ;

	mape_saplt_exchange_index(ex_row1, ex_col1, ex_row2, ex_col2) ;

	g1 = mape_mspplt_rc_group(ex_row1, ex_col1) ;
	g2 = mape_mspplt_rc_group(ex_row2, ex_col2) ;

	diff = mape_saplt_prob_change_group(g1, g2) ;
	total_prob += diff ;
}

/*
 * mape_saplt_sa_accept
 *
 * SA package callback function of accept the move. 
 */
void 
mape_saplt_sa_accept_s(ac)
int ac ;
{
	double diff ;
	node_t *node1 ;
	node_t *node2 ;

	if(ac)
		return ;

	mape_saplt_exchange_index(ex_row1, ex_col1, ex_row2, ex_col2) ;

	node1 = mape_saplt_index_node(ex_row1, ex_col1) ;
	node2 = mape_saplt_index_node(ex_row2, ex_col2) ;

	/* change the wire length */
	diff = mape_saplt_wire_change(node1, node2) ;
	wire_length += diff ;
}

/*
 * mape_saplt_node_wire_update
 *
 * update the node's wire length and return the increase. 
 */
int
mape_saplt_node_wire_update(node)
node_t *node ;
{
	int rtv, w, i;
	node_t *fanin ;

	if(mape_mspplt_null_node(node))
		return 0 ;
	if(!node)
		return 0 ;
	if(!node->mape)
		return 0 ;
	if(!((mape_t *)(node->mape))->plt)
		return 0 ;
	
	w = mape_saplt_net_wire_length(node) ;
	rtv = w - ((mape_t *)(node->mape))->plt->wire ;
	((mape_t *)(node->mape))->plt->wire = w ;
	foreach_fanin(node, i, fanin){
		if(!fanin->mape)	
			continue ;
		if(!((mape_t *)(fanin->mape))->plt)
			continue ;
		w = mape_saplt_net_wire_length(fanin) ;
		rtv += w - ((mape_t *)(fanin->mape))->plt->wire ;
		((mape_t *)(fanin->mape))->plt->wire = w ; 
	}
	return rtv ;
}


/*
* mape_saplt_wire_change_gate
*
* the change of the two combinational gates. 
*/
double
mape_saplt_wire_change_gate(n1, n2)
node_t *n1, *n2 ;
{
	int rtv = 0 ;
	int w, i ; 
	node_t *fanin ;

	/* process n1 */
	rtv += mape_saplt_node_wire_update(n1) ;
	rtv += mape_saplt_node_wire_update(n2) ;
	return (double)rtv ;
}

/*
 * mape_saplt_wire_change_latch
 *
 * the change of the two latches. 
 */
double
mape_saplt_wire_change_latch(l1, l2)
node_t *l1, *l2 ;
{
	return mape_saplt_wire_change_gate(l1, l2) ;
}


/*
 * mape_saplt_wire_change_gate_latch
 *
 * the change of 4 gates with one latch
 */
double
mape_saplt_wire_change_gate_latch(node1, node2)
node_t *node1, *node2 ;
{
	int r, c ;
	node_t *node, *latch, *node_r, *node_d, *node_c, *fanin ;
	int w, rtv, i ; 

	if(mape_node_is_latch(node1)){
		latch = node1 ;
		node = node2 ;
	}else if(mape_node_is_latch(node2)){
		latch = node2 ;
		node = node1 ;
	}else
		mape_error("wrong pointer!\n", __FILE__, __LINE__) ;

	r = ((mape_t *)(node->mape))->plt->row ;
	c = ((mape_t *)(node->mape))->plt->col ;
	node_r = plt_array[(r-1)*plt_size + c] ;
	node_d = plt_array[r*plt_size + c - 1] ;
	node_c = plt_array[r*plt_size + c] ;

	rtv = 0 ;
	rtv += mape_saplt_node_wire_update(latch) ;
	rtv += mape_saplt_node_wire_update(node) ;
	rtv += mape_saplt_node_wire_update(node_r) ;
	rtv += mape_saplt_node_wire_update(node_c) ;
	rtv += mape_saplt_node_wire_update(node_d) ;

	return (double)rtv ;
}


/*
 * mape_saplt_wire_change
 *
 * the change of the wire length arised by the 
 * exchange of the two nodes. 
 */
double
mape_saplt_wire_change(node1, node2) 
node_t *node1, *node2 ; 
{
	if(!mape_node_is_latch(node1) && !mape_node_is_latch(node2))
		return mape_saplt_wire_change_gate(node1, node2) ;

	/* two nodes are all latches */
	if(mape_node_is_latch(node1) && mape_node_is_latch(node2))
		return mape_saplt_wire_change_latch(node1, node2) ;

	/* one is combinational get and another is latch */
	return mape_saplt_wire_change_gate_latch(node1, node2) ;
}

/*
 * mape_saplt_node_prob_update
 *
 * update node's prob. 
 */
double
mape_saplt_node_prob_update(node)
node_t *node ;
{
	double rtv ;
	double prob ;
	int i, j;
	node_t *p ;
	int step ;
	int row, col, row1, col1 ;

	if(!node)
		return 0.0 ;
	if(!node->mape)
		return 0.0 ;
	if(((mape_t *)(node->mape))->plt == NULL)
		return 0.0 ;

	prob = mape_saplt_node_prob(node) ;
	rtv = prob - ((mape_t *)(node->mape))->plt->prob ;
	((mape_t *)(node->mape))->plt->prob = prob ;

	step = (FIX_GROUP_SIZE/2) + (FIX_GROUP_SIZE%2 == 0?0:1) ;
	for(i=-step;i<=step;i++)
		for(j=-step;j<=step;j++){
			row1 = row + i ;
			col1 = col + i ;
			if(row1<=0 || col1 <= 0 || row1 > plt_size || col1 > plt_size)
				continue ;
			if(row == row1 && col == col1)
				continue ;
			p = mape_saplt_index_node(row1, col1) ;
			if(!p)
				continue ;
			if(!p->mape)
				continue ;
			if(((mape_t*)(p->mape))->plt == NULL)
				continue ;
			prob = mape_saplt_node_prob(p) ;
			rtv += prob - ((mape_t *)(p->mape))->plt->prob ;
			((mape_t *)(p->mape))->plt->prob = prob ;
		}
	return rtv ;
}

/*
 * mape_saplt_prob_change. 
 *
 * change the node's probability. 
 */
double
mape_saplt_prob_change(n1, n2)
node_t *n1, *n2 ;
{
	double rtv = 0.0 ;

	rtv += mape_saplt_node_prob_update(n1) ;
	rtv += mape_saplt_node_prob_update(n2) ;
	return (double)rtv ;
}

/*
 * mape_saplt_prob_change_group. 
 *
 * change the group's probability
 */
double
mape_saplt_prob_change_group(gn1, gn2)
int gn1, gn2 ;
{
	double rtv, g_p ;
	int count = 0 ;
	lsList group ;

	rtv = 0.0 ;
	group = lsCreate() ;
	mape_sassc_group_start(FIX_GROUP) ;
	while(mape_sassc_group_next(group)){
		if(count == gn1 || count == gn2){
			g_p = mape_saplt_group_prob(group) ;
			rtv += (g_p - group_prob[count]) ;
			group_prob[count] = g_p ;
		}
		count ++ ;
	}
	lsDestroy(group, NULL) ;
	return rtv ;
}

/*
 * mape_saplt_prob_add
 *
 * add the prob to a node's prob. 
 */
void 
mape_saplt_prob_add(node, sum)
node_t *node ;
double *sum ;
{
	lsList st_list ;
	lsGen gen ;
	mape_mspplt_st_t *st ;

	if(!node)
		return ;
	if(!node->mape)
		return  ;
	if(((mape_t *)(node->mape))->plt == NULL)
		return  ;
	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(!st_list)
		return  ;

	lsForeachItem(st_list, gen, st)
		sum[st->step] += (double)(st->rp + st->fp) ;
}

/*
 * mape_saplt_node_prob
 *
 * calculate the probability for the node to others. 
 */
double
mape_saplt_node_prob(node)
node_t *node ;
{
	lsList st_list ;
	lsGen gen ;
	int i, j ;
	double sum[128] ;
	mape_mspplt_st_t *st ;
	int step ;
	int row, col , row1, col1;
	double rtv = 0.0 ;
	double rfp ;

	if(!node)
		return 0.0 ;
	if(!node->mape)
		return 0.0 ;
	if(((mape_t *)(node->mape))->plt == NULL)
		return 0.0 ;

	st_list = ((mape_t *)(node->mape))->plt->st_list ;
	if(!st_list)
		return 0.0 ;

	row = ((mape_t *)(node->mape))->plt->row ;
	col = ((mape_t *)(node->mape))->plt->col ;

	memset(sum, 0, sizeof(double)*128) ;
	step = (FIX_GROUP_SIZE/2) + (FIX_GROUP_SIZE%2 == 0?0:1) ;
	for(i=-step;i<=step;i++)
		for(j=-step;j<=step;j++){
			row1 = row + i ;
			col1 = col + i ;
			if(row1<=0 || col1 <= 0 || row1 > plt_size || col1 > plt_size)
				continue ;
			node = mape_saplt_index_node(row1, col1) ;
			mape_saplt_prob_add(node, sum) ;
		}
	lsForeachItem(st_list, gen, st){
		rfp = (double)(st->fp + st->rp) ;
		rtv += rfp*sum[st->step] ;
	}
	return rtv ;
}

/*
 * mape_saplt_group_prob
 *
 * calculate the group's probability. 
 */
double 
mape_saplt_group_prob(group)
lsList group ;
{
	lsGen gen, gen1 ;
	node_t *node ;
	double x[128] ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	int step, i ;

	memset(x, 0, sizeof(double)*128) ;
	lsForeachItem(group, gen, node){
		if(!node)
			continue ;
		if(!node->mape)
			continue ;
		if(!((mape_t*)(node->mape))->plt)
			continue ;
		st_list = ((mape_t*)(node->mape))->plt->st_list ;
		if(!st_list)
			continue ;
		lsForeachItem(st_list, gen1, st)
			x[st->step] += (st->rp + st->fp) ;
	}

	step = 0 ;
	for(i=0;i<128;i++){
		if(x[i] > x[step])
			step = i ;
	}	
	return x[step] ;
}



/*
 * mape_saplt_prob 
 *
 * probability calculation. 
 */
double 
mape_saplt_prob(void)
{
	/*
	network_t *network = global_network ;
	lsGen gen ;
	node_t *node ;
	double t_p, n_p ;

	t_p = 0.0 ;
	foreach_node(network, gen, node){
		n_p = mape_saplt_node_prob(node) ;
		t_p += n_p ;
		if(node->mape)
			if(((mape_t *)(node->mape))->plt)
				((mape_t *)(node->mape))->plt->prob = n_p ;
	}
	return (double)t_p ;
	*/
	lsList group ;
	double t_p, g_p ;
	double x[2048] ;
	int i ;

	group = lsCreate() ;
	t_p = 0.0 ;
	mape_sassc_group_start(FIX_GROUP) ;
	group_count = 0 ;
	while(mape_sassc_group_next(group)){
		g_p = mape_saplt_group_prob(group) ;
		x[group_count++] = g_p ;
		t_p += g_p ;
	}
	group_prob = ALLOC(double, group_count) ;
	for(i=0;i<group_count;i++)
		group_prob[i] = x[i] ;
	lsDestroy(group, NULL) ;
	return t_p ;
}

/*
 * mape_saplt_prob_cost
 *
 * cost for the prob
 */
double 
mape_saplt_prob_cost(nouse)
char *nouse ;
{
	double diff ;
	int g1, g2 ;

	if(total_prob <= 0.0){
		total_prob = mape_saplt_prob() ;
	}else{
		g1 = mape_mspplt_rc_group(ex_row1, ex_col1) ;	
		g2 = mape_mspplt_rc_group(ex_row2, ex_col2) ;
		if(g1 != g2){
			diff = mape_saplt_prob_change_group(g1, g2) ;
			total_prob += diff ;
		}
	}
	return  total_prob ;
}

/*
 * mape_saplt_mix_cost
 *
 *
 * combinational cost for total wire length and 
 * switching probability.
 */
double
mape_saplt_mix_cost(nouse)
char *nouse ;
{
	return mape_saplt_cost(nouse) + mape_saplt_prob_cost(nouse) ;
}


/*
 * mape_saplt_cost
 *
 * get the cost for the network
 */
double
mape_saplt_cost(nouse)
char *nouse ;
{
	int diff ;
	double cal ;
	node_t *node1, *node2 ;

	if(wire_length <= 0.0){
		wire_length = mape_saplt_wire_length(nouse) ;
	}else{
		node1 = mape_saplt_index_node(ex_row1, ex_col1) ;
		node2 = mape_saplt_index_node(ex_row2, ex_col2) ;
		diff = mape_saplt_wire_change(node1, node2) ;
		wire_length += (double)diff ;
	}
	/*
	cal = mape_saplt_wire_length(nouse) ;
	if(cal != wire_length){
		fprintf(mapeout, "cal=%lf, wire_length=%lf, diff=%d\n", cal, wire_length, diff) ;
		mape_error("wire length calculate error !", __FILE__, __LINE__) ;
	}
	*/
	return wire_length ;
}

/*
 * mape_saplt_cal_assc
 *
 * calculate the average ssc for all possible groups. 
 */
double 
mape_saplt_cal_ssc(network, assc, mssc, group_type)
network_t *network ;
double *assc, *mssc ;
int group_type ;
{
	lsList group;
	lsList switch_list ;
	lsGen gen ;
	node_t *node;
	double sum = 0.0 ;
	double ssc = 0.0 ;
	int count = 0;
	double time ;

	mape_curlib_set_unit_mode(TRUE) ;
	group = lsCreate() ;
	switch_list = lsCreate() ;

	*mssc = 0.0;
	mape_sassc_group_start(group_type) ;
	while(mape_sassc_group_next(group)){
		lsForeachItem(group, gen, node)
			fprintf(mapeout, "%-5s", node?node->short_name:"NA") ;
		ssc = mape_sassc_group_max_ssc(group, UNIT_DELAY_MODEL, &time, switch_list) ;
		fprintf(mapeout, "\n\tssc=%.2lf, time=%.2lf\n", ssc, time)  ;
		/*
		fprintf(mapeout, "\n\tswitch:") ;
		lsForeachItem(switch_list, gen, node)
			fprintf(mapeout, "%7s(%d)", node->short_name, ((mape_t *)(node->mape))->plt->st_value) ;
		fprintf(mapeout, "\n") ;
		*/
		sum += ssc ;
		count ++ ;
		*mssc = MAX(*mssc, ssc) ;
	}
	*assc  = (double)(sum/(double)(count)) ;
	lsDestroy(group, NULL) ;
	lsDestroy(switch_list, NULL) ;
}

/*
 * using simulated annealing algorithm to 
 * get a optimum solution for placement in 
 * terms of the wire length. 
 */
void 
mape_saplt(network)
network_t *network ;
{
	double t_w  ;
	char *sol ;
	double assc, mssc ;

	srandom(util_cpu_time()) ;
	mape_saplt_init_plt(network) ;

	//mape_saplt_dump() ;
	mape_sa_set(50, 0.9, mape_saplt_sa_create, mape_saplt_sa_destroy, 
		mape_saplt_cost, mape_saplt_move_s, mape_saplt_sa_equilibrum, 
		mape_saplt_sa_stop, mape_saplt_sa_accept_s) ;

	sol = (char *)network ;
	mape_sa_start(&sol) ; 
	mape_saplt_dump() ;
	mape_saplt_cal_ssc(network, &assc, &mssc, MSP_GROUP) ;
	fprintf(mapeout, "After SATWL TWL=%lf, ASSC(MSG)=%lf, MSSC(MSG)=%lf\n", mape_saplt_wire_length((char *)network), assc, mssc) ; 
}
