/*
 * mape_sosa.c
 *
 *
 * Using SA to cluster gates which otains an optimized  
 * sleep transistor Area. 
 *
 * Changbo Long. (clong@cae.wisc.edu). Nov. 23, 2002. 
 */
#include "mape_int.h"

/* the cluster structure representing clusters in circuit */  

typedef struct mape_sosa_cluster_struct {
	lsList nodes ;
	double cost ;
	double o_cost ;
	node_t *out ;
	node_t *in ;
} mape_sosa_cluster_t ;

extern char * 	mape_sosa_sa_create(void) ;
extern double 	mape_sosa_sa_cost(char *) ;
extern void 	mape_sosa_sa_destroy(char *) ;
extern void 	mape_sosa_sa_move(char *, char *) ;
extern int 	mape_sosa_sa_equilibrum(int, int, int) ;
extern int 	mape_sosa_sa_stop(double, int, double) ;
extern void 	mape_sosa_sa_accept(int) ;

lsList sosa_clusters = NULL;
mape_sosa_cluster_t *change1, *change2 ;
int cluster_num ;

/*
 * mape_sosa_create_cluster_t
 *
 * create the sosa_cluster_t
 */
mape_sosa_cluster_t *
mape_sosa_create_cluster_t(void) 
{
	mape_sosa_cluster_t *rtv ;

	rtv 		= ALLOC(mape_sosa_cluster_t, 1);
	rtv->nodes	= NULL ;			
	rtv->cost	= 0.0;
	rtv->o_cost	= 0.0 ;
	rtv->out 	= NULL ;
	rtv->in		= NULL ;

	return rtv ;
}

/*
 * mape_sosa_sa_create
 *
 * jsut return the groups address. 
 */
char *
mape_sosa_sa_create(void)
{
	return (char *) sosa_clusters ;
}

/*
 * mape_sosa_sa_destroy
 *
 * do nothing. 
 */
void 
mape_sosa_sa_destroy(nouse)
char *nouse ;
{
	return ;
}

/*
 * mape_sosa_sa_cost. 
 *
 * the cost function. 
 */
double 
mape_sosa_sa_cost(nouse)
char *nouse ;
{
	lsGen gen ;
	lsList l ;
	double rtv = 0.0, tmp ;
	mape_sosa_cluster_t *c ;

	lsForeachItem(sosa_clusters, gen, c){
		if(c->in == NULL && c->out == NULL){
			rtv += c->cost ; 
		}else{
			l = lsCopy(c->nodes, NULL) ;
			lsNewEnd(l, (lsGeneric)(c->in), LS_NH) ;

			mape_pbssce_set_use_cant(TRUE) ;
			c->cost = mape_pbssce_itr_estimate_current(global_network, l) ;
			rtv += c->cost ;
			lsDestroy(l, NULL) ;
		}
	}
	//fprintf(mapeout, "sa_cost calculation: %.2lf\n", -rtv) ;
	return -rtv ;
}

/*
 * mape_sosa_optimal_cost. 
 *
 * calculate the optimal cost, which is the maximum current 
 * value for all gates ;
 */
double
mape_sosa_optimal_cost(network)
network_t * network ;
{
	lsList l, vpl, vol_points ;
	lsGen gen ;
	node_t *node ;
	double rtv = 0.0, ncur = 0.0, t ;
	lsGeneric tmp ;


	l = lsCreate() ;
	vpl = lsCreate() ;
	foreach_node(network, gen, node) {
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		lsNewEnd(l, (lsGeneric)node, LS_NH) ; 
	}
	for(;;){
		mape_cur_clear(network) ;
		if(!mape_disksim_read(network))
			break ;
		while(lsLength(vpl))
			lsDelEnd(vpl, &tmp) ;
		lsForeachItem(l, gen, node){
			if((vol_points = mape_cur_top_vol_points(node)) != NULL)
				lsNewEnd(vpl, (lsGeneric)vol_points, LS_NH) ;
		}
		mape_pbssce_lists_max_current(vpl, &ncur, &t) ;
		if(fabs(ncur) > fabs(rtv)) 
			rtv = ncur ;
	}
	lsDestroy(vpl, NULL) ;
	lsDestroy(l, NULL) ;
	return rtv ;
}

/*
 * mape_sosa_disksim_cost
 *
 * calculate the cost according to the simulation
 * stored in the disk. 
 */
double 
mape_sosa_disksim_cost(gs)
lsList gs ;
{
	lsGen gen, gen1 ;
	mape_sosa_cluster_t *c ;
	double x[2048] ;
	int i, j ;
	double rtv, ncur ;
	node_t *node ;
	double t ;
	lsList vpl, vol_points ;
	lsGeneric tmp ;


	vpl = lsCreate() ;
	memset(x, 0, sizeof(double)*2048) ;
	j = 0 ;
	for(;;){
		j ++ ;
		mape_cur_clear(global_network) ;
		if(!mape_disksim_read(global_network))
			break ;
		i = 0 ;
		lsForeachItem(gs, gen, c){
			while(lsLength(vpl))
				lsDelEnd(vpl, (lsGeneric *)&tmp) ;
			lsForeachItem(c->nodes, gen1, node){
				if((vol_points = mape_cur_top_vol_points(node)) != NULL)
					lsNewEnd(vpl, (lsGeneric)vol_points, LS_NH) ;
			}
			mape_pbssce_lists_max_current(vpl, &ncur, &t) ;
			if(fabs(ncur) > fabs(x[i]))
				x[i] = ncur ;
			i ++ ;
		}
		/*
		if(j % 500 == 0){
			fprintf(mapeout, "%dth disksim cost \n", j) ; 
			for(i=0;i<lsLength(gs);i++){
				fprintf(mapeout, "x[%d] = %.2f\n", i, x[i]) ;
			}
		}
		*/
	}
	rtv = 0.0 ;
	i = 0 ;
	lsForeachItem(gs, gen, c){ 
		lsForeachItem(c->nodes, gen1, node)
			fprintf(mapeout, "%s ", node->name) ;
		fprintf(mapeout, "\n\t%.2f\n", x[i]) ;
		rtv += x[i] ;
		i++ ;
	}
	lsDestroy(vpl, NULL) ;
	return rtv ;
}

/*
 * mape_sosa_random_take
 *
 * random take a node out of the list. 
 */
node_t * 
mape_sosa_random_take(l)
lsList l ;
{
	long v ;
	int pos, i ;
	node_t *node ;
	node_t *rtv ;
	lsGen gen ;

	v = random() ;
	pos = (int)(v%lsLength(l)) ;
	i = 0 ;
	lsForeachItem(l, gen, node) {
		if(i == pos){
			lsDelBefore(gen, (lsGeneric *)&rtv) ;
			lsFinish(gen) ;
			return rtv ;
		}
		i++ ;
	}
	return NULL ;
}

/*
 * mape_sosa_sa_move
 *
 * exchange two gates. 
 */
void 
mape_sosa_sa_move(from, to)
char *from, *to ;
{
	long v ;		
	int x1, x2 ;
	lsGen gen, gen1, gen2 ;
	int i, j ;
	mape_sosa_cluster_t *c, *c1 = NULL, *c2 = NULL ;
	node_t *n1 = NULL, *n2 = NULL ;

	v = random() ;
	x1 = (int)(v%cluster_num) ;
	do{
		v = random() ;
		x2 = (int)(v%cluster_num) ;
	}while(x1 == x2) ;

	i = 0 ;
	lsForeachItem(sosa_clusters, gen, c){
		if(c1 && c2){
			lsFinish(gen) ;
			break ;
		}
		if(i == x1) c1 = c ;
		if(i == x2) c2 = c ;
		i ++ ;
	}
	if(c1 == NULL || c2 == NULL)
		mape_error("c1 and c2 should not be zero!", __FILE__, __LINE__) ;

	n1 = mape_sosa_random_take(c1->nodes) ;
	n2 = mape_sosa_random_take(c2->nodes) ;
	c1->in = n2 ;
	c1->out = n1 ;
	c2->in = n1 ;
	c2->out = n2 ;
	change1 = c1 ;
	change2 = c2 ;
}

/*
 * mape_sosa_sa_accept. 
 *
 * callback function for informing the accept of the move. 
 */
void 
mape_sosa_sa_accept(ac)
int ac ;
{
	lsGen gen ;

	if(ac){
		change1->o_cost = change1->cost ; 	
		lsNewEnd(change1->nodes, (lsGeneric)(change1->in), LS_NH) ;
		mape_pbssce_node_destroy_cantl(change1->in) ;
		change2->o_cost = change2->cost ; 	
		lsNewEnd(change2->nodes, (lsGeneric)(change2->in), LS_NH) ;
		mape_pbssce_node_destroy_cantl(change2->in) ;
	}else{
		change1->cost = change1->o_cost ; 	
		lsNewEnd(change1->nodes, (lsGeneric)(change1->out), LS_NH) ;
		change2->cost = change2->o_cost ; 	
		lsNewEnd(change2->nodes, (lsGeneric)(change2->out), LS_NH) ;
	}
	change1->in = change1->out = NULL ;
	change2->in = change2->out = NULL ;
}

/*
 * mape_sosa_sa_equilibrum
 *
 * equilibrum stop
 */
int 
mape_sosa_sa_equilibrum(up, total, reject)
int up, total, reject ;
{
	int times  ;
	times = (int)sqrt(cluster_num) ;

	if(up >= 100*times || total >= 200*times) 
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_sosa_stop
 *
 * stop the process of the sa. 
 */
int 
mape_sosa_sa_stop(temp, total, reject_ratio)
double temp, reject_ratio ; 
int total ;
{
	if(temp <= 0.1)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_sosa_init
 *
 * initialize the sosa package. 
 */
void 
mape_sosa_init(network)
network_t *network ;
{
	lsList groups, g;
	lsGen gen, gen1 ;
	mape_sosa_cluster_t *c ;
	double cost, total ;
	node_t *node ;

	if(sosa_clusters == NULL)		
		sosa_clusters = lsCreate() ;
	groups = mape_group_get_groups(network) ;

	mape_curlib_set_various_mode(TRUE) ;
	mape_pbssce_build_coswitch_db(network) ;

	total = 0.0  ;
	lsForeachItem(groups, gen, g){
		c = mape_sosa_create_cluster_t() ;
		c->nodes = g ;
		c->cost = mape_pbssce_itr_estimate_current(network, g) ;
		fprintf(mapeout, "c->cost = %.2lf\n", c->cost) ;
		c->o_cost = c->cost ;
		total += c->cost ;
		lsNewEnd(sosa_clusters, (lsGeneric)c, LS_NH) ;
	}
	/* test, get the current for the secnd time */
	/*
	fprintf(mapeout, "Total = %.2lf\nSecond time: \n", total) ;
	foreach_node(global_network, gen, node)	
		mape_pbssce_node_destroy_cantl(node) ;
	total = 0.0 ;
	lsForeachItem(sosa_clusters, gen, c){
		cost = mape_pbssce_itr_estimate_current(network, c->nodes) ;
		total += cost ;
		fprintf(mapeout, "cost = %.2f\n", cost) ;
	}
	fprintf(mapeout, "total = %.2f\n", total) ;
	*/
	cluster_num = lsLength(sosa_clusters);
}

/*
 * mape_sosa
 *
 * sleep transistor optimization by sa.  
 */
void
mape_sosa(network)
network_t *network ;
{
	double t_w, cost, t ;
	char *sol ;
	lsGen gen  ;
	node_t *node ;
	mape_sosa_cluster_t *c ;
	lsGen gen1 ;

	mape_sosa_init(network) ;
	srandom(util_cpu_time()) ;
	sol = (char *)sosa_clusters ;
	fprintf(mapeout, "Optimal cost: %.2f\n", mape_sosa_optimal_cost(network)) ;
	fprintf(mapeout, "Cost before SA: %.2f\n", mape_sosa_sa_cost(sol)) ;
	//fprintf(mapeout, "Disksim Cost brefore SA: %.2f\n", mape_sosa_disksim_cost(sosa_clusters)) ;

	mape_sa_set(75, 0.5, mape_sosa_sa_create, mape_sosa_sa_destroy, 
		mape_sosa_sa_cost, mape_sosa_sa_move, mape_sosa_sa_equilibrum, 
		mape_sosa_sa_stop, mape_sosa_sa_accept) ;
	sol = (char *)sosa_clusters ;
	mape_sa_start(&sol) ;

	/* clear all node's cantl list and calculate the cost */
	foreach_node(global_network, gen, node)	
		mape_pbssce_node_destroy_cantl(node) ;
	cost = 0.0 ;
	lsForeachItem(sosa_clusters, gen, c){
		t = mape_pbssce_itr_estimate_current(global_network, c->nodes) ;
		lsForeachItem(c->nodes, gen1, node)
			fprintf(mapeout, "%s ", node->name) ;
		fprintf(mapeout, "\n\t: %.2f\n", t) ; 
		cost += t ;
	}
	fprintf(mapeout, "total Cost after  SA: %.2f\n", cost) ; 

	mape_disksim_read_reset(global_network) ;
	fprintf(mapeout, "Disksim Cost after SA: %.2f\n", mape_sosa_disksim_cost(sosa_clusters)) ;
}
