/*
 * mape_spice.c
 *
 * dump the circuit with spice format for simulation the circuit. 
 *
 * Changbo Long.  Dec 30, 2001. 
 */
#include "mape_int.h"
#include "ga_alg.h"

#define SPICE_HEAD "/home/clong/work/sis/sis_uw/mape/src/spice/head.sp"
#define SPICE_TAIL "/home/clong/work/sis/sis_uw/mape/src/spice/tail.sp"
#define CLOCK_WAVE "pulse (0 1.3 0.5n 0.05n 0.05n 0.45n 1n)"
#define CLOCK_WAVE_BAR "pulse (0 1.3 0n 0.05n 0.05n 0.45n 1n)"
#define INPUT_WAVE "pulse (0 1.3 0n 0.05n 0.05n 0.95n 2n)"
#define INPUT_WAVE_R "pulse (0 1.3 1n 0.05n 0.05n 0.95n 2n)"
#define INPUT_WAVE_FH "pulse (0 1.3 "
#define INPUT_WAVE_SH " 0.05n 0.05n 0.95n 2n)"
#define VVDD "virtual_vdd_node"
#define VGND "virtual_gnd_node"

chromosome_t *ga_sol = NULL ;

extern mape_curlib_gate_t *mape_cur_curlib_gate(node_t *) ;
extern void mape_curga_parse_code_unit(chromosome_t *, int, int *, int *) ;

char *
mape_spice_convert_name(name)
char *name ;
{
	static char rtv[128] ;
	static int count = 0 ;
	char *p, *r ;

	p = name ;
	r = rtv ;

	if(strlen(name) >= 6){
		sprintf(rtv, "MAPE%d", count) ;
		count ++ ;
		return rtv ;
	}

	while(*p != '\0'){
		/*
		if(*p == 'G' || *p == 'A' || *p == 'T')
			*(r++) = '9' ;
		else*/ if(*p != '[' && *p != ']' && *p != '(' && *p != ')') 
			*(r++) = *p ;
		p ++ ;
	}
	*r = '\0' ;
	return rtv ;
}


void 
mape_print_file(filename)
char *filename ;
{
	FILE *fp ;
	int ch ;

	if((fp = fopen(filename, "r")) == NULL)
		mape_error("can't open file\n", __FILE__, __LINE__) ;
	
	while(TRUE){
		ch = fgetc(fp) ;
		if(ch == EOF)
			break ;
		fputc(ch, mapeout) ; 
	}
}

/*
 * mape_spice_dump_node. 
 *
 * dump each gate in the network except PI and PO. 
 */
void 
mape_spice_dump_node(node, verb)
node_t *node ;
int verb ;
{
	int i ;
	node_t *fanin ;
	mape_curlib_gate_t *curlib_gate ;
	char *n_name ;

	if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
		return ;

	fprintf(mapeout, "X%s\t", mape_spice_convert_name(node->short_name)) ;
	fprintf(mapeout, "N%s\t", mape_spice_convert_name(node->short_name)) ;

	if(mape_node_is_latch(node))
		fprintf(mapeout, "N%s_bar\t", mape_spice_convert_name(node->short_name)) ;

	foreach_fanin(node, i, fanin)
		fprintf(mapeout,"N%s\t", mape_spice_convert_name(fanin->short_name)) ;

	if(mape_node_is_latch(node))
		fprintf(mapeout, "clk\tclk_bar\t") ;

	if(verb)
		fprintf(mapeout, "VDD_%s\t%s\t", mape_spice_convert_name(node->short_name), VGND) ;
	else 
		fprintf(mapeout, "%s\t%s\t",  VVDD, VGND) ;
	
	curlib_gate = mape_cur_curlib_gate(node) ;
	fprintf(mapeout, "%s\n", curlib_gate->name) ;
}

/*
 * mape_spice_dump
 *
 * dump the circuit information. 
 */
void 
mape_spice_dump(network, verb)
network_t *network ;
int verb ;
{
	lsGen gen ;
	node_t *node ;

	foreach_node(network, gen, node)
		mape_spice_dump_node(node, verb) ;
}

/*
 * mape_spice_set_ga_solution
 *
 * set the ga's solution for outputing the value
 * for PIs. 
 */
void 
mape_spice_set_ga_solution(chromosome_t *sol)
{
	ga_sol = sol ;
}

/*
 * mape_spice_pi_trans_time
 *
 * return the transition time of the PI. 
 */
char *
mape_spice_pi_trans_time(i, rtv)
int i;
char *rtv ;
{
	int switch_type, pos, count, chr_pos ;
	int pi_pos_unit ;

	if(ga_sol == NULL){
		sprintf(rtv, "0n") ;
		return rtv ;
	}

	count = get_gene_count() ;
	pi_pos_unit = (user_option->curga_option.bit_unit)*HALF_SLOP_TIME/100 ;
	chr_pos = i*(1+user_option->curga_option.pi_bit) ;
	mape_curga_parse_code_unit(ga_sol, chr_pos, &switch_type, &pos) ;
	if(switch_type)
		sprintf(rtv, "%lfn", ((float)(pos*pi_pos_unit))/1000.0 + 1.0) ;
	else 
		sprintf(rtv, "%lfn", ((float)(pos*pi_pos_unit))/1000.0) ;
	return rtv ;
}

/*
 * mape_spice_dump_input
 * 
 * dump the circuit's input wave. 
 */
void 
mape_spice_dump_input(network, verb)
network_t *network ;
int verb ;
{
	lsGen gen ;
	node_t *pi ;
	int i = 0 ;
	node_t *node ;
	char tmp[128] ;

	if(verb){
		foreach_node(network, gen, node){
			if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
				continue ;
			fprintf(mapeout, "VVDD_%s\t", mape_spice_convert_name(node->short_name)) ;
			fprintf(mapeout, "VDD_%s\t0\t3V\n", mape_spice_convert_name(node->short_name)) ;
		}
	}
	/*
	foreach_primary_input(network, gen, pi){
		fprintf(mapeout, "vin%d\tN%s\t0\t%s\n", i, mape_spice_convert_name(pi->short_name), INPUT_WAVE) ;
		i++ ;
	}
	*/
	foreach_primary_input(network, gen, pi){
		fprintf(mapeout, "vin%d\tN%s\t0\t%s%s%s\n", i, mape_spice_convert_name(pi->short_name), INPUT_WAVE_FH, 
			mape_spice_pi_trans_time(i, tmp), INPUT_WAVE_SH) ;
		i++ ;
	}
	/* dump clock */
	fprintf(mapeout, "vclk\tclk\t0\t%s\n", CLOCK_WAVE) ; 
	fprintf(mapeout, "vclk_bar\tclk_bar\t0\t%s\n", CLOCK_WAVE_BAR) ; 
}

/*
 * mape_spice_curitr_input
 * 
 * dump the circuit's input wave after maximum estimation. 
 */
void 
mape_spice_dump_curitr_input(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;
	int i = 0 ;
	node_t *node ;

	foreach_primary_input(network, gen, pi){
		if(mape_curitr_pi_value(pi) == MAPE_FALL)
			fprintf(mapeout, "vin%d\t%s\t0\t%s\n", i, mape_spice_convert_name(pi->short_name), INPUT_WAVE) ;
		else
			fprintf(mapeout, "vin%d\t%s\t0\t%s\n", i, mape_spice_convert_name(pi->short_name), INPUT_WAVE_R) ;
		i++ ;
	}
}

/*
 * output a spice file for simulation. 
 * internal use. 
 */
void 
mape_spice(network)
network_t *network ;
{
	mape_print_file(SPICE_HEAD) ;
	/*
	mape_cur_test_filter(network) ;
	*/
	mape_spice_dump(network, user_option->verb) ;
	mape_spice_dump_input(network, user_option->verb) ;
	mape_print_file(SPICE_TAIL) ;
}

/*
 * mape_spice_dump_node. 
 *
 * dump each gate in the network except PI and PO. 
 */
void 
mape_spigrp_dump_node(node, index)
node_t *node ;
int index ;
{
	int i ;
	node_t *fanin ;
	mape_curlib_gate_t *curlib_gate ;

	if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
		return ;
	
	fprintf(mapeout, "X%s\t", mape_spice_convert_name(node->short_name)) ;
	fprintf(mapeout, "N%s\t", mape_spice_convert_name(node->short_name)) ;

	curlib_gate = mape_cur_curlib_gate(node) ;
	if(mape_node_is_latch(node))
		fprintf(mapeout, "N%s_bar\t", mape_spice_convert_name(node->short_name)) ;

	foreach_fanin(node, i, fanin)
		fprintf(mapeout,"N%s\t", mape_spice_convert_name(fanin->short_name)) ;

	if(mape_node_is_latch(node)){
		fprintf(mapeout, "clk\tclk_bar\t") ;
	}

	fprintf(mapeout, "part_vdd_node_%d\t%s\t", index, VGND) ;
	fprintf(mapeout, "%s\n", curlib_gate->name) ;
}

/*
 * mape_spigrp_dump_group
 *
 * dump a group of node. 
 */
void 
mape_spigrp_dump_group(group, index)
lsList group ;
int index ;
{
	lsGen gen ;
	node_t *node ;

	lsForeachItem(group, gen, node)
		mape_spigrp_dump_node(node, index) ;
}

/*
 * mape_spigrp_dump_part_vdd
 *
 * dump the part vdd node. 
 */
void 
mape_spigrp_dump_part_vdd(index)
int index ;
{
	int i ;

	for(i=1;i<=index;i++)
		fprintf(mapeout, "Vpart_vdd_source_%d  part_vdd_node_%d 0 3V\n", i, i) ;
}

/*
 * mape_spigrp
 *
 *
 * output a spice file for simulation, 
 * the network has been partitioned. 
 */
void 
mape_spigrp(network, groups)
network_t *network ;
lsList groups ;
{
	lsList group ;
	lsHandle h ;
	node_t *node ;
	lsGen gen, gen1 ;
	int i = 0 ;

	mape_print_file(SPICE_HEAD) ;
	lsForeachItem(groups, gen, group){
		i++ ;
		mape_spigrp_dump_group(group, i) ;
	}
	mape_spigrp_dump_part_vdd(i) ;
	mape_spice_dump_input(network, FALSE) ;
	mape_print_file(SPICE_TAIL) ;
}
