/*
 * mape_upb.c
 *
 * Calculation of the up boundary of the switching activity. 
 * Just a try. :)
 *
 * Changbo Long. Nov. 19, 2001
 */
#include "mape_int.h"

#define GROUP_SIZE 8
#define DONT_CARE_DISTANCE 12
#define NO_LABEL -2

#define FIRST_SIDE  1
#define SECOND_SIDE 2

lsList tie_list ; /*store all the node whose transistion is tied together*/
int upb_gain = 0 ;

typedef struct mape_non_trans_struct{
	lsList global ;		/*global non-transistion list */
	int    global_gain ;	/*the global non-transistion list's gain*/
	lsList min ;	/*minimum non-transistion list in current exhaustive estimation*/			
	int    min_gain ;	/*minimum non-transistion node's gain */  	
	lsList cur ;		/*the non-transistion list in this time estimation*/
	int    cur_gain ; 	/*current non-transistion list's gain */
}mape_non_trans_t ;

extern void mape_upb_dump_tie(lsList) ;
static void mape_upb_dump(void) ;
static mape_node_value_t mape_upb_get_node_value(node_t *) ;
static void mape_upb_set_node_value(node_t *, mape_node_value_t) ;


/*
 * mape_upb_create_non_trans_t
 *
 *
 * create the mape_non_trans_t 
 */
mape_non_trans_t *
mape_upb_create_non_trans_t(void)
{
	mape_non_trans_t *rtv ;

	rtv 			= ALLOC(mape_non_trans_t, 1) ;
	rtv->global 		= lsCreate() ;
	rtv->global_gain	= 0 ;
	rtv->min		= lsCreate() ;
	rtv->cur		= lsCreate() ;
	rtv->min_gain		= 9999 ;
	rtv->cur_gain		= 0 ;

	return rtv ;
}

/*
 * mape_node_in_list
 *
 * return true if node in the list. 
 */
int
mape_node_in_list(node, list)
node_t *node ;
lsList list ;
{
	lsGen gen ;
	node_t *p ;
	lsHandle h ;
	int find = FALSE ;

	if(!list || !node)
		return FALSE ;

	gen = lsStart(list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		if(node == p){
			find = TRUE ;
			break ;
		}
	}
	lsFinish(gen) ;
	return find ;
}

/*
 * mape_upb_find_tie
 *
 * is a node in a tie list? 
 */
lsList 
mape_upb_find_tie(node, hr)
node_t *node ;
lsHandle *hr ;
{
	lsList tie ;
	lsGen gen ;
	lsHandle h ;
	int find = FALSE ;

	gen = lsStart(tie_list) ;
	while(lsNext(gen, (lsGeneric *)&tie, &h) == LS_OK){
		if(mape_node_in_list(node, tie)){
			find = TRUE ;
			break ;
		}
	}
	lsFinish(gen) ;
	if(find){
		*hr = h ;
		return tie ;
	}else 
		return NULL ;
}

/*
 * mape_upb_merge
 *
 * mage two ties together and delete the later from the tie_list. 
 */
void 
mape_upb_merge(tie1, tie2, h2)
lsList tie1, tie2 ;
lsHandle h2 ;
{
	lsGen gen ;
	node_t *node ;
	lsList tmp ;

	while(lsLength(tie2)){
		lsDelEnd(tie2, (lsGeneric *)&node) ;
		LS_ASSERT(lsNewEnd(tie1, (lsGeneric)node, LS_NH)) ;
	}
	LS_ASSERT(lsRemoveItem(h2, (lsGeneric *)&tmp)) ; 
}

/*
 * mape_upb_push_tie 
 *
 * push a tie into the tie_list. 
 */
void 
mape_upb_push_tie(nodes)
lsList nodes ;
{
	lsGen gen ;
	node_t *p ;
	lsList tie = NULL ;
	lsList p_tie = NULL ;
	lsHandle h, ht, p_ht ;

	if(!tie_list)
		tie_list = lsCreate() ;

	/*
	fprintf(mapeout, "\n\n\nBefore push tie:\n") ;
	mape_upb_dump() ;
	fprintf(mapeout, "Push tie: \n") ;
	mape_upb_dump_tie(nodes) ;
	*/

	gen = lsStart(nodes) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		if((tie = mape_upb_find_tie(p, &ht)) == NULL)
			continue ;
		if(p_tie != NULL && tie != p_tie)
			mape_upb_merge(tie, p_tie, p_ht); 
		p_tie = tie ;
		p_ht = ht ;
	}
	lsFinish(gen) ;

	tie = p_tie ;

	if(tie){
		while(lsLength(nodes)){
			lsDelEnd(nodes, (lsGeneric *)&p) ;
			if(!mape_node_in_list(p, tie)){
				LS_ASSERT(lsNewEnd(tie, (lsGeneric)p, LS_NH)) ;
			}
		}
		lsDestroy(nodes, NULL) ;
	}else{
		lsNewEnd(tie_list, (lsGeneric)nodes, LS_NH) ;
	}
	/*
	fprintf(mapeout, "After push tie:\n") ;
	mape_upb_dump() ;
	*/
}

/*
 * mape_upb_get_gain
 *
 * get the gain of the node. 
 */
void
mape_upb_get_gain(node, pos)
node_t *node ;
int pos ;
{
	node_t *fanin ;

	if(pos == -1)
		upb_gain += node_num_fanout(node) ;
	else 
		upb_gain += MIN(node_num_fanout(node), node_num_fanout(node_get_fanin(node, pos))) ;
	fprintf(mapeout, "node %s has the control value, upb_gain=%d \n", node->name, upb_gain) ;
}

/*
 * mape_upb_stuck_node
 *
 * find the stuck node in the circuit. 
 */
void 
mape_upb_stuck_node(node)
node_t *node ;
{
	fprintf(mapeout, "Stuck node %s detected %s\n", node->name) ;
}

/*
 * mape_upb_make_tie_list
 *
 *
 * for one specified node, find all it's fanin's 
 * tie. 
 */
void 
mape_upb_make_tie_list(node)
node_t *node ;
{
	lsGen gen ;
	node_function_t func ;
	node_t *fanin ;
	int pos , cv;
	lsList nodes = NULL ;

	if(!(node->F))
		return ;
	func = node_function(node) ;
	if(func == NODE_PI)
		return ;
	if(func == NODE_INV)
		return ;
	if(func == NODE_1 || func == NODE_0){
		mape_upb_stuck_node(node) ;
		return ;
	}

	for(pos=0;pos<node->nin;pos++){
		switch(mape_node_ctrl_input(node->F, pos, &cv)){
			case NO_CTRL_VALUE:
				break ;
			case NORMAL_CTRL_VALUE:
			case REVERSE_CTRL_VALUE:
				if(nodes == NULL)
					nodes = lsCreate() ;
				fanin = node_get_fanin(node, pos) ;		
				if(!mape_node_in_list(fanin, nodes))
					lsNewEnd(nodes, (lsGeneric)fanin, LS_NH) ;
				if(!mape_node_in_list(node, nodes))
					lsNewEnd(nodes, (lsGeneric)node, LS_NH) ;
				break ;
		}
	}
	if(nodes)
		mape_upb_push_tie(nodes) ;
}

/*
 * mape_upb_dump_tie
 *
 * dump the tie. 
 */
void 
mape_upb_dump_tie(tie)
lsList tie ;
{
	lsGen gen ;
	node_t *node ;
	int i , total;

	if(!tie)
		return ;
	i = 0 ;
	total = lsLength(tie) ;

	fprintf(mapeout, "\t") ;
	lsForeachItem(tie, gen, node){
		i++ ;
		fprintf(mapeout, "%s(%d)", node->name, node_num_fanout(node)) ;
		if(i < total)
			fprintf(mapeout, ",") ;
		if(i%5 == 0)	
			fprintf(mapeout, "\n\t") ;
	}
}


/*
 * mape_upb_dump
 *
 * dump the information of the nodes in the tie
 * list. 
 */
void 
mape_upb_dump(void)
{
	lsGen gen ;
	lsList tie ;

	fprintf(mapeout, "Now dumping the tie list(%d)\n", lsLength(tie_list)) ;
	lsForeachItem(tie_list, gen, tie){
		fprintf(mapeout, "\n\n") ;
		mape_upb_dump_tie(tie) ;
	}
	fprintf(mapeout, "\n") ;
}

/*
 * mape_upb_make_ppi
 *
 *
 * make the pseduo primary input from a tie. 
 */
void 
mape_upb_make_ppi(tie, ppi)
lsList tie, ppi ;
{
	lsGen gen ;
	node_t *node ;
	node_t *fanin ;
	int i ;
	int find ;

	lsForeachItem(tie, gen, node){
		i = 0 ;
		find = FALSE ;
		foreach_fanin(node, i, fanin)
			if(mape_node_in_list(fanin, tie)){
				find = TRUE ;
			}
		if(!find)
			lsNewEnd(ppi, (lsGeneric)node, LS_NH) ;
	}
}

/*
 * mape_upb_clean_do
 *
 *
 * clean the value of the previous simulation. 
 */
int
mape_upb_clean_do(node, arg)
node_t *node ;
char *arg ;
{
	if(mape_upb_get_node_value(node) == MAPE_U_U){
		/*fprintf(mapeout, "clean node: %s(%s)\n", node->name, mape_print_node_value(mape_upb_get_node_value(node))) ;*/
		return FALSE;
	} else {
		/*fprintf(mapeout, "clean node: %s(%s)\n", node->name, mape_print_node_value(mape_upb_get_node_value(node))) ;*/
		mape_upb_set_node_value(node, MAPE_U_U)  ;
		return TRUE ;
	}
}

/*
 * mape_upb_clean_simulate
 *
 * exhaustively simulate the ppi to get the 
 * minimum transistion loss. 
 */
int 
mape_upb_clean_simulate(ppi)
lsList ppi ;
{
	lsGen gen ;
	node_t * node ;

	mape_imp_reset(mape_upb_clean_do, (char *)NULL) ;
	lsForeachItem(ppi, gen, node)
		mape_imp_push(node);
	mape_imp_start() ;
}

/*
 * mape_upb_get_node_value
 *
 * get node's value
 */
mape_node_value_t
mape_upb_get_node_value(node)
node_t *node ;
{
	return	mape_rdm_value(node) ;
}

/*
 * mape_upb_set_node_value
 *
 * set node's value
 */
void 
mape_upb_set_node_value(node, value)
node_t *node ;
mape_node_value_t value ;
{
	mape_push_rdm(node, value);
}

/*
 * mape_upb_set_ppi
 *
 * set the ppi's value according to the index.
 */
void 
mape_upb_set_ppi(list, index)
lsList list ;
int index ;
{
	int r, s;
	lsGen gen ;
	node_t *node ;
	
	r = index ;
	lsForeachItem(list, gen, node){
		s = r%4;
		r = r/4;
		switch(s){
			case 0:
				mape_upb_set_node_value(node, MAPE_FALL) ;
				break ;
			case 1:
				mape_upb_set_node_value(node, MAPE_RISE) ;
				break ;
			case 2:
				mape_upb_set_node_value(node, MAPE_1_1) ;
				break ;
			case 3:
				mape_upb_set_node_value(node, MAPE_0_0) ;
				break ;
		}
	}
}

/*
 * mape_upb_simulate_node. 
 *
 *
 * simulate the node. 
 */
mape_node_value_t
mape_upb_simulate_node(node)
node_t *node ;
{
	int value ;
	pcube input ;
	int i ;
	int oF, oS ; //first and second output. 
	mape_node_value_t o ;
	node_t *p ;

	o = mape_upb_get_node_value(node) ;

	if(mape_value_determined(o))
		return o ;

	cube.num_vars = node->nin ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_rdm_value_first(p) ;
		PUTINPUT(input, i, value) ;
	}
	oF = mape_set_simulate_node(input, node) ; 
	foreach_fanin(node, i, p){
		value = mape_rdm_value_second(p) ;
		PUTINPUT(input, i, value) ;
	}
	oS = mape_set_simulate_node(input, node) ; 
	o = mape_value_convert(oF, oS) ;

	free_cube(input) ;
	return o ;
}

/*
 * mape_upb_exhaustive_do
 *
 * exhaustively simulate the ppi's value
 */
int 
mape_upb_exhaustive_do(node, arg)
node_t *node ;
char *arg ;
{
	mape_non_trans_t *non_trans  ;
	mape_node_value_t value ;

	non_trans = (mape_non_trans_t *)arg ;
	if(non_trans->cur_gain >= non_trans->min_gain)
		return FALSE ;

	value = mape_upb_simulate_node(node) ;
	/*
	fprintf(mapeout, "get node's value %s<=%s, cur_gain=%d\n", node->name, mape_print_node_value(value), non_trans->cur_gain) ;
	*/

	if(value == MAPE_U_U)
		return FALSE ;
	
	if(value == MAPE_0_0 || value == MAPE_1_1){
		if(!mape_node_in_list(node, non_trans->global)){
			lsNewEnd(non_trans->cur, (lsGeneric)node, LS_NH);
			non_trans->cur_gain += node_num_fanout(node) ;
		}
	}
	mape_upb_set_node_value(node, value);
	return TRUE ;
}

/*
 * mape_upb_simulate_ppi
 *
 * simulate the ppi's value. 
 */
void 
mape_upb_simulate_ppi(list, non_trans)
lsList list ;
mape_non_trans_t *non_trans; 
{
	lsGen gen  ;
	node_t *node ;

	int i=0 ;
	int length = lsLength(list) ;
	fprintf(mapeout, "simulate ppi: ") ;
	lsForeachItem(list, gen, node){
		i++ ;
		fprintf(mapeout, "%s(%s)", node->name, mape_print_node_value(mape_upb_get_node_value(node))) ;
		if(i<length) 
			fprintf(mapeout, ",\t") ;
	}
	fprintf(mapeout, "\n") ;

	mape_imp_reset(mape_upb_exhaustive_do, (char *)non_trans) ;
	lsForeachItem(list, gen, node)
		mape_imp_push(node) ;
	mape_imp_start() ;
}

/*
 * mape_upb_check_min
 *
 *
 * check if the current result is less than 
 * the previous minimum result. If, substitue 
 * the previous minimum result. 
 */
void 
mape_upb_check_min(non_trans)
mape_non_trans_t *non_trans ;
{
	node_t *node ;

	/*
	fprintf(mapeout, "Checking the min gain: cur_gain=%d, min_gain=%d\n", 
		non_trans->cur_gain, non_trans->min_gain) ;
	*/

	if(non_trans->cur_gain >= non_trans->min_gain){
		while(lsLength(non_trans->cur))
			LS_ASSERT(lsDelEnd(non_trans->cur, (lsGeneric *)&node)) ;
		non_trans->cur_gain = 0 ;
		return ;
	}
	
	non_trans->min_gain = non_trans->cur_gain ;
	non_trans->cur_gain = 0 ;

	while(lsLength(non_trans->min))
		LS_ASSERT(lsDelEnd(non_trans->min, (lsGeneric *)&node)) ;
	while(lsLength(non_trans->cur)){
		LS_ASSERT(lsDelEnd(non_trans->cur, (lsGeneric *)&node)) ;
		lsNewEnd(non_trans->min, (lsGeneric)node, LS_NH) ;
	}
}

/*
 * mape_upb_exhaustive_simulate
 *
 *
 * carry out the exhaustive simulate for a group of 
 * ppi list. 
 */
void 
mape_upb_exhaustive_simulate(list, non_trans)
lsList list ;
mape_non_trans_t *non_trans ; /*non_transistion node's list*/
{
	int i ;
	int total = 1 ;
	lsGen gen ;
	node_t *node ;

	for(i=0;i<lsLength(list);i++)
		total *= 4 ;
	for(i=0;i<total;i++){
		mape_upb_set_ppi(list, i) ;
		mape_upb_simulate_ppi(list, non_trans) ;
		mape_upb_clean_simulate(list) ;
		mape_upb_check_min(non_trans) ;
		if(non_trans->min_gain == 0)
			break ;
	}

	/*add the node in the min  to the global*/
	non_trans->min_gain = 9999 ;
	while(lsLength(non_trans->min)){
		LS_ASSERT(lsDelEnd(non_trans->min, (lsGeneric *)&node)) ;
		if(!mape_node_in_list(node, non_trans->global)){
			lsNewEnd(non_trans->global, (lsGeneric)node, LS_NH)  ;
			non_trans->global_gain += node_num_fanout(node) ;
		}
	}
}

/*
 * mape_upb_node_label
 *
 * return the node's label 
 */
int
mape_upb_node_label(node)
node_t *node ;
{
	if(node->mape == NULL)
		return NO_LABEL ;
	
	return ((mape_t *)(node->mape))->visit  ;
}

/*
 * mape_upb_set_ppi_index
 *
 * label the node. 
 */
void
mape_upb_set_ppi_index(node, index)
node_t *node ;
int index ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;
	
	((mape_t *)(node->mape))->C0 = index ;
}

/*
 * mape_upb_get_ppi_index
 *
 * label the node. 
 */
int
mape_upb_get_ppi_index(node)
node_t *node ;
{
	if(node->mape == NULL)
		mape_error("get ppi's index\n", __FILE__, __LINE__) ;
	
	return ((mape_t *)(node->mape))->C0  ;
}

/*
 * mape_upb_label_node
 *
 * label the node. 
 */
void
mape_upb_label_node(node, label)
node_t *node ;
int label ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;
	
	((mape_t *)(node->mape))->visit = label ;
}

/*
 * mape_upb_next_connect
 *
 * get the next connection for a node in tie. 
 */
void 
mape_upb_next_connect(node, d, next, tie)
node_t *node;
lsList next, tie ;
int d ;
{
	int label ;
	lsGen gen ;
	int i ;
	node_t *p ;

	mape_upb_label_node(node, d) ;
	
	foreach_fanout(node, gen, p){
		if(!mape_node_in_list(p, tie))
			continue ;
		if(mape_upb_node_label(p) != NO_LABEL)
			continue ;
		lsNewEnd(next, (lsGeneric)p, LS_NH) ;
	}

	foreach_fanin(node, i, p){
		if(!mape_node_in_list(p, tie))
			continue ;
		if(mape_upb_node_label(p) != NO_LABEL)
			continue ;
		lsNewEnd(next, (lsGeneric)p, LS_NH) ;
	}
}

/*
 * mape_upb_node_distance. 
 *
 * 
 * calculate the distance between 
 * two node in the tie. 
 */
int 
mape_upb_node_distance(tie, n1, n2)
lsList tie ;
node_t *n1,  *n2 ;
{
	int d, rtv ;
	lsList front, next ;
	lsGen gen ;
	int find = FALSE;
	node_t *f_node ;

	front = lsCreate() ;
	next = lsCreate() ;

	d = 0 ;
	/* first clean all the visit mark in the tie's list */
	lsForeachItem(tie, gen, f_node)
		mape_upb_label_node(f_node, NO_LABEL) ;

	lsNewEnd(front, (lsGeneric)n1, LS_NH) ;
	while(d <= DONT_CARE_DISTANCE && !find){
		while(lsLength(front)){
			LS_ASSERT(lsDelEnd(front, (lsGeneric *)&f_node)) ;

			if(f_node == n2){
				find = TRUE ;
				rtv = d ;
			}else{
				mape_upb_next_connect(f_node, d, next, tie) ;
			}
		}
		d ++ ;
		if(find)
			break ;
		while(lsLength(next)){
			LS_ASSERT(lsDelEnd(next, (lsGeneric *)&f_node)) ;
			lsNewEnd(front, (lsGeneric)f_node, LS_NH) ;
		}
	}
	lsDestroy(front, NULL) ;
	lsDestroy(next, NULL) ;

	if(!find)
		return DONT_CARE_DISTANCE;
	else 
		return rtv ;
}

/*
 * mape_upb_cal_distance
 *
 * 
 * calculate the distance of any pairs of
 * the node in the node's list. 
 */
int *
mape_upb_cal_distance(tie, ppi)
lsList tie, ppi ;
{
	int length ;
	int *rtv  ;
	int i, j, distance ;
	node_t *n1 , *n2 ;
	lsGen gen1, gen2 ;

	if(!ppi || !tie)
		mape_error("invalid ppi or tie value\n", __FILE__, __LINE__) ;
	if(!lsLength(tie) || !lsLength(ppi))
		mape_error("ppi or tie is empty\n", __FILE__, __LINE__) ;

	length = lsLength(ppi) ;
	rtv = ALLOC(int, length*length) ;

	i =0 ;
	lsForeachItem(ppi, gen1, n1){
		j = 0 ;
		lsForeachItem(ppi, gen2, n2){
			if(j == i)
				distance = 0 ;
			else if(j < i)
				distance = rtv[j*length + i] ;
			else 
				distance = mape_upb_node_distance(tie, n1, n2) ;

			rtv[i*length+j] = distance ;
			j++;
		}
		i++ ;
	}
	return rtv ;
}

/*
 * mape_upb_max_distance_pair
 *
 *
 * get the maximum distance pair. 
 */
void 
mape_upb_max_distance_pair(d, ppi, n1, n2)
int *d ;
lsList ppi ;
node_t **n1, **n2 ;
{
	int distance ; 
	node_t *p1, *p2 ;
	lsGen gen1, gen2 ;
	int i,j, length ;
	int max = 0; 
	lsHandle h1, h2, h1_m, h2_m ; 

	length = lsLength(ppi) ;
	i = 0, j = 0 ;

	/* find the two nodes with the maximum distance */
	gen1 = lsStart(ppi) ;
	while(lsNext(gen1, (lsGeneric *)&p1, &h1) == LS_OK){
		gen2 = lsStart(ppi) ;
		while(lsNext(gen2, (lsGeneric *)&p2, &h2) == LS_OK){
			distance = d[i*length + j] ;

			if(distance >= DONT_CARE_DISTANCE){
				lsFinish(gen2) ;
				lsFinish(gen1) ;
				LS_ASSERT(lsRemoveItem(h1, (lsGeneric *)n1)) ;
				LS_ASSERT(lsRemoveItem(h2, (lsGeneric *)n2)) ;
				return ;
			}
			if(distance > max){
				max = distance ;
				h1_m = h1 ;
				h2_m = h2 ;
			}
			j ++ ;
		}
		lsFinish(gen2) ;
			
		i++ ;
		j = 0 ;
	}
	lsFinish(gen1) ;
	LS_ASSERT(lsRemoveItem(h1_m, (lsGeneric *)n1)) ;
	LS_ASSERT(lsRemoveItem(h2_m, (lsGeneric *)n2)) ;
}

/*
 * return the total distance from one 
 * node to another. 
 *
 */
int 
mape_upb_total_distance(d, length, node, list)
int *d, length ;
node_t *node ;
lsList list ;
{
	int i , j ;
	lsGen gen ;
	node_t *p ;
	int total = 0  ;

	i = mape_upb_get_ppi_index(node) ;
	lsForeachItem(list, gen, p){
		j = mape_upb_get_ppi_index(p) ;
		total += d[i*length + j] ;
	}
	return total ;
}

/*
 * mape_upb_max_unbanlance
 *
 *
 * find the node with the maximum unbanlance 
 * to all the p1 and p2's list. 
 */
int
mape_upb_max_unbanlance(d, length, ppi, p1, p2, node)
int *d, length ;
lsList ppi, p1, p2 ;
node_t **node ;
{
	int diff = 0 ;
	lsGen gen ;
	node_t *p  ;
	lsHandle h, hm ;
	int max = -1 ;
	int rtv ;

	gen = lsStart(ppi) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		diff = mape_upb_total_distance(d, length, p, p1) - mape_upb_total_distance(d,length, p, p2) ;
		if((diff>0?diff:-diff)>max){
			max = (diff>0?diff:-diff) ;
			hm = h ;
			rtv = (diff>0?SECOND_SIDE:FIRST_SIDE) ;
		}
	}
	lsFinish(gen) ;

	LS_ASSERT(lsRemoveItem(hm, (lsGeneric *)node)) ;
	return rtv ;
}

/*
 * mape_upb_bipartition_build_distance. 
 *
 *
 * build the distance matrix of the a list. 
 */
void 
mape_upb_bipartition_build_distance(d, length, list, d_list)
int *d, *d_list, length ;
lsList list ;
{
	int i, j;
	lsGen gen1, gen2 ;
	node_t *n1 , *n2;
	int n_length ;

	i = 0, j = 0 ;
	n_length = lsLength(list) ;
	lsForeachItem(list, gen1, n1){
		lsForeachItem(list, gen2, n2){
			d[i*n_length + i] = d[mape_upb_get_ppi_index(n1)*length + 
					      mape_upb_get_ppi_index(n2)] ;
			j ++ ;
		}
		i++ ;
		j = 0 ;
	}
}


/*
 * mape_upb_bipartition_ppi
 *
 *
 * split the ppi into 2 sub set. 
 */
void 
mape_upb_bipartition_ppi(d, ppi, d1, p1, d2, p2)
int *d ; 		/*distance vector for ppi */
int **d1, **d2 ; 	/*returned distance vectors for p1 and p2*/ 
lsList ppi, p1, p2 ;
{
	node_t *n1, *n2 ;
	int i = 0 ;
	lsGen gen ;
	int length ;

	/*set the index for all the node in ppi */
	lsForeachItem(ppi, gen, n1){
		mape_upb_set_ppi_index(n1, i) ;
		i++ ;
	}

	length = lsLength(ppi) ;
		
	/* pop out the maximum distance node pair*/
	mape_upb_max_distance_pair(d, ppi, &n1, &n2);
	
	/* push max1 into p1 and max2 into p2 */
	lsNewEnd(p1, (lsGeneric)n1, LS_NH) ;
	lsNewEnd(p2, (lsGeneric)n2, LS_NH) ;

	/* for each node , find the node with maximum unbalance*/
	while(lsLength(ppi)){
		if(mape_upb_max_unbanlance(d,length, ppi, p1, p2, &n1) == FIRST_SIDE){
			lsNewEnd(p1, (lsGeneric)n1, LS_NH);
		}else{
			lsNewEnd(p2, (lsGeneric)n1, LS_NH);
		}
	}

	/* building the return node's distance matrix */
	*d1 = ALLOC(int, lsLength(p1)*lsLength(p1)) ;
	*d2 = ALLOC(int, lsLength(p1)*lsLength(p2)) ;
	mape_upb_bipartition_build_distance(d, length, p1, *d1) ;
	mape_upb_bipartition_build_distance(d, length, p2, *d2) ;
}

/*
 * mape_upb_group_ppi
 *
 * take out a ppi's group. 
 * if no node in ppi list, return FALSE ;
 */
void
mape_upb_group_ppi(d, ppi, groups, size)
int *d ;
lsList ppi ;
lsList groups ;
int  size ;
{
	node_t *node ;
	lsList g_ppi ;
	lsList p1, p2;
	int *d1, *d2 ;

	/*no node at all */
	if(lsLength(ppi) == 0)
		return  ;

	/* one group */
	if(lsLength(ppi) <= size){
		g_ppi = lsCreate() ;
		while(lsLength(ppi)){
			LS_ASSERT(lsDelEnd(ppi, (lsGeneric *)&node)) ;
			lsNewEnd(g_ppi, (lsGeneric)node, LS_NH) ;
		}
		lsNewEnd(groups, (lsGeneric)g_ppi, LS_NH) ;
		return ;
	}

	/*2 groups*/
	p1 = lsCreate() ;
	p2 = lsCreate() ;
	mape_upb_bipartition_ppi(d, ppi, &d1, p1, &d2, p2) ;

	mape_upb_group_ppi(d1, p1, groups, size) ;
	mape_upb_group_ppi(d2, p2, groups, size) ;
	lsDestroy(p1, NULL) ;
	lsDestroy(p2, NULL) ;
	free(d1) ;
	free(d2) ;
}

/*
 * mape_upb_min_loss
 *
 * get the minimum loss of the transistion 
 * for a tie. 
 */
void
mape_upb_min_loss(tie, non_trans)
lsList tie ;
mape_non_trans_t *non_trans ;
{
	lsList ppi;
	lsList g_ppi ; /*grouped ppi, which is used to do exhaustive simulation */
	int i ;
	node_t *node ;
	node_t *tmp ;
	lsGen gen ;
	lsList groups ;
	int *d ;
	int j ;

	ppi = lsCreate() ;
	groups = lsCreate() ;

	/* make the ppi out from all the targe nodes */
	mape_upb_make_ppi(tie, ppi);
	fprintf(mapeout, "The ppi list:\n") ;
	mape_upb_dump_tie(ppi) ;

	/* make the distance vector for all the ppi*/
	d = mape_upb_cal_distance(tie, ppi) ;

	/* checking the distance matrix */
	for(i=0;i<lsLength(ppi);i++){
		for(j=0;j<lsLength(ppi);j++){
			fprintf(mapeout, "%-3d", d[i*lsLength(ppi)+j]) ;
		}
		fprintf(mapeout, "\n") ;
	}


	/* group the ppis together */
	mape_upb_group_ppi(d, ppi, groups, GROUP_SIZE);

	/* don't use any longer */
	free(d) ;

	/*
	fprintf(mapeout, "\ntie:\n") ;
	mape_upb_dump_tie(tie) ;
	fprintf(mapeout, "\nppi:\n") ;
	mape_upb_dump_tie(ppi) ;
	*/

	lsForeachItem(groups, gen, g_ppi){
		fprintf(mapeout, "\ngroup: ") ;
		/*
		mape_upb_exhaustive_simulate(g_ppi, non_trans) ;
		*/
		while(lsLength(g_ppi)){
			lsDelEnd(g_ppi, (lsGeneric *)&tmp) ;
			fprintf(mapeout, " %s(%d)", tmp->name, mape_upb_get_ppi_index(tmp)) ;
		}
		lsDestroy(g_ppi, NULL) ;
	}
	fprintf(mapeout, "\n") ;

	lsDestroy(ppi, NULL) ;
	lsDestroy(groups, NULL) ;
}

/*
 * mape_upb_single_connect
 *
 * is the node a single connect to the tie? 
 */
int 
mape_upb_single_connect(node, tie)
node_t *node ;
lsList tie ;
{
	int c_n = 0 ;
	int i ;
	node_t *p ;
	lsGen gen ;

	foreach_fanin(node, i, p){
		if(mape_node_in_list(p, tie)){
			c_n++ ;
			if(c_n >= 2)
				return FALSE ;
		}
	}

	foreach_fanout(node, gen, p){
		if(mape_node_in_list(p, tie)){
			c_n++ ;
			if(c_n >= 2)
				return FALSE ;
		}
	}

	if(c_n <= 1)
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_upb_concise_tie
 *
 * concise a special tie. return TRUE means
 * after concise the tie is not empty, otherwise
 * it is empty. 
 */
int 
mape_upb_concise_tie(tie)
lsList tie ;
{
	int rtv ;
	int find = FALSE;
	node_t *node, *tmp ;
	lsHandle h ;
	lsGen gen ;

	if(lsLength(tie) == 0)
		return FALSE ;

	gen = lsStart(tie);
	while(lsNext(gen, (lsGeneric *)&node, &h) == LS_OK){
		if(mape_upb_single_connect(node, tie)){
			LS_ASSERT(lsDelBefore(gen,(lsGeneric *)&tmp)) ;
			find = TRUE ;
		}
	}
	lsFinish(gen) ;

	if(!find)
		return TRUE ;
	else 
		return mape_upb_concise_tie(tie) ;
}

/*
 * mape_upb_concise
 *
 * make the tie list concise. 
 */
void 
mape_upb_concise(void)
{
	lsGen gen ;
	lsList tie, tmp ;
	lsHandle h ;

	gen = lsStart(tie_list) ;
	while(lsNext(gen, (lsGeneric *)&tie, &h) == LS_OK){
		if(!mape_upb_concise_tie(tie))
			LS_ASSERT(lsDelBefore(gen,(lsGeneric *)&tmp)) ;
	}
	lsFinish(gen) ;
}

/*
 * mape_upb
 *
 * maximum power upper boundary estimation. 
 */
void
mape_upb(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;
	int gain = 0;
	mape_non_trans_t *non_trans ;
	lsList tie ;

	dump_network(network) ;
	non_trans = mape_upb_create_non_trans_t() ;

	foreach_node(network, gen, node)
		mape_upb_make_tie_list(node) ;
	mape_upb_concise() ;
	mape_upb_dump() ;

	if(tie_list){
		lsForeachItem(tie_list, gen, tie)
			mape_upb_min_loss(tie, non_trans) ;
	}

	fprintf(mapeout, "Global gain = %d\n", non_trans->global_gain) ;
	lsDestroy(non_trans->global, NULL) ;
	lsDestroy(non_trans->min, NULL) ;
	lsDestroy(non_trans->cur, NULL) ;
	free(non_trans) ;
}
