/*
 * mape_util.c
 * 
 * in the file, some basic function has been provided 
 * These utilities includes the levelization of the circuit. 
 * 
 * Changbo Long Nov. 26, 2001. (clong@cae.wisc.edu)
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include "mape_int.h"

#define CUBE_SHARP_0 0
#define CUBE_SHARP_1 1
#define CUBE_SHARP_Z 2
#define CUBE_SHARP_Y 3

long acc_time ;         /*accumulated time*/
long one_time ;         /*time for one calculation*/

long acc_space ;	/*accumulated space*/
long one_space ;	/*space for one calculation*/ 


/* the conflict table for two node_value */
int mape_conflict_table[9][9] = {
	{0,1,0,1,1,0,1,0,1}, 
	{1,0,1,0,1,0,1,1,0}, 
	{0,1,0,0,0,0,1,0,1}, 
	{1,0,0,0,0,0,1,1,0}, 
	{1,1,0,0,0,0,1,1,1}, 
	{0,0,0,0,0,0,0,0,0}, 
	{1,1,1,1,1,0,0,0,0}, 
	{0,1,0,1,1,0,0,0,0}, 
	{1,0,1,0,1,0,0,0,0}
};

/*
 * mape_error
 *
 * printf error message and output
 */
void
mape_error(s, file, line)
char *s, *file; 
int line ;
{
	fprintf(stderr, "%s(line %d) Error: %s\n", file, line, s) ;
	exit(1) ;
}

/*
 * debug message.
 */
void 
mape_debug(s)
char *s ;
{
	fprintf(mapeout, "%s", s) ;
	fflush(mapeout) ;
}

/*
 * copied from sis
 */
static int intcpy(d, s, n)
register unsigned int *d, *s;
register long n;
{
	register int i;
	for(i = 0; i < n; i++) {
		*d++ = *s++;
	}
}

/*
 * mape_set_node_level
 *
 * assign the level to a node into the mape. 
 */
void 
mape_set_node_level(node, level)
node_t *node ;
int level ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)(node->mape))->level = level ;
}

/*
 * assign_level.  
 *
 * Copied from speed_util.c
 */
int
mape_assign_level(node)
node_t *node;
{
	node_t *fanin;
	int i, level, new_level;
	char *dummy;

	new_level = (int)(node->mape == NULL?0:((mape_t *)(node->mape))->level) ;
	if (node->type == PRIMARY_INPUT || new_level > 0) {
		return new_level ;
	} else {
		level = 0;
		foreach_fanin(node, i, fanin) {
			level = MAX(level, mape_assign_level(fanin));
		}
	}
	level++;

	mape_set_node_level(node, level) ;
	return level;
}

/*
 * mape_levelize
 *
 * levelizaiton of the whole circuit. 
 *
 */
mape_levelize(network)
network_t *network ;
{
	lsGen gen ;
	node_t *p ;

	foreach_primary_output(network, gen, p)
		mape_assign_level(p) ;
}


/*
 * mape_get_node_level
 *
 * get the level of every node in the circuit. 
 *
 */
int 
mape_get_node_level(node)
node_t *node ;
{
	static int maked = FALSE ;

	if(!maked){
		mape_levelize(node->network) ;
		maked = TRUE ;
	}

	if(node->mape == NULL)
		return 0 ;
	return ((mape_t *)(node->mape))->level ;
}

/*
 * mape_level_cmp
 *
 * for array_sort, compare the level of two node.  
 */
int
mape_level_cmp(a1, a2)
lsGeneric a1 ;
lsGeneric a2 ;
{
	int l1 = mape_get_node_level((node_t *)a1) ; 
	int l2 = mape_get_node_level((node_t *)a2) ; 
	return l1-l2 ;
}

/*
 * mape_fanout_cmp
 *
 * for array_sort. compare the fanout number of two node. 
 */
int
mape_fanout_cmp(a1, a2)
char **a1 ;
char **a2 ;
{
	int n1 = lsLength( ((node_t *)(*a1))->fanout ) ;
	int n2 = lsLength( ((node_t *)(*a2))->fanout ) ;
	return n1-n2 ;
}

/*
 * mape_node_clear_reps
 *
 * clear all the node inside the reps. 
 */
void 
mape_node_clear_reps(node)
node_t *node ;
{
	node_t *tmp ;
	lsList reps ;

	if(!node->mape)
		return ;
	reps = ((mape_t *)(node->mape))->reps ;

	if(!reps)
		return ;

	while(lsLength(reps)) 
		LS_ASSERT(lsDelEnd(reps, (lsGeneric *)&tmp)) ;
}

/*
 * mape_node_push_reps
 *
 * push a node into node's reps list. 
 */
void 
mape_node_push_reps(node, p)
node_t *node , *p ;
{
	lsList reps ;

	if(!p)
		return ;
	if(!node->mape){
		node->mape = (char *)mape_create_mape_t() ;
		((mape_t*)(node->mape))->reps = lsCreate() ;
	}else
		if(!((mape_t*)(node->mape))->reps)
			((mape_t*)(node->mape))->reps = lsCreate() ;
	
	reps = ((mape_t*)(node->mape))->reps ;
	if(!mape_node_in_list(p, reps)){
		lsNewEnd(reps, (lsGeneric)p, LS_NH) ;
		//fprintf(mapeout, "push pi %s into node %s's reps list\n", p->name, node->name) ;
	}
}

/*
 * mape_node_push_reps_list. 
 *
 * push a list of node into the node's reps list. 
 */
void 
mape_node_push_reps_list(node, list)
node_t *node ;
lsList list ;
{
	lsGen gen ;
	node_t *p ;

	if(!list)
		return ;
	if(lsLength(list) == 0)
		return ;

	if(!node->mape){
		node->mape = (char *)mape_create_mape_t() ;
		((mape_t*)(node->mape))->reps = lsCreate() ;
	}else
		if(!((mape_t*)(node->mape))->reps)
			((mape_t*)(node->mape))->reps = lsCreate() ;

	if(!((mape_t*)(node->mape))->reps)
		((mape_t*)(node->mape))->reps = lsCreate() ;
	lsForeachItem(list, gen, p)
		mape_node_push_reps(node, p) ;
}

/*
 * mape_create_mape_t ;
 *
 * create a new mape_t structure.
 */
mape_t *
mape_create_mape_t(void)
{
	mape_t *rtv ;

	rtv 			= ALLOC(mape_t, 1) ;
	rtv->values 		= lsCreate() ;
	rtv->cur_values		= NULL ;
	rtv->pi_fanin		= NULL ;
	rtv->rdm_v 		= MAPE_U_U ;
	rtv->rdm_max 		= MAPE_U_U ;
	rtv->level 		= 0 ;
	rtv->visit 		= FALSE ;
	rtv->C0			= 0 ;
	rtv->C1			= 0 ;
	rtv->reps		= NULL ;
	rtv->curlib_gate	= NULL ;
	rtv->plt		= NULL ;
	rtv->cs			= NULL ;
	rtv->depsa		= NULL ;

	rtv->sim.curr		= NULL ;
	rtv->sim.hc		= NULL ;
	rtv->sim.hp		= NULL ;

	rtv->pbssce.tmpr	= 0.0 ;
	rtv->pbssce.cl		= NULL ;
	rtv->pbssce.gen		= NULL ;
	rtv->pbssce.zol		= NULL ;
	rtv->pbssce.curzo	= NULL ;
	rtv->pbssce.sw_static	= NULL ;
	rtv->pbssce.cantl	= NULL ;

	return rtv ;
}

/*
 * mape_create_imp_value
 *
 * create a new mape_imp_value_t structure. 
 */
mape_imp_value_t *
mape_create_imp_value(value, pi, pi_value)
mape_node_value_t value ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_imp_value_t *rtv ;

	rtv 		= ALLOC(mape_imp_value_t, 1) ;
	rtv->value 	= value ;
	rtv->pi 	= pi ;	
	rtv->pi_value	= pi_value ;

	return rtv ;
}

/*
 * mape_free_mape_t ;
 *
 * free a mape_t structure. 
 */
void 
mape_free_mape_t(v)
mape_t *v ;
{
	lsDestroy(v->values, NULL) ;
	FREE(v) ;
}

/*
 * mape_pop
 *
 * pop a value from the node's stack
 */
int 
mape_pop(node, pi, pi_value)
node_t *node ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	lsList values ;
	mape_imp_value_t  *imp_value ;
	char txt[128] ;

	if(node == NULL)
		mape_error("stupid programming! email clong@cae.wisc.edu for debuggering ", 
			__FILE__, __LINE__) ;
	if(node->mape == NULL)
		return FALSE ;

	if(((mape_t *)(node->mape))->values == NULL) 
		mape_error("stupid programming! email clong@cae.wisc.edu for debuggering ", 
			__FILE__, __LINE__) ;
		
	values = ((mape_t *)(node->mape))->values ;
	if(lsLength(values) == 0)
		return FALSE;

	(void)lsLastItem(values, (lsGeneric *)&imp_value, LS_NH);

	if(imp_value->pi != pi)
		return FALSE;
	if(imp_value->pi_value != pi_value)
		return FALSE;
		//mape_error("Do you imply from one pi twice?", __FILE__, __LINE__) ; 

	(void)lsDelEnd(values, (lsGeneric *)&imp_value) ;
	/*
	fprintf(mapeout, "pop %s from %s \n", 
		mape_node_value_string(imp_value->value, txt), node->name);
	fflush(mapeout) ;
	*/
	FREE(imp_value) ;
	return TRUE ;
}

/*
 * mape_rdm_value_first.  
 *
 * return the first part of the random stack's value 
 */
int
mape_rdm_value_first(node)
node_t *node ;
{
	mape_node_value_t value = mape_rdm_value(node) ;
	switch(value){
		case MAPE_FALL:
		case MAPE_1_1:
		case MAPE_1_U:
			return ONE ;
		case MAPE_0_0:
		case MAPE_RISE:
		case MAPE_0_U:
			return ZERO ;
		case MAPE_U_0:
		case MAPE_U_1:
		case MAPE_U_U:
			return TWO ;
		default: 
			dump_network(global_network) ;
			fflush(mapeout) ;
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_rdm_value_first.  
 *
 * return the first part of the random stack's value 
 */
int
mape_rdm_value_second(node)
node_t *node ;
{
	mape_node_value_t value = mape_rdm_value(node) ;
	switch(value){
		case MAPE_RISE:
		case MAPE_U_1:
		case MAPE_1_1:
			return ONE ;
		case MAPE_FALL:
		case MAPE_U_0:
		case MAPE_0_0:
			return ZERO ;
		case MAPE_U_U:
		case MAPE_0_U:
		case MAPE_1_U:
			return TWO ;
		default: 
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_rdm_value_max
 *
 * return the node's random-max stack. 
 */
mape_node_value_t
mape_rdm_value_max(node)
node_t *node ;
{
	if(node->mape == NULL)
		return MAPE_U_U ;

	return ((mape_t *)(node->mape))->rdm_max ;
}

/*
 * mape_rdm_value
 *
 * return the node's random stack
 */
mape_node_value_t
mape_rdm_value(node)
node_t *node ;
{
	if(node->mape == NULL)
		return MAPE_U_U ;

	return ((mape_t *)(node->mape))->rdm_v ;
}

/*
 * mape_push_rdm_max
 *
 * push a value into a node's random_max stack. 
 */
void 
mape_push_rdm_max(node, value)
node_t *node; 
mape_node_value_t value ;
{
	((mape_t *)(node->mape))->rdm_max = value ;
}

/*
 * mape_push_rdm
 *
 * push a value into a node's random stack. 
 */
void 
mape_push_rdm(node, value)
node_t *node; 
mape_node_value_t value ;
{
	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)(node->mape))->rdm_v = value ;
}

/*
 * mape_push
 *
 * push a value into the stack of node. 
 */
void 
mape_push(node, value, pi, pi_value)
node_t *node; 
mape_node_value_t value ;
node_t *pi ;
mape_node_value_t pi_value ;
{
	lsList values ;
	mape_imp_value_t  *imp_value ;
	char txt[128] ;

	/*
	fprintf(mapeout, "push %s into %p \n", 
		mape_node_value_string(value, txt), node);
	fflush(mapeout) ;
	*/

	if(node->mape == NULL)
		node->mape = (char *)mape_create_mape_t() ;

	values = ((mape_t *)(node->mape))->values ;

	imp_value = mape_create_imp_value(value, pi, pi_value) ;
	lsNewEnd(values, (lsGeneric)imp_value, NULL) ; 
}

/*
 * mape_get_value_index
 *
 * return the index to a specified value. 
 */
int
mape_get_value_index(v)
mape_node_value_t v ;
{
	switch (v){
		case MAPE_FALL:
			return 0 ;
		case MAPE_RISE:
			return 1 ;
		case MAPE_U_0:
			return 2 ;
		case MAPE_0_U:
			return 3 ;
		case MAPE_0_0:
			return 4 ;
		case MAPE_U_U:
			return 5 ;
		case MAPE_1_1:
			return 6 ;
		case MAPE_1_U:
			return 7 ;
		case MAPE_U_1:
			return 8 ;
	}
}

/*
 * mape_3_to_9
 *
 * 3-value converted to 9-value 
 *
 */
mape_node_value_t 
mape_3_to_9(v)
int v ;
{
	switch(v){
		case ONE: 
			return MAPE_1_1 ;
		case ZERO: 
			return MAPE_0_0 ;
		case TWO: 
		default:
			return MAPE_U_U ;
	}
}

/*
 * mape_set_value_not
 *
 * return the not set's value. 
 */
int 
mape_set_value_not(v)
int v; 
{
	switch(v){
		case ONE: 
			return ZERO ;
		case ZERO: 
			return ONE ;
		case TWO: 
		default:
			return TWO ;
	}
}

/*
 * mape_value_convert
 *
 */
mape_node_value_t 
mape_value_convert(first, second)
int first ;
int second ;
{
	if(first == ZERO){
		if(second == ZERO)
			return MAPE_0_0 ;
		else if(second == ONE)
			return MAPE_RISE ;
		else 
			return MAPE_0_U ;
	}else if(first == ONE){
		if(second == ZERO)
			return MAPE_FALL ;
		else if(second == ONE)
			return MAPE_1_1 ;
		else 
			return MAPE_1_U ;
	}else{
		if(second == ZERO)
			return MAPE_U_0 ;
		else if(second == ONE)
			return MAPE_U_1 ;
		else 
			return MAPE_U_U ;
	}
}

/*
 * mape_value_second_not
 *
 * return the index to a specified value. 
 */
mape_node_value_t
mape_value_second_not(v)
mape_node_value_t v ;
{
	switch (v){
		case MAPE_FALL:
			return MAPE_1_1 ;
		case MAPE_RISE:
			return MAPE_0_0 ;
		case MAPE_U_0:
			return MAPE_U_1 ;
		case MAPE_0_U:
			return MAPE_0_U;
		case MAPE_0_0:
			return MAPE_RISE ;
		case MAPE_U_U:
			return MAPE_U_U ;
		case MAPE_1_1:
			return MAPE_FALL ;
		case MAPE_1_U:
			return MAPE_1_U ;
		case MAPE_U_1:
			return MAPE_U_0 ;
	}
}

/*
 * mape_value_first_not
 *
 * return the index to a specified value. 
 */
mape_node_value_t
mape_value_first_not(v)
mape_node_value_t v ;
{
	switch (v){
		case MAPE_FALL:
			return MAPE_0_0 ;
		case MAPE_RISE:
			return MAPE_1_1 ;
		case MAPE_U_0:
			return MAPE_U_0 ;
		case MAPE_0_U:
			return MAPE_1_U;
		case MAPE_0_0:
			return MAPE_FALL ;
		case MAPE_U_U:
			return MAPE_U_U ;
		case MAPE_1_1:
			return MAPE_RISE ;
		case MAPE_1_U:
			return MAPE_0_U ;
		case MAPE_U_1:
			return MAPE_U_1 ;
	}
}

/*
 * mape_value_not
 *
 * return the index to a specified value. 
 */
mape_node_value_t
mape_value_not(v)
mape_node_value_t v ;
{
	switch (v){
		case MAPE_FALL:
			return MAPE_RISE ;
		case MAPE_RISE:
			return MAPE_FALL ;
		case MAPE_U_0:
			return MAPE_U_1 ;
		case MAPE_0_U:
			return MAPE_1_U;
		case MAPE_0_0:
			return MAPE_1_1 ;
		case MAPE_U_U:
			return MAPE_U_U ;
		case MAPE_1_1:
			return MAPE_0_0 ;
		case MAPE_1_U:
			return MAPE_0_U ;
		case MAPE_U_1:
			return MAPE_U_0 ;
	}
}

/*
 * mape_conflict
 *
 * if the two values are conflict, return 1 ;
 * else return 0 ;
 */
int 
mape_conflict(v1, v2)
mape_node_value_t v1; 
mape_node_value_t v2;
{
	int ind1 = mape_get_value_index(v1) ;
	int ind2 = mape_get_value_index(v2) ;

	return mape_conflict_table[v1][v2] ;
}

/*
 * mape_value_second
 *
 * return the second value.
 */
int 
mape_value_second(value) 
mape_node_value_t value ;
{
	switch(value){
		case MAPE_RISE:
		case MAPE_U_1:
		case MAPE_1_1:
			return ONE ;
		case MAPE_FALL:
		case MAPE_U_0:
		case MAPE_0_0:
			return ZERO ;
		case MAPE_U_U:
		case MAPE_0_U:
		case MAPE_1_U:
			return TWO ;
		default: 
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_value_first
 *
 * return the first value 
 */
int 
mape_value_first(value) 
mape_node_value_t value ;
{
	switch(value){
		case MAPE_FALL:
		case MAPE_1_1:
		case MAPE_1_U:
			return ONE ;
		case MAPE_0_0:
		case MAPE_RISE:
		case MAPE_0_U:
			return ZERO ;
		case MAPE_U_0:
		case MAPE_U_1:
		case MAPE_U_U:
			return TWO ;
		default: 
			dump_network(global_network) ;
			fflush(mapeout) ;
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_top_second
 *
 * return the second value in the top of the node. 
 */
int 
mape_top_second(node) 
node_t *node ;
{
	mape_node_value_t value = mape_top(node) ;
	switch(value){
		case MAPE_RISE:
		case MAPE_U_1:
		case MAPE_1_1:
			return ONE ;
		case MAPE_FALL:
		case MAPE_U_0:
		case MAPE_0_0:
			return ZERO ;
		case MAPE_U_U:
		case MAPE_0_U:
		case MAPE_1_U:
			return TWO ;
		default: 
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_top_first
 *
 * return the first value in the top of the node. 
 */
int 
mape_top_first(node) 
node_t *node ;
{
	mape_node_value_t value = mape_top(node) ;
	switch(value){
		case MAPE_FALL:
		case MAPE_1_1:
		case MAPE_1_U:
			return ONE ;
		case MAPE_0_0:
		case MAPE_RISE:
		case MAPE_0_U:
			return ZERO ;
		case MAPE_U_0:
		case MAPE_U_1:
		case MAPE_U_U:
			return TWO ;
		default: 
			dump_network(global_network) ;
			fflush(mapeout) ;
			mape_error("Unexpered value", __FILE__, __LINE__) ;
	}
}

/*
 * mape_top 
 *
 * return the top in the node's value stack. 
 */
mape_node_value_t 
mape_top(node)
node_t *node ;
{
	lsList values ; 
	mape_imp_value_t *rtv ;

	if(node->mape == NULL)
		return MAPE_U_U ; 

	values = ((mape_t *)(node->mape))->values ;
	if(lsLength(values) == 0)
		return MAPE_U_U ; 
	(void) lsLastItem(values, (lsGeneric *)&rtv, LS_NH) ; 
		
	return (mape_node_value_t) (rtv->value) ;
}

/*
 * mape_top_pi_value 
 *
 * return the pi's value in the top of the node's value stack. 
 */
mape_node_value_t 
mape_top_pi_value(node)
node_t *node ;
{
	lsList values ; 
	mape_imp_value_t *rtv ;

	if(node->mape == NULL)
		return MAPE_U_U ; 

	values = ((mape_t *)(node->mape))->values ;
	if(lsLength(values) == 0)
		return MAPE_U_U ; 
	(void) lsLastItem(values, (lsGeneric *)&rtv, LS_NH) ; 
		
	return (mape_node_value_t) (rtv->pi_value) ;
}

/*
 * mape_top_pi
 *
 * return which pi implication to create the node top's value
 */
node_t *
mape_top_pi(node)
node_t *node ;
{
	lsList values ; 
	mape_imp_value_t *rtv ;

	if(node->mape == NULL)
		return NULL ; 

	values = ((mape_t *)(node->mape))->values ;
	if(lsLength(values) == 0)
		return NULL ; 
	(void) lsLastItem(values, (lsGeneric *)&rtv, LS_NH) ; 
		
	return (node_t *) (rtv->pi) ;
}

/* 
 * mape_sf_append
 *
 * append the sets of B to the end of A, clear the sets in B. 
 */
pset_family mape_sf_append(A, B)
pset_family A, B;
{
	long asize = A->count * A->wsize;
	long bsize = B->count * B->wsize;

	if (A->sf_size != B->sf_size) fatal("sf_append: sf_size mismatch");
	A->capacity = A->count + B->count;
	A->data = REALLOC(unsigned int, A->data, (long) A->capacity * A->wsize);
	intcpy(A->data + asize, B->data, bsize);
	A->count += B->count;
	A->active_count += B->active_count;

	/* clean B */
	B->count = 0;
	B->active_count = 0;
	return A;
}

/*
 * mape_sf_append_cube
 *
 * append a set into the set family A
 */
pset_family mape_sf_append_set(A, b)
pset_family A ;
pset b ;
{
	long asize = A->count * A->wsize;

	/*
	if(A->wsize != SIZE(b)){
		fprintf(stderr, "%d != %d\n", A->wsize, SIZE(b)) ;
		mape_error("Size Family doesn't match with cube's size", __FILE__, __LINE__) ;
	}
	*/

	set_copy(A->data + asize, b) ;
	A->count += 1;
	return A;
}

/*
 * mape_node_ctrl_input
 *
 * is there a ctrled value in the input of the node. 
 */
int 
mape_node_ctrl_input(cover, pos, cv)
pset_family cover ;
int pos ;
int *cv ;
{
	pset tmp ;
	pset_family sf_tmp ;
	
	int i ;
	int j ;

	if(cover == NULL)
		return NO_CTRL_VALUE ;

	/* seach for the reverse part first */
	tmp = sf_or(cover);
	switch(GETINPUT(tmp, pos)){
		case ZERO:
			*cv = ONE ;
			set_free(tmp) ;
			return REVERSE_CTRL_VALUE ;
		case ONE:
			*cv = ZERO ;
			set_free(tmp) ;
			return REVERSE_CTRL_VALUE ;
		case TWO:
		default:
			break ;
	}

	/* search for the non-reverse part */

	/*construct 0XXXX, X0XXX, ... */
	for(j=0;j<cube.num_vars;j++){
		if(j == pos)
			PUTINPUT(tmp, j, ZERO) ;
		else 
			PUTINPUT(tmp, j, TWO) ;
	}
	sf_tmp = cb_sharp(tmp, cover);
	if(sf_tmp->count == 0){
		*cv = ZERO, 
		set_free(tmp) ;
		sf_free(sf_tmp) ;
		return NORMAL_CTRL_VALUE ;
	}

	/*construct 1XXXX, X1XXX, ... */
	for(j=0;j<cube.num_vars;j++){
		if(j == pos)
			PUTINPUT(tmp, j, ONE) ;
		else 
			PUTINPUT(tmp, j, TWO) ;
	}
	sf_tmp = cb_sharp(tmp, cover);
	if(sf_tmp->count == 0){
		*cv = ONE, 
		set_free(tmp) ;
		sf_free(sf_tmp) ;
		return NORMAL_CTRL_VALUE ;
	}

	set_free(tmp);
	sf_free(sf_tmp) ;
	return NO_CTRL_VALUE ;
}

/*
 * mape_node_control_value. 
 *
 * get a control value for a cover. 
 * 
 * return values: 
 *	NO_CTRL_VALUE 		: no control value exist. 
 *	NORMAL_CTRL_VALUE 	: there is a control value and output value is same as c_v. 
 *	REVERSE_CTRL_VALUE	: the output value is reverse of the control value. 
 */
int 
mape_node_ctrl_value(cover, pos, cv)
pset_family cover ;
int *pos ; 
int *cv ;  
{
	pset tmp = NULL ;
	pset_family sf_tmp = NULL ;
	
	int i ;
	int j ;

	/* seach for the reverse part first */
	tmp = sf_or(cover);
	for(i=0;i<cube.num_vars;i++){
		switch(GETINPUT(tmp, i)){
			case ZERO:
				*pos = i ;
				*cv = ONE ;
				if(tmp)
					set_free(tmp) ;
				return REVERSE_CTRL_VALUE ;
			case ONE:
				*pos = i ;
				*cv = ZERO ;
				if(tmp)
					set_free(tmp) ;
				return REVERSE_CTRL_VALUE ;
			case TWO:
			default:
				break ;
		}
	}

	/* search for the non-reverse part */
	for(i=0;i<cube.num_vars;i++){

		/*construct 0XXXX, X0XXX, ... */
		for(j=0;j<cube.num_vars;j++){
			if(j == i)
				PUTINPUT(tmp, j, ZERO) ;
			else 
				PUTINPUT(tmp, j, TWO) ;
		}
		sf_tmp = cb_sharp(tmp, cover);
		if(sf_tmp->count == 0){
			*cv = ZERO, 
			*pos = i ;
			if(tmp)
				set_free(tmp) ;
			if(sf_tmp)
				sf_free(sf_tmp) ;
			return NORMAL_CTRL_VALUE ;
		}

		/*construct 1XXXX, X1XXX, ... */
		for(j=0;j<cube.num_vars;j++){
			if(j == i)
				PUTINPUT(tmp, j, ONE) ;
			else 
				PUTINPUT(tmp, j, TWO) ;
		}
		sf_tmp = cb_sharp(tmp, cover);
		if(sf_tmp->count == 0){
			*cv = ONE, 
			*pos = i ;
			if(tmp)
				set_free(tmp) ;
			if(sf_tmp)
				sf_free(sf_tmp) ;
			return NORMAL_CTRL_VALUE ;
		}
	}

	if(tmp)
		set_free(tmp);
	if(sf_tmp)
		sf_free(sf_tmp) ;
	return NO_CTRL_VALUE ;
}

/*
 * mape_timer_reset
 *
 * reset the timer for global timing calculation
 */
void
mape_timer_reset(void)
{
	acc_time = 0 ;
}

/*
 * mape_timer_start(void)
 *
 * start record the time
 */
void 
mape_timer_start(void)
{
	one_time = util_cpu_time() ;
}

/*
 * mape_timer_stop(void)
 *
 * stop record the time. 
 */
void 
mape_timer_stop(void)
{
	acc_time += (util_cpu_time() - one_time) ;
}

/*
 * mape_timer_report
 *
 * report the time accumulated. 
 */
void 
mape_timer_report(s)
char *s ;
{
	if(s == NULL)
		fprintf(mapeout, "%s\n", util_print_time(acc_time)) ;
	else 
		fprintf(mapeout, "%s:%s\n", s, util_print_time(acc_time)) ;
}

/*
 * mape_spacer_reset
 *
 * reset the space calculator
 */
void 
mape_spacer_reset(void)
{
	acc_space = 0 ;
}

/*
 * mape_spacer_start
 *
 * space calculator start 
 */
void 
mape_spacer_start(void)
{
	one_space = (long)sbrk(0) ;
	printf("one_space=%ld\n", one_space) ;
}

/*
 * mape_spacer_stop
 *
 * space calculator stop
 */
void 
mape_spacer_stop(void)
{
	acc_space += (long)sbrk(0) - one_space ;
	printf("acc_space=%ld\n", acc_space) ;
}

/*
 * mape_timer_report
 *
 * report the time accumulated. 
 */
void 
mape_spacer_report(s)
char *s ;
{
	if(s == NULL)
		fprintf(mapeout, "%ld\n", acc_space) ;
	else 
		fprintf(mapeout, "%s:%ld\n", s, acc_space) ;
}

/*
 * mape_clear_node_list 
 *
 * clear the node list. 
 */
void 
mape_clear_node_list(list)
lsList list ;
{
	node_t *node ;

	if(!list)
		return ;

	while(lsLength(list))
		lsDelEnd(list, (lsGeneric *)&node) ;
}

/*
 * mape_check_dir
 * 
 * check if directory is exist and accessable. 
 */
int 
mape_check_dir(dir)
char *dir;
{
	struct stat statbuf;

	if (! dir)
		return 0;
	else if (stat(dir, &statbuf) < 0)
		return 0;
	else if ((statbuf.st_mode & S_IFMT) != S_IFDIR)
		return 0;
	else if (access(dir, W_OK | X_OK) < 0)
		return 0;
	else
		return 1;
}

/*
 * mape_check_file
 * 
 * check if directory is exist and accessable. 
 */
int 
mape_check_regfile(path)
char *path;
{
	struct stat statbuf;

	if (! path)
		return 0;
	else if (stat(path, &statbuf) < 0)
		return 0;
	else if ((statbuf.st_mode & S_IFMT) != S_IFREG)
		return 0;
	else
		return 1;
}
