
# line 11 "read_eqn.y"
#include "sis.h"
#include "io_int.h"
#include <setjmp.h>

#ifdef YYBYACC
extern char *EQN_yytext;
#else
#if defined(hpux) || defined(__osf__)
extern unsigned char EQN_yytext[];
#else
extern char EQN_yytext[];
#endif
#endif

static network_t *global_network;
static lsList po_list;

static int EQN_yyerror();
#ifndef FLEX_SCANNER
#undef EQN_yywrap
static int input();
static int unput();
static int EQN_yywrap();
#endif


static void
do_assign(name, expr)
char *name;
node_t *expr;
{
    char errmsg[1024];
    node_t *node;

    node = read_find_or_create_node(global_network, name);
    if (node_function(node) != NODE_UNDEFINED) {
	(void) sprintf(errmsg, "Attempt to redefine '%s'\n", name);
	EQN_yyerror(errmsg);   /* never returns */
    }
    FREE(name);
    node_replace(node, expr);
}


static node_t *
do_sexpr_list(list, func)
array_t *list;
node_t *(*func)();
{
    int i;
    node_t *node, *node1, *node2;

    node1 = array_fetch(node_t *, list, 0);
    node = node_dup(node1);
    node_free(node1);
    for(i = 1; i < array_n(list); i++) {
	node1 = node;
	node2 = array_fetch(node_t *, list, i);
	node = (*func)(node1, node2);
	node_free(node1);
	node_free(node2);
    }
    array_free(list);
    return node;
}


# line 79 "read_eqn.y"
typedef union
#ifdef __cplusplus
	YYSTYPE
#endif
 {
    char *strval;
    node_t *node;
    array_t *array;
} YYSTYPE;
# define OPR_OR 257
# define OPR_AND 258
# define CONST_ZERO 259
# define CONST_ONE 260
# define IDENTIFIER 261
# define LPAREN 262
# define OPR_XOR 263
# define OPR_XNOR 264
# define OPR_NOT 265
# define OPR_NOT_POST 266
# define NAME 267
# define INORDER 268
# define OUTORDER 269
# define PASS 270
# define ASSIGN 271
# define SEMI 272
# define RPAREN 273
# define END 274

#include <inttypes.h>

#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#define	YYCONST	const
#else
#include <malloc.h>
#include <memory.h>
#define	YYCONST
#endif

#include <values.h>

#if defined(__cplusplus) || defined(__STDC__)

#if defined(__cplusplus) && defined(__EXTERN_C__)
extern "C" {
#endif
#ifndef EQN_yyerror
#if defined(__cplusplus)
	void EQN_yyerror(YYCONST char *);
#endif
#endif
#ifndef EQN_yylex
	int EQN_yylex(void);
#endif
	int EQN_yyparse(void);
#if defined(__cplusplus) && defined(__EXTERN_C__)
}
#endif

#endif

#define EQN_yyclearin EQN_yychar = -1
#define EQN_yyerrok EQN_yyerrflag = 0
extern int EQN_yychar;
extern int EQN_yyerrflag;
YYSTYPE EQN_yylval;
YYSTYPE EQN_yyval;
typedef int EQN_yytabelem;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
#if YYMAXDEPTH > 0
int EQN_yy_EQN_yys[YYMAXDEPTH], *EQN_yys = EQN_yy_EQN_yys;
YYSTYPE EQN_yy_EQN_yyv[YYMAXDEPTH], *EQN_yyv = EQN_yy_EQN_yyv;
#else	/* user does initial allocation */
int *EQN_yys;
YYSTYPE *EQN_yyv;
#endif
static int EQN_yymaxdepth = YYMAXDEPTH;
# define YYERRCODE 256

# line 240 "read_eqn.y"


#include "eqnlex.c"

static char *string;
static int file_mode;
static int nbuffer;
static char buffer[20];
static int last_was_newline;
static jmp_buf jmpbuf;

#ifndef FLEX_SCANNER
static int
input()
{
    register int c;
    register char *prompt;

    if (nbuffer > 0) {
	c = buffer[--nbuffer];
    } else if (file_mode) {
	if (last_was_newline && EQN_yyin == stdin) {
	    prompt = com_get_flag("prompt2");
	    if (prompt != NULL) fputs(prompt,stdout);
	}
	c = getc(EQN_yyin);
    } else {
	c = *string;
	if (c != 0) string++;
    }

    last_was_newline = (c == '\n');
    if (last_was_newline) read_lineno++;
    if (c == EOF) c = 0;
    return c;
}

static int
unput(c)
int c;
{
    if (nbuffer < 20) buffer[nbuffer++] = c;
    if (c == '\n') read_lineno--;
}
#else

char *eqn_inp_ptr;
char* eqn_inp_lim;

int eqn_input(char* buf, int max_size)
{
    char c;
    int i, n;
	 
    if (! file_mode) {
	if (max_size > (eqn_inp_lim - eqn_inp_ptr))
	    n = eqn_inp_lim - eqn_inp_ptr;
	else
	    n= max_size;
	if (n > 0) {
	    for (i = 0; i < n; i++)
		if (eqn_inp_ptr[i] == '\n') {
		    read_lineno++;
		    i++;
		    break;
		}
	    memcpy (buf, eqn_inp_ptr, i);
	    eqn_inp_ptr += i;
	}
    } else {
	c = getc(EQN_yyin);
	if (c == '\n')
	    read_lineno++;
	buf[0] = (c == EOF) ? (n = 0, YY_NULL) : (n = 1, c);
    }
    return n;
}

#endif

static int
equation_setup_string(s)
char *s;
{
    read_register_filename(NIL(char));
    file_mode = 0;
    string = s;
#ifdef FLEX_SCANNER
    EQN_yy_switch_to_buffer(
                 EQN_yy_create_buffer( EQN_yyin, YY_BUF_SIZE ) );
    EQN_yyrestart(EQN_yyin); 
    eqn_inp_lim = string + strlen(string);
    eqn_inp_ptr = string;
#endif
    nbuffer = 0;
}


static int
equation_setup_file(fp)
FILE *fp;
{
    EQN_yyin = fp;
#ifdef FLEX_SCANNER
    EQN_yy_switch_to_buffer(
                 EQN_yy_create_buffer( EQN_yyin, YY_BUF_SIZE ) );
    EQN_yyrestart(EQN_yyin); 
    BEGIN(INITIAL);
#endif
    file_mode = 1;
    nbuffer = 0;
    last_was_newline = 1;
}


static int
EQN_yyerror(errmsg)
char *errmsg;
{
    read_error(errmsg);
    longjmp(jmpbuf, 1);
}

#undef EQN_yywrap
static int
EQN_yywrap()
{
    return 1;
}


network_t *
read_eqn(fp)
FILE *fp;
{
    error_init();

    if (setjmp(jmpbuf)) {
	/* syntax error -- return from EQN_yyerror() */
	LS_ASSERT(lsDestroy(po_list, (void (*)()) 0));
	network_free(global_network);
	return 0;

    } else {
	global_network = network_alloc();
	read_filename_to_netname(global_network, read_filename);
	po_list = lsCreate();
	equation_setup_file(fp);
	(void) EQN_yyparse();

	if (!read_check_io_list(global_network, po_list)) {
	    network_free(global_network);
	    return 0;
	}
	read_hack_outputs(global_network, po_list);
	LS_ASSERT(lsDestroy(po_list, (void (*)()) 0));

	if (! network_is_acyclic(global_network)) {
	    network_free(global_network);
	    return 0;
	}

	return global_network;
    }
}

network_t *
read_eqn_string(s)
char *s;
{
    error_init();

    if (setjmp(jmpbuf)) {
	/* syntax error -- return from EQN_yyerror() */
	LS_ASSERT(lsDestroy(po_list, (void (*)()) 0));
	network_free(global_network);
	return 0;

    } else {
	global_network = network_alloc();
	po_list = lsCreate();
	equation_setup_string(s);
	(void) EQN_yyparse();

	if (!read_check_io_list(global_network, po_list)) {
	    network_free(global_network);
	    return 0;
	}
	read_hack_outputs(global_network, po_list);
	LS_ASSERT(lsDestroy(po_list, (void (*)()) 0));

	if (! network_is_acyclic(global_network)) {
	    network_free(global_network);
	    return 0;
	}

	return global_network;
    }
}
static YYCONST EQN_yytabelem EQN_yyexca[] ={
-1, 0,
	272, 7,
	-2, 1,
-1, 1,
	0, -1,
	-2, 0,
-1, 2,
	272, 7,
	-2, 2,
	};
# define YYNPROD 39
# define YYLAST 153
static YYCONST EQN_yytabelem EQN_yyact[]={

    53,    54,    41,    42,    12,    43,    56,    57,    40,    58,
    63,    12,    16,    52,    38,    37,    68,     9,    10,    11,
     6,     7,    12,    33,    13,     6,     7,    12,    33,    20,
    19,    22,    23,    18,    51,    29,    28,    21,    24,    49,
    17,    27,    26,    18,    53,    54,    41,    42,    12,    43,
    56,    57,    40,    58,     6,     7,    12,     5,    12,    25,
    47,    46,     9,    10,    11,    54,    41,    42,    12,    43,
    56,    57,    40,    58,    56,    57,    40,    58,    41,    42,
    12,    43,    40,    58,    40,    20,    19,    22,    23,     6,
     7,    12,    33,    21,    55,    44,     8,    31,    15,     2,
     4,    32,     3,    14,     1,     0,     0,     0,     0,     0,
     0,     0,     0,     0,    30,    34,    34,    34,    35,    39,
     0,    45,     0,    36,     0,     0,    34,    34,     0,     0,
     0,    34,    48,    50,     0,    59,     0,    50,    60,     0,
     0,     0,    61,    62,     0,     0,     0,     0,    64,    65,
     0,    66,    67 };
static YYCONST EQN_yytabelem EQN_yypact[]={

  -205,-10000000,  -250,-10000000,  -232,  -228,-10000000,-10000000,  -233,  -212,
  -229,  -230,-10000000,  -236,  -237,  -233,  -238,-10000000,  -203,  -170,
  -170,  -170,  -258,  -259,  -181,  -203,-10000000,-10000000,-10000000,-10000000,
  -170,  -234,-10000000,  -172,-10000000,  -239,  -260,-10000000,-10000000,  -213,
  -181,-10000000,-10000000,  -181,-10000000,-10000000,  -203,  -203,  -263,-10000000,
-10000000,-10000000,-10000000,  -181,  -181,  -189,  -181,  -181,-10000000,-10000000,
  -257,-10000000,-10000000,-10000000,  -193,  -189,  -183,  -183,-10000000 };
static YYCONST EQN_yytabelem EQN_yypgo[]={

     0,    95,    94,   101,    97,   104,    99,   100,    61,    60 };
static YYCONST EQN_yytabelem EQN_yyr1[]={

     0,     5,     5,     5,     5,     6,     6,     7,     7,     7,
     7,     7,     7,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     3,     3,     3,     3,     3,     3,
     3,     3,     4,     4,     1,     8,     8,     9,     9 };
static YYCONST EQN_yytabelem EQN_yyr2[]={

     0,     0,     2,     6,     3,     4,     6,     0,     7,     6,
     6,     7,    11,     7,     7,     5,     7,     7,     5,     5,
     3,     3,     7,     3,     9,     9,     9,     3,     3,     7,
     7,     3,     3,     5,     3,     0,     5,     0,     5 };
static YYCONST EQN_yytabelem EQN_yychk[]={

-10000000,    -5,    -6,    -3,    -7,   262,   259,   260,    -1,   267,
   268,   269,   261,   274,    -7,    -1,   262,   272,   271,   258,
   257,   265,   259,   260,   271,   271,   271,   271,   272,   272,
    -1,    -4,    -3,   262,    -1,    -4,    -3,   273,   273,    -2,
   265,   259,   260,   262,    -1,    -1,    -8,    -9,    -3,   273,
    -3,   273,   273,   257,   258,    -2,   263,   264,   266,    -2,
    -2,    -1,    -1,   273,    -2,    -2,    -2,    -2,   273 };
static YYCONST EQN_yytabelem EQN_yydef[]={

    -2,    -2,    -2,     4,     0,     0,    27,    28,    31,     0,
     0,     0,    34,     0,     0,     0,     0,     5,     0,     0,
     0,     0,     0,     0,     0,     0,    35,    37,     3,     6,
     0,     0,    32,     0,    31,     0,     0,    29,    30,    11,
     0,    20,    21,     0,    23,     8,     9,    10,     0,    24,
    33,    25,    26,     0,     0,    15,     0,     0,    18,    19,
     0,    36,    38,    12,    13,    14,    16,    17,    22 };
typedef struct
#ifdef __cplusplus
	EQN_yytoktype
#endif
{ char *t_name; int t_val; } EQN_yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#if YYDEBUG

EQN_yytoktype EQN_yytoks[] =
{
	"OPR_OR",	257,
	"OPR_AND",	258,
	"CONST_ZERO",	259,
	"CONST_ONE",	260,
	"IDENTIFIER",	261,
	"LPAREN",	262,
	"OPR_XOR",	263,
	"OPR_XNOR",	264,
	"OPR_NOT",	265,
	"OPR_NOT_POST",	266,
	"NAME",	267,
	"INORDER",	268,
	"OUTORDER",	269,
	"PASS",	270,
	"ASSIGN",	271,
	"SEMI",	272,
	"RPAREN",	273,
	"END",	274,
	"-unknown-",	-1	/* ends search */
};

char * EQN_yyreds[] =
{
	"-no such reduction-",
	"program : /* empty */",
	"program : prog",
	"program : prog END SEMI",
	"program : sexpr",
	"prog : stat SEMI",
	"prog : prog stat SEMI",
	"stat : /* empty */",
	"stat : NAME ASSIGN identifier",
	"stat : INORDER ASSIGN input_list",
	"stat : OUTORDER ASSIGN output_list",
	"stat : identifier ASSIGN expr",
	"stat : LPAREN ASSIGN identifier sexpr RPAREN",
	"expr : expr OPR_OR expr",
	"expr : expr OPR_AND expr",
	"expr : expr expr",
	"expr : expr OPR_XOR expr",
	"expr : expr OPR_XNOR expr",
	"expr : expr OPR_NOT_POST",
	"expr : OPR_NOT expr",
	"expr : CONST_ZERO",
	"expr : CONST_ONE",
	"expr : LPAREN expr RPAREN",
	"expr : identifier",
	"sexpr : LPAREN OPR_AND sexpr_list RPAREN",
	"sexpr : LPAREN OPR_OR sexpr_list RPAREN",
	"sexpr : LPAREN OPR_NOT sexpr RPAREN",
	"sexpr : CONST_ZERO",
	"sexpr : CONST_ONE",
	"sexpr : LPAREN CONST_ZERO RPAREN",
	"sexpr : LPAREN CONST_ONE RPAREN",
	"sexpr : identifier",
	"sexpr_list : sexpr",
	"sexpr_list : sexpr_list sexpr",
	"identifier : IDENTIFIER",
	"input_list : /* empty */",
	"input_list : input_list identifier",
	"output_list : /* empty */",
	"output_list : output_list identifier",
};
#endif /* YYDEBUG */
# line	1 "/usr/ccs/bin/yaccpar"
/*
 * Copyright (c) 1993 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)yaccpar	6.16	99/01/20 SMI"

/*
** Skeleton parser driver for yacc output
*/

/*
** yacc user known macros and defines
*/
#define YYERROR		goto EQN_yyerrlab
#define YYACCEPT	return(0)
#define YYABORT		return(1)
#define YYBACKUP( newtoken, newvalue )\
{\
	if ( EQN_yychar >= 0 || ( EQN_yyr2[ EQN_yytmp ] >> 1 ) != 1 )\
	{\
		EQN_yyerror( "syntax error - cannot backup" );\
		goto EQN_yyerrlab;\
	}\
	EQN_yychar = newtoken;\
	EQN_yystate = *EQN_yyps;\
	EQN_yylval = newvalue;\
	goto EQN_yynewstate;\
}
#define YYRECOVERING()	(!!EQN_yyerrflag)
#define YYNEW(type)	malloc(sizeof(type) * EQN_yynewmax)
#define YYCOPY(to, from, type) \
	(type *) memcpy(to, (char *) from, EQN_yymaxdepth * sizeof (type))
#define YYENLARGE( from, type) \
	(type *) realloc((char *) from, EQN_yynewmax * sizeof(type))
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
int EQN_yydebug;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
#define YYFLAG		(-10000000)

/*
** global variables used by the parser
*/
YYSTYPE *EQN_yypv;			/* top of value stack */
int *EQN_yyps;			/* top of state stack */

int EQN_yystate;			/* current state */
int EQN_yytmp;			/* extra var (lasts between blocks) */

int EQN_yynerrs;			/* number of errors */
int EQN_yyerrflag;			/* error recovery flag */
int EQN_yychar;			/* current input token number */



#ifdef YYNMBCHARS
#define YYLEX()		EQN_yycvtok(EQN_yylex())
/*
** EQN_yycvtok - return a token if i is a wchar_t value that exceeds 255.
**	If i<255, i itself is the token.  If i>255 but the neither 
**	of the 30th or 31st bit is on, i is already a token.
*/
#if defined(__STDC__) || defined(__cplusplus)
int EQN_yycvtok(int i)
#else
int EQN_yycvtok(i) int i;
#endif
{
	int first = 0;
	int last = YYNMBCHARS - 1;
	int mid;
	wchar_t j;

	if(i&0x60000000){/*Must convert to a token. */
		if( EQN_yymbchars[last].character < i ){
			return i;/*Giving up*/
		}
		while ((last>=first)&&(first>=0)) {/*Binary search loop*/
			mid = (first+last)/2;
			j = EQN_yymbchars[mid].character;
			if( j==i ){/*Found*/ 
				return EQN_yymbchars[mid].tvalue;
			}else if( j<i ){
				first = mid + 1;
			}else{
				last = mid -1;
			}
		}
		/*No entry in the table.*/
		return i;/* Giving up.*/
	}else{/* i is already a token. */
		return i;
	}
}
#else/*!YYNMBCHARS*/
#define YYLEX()		EQN_yylex()
#endif/*!YYNMBCHARS*/

/*
** EQN_yyparse - return 0 if worked, 1 if syntax error not recovered from
*/
#if defined(__STDC__) || defined(__cplusplus)
int EQN_yyparse(void)
#else
int EQN_yyparse()
#endif
{
	register YYSTYPE *EQN_yypvt = 0;	/* top of value stack for $vars */

#if defined(__cplusplus) || defined(lint)
/*
	hacks to please C++ and lint - goto's inside
	switch should never be executed
*/
	static int __yaccpar_lint_hack__ = 0;
	switch (__yaccpar_lint_hack__)
	{
		case 1: goto EQN_yyerrlab;
		case 2: goto EQN_yynewstate;
	}
#endif

	/*
	** Initialize externals - EQN_yyparse may be called more than once
	*/
	EQN_yypv = &EQN_yyv[-1];
	EQN_yyps = &EQN_yys[-1];
	EQN_yystate = 0;
	EQN_yytmp = 0;
	EQN_yynerrs = 0;
	EQN_yyerrflag = 0;
	EQN_yychar = -1;

#if YYMAXDEPTH <= 0
	if (EQN_yymaxdepth <= 0)
	{
		if ((EQN_yymaxdepth = YYEXPAND(0)) <= 0)
		{
			EQN_yyerror("yacc initialization error");
			YYABORT;
		}
	}
#endif

	{
		register YYSTYPE *EQN_yy_pv;	/* top of value stack */
		register int *EQN_yy_ps;		/* top of state stack */
		register int EQN_yy_state;		/* current state */
		register int  EQN_yy_n;		/* internal state number info */
	goto EQN_yystack;	/* moved from 6 lines above to here to please C++ */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	EQN_yynewstate:
		EQN_yy_pv = EQN_yypv;
		EQN_yy_ps = EQN_yyps;
		EQN_yy_state = EQN_yystate;
		goto EQN_yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	EQN_yystack:
		EQN_yy_pv = EQN_yypv;
		EQN_yy_ps = EQN_yyps;
		EQN_yy_state = EQN_yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	EQN_yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#if YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( EQN_yydebug )
		{
			register int EQN_yy_i;

			printf( "State %d, token ", EQN_yy_state );
			if ( EQN_yychar == 0 )
				printf( "end-of-file\n" );
			else if ( EQN_yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( EQN_yy_i = 0; EQN_yytoks[EQN_yy_i].t_val >= 0;
					EQN_yy_i++ )
				{
					if ( EQN_yytoks[EQN_yy_i].t_val == EQN_yychar )
						break;
				}
				printf( "%s\n", EQN_yytoks[EQN_yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++EQN_yy_ps >= &EQN_yys[ EQN_yymaxdepth ] )	/* room on stack? */
		{
			/*
			** reallocate and recover.  Note that pointers
			** have to be reset, or bad things will happen
			*/
			long EQN_yyps_index = (EQN_yy_ps - EQN_yys);
			long EQN_yypv_index = (EQN_yy_pv - EQN_yyv);
			long EQN_yypvt_index = (EQN_yypvt - EQN_yyv);
			int EQN_yynewmax;
#ifdef YYEXPAND
			EQN_yynewmax = YYEXPAND(EQN_yymaxdepth);
#else
			EQN_yynewmax = 2 * EQN_yymaxdepth;	/* double table size */
			if (EQN_yymaxdepth == YYMAXDEPTH)	/* first time growth */
			{
				char *newEQN_yys = (char *)YYNEW(int);
				char *newEQN_yyv = (char *)YYNEW(YYSTYPE);
				if (newEQN_yys != 0 && newEQN_yyv != 0)
				{
					EQN_yys = YYCOPY(newEQN_yys, EQN_yys, int);
					EQN_yyv = YYCOPY(newEQN_yyv, EQN_yyv, YYSTYPE);
				}
				else
					EQN_yynewmax = 0;	/* failed */
			}
			else				/* not first time */
			{
				EQN_yys = YYENLARGE(EQN_yys, int);
				EQN_yyv = YYENLARGE(EQN_yyv, YYSTYPE);
				if (EQN_yys == 0 || EQN_yyv == 0)
					EQN_yynewmax = 0;	/* failed */
			}
#endif
			if (EQN_yynewmax <= EQN_yymaxdepth)	/* tables not expanded */
			{
				EQN_yyerror( "yacc stack overflow" );
				YYABORT;
			}
			EQN_yymaxdepth = EQN_yynewmax;

			EQN_yy_ps = EQN_yys + EQN_yyps_index;
			EQN_yy_pv = EQN_yyv + EQN_yypv_index;
			EQN_yypvt = EQN_yyv + EQN_yypvt_index;
		}
		*EQN_yy_ps = EQN_yy_state;
		*++EQN_yy_pv = EQN_yyval;

		/*
		** we have a new state - find out what to do
		*/
	EQN_yy_newstate:
		if ( ( EQN_yy_n = EQN_yypact[ EQN_yy_state ] ) <= YYFLAG )
			goto EQN_yydefault;		/* simple state */
#if YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		EQN_yytmp = EQN_yychar < 0;
#endif
		if ( ( EQN_yychar < 0 ) && ( ( EQN_yychar = YYLEX() ) < 0 ) )
			EQN_yychar = 0;		/* reached EOF */
#if YYDEBUG
		if ( EQN_yydebug && EQN_yytmp )
		{
			register int EQN_yy_i;

			printf( "Received token " );
			if ( EQN_yychar == 0 )
				printf( "end-of-file\n" );
			else if ( EQN_yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( EQN_yy_i = 0; EQN_yytoks[EQN_yy_i].t_val >= 0;
					EQN_yy_i++ )
				{
					if ( EQN_yytoks[EQN_yy_i].t_val == EQN_yychar )
						break;
				}
				printf( "%s\n", EQN_yytoks[EQN_yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( EQN_yy_n += EQN_yychar ) < 0 ) || ( EQN_yy_n >= YYLAST ) )
			goto EQN_yydefault;
		if ( EQN_yychk[ EQN_yy_n = EQN_yyact[ EQN_yy_n ] ] == EQN_yychar )	/*valid shift*/
		{
			EQN_yychar = -1;
			EQN_yyval = EQN_yylval;
			EQN_yy_state = EQN_yy_n;
			if ( EQN_yyerrflag > 0 )
				EQN_yyerrflag--;
			goto EQN_yy_stack;
		}

	EQN_yydefault:
		if ( ( EQN_yy_n = EQN_yydef[ EQN_yy_state ] ) == -2 )
		{
#if YYDEBUG
			EQN_yytmp = EQN_yychar < 0;
#endif
			if ( ( EQN_yychar < 0 ) && ( ( EQN_yychar = YYLEX() ) < 0 ) )
				EQN_yychar = 0;		/* reached EOF */
#if YYDEBUG
			if ( EQN_yydebug && EQN_yytmp )
			{
				register int EQN_yy_i;

				printf( "Received token " );
				if ( EQN_yychar == 0 )
					printf( "end-of-file\n" );
				else if ( EQN_yychar < 0 )
					printf( "-none-\n" );
				else
				{
					for ( EQN_yy_i = 0;
						EQN_yytoks[EQN_yy_i].t_val >= 0;
						EQN_yy_i++ )
					{
						if ( EQN_yytoks[EQN_yy_i].t_val
							== EQN_yychar )
						{
							break;
						}
					}
					printf( "%s\n", EQN_yytoks[EQN_yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register YYCONST int *EQN_yyxi = EQN_yyexca;

				while ( ( *EQN_yyxi != -1 ) ||
					( EQN_yyxi[1] != EQN_yy_state ) )
				{
					EQN_yyxi += 2;
				}
				while ( ( *(EQN_yyxi += 2) >= 0 ) &&
					( *EQN_yyxi != EQN_yychar ) )
					;
				if ( ( EQN_yy_n = EQN_yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( EQN_yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( EQN_yyerrflag )
			{
			case 0:		/* new error */
				EQN_yyerror( "syntax error" );
				goto skip_init;
			EQN_yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				EQN_yy_pv = EQN_yypv;
				EQN_yy_ps = EQN_yyps;
				EQN_yy_state = EQN_yystate;
			skip_init:
				EQN_yynerrs++;
				/* FALLTHRU */
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				EQN_yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( EQN_yy_ps >= EQN_yys )
				{
					EQN_yy_n = EQN_yypact[ *EQN_yy_ps ] + YYERRCODE;
					if ( EQN_yy_n >= 0 && EQN_yy_n < YYLAST &&
						EQN_yychk[EQN_yyact[EQN_yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						EQN_yy_state = EQN_yyact[ EQN_yy_n ];
						goto EQN_yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#if YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( EQN_yydebug )
						printf( _POP_, *EQN_yy_ps,
							EQN_yy_ps[-1] );
#	undef _POP_
#endif
					EQN_yy_ps--;
					EQN_yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#if YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( EQN_yydebug )
				{
					register int EQN_yy_i;

					printf( "Error recovery discards " );
					if ( EQN_yychar == 0 )
						printf( "token end-of-file\n" );
					else if ( EQN_yychar < 0 )
						printf( "token -none-\n" );
					else
					{
						for ( EQN_yy_i = 0;
							EQN_yytoks[EQN_yy_i].t_val >= 0;
							EQN_yy_i++ )
						{
							if ( EQN_yytoks[EQN_yy_i].t_val
								== EQN_yychar )
							{
								break;
							}
						}
						printf( "token %s\n",
							EQN_yytoks[EQN_yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( EQN_yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				EQN_yychar = -1;
				goto EQN_yy_newstate;
			}
		}/* end if ( EQN_yy_n == 0 ) */
		/*
		** reduction by production EQN_yy_n
		** put stack tops, etc. so things right after switch
		*/
#if YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( EQN_yydebug )
			printf( "Reduce by (%d) \"%s\"\n",
				EQN_yy_n, EQN_yyreds[ EQN_yy_n ] );
#endif
		EQN_yytmp = EQN_yy_n;			/* value to switch over */
		EQN_yypvt = EQN_yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using EQN_yy_state here as temporary
		** register variable, but why not, if it works...
		** If EQN_yyr2[ EQN_yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto EQN_yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int EQN_yy_len = EQN_yyr2[ EQN_yy_n ];

			if ( !( EQN_yy_len & 01 ) )
			{
				EQN_yy_len >>= 1;
				EQN_yyval = ( EQN_yy_pv -= EQN_yy_len )[1];	/* $$ = $1 */
				EQN_yy_state = EQN_yypgo[ EQN_yy_n = EQN_yyr1[ EQN_yy_n ] ] +
					*( EQN_yy_ps -= EQN_yy_len ) + 1;
				if ( EQN_yy_state >= YYLAST ||
					EQN_yychk[ EQN_yy_state =
					EQN_yyact[ EQN_yy_state ] ] != -EQN_yy_n )
				{
					EQN_yy_state = EQN_yyact[ EQN_yypgo[ EQN_yy_n ] ];
				}
				goto EQN_yy_stack;
			}
			EQN_yy_len >>= 1;
			EQN_yyval = ( EQN_yy_pv -= EQN_yy_len )[1];	/* $$ = $1 */
			EQN_yy_state = EQN_yypgo[ EQN_yy_n = EQN_yyr1[ EQN_yy_n ] ] +
				*( EQN_yy_ps -= EQN_yy_len ) + 1;
			if ( EQN_yy_state >= YYLAST ||
				EQN_yychk[ EQN_yy_state = EQN_yyact[ EQN_yy_state ] ] != -EQN_yy_n )
			{
				EQN_yy_state = EQN_yyact[ EQN_yypgo[ EQN_yy_n ] ];
			}
		}
					/* save until reenter driver code */
		EQN_yystate = EQN_yy_state;
		EQN_yyps = EQN_yy_ps;
		EQN_yypv = EQN_yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( EQN_yytmp )
	{
		
case 4:
# line 100 "read_eqn.y"
{ do_assign(util_strsav("SILLY"), EQN_yypvt[-0].node); } break;
case 8:
# line 109 "read_eqn.y"
{
			network_set_name(global_network, EQN_yypvt[-0].strval);
			FREE(EQN_yypvt[-0].strval);
		    } break;
case 11:
# line 119 "read_eqn.y"
{ do_assign(EQN_yypvt[-2].strval, EQN_yypvt[-0].node); } break;
case 12:
# line 122 "read_eqn.y"
{ do_assign(EQN_yypvt[-2].strval, EQN_yypvt[-1].node); } break;
case 13:
# line 127 "read_eqn.y"
{ EQN_yyval.node = node_or(EQN_yypvt[-2].node, EQN_yypvt[-0].node); node_free(EQN_yypvt[-2].node); node_free(EQN_yypvt[-0].node);} break;
case 14:
# line 130 "read_eqn.y"
{ EQN_yyval.node = node_and(EQN_yypvt[-2].node, EQN_yypvt[-0].node); node_free(EQN_yypvt[-2].node); node_free(EQN_yypvt[-0].node);} break;
case 15:
# line 133 "read_eqn.y"
{ EQN_yyval.node = node_and(EQN_yypvt[-1].node, EQN_yypvt[-0].node); node_free(EQN_yypvt[-1].node); node_free(EQN_yypvt[-0].node);} break;
case 16:
# line 136 "read_eqn.y"
{ EQN_yyval.node = node_xor(EQN_yypvt[-2].node, EQN_yypvt[-0].node); node_free(EQN_yypvt[-2].node); node_free(EQN_yypvt[-0].node);} break;
case 17:
# line 139 "read_eqn.y"
{ EQN_yyval.node = node_xnor(EQN_yypvt[-2].node, EQN_yypvt[-0].node); node_free(EQN_yypvt[-2].node); node_free(EQN_yypvt[-0].node);} break;
case 18:
# line 142 "read_eqn.y"
{ EQN_yyval.node = node_not(EQN_yypvt[-1].node); node_free(EQN_yypvt[-1].node); } break;
case 19:
# line 145 "read_eqn.y"
{ EQN_yyval.node = node_not(EQN_yypvt[-0].node); node_free(EQN_yypvt[-0].node); } break;
case 20:
# line 148 "read_eqn.y"
{ EQN_yyval.node = node_constant(0); } break;
case 21:
# line 151 "read_eqn.y"
{ EQN_yyval.node = node_constant(1); } break;
case 22:
# line 154 "read_eqn.y"
{ EQN_yyval.node = EQN_yypvt[-1].node; } break;
case 23:
# line 157 "read_eqn.y"
{
			node_t *node;
			node = read_find_or_create_node(global_network, EQN_yypvt[-0].strval);
			EQN_yyval.node = node_literal(node, 1);
			FREE(EQN_yypvt[-0].strval);
		    } break;
case 24:
# line 167 "read_eqn.y"
{ EQN_yyval.node = do_sexpr_list(EQN_yypvt[-1].array, node_and); } break;
case 25:
# line 170 "read_eqn.y"
{ EQN_yyval.node = do_sexpr_list(EQN_yypvt[-1].array, node_or); } break;
case 26:
# line 173 "read_eqn.y"
{ EQN_yyval.node = node_not(EQN_yypvt[-1].node); node_free(EQN_yypvt[-1].node); } break;
case 27:
# line 176 "read_eqn.y"
{ EQN_yyval.node = node_constant(0); } break;
case 28:
# line 179 "read_eqn.y"
{ EQN_yyval.node = node_constant(1); } break;
case 29:
# line 182 "read_eqn.y"
{ EQN_yyval.node = node_constant(0); } break;
case 30:
# line 185 "read_eqn.y"
{ EQN_yyval.node = node_constant(1); } break;
case 31:
# line 188 "read_eqn.y"
{
			node_t *node;
			node = read_find_or_create_node(global_network, EQN_yypvt[-0].strval);
			EQN_yyval.node = node_literal(node, 1);
			FREE(EQN_yypvt[-0].strval);
		    } break;
case 32:
# line 198 "read_eqn.y"
{
			EQN_yyval.array = array_alloc(node_t *, 10);
			array_insert_last(node_t *, EQN_yyval.array, EQN_yypvt[-0].node);
		    } break;
case 33:
# line 203 "read_eqn.y"
{
			array_insert_last(node_t *, EQN_yypvt[-1].array, EQN_yypvt[-0].node);
		    } break;
case 34:
# line 208 "read_eqn.y"
{ EQN_yyval.strval = util_strsav(EQN_yytext); } break;
case 36:
# line 214 "read_eqn.y"
{
			node_t *node;
			char errmsg[1024];
			node = read_find_or_create_node(global_network, EQN_yypvt[-0].strval);
			if (node_function(node) != NODE_UNDEFINED) {
			    (void) sprintf(errmsg, 
				"Attempt to redefine '%s'\n", EQN_yypvt[-0].strval);
			    EQN_yyerror(errmsg);   /* never returns */
			}
			network_change_node_type(global_network, 
							node, PRIMARY_INPUT);
			FREE(EQN_yypvt[-0].strval);
		    } break;
case 38:
# line 232 "read_eqn.y"
{
			node_t *node;
			node = read_find_or_create_node(global_network, EQN_yypvt[-0].strval);
			LS_ASSERT(lsNewEnd(po_list, (lsGeneric) node, LS_NH));
			FREE(EQN_yypvt[-0].strval);
		    } break;
# line	531 "/usr/ccs/bin/yaccpar"
	}
	goto EQN_yystack;		/* reset registers in driver code */
}

